Package: PKNCA
Type: Package
Title: Perform Pharmacokinetic Non-Compartmental Analysis
Version: 0.12.1
Authors@R: c(
    person("Bill", "Denney", email="wdenney@humanpredictions.com", role=c("aut", "cre"), comment=c(ORCID="0000-0002-5759-428X")),
    person("Clare", "Buckeridge", email="clare.buckeridge@pfizer.com", role="aut"),
    person("Gerardo Jose", "Rodriguez", email="gerardo.jrac@gmail.com", role="aut", comment=c(ORCID="0000-0003-1413-0060")),
    person("Sridhar", "Duvvuri", role="ctb"))
Imports: checkmate, dplyr (>= 0.5.0), digest, nlme, purrr, rlang,
        stats, tidyr, tibble, utils, lifecycle
Suggests: covr, cowplot, ggplot2, knitr, labeling, pander, pmxTools,
        rmarkdown, spelling, testthat (>= 3.0.0), units, withr
Description: Compute standard Non-Compartmental Analysis (NCA) parameters for
    typical pharmacokinetic analyses and summarize them.
License: AGPL-3
URL: https://humanpred.github.io/pknca/,
        https://github.com/humanpred/pknca,
        http://humanpred.github.io/pknca/
BugReports: https://github.com/humanpred/pknca/issues
NeedsCompilation: no
VignetteBuilder: knitr
RoxygenNote: 7.3.2
Config/testthat/edition: 3
Encoding: UTF-8
Language: en-US
Packaged: 2025-08-19 00:38:28 UTC; bill
Author: Bill Denney [aut, cre] (ORCID: <https://orcid.org/0000-0002-5759-428X>),
  Clare Buckeridge [aut],
  Gerardo Jose Rodriguez [aut] (ORCID:
    <https://orcid.org/0000-0003-1413-0060>),
  Sridhar Duvvuri [ctb]
Maintainer: Bill Denney <wdenney@humanpredictions.com>
Repository: CRAN
Date/Publication: 2025-08-19 04:40:02 UTC
Built: R 4.6.0; ; 2026-04-26 16:06:45 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(PKNCAconc,data.frame)
S3method(PKNCAconc,default)
S3method(PKNCAconc,tbl_df)
S3method(PKNCAdata,PKNCAconc)
S3method(PKNCAdata,PKNCAdose)
S3method(PKNCAdata,default)
S3method(PKNCAdose,data.frame)
S3method(PKNCAdose,default)
S3method(PKNCAdose,tbl_df)
S3method(as.data.frame,PKNCAconc)
S3method(as.data.frame,PKNCAdose)
S3method(as.data.frame,PKNCAresults)
S3method(as_PKNCAconc,PKNCAconc)
S3method(as_PKNCAconc,PKNCAdata)
S3method(as_PKNCAconc,PKNCAresults)
S3method(as_PKNCAdata,PKNCAdata)
S3method(as_PKNCAdata,PKNCAresults)
S3method(as_PKNCAdose,PKNCAdata)
S3method(as_PKNCAdose,PKNCAdose)
S3method(as_PKNCAdose,PKNCAresults)
S3method(as_PKNCAresults,PKNCAdata)
S3method(as_PKNCAresults,PKNCAresults)
S3method(dplyr::group_vars,PKNCAconc)
S3method(dplyr::group_vars,PKNCAdata)
S3method(dplyr::group_vars,PKNCAdose)
S3method(dplyr::group_vars,PKNCAresults)
S3method(exclude,default)
S3method(filter,PKNCAconc)
S3method(filter,PKNCAdose)
S3method(filter,PKNCAresults)
S3method(formula,PKNCAconc)
S3method(formula,PKNCAdose)
S3method(full_join,PKNCAconc)
S3method(full_join,PKNCAdose)
S3method(full_join,PKNCAresults)
S3method(getDataName,PKNCAconc)
S3method(getDataName,PKNCAdose)
S3method(getDataName,PKNCAresults)
S3method(getDepVar,PKNCAconc)
S3method(getDepVar,PKNCAdose)
S3method(getGroups,PKNCAconc)
S3method(getGroups,PKNCAdata)
S3method(getGroups,PKNCAdose)
S3method(getGroups,PKNCAresults)
S3method(getIndepVar,PKNCAconc)
S3method(getIndepVar,PKNCAdose)
S3method(group_by,PKNCAconc)
S3method(group_by,PKNCAdose)
S3method(group_by,PKNCAresults)
S3method(inner_join,PKNCAconc)
S3method(inner_join,PKNCAdose)
S3method(inner_join,PKNCAresults)
S3method(is_sparse_pk,PKNCAconc)
S3method(is_sparse_pk,PKNCAdata)
S3method(is_sparse_pk,PKNCAresults)
S3method(left_join,PKNCAconc)
S3method(left_join,PKNCAdose)
S3method(left_join,PKNCAresults)
S3method(model.frame,PKNCAconc)
S3method(model.frame,PKNCAdose)
S3method(mutate,PKNCAconc)
S3method(mutate,PKNCAdose)
S3method(mutate,PKNCAresults)
S3method(print,PKNCAconc)
S3method(print,PKNCAdata)
S3method(print,PKNCAdose)
S3method(print,provenance)
S3method(print,summary_PKNCAresults)
S3method(right_join,PKNCAconc)
S3method(right_join,PKNCAdose)
S3method(right_join,PKNCAresults)
S3method(setDuration,PKNCAconc)
S3method(setDuration,PKNCAdose)
S3method(setRoute,PKNCAdose)
S3method(signifString,data.frame)
S3method(signifString,default)
S3method(summary,PKNCAconc)
S3method(summary,PKNCAdata)
S3method(summary,PKNCAdose)
S3method(summary,PKNCAresults)
S3method(superposition,PKNCAconc)
S3method(superposition,numeric)
S3method(time_calc,POSIXt)
S3method(time_calc,difftime)
S3method(time_calc,numeric)
S3method(ungroup,PKNCAconc)
S3method(ungroup,PKNCAdose)
S3method(ungroup,PKNCAresults)
export(PKNCA.choose.option)
export(PKNCA.options)
export(PKNCA.options.describe)
export(PKNCA.set.summary)
export(PKNCA_impute_method_start_cmin)
export(PKNCA_impute_method_start_conc0)
export(PKNCA_impute_method_start_predose)
export(PKNCAconc)
export(PKNCAdata)
export(PKNCAdose)
export(PKNCAresults)
export(add.interval.col)
export(addProvenance)
export(adj.r.squared)
export(as_PKNCAconc)
export(as_PKNCAdata)
export(as_PKNCAdose)
export(as_PKNCAresults)
export(as_sparse_pk)
export(assert_intervals)
export(business.cv)
export(business.geocv)
export(business.geomean)
export(business.max)
export(business.mean)
export(business.median)
export(business.min)
export(business.range)
export(business.sd)
export(check.conc.time)
export(check.conversion)
export(check.interval.specification)
export(checkProvenance)
export(choose.auc.intervals)
export(clean.conc.blq)
export(clean.conc.na)
export(cov_holder)
export(exclude)
export(exclude_nca_by_param)
export(exclude_nca_count_conc_measured)
export(exclude_nca_max.aucinf.pext)
export(exclude_nca_min.hl.adj.r.squared)
export(exclude_nca_min.hl.r.squared)
export(exclude_nca_span.ratio)
export(exclude_nca_tmax_0)
export(exclude_nca_tmax_early)
export(extrapolate.conc)
export(filter)
export(find.tau)
export(full_join)
export(geocv)
export(geomean)
export(geosd)
export(get.best.model)
export(get.interval.cols)
export(get.parameter.deps)
export(getDepVar)
export(getGroups)
export(getIndepVar)
export(get_halflife_points)
export(group_by)
export(inner_join)
export(interp.extrap.conc)
export(interp.extrap.conc.dose)
export(interpolate.conc)
export(is_sparse_pk)
export(left_join)
export(mutate)
export(pk.business)
export(pk.calc.ae)
export(pk.calc.auc)
export(pk.calc.auc.all)
export(pk.calc.auc.inf)
export(pk.calc.auc.inf.obs)
export(pk.calc.auc.inf.pred)
export(pk.calc.auc.last)
export(pk.calc.aucabove)
export(pk.calc.aucint)
export(pk.calc.aucint.all)
export(pk.calc.aucint.inf.obs)
export(pk.calc.aucint.inf.pred)
export(pk.calc.aucint.last)
export(pk.calc.auciv)
export(pk.calc.auciv_pbext)
export(pk.calc.aucpext)
export(pk.calc.aumc)
export(pk.calc.aumc.all)
export(pk.calc.aumc.inf)
export(pk.calc.aumc.inf.obs)
export(pk.calc.aumc.inf.pred)
export(pk.calc.aumc.last)
export(pk.calc.auxc)
export(pk.calc.c0)
export(pk.calc.cav)
export(pk.calc.ceoi)
export(pk.calc.cl)
export(pk.calc.clast.obs)
export(pk.calc.clr)
export(pk.calc.cmax)
export(pk.calc.cmin)
export(pk.calc.count_conc)
export(pk.calc.count_conc_measured)
export(pk.calc.cstart)
export(pk.calc.ctrough)
export(pk.calc.deg.fluc)
export(pk.calc.dn)
export(pk.calc.f)
export(pk.calc.fe)
export(pk.calc.half.life)
export(pk.calc.kel)
export(pk.calc.mrt)
export(pk.calc.mrt.iv)
export(pk.calc.mrt.md)
export(pk.calc.ptr)
export(pk.calc.sparse_auc)
export(pk.calc.sparse_auclast)
export(pk.calc.swing)
export(pk.calc.tfirst)
export(pk.calc.thalf.eff)
export(pk.calc.time_above)
export(pk.calc.tlag)
export(pk.calc.tlast)
export(pk.calc.tmax)
export(pk.calc.totdose)
export(pk.calc.vss)
export(pk.calc.vz)
export(pk.nca)
export(pk.nca.interval)
export(pk.tss)
export(pk.tss.monoexponential)
export(pk.tss.stepwise.linear)
export(pknca_units_table)
export(right_join)
export(roundString)
export(roundingSummarize)
export(setDuration)
export(setRoute)
export(set_intervals)
export(signifString)
export(sparse_auc_weight_linear)
export(sparse_mean)
export(superposition)
export(time_calc)
export(ungroup)
export(var_sparse_auc)
importFrom(dplyr,"%>%")
importFrom(dplyr,filter)
importFrom(dplyr,full_join)
importFrom(dplyr,group_by)
importFrom(dplyr,inner_join)
importFrom(dplyr,left_join)
importFrom(dplyr,mutate)
importFrom(dplyr,right_join)
importFrom(dplyr,ungroup)
importFrom(lifecycle,deprecated)
importFrom(nlme,getGroups)
importFrom(rlang,.data)
importFrom(stats,formula)
importFrom(stats,model.frame)

<!-- badges: start -->
[![CRAN status](https://www.r-pkg.org/badges/version/PKNCA)](https://CRAN.R-project.org/package=PKNCA)
[![R-CMD-check](https://github.com/humanpred/pknca/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/humanpred/pknca/actions/workflows/R-CMD-check.yaml)
[![Codecov test coverage](https://codecov.io/gh/humanpred/pknca/graph/badge.svg)](https://app.codecov.io/gh/humanpred/pknca)
[![CodeFactor](https://www.codefactor.io/repository/github/humanpred/pknca/badge?style=plastic)](https://www.codefactor.io/repository/github/humanpred/pknca)
[![CII Best Practices](https://bestpractices.coreinfrastructure.org/projects/2054/badge)](https://bestpractices.coreinfrastructure.org/projects/2054)
<!-- badges: end -->

The PKNCA R Package <img src="man/figures/pknca-hex.png" align="right" />
=====

The PKNCA R package is designed to perform all noncompartmental
analysis (NCA) calculations for pharmacokinetic (PK) data.  The
package is broadly separated into two parts (calculation and summary)
with some additional housekeeping functions.

The primary and secondary goals of the PKNCA package are to 1) only
give correct answers to the specific questions being asked and 2)
automate as much as possible to simplify the task of the analyst. When
automation would leave ambiguity or make a choice that the analyst may
have an alternate preference for, it is either not used, is possible
to override or raises an error message.

Note that backward compatibility will not be guaranteed until version
1.0.  Argument and function changes will continue until then.  These
will be especially noticeable around the inclusion of IV NCA parameters and additional specifications of the drug, which include dose amount and route of administration.

# Citation

Citation information for the PKNCA package is available with a call to
`citation(package="PKNCA")`.  The preferred citation until publication
of version 1.0 is below:

Denney W, Duvvuri S and Buckeridge C (2015). "Simple, Automatic
Noncompartmental Analysis: The PKNCA R Package." _Journal of
Pharmacokinetics and Pharmacodynamics_, *42*(1), pp. 11-107,S65. ISSN
1573-8744, doi: 10.1007/s10928-015-9432-2, <URL:
https://humanpred.github.io/pknca/ >.

# Installation

## From CRAN

The current stable version of PKNCA is available on CRAN.  You can
install it and its dependencies using the following command:

    install.packages("PKNCA")

## From GitHub

To install the development version from GitHub, type the following commands:

    install.packages("remotes")
    remotes::install_github("humanpred/pknca")

# Calculating parameters

    # Load the package
    library(PKNCA)
    # Set the business rule options with the PKNCA.options() function
    # Load your concentration-time data
    conc_raw <- read.csv("myconc.csv", stringsAsFactors=FALSE)
    # Load your dose data
    dose_raw <- read.csv("mydose.csv", stringsAsFactors=FALSE)
    # Put your concentration data into a PKNCAconc object
    o_conc <- PKNCAconc(data=conc_raw,
                        formula=conc~time|treatment+subject/analyte)
    # Put your dose data into a PKNCAdose object
    o_dose <- PKNCAdose(data=dose_raw,
                        formula=dose~time|treatment+subject)
    # Combine the two (and automatically determine the intervals of
    # interest
    o_data <- PKNCAdata(o_conc, o_dose)
    # Compute the NCA parameters
    o_results <- pk.nca(o_data)
    # Summarize the results
    summary(o_results)

More help is available in the function help files. Be sure to look at the PKNCA.options function for choices on making PKNCA conform to your company’s business rules on calculation and summarization.

# Feature requests

Please use the github issues page
(https://github.com/humanpred/pknca/issues) to make feature requests
and bug reports.

## ----setup, echo=FALSE, include=FALSE-----------------------------------------
library(PKNCA)
library(knitr)

## ----showtheoph---------------------------------------------------------------
## It is always a good idea to look at the data
knitr::kable(head(datasets::Theoph))

## ----setupconcdose------------------------------------------------------------
## By default it is groupedData; convert it to a data frame for use
conc_obj <- PKNCAconc(as.data.frame(datasets::Theoph), conc~Time|Subject)

## Dosing data needs to only have one row per dose, so subset for
## that first.
d_dose <- unique(datasets::Theoph[datasets::Theoph$Time == 0,
                                  c("Dose", "Time", "Subject")])
knitr::kable(d_dose,
             caption="Example dosing data extracted from theophylline data set")
dose_obj <- PKNCAdose(d_dose, Dose~Time|Subject)

## ----autointervals------------------------------------------------------------
data_obj_automatic <- PKNCAdata(conc_obj, dose_obj)
knitr::kable(PKNCA.options("single.dose.aucs"))
knitr::kable(data_obj_automatic$intervals)

## ----manualintervals----------------------------------------------------------
intervals_manual <- data.frame(start=0,
                               end=Inf,
                               cmax=TRUE,
                               tmax=TRUE,
                               aucinf.obs=TRUE,
                               auclast=TRUE)
data_obj_manual <- PKNCAdata(conc_obj, dose_obj,
                             intervals=intervals_manual)
knitr::kable(data_obj_manual$intervals)

## ----calculationauto----------------------------------------------------------
results_obj_automatic <- pk.nca(data_obj_automatic)
knitr::kable(head(as.data.frame(results_obj_automatic)))

## ----calculationautoshow, eval=FALSE------------------------------------------
# summary(results_obj_automatic)

## ----calculationautoshowpretty, echo=FALSE------------------------------------
## Make a pretty table instead of the data.frame preformatted printout
knitr::kable(summary(results_obj_automatic))

## ----calculationmanual--------------------------------------------------------
results_obj_manual <- pk.nca(data_obj_manual)
knitr::kable(head(as.data.frame(results_obj_manual)))

## ----calculationmanualshow, eval=FALSE----------------------------------------
# summary(results_obj_manual)

## ----calculationmanualshowpretty, echo=FALSE----------------------------------
## Make a pretty table instead of the data.frame preformatted printout
knitr::kable(summary(results_obj_manual))

## ----superposition------------------------------------------------------------
d_conc <- PKNCAconc(as.data.frame(Theoph), conc~Time|Subject)
conc_obj_multi <-
  PKNCAconc(
    superposition(d_conc,
                  tau=168,
                  dose.times=seq(0, 144, by=24),
                  n.tau=1,
                  check.blq=FALSE),
    conc~time|Subject)
conc_obj_multi
dose_obj_multi <- PKNCAdose(expand.grid(Subject=unique(as.data.frame(conc_obj_multi)$Subject),
                                      time=seq(0, 144, by=24)),
                          ~time|Subject)
dose_obj_multi

## ----multi_auto_choose_intervals----------------------------------------------
data_obj <- PKNCAdata(conc_obj_multi, dose_obj_multi)
data_obj$intervals[,c("Subject", "start", "end")]

## ----multi_manual_choose_intervals--------------------------------------------
intervals_manual <- data.frame(start=c(0, 144),
                               end=c(24, 168),
                               cmax=TRUE,
                               auclast=TRUE)
data_obj <- PKNCAdata(conc_obj_multi, dose_obj_multi,
                      intervals=intervals_manual)
data_obj$intervals

## ----multi_calc---------------------------------------------------------------
results_obj <- pk.nca(data_obj)
print(results_obj)

summary(results_obj)

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(echo=TRUE,
                      fig.width=6,
                      fig.height=4)
library(cowplot)
library(PKNCA)
library(knitr)
library(ggplot2)
source("../tests/testthat/helper-generate_data.R")
scale_colour_discrete <- scale_colour_hue
scale_fill_discrete <- scale_fill_hue

## ----check-ggplot, include=!requireNamespace("ggplot2"), results="asis"-------
cat("ggplot2 is required for this vignette to work correctly.  Please install the ggplot2 library and retry building the vignette.")

## ----intro_interval_plot, echo=FALSE------------------------------------------
# Simulate concentration-time data and setup the PKNCAconc object
d_conc <-
  PKNCAconc(
    generate.conc(nsub=1,
                  ntreat=1,
                  time.points=c(0, 1, 2, 4, 6, 8, 12, 24, 36, 48),
                  nstudies=1,
                  nanalytes=1,
                  resid=0),
    conc~time|treatment+ID)
print(d_conc)
# Use superposition to simulate multiple doses
d_conc_multi <-
  superposition(d_conc,
                tau=168,
                dose.times=seq(0, 168-24, by=24),
                n.tau=1)

## ----intro_interval_plot-visualization, eval=requireNamespace("ggplot2")------
# Plot the concentration-time data and the interval
ggplot(d_conc_multi, aes(x=time, y=conc)) +
  geom_ribbon(data=d_conc_multi[d_conc_multi$time >= 144,],
              aes(ymax=conc, ymin=0),
              fill="skyblue") +
  geom_point() + geom_line() +
  scale_x_continuous(breaks=seq(0, 168, by=12)) +
  scale_y_continuous(limits=c(0, NA)) +
  labs(x="Time Since First Dose (hr)",
       y="Concentration\n(arbitrary units)")

## ----intro_interval_spec------------------------------------------------------
intervals_manual <- data.frame(start=144, end=168, auclast=TRUE)
knitr::kable(intervals_manual)

PKNCAdata(d_conc, intervals=intervals_manual)

## ----select_group-------------------------------------------------------------
intervals_manual <-
  data.frame(
    treatment=c("Drug 1 Single", "Drug 1 Multiple", "Drug 1 Multiple"),
    start=c(0, 0, 216),
    end=c(Inf, 24, 240),
    aucinf.obs=c(TRUE, FALSE, FALSE),
    auclast=c(FALSE, TRUE, TRUE)
  )
knitr::kable(intervals_manual)

## ----infinity_interval_plot, echo=FALSE---------------------------------------
# Simulate concentration-time data and setup the PKNCAconc object
d_conc <-
  PKNCAconc(
    generate.conc(nsub=1,
                  ntreat=1,
                  time.points=c(0, 1, 2, 4, 6, 8, 12, 24, 36, 48),
                  nstudies=1,
                  nanalytes=1,
                  resid=0),
    conc~time|treatment+ID)
print(d_conc)

## ----infinity_interval_plot-visualization, eval=requireNamespace("ggplot2")----
# Use superposition to simulate multiple doses
ggplot(as.data.frame(d_conc)[as.data.frame(d_conc)$time <= 48,], aes(x=time, y=conc)) +
  geom_ribbon(data=as.data.frame(d_conc),
              aes(ymax=conc, ymin=0),
              fill="skyblue") +
  geom_point() + geom_line() +
  scale_x_continuous(breaks=seq(0, 72, by=12)) +
  scale_y_continuous(limits=c(0, NA)) +
  labs(x="Time Since First Dose (hr)",
       y="Concentration\n(arbitrary units)")

## ----infinity_interval_spec---------------------------------------------------
intervals_manual <-
  data.frame(
    start=0,
    end=Inf,
    auclast=TRUE,
    aucinf.obs=TRUE
  )
print(intervals_manual)

my.data <- PKNCAdata(d_conc, intervals=intervals_manual)

## ----multiple_intervals_plot, echo=FALSE--------------------------------------
# Simulate concentration-time data and setup the PKNCAconc object
d_conc <-
  PKNCAconc(
    generate.conc(nsub=1,
                  ntreat=1,
                  time.points=c(0, 1, 2, 4, 6, 8, 12, 24, 36, 48),
                  nstudies=1,
                  nanalytes=1,
                  resid=0),
    conc~time|treatment+ID)
print(d_conc)
# Use superposition to simulate multiple doses
d_conc_multi <-
  superposition(d_conc,
                tau=168,
                dose.times=seq(0, 168-24, by=24),
                n.tau=1)

## ----multiple_intervals_plot-visualization, eval=requireNamespace("ggplot2")----
# Plot the concentration-time data and the interval
ggplot(d_conc_multi, aes(x=time, y=conc)) +
  geom_ribbon(data=d_conc_multi[d_conc_multi$time <= 24,],
              aes(ymax=conc, ymin=0),
              fill="skyblue") +
  geom_ribbon(data=d_conc_multi[d_conc_multi$time >= 144,],
              aes(ymax=conc, ymin=0),
              fill="lightgreen") +
  geom_point() + geom_line() +
  scale_x_continuous(breaks=seq(0, 168, by=12)) +
  scale_y_continuous(limits=c(0, NA)) +
  labs(x="Time Since First Dose (hr)",
       y="Concentration\n(arbitrary units)")

## ----multiple_intervals_spec--------------------------------------------------
intervals_manual <-
  data.frame(
    start=c(0, 144),
    end=c(24, 168),
    auclast=TRUE
  )
knitr::kable(intervals_manual)

my.data <- PKNCAdata(d_conc, intervals=intervals_manual)

## ----overlapping_intervals_plot, echo=FALSE-----------------------------------
# Simulate concentration-time data and setup the PKNCAconc object
d_conc <-
  PKNCAconc(
    generate.conc(nsub=1,
                  ntreat=1,
                  time.points=c(0, 1, 2, 4, 6, 8, 12, 24, 36, 48),
                  nstudies=1,
                  nanalytes=1,
                  resid=0),
    conc~time|treatment+ID)
print(d_conc)

## ----overlapping_intervals_plot-visualization, eval=requireNamespace("ggplot2")----
# Use superposition to simulate multiple doses
ggplot(as.data.frame(d_conc), aes(x=time, y=conc)) +
  geom_ribbon(data=as.data.frame(d_conc),
              aes(ymax=conc, ymin=0),
              fill="lightgreen",
              alpha=0.5) +
  geom_ribbon(data=as.data.frame(d_conc)[as.data.frame(d_conc)$time <= 24,],
              aes(ymax=conc, ymin=0),
              fill="skyblue",
              alpha=0.5) +
  geom_point() + geom_line() +
  scale_x_continuous(breaks=seq(0, 168, by=12)) +
  scale_y_continuous(limits=c(0, NA)) +
  labs(x="Time Since First Dose (hr)",
       y="Concentration\n(arbitrary units)")

## ----overlapping_intervals_spec-----------------------------------------------
intervals_manual <-
  data.frame(
    start=0,
    end=c(24, Inf),
    auclast=TRUE,
    aucinf.obs=c(FALSE, TRUE)
  )
knitr::kable(intervals_manual)

my.data <- PKNCAdata(d_conc, intervals=intervals_manual)

## ----interval_yes_no, echo=FALSE, eval=requireNamespace("ggplot2")------------
ggplot_intervals <- function(definition, intervals) {
  intervals$within <-
    c("No", "Yes")[intervals$interval.start >= min(definition$start.x) &
                     intervals$interval.end >= max(definition$start.x)]
  intervals$within <-
    factor(intervals$within, levels=c("Yes", "No"))
  ggplot(intervals,
         aes(x=interval.start,
             xend=interval.end,
             y=y,
             yend=y,
             linetype=within)) +
    geom_segment() +
    geom_segment(data=definition,
                 mapping=aes(x=start.x, xend=start.x, y=start.y, yend=end.y),
                 arrow=arrow(),
                 inherit.aes=FALSE) +
    geom_point(aes(x=interval.start, y=1),
               shape="|",
               size=4) +
    geom_point(aes(x=max(interval.end), y=1),
               shape="|",
               size=4) +
    scale_y_continuous(breaks=NULL) +
    labs(x="Time",
         y=NULL)
}

interval_definition <-
  data.frame(start.x=c(0, 24),
             start.y=0.5,
             end.y=0.9)
show_intervals <-
  data.frame(interval.start=c(0, 4, 12, 24),
             interval.end=c(4, 12, 24, 48),
             y=1)
ggplot_intervals(interval_definition,
                 show_intervals)

interval_definition <-
  data.frame(start.x=c(0, 16),
             start.y=0.5,
             end.y=0.9)
show_intervals <-
  data.frame(interval.start=c(0, 4, 12, 24),
             interval.end=c(4, 12, 24, 48),
             y=1)
ggplot_intervals(interval_definition,
                 show_intervals)


## ----parameters-available, echo=FALSE-----------------------------------------
interval_spec <- get.interval.cols()
interval_spec <- interval_spec[sort(names(interval_spec))]

get_function_for_calc <- function(x) {
  if (is.na(x$FUN)) {
    if (is.null(x$depends)) {
      "(none)"
    } else {
      paste("See the parameter name", x$depends)
    }
  } else {
    x$FUN
  }
}

kable(
  data.frame(
    `Parameter Name`=names(interval_spec),
    `Unit Type`=vapply(X = interval_spec, FUN = "[[", "unit_type", FUN.VALUE = ""),
    `Parameter Description`=vapply(X = interval_spec, FUN = "[[", "desc", FUN.VALUE = ""),
    `Function for Calculation`=vapply(X = interval_spec, FUN = get_function_for_calc, FUN.VALUE = ""),
    check.names=FALSE,
    stringsAsFactors=FALSE
  ),
  row.names=FALSE
)

## ----setup, echo=FALSE, include=FALSE-----------------------------------------
library(PKNCA)
library(dplyr)

## ----setup-example------------------------------------------------------------
library(PKNCA)
suppressPackageStartupMessages(library(dplyr))

d_conc <-
  as.data.frame(datasets::Theoph) %>%
  mutate(Subject=as.numeric(as.character(Subject)))
## Generate the dosing data
d_dose <- d_conc[d_conc$Time == 0,]
d_dose$Time <- 0

conc_obj <-
  PKNCAconc(
    d_conc,
    conc~Time|Subject
  )
dose_obj <-
  PKNCAdose(
    d_dose,
    Dose~Time|Subject
  )
data_obj <- PKNCAdata(conc_obj, dose_obj)
results_obj <- pk.nca(data_obj)

## ----exclude-function---------------------------------------------------------
results_excl_span <- exclude(results_obj, FUN=exclude_nca_span.ratio())

# Without any exclusions applied, the 'exclude' column is all NA.
as.data.frame(results_obj) %>%
  filter(Subject == 1)
# With exclusions applied, the 'exclude' column has the reason for exclusion.
as.data.frame(results_excl_span) %>%
  filter(Subject == 1)

## ----exclude-specific---------------------------------------------------------
mask_exclude_cmax <-
  results_obj %>%
  as.data.frame() %>%
  dplyr::mutate(
    mask_exclude=Subject == 1 & PPTESTCD == "cmax"
  ) %>%
  "[["("mask_exclude")
results_excl_specific <-
  exclude(
    results_obj,
    mask=mask_exclude_cmax,
    reason="Cmax was actually above the ULOQ"
  )

# Without any exclusions applied, the 'exclude' column is all NA.
results_obj %>%
  as.data.frame() %>%
  filter(Subject == 1)
# With exclusions applied, the 'exclude' column has the reason for exclusion.
results_excl_specific %>%
  as.data.frame() %>%
  filter(Subject == 1)

## ----exclude-multi------------------------------------------------------------
mask_exclude_lz <-
  results_obj %>%
  as.data.frame() %>%
  dplyr::mutate(
    mask_exclude=Subject == 1 & PPTESTCD == "lambda.z"
  ) %>%
  "[["("mask_exclude")

# Starting from the exclusion example above where short span ratios were
# excluded, exclude Cmax for Subject 1, too.
results_excl_multi <-
  exclude(
    results_excl_span,
    mask=mask_exclude_cmax,
    reason="Cmax was actually above the ULOQ"
  )
results_excl_multi <-
  exclude(
    results_excl_multi,
    mask=mask_exclude_lz,
    reason="Issue with lambda.z fit"
  )

# With exclusions applied, the 'exclude' column has the reason for exclusion.
# More than one reason may appear if more than one exclusion is applied.
results_excl_multi %>%
  as.data.frame() %>%
  filter(Subject == 1)

## ----summary------------------------------------------------------------------
summary(results_obj)

## ----summary-exclusion--------------------------------------------------------
summary(results_excl_span)

## ----listing------------------------------------------------------------------
as.data.frame(results_obj) %>%
  head(20)

## ----listing-exclusion--------------------------------------------------------
as.data.frame(results_excl_span) %>%
  head(20)

## ----check-ggplot, include=!requireNamespace("ggplot2"), results="asis"-------
cat("ggplot2 is required for this vignette to work correctly.  Please install the ggplot2 library and retry building the vignette.")

## ----setup--------------------------------------------------------------------
suppressPackageStartupMessages({
  library(PKNCA)
  library(dplyr)
  library(cowplot)
  library(knitr)
  library(ggplot2)
})
scale_colour_discrete <- scale_colour_hue
scale_fill_discrete <- scale_fill_hue

my_conc <- data.frame(conc=c(0, 2.5, 3, 2, 1.5, 1.2, 1.1, 0, 0),
                      time=c(0:5, 8, 12, 24),
                      subject=1)
my_conc$BLQ <- my_conc$conc == 0
my_conc$measured <- TRUE

## ----setup-visualization, eval=requireNamespace("ggplot2")--------------------
ggplot(my_conc,
       aes(x=time,
           y=conc,
           shape=BLQ,
           group=subject)) +
  geom_line() +
  geom_point(size=4) +
  scale_x_continuous(breaks=my_conc$time) +
  theme(legend.position=c(0.8, 0.8))

## ----calc_nca-----------------------------------------------------------------
conc_obj <- PKNCAconc(my_conc, conc~time|subject)
data_obj <- PKNCAdata(data.conc=conc_obj,
                      intervals=data.frame(start=0,
                                           end=24,
                                           aucall=TRUE,
                                           auclast=TRUE,
                                           aucinf.pred=TRUE,
                                           aucinf.obs=TRUE))
results_obj <- pk.nca(data_obj)
kable(as.data.frame(results_obj))

## ----auclast------------------------------------------------------------------
tlast <- pk.calc.tlast(conc=my_conc$conc,
                       time=my_conc$time)
tlast

my_conc$include_auclast <- my_conc$time <= tlast

## ----auclast-visualization, eval=requireNamespace("ggplot2")------------------
ggplot(my_conc,
       aes(x=time,
           y=conc,
           shape=BLQ,
           group=subject)) +
  geom_ribbon(data=my_conc[my_conc$include_auclast,],
              aes(ymin=0, ymax=conc),
              fill="lightblue") +
  geom_line() +
  geom_point(size=4) +
  scale_x_continuous(breaks=my_conc$time) +
  theme(legend.position=c(0.8, 0.8))

## ----aucall-------------------------------------------------------------------
first_after_tlast <- my_conc$time[my_conc$time > tlast][1]
first_after_tlast

my_conc$include_aucall <- my_conc$time <= first_after_tlast

## ----aucall-visualization, eval=requireNamespace("ggplot2")-------------------
ggplot(my_conc,
       aes(x=time,
           y=conc,
           shape=BLQ,
           group=subject)) +
  geom_ribbon(data=my_conc[my_conc$include_aucall,],
              aes(ymin=0, ymax=conc),
              fill="lightblue") +
  geom_line() +
  geom_point(size=4) +
  scale_x_continuous(breaks=my_conc$time) +
  theme(legend.position=c(0.8, 0.8))

## ----aucinf, fig.width=6------------------------------------------------------
# Add one row to illustrate extrapolation beyond observed data
my_conc <-
  rbind(my_conc,
        data.frame(conc=NA,
                   time=36,
                   subject=1,
                   BLQ=NA,
                   measured=FALSE,
                   include_auclast=FALSE,
                   include_aucall=FALSE))
# Extrapolate concentrations for aucinf.obs
my_conc$conc_aucinf.obs <- my_conc$conc
my_conc$conc_aucinf.obs[my_conc$BLQ | is.na(my_conc$BLQ)] <-
  interp.extrap.conc(conc=my_conc$conc,
                     time=my_conc$time,
                     time.out=my_conc$time[my_conc$BLQ | is.na(my_conc$BLQ)],
                     lambda.z=as.data.frame(results_obj)$PPORRES[as.data.frame(results_obj)$PPTESTCD %in% "lambda.z"])

# Extrapolate concentrations for aucinf.pred
my_conc$conc_aucinf.pred <- my_conc$conc
my_conc$conc_aucinf.pred[my_conc$BLQ | is.na(my_conc$BLQ)] <-
  interp.extrap.conc(conc=my_conc$conc,
                     time=my_conc$time,
                     time.out=my_conc$time[my_conc$BLQ | is.na(my_conc$BLQ)],
                     lambda.z=as.data.frame(results_obj)$PPORRES[as.data.frame(results_obj)$PPTESTCD %in% "lambda.z"],
                     clast=as.data.frame(results_obj)$PPORRES[as.data.frame(results_obj)$PPTESTCD %in% "clast.pred"])
my_conc$conc_aucinf.pred[my_conc$time == tlast] <-
  as.data.frame(results_obj)$PPORRES[as.data.frame(results_obj)$PPTESTCD %in% "clast.pred"]

## ----aucinf-visualization, eval=requireNamespace("ggplot2")-------------------
ggplot(my_conc[!is.na(my_conc$conc),],
       aes(x=time,
           y=conc,
           #shape=BLQ,
           group=subject)) +
  geom_ribbon(data=my_conc,
              aes(ymin=0, ymax=conc_aucinf.obs),
              fill="lightblue") +
  geom_line() +
  #geom_point(size=2) +
  scale_x_continuous(breaks=my_conc$time) +
  theme(legend.position=c(0.8, 0.8)) +
  labs(title="Extrapolation using AUCinf,obs")

ggplot(my_conc[!is.na(my_conc$conc),],
       aes(x=time,
           y=conc,
           #shape=BLQ,
           group=subject)) +
  geom_ribbon(
    data=arrange(
      bind_rows(mutate(my_conc,
                       before=FALSE),
                mutate(filter(my_conc, time == tlast),
                       conc_aucinf.pred=conc,
                       before=TRUE)),
      time, desc(before)),
    aes(ymin=0,
        ymax=conc_aucinf.pred),
    fill="lightblue") +
  geom_line() +
  #geom_point(size=2) +
  scale_x_continuous(breaks=my_conc$time) +
  theme(legend.position=c(0.8, 0.8)) +
  labs(title="Extrapolation using AUCinf,pred")

## ----partial_auc_observed-----------------------------------------------------
# Interpolation not required
data_obs_obj <- PKNCAdata(conc_obj, intervals=data.frame(start=0, end=2, auclast=TRUE))
results_obs_obj <- pk.nca(data_obs_obj)
kable(as.data.frame(results_obs_obj))

## ----partial_auc_interpolated-------------------------------------------------
# Interpolation required
my_conc_interp <-
  arrange(
    bind_rows(
      my_conc,
      data.frame(conc=interp.extrap.conc(conc=my_conc$conc, time=my_conc$time, time.out=1.5),
                 time=1.5,
                 subject=1)),
    time)
kable(my_conc_interp)

conc_interp_obj <- PKNCAconc(my_conc_interp, conc~time|subject)
data_interp_obj <- PKNCAdata(conc_interp_obj, intervals=data.frame(start=0, end=1.5, auclast=TRUE))
results_interp <- pk.nca(data_interp_obj)
as.data.frame(results_interp)

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(echo=TRUE,
                      fig.width=6,
                      fig.height=4)
library(PKNCA)
library(knitr)
library(testthat)
library(dplyr)

run_tests <- dir.exists(system.file(package="PKNCA", "tests/"))
testing_results <- "Tests were not run because tests are not installed."

## ----validation, include=FALSE, eval=run_tests--------------------------------
# testresult <- test_package(package="PKNCA")
# test_log <- as.data.frame(testresult)
# failed_tests <- sum(test_log$failed)
# testing_results <-
#   if (failed_tests == 0) {
#     "All tests passed."
#   } else {
#     "Some tests failed (see below)."
#   }

## ----testing_log, results="asis", echo=FALSE, eval=run_tests------------------
# cat("# Testing Log\n\n")
# 
# all_contexts <- unique(test_log$context)
# test_log_printing <-
#   test_log %>%
#   select(context, file, test, nb, failed) %>%
#   rename(`Testing Filename`=file,
#          `Test Description`=test,
#          `Number of Tests`=nb,
#          `Number of Failed Tests`=failed)
# 
# for (current_context in all_contexts) {
#   current_log <-
#     test_log_printing[
#       test_log_printing$context %in% current_context,
#       setdiff(names(test_log_printing), "context"), # context is in the header
#       drop=FALSE]
#   cat("## ", current_context, "\n\n")
#   print(kable(x=current_log,
#               row.names=FALSE))
#   cat("\n\n")
# }

## ----session_info-------------------------------------------------------------
Sys.Date()
sessionInfo()

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(echo = FALSE)
requireNamespace("pmxTools")
library(PKNCA)
library(dplyr)
library(ggplot2)
library(tidyr)
library(purrr)
breaks_hours <- function(n=5, Q=c(1, 6, 4, 12, 2, 24, 168), ...) {
  n_default <- n
  Q_default <- Q
  function(x, n = n_default, Q=Q_default) {
    x <- x[is.finite(x)]
    if (length(x) == 0) {
      return(numeric())
    }
    rng <- range(x)
    labeling::extended(rng[1], rng[2], m=n, Q=Q, ...)
  }
}

scale_x_hours <- function(..., breaks=breaks_hours()) {
  ggplot2::scale_x_continuous(..., breaks=breaks)
}

## ----fig.width=6, fig.height=4------------------------------------------------
conc <-
  datasets::Theoph %>%
  filter(Subject %in% 1)
ggplot(conc, aes(x=Time, y=conc)) +
  geom_line() +
  scale_x_hours()

## -----------------------------------------------------------------------------
conc_data <-
  withr::with_seed(5, {
    data.frame(
      Subject=rep(1:2, each=6),
      Treatment=rep(c("A", "B", "A", "B"), each=3),
      Time=rep(c(0, 2, 8), 4),
      Conc=rep(c(0, 2, 0.5), 4)*exp(rnorm(n=12, sd=0.05))
    )
  })

## -----------------------------------------------------------------------------
pander::pander(conc_data %>% filter(Subject == 1))

## -----------------------------------------------------------------------------
pander::pander(conc_data %>% filter(Subject == 2))

## -----------------------------------------------------------------------------
dose_data <-
  data.frame(
    Subject=rep(1:2, each=2),
    Treatment=c("A", "B", "A", "B"),
    Time=0,
    Dose=10
  )

## -----------------------------------------------------------------------------
pander::pander(dose_data %>% filter(Subject == 1))

## -----------------------------------------------------------------------------
pander::pander(dose_data %>% filter(Subject == 2))

## ----echo=TRUE----------------------------------------------------------------
d_interval_1 <-
  data.frame(
    start=0, end=8,
    cmax=TRUE, tmax=TRUE, auclast=TRUE
  )

## -----------------------------------------------------------------------------
pander::pander(d_interval_1)

## ----echo=TRUE----------------------------------------------------------------
library(dplyr)
library(ggplot2)
library(tidyr)
library(purrr)
library(PKNCA)
# Concentration data setup
d_conc <-
  datasets::Theoph %>%
  filter(Subject %in% 1)
o_conc <- PKNCAconc(conc~Time, data=d_conc)
# Setup intervals for calculation
d_intervals <- data.frame(start=0, end=24, cmax=TRUE, tmax=TRUE,
                          auclast=TRUE, aucint.inf.obs=TRUE)
# Combine concentration and dose
o_data <- PKNCAdata(o_conc, intervals=d_intervals)
# Calculate the results (suppressMessages() hides a message that isn't needed now)
o_result <- suppressMessages(pk.nca(o_data))
# summary(o_result)

## ----echo=TRUE----------------------------------------------------------------
# Concentration data setup
d_conc <-
  datasets::Theoph %>%
  filter(Subject %in% 1)
o_conc <- PKNCAconc(conc~Time, data=d_conc)
# Dose data setup
d_dose <-
  datasets::Theoph %>%
  filter(Subject %in% 1) %>%
  filter(Time == 0)
o_dose <- PKNCAdose(Dose~Time, data=d_dose)
# Combine concentration and dose
o_data <- PKNCAdata(o_conc, o_dose)
# Calculate the results
o_result <- pk.nca(o_data)

## ----echo=TRUE----------------------------------------------------------------
# Load your dataset as a data.frame
d_conc <-
  datasets::Theoph %>%
  filter(Subject %in% 1)
# Take a look at the data
pander::pander(head(d_conc, 2))
# Define the PKNCAconc object indicating the concentration and time columns, the
# dataset, and any other options.
o_conc <- PKNCAconc(conc~Time, data=d_conc)

## ----echo=TRUE----------------------------------------------------------------
# Load your dataset as a data.frame
d_dose <-
  datasets::Theoph %>%
  filter(Subject %in% 1) %>%
  filter(Time == 0)
# Take a look at the data
pander::pander(d_dose)
# Define the PKNCAdose object indicating the dose amount and time columns, the
# dataset, and any other options.
o_dose <- PKNCAdose(Dose~Time, data=d_dose)

## ----echo=TRUE----------------------------------------------------------------
# Combine the PKNCAconc and PKNCAdose objects.  You can add interval
# specifications and calculation options here.
o_data <- PKNCAdata(o_conc, o_dose)
# Calculate the results
o_result <- pk.nca(o_data)

## ----echo=TRUE----------------------------------------------------------------
# Look at summarized results
pander::pander(summary(o_result))

## ----echo=TRUE----------------------------------------------------------------
# Look at individual results
pander::pander(head(
  as.data.frame(o_result),
  n=3
))

## ----interval-vs-groups-setup, echo=FALSE-------------------------------------
last_dose_time <- 24
dose_interval <- 8
dose_times <- seq(0, last_dose_time-dose_interval, by=dose_interval)
d_conc_superposition <-
  superposition(
    o_conc,
    dose.times=dose_times,
    tau=last_dose_time,
    check.blq=FALSE,
    n.tau=1
  )

## ----fig.width=4, fig.height=4------------------------------------------------
d_intervals <-
  tibble(
    start=dose_times,
    end=dose_times + dose_interval
  ) %>%
  mutate(
    name=sprintf("Interval %g", row_number()),
    height=max(d_conc_superposition$conc)*1.03,
    width=dose_interval,
    x=(start+end)/2,
    y=height/2
  )
d_interval_arrows <-
  d_conc_superposition %>%
  filter(time != 0 & time %in% dose_times) %>%
  mutate(
    name1=sprintf("Interval %g", row_number()),
    name2=sprintf("Interval %g", row_number() + 1),
  )
ggplot(d_conc_superposition, aes(x=time, y=conc)) +
  geom_tile(
    data=d_intervals,
    aes(x=x, y=y, width=width, height=height, colour=name, fill=name),
    alpha=0.2,
    inherit.aes=FALSE,
    show.legend=FALSE
  ) +
  geom_segment(
    data=d_interval_arrows,
    aes(x=time - 0.8, xend=time - 0.1, y=conc-2.1, yend=conc - 0.1, colour=name2),
    arrow=arrow(length=unit(0.1, "inches")),
    inherit.aes=FALSE,
    show.legend=FALSE
  ) +
  geom_segment(
    data=d_interval_arrows,
    aes(x=time + 0.8, xend=time + 0.1, y=conc-2.1, yend=conc - 0.1, colour=name1),
    arrow=arrow(length=unit(0.1, "inches")),
    inherit.aes=FALSE,
    show.legend=FALSE
  ) +
  geom_line() +
  geom_point() +
  scale_x_hours() +
  labs(
    title=sprintf("Dosing Q%gH", dose_interval)
  )

## -----------------------------------------------------------------------------
PKNCA.options("single.dose.aucs") %>%
  select(c(all_of(c("start", "end")), where(~is.logical(.x) && any(.x)))) %>%
  pander::pander()

## ----echo=TRUE----------------------------------------------------------------
d_conc <-
  datasets::Theoph %>%
  mutate(
    Treatment=
      case_when(
        Dose <= median(Dose)~"Low dose",
        TRUE~"High dose"
      )
  )
# The study was single-dose
d_dose <-
  d_conc %>%
  select(Treatment, Subject, Dose) %>%
  unique() %>%
  mutate(dose_time=0)

## ----echo=TRUE----------------------------------------------------------------
o_conc <- PKNCAconc(conc~Time|Treatment+Subject, data=d_conc)
try({
  o_data <- PKNCAdata(o_conc)
  summary(pk.nca(o_data))
})

## ----echo=TRUE----------------------------------------------------------------
o_conc <- PKNCAconc(conc~Time|Treatment+Subject, data=d_conc)
d_intervals <- data.frame(start=0, end=Inf, cmax=TRUE, tmax=TRUE,
                          half.life=TRUE, aucinf.obs=TRUE)
o_data_manual_intervals <- PKNCAdata(o_conc, intervals=d_intervals)
summary(pk.nca(o_data_manual_intervals))

## ----echo=TRUE----------------------------------------------------------------
o_conc <- PKNCAconc(conc~Time|Treatment+Subject, data=d_conc)
o_dose <- PKNCAdose(Dose~dose_time|Treatment+Subject, data=d_dose)
o_data_auto_intervals <- PKNCAdata(o_conc, o_dose)
o_data_auto_intervals$intervals$aucint.inf.obs <- TRUE
summary(pk.nca(o_data_auto_intervals))

## ----auc-considerations-setup, warning=FALSE----------------------------------
d_conc <-
  datasets::Theoph %>%
  filter(Subject == 1)
o_conc <- PKNCAconc(conc~Time, data=d_conc)
d_interval_int <- data.frame(start=0, end=Inf, half.life=TRUE)
o_data_int <- PKNCAdata(o_conc, intervals=d_interval_int)
o_nca_int <- suppressMessages(pk.nca(o_data_int))
lambda_z_int <-
  o_nca_int %>%
  as.data.frame() %>%
  filter(PPTESTCD %in% "lambda.z") %>%
  "[["("PPORRES")

d_interval_inf <- data.frame(start=0, end=24, half.life=TRUE)
o_data_inf <- PKNCAdata(o_conc, intervals=d_interval_inf)
o_nca_inf <- suppressMessages(pk.nca(o_data_inf))
lambda_z_inf <-
  o_nca_inf %>%
  as.data.frame() %>%
  filter(PPTESTCD %in% "lambda.z") %>%
  "[["("PPORRES")

tlast <- 
  o_nca_inf %>%
  as.data.frame() %>%
  filter(PPTESTCD %in% "tlast") %>%
  "[["("PPORRES")

d_auc_calcs <-
  d_conc %>%
  bind_rows(
    tibble(Time=seq(12, 60))
  ) %>%
  mutate(
    conc_all_int=
      interp.extrap.conc(
        conc=conc[!is.na(conc)],
        time=Time[!is.na(conc)],
        time.out=Time,
        lambda.z=lambda_z_int
      ),
    conc_all_inf=
      interp.extrap.conc(
        conc=conc[!is.na(conc) & Time <= 24],
        time=Time[!is.na(conc) & Time <= 24],
        time.out=Time,
        lambda.z=lambda_z_inf
      ),
    conc_last=
      case_when(
        Time <= 24~conc,
        TRUE~NA_real_
      ),
    conc_int=
      case_when(
        Time <= 24 & Time >= tlast~conc_all_int,
        TRUE~NA_real_
      ),
    conc_inf=
      case_when(
        Time >= tlast~conc_all_inf,
        TRUE~NA_real_
      )
  ) %>%
  arrange(Time)
auc_figure_time_max <- 36
p_auc_calcs <-
  ggplot(d_auc_calcs, aes(x=Time, y=conc)) +
  # AUCinf (with a work-around for https://github.com/tidyverse/ggplot2/issues/4661)
  geom_area(
    data=d_auc_calcs %>% filter(Time <= auc_figure_time_max),
    aes(y=conc_inf, colour="AUCinf", fill="AUCinf"),
    alpha=0.2,
    na.rm=TRUE
  ) +
  geom_line(
    data=d_auc_calcs,
    aes(y=conc_inf, colour="AUCinf"),
    na.rm=TRUE
  ) +
  # AUCint
  geom_area(
    aes(y=conc_int, colour="AUCint", fill="AUCint"),
    alpha=0.2,
    na.rm=TRUE
  ) +
  # AUClast
  geom_area(
    aes(y=conc_last, colour="AUClast", fill="AUClast"),
    na.rm=TRUE
  ) +
  geom_point(show.legend=FALSE,
    na.rm=TRUE) +
  geom_line(show.legend=FALSE,
    na.rm=TRUE) +
  geom_vline(xintercept=24, linetype="63") +
  scale_x_continuous(breaks=seq(0, auc_figure_time_max, by=6)) +
  coord_cartesian(xlim=c(0, auc_figure_time_max)) +
  labs(
    colour="AUC type",
    fill="AUC type"
  )

## ----warning=FALSE, out.width="100%"------------------------------------------
p_auc_calcs

## ----warning=FALSE, out.width="100%"------------------------------------------
p_auc_calcs

## ----warning=FALSE, out.width="100%"------------------------------------------
p_auc_calcs

## ----eval=FALSE, echo=TRUE----------------------------------------------------
# library(PKNCA)
# 
# d_conc <- read.csv("c:/tmp/whale_conc.csv")
# d_dose <- read.csv("c:/tmp/whale_dose.csv")
# head(d_conc)
# head(d_dose)
# 
# o_conc <- PKNCAconc(concentration~time|Animal, data=d_conc)
# o_dose <- PKNCAdose(dose~time|Animal, data=d_dose)
# o_data <- PKNCAdata(o_conc, o_dose)
# o_data$intervals
# o_nca <- pk.nca(o_data)
# summary(o_nca)
# summary(o_nca, drop.group=c())
# as.data.frame(o_nca)

## ----exclude-example-1, echo=TRUE---------------------------------------------
d_before_exclude <-
 data.frame(
  time=0:4,
  conc=c(0, 2, 1, 0.5, 0.25),
  not_this=c(NA, "Not this", rep(NA, 3))
 )
o_conc <-
 PKNCAconc(
  data=d_before_exclude,
  conc~time,
  exclude="not_this"
 )

## ----exclude-example-2, echo=TRUE---------------------------------------------
pander::pander(
  d_before_exclude %>%
    filter(is.na(not_this))
)

## ----echo=TRUE, eval=FALSE----------------------------------------------------
# o_conc <- PKNCAconc(data=d_before_exclude, conc~time, exclude="not_this")

## ----echo=TRUE----------------------------------------------------------------
d_urine <-
  data.frame(
    conc=c(1, 2, 3),
    urine_volume=c(200, 100, 300),
    time=c(1, 2, 3)
  )
o_conc <- PKNCAconc(data=d_urine, conc~time, volume="urine_volume")
d_intervals <- data.frame(start=0, end=24, ae=TRUE)
o_data <- PKNCAdata(o_conc, intervals=d_intervals)
o_nca <- suppressMessages(pk.nca(o_data))
summary(o_nca)

## ----echo=TRUE----------------------------------------------------------------
o_conc <- PKNCAconc(data=data.frame(conc=2^-(1:4), time=0:3), conc~time)
o_data <- PKNCAdata(o_conc, intervals=data.frame(start=0, end=Inf, cmax=TRUE))
o_nca <- suppressMessages(pk.nca(o_data))
as.data.frame(o_nca)

## ----echo=TRUE----------------------------------------------------------------
o_data <-
  PKNCAdata(
    o_conc,
    intervals=
      data.frame(
        start=0, end=Inf,
        aucinf.obs=TRUE
      )
  )
o_nca <- suppressMessages(pk.nca(o_data))

## ----echo=TRUE----------------------------------------------------------------
as.data.frame(o_nca)

## ----fig.height=3-------------------------------------------------------------
d_prep <-
  datasets::Theoph %>%
  filter(Subject == 1) %>%
  mutate(
    conc=
      case_when(
        Time == 0~0,
        TRUE~conc
      )
  )
d_plot <-
  superposition(
    conc=d_prep$conc,
    time=d_prep$Time,
    tau=48,
    n.tau=1,
    dose.times=24*(0:1)
  ) %>%
  # Pretend that we missed the predose sample
  filter(
    !between(time, 23, 24.5)
  )
# Pretend that there was a dose at 24
ggplot(d_plot, aes(x=time, y=conc)) +
  geom_point() + geom_line() +
  geom_vline(xintercept=24) +
  scale_x_hours()

## ----exclude-setup------------------------------------------------------------
o_conc <- PKNCAconc(data=data.frame(conc=2^-(1:4), time=0:3), conc~time)
o_data <-
  PKNCAdata(
    o_conc,
    intervals=
      data.frame(
        start=0, end=Inf,
        aucinf.obs=TRUE
      )
  )
o_nca <- suppressMessages(pk.nca(o_data))

## ----exclude-not-best, echo=TRUE----------------------------------------------
as.data.frame(o_nca) %>%
  filter(PPTESTCD != "half.life")

## ----exclude-best, echo=TRUE--------------------------------------------------
o_nca_excluded <-
  o_nca %>%
  exclude(FUN=exclude_nca_span.ratio(3))
as.data.frame(o_nca_excluded)

## ----echo=TRUE----------------------------------------------------------------
summary(o_nca)
summary(o_nca_excluded)

## ----echo=TRUE----------------------------------------------------------------
# Subject 2 is selected for a BLQ time=0 concentration
d_prep <-
  datasets::Theoph %>%
  filter(Subject == 2)
# Superposition to steady-state is the default
d_ss <-
  superposition(
    conc=d_prep$conc,
    time=d_prep$Time,
    tau=24
  )
# Going to steady-state is also an option
# (n.tau=2 means the second dose)
d_second_dose <-
  superposition(
    conc=d_prep$conc,
    time=d_prep$Time,
    tau=24,
    n.tau=2
  )

## ----echo=TRUE----------------------------------------------------------------
# Want the profile for the first two doses
# together?
d_first_two <-
  superposition(
    conc=d_prep$conc,
    time=d_prep$Time,
    tau=48, # 48 hours
    n.tau=1, # One tau interval (0 to 48 hours)
    dose.times=c(0, 24)
  )

## ----fig.height=3, fig.width=3------------------------------------------------
ggplot(d_ss, aes(x=time, y=conc)) +
  geom_point() + geom_line() +
  scale_y_continuous(limits=c(0, NA))

## ----echo=TRUE----------------------------------------------------------------
dose_times <- seq(0, 96-1, by=6)
d_multidose <-
  superposition(
    conc=d_prep$conc,
    time=d_prep$Time,
    tau=96, # 48 hours
    n.tau=1, # One tau interval (0 to 48 hours)
    dose.times=dose_times
  )
pk.tss.monoexponential(
  conc=d_multidose$conc, time=d_multidose$time, subject=rep(1, nrow(d_multidose)),
  time.dosing=dose_times, subject.dosing=rep(1, length(dose_times)),
  output="single"
)

## ----echo=TRUE----------------------------------------------------------------
o_conc <- PKNCAconc(conc~Time|Subject, data=datasets::Theoph)
d_plot <-
  grouped_df(data=datasets::Theoph, vars=names(getGroups(o_conc))) %>%
  nest() %>%
  mutate(
    figure=
      lapply(
        pmap(.l=list(data=data), .f=ggplot,aes(x=Time, y=conc)),
        FUN="+",
        geom_line()
      )
  )
# d_plot$figure

## ----echo=TRUE----------------------------------------------------------------
pander::pander(summary(o_nca))

## ----echo=TRUE, eval=FALSE----------------------------------------------------
# pander::pander(as.data.frame(o_nca))

## ----echo=TRUE----------------------------------------------------------------
d_conc <-
  datasets::Theoph %>%
  rename(time=Time) %>%
  mutate(
    Subject=as.character(Subject)
  )
d_multidose <-
  PKNCAconc(conc~time|Subject, data=d_conc) %>%
  superposition(tau=24, check.blq=FALSE)
d_singledose_single_analyte <-
  d_conc %>%
  mutate(
    Study_Part="Single"
  )
d_multidose_single_analyte <-
  d_conc %>%
  mutate(Day=1) %>%
  bind_rows(
    d_multidose %>% mutate(time=time + 120, Day=6)
  ) %>%
  mutate(
    Study_Part="Multiple"
  )

## ----echo=TRUE----------------------------------------------------------------
d_single_multi_conc <- bind_rows(d_singledose_single_analyte, d_multidose_single_analyte)
d_single_multi_dose <-
  d_single_multi_conc %>%
  filter(
    (Study_Part %in% "Single" & time == 0) |
      (Study_Part %in% "Multiple" & (time %% 24) == 0)
  )

## ----echo=TRUE----------------------------------------------------------------
o_conc <- PKNCAconc(data=d_single_multi_conc, conc~time|Study_Part+Subject)
o_dose <- PKNCAdose(data=d_single_multi_dose, Dose~time|Study_Part+Subject)
o_data <- PKNCAdata(o_conc, o_dose)
o_data$intervals %>% select(-Subject) %>% unique() %>% as.data.frame()
o_nca <- pk.nca(o_data)

## ----echo=TRUE----------------------------------------------------------------
d_intervals <-
  data.frame(
    start=0,
    end=24,
    Subject=c("1", "2"),
    Study_Part="Single",
    aucinf.obs=TRUE
  )
o_data <- PKNCAdata(o_conc, o_dose, intervals=d_intervals)
o_nca <- pk.nca(o_data)
summary(o_nca)

## ----echo=TRUE----------------------------------------------------------------
# Find the time closest to 12 hours
d_intervals_prep <-
  d_single_multi_conc %>%
  filter(Study_Part == "Single") %>%
  mutate(
    time_deviation=abs(time-12)
  ) %>%
  group_by(Subject, Study_Part) %>%
  filter(time %in% time[time_deviation == min(time_deviation)])
d_intervals <-
  d_intervals_prep %>%
  select(Study_Part, Subject, end=time) %>%
  mutate(
    start=0,
    aucinf.obs=TRUE
  )
o_data <- PKNCAdata(o_conc, o_dose, intervals=d_intervals)

o_nca <- pk.nca(o_data)
summary(o_nca, drop.group=c("Subject", "end"))

## ----echo=TRUE----------------------------------------------------------------
d_single_multi_conc_multi_analyte <-
  bind_rows(
    d_single_multi_conc %>% mutate(Analyte="Parent"),
    d_single_multi_conc %>%
      mutate(
        Analyte="Metabolite",
        conc=conc/2
      )
  )
o_conc <-
  PKNCAconc(
    data=d_single_multi_conc_multi_analyte,
    conc~time|Study_Part+Subject/Analyte
  )
o_dose <- PKNCAdose(data=d_single_multi_dose, Dose~time|Study_Part+Subject)
o_data <- PKNCAdata(o_conc, o_dose)
o_nca <- pk.nca(o_data)
summary(o_nca)

## ----setup, echo=FALSE, include=FALSE-----------------------------------------
library(PKNCA)
library(knitr)
library(ggplot2)
scale_colour_discrete <- scale_colour_hue
scale_fill_discrete <- scale_fill_hue
scale_colour_ordinal <- scale_colour_hue
scale_fill_ordinal <- scale_fill_hue

## ----check-ggplot, include=!requireNamespace("ggplot2"), results="asis"-------
cat("ggplot2 is required for this vignette to work correctly.  Please install the ggplot2 library and retry building the vignette.")

## ----showtheoph---------------------------------------------------------------
## It is always a good idea to look at the data
knitr::kable(head(datasets::Theoph))

## ----setupconc----------------------------------------------------------------
## By default it is groupedData; convert it to a data frame for use
conc_obj <- PKNCAconc(as.data.frame(datasets::Theoph), conc~Time|Subject)

## ----superposition, error=TRUE------------------------------------------------
try({
steady_state <- superposition(conc_obj, tau=24)
})

## ----findnonzero--------------------------------------------------------------
knitr::kable(subset(datasets::Theoph, Time == 0 & conc > 0),
             caption="Nonzero predose measurements",
             row.names=FALSE)

## ----allownonzero-------------------------------------------------------------
## Correct nonzero concentrations at time 0 to be BLQ.
theoph_corrected <- as.data.frame(datasets::Theoph)
theoph_corrected$conc[theoph_corrected$Time == 0] <- 0
conc_obj_corrected <- PKNCAconc(theoph_corrected, conc~Time|Subject)

## Calculate the new steady-state concentrations with 24 hour dosing
steady_state <- superposition(conc_obj_corrected, tau=24)
knitr::kable(head(steady_state, n=14),
             caption="Superposition at steady-state")

## ----UnsteadyState------------------------------------------------------------
## Calculate the unsteady-state concentrations with 24 hour dosing
unsteady_state <- superposition(conc_obj_corrected, tau=24, n.tau=2)
knitr::kable(head(unsteady_state, n=14),
             caption="Superposition before steady-state")

## ----ComplexInterval----------------------------------------------------------
## Calculate the new steady-state concentrations with 24 hour dosing
complex_interval_steady_state <- superposition(conc_obj_corrected, tau=24, dose.times=c(0, 2, 4))
knitr::kable(head(complex_interval_steady_state, n=10),
             caption="Superposition at steady-state with complex dosing")

## ----ComplexInterval-visualization, eval=requireNamespace("ggplot2")----------
ggplot(complex_interval_steady_state,
       aes(y=conc, x=time, colour=Subject)) +
  geom_line()

## ----to_steady_state----------------------------------------------------------
up_to_steady_state <- superposition(conc_obj_corrected,
                                    tau=4*24,
                                    n.tau=1,
                                    dose.times=seq(0, 3*24, by=12))

## ----to_steady_state-visualization, eval=requireNamespace("ggplot2")----------
ggplot(up_to_steady_state, aes(x=time, y=conc, colour=Subject)) +
  geom_line()

## ----time-point-selection-----------------------------------------------------
steady_state$time[steady_state$Subject == 1]
sum(steady_state$Subject == 1)
complex_interval_steady_state$time[complex_interval_steady_state$Subject == 1]
sum(complex_interval_steady_state$Subject == 1)

## ----setup, echo=FALSE, include=FALSE-----------------------------------------
library(PKNCA)

## ----setup_data---------------------------------------------------------------
library(PKNCA)
library(dplyr, quietly=TRUE)

## Load the PK concentration data
d_conc <- datasets::Theoph
## Generate the dosing data
d_dose <- d_conc[d_conc$Time == 0,]

## Create a concentration object specifying the concentration, time, and
## subject columns.  (Note that any number of grouping levels is
## supported; you are not restricted to just grouping by subject.)
conc_obj <-
  PKNCAconc(
    d_conc,
    conc~Time|Subject
  )
## Create a dosing object specifying the dose, time, and subject
## columns.  (Note that the grouping factors should be the same as or a
## subset of the grouping factors for concentration, and the grouping
## columns must have the same names between concentration and dose
## objects.)
dose_obj <-
  PKNCAdose(
    d_dose,
    Dose~Time|Subject
  )
## Combine the concentration and dosing information both to
## automatically define the intervals for NCA calculation and provide
## doses for calculations requiring dose.
data_obj <- PKNCAdata(conc_obj, dose_obj)

## Calculate the NCA parameters
results_obj <- pk.nca(data_obj)

## Summarize the results
summary(results_obj)

## -----------------------------------------------------------------------------
PKNCA.options()

## ----eval=FALSE---------------------------------------------------------------
# PKNCA.options(default=TRUE)

## ----eval=FALSE---------------------------------------------------------------
# PKNCA.set.summary(
#   name = "cmax",
#   description = "geometric mean and geometric coefficient of variation",
#   point = business.geomean,
#   spread = business.geocv,
#   rounding = list(signif=3)
# )

## ----eval=FALSE---------------------------------------------------------------
# PKNCA.set.summary(
#   name = "tmax",
#   description = "median and range",
#   point = business.median,
#   spread = business.range,
#   rounding = list(round=2)
# )

## ----custom_summary_fun, eval=FALSE-------------------------------------------
# median_na <- function(x) {
#   median(x, na.rm = TRUE)
# }
# quantprob_na <- function(x) {
#   quantile(x, probs = c(0.05, 0.95), na.rm=TRUE)
# }
# PKNCA.set.summary(
#   name="auclast",
#   description = "median and 5th to 95th percentile",
#   point=median_na,
#   spread=quantprob_na,
#   rounding=list(signif=3)
# )

## ----multi_summary_settings, eval=FALSE---------------------------------------
# median_na <- function(x) {
#   median(x, na.rm=TRUE)
# }
# quantprob_na <- function(x) {
#   quantile(x, probs=c(0.05, 0.95), na.rm=TRUE)
# }
# PKNCA.set.summary(
#   name=c("auclast", "cmax", "tmax", "half.life", "aucinf.pred"),
#   description = "median and 5th to 95th percentile",
#   point=median_na,
#   spread=quantprob_na,
#   rounding=list(signif=3)
# )

## ----grouping, eval=FALSE-----------------------------------------------------
# ## Generate a faux multi-study, multi-analyte dataset.
# d_conc_parent <- d_conc
# d_conc_parent$Subject <- as.numeric(as.character(d_conc_parent$Subject))
# d_conc_parent$Study <- d_conc_parent$Subject <= 6
# d_conc_parent$Analyte <- "Parent"
# d_conc_metabolite <- d_conc_parent
# d_conc_metabolite$conc <- d_conc_metabolite$conc/2
# d_conc_metabolite$Analyte <- "Metabolite"
# d_conc_both <- rbind(d_conc_parent, d_conc_metabolite)
# d_conc_both <- d_conc_both[with(d_conc_both, order(Study, Subject, Time, Analyte)),]
# d_dose_both <- d_conc_both[d_conc_both$Time == 0 & d_conc_both$Analyte %in% "Parent",
#                            c("Study", "Subject", "Dose", "Time")]
# 
# ## Create a concentration object specifying the concentration, time,
# ## study, and subject columns.  (Note that any number of grouping
# ## levels is supporting; you are not restricted to this list.)
# conc_obj <- PKNCAconc(d_conc_both,
#                       conc~Time|Study+Subject/Analyte)
# ## Create a dosing object specifying the dose, time, study, and
# ## subject columns.  (Note that the grouping factors should be a
# ## subset of the grouping factors for concentration, and the columns
# ## must have the same names between concentration and dose objects.)
# dose_obj <- PKNCAdose(d_dose_both,
#                      Dose~Time|Study+Subject)
# 
# # Perform and summarize the PK data as previously described
# data_obj <- PKNCAdata(conc_obj, dose_obj)
# results_obj <- pk.nca(data_obj)
# summary(results_obj)

## -----------------------------------------------------------------------------
intervals <-
  data.frame(
    start=0, end=c(24, Inf),
    cmax=c(FALSE, TRUE),
    tmax=c(FALSE, TRUE),
    auclast=TRUE,
    aucinf.obs=c(FALSE, TRUE)
  )

## ----asis=TRUE, echo=FALSE----------------------------------------------------
knitr::kable(PKNCA.options()$single.dose.aucs)

## -----------------------------------------------------------------------------
## find.tau can work when all doses have the same interval...
dose_times <- seq(0, 168, by=24)
print(dose_times)
PKNCA::find.tau(dose_times)

## or when the doses have mixed intervals (10 and 24 hours).
dose_times <- sort(c(seq(0, 168, by=24),
                     seq(10, 178, by=24)))
print(dose_times)
PKNCA::find.tau(dose_times)

## ----eval=FALSE---------------------------------------------------------------
# intervals_manual <-
#   data.frame(
#     start=0, end=c(24, Inf),
#     cmax=c(FALSE, TRUE),
#     tmax=c(FALSE, TRUE),
#     auclast=TRUE,
#     aucinf.obs=c(FALSE, TRUE)
#   )
# data_obj <-
#   PKNCAdata(
#     conc_obj, dose_obj,
#     intervals=intervals_manual
#   )

## ----eval=FALSE---------------------------------------------------------------
# data_obj <- PKNCAdata(conc_obj, dose_obj)
# intervals_manual <- data_obj$intervals
# intervals_manual$aucinf.obs[1] <- TRUE
# data_obj$intervals <- intervals_manual

## ----eval=FALSE---------------------------------------------------------------
# data_obj <- PKNCAdata(conc_obj, dose_obj, options = list(keep_interval_cols = "dosetype"))

## ----eval=FALSE---------------------------------------------------------------
# summary(o_nca)

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(echo = TRUE)
library(PKNCA)
library(dplyr)

## ----generatecombs, echo=FALSE------------------------------------------------
method.choices <- names(PKNCA:::interp.extrap.conc.dose.select)
method.choices <-
  factor(method.choices, levels=method.choices, ordered=TRUE)
         
all_combs <-
  expand.grid(event_before=setdiff(unlist(PKNCA:::event_choices_interp.extrap.conc.dose), "output_only"),
              event=setdiff(unlist(PKNCA:::event_choices_interp.extrap.conc.dose), "none"),
              event_after=setdiff(unlist(PKNCA:::event_choices_interp.extrap.conc.dose), "output_only"),
              Method="",
              stringsAsFactors=FALSE)

for (n in method.choices) {
  mask <-
    do.call(PKNCA:::interp.extrap.conc.dose.select[[n]]$select,
            list(x=all_combs),
            envir=environment(pk.nca))
  all_combs$Method[mask] <- n
}

all_combs <-
  all_combs[do.call(order,
                    args=append(as.list(all_combs), list(na.last=FALSE))),]

## ----methodsummary, results="asis", echo=FALSE--------------------------------
methodorder <- names(sort(summary(factor(all_combs$Method)), decreasing=TRUE))

for (n in methodorder) {
  cat("## ", n, "\n\n", sep="")
  cat(PKNCA:::interp.extrap.conc.dose.select[[n]]$description, "\n\n", sep="")
  print(knitr::kable(
    all_combs[all_combs$Method %in% n,
              c("event_before","event", "event_after")],
    row.names=FALSE,
    col.names=c("Event Before", "Event At", "Event After")))
  cat("\n")
}

## ----methodlisting, echo=FALSE------------------------------------------------
knitr::kable(
  all_combs[,c("event_before", "event", "event_after", "Method")],
    row.names=FALSE,
    col.names=c("Event Before", "Event At", "Event After", "Method Used"))

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
library(PKNCA)
library(knitr)
library(ggplot2)
scale_colour_discrete <- scale_colour_hue
scale_fill_discrete <- scale_fill_hue
scale_colour_ordinal <- scale_colour_hue
scale_fill_ordinal <- scale_fill_hue

## ----superposition, error=TRUE------------------------------------------------
try({
library(PKNCA)
theoph_corrected <- as.data.frame(datasets::Theoph)
theoph_corrected$conc[theoph_corrected$Time == 0] <- 0
conc_obj <- PKNCAconc(theoph_corrected, conc~Time|Subject)
steady_state <- superposition(conc_obj, dose.times = seq(0, 168 - 12, by=12), tau=168, n.tau=1)
# Add some noise to the data so that it seems more reasonable
steady_state_noise <- steady_state
steady_state_noise$conc <-
  withr::with_seed(
    seed = 5,
    steady_state_noise$conc*exp(rnorm(nrow(steady_state_noise), mean = 0, sd = 0.1))
  )
})

## ----superposition-fig--------------------------------------------------------
library(ggplot2)
ggplot(steady_state_noise, aes(x=time, y=conc, groups=Subject)) + geom_line()

## ----tss-mono-----------------------------------------------------------------
tss_mono <-
  pk.tss.monoexponential(
    conc = steady_state_noise$conc,
    time = steady_state_noise$time,
    subject = steady_state_noise$Subject,
    time.dosing = seq(0, 168 - 12, by=12)
  )
tss_mono

## ----tss-step-----------------------------------------------------------------
tss_step <-
  pk.tss.stepwise.linear(
    conc = steady_state_noise$conc,
    time = steady_state_noise$time,
    subject = steady_state_noise$Subject,
    time.dosing = seq(0, 168 - 12, by=12)
  )
tss_step

## ----setup, echo=FALSE, include=FALSE-----------------------------------------
library(PKNCA)
library(dplyr)

## ----check-ggplot, include=!requireNamespace("dplyr"), results="asis"---------
cat("dplyr is required for this vignette to work correctly.  Please install the dplyr library and retry building the vignette.")

## ----normal-example-----------------------------------------------------------
# Perform calculations for subject 1, only
data_conc <- as.data.frame(datasets::Theoph)[datasets::Theoph$Subject == 1, ]

# Keep all points
conc_obj <-
  PKNCAconc(
    data_conc,
    conc~Time|Subject
  )

# Only calculate half-life and parameters required for half-life
current_intervals <- data.frame(start=0, end=Inf, half.life=TRUE)
data_obj <- PKNCAdata(conc_obj, intervals=current_intervals)
result_obj <- pk.nca(data_obj)

# Extract the results for subject 1 
as.data.frame(result_obj)

## ----exclude-points-----------------------------------------------------------
data_conc$exclude_hl <- data_conc$Time == 12.12
# Confirm that we will be excluding exactly one point
stopifnot(sum(data_conc$exclude_hl) == 1)

# Drop one point
conc_obj_exclude1 <-
  PKNCAconc(
    data_conc,
    conc~Time|Subject,
    exclude_half.life="exclude_hl"
  )

data_obj_exclude1 <- PKNCAdata(conc_obj_exclude1, intervals=current_intervals)

# Perform the calculations
result_obj_exclude1 <- pk.nca(data_obj_exclude1)

# Results differ when excluding the 12-hour point for subject 1 (compare to
# example in the previous section)
as.data.frame(result_obj_exclude1)

## ----include-points-----------------------------------------------------------
data_conc$include_hl <- data_conc$Time > 3
# Confirm that we will be excluding exactly one point
stopifnot(sum(data_conc$include_hl) == 6)

# Drop one point
conc_obj_include6 <-
  PKNCAconc(
    data_conc,
    conc~Time|Subject,
    include_half.life="include_hl"
  )

data_obj_include6 <- PKNCAdata(conc_obj_include6, intervals=current_intervals)

# Perform the calculations
result_obj_include6 <- pk.nca(data_obj_include6)

# Results differ when including 6 points (compare to example in the previous
# section)
as.data.frame(result_obj_include6)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----results='markup'---------------------------------------------------------
library(PKNCA)
cat(paste(
  "*", ls("package:PKNCA", pattern = "^PKNCA_impute_method")
), sep = "\n")

## ----impute-full-data---------------------------------------------------------
library(PKNCA)
# Remove time 0 to illustrate that imputation works
d_conc <- as.data.frame(datasets::Theoph)[!datasets::Theoph$Time == 0, ]
conc_obj <- PKNCAconc(d_conc, conc~Time|Subject)
d_dose <- unique(datasets::Theoph[datasets::Theoph$Time == 0,
                                  c("Dose", "Time", "Subject")])
dose_obj <- PKNCAdose(d_dose, Dose~Time|Subject)
data_obj <- PKNCAdata(conc_obj, dose_obj, impute = "start_predose,start_conc0")
nca_obj <- pk.nca(data_obj)
summary(nca_obj)

## ----impute-by-interval-------------------------------------------------------
library(PKNCA)
# Remove time 0 to illustrate that imputation works
d_conc <- as.data.frame(datasets::Theoph)[!datasets::Theoph$Time == 0, ]
conc_obj <- PKNCAconc(d_conc, conc~Time|Subject)
d_dose <- unique(datasets::Theoph[datasets::Theoph$Time == 0,
                                  c("Dose", "Time", "Subject")])
dose_obj <- PKNCAdose(d_dose, Dose~Time|Subject)

d_intervals <-
  data.frame(
    start=0, end=c(24, 24.1),
    auclast=TRUE,
    impute=c(NA, "start_conc0")
  )

data_obj <- PKNCAdata(conc_obj, dose_obj, intervals = d_intervals, impute = "impute")
nca_obj <- pk.nca(data_obj)
# PKNCA does not impute time 0 by default, so AUClast in the 0-24 interval is
# not calculated
summary(nca_obj)

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(echo = TRUE)
library(PKNCA)

## ----options, echo=FALSE, results='asis'--------------------------------------
PKNCA.options(default=TRUE)
opts <- PKNCA.options()
for (n in names(opts)) {
  cat("## ", n, "\n", sep="")
  cat("\n")
  cat(PKNCA:::PKNCA.options.describe(n), "\n", sep="")
  cat("\n")
  if (is.character(opts[[n]]) | is.numeric(opts[[n]]) | is.factor(opts[[n]]) | is.logical(opts[[n]])) {
    cat("The default value is: ", opts[[n]], "\n", sep="")
  } else if (is.data.frame(opts[[n]])) {
    print(knitr::kable(opts[[n]]))
  } else {
    # Probably a list
    print(opts[[n]])
  }
  cat("\n")
}

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup--------------------------------------------------------------------
suppressPackageStartupMessages(
  library(PKNCA)
)

## ----create-conc-dose-obj-auto------------------------------------------------
d_conc <- as.data.frame(datasets::Theoph)
d_conc$concu_col <- "mg/L"
d_conc$timeu_col <- "hr"
d_dose <- datasets::Theoph[datasets::Theoph$Time == 0, c("Dose", "Time", "Subject")]
d_dose$doseu_col <- "mg/kg"
o_conc <- PKNCAconc(d_conc, conc~Time|Subject, concu = "concu_col", timeu = "timeu_col")
o_dose <- PKNCAdose(d_dose, Dose~Time|Subject, doseu = "doseu_col")

## ----create-data-obj-auto-----------------------------------------------------
o_data <- PKNCAdata(o_conc, o_dose)
o_nca <- pk.nca(o_data)
summary(o_nca)

## ----create-auto-direct-------------------------------------------------------
d_conc <- as.data.frame(datasets::Theoph)
d_dose <- datasets::Theoph[datasets::Theoph$Time == 0, c("Dose", "Time", "Subject")]
o_conc <- PKNCAconc(d_conc, conc~Time|Subject, concu = "mg/L", timeu = "hr")
o_dose <- PKNCAdose(d_dose, Dose~Time|Subject, doseu = "mg/kg")
o_data <- PKNCAdata(o_conc, o_dose)
o_nca <- pk.nca(o_data)
summary(o_nca)

## ----create-auto-direct-pref--------------------------------------------------
d_conc <- as.data.frame(datasets::Theoph)
d_dose <- datasets::Theoph[datasets::Theoph$Time == 0, c("Dose", "Time", "Subject")]
o_conc <- PKNCAconc(d_conc, conc~Time|Subject, concu = "mg/L", timeu = "hr", concu_pref = "ug/L", timeu_pref = "day")
o_dose <- PKNCAdose(d_dose, Dose~Time|Subject, doseu = "mg/kg")
o_data <- PKNCAdata(o_conc, o_dose)
o_nca <- pk.nca(o_data)
summary(o_nca)

## ----create-conc-dose-obj-----------------------------------------------------
o_conc <- PKNCAconc(as.data.frame(datasets::Theoph), conc~Time|Subject)
d_dose <- datasets::Theoph[datasets::Theoph$Time == 0, c("Dose", "Time", "Subject")]
o_dose <- PKNCAdose(d_dose, Dose~Time|Subject)

## ----create-data-obj----------------------------------------------------------
d_units <-
  pknca_units_table(
    concu="mg/L", doseu="mg/kg", timeu="hr",
    # use molar units for concentrations and AUCs
    conversions=
      data.frame(
        PPORRESU=c("(mg/kg)/(hr*mg/L)", "(mg/kg)/(mg/L)", "mg/L", "hr*mg/L"),
        PPSTRESU=c("L/hr/kg", "L/kg", "mmol/L", "hr*mmol/L"),
        conversion_factor=c(NA, NA, 1/180.164, 1/180.164)
      )
  )

o_data <- PKNCAdata(o_conc, o_dose, units=d_units)
o_nca <- pk.nca(o_data)
summary(o_nca)

## ----create-units-auto--------------------------------------------------------
d_units_auto <- pknca_units_table(concu="ng/mL", doseu="mg", amountu="mg", timeu="hr")
# Show a selection of the units generated
d_units_auto[d_units_auto$PPTESTCD %in% c("cmax", "tmax", "auclast", "cl.obs", "vd.obs"), ]

## ----create-units-semi-manual-------------------------------------------------
d_units_clean <-
  pknca_units_table(
    concu="ng/mL", doseu="mg", amountu="ng", timeu="hr",
    conversions=
      data.frame(
        PPORRESU=c("mg/(hr*ng/mL)", "mg/(ng/mL)", "hr", "ng/mg"),
        PPSTRESU=c("L/hr", "L", "day", "fraction")
      )
  )
# Show a selection of the units generated
d_units_clean[d_units_clean$PPTESTCD %in% c("cmax", "tmax", "auclast", "cl.obs", "vd.obs", "fe"), ]

## ----create-units-manual------------------------------------------------------
d_units_clean_manual <-
  pknca_units_table(
    concu="ng/mL", doseu="mg", amountu="mg", timeu="hr",
    conversions=
      data.frame(
        PPORRESU=c("mg/(hr*ng/mL)", "mg/(ng/mL)", "hr", "ng/mL"),
        PPSTRESU=c("L/hr", "L", "day", "nmol/L"),
        conversion_factor=c(NA, NA, NA, 1000/123)
      )
  )
# Show a selection of the units generated
d_units_clean_manual[d_units_clean_manual$PPTESTCD %in% c("cmax", "tmax", "auclast", "cl.obs", "vd.obs"), ]

## ----different-units-conc-dose-setup------------------------------------------
d_conc_theoph <- as.data.frame(datasets::Theoph)
d_conc_theoph$Analyte <- "Theophylline"
# Approximately 6% of theophylline is metabolized to caffeine
# (https://www.pharmgkb.org/pathway/PA165958541).  Let's pretend that means it
# has 6% of the theophylline concentration at all times.
d_conc_caffeine <- as.data.frame(datasets::Theoph)
d_conc_caffeine$conc <- 0.06*d_conc_caffeine$conc
d_conc_caffeine$Analyte <- "Caffeine"
d_conc <- rbind(d_conc_theoph, d_conc_caffeine)

d_dose <- unique(datasets::Theoph[datasets::Theoph$Time == 0,
                                  c("Dose", "Time", "Subject")])

## ----different-units-units-setup----------------------------------------------
d_units_theoph <-
  pknca_units_table(
    concu="mg/L", doseu="mg/kg", timeu="hr",
    # use molar units for concentrations and AUCs
    conversions=
      data.frame(
        PPORRESU=c("(mg/kg)/(hr*mg/L)", "(mg/kg)/(mg/L)", "mg/L", "hr*mg/L"),
        PPSTRESU=c("L/hr/kg", "L/kg", "mmol/L", "hr*mmol/L"),
        conversion_factor=c(NA, NA, 1/180.164, 1/180.164)
      )
  )
d_units_theoph$Analyte <- "Theophylline"
d_units_caffeine <-
  pknca_units_table(
    concu="mg/L", doseu="mg/kg", timeu="hr",
    # use molar units for concentrations and AUCs
    conversions=
      data.frame(
        PPORRESU=c("(mg/kg)/(hr*mg/L)", "(mg/kg)/(mg/L)", "mg/L", "hr*mg/L"),
        PPSTRESU=c("L/hr/kg", "L/kg", "mmol/L", "hr*mmol/L"),
        conversion_factor=c(NA, NA, 1/194.19, 1/194.19)
      )
  )
d_units_caffeine$Analyte <- "Caffeine"
d_units <- rbind(d_units_theoph, d_units_caffeine)

## ----different-units-calc-----------------------------------------------------
o_conc <- PKNCAconc(d_conc, conc~Time|Subject/Analyte)
o_dose <- PKNCAdose(d_dose, Dose~Time|Subject)
o_data <- PKNCAdata(o_conc, o_dose, units=d_units)
o_nca <- pk.nca(o_data)
summary(o_nca)

## ----knitr-setup, include = FALSE---------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  warning = FALSE,
  comment = "#>"
)

## ----setup, include = FALSE---------------------------------------------------
library(PKNCA)
library(ggplot2)
auc_methods <- c("lin up/log down", "linear", "lin-log")
auc_types <- c("AUCinf", "AUClast", "AUCall")

comma_and <- function(x) {
  paste(
    paste(x[-length(x)], collapse = ", "),
    ngettext(length(x) - 1, ", and", "and"),
    x[length(x)]
  )
}

## ----setup-example, include=FALSE---------------------------------------------
examples <-
  list(
    "Clast is above the LLOQ" =
      data.frame(
        conc = c(0, 1.8, 3, 2, 1, 0.5, 0.25),
        time = 0:6
      ),
    "Clast is below the LLOQ" =
      data.frame(
        conc = c(0, 1.8, 3, 2, 1, 0.5, 0),
        time = 0:6
      ),
    "Concentrations increase and decrease after T~max~ with zeros in the middle" =
      data.frame(
        conc = c(0, 1.8, 0, 0, 3, 2, 2.5, 0, 0, 2.5, 1, 0.5, 0),
        time = 0:12
      )
  )

## ----example-figures, fig.cap = caption, echo = FALSE-------------------------
caption <- character()
for (current_example_nm in names(examples)) {
  for (current_method in auc_methods) {
    for (current_type in auc_types) {
      caption <-
        c(
          caption,
          sprintf(
            "Example PK where %s; %s interpolation method; %s extrapolation method (AUC type); and keeping all BLQ values (not default BLQ handling)",
            current_example_nm, current_method, current_type
          )
        )
      current_example <- examples[[current_example_nm]]

      # Handle BLQ
      current_example_noblq <- clean.conc.blq(conc = current_example$conc, time = current_example$time)
      current_example_noblq$BLQ <- "Above LLOQ"
      current_example_points <-
        merge(current_example, current_example_noblq, all = TRUE)
      current_example_points$BLQ[is.na(current_example_points$BLQ)] <- "Below LLOQ"
      
      # Calculate some NCA parameters required for interpolation and extrapolation
      tlast <- pk.calc.tlast(conc = current_example$conc, time = current_example$time)
      half_life <- pk.calc.half.life(conc = current_example$conc, time = current_example$time)
      
      # Define the methods used for interpolation and extrapolation
      current_example$method <-
        factor(
          PKNCA:::choose_interval_method(
            conc = current_example$conc,
            time = current_example$time,
            tlast = tlast,
            method = current_method,
            auc.type = current_type,
            options = list()
          ),
          levels = c("zero", "linear", "log", "extrap_log")
        )
      stopifnot(!any(is.na(current_example$method)))
      current_example_interpolate <-
        data.frame(
          time =
            sort(c(
              # make lines almost directly down at the transition after Tlast to
              # zero, where applicable
              current_example$time + sqrt(.Machine$double.eps),
              seq(min(current_example$time), max(current_example$time) + diff(current_example$time)[1], length.out = 200)
            ))
        )
      current_example_interpolate$conc <-
        interp.extrap.conc(
          conc = current_example$conc,
          time = current_example$time,
          time.out = current_example_interpolate$time,
          lambda.z = half_life$lambda.z,
          method = current_method,
          auc.type = current_type,
          conc.blq = "keep"
        )
      current_example$time_after <- c(current_example$time[-1], Inf)
      if (current_example$method[nrow(current_example)] == "zero") {
        current_example_plot <- current_example
      } else {
        current_example_before_tlast <- current_example[current_example$time < tlast, ]
        current_example_after_tlast <- current_example[current_example$time >= tlast, ][1, ]
        current_example_after_tlast$method <- current_example$method[nrow(current_example)]
        current_example_after_tlast$time_after <- Inf
        current_example_plot <-
          dplyr::bind_rows(
            current_example_before_tlast,
            current_example_after_tlast
          )
      }
      print(
        ggplot(current_example_points, aes(x = time, y = conc)) +
          geom_rect(data = current_example_plot, aes(xmin = time, xmax = time_after, ymin = -Inf, ymax = Inf, fill = method), alpha = 0.2) +
          geom_point(aes(shape = BLQ)) +
          geom_line(data = current_example_interpolate) +
          scale_colour_hue(drop = FALSE)
      )
    }
  }
}

#' Compute NCA parameters for each interval for each subject.
#'
#' The `pk.nca` function computes the NCA parameters from a `PKNCAdata` object.
#' All options for the calculation and input data are set in prior functions
#' (`PKNCAconc`, `PKNCAdose`, and `PKNCAdata`).  Options for calculations are
#' set either in `PKNCAdata` or with the current default options in
#' `PKNCA.options`.
#'
#' When performing calculations, all time results are relative to the start of
#' the interval.  For example, if an interval starts at 168 hours, ends at 192
#' hours, and and the maximum concentration is at 169 hours, `tmax=169-168=1`.
#'
#' @param data A PKNCAdata object
#' @param verbose Indicate, by `message()`, the current state of calculation.
#' @returns A `PKNCAresults` object.
#' @seealso [PKNCAdata()], [PKNCA.options()], [summary.PKNCAresults()],
#'   [as.data.frame.PKNCAresults()], [exclude()]
#' @export
pk.nca <- function(data, verbose=FALSE) {
  assert_PKNCAdata(data)
  results <- data.frame()
  if (nrow(data$intervals) > 0) {
    if (verbose) message("Setting up options")
    # Merge the options into the default options.
    tmp_options <- PKNCA.options()
    tmp_options[names(data$options)] <- data$options
    data$options <- tmp_options
    splitdata <- full_join_PKNCAdata(data)
    group_info <-
      splitdata[
        ,
        setdiff(names(splitdata), c("data_conc", "data_sparse_conc", "data_dose", "data_intervals")),
        drop=FALSE
      ]
    # Calculate the results
    if (verbose) message("Starting dense PK NCA calculations.")
    results_dense <-
      purrr::pmap(
        .l = list(
          data_conc = splitdata$data_conc,
          data_dose = splitdata$data_dose,
          data_intervals = splitdata$data_intervals
        ),
        .f = pk.nca.intervals,
        options = data$options,
        impute = data$impute,
        verbose = verbose,
        sparse = FALSE,
        .progress = data$options$progress
      )
    if (verbose) message("Combining completed dense PK calculation results.")
    results <- pk_nca_result_to_df(group_info, results_dense)
    if (is_sparse_pk(data)) {
      if (verbose) message("Starting sparse PK NCA calculations.")
      results_sparse <-
        purrr::pmap(
          .l=list(
            data_conc=splitdata$data_sparse_conc,
            data_dose=splitdata$data_dose,
            data_intervals=splitdata$data_intervals
          ),
          .f=pk.nca.intervals,
          options=data$options,
          impute=data$impute,
          verbose=verbose,
          sparse=TRUE
        )
      if (verbose) message("Combining completed sparse PK calculation results.")
      results <-
        dplyr::bind_rows(
          results,
          pk_nca_result_to_df(group_info, results_sparse)
        )
    }
  }
  PKNCAresults(
    result=results,
    data=data,
    exclude="exclude"
  )
}

#' Convert the grouping info and list of results for each group into a results
#' data.frame
#'
#' @param group_info A data.frame of grouping columns
#' @param result A list of data.frames with the results from NCA parameter
#'   calculations
#' @return A data.frame with group_info and result combined, warnings filtered
#'   out, and results unnested.
#' @keywords Internal
pk_nca_result_to_df <- function(group_info, result) {
  ret <- group_info
  ret$data_result <- result
  # Gather, report, and remove warnings
  mask_warning <- vapply(X=ret$data_result, inherits, what="warning", TRUE)
  ret_warnings <- ret[mask_warning, ]
  if (nrow(ret_warnings) > 0) {
    group_names <- setdiff(names(ret_warnings), "data_result")
    # Tell the user where the warning comes from
    warning_preamble <-
      do.call(
        what=paste,
        args=
          append(
            lapply(
              X=group_names,
              FUN=function(x) paste(x, ret_warnings[[x]], sep="=")
            ),
            list(sep="; ")
          )
      )
    invisible(lapply(
      X=seq_along(warning_preamble),
      FUN=function(idx) {
        warning_prep <- ret_warnings$data_result[[idx]]
        warning_prep$message <- paste(warning_preamble[idx], warning_prep$message, sep=": ")
        warning(warning_prep)
      }
    ))
  }
  ret_nowarning <- ret[!mask_warning, ]
  # Generate the outputs
  if (nrow(ret_nowarning) == 0) {
    rlang::warn(
      message = "All results generated warnings or errors; no results generated",
      class = "pknca_all_warnings_no_results"
    )
    results <- data.frame()
  } else {
    results <- tidyr::unnest(ret_nowarning, cols="data_result")
    rownames(results) <- NULL
  }
  results
}

filter_interval <- function(data, start, end, include_na=FALSE, include_end=TRUE) {
  mask_na <- include_na & is.na(data$time)
  mask_keep_start <- start <= data$time
  mask_keep_end <-
    if (include_end) {
      data$time <= end
    } else {
      data$time < end
    }
  mask_time <- mask_keep_start & mask_keep_end
  data[mask_na | mask_time, ]
}

#' Determine if there are any sparse or dense calculations requested within an interval
#'
#' @param interval An interval specification
#' @inheritParams PKNCAconc
#' @return A logical value indicating if the interval requests any sparse (if
#'   `sparse=TRUE`) or dense (if `sparse=FALSE`) calculations.
#' @keywords Internal
any_sparse_dense_in_interval <- function(interval, sparse) {
  all_intervals <- get.interval.cols()
  interval_subset <- interval[, names(interval) %in% names(all_intervals)]
  requested <- vapply(X = interval_subset, FUN = isTRUE, FUN.VALUE = TRUE)
  # Extract if the parameters to be calculated (`names(requested[requested])`)
  # are sparse, and compare that to if the request is for sparse or dense
  any(
    vapply(
      X=all_intervals[names(requested[requested])],
      FUN="[[",
      "sparse",
      FUN.VALUE = TRUE
    ) %in% sparse
  )
}

# Subset data down to just the times of interest and then pass it
# further to the calculation routines.
#
# This is simply a helper for pk.nca
#' Compute NCA for multiple intervals
#'
#' @param data_conc A data.frame or tibble with standardized column names as
#'   output from `prepare_PKNCAconc()`
#' @param data_dose A data.frame or tibble with standardized column names as
#'   output from `prepare_PKNCAdose()`
#' @param data_intervals A data.frame or tibble with standardized column names
#'   as output from `prepare_PKNCAintervals()`
#' @param impute The column name in `data_intervals` to use for imputation
#' @inheritParams PKNCAdata
#' @inheritParams pk.nca
#' @inheritParams pk.nca.interval
#' @return A data.frame with all NCA results
pk.nca.intervals <- function(data_conc, data_dose, data_intervals, sparse,
                             options, impute, verbose=FALSE) {
  if (is.null(data_conc) || (nrow(data_conc) == 0)) {
    # No concentration data; potentially placebo data
    return(rlang::warning_cnd(class="pknca_no_conc_data", message="No concentration data"))
  } else if (is.null(data_intervals) || (nrow(data_intervals) == 0)) {
    # No intervals; potentially placebo data
    return(rlang::warning_cnd(class="pknca_no_intervals", message="No intervals for data"))
  }
  ret <- data.frame()
  for (i in seq_len(nrow(data_intervals))) {
    current_interval <- data_intervals[i, , drop=FALSE]
    has_calc_sparse_dense <- any_sparse_dense_in_interval(current_interval, sparse=sparse)
    # Choose only times between the start and end.
    conc_data_interval <- filter_interval(data_conc, start=data_intervals$start[i], end=data_intervals$end[i])
    # Sort the data in time order
    conc_data_interval <- conc_data_interval[order(conc_data_interval$time),]
    NA_data_dose_ <- data.frame(dose=NA_real_, time=NA_real_, duration=NA_real_, route=NA_real_)
    if (is.null(data_dose) || identical(data_dose, NA)) {
      data_dose <- dose_data_interval <- NA_data_dose_
    } else {
      # include_end=FALSE so that a dose at the end of an interval is not included
      dose_data_interval <-
        filter_interval(
          data_dose,
          start=data_intervals$start[i],
          end=data_intervals$end[i],
          include_na=TRUE, include_end=FALSE
        )
    }
    if (nrow(dose_data_interval) > 0) {
      dose_data_interval <- dose_data_interval[order(dose_data_interval$time),]
    } else {
      # When all data are filtered out
      dose_data_interval <- NA_data_dose_
    }
    # Setup for detailed error reporting in case it's needed
    error_preamble <-
      paste(
        "Error with interval",
        paste(
          c("start", "end"),
          unlist(current_interval[, c("start", "end")]),
          sep="=", collapse=", ")
      )
    if (nrow(conc_data_interval) == 0) {
      warning(paste(error_preamble, "No data for interval", sep=": "))
    } else if (!has_calc_sparse_dense) {
      if (verbose) message("No ", ifelse(sparse, "sparse", "dense"), " calculations requested for an interval")
    } else {
      impute_method <- get_impute_method(intervals = current_interval, impute = impute)
      args <- list(
        # Interval-level data
        conc=conc_data_interval$conc,
        time=conc_data_interval$time,
        volume=conc_data_interval$volume,
        duration.conc=conc_data_interval$duration,
        dose=dose_data_interval$dose,
        time.dose=dose_data_interval$time,
        duration.dose=dose_data_interval$duration,
        route=dose_data_interval$route,
        impute_method=impute_method,
        # Group-level data
        conc.group=data_conc$conc,
        time.group=data_conc$time,
        volume.group=data_conc$volume,
        duration.conc.group=data_conc$duration,
        dose.group=data_dose$dose,
        time.dose.group=data_dose$time,
        duration.dose.group=data_dose$duration,
        route.group=data_dose$route,
        # Generic data
        sparse=sparse,
        interval=current_interval,
        options=options)
      if ("subject" %in% names(conc_data_interval)) {
        args$subject <- conc_data_interval$subject
      }
      if ("include_half.life" %in% names(conc_data_interval)) {
        args$include_half.life <- conc_data_interval$include_half.life
      }
      if ("exclude_half.life" %in% names(conc_data_interval)) {
        args$exclude_half.life <- conc_data_interval$exclude_half.life
      }
      # Try the calculation
      if (!is.null(PKNCA.options()$debug)) {
        # debugging mode does not need coverage
        calculated_interval <- do.call(pk.nca.interval, args) # nocov
      } else {
        calculated_interval <-
          tryCatch(
            do.call(pk.nca.interval, args),
            error=function(e) {
              e$message <- paste("Please report a bug.\n", error_preamble, e$message, sep=": ") # nocov
              stop(e) # nocov
            }
          )
      }
      # Add all the new data into the output
      new_ret <-
        cbind(
          # The rep(1, ...) is to fix #381 where attributes on an interval
          # column cause cbind to fail
          current_interval[
            rep(1, nrow(calculated_interval)),
            c("start", "end", options$keep_interval_cols)
          ],
          calculated_interval,
          row.names=NULL
        )
      ret <- rbind(ret, new_ret)
    }
  }
  ret
}

#' Compute all PK parameters for a single concentration-time data set
#'
#' For one subject/time range, compute all available PK parameters. All the
#' internal options should be set by [PKNCA.options()] prior to running.  The
#' only part that changes with a call to this function is the `conc`entration
#' and `time`.
#'
#' @inheritParams assert_conc_time
#' @inheritParams PKNCA.choose.option
#' @param conc.group All concentrations measured for the group
#' @param time.group Time of all concentrations measured for the group
#' @param volume,volume.group The volume (or mass) of the concentration
#'   measurement for the current interval or all data for the group (typically
#'   for urine and fecal measurements)
#' @param duration.conc,duration.conc.group The duration of the concentration
#'   measurement for the current interval or all data for the group (typically
#'   for urine and fecal measurements)
#' @param dose,dose.group Dose amount (may be a scalar or vector) for the
#'   current interval or all data for the group
#' @param time.dose Time of the dose for the current interval (must be the same
#'   length as `dose`)
#' @param time.dose.group Time of the dose for all data for the group (must be
#'   the same length as `dose.group`)
#' @param duration.dose The duration of the dose administration for the current
#'   interval (typically zero for extravascular and intravascular bolus and
#'   nonzero for intravascular infusion)
#' @param duration.dose.group The duration of the dose administration for all
#'   data for the group (typically zero for extravascular and intravascular
#'   bolus and nonzero for intravascular infusion)
#' @param route,route.group The route of dosing for the current interval or all
#'   data for the group
#' @param impute_method The method to use for imputation as a character string
#' @param interval One row of an interval definition (see
#'   [check.interval.specification()] for how to define the interval.
#' @param include_half.life An optional boolean vector of the concentration
#'   measurements to include in the half-life calculation. If given, no
#'   half-life point selection will occur.
#' @param exclude_half.life An optional boolean vector of the concentration
#'   measurements to exclude from the half-life calculation.
#' @param subject Subject identifiers (used for sparse calculations)
#' @param sparse Should only sparse calculations be performed (TRUE) or only
#'   dense calculations (FALSE)?
#' @returns A data frame with the start and end time along with all PK
#'   parameters for the `interval`
#'
#' @seealso [check.interval.specification()]
#' @export
pk.nca.interval <- function(conc, time, volume, duration.conc,
                            dose, time.dose, duration.dose, route,
                            conc.group=NULL, time.group=NULL, volume.group=NULL, duration.conc.group=NULL,
                            dose.group=NULL, time.dose.group=NULL, duration.dose.group=NULL, route.group=NULL,
                            impute_method=NA_character_,
                            include_half.life=NULL, exclude_half.life=NULL,
                            subject, sparse, interval, options=list()) {
  if (!is.data.frame(interval)) {
    stop("Please report a bug.  Interval must be a data.frame")
  }
  if (nrow(interval) != 1) {
    stop("Please report a bug.  Interval must be a one-row data.frame")
  }
  if (!all(is.na(impute_method))) {
    impute_funs <- PKNCA_impute_fun_list(impute_method)
    stopifnot(length(impute_funs) == 1)
    impute_data <- data.frame(conc=conc, time=time)
    for (current_fun_nm in impute_funs[[1]]) {
      impute_args <- as.list(impute_data)
      impute_args$start <- interval$start[1]
      impute_args$end <- interval$end[1]
      impute_args$conc.group <- conc.group
      impute_args$time.group <- time.group
      impute_args$options <- options
      impute_data <- do.call(current_fun_nm, args=impute_args)
    }
    conc <- impute_data$conc
    time <- impute_data$time
  }
  # Prepare the return value using SDTM names
  ret <- data.frame(PPTESTCD=NA, PPORRES=NA)[-1,]
  # Determine exactly what needs to be calculated in what order. Start with the
  # interval specification and find any dependencies that are not listed for
  # calculation.  Then loop over the calculations in order confirming what needs
  # to be passed from a previous calculation to a later calculation.
  all_intervals <- get.interval.cols()
  # Set the dose to NA if its length is zero
  if (length(dose) == 0) {
    stop("Please report a bug. Length of dose should not be zero.") # nocov
  }
  # Make sure that we calculate all of the dependencies.  Do this in
  # reverse order for dependencies of dependencies.
  for (n in rev(names(all_intervals))) {
    if (interval[[n]]) {
      interval[all_intervals[[n]]$depends] <- TRUE
    }
  }
  # Do the calculations
  for (n in names(all_intervals)) {
    request_to_calculate <- as.logical(interval[[n]])
    has_calculation_function <- !is.na(all_intervals[[n]]$FUN)
    is_correct_sparse_dense <- all_intervals[[n]]$sparse == sparse
    if (request_to_calculate & has_calculation_function & is_correct_sparse_dense) {
      call_args <- list()
      exclude_from_argument <- character(0)
      # Prepare to call the function by setting up its arguments.
      # Define the required arguments (arglist), and ignore the "..." argument
      # if it exists.
      arglist <- setdiff(names(formals(get(all_intervals[[n]]$FUN))),
                         "...")
      arglist <- stats::setNames(object=as.list(arglist), arglist)
      arglist[names(all_intervals[[n]]$formalsmap)] <- all_intervals[[n]]$formalsmap
      # Drop arguments that were set to NULL by the formalsmap
      arglist <- arglist[!vapply(X = arglist, FUN = is.null, FUN.VALUE = TRUE)]
      for (arg_formal in names(arglist)) {
        arg_mapped <- arglist[[arg_formal]]
        if (arg_mapped == "conc") {
          call_args[[arg_formal]] <- conc
        } else if (arg_mapped == "time") {
          # Realign the time to be relative to the start of the
          # interval
          call_args[[arg_formal]] <- time - interval$start[1]
        } else if (arg_mapped == "volume") {
          call_args[[arg_formal]] <- volume
        } else if (arg_mapped == "duration.conc") {
          call_args[[arg_formal]] <- duration.conc
        } else if (arg_mapped == "dose") {
          call_args[[arg_formal]] <- dose
        } else if (arg_mapped == "time.dose") {
          # Realign the time to be relative to the start of the
          # interval
          call_args[[arg_formal]] <- time.dose - interval$start[1]
        } else if (arg_mapped == "duration.dose") {
          call_args[[arg_formal]] <- duration.dose
        } else if (arg_mapped == "route") {
          call_args[[arg_formal]] <- route
        } else if (arg_mapped == "conc.group") {
          call_args[[arg_formal]] <- conc.group
        } else if (arg_mapped == "time.group") {
          # Don't realign the time to be relative to the start of the
          # interval
          call_args[[arg_formal]] <- time.group
        } else if (arg_mapped == "volume.group") {
          call_args[[arg_formal]] <- volume.group
        } else if (arg_mapped == "duration.conc.group") {
          call_args[[arg_formal]] <- duration.conc.group
        } else if (arg_mapped == "dose.group") {
          call_args[[arg_formal]] <- dose.group
        } else if (arg_mapped == "time.dose.group") {
          # Realign the time to be relative to the start of the
          # interval
          call_args[[arg_formal]] <- time.dose.group
        } else if (arg_mapped == "duration.dose.group") {
          call_args[[arg_formal]] <- duration.dose.group
        } else if (arg_mapped == "route.group") {
          call_args[[arg_formal]] <- route.group
        } else if (arg_mapped == "subject") {
          call_args[[arg_formal]] <- subject
        } else if (arg_mapped %in% c("start", "end")) {
          # Provide the start and end of the interval if they are requested
          call_args[[arg_formal]] <- interval[[arg_mapped]]
        } else if (arg_mapped == "options") {
          call_args[[arg_formal]] <- options
        } else if (any(mask_arg <- ret$PPTESTCD %in% arg_mapped)) {
          call_args[[arg_formal]] <- ret$PPORRES[mask_arg]
          exclude_from_argument <-
            c(exclude_from_argument, ret$exclude[mask_arg])
        } else if (!is.null(interval[[arg_mapped]])) {
          call_args[[arg_formal]] <- interval[[arg_mapped]]
        } else {
          # Give an error if there is not a default argument.
          if (inherits(formals(get(all_intervals[[n]]$FUN))[[arg_formal]], "name")) {
            arg_text <- # nocov start
              if (arg_formal == arg_mapped) {
                sprintf("'%s'", arg_formal)
              } else {
                sprintf("'%s' mapped to '%s'", arg_formal, arg_mapped)
              }
            stop(sprintf(
              "Cannot find argument %s for NCA function '%s'",
              arg_text, all_intervals[[n]]$FUN)
            ) # nocov end
          }
        }
      }
      # Apply manual inclusion and exclusion
      if (n %in% "half.life") {
        if (!is.null(include_half.life) && !all(is.na(include_half.life))) {
          include_tf <- include_half.life %in% TRUE
          call_args$conc <- call_args$conc[include_tf]
          call_args$time <- call_args$time[include_tf]
          call_args$manually.selected.points <- TRUE
        } else if (!is.null(exclude_half.life) && !all(is.na(exclude_half.life))) {
          exclude_tf <- exclude_half.life %in% TRUE
          call_args$conc <- call_args$conc[!exclude_tf]
          call_args$time <- call_args$time[!exclude_tf]
        }
      }
      # Do the calculation
      tmp_result <- do.call(all_intervals[[n]]$FUN, call_args)
      # The handling of the exclude column is documented in the
      # "Writing-Parameter-Functions.Rmd" vignette.  Document any changes to
      # this section of code there.
      exclude_reason <-
        stats::na.omit(c(
          exclude_from_argument, attr(tmp_result, "exclude")
        ))
      exclude_reason <-
        if (identical(attr(tmp_result, "exclude"), "DO NOT EXCLUDE")) {
          NA_character_
        } else if (length(exclude_reason) > 0) {
          paste(exclude_reason, collapse="; ")
        } else {
          NA_character_
        }
      # If the function returns a data frame, save all the returned values,
      # otherwise, save the value returned.
      if (is.data.frame(tmp_result)) {
        # if (uses_units) {
        #   # Convert to mixed_units so that rbind will work
        #   for (nm in names(tmp_result)) {
        #     if (inherits(tmp_result[[nm]], "units")) {
        #       tmp_result[[nm]] <- units::mixed_units(tmp_result[[nm]])
        #     } else {
        #       # unitless
        #       tmp_result[[nm]] <- units::mixed_units(tmp_result[[nm]], "")
        #     }
        #   }
        # }
        tmp_testcd <- names(tmp_result)
        tmp_result <- unlist(tmp_result, use.names=FALSE, recursive=FALSE)
      } else {
        # if (uses_units) {
        #   if (inherits(tmp_result, "units")) {
        #     # I() due to https://github.com/r-quantities/units/issues/309
        #     tmp_result <- I(units::mixed_units(tmp_result))
        #   } else {
        #     # unitless
        #     # I() due to https://github.com/r-quantities/units/issues/309
        #     tmp_result <- I(units::mixed_units(tmp_result, ""))
        #   }
        # }
        tmp_testcd <- n
      }
      single_result <-
        data.frame(
          PPTESTCD=tmp_testcd,
          PPORRES=tmp_result,
          exclude=exclude_reason,
          stringsAsFactors=FALSE
        )
      ret <- rbind(ret, single_result)
    }
  }
  ret
}

test_that("pk.nca", {
  # Note that generate.conc sets the random seed, so it doesn't have to happen
  # here.
  tmpconc <- generate.conc(2, 1, 0:24)
  tmpdose <- generate.dose(tmpconc)
  myconc <- PKNCAconc(tmpconc, formula=conc~time|treatment+ID)
  mydose <- PKNCAdose(tmpdose, formula=dose~time|treatment+ID)
  mydata <- PKNCAdata(myconc, mydose)
  myresult <- pk.nca(mydata)

  expect_equal(names(myresult),
               c("result", "data", "columns"),
               info="Make sure that the result has the expected names (and only the expected names) in it.")
  expect_true(checkProvenance(myresult),
              info="Provenance works on results")

  mydata.failure <- mydata
  # There's no way to automatically make a PKNCAdata object with no intervals,
  # but we want to ensure that users cannot cause this error by playing in the
  # internals.
  mydata.failure$intervals <- data.frame()
  expect_warning(myresult.failure <- pk.nca(mydata.failure),
                 regexp="No intervals given; no calculations will be done.",
                 info="An empty result is returned if there are no intervals")

  tmpconc <- generate.conc(2, 1, 0:24)
  tmpdose <- generate.dose(tmpconc)
  myconc <- PKNCAconc(tmpconc, formula=conc~time|treatment+ID)
  mydose.nodose <- PKNCAdose(tmpdose, formula=~time|treatment+ID)
  mydata.nodose <- PKNCAdata(myconc, mydose.nodose)
  expect_equal(
    pk.nca(mydata.nodose)$result,
    myresult$result,
    info="missing dose information is handled without an issue"
  )

  # Test each of the pieces for myresult for accuracy

  expect_equal(myresult$data, {
    tmp <- mydata
    # The options should be the default options after the calculations are done.
    tmp$options <- PKNCA.options()
    tmp
  }, info="The data is just a copy of the input data plus an instantiation of the PKNCA.options")

  verify.result <-
    tibble::tibble(
      treatment="Trt 1",
      ID=rep(c(1, 2), each=15),
      start=0,
      end=c(24, rep(Inf, 14),
            24, rep(Inf, 14)),
      PPTESTCD=rep(c("auclast", "cmax", "tmax", "tlast", "clast.obs",
                     "lambda.z", "r.squared", "adj.r.squared",
                     "lambda.z.time.first", "lambda.z.time.last",
                     "lambda.z.n.points", "clast.pred", "half.life",
                     "span.ratio", "aucinf.obs"),
                   times=2),
      PPORRES=c(13.54, 0.9998, 4.000, 24.00, 0.3441,
                0.04297, 0.9072, 0.9021, 5.000, 24.00,
                20.00, 0.3356, 16.13, 1.178,
                21.55, 14.03, 0.9410, 2.000,
                24.00, 0.3148, 0.05689, 0.9000, 0.8944,
                5.000, 24.00, 20.00, 0.3011, 12.18,
                1.560, 19.56),
      exclude=NA_character_
    )
  expect_equal(
    myresult$result,
    verify.result,
    tolerance=0.001,
    info=paste("The specific order of the levels isn't important--",
               "the fact that they are factors and that the set",
               "doesn't change is important.")
  )

  # Specifying new intervals
  mydata.newinterval <-
      PKNCAdata(myconc, mydose,
                intervals=data.frame(start=0, end=c(24, Inf),
                                     auclast=c(TRUE, FALSE),
                                     aucinf.obs=c(FALSE, TRUE),
                                     cmax=c(FALSE, TRUE),
                                     tmax=c(FALSE, TRUE),
                                     half.life=c(FALSE, TRUE)))
  myresult.newinterval <- pk.nca(mydata)
  expect_equal(myresult.newinterval$result,
               myresult$result,
               info="Intervals can be specified manually, and will apply across appropriate parts of the grouping variables.")


  # Dosing not at time 0
  tmpconc.multi <- generate.conc(2, 1, 0:24)
  tmpdose.multi <- generate.dose(tmpconc.multi)
  tmpconc.multi$time <- tmpconc.multi$time + 2
  tmpdose.multi$time <- tmpdose.multi$time + 2
  myconc.multi <- PKNCAconc(tmpconc.multi, conc~time|treatment+ID)
  mydose.multi <- PKNCAdose(tmpdose.multi, dose~time|treatment+ID)
  mydata.multi <- PKNCAdata(myconc.multi, mydose.multi)
  myresult.multi <- pk.nca(mydata.multi)

  verify.result.multi <- verify.result
  verify.result.multi$start <- verify.result.multi$start + 2
  verify.result.multi$end <- verify.result.multi$end + 2
  expect_equal(myresult.multi$result, verify.result.multi,
               tolerance=0.001,
               info="Shifted dosing works the same as un-shifted where time parameters like tmax and tlast are reported relative to the start of the interval")

  tmpconc <- generate.conc(2, 1, 0:24)
  tmpdose <- generate.dose(tmpconc)
  myconc <- PKNCAconc(tmpconc, conc~time|treatment+ID)
  mydose <- PKNCAdose(tmpdose, dose~time|treatment+ID)
  mydata <- PKNCAdata(myconc, mydose,
                      intervals=data.frame(start=0, end=Inf, cmax=TRUE))
  myresult <- pk.nca(mydata)
  expect_equal(myresult$result$PPORRES,
               c(0.99981, 0.94097), tolerance=0.00001,
               info="Calculations work with a single row of intervals and a single parameter requested")

  mydata <- PKNCAdata(myconc, mydose,
                      intervals=data.frame(start=0, end=Inf, cl.obs=TRUE))
  myresult <- pk.nca(mydata)
  expect_equal(subset(myresult$result, PPTESTCD %in% "cl.obs")$PPORRES,
               c(0.04640, 0.05111), tolerance=0.0001,
               info="PK intervals work with passing in dose as a parameter")

  tmpconc <- generate.conc(2, 1, 0:24)
  tmpdose <- generate.dose(tmpconc)
  myconc <- PKNCAconc(tmpconc, conc~time|treatment+ID)
  mydose <- PKNCAdose(tmpdose, dose~time|treatment+ID)
  mydata <- PKNCAdata(myconc, mydose,
                      intervals=data.frame(start=0, end=24, cmax=TRUE, cav=TRUE))
  myresult <- pk.nca(mydata)
  expect_equal(subset(myresult$result, PPTESTCD %in% "cav")$PPORRES,
               c(0.5642, 0.5846), tolerance=0.0001,
               info="PK intervals work with passing in start and end as parameters")

  # Ensure that the correct number of doses are included in parameters that use dosing.
  tmpconc <- generate.conc(2, 1, 0:24)
  tmpdose <- generate.dose(tmpconc)
  tmpdose$time <- NULL
  tmpdose <- merge(tmpdose, data.frame(time=c(0, 6, 12, 18, 24)))
  myconc <- PKNCAconc(tmpconc, conc~time|treatment+ID)
  mydose <- PKNCAdose(tmpdose, dose~time|treatment+ID)
  mydata <- PKNCAdata(myconc, mydose,
                      intervals=data.frame(start=0, end=24, cl.obs=TRUE))
  myresult <- pk.nca(mydata)
  expect_equal(myresult$result$PPORRES[myresult$result$PPTESTCD %in% "cl.obs"],
               4/myresult$result$PPORRES[myresult$result$PPTESTCD %in% "aucinf.obs"],
               tolerance=0.0001,
               info="The correct number of doses is selected for an interval (>=start and <end), 4 doses and not 5")

  mydata <- PKNCAdata(myconc, mydose,
                      intervals=data.frame(start=0, end=6, cl.last=TRUE))
  myresult <- pk.nca(mydata)
  expect_equal(myresult$result$PPORRES[myresult$result$PPTESTCD %in% "cl.last"],
               1/myresult$result$PPORRES[myresult$result$PPTESTCD %in% "auclast"],
               tolerance=0.0001,
               info="The correct number of doses is selected for an interval (>=start and <end), 1 dose and not 5")

  mydata <- PKNCAdata(myconc, mydose,
                      intervals=data.frame(start=1, end=6, cl.last=TRUE))
  myresult <- pk.nca(mydata)
  expect_equal(myresult$result$PPORRES[myresult$result$PPTESTCD %in% "cl.last"],
               NA/myresult$result$PPORRES[myresult$result$PPTESTCD %in% "auclast"],
               tolerance=0.0001,
               info="The correct number of doses is selected for an interval (>=start and <end), no doses selected")

})

test_that("verbose pk.nca", {
  tmpconc <- generate.conc(nsub=1, ntreat=1, 0:4)
  tmpdose <- generate.dose(tmpconc)
  myconc <- PKNCAconc(tmpconc, formula=conc~time)
  mydose <- PKNCAdose(tmpdose, formula=dose~time)
  mydata <- PKNCAdata(myconc, mydose)
  expect_message(expect_message(expect_message(
    suppressWarnings(pk.nca(mydata, verbose=TRUE)),
    regexp = "Setting up options"),
    regexp = "Starting dense PK NCA calculations"),
    regexp = "Combining completed dense PK calculation results"
  )
  expect_message(
    suppressWarnings(pk.nca(mydata, verbose=FALSE)),
    NA
  )
})

test_that("pk.nca warnings", {
  tmpconc <- generate.conc(nsub=1, ntreat=1, 0:4)
  tmpdose <- generate.dose(tmpconc)
  myconc <- PKNCAconc(tmpconc, formula=conc~time)
  mydose <- PKNCAdose(tmpdose, formula=dose~time)
  mydata <- PKNCAdata(myconc, mydose, intervals=data.frame(start=24, end=48, cmax=TRUE))
  expect_warning(
    pk.nca(mydata),
    regexp="No data for interval"
  )
})

test_that("pk.nca.interval errors", {
  expect_error(
    pk.nca.interval(interval="A"),
    regexp="Interval must be a data.frame"
  )
  expect_error(
    pk.nca.interval(interval=data.frame()),
    regexp="Interval must be a one-row data.frame"
  )
})

test_that("Calculations when dose time is missing", {
  # Ensure that the correct number of doses are included in parameters that use dosing.
  tmpconc <- generate.conc(2, 1, 0:24)
  tmpdose <- generate.dose(tmpconc)
  myconc <- PKNCAconc(tmpconc, conc~time|treatment+ID)
  mydose <- PKNCAdose(tmpdose, dose~.|treatment+ID)
  mydata <- PKNCAdata(myconc, mydose,
                      intervals=data.frame(start=0, end=24, cl.obs=TRUE))
  myresult <- pk.nca(mydata)
  expect_equal(
    myresult$result$PPORRES[myresult$result$PPTESTCD %in% "cl.obs"],
    1/myresult$result$PPORRES[myresult$result$PPTESTCD %in% "aucinf.obs"],
    info="The correct number of doses is selected for an interval (>=start and <end), 4 doses and not 5"
  )
})

test_that("Calculations when no dose info is given", {
  tmpconc <- generate.conc(2, 1, 0:24)
  myconc <- PKNCAconc(tmpconc, formula=conc~time|treatment+ID)
  mydata <- PKNCAdata(myconc, intervals=data.frame(start=0, end=24, cmax=TRUE, cl.last=TRUE))
  expect_message(
    myresult <- pk.nca(mydata),
    regexp="No dose information provided, calculations requiring dose will return NA.",
    info="Dosing information not required."
  )
  expect_equal(
    myresult$result,
    tibble::tibble(
      treatment="Trt 1",
      ID=rep(1:2, each=3),
      start=0,
      end=24,
      PPTESTCD=rep(c("auclast", "cmax", "cl.last"), 2),
      PPORRES=c(13.5417297156528, 0.999812606062292, NA,
                14.0305397438242, 0.94097296083447, NA),
      exclude=NA_character_
    )
  )
})

test_that("pk.nca with exclusions", {
  # Note that generate.conc sets the random seed, so it doesn't have to happen
  # here.
  tmpconc <- generate.conc(2, 1, 0:24)
  tmpdose <- generate.dose(tmpconc)
  myconc <- PKNCAconc(tmpconc, formula=conc~time|treatment+ID)
  mydose <- PKNCAdose(tmpdose, formula=dose~time|treatment+ID)
  mydata <- PKNCAdata(myconc, mydose)
  myresult <- pk.nca(mydata)
  tmpconc.excl <- tmpconc
  tmpconc.excl$excl <- NA_character_
  tmpconc.excl$excl[5] <- "test exclusion"
  myconc.excl <- PKNCAconc(tmpconc.excl,
                           formula=conc~time|treatment+ID,
                           exclude="excl")
  mydata.excl <- PKNCAdata(myconc.excl, mydose)
  myresult.excl <- pk.nca(mydata.excl)
  expect_true(identical(myresult$result[myresult$result$ID %in% 2,],
                        myresult.excl$result[myresult.excl$result$ID %in% 2,]),
               info="Results are unchanged for the subject who has the same data")
  expect_false(identical(myresult$result[myresult$result$ID %in% 1,],
                         myresult.excl$result[myresult.excl$result$ID %in% 1,]),
               info="Results are changed for the subject who has the same data")
  expect_equal(myresult.excl$result$PPORRES[myresult.excl$result$ID %in% 1 & myresult.excl$result$PPTESTCD %in% "cmax"],
               max(tmpconc.excl$conc[tmpconc.excl$ID %in% 1 & is.na(tmpconc.excl$excl)]),
               info="Cmax is affected by the exclusion")
})

test_that("pk.calc.all with duration.dose required", {
  tmpconc <- generate.conc(2, 1, 0:24)
  tmpdose <- generate.dose(tmpconc)
  tmpdose$duration_dose <- 0.1
  myconc <- PKNCAconc(tmpconc, formula=conc~time|treatment+ID)
  mydose <- PKNCAdose(tmpdose, formula=dose~time|treatment+ID, duration="duration_dose", route="intravascular")
  mydata <- PKNCAdata(myconc, mydose,
                      intervals=data.frame(start=0, end=24,
                                           mrt.iv.last=TRUE))
  myresult <- pk.nca(mydata)
  expect_equal(myresult$result$PPORRES[myresult$result$PPTESTCD %in% "mrt.iv.last"],
               c(10.36263, 10.12515),
               tolerance=1e-5,
               info="duration.dose is used when requested")
})

test_that("pk.calc.all with duration.conc required", {
  tmpconc <- generate.conc(2, 1, 0:24)
  tmpdose <- generate.dose(tmpconc)
  tmpconc$duration_conc <- 0.1
  myconc <- PKNCAconc(tmpconc, formula=conc~time|treatment+ID, duration="duration_conc")
  mydose <- PKNCAdose(tmpdose, formula=dose~time|treatment+ID, route="intravascular")
  mydata <- PKNCAdata(myconc, mydose,
                      intervals=data.frame(start=0, end=24,
                                           mrt.iv.last=TRUE))
  myresult <- pk.nca(mydata)
  expect_equal(myresult$result$PPORRES[myresult$result$PPTESTCD %in% "mrt.iv.last"],
               c(10.41263, 10.17515),
               tolerance=1e-5,
               info="duration.conc is used when requested")
})

test_that("half life inclusion and exclusion", {
  tmpconc <- generate.conc(2, 1, 0:24)
  tmpdose <- generate.dose(tmpconc)
  tmpconc$include_hl <- tmpconc$time <= 22
  tmpconc$exclude_hl <- tmpconc$time == 22
  myconc <- PKNCAconc(tmpconc, formula=conc~time|treatment+ID)
  myconc_incl <- PKNCAconc(tmpconc, formula=conc~time|treatment+ID,
                           include_half.life="include_hl")
  myconc_excl <- PKNCAconc(tmpconc, formula=conc~time|treatment+ID,
                           exclude_half.life="exclude_hl")
  mydose <- PKNCAdose(tmpdose, formula=dose~time|treatment+ID)
  mydata <-  PKNCAdata(myconc, mydose,
                       intervals=data.frame(start=0, end=24, half.life=TRUE))
  mydata_incl <- PKNCAdata(myconc_incl, mydose,
                           intervals=data.frame(start=0, end=24, half.life=TRUE))
  mydata_excl <- PKNCAdata(myconc_excl, mydose,
                           intervals=data.frame(start=0, end=24, half.life=TRUE))
  myresult <- pk.nca(mydata)
  myresult_incl <- pk.nca(mydata_incl)
  myresult_excl <- pk.nca(mydata_excl)
  expect_false(identical(myresult$result, myresult_excl$result))
  expect_false(identical(myresult$result, myresult_incl$result))
})

test_that("include_half.life and exclude_half.life work with NAs treated as missing for all NA and as FALSE for partial NA (#372)", {
  # Partial NA include_hl is used
  d_conc_incl <- data.frame(conc = c(1, 0.6, 0.3, 0.25, 0.15, 0.1), time = 0:5, include_hl = c(FALSE, NA, TRUE, TRUE, TRUE, TRUE))
  o_conc_incl <- PKNCAconc(d_conc_incl, conc~time, include_half.life = "include_hl")
  o_data_incl <- PKNCAdata(o_conc_incl, intervals = data.frame(start = 0, end = Inf, half.life = TRUE))
  suppressMessages(o_nca_incl <- pk.nca(o_data_incl))
  expect_equal(as.data.frame(o_nca_incl, out_format = "wide")$half.life, 1.820879, tolerance = 0.00001)

  # All FALSE include_hl is used
  d_conc_false <- data.frame(conc = c(1, 0.6, 0.3, 0.25, 0.15, 0.1), time = 0:5, include_hl = FALSE)
  o_conc_false <- PKNCAconc(d_conc_false, conc~time, include_half.life = "include_hl")
  o_data_false <- PKNCAdata(o_conc_false, intervals = data.frame(start = 0, end = Inf, half.life = TRUE))
  suppressWarnings(suppressMessages(o_nca_false <- pk.nca(o_data_false)))
  d_nca_false <- as.data.frame(o_nca_false)
  expect_equal(d_nca_false$PPORRES[d_nca_false$PPTESTCD %in% "half.life"], NA_real_)

  # All NA include_hl is ignored
  d_conc <- data.frame(conc = c(1, 0.6, 0.3, 0.25, 0.15, 0.1), time = 0:5, include_hl = NA)
  o_conc <- PKNCAconc(d_conc, conc~time, include_half.life = "include_hl")
  o_data <- PKNCAdata(o_conc, intervals = data.frame(start = 0, end = Inf, half.life = TRUE))
  suppressMessages(o_nca <- pk.nca(o_data))
  expect_equal(as.data.frame(o_nca, out_format = "wide")$half.life, 1.512942, tolerance = 0.00001)

  # Partial NA include_hl is used
  d_conc_excl <- data.frame(conc = c(1, 0.6, 0.3, 0.25, 0.15, 0.1), time = 0:5, exclude_hl = c(FALSE, NA, TRUE, TRUE, TRUE, TRUE))
  o_conc_excl <- PKNCAconc(d_conc_excl, conc~time, exclude_half.life = "exclude_hl")
  o_data_excl <- PKNCAdata(o_conc_excl, intervals = data.frame(start = 0, end = Inf, half.life = TRUE))
  suppressWarnings(suppressMessages(o_nca_excl <- pk.nca(o_data_excl)))
  d_nca_excl <- as.data.frame(o_nca_excl)
  expect_equal(d_nca_excl$PPORRES[d_nca_excl$PPTESTCD %in% "half.life"], NA_real_)

  # All NA exclude_hl is ignored
  d_conc <- data.frame(conc = c(1, 0.6, 0.3, 0.25, 0.15, 0.1), time = 0:5, exclude_hl = NA)
  o_conc <- PKNCAconc(d_conc, conc~time, exclude_half.life = "exclude_hl")
  o_data <- PKNCAdata(o_conc, intervals = data.frame(start = 0, end = Inf, half.life = TRUE))
  suppressMessages(o_nca <- pk.nca(o_data))
  expect_equal(as.data.frame(o_nca, out_format = "wide")$half.life, 1.512942, tolerance = 0.00001)

  # All FALSE exclude_hl is used
  d_conc_false <- data.frame(conc = c(1, 0.6, 0.3, 0.25, 0.15, 0.1), time = 0:5, exclude_hl = FALSE)
  o_conc_false <- PKNCAconc(d_conc_false, conc~time, exclude_half.life = "exclude_hl")
  o_data_false <- PKNCAdata(o_conc_false, intervals = data.frame(start = 0, end = Inf, half.life = TRUE))
  suppressWarnings(suppressMessages(o_nca_false <- pk.nca(o_data_false)))
  d_nca_false <- as.data.frame(o_nca_false)
  expect_equal(d_nca_false$PPORRES[d_nca_false$PPTESTCD %in% "half.life"], 1.512942, tolerance = 0.00001)
})

test_that("No interval requested (e.g. for placebo)", {
  tmpconc <- generate.conc(2, 1, 0:24)
  tmpdose <- generate.dose(tmpconc)
  myconc <- PKNCAconc(tmpconc, formula=conc~time|treatment+ID)
  mydose <- PKNCAdose(tmpdose, formula=dose~time|treatment+ID)
  mydata <-
    PKNCAdata(
      myconc, mydose,
      intervals=
        data.frame(
          treatment="Trt 3", start=0, end=24, cmax=TRUE,
          stringsAsFactors=FALSE
        )
    )
  expect_warning(expect_warning(expect_warning(expect_warning(
    myresult <- pk.nca(mydata),
    class = "pknca_no_intervals"),
    class = "pknca_no_intervals"),
    class = "pknca_no_conc_data"),
    class = "pknca_all_warnings_no_results"
  )
  expect_equal(
    nrow(as.data.frame(myresult)),
    0,
    info="No rows were generated when no intervals applied"
  )
})

test_that("Volume-related calculations", {
  tmpconc <- generate.conc(2, 1, c(4, 12, 24))
  tmpconc$conc <- seq_len(nrow(tmpconc))
  tmpconc$vol <- 2
  tmpdose <- generate.dose(tmpconc)
  myconc <- PKNCAconc(tmpconc, formula=conc~time|treatment+ID, volume="vol")
  mydose <- PKNCAdose(tmpdose, formula=dose~time|treatment+ID)
  mydata <-  PKNCAdata(myconc, mydose,
                       intervals=data.frame(treatment="Trt 1", start=0, end=24,
                                            ae=TRUE, fe=TRUE,
                                            stringsAsFactors=FALSE))
  myresult <- pk.nca(mydata)
  expect_equal(as.data.frame(myresult)[["PPORRES"]], c(12, 12, 30, 30),
              info="ae and fe are correctly calculated")
  tmpdose2 <- tmpdose
  tmpdose2$dose <- 2
  mydose2 <- PKNCAdose(tmpdose2, formula=dose~time|treatment+ID)
  mydata2 <-  PKNCAdata(myconc, mydose2,
                       intervals=data.frame(treatment="Trt 1", start=0, end=24,
                                            ae=TRUE, fe=TRUE,
                                            stringsAsFactors=FALSE))
  myresult2 <- pk.nca(mydata2)
  expect_equal(as.data.frame(myresult2)[["PPORRES"]], c(12, 6, 30, 15),
               info="fe respects dose")
})

test_that("pk.nca can calculate values with group-level data", {
  tmpconc_impute <- generate.conc(2, 1, 0:24)
  # This is what will happen in the imputation
  tmpconc_observe_05 <- tmpconc_impute[tmpconc_impute$time %in% 0,]
  tmpconc_observe_05$time <- 0.5
  tmpconc_observe <- rbind(tmpconc_impute, tmpconc_observe_05)
  tmpconc_observe <- tmpconc_observe[order(tmpconc_observe$treatment, tmpconc_observe$ID, tmpconc_observe$time),]
  tmpdose <- generate.dose(tmpconc_impute)
  tmpdose$time <- 0.5

  myconc_impute <- PKNCAconc(tmpconc_impute, formula=conc~time|treatment+ID)
  myconc_observe <- PKNCAconc(tmpconc_observe, formula=conc~time|treatment+ID)
  mydose <- PKNCAdose(tmpdose, formula=dose~time|treatment+ID)
  mydata_impute <-
    PKNCAdata(myconc_impute, mydose,
              intervals=data.frame(treatment="Trt 1", start=0, end=24,
                                   aucint.last.dose=TRUE,
                                   stringsAsFactors=FALSE))
  mydata_observe <-
    PKNCAdata(myconc_observe, mydose,
              intervals=data.frame(treatment="Trt 1", start=0, end=24,
                                   auclast=TRUE,
                                   stringsAsFactors=FALSE))
  myres_impute <- pk.nca(mydata_impute)
  myres_observe <- pk.nca(mydata_observe)
  expect_equal(as.data.frame(myres_impute)$PPORRES,
               as.data.frame(myres_observe)$PPORRES,
               info="Manually imputing values gives the same result as aucint")
})

test_that("Missing dose info for some subjects gives a warning, not a difficult-to-interpret error", {
  tmpconc <- generate.conc(2, 1, 0:24)
  tmpdose <- generate.dose(tmpconc)[1,]
  myconc <- PKNCAconc(tmpconc, formula=conc~time|treatment+ID)
  mydose <- PKNCAdose(tmpdose, formula=dose~time|treatment+ID)
  mydata <-  PKNCAdata(myconc, mydose,
                       intervals=data.frame(start=0, end=24,
                                            cl.last=TRUE))
  myresult <- pk.nca(mydata)
  expect_true(all(is.na(myresult$result[["PPORRES"]]) == c(FALSE, FALSE, FALSE, TRUE)) &
                all(myresult$result[["PPTESTCD"]] == rep(c("auclast", "cl.last"), 2)),
              info="cl.last is not calculated when dose information is missing, but only for the subject where dose info is missing.")
})

# Fix issue #68
test_that("Ensure that options are respected during pk.nca call", {
  doses <- data.frame(ID=1:2, Time=0, Dose=0.5)

  conc.data <- c(0, 1, 2, 1.3, 0.4, 0.35, 0.125)
  time.data <- c(0, 1, 2, 4,   8,   24,   48)
  concs <- merge(doses["ID"], data.frame(Conc=conc.data, Time=time.data))

  myconc <- PKNCA::PKNCAconc(concs, formula=Conc~Time|ID)
  mydose <- PKNCA::PKNCAdose(doses, formula=Dose~Time|ID)

  myintervals <- data.frame(start=c(0,0,0),
                            end=c(24,48,Inf),
                            auclast=TRUE,
                            aucinf.obs=TRUE,
                            aucinf.pred=TRUE,
                            aumclast=TRUE,
                            aumcall=TRUE,
                            half.life=TRUE)

  linear.mydata <- PKNCA::PKNCAdata(myconc, mydose, intervals = myintervals,
                                    options = list(auc.method = "linear"))
  linear.results <- PKNCA::pk.nca(linear.mydata)

  linlog.mydata <- PKNCA::PKNCAdata(myconc, mydose, intervals = myintervals,
                                    options = list(auc.method = "lin up/log down"))
  linlog.results <- PKNCA::pk.nca(linlog.mydata)
  expect_true(
    all.equal(
      linear.results$result$PPORRES[linear.results$result$PPTESTCD %in% "aucinf.obs" &
                                      linear.results$result$ID %in% 1 &
                                      linear.results$result$end %in% Inf],
      24.54319,
      tolerance=0.0001) &
      all.equal(
        linlog.results$result$PPORRES[linlog.results$result$PPTESTCD %in% "aucinf.obs" &
                                        linlog.results$result$ID %in% 1 &
                                        linlog.results$result$end %in% Inf],
        23.68317,
        tolerance=0.0001),
    info="linear and loglinear effects are calculated differently."
  )
})

test_that("Can calculate parameters requiring extra arguments", {
  o_conc <- PKNCAconc(conc~time, data=data.frame(conc=c(1:3, 2:1), time=0:4))
  d_intervals <- data.frame(start=0, end=4, time_above=TRUE, conc_above=2)
  o_data <- PKNCAdata(o_conc, intervals=d_intervals, options=list(auc.method="linear"))
  o_nca <- suppressMessages(pk.nca(o_data))
  expect_equal(as.data.frame(o_nca)$PPORRES, 2)
})

test_that("calculate with sparse data", {
  d_sparse <-
    data.frame(
      id = c(1L, 2L, 3L, 1L, 2L, 3L, 1L, 2L, 3L, 4L, 5L, 6L, 4L, 5L, 6L, 7L, 8L, 9L, 7L, 8L, 9L),
      conc = c(0, 0, 0,  1.75, 2.2, 1.58, 4.63, 2.99, 1.52, 3.03, 1.98, 2.22, 3.34, 1.3, 1.22, 3.54, 2.84, 2.55, 0.3, 0.0421, 0.231),
      time = c(0, 0, 0, 1, 1, 1, 6, 6, 6, 2, 2, 2, 10, 10, 10, 4, 4, 4, 24, 24, 24),
      dose = c(100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100)
    )
  o_conc_sparse <- PKNCAconc(d_sparse, conc~time|id, sparse=TRUE)

  d_intervals <-
    data.frame(
      start=0,
      end=24,
      aucinf.obs=TRUE,
      cmax=TRUE,
      sparse_auclast=TRUE
    )
  o_data_sparse <- PKNCAdata(o_conc_sparse, intervals=d_intervals)
  suppressMessages(
    expect_warning(expect_warning(
      o_nca <- pk.nca(o_data_sparse),
      class = "pknca_sparse_df_multi"),
      class = "pknca_halflife_too_few_points"
    )
  )
  df_result <- as.data.frame(o_nca)
  expect_true("sparse_auclast" %in% df_result$PPTESTCD)
  expect_equal(df_result$PPORRES[df_result$PPTESTCD %in% "sparse_auclast"], 39.4689)
  sum_o_nca <- summary(o_nca)
  expect_s3_class(sum_o_nca, "summary_PKNCAresults")

  # Mixed sparse and dense calculations when only one type is requested in an
  # interval works The example below has dense-only; sparse and dense; and and
  # sparse-only.
  d_intervals_mixed <-
    data.frame(
      start=0,
      end=c(23, 24, 25),
      cmax=c(TRUE, TRUE, FALSE),
      sparse_auclast=c(FALSE, TRUE, TRUE)
    )
  o_data_sparse_mixed <- PKNCAdata(o_conc_sparse, intervals=d_intervals_mixed)
  suppressMessages(
    expect_warning(expect_warning(
      o_nca_sparse_mixed <- pk.nca(o_data_sparse_mixed),
      class = "pknca_sparse_df_multi"),
      class = "pknca_sparse_df_multi"
    )
  )
  df_result_sparse_mixed <- as.data.frame(o_nca_sparse_mixed)
  expect_true("sparse_auclast" %in% df_result_sparse_mixed$PPTESTCD)
  expect_equal(df_result_sparse_mixed$PPORRES[df_result_sparse_mixed$PPTESTCD %in% "sparse_auclast"], rep(39.4689, 2))
  suppressMessages(
    expect_message(
      expect_warning(expect_warning(
        o_nca_sparse_mixed <- pk.nca(o_data_sparse_mixed, verbose=TRUE),
        class = "pknca_sparse_df_multi"),
        class = "pknca_sparse_df_multi"
      ),
      regexp="No sparse calculations requested for an interval"
    )
  )

  # Sparse data with multiple treatments, confirm the correct number of rows of
  # outputs are created.
  d_sparse_200 <- d_sparse
  d_sparse_200$dose <- 200
  d_sparse_multi_trt <- rbind(d_sparse, d_sparse_200)
  d_sparse_multi_trt$dose_grp <- d_sparse_multi_trt$dose
  o_conc_sparse_multi_trt <- PKNCAconc(d_sparse_multi_trt, conc~time|dose_grp+id, sparse=TRUE)
  d_intervals_mixed <-
    data.frame(
      start=0,
      end=c(23, 24, 25),
      cmax=c(TRUE, TRUE, FALSE),
      sparse_auclast=c(FALSE, TRUE, TRUE)
    )
  d_dose_sparse_multi_trt <- unique(d_sparse_multi_trt[, c("id", "dose")])
  d_dose_sparse_multi_trt$time <- 0
  d_dose_sparse_multi_trt$dose_grp <- d_dose_sparse_multi_trt$dose
  o_dose_sparse_multi_trt <- PKNCAdose(d_dose_sparse_multi_trt, dose~time|dose_grp+id)
  o_data_sparse_multi_trt <- PKNCAdata(o_conc_sparse_multi_trt, o_dose_sparse_multi_trt, intervals=d_intervals_mixed)
  suppressMessages(
    expect_warning(expect_warning(expect_warning(expect_warning(
      o_nca_sparse_multi_trt <- pk.nca(o_data_sparse_multi_trt),
      class = "pknca_sparse_df_multi"),
      class = "pknca_sparse_df_multi"),
      class = "pknca_sparse_df_multi"),
      class = "pknca_sparse_df_multi"
    )
  )
  expect_equal(nrow(as.data.frame(o_nca_sparse_multi_trt)), 16)

  # Correct detection of mixed doses within a sparse dose group when there are
  # no grouping columns other than subject
  d_sparse_multi_trt_bad_dose_single <- d_sparse_multi_trt[d_sparse_multi_trt$dose == 100, ]
  d_dose_sparse_multi_trt_bad_dose_single <- unique(d_sparse_multi_trt_bad_dose_single[, c("id", "dose")])
  d_dose_sparse_multi_trt_bad_dose_single$time <- 0
  d_dose_sparse_multi_trt_bad_dose_single$dose[1] <- d_dose_sparse_multi_trt_bad_dose_single$dose[1] + 1
  o_conc_sparse_multi_trt_bad_dose_single <- PKNCAconc(d_sparse_multi_trt_bad_dose_single, conc~time|id, sparse=TRUE)
  o_dose_sparse_multi_trt_bad_dose_single <- PKNCAdose(d_dose_sparse_multi_trt_bad_dose_single, dose~time|id)
  o_data_sparse_multi_trt_bad_dose_single <- PKNCAdata(o_conc_sparse_multi_trt_bad_dose_single, o_dose_sparse_multi_trt_bad_dose_single, intervals=d_intervals_mixed)
  expect_error(
    pk.nca(o_data_sparse_multi_trt_bad_dose_single),
    regexp="With sparse PK, all subjects in a group must have the same dosing information.*Not all subjects have the same dosing information"
  )

  # Correct detection of mixed doses within a sparse dose group
  d_dose_sparse_multi_trt_bad_dose <- d_dose_sparse_multi_trt
  d_dose_sparse_multi_trt_bad_dose$dose[1] <- d_dose_sparse_multi_trt$dose[1] + 1
  o_dose_sparse_multi_trt_bad_dose <- PKNCAdose(d_dose_sparse_multi_trt_bad_dose, dose~time|dose_grp+id)
  o_data_sparse_multi_trt_bad_dose <- PKNCAdata(o_conc_sparse_multi_trt, o_dose_sparse_multi_trt_bad_dose, intervals=d_intervals_mixed)
  expect_error(
    pk.nca(o_data_sparse_multi_trt_bad_dose),
    regexp="With sparse PK, all subjects in a group must have the same dosing information.*Not all subjects have the same dosing information for this group: +dose_grp=100"
  )
  # Correct detection of mixed doses within a sparse dose group when there are no groups
})

test_that("Unexpected interval columns now not cause an error (#238)", {
  d_conc <-
    data.frame(
      ID = 1L,
      time = 0:6,
      conc = c(0, 0.7, 0.71, 0.85, 1, 0.76, 0.74)
    )
  d_dose <- data.frame(dose = 1)
  d_intervals <- data.frame(start = 0, end = 6, cmax = TRUE, aucinf = TRUE)
  o_conc <- PKNCAconc(d_conc, formula = conc~time|ID)
  o_dose <- PKNCAdose(d_dose, formula = dose~.)
  expect_error(PKNCAdata(o_conc, o_dose, intervals = d_intervals),
               "The following columns in 'intervals' are not allowed:"
  )
})

test_that("aucint works within pk.calc.all for all zero concentrations with interpolated or extrapolated concentrations", {
  # AUCint.inf.obs
  d_interval <- data.frame(start = 0, end = 4, aucint.inf.obs = TRUE)
  d_conctime <- data.frame(conc = c(0, 0, 0, 0), time = 0:3)
  o_conc <- PKNCAconc(d_conctime, conc~time)
  o_data <- PKNCAdata(o_conc, intervals = d_interval)
  suppressWarnings(suppressMessages(
    o_nca <- pk.nca(o_data)
  ))
  results <- setNames(as.data.frame(o_nca)$PPORRES, nm = as.data.frame(o_nca)$PPTESTCD)
  zero_names <- c("clast.obs", "aucint.inf.obs")
  na_names <- setdiff(names(results), zero_names)
  expect_equal(
    results[zero_names],
    setNames(rep(0, length(zero_names)), zero_names)
  )
  expect_equal(
    results[na_names],
    setNames(rep(NA_real_, length(na_names)), na_names)
  )

  # AUCint.inf.pred
  d_interval <- data.frame(start = 0, end = 4, aucint.inf.pred = TRUE)
  d_conctime <- data.frame(conc = c(0, 0, 0, 0), time = 0:3)
  o_conc <- PKNCAconc(d_conctime, conc~time)
  o_data <- PKNCAdata(o_conc, intervals = d_interval)
  suppressWarnings(suppressMessages(
    o_nca <- pk.nca(o_data)
  ))
  expect_equal(
    as.data.frame(o_nca)$PPORRES,
    c(rep(NA_real_, 11), 0)
  )
})

test_that("The option keep_interval_cols is respected", {
  d_interval <- data.frame(start = 0, end = 4, cmax = TRUE, foo = "A")
  d_conctime <- data.frame(conc = c(0, 0, 0, 0), time = 0:3)
  o_conc <- PKNCAconc(d_conctime, conc~time)
 expect_error(PKNCAdata(o_conc, intervals = d_interval),
              "The following columns in 'intervals' are not allowed:")

  o_data <- PKNCAdata(o_conc, intervals = d_interval, options = list(keep_interval_cols = "foo"))
  suppressWarnings(suppressMessages(
    o_nca <- pk.nca(o_data)
  ))
  expect_equal(o_nca$result$foo, "A")
  expect_true("foo" %in% names(summary(o_nca)))
})

test_that("dose is calculable", {
  tmpconc <- generate.conc(2, 1, 0:24)
  tmpdose <- generate.dose(tmpconc)
  myconc <- PKNCAconc(tmpconc, formula=conc~time|treatment+ID)
  mydose <- PKNCAdose(tmpdose, formula=dose~time|treatment+ID)
  mydata <- PKNCAdata(myconc, mydose, intervals = data.frame(start = 0, end = Inf, totdose = TRUE))
  myresult <- pk.nca(mydata)

  # One dose in the interval
  expect_equal(as.data.frame(myresult)$PPORRES, rep(1, 2))
  expect_equal(as.data.frame(myresult)$PPTESTCD, rep("totdose", 2))

  # Don't give dose data
  mydata <- PKNCAdata(myconc, intervals = data.frame(start = 0, end = Inf, totdose = TRUE))
  suppressMessages(myresult <- pk.nca(mydata))
  expect_equal(as.data.frame(myresult)$PPORRES, rep(NA_real_, 2))
  expect_equal(as.data.frame(myresult)$PPTESTCD, rep("totdose", 2))

  # Multiple doses in the interval
  tmpdose_second <- tmpdose
  tmpdose_second$time <- 1
  mydose <- PKNCAdose(rbind(tmpdose, tmpdose_second), formula=dose~time|treatment+ID)
  mydata <- PKNCAdata(myconc, mydose, intervals = data.frame(start = 0, end = Inf, totdose = TRUE))
  myresult <- pk.nca(mydata)
  expect_equal(as.data.frame(myresult)$PPORRES, rep(2, 2))
  expect_equal(as.data.frame(myresult)$PPTESTCD, rep("totdose", 2))
})

test_that("do not give rbind error when interval columns have attributes (#381)", {
  o_conc <- PKNCAconc(data = data.frame(conc = 1, time = 0), conc~time)
  d_interval <- data.frame(start = 0, end = Inf, cmax = TRUE)

  d_interval <- data.frame(start = 0, end = Inf, cmax = TRUE, tmax = TRUE)
  attr(d_interval$start, "label") <- "start"
  o_data <- PKNCAdata(o_conc, intervals = d_interval)
  suppressMessages(o_nca <- pk.nca(o_data))
  # interval attributes are preserved
  expect_equal(
    attributes(as.data.frame(o_nca)$start),
    list(label = "start")
  )
})

test_that("PKNCAdata", {
  tmp.conc <- generate.conc(nsub=5, ntreat=2, time.points=0:24)
  tmp.conc.analyte <- generate.conc(nsub=5, ntreat=2, time.points=0:24,
                                    nanalytes=2)
  tmp.conc.study <- generate.conc(nsub=5, ntreat=2, time.points=0:24,
                                  nstudies=2)
  tmp.conc.analyte.study <- generate.conc(nsub=5, ntreat=2, time.points=0:24,
                                          nanalytes=2, nstudies=2)
  tmp.dose <- generate.dose(tmp.conc)
  tmp.dose.analyte <- generate.dose(tmp.conc.analyte)
  tmp.dose.study <- generate.dose(tmp.conc.study)
  tmp.dose.analyte.study <- generate.dose(tmp.conc.analyte.study)
  obj.conc <-
    PKNCAconc(tmp.conc, formula=conc~time|treatment+ID)
  obj.conc.analyte <-
    PKNCAconc(tmp.conc.analyte,
              formula=conc~time|treatment+ID/analyte)
  obj.conc.study <-
    PKNCAconc(tmp.conc.study,
              formula=conc~time|study+treatment+ID)
  obj.conc.analyte.study <-
    PKNCAconc(tmp.conc.analyte.study,
              formula=conc~time|study+treatment+ID/analyte)

  obj.dose <- PKNCAdose(tmp.dose, formula=dose~time|treatment+ID)
  obj.dose.analyte <- PKNCAdose(tmp.dose.analyte, formula=dose~time|treatment+ID)
  obj.dose.study <- PKNCAdose(tmp.dose.study, formula=dose~time|study+treatment+ID)
  obj.dose.analyte.study <- PKNCAdose(tmp.dose.analyte.study, formula=dose~time|study+treatment+ID)

  expect_equal(PKNCAdata(obj.conc, obj.dose),
               PKNCAdata(obj.dose, obj.conc),
               info="Input arguments are reversible")
  expect_equal(PKNCAdata(obj.conc.analyte, obj.dose),
               PKNCAdata(obj.dose, obj.conc.analyte),
               info="Combination of dose and analyte works")
  expect_equal(PKNCAdata(data.conc=tmp.conc, formula.conc=conc~time|treatment+ID,
                         data.dose=tmp.dose, formula.dose=dose~time|treatment+ID),
               PKNCAdata(obj.conc, obj.dose),
               info="Concentration and dose data can be created on the fly")

  # Input checking
  expect_error(PKNCAdata(obj.conc, obj.dose, options="a"),
               regexp="options must be a list.",
               info="Option class")
  expect_error(PKNCAdata(obj.conc, obj.dose, options=list(1)),
               regexp="options must have names.",
               info="Option structure")
  expect_error(PKNCAdata(obj.conc, obj.dose, options=list(foo=1)),
               regexp="Invalid setting for PKNCA.*foo",
               info="Option names")

  # Single dose AUCs are appropriately selected
  expect_equal(
    PKNCAdata(obj.conc, obj.dose),
    {
      tmp.intervals <- tibble::as_tibble(merge(PKNCA.options("single.dose.aucs"), tmp.dose))
      tmp.intervals <- tmp.intervals[order(tmp.intervals$treatment, tmp.intervals$ID),]
      tmp.intervals$time <- NULL
      tmp.intervals$dose <- NULL
      tmp <- list(
        conc=obj.conc,
        dose=obj.dose,
        options=list(),
        intervals=tmp.intervals,
        impute=NA_character_
      )
      class(tmp) <- c("PKNCAdata", "list")
      tmp
    },
    ignore_attr=FALSE,
    info="Selection of single dose AUCs"
  )

  tmp.conc <- generate.conc(nsub=5, ntreat=2, time.points=0:24)
  tmp.dose <- generate.dose(tmp.conc)
  tmp.conc <- tmp.conc[!(tmp.conc$ID %in% 1),]
  obj.conc <-
    PKNCAconc(tmp.conc, formula=conc~time|treatment+ID)
  obj.dose <- PKNCAdose(tmp.dose, formula=dose~time|treatment+ID)
  expect_warning(expect_warning(
    PKNCAdata(obj.conc, obj.dose),
    class = "pknca_no_intervals_generated"),
    class = "pknca_no_intervals_generated",
    info="Missing concentration data with dose data gives a warning."
  )

  expect_warning(expect_warning(expect_warning(
    PKNCAdata(obj.conc, obj.dose, formula.conc=a~b),
    class = "pknca_dataconc_formulaconc"),
    class = "pknca_no_intervals_generated"),
    class = "pknca_no_intervals_generated"
  )
  expect_warning(expect_warning(expect_warning(
    PKNCAdata(obj.conc, obj.dose, formula.dose=a~b),
    class = "pknca_dataconc_formuladose"),
    class = "pknca_no_intervals_generated"),
    class = "pknca_no_intervals_generated"
  )
})

test_that("PKNCAdata with no or limited dose information", {
  tmp.conc <- generate.conc(nsub=5, ntreat=2, time.points=0:24)
  obj.conc <-
    PKNCAconc(tmp.conc, formula=conc~time|treatment+ID)

  expect_error(PKNCAdata(obj.conc),
               regexp="If data.dose is not given, intervals must be given",
               info="One of dose and intervals is required (no dose)")
  expect_error(PKNCAdata(obj.conc, data.dose=NA),
               regexp="If data.dose is not given, intervals must be given",
               info="One of dose and intervals is required (NA dose)")
  expect_equal(
    PKNCAdata(obj.conc, intervals=data.frame(start=0, end=24, aucinf.obs=TRUE)),
    {
      tmp <-
        list(
          conc=obj.conc,
          dose=NA,
          options=list(),
          intervals=check.interval.specification(
            data.frame(start=0, end=24, aucinf.obs=TRUE)),
          impute=NA_character_
        )
      class(tmp) <- c("PKNCAdata", "list")
      tmp
    }
  )

  tmp.conc <- generate.conc(nsub=5, ntreat=2, time.points=0:24)
  tmp.dose <- generate.dose(tmp.conc)
  obj.conc <-
    PKNCAconc(tmp.conc, formula=conc~time|treatment+ID)
  obj.dose <- PKNCAdose(tmp.dose, formula=dose~.|treatment+ID)
  expect_error(PKNCAdata(obj.conc, obj.dose),
               regexp="Dose times were not given, so intervals must be manually specified.",
               info="No dose times requires intervals.")
})

test_that("print.PKNCAdata", {
  tmp.conc <- generate.conc(nsub=2, ntreat=2, time.points=0:24)
  tmp.dose <- generate.dose(tmp.conc)
  obj.conc <- PKNCAconc(tmp.conc, formula=conc~time|treatment+ID)
  obj.dose <- PKNCAdose(tmp.dose, formula=dose~time|treatment+ID)
  obj.data.nodose <- PKNCAdata(obj.conc,
                               intervals=data.frame(start=0, end=24, aucinf.obs=TRUE))
  obj.data.nodose.opt <-
    PKNCAdata(obj.conc,
              intervals=data.frame(start=0, end=24, aucinf.obs=TRUE),
              options=list(min.hl.r.squared=0.95))
  obj.data.dose <- PKNCAdata(obj.conc, data.dose=obj.dose)
  obj.data.units <- PKNCAdata(obj.conc,
                               intervals=data.frame(start=0, end=24, aucinf.obs=TRUE))
  obj.data.units$units <- "mg"

  expect_output(print.PKNCAdata(obj.data.nodose),
                regexp="Formula for concentration:
 conc ~ time | treatment + ID
With 2 subjects defined in the 'ID' column.
Nominal time column is not specified.

First 6 rows of concentration data:
 treatment ID time      conc exclude
     Trt 1  1    0 0.0000000    <NA>
     Trt 1  1    1 0.7052248    <NA>
     Trt 1  1    2 0.7144320    <NA>
     Trt 1  1    3 0.8596094    <NA>
     Trt 1  1    4 0.9998126    <NA>
     Trt 1  1    5 0.7651474    <NA>
No dosing information.

With 1 rows of interval specifications.
No options are set differently than default.",
                info="Generic print.PKNCAdata works with no dosing")
  expect_output(print.PKNCAdata(obj.data.dose),
                regexp="Formula for concentration:
 conc ~ time | treatment + ID
With 2 subjects defined in the 'ID' column.
Nominal time column is not specified.

First 6 rows of concentration data:
 treatment ID time      conc exclude
     Trt 1  1    0 0.0000000    <NA>
     Trt 1  1    1 0.7052248    <NA>
     Trt 1  1    2 0.7144320    <NA>
     Trt 1  1    3 0.8596094    <NA>
     Trt 1  1    4 0.9998126    <NA>
     Trt 1  1    5 0.7651474    <NA>
Formula for dosing:
 dose ~ time | treatment + ID
Nominal time column is not specified.

Data for dosing:
 treatment ID dose time exclude         route duration
     Trt 1  1    1    0    <NA> extravascular        0
     Trt 1  2    1    0    <NA> extravascular        0
     Trt 2  1    2    0    <NA> extravascular        0
     Trt 2  2    2    0    <NA> extravascular        0
With 1 rows of interval specifications.
No options are set differently than default.",
                info="Generic print.PKNCAdata works with dosing")

  expect_output(print.PKNCAdata(obj.data.nodose.opt),
                regexp="Formula for concentration:
 conc ~ time | treatment + ID
With 2 subjects defined in the 'ID' column.
Nominal time column is not specified.

First 6 rows of concentration data:
 treatment ID time      conc exclude
     Trt 1  1    0 0.0000000    <NA>
     Trt 1  1    1 0.7052248    <NA>
     Trt 1  1    2 0.7144320    <NA>
     Trt 1  1    3 0.8596094    <NA>
     Trt 1  1    4 0.9998126    <NA>
     Trt 1  1    5 0.7651474    <NA>
No dosing information.

With 1 rows of interval specifications.
Options changed from default are:
$min.hl.r.squared
[1] 0.95",
                info="Generic print.PKNCAdata works with no dosing and with options changed")
  expect_output(print.PKNCAdata(obj.data.units),
                regexp="Formula for concentration:
 conc ~ time | treatment + ID
With 2 subjects defined in the 'ID' column.
Nominal time column is not specified.

First 6 rows of concentration data:
 treatment ID time      conc exclude
     Trt 1  1    0 0.0000000    <NA>
     Trt 1  1    1 0.7052248    <NA>
     Trt 1  1    2 0.7144320    <NA>
     Trt 1  1    3 0.8596094    <NA>
     Trt 1  1    4 0.9998126    <NA>
     Trt 1  1    5 0.7651474    <NA>
No dosing information.
With units
With 1 rows of interval specifications.
No options are set differently than default.",
                info="Generic print.PKNCAdata works with no dosing")
})

test_that("summary.PKNCAdata", {
  tmp.conc <- generate.conc(nsub=2, ntreat=2, time.points=0:24)
  tmp.dose <- generate.dose(tmp.conc)
  obj.conc <- PKNCAconc(tmp.conc, formula=conc~time|treatment+ID)
  obj.dose <- PKNCAdose(tmp.dose, formula=dose~time|treatment+ID)
  obj.data.nodose <- PKNCAdata(obj.conc,
                               intervals=data.frame(start=0, end=24, aucinf.obs=TRUE))

  expect_output(summary(obj.data.nodose),
                regexp="Formula for concentration:
 conc ~ time | treatment + ID
With 2 subjects defined in the 'ID' column.
Nominal time column is not specified.

Group summary:
 Group Name Count
  treatment     2
         ID     4

First 6 rows of concentration data:
 treatment ID time      conc exclude
     Trt 1  1    0 0.0000000    <NA>
     Trt 1  1    1 0.7052248    <NA>
     Trt 1  1    2 0.7144320    <NA>
     Trt 1  1    3 0.8596094    <NA>
     Trt 1  1    4 0.9998126    <NA>
     Trt 1  1    5 0.7651474    <NA>
No dosing information.

With 1 rows of interval specifications.
No options are set differently than default.",
                info="Generic summary.PKNCAdata works.")
})

test_that("no intervals auto-determined (Fix GitHub issue #84)", {
  tmp_conc <-
    data.frame(
      Subject=1,
      Treatment=c(1, rep(2, 6)),
      Time=c(0, 1:6),
      Conc=1
    )
  tmp_dose <-
    data.frame(
      Subject=1,
      Treatment=c(1, 2),
      Time=c(0, 1),
      Dose=1
    )

  interval_1 <- PKNCA.options("single.dose.aucs")[c(1:2, 1:2),]
  interval_1$start <- rep(0:1, each=2)
  interval_1$end <- c(interval_1$end[1:2], interval_1$end[3:4] + 1)
  interval_1 <- cbind(interval_1, data.frame(Treatment=rep(1:2, each=2), Subject=1))
  two_single_dose_treatments <-
    PKNCAdata(
      PKNCAconc(data=tmp_conc, Conc~Time|Treatment+Subject),
      PKNCAdose(data=tmp_dose, Dose~Time|Treatment+Subject)
    )
  expect_equal(
    two_single_dose_treatments$intervals,
    interval_1,
    ignore_attr=TRUE
  )
  interval_2 <-
    check.interval.specification(
      tibble::tibble(
        start=1, end=c(2, Inf),
        auclast=c(TRUE, FALSE),
        cmax=c(TRUE, FALSE),
        tmax=c(TRUE, FALSE),
        half.life=c(FALSE, TRUE),
        Treatment=2,
        Subject=1
      )
    )
  expect_warning(
    two_multiple_dose_treatments <-
      PKNCAdata(
        PKNCAconc(data=tmp_conc, Conc~Time|Treatment+Subject),
        PKNCAdose(data=tmp_dose, Dose~Time|Subject)
      ),
    regexp="No intervals generated"
  )
  expect_equal(
    two_multiple_dose_treatments$intervals,
    interval_2
  )
})

test_that("Ensure that unexpected arguments to PKNCAdata give an error (related to issue #83)", {
  tmp.conc <- generate.conc(nsub=2, ntreat=1, time.points=0:24)
  tmp.dose <- generate.dose(tmp.conc)
  obj.conc <-
    PKNCAconc(tmp.conc, formula=conc~time|treatment+ID)
  obj.dose <-
    PKNCAdose(tmp.dose, formula=dose~time|treatment+ID)
  expect_error(mydata <- PKNCAdata(obj.conc, obj.dose, 1),
               regexp="Unknown argument")
})

test_that("intervals may be a tibble", {
  tmp.conc <- generate.conc(nsub=2, ntreat=1, time.points=0:24)
  tmp.dose <- generate.dose(tmp.conc)
  obj.conc <-
    PKNCAconc(tmp.conc, formula=conc~time|treatment+ID)
  obj.dose <-
    PKNCAdose(tmp.dose, formula=dose~time|treatment+ID)
  intervals <- data.frame(start=0, end=24, aucinf.obs=TRUE)
  mydata_tibble <- PKNCAdata(obj.conc, obj.dose, intervals=dplyr::as_tibble(intervals))
  mydata <- PKNCAdata(obj.conc, obj.dose, intervals=intervals)
  expect_equal(
    as.data.frame(pk.nca(mydata_tibble)),
    as.data.frame(pk.nca(mydata))
  )
})

test_that("PKNCAdata units (#336)", {
  # Typical use
  d_conc <- data.frame(conc = 1, time = 0, concu_x = "A", timeu_x = "B", amountu_x = "C")
  d_dose <- data.frame(dose = 1, time = 0, doseu_x = "D")

  o_conc <- PKNCAconc(data = d_conc, conc~time, concu = "concu_x", timeu = "timeu_x")
  o_dose <- PKNCAdose(data = d_dose, dose~time, doseu = "doseu_x")
  o_data <- PKNCAdata(o_conc, o_dose)
  expect_equal(
    o_data$units,
    pknca_units_table(concu = "A", doseu = "D", timeu = "B")
  )
  suppressWarnings(o_nca <- pk.nca(o_data))
  expect_true("Cmax (A)" %in% names(summary(o_nca)))

  # NA unit values are ignored
  d_conc <- data.frame(conc = 1, time = 0:1, concu_x = c("A", NA), timeu_x = "B", amountu_x = "C")
  d_dose <- data.frame(dose = 1, time = 0, doseu_x = "D")

  o_conc <- PKNCAconc(data = d_conc, conc~time, concu = "concu_x", timeu = "timeu_x")
  o_dose <- PKNCAdose(data = d_dose, dose~time, doseu = "doseu_x")
  o_data <- PKNCAdata(o_conc, o_dose)
  expect_equal(
    o_data$units,
    pknca_units_table(concu = "A", doseu = "D", timeu = "B")
  )
  suppressWarnings(o_nca <- pk.nca(o_data))
  expect_true("Cmax (A)" %in% names(summary(o_nca)))

  # multiple unit values cause an error
  d_conc <- data.frame(conc = 1, time = 0:1, concu_x = c("A", "C"), timeu_x = "B", amountu_x = "C")
  d_dose <- data.frame(dose = 1, time = 0, doseu_x = "B")

  o_conc <- PKNCAconc(data = d_conc, conc~time, concu = "concu_x")
  o_dose <- PKNCAdose(data = d_dose, dose~time, doseu = "doseu_x")
  expect_error(
    PKNCAdata(o_conc, o_dose),
    regexp = "Only one unit may be provided at a time: A, C"
  )
})

test_that("getGroups works", {
  # Check that it works with grouping [contains only the grouping column(s)]
  o_conc_group <- PKNCAconc(as.data.frame(datasets::Theoph), conc~Time|Subject)
  data_group <- as.data.frame(datasets::Theoph)
  expected_group <- data.frame(Subject = data_group$Subject)
  o_data_group <- PKNCAdata(o_conc_group, intervals = data.frame(start = 0, end = 1, cmax = TRUE))
  expect_equal(getGroups(o_data_group), expected_group)

  # Check that it works without groupings as expected [empty]
  o_conc_nongroup <- PKNCAconc(as.data.frame(datasets::Theoph)[datasets::Theoph$Subject == 1,], conc~Time)
  o_data_nogroup <- PKNCAdata(o_conc_nongroup, intervals = data.frame(start = 0, end = 1, cmax = TRUE))

  # It should be an empty data.frame with 11 rows
  expect_equal(getGroups(o_data_nogroup), data.frame(A = 1:11)[, -1])
})

test_that("group_vars.PKNCAdata", {
  o_conc_group <- PKNCAconc(as.data.frame(datasets::Theoph), conc~Time|Subject)
  o_data_group <- PKNCAdata(o_conc_group, intervals = data.frame(start = 0, end = 1, cmax = TRUE))

  expect_equal(dplyr::group_vars(o_data_group), "Subject")

  # Check that it works without groupings as expected [empty]
  o_conc_nongroup <- PKNCAconc(as.data.frame(datasets::Theoph)[datasets::Theoph$Subject == 1,], conc~Time)
  o_data_nogroup <- PKNCAdata(o_conc_nongroup, intervals = data.frame(start = 0, end = 1, cmax = TRUE))

  expect_equal(dplyr::group_vars(o_data_nogroup), character(0))
})

#' Combine PKNCAconc and PKNCAdose objects
#'
#' The function is inspired by `dplyr::full_join`, but it has different
#' semantics.
#'
#' @param o_conc a PKNCAconc object
#' @param o_dose a PKNCAdose object or `NA`
#' @param extra_cols_conc Additional columns to include in the concentration data
#' @returns A tibble with columns for the groups, "data_conc" (the concentration
#'   data), and "data_dose" (the dosing data).  If `is.na(o_dose)`, "data_dose"
#'   will be `NA`.
#' @family Combine PKNCA objects
#' @keywords Internal
#' @noRd
full_join_PKNCAconc_PKNCAdose <- function(o_conc, o_dose, extra_cols_conc = character()) {
  stopifnot(inherits(x=o_conc, what="PKNCAconc"))
  if (identical(o_dose, NA)) {
    message("No dose information provided, calculations requiring dose will return NA.")
    n_dose <- tibble::tibble(data_dose=list(NA))
  } else {
    stopifnot(inherits(x=o_dose, what="PKNCAdose"))
    n_dose <- prepare_PKNCAdose(o_dose, sparse=is_sparse_pk(o_conc), subject_col=o_conc$columns$subject)
  }
  n_conc <- prepare_PKNCAconc(o_conc, extra_cols = extra_cols_conc)
  shared_groups <- intersect(names(n_conc), names(n_dose))
  if (length(shared_groups) > 0) {
    dplyr::full_join(n_conc, n_dose, by=shared_groups)
  } else {
    tidyr::crossing(n_conc, n_dose)
  }
}

#' Convert a PKNCAdata object into a data.frame for analysis
#'
#' The function is inspired by `dplyr::full_join`, but it has different
#' semantics.
#'
#' @param x The PKNCAdata object
#' @inheritParams full_join_PKNCAconc_PKNCAdose
#' @returns A tibble with columns the grouping variables, "data_conc" for
#'   concentration data, "data_dose" for dosing data, and "data_intervals" for
#'   intervals data.
#' @family Combine PKNCA objects
#' @keywords Internal
#' @noRd
full_join_PKNCAdata <- function(x, extra_conc_cols = character()) {
  conc_dose <- full_join_PKNCAconc_PKNCAdose(o_conc = x$conc, o_dose = x$dose, extra_cols_conc = extra_conc_cols)
  n_i <-
    prepare_PKNCAintervals(
      .dat=x$intervals,
      vars=setdiff(names(conc_dose), c("data_conc", "data_dose", "data_intervals"))
    )
  shared_groups <- intersect(names(conc_dose), names(n_i))
  if (length(shared_groups) > 0) {
    ret <- dplyr::full_join(conc_dose, n_i, by=shared_groups)
  } else {
    ret <- tidyr::crossing(conc_dose, n_i)
  }
  ret
}

#' Prepare a PKNCA object and drop unnecessary columns
#'
#' @param .dat The PKNCA object to prepare as a nested tibble
#' @param ...,.names_sep,.key Ignored
#' @return A nested tibble with a column named "data_conc" containing the
#'   concentration data and a column
#' @family Combine PKNCA objects
#' @keywords Internal
#' @noRd
prepare_PKNCA_general <- function(.dat, cols, exclude, group_cols, data_name, insert_if_missing=list()) {
  check_reserved_column_names(.dat)
  intermediate_group_cols <-
    if (length(group_cols) > 0) {
      paste0("group", seq_along(group_cols))
    } else {
      character(0)
    }
  data_no_exclude <-
    as.data.frame(.dat[
      is.na(normalize_exclude(.dat[[exclude]])),,
      drop=FALSE
    ])
  data_standard <-
    standardize_column_names(
      x=data_no_exclude,
      cols=cols,
      group_cols=group_cols,
      insert_if_missing=insert_if_missing
    )
  # data_conc is used since it is reserved, and it will be replaced on the next
  # line.
  as_nest <- tidyr::nest(data_standard, data_conc=!dplyr::all_of(intermediate_group_cols))
  names(as_nest)[names(as_nest) %in% "data_conc"] <- data_name
  ret <- restore_group_col_names(as_nest, group_cols=group_cols)
  ret
}

prepare_PKNCAconc_dense <- function(.dat, needed_cols, group_cols_selected) {
  ret <-
    prepare_PKNCA_general(
      .dat=.dat$data,
      exclude=.dat$columns$exclude,
      cols=needed_cols,
      data_name="data_conc",
      group_cols=group_cols_selected
    )
  ret
}

prepare_PKNCAconc_sparse <- function(.dat, needed_cols, group_cols_selected) {
  needed_cols$subject <- .dat$columns$subject
  ret <-
    prepare_PKNCA_general(
      .dat=.dat$data_sparse,
      exclude=.dat$columns$exclude,
      cols=needed_cols,
      data_name="data_sparse_conc",
      group_cols=setdiff(group_cols_selected, .dat$columns$subject)
    )
  # Generate the mean profile for non-sparse parameters
  ret$data_conc <-
    ret$data_sparse_conc %>%
    lapply(FUN=as_sparse_pk) %>%
    lapply(FUN=sparse_mean) %>%
    lapply(FUN=sparse_to_dense_pk)
  ret
}

prepare_PKNCAconc <- function(.dat, extra_cols = character()) {
  # Remove rows to be excluded from all calculations
  # Drop unnecessary column names
  needed_cols <-
    list(
      conc=.dat$columns$concentration,
      time=.dat$columns$time,
      volume=.dat$columns$volume,
      duration=.dat$columns$duration,
      include_half.life=.dat$columns$include_half.life,
      exclude_half.life=.dat$columns$exclude_half.life
    )
  needed_cols <- append(needed_cols, stats::setNames(nm = extra_cols))
  data_name <- getDataName(.dat)
  group_cols_selected <- unlist(.dat$columns$groups)
  if (is_sparse_pk(.dat)) {
    ret <-
      prepare_PKNCAconc_sparse(
        .dat=.dat,
        needed_cols=needed_cols,
        group_cols_selected=group_cols_selected
      )
  } else if (data_name == "data") {
    ret <-
      prepare_PKNCAconc_dense(
        .dat=.dat,
        needed_cols=needed_cols,
        group_cols_selected=group_cols_selected
      )
  } else {
    stop("Please report this as a bug: Invalid data_name") # nocov
  }
  ret
}

#' @describeIn prepare_PKNCAconc Nest a PKNCAdose object
#'
#' @param sparse Is the data for sparse PK?
#' @param subject_col The column name indicating the subject identifier (to be
#'   dropped from groups with sparse PK)
#' @family Combine PKNCA objects
#' @keywords Internal
#' @noRd
prepare_PKNCAdose <- function(.dat, sparse, subject_col) {
  ret <- prepare_PKNCAdose_general(.dat)
  if (sparse && (length(subject_col) == 1) && (subject_col %in% names(ret))) {
    # Verify that all subjects in a group had the same data_dose and then drop
    # the subjects
    # ret_grp will have one column named "sparse_group_check" with one row per ID and all of the dosing information within a group and
    ret_grp <-
      tidyr::nest(
        ret,
        sparse_group_check=!setdiff(names(ret), c("data_dose", subject_col))
      )
    ret_grp$data_dose <- rep(list(NULL), nrow(ret_grp))
    for (current_row in seq_len(nrow(ret_grp))) {
      # Dosing information for all subjects are identical to the first subject
      all_match <-
        all(vapply(
          X=ret_grp$sparse_group_check[[current_row]]$data_dose,
          FUN=identical,
          y=ret_grp$sparse_group_check[[current_row]]$data_dose[[1]],
          FUN.VALUE = TRUE
        ))
      if (all_match) {
        # Drop the subject identifier from the dosing
        ret_grp$data_dose[[current_row]] <- ret_grp$sparse_group_check[[current_row]]$data_dose[[1]]
      } else {
        names_to_print <- setdiff(names(ret_grp), c("sparse_group_check", "data_dose"))
        msg_error_row <- paste(names_to_print, unlist(ret_grp[current_row, names_to_print]), sep="=", collapse="; ")
        msg_error <-
          if (length(names_to_print) > 0) {
            paste(
              "Not all subjects have the same dosing information for this group: ",
              msg_error_row
            )
          } else {
            "Not all subjects have the same dosing information."
          }
        stop(
          "With sparse PK, all subjects in a group must have the same dosing information.\n",
          msg_error
        )
      }
    }
    ret <- ret_grp[, setdiff(names(ret_grp), "sparse_group_check")]
  }
  ret
}

#' @describeIn prepare_PKNCAconc Nest a PKNCAdose object
#' @family Combine PKNCA objects
#' @keywords Internal
#' @noRd
prepare_PKNCAdose_general <- function(.dat) {
  dose_col <- .dat$columns$dose
  time_col <- .dat$columns$time
  if (length(dose_col) == 0) dose_col <- NULL
  if (length(time_col) == 0) time_col <- NULL
  needed_cols <-
    list(
      dose=dose_col,
      time=time_col,
      duration=.dat$columns$duration,
      route=.dat$columns$route
    )
  ret <-
    prepare_PKNCA_general(
      .dat=.dat$data,
      exclude=.dat$columns$exclude,
      cols=needed_cols,
      data_name="data_dose",
      group_cols=unlist(.dat$columns$groups),
      insert_if_missing=list(dose=NA, time=NA)
    )
  ret
}

#' @describeIn prepare_PKNCAconc Nest a PKNCAdose object
#' @noRd
#' @family Combine PKNCA objects
#' @keywords Internal
prepare_PKNCAintervals <- function(.dat, vars=character(0)) {
  check_reserved_column_names(.dat)
  .dat <- tibble::as_tibble(.dat)
  vars <- intersect(vars, names(.dat))
  if (length(vars) == 0) {
    as_nest <- tibble::tibble(data_intervals=list(.dat))
  } else {
    as_nest <- tidyr::nest(.dat, data_intervals=!dplyr::all_of(vars))
  }
  as_nest
}

#' Confirm that PKNCA reserved column names are not in a data.frame
#'
#' @param x A data.frame or similar object
#' @return NULL (generate an error if a reserved name is present)
#' @keywords Internal
#' @noRd
check_reserved_column_names <- function(x) {
  reserved_names <- c("data_conc", "data_dose", "data_intervals", "data_results")
  overlap <- intersect(reserved_names, names(x))
  if (length(overlap) > 0) {
    msg <-
      paste(
        ngettext(length(overlap), msg1="The column", msg2="The columns"),
        paste0("'", overlap, "'", collapse=", "),
        ngettext(length(overlap), msg1="is", msg2="are"),
        "reserved for internal use in PKNCA.  Change the",
        ngettext(length(overlap), msg1="name", msg2="names"),
        "and retry."
      )
    stop(msg)
  }
}

#' Standardize column names and drop unnecessary columns from a data.frame or
#' tibble
#'
#' @param x The data.frame or tibble
#' @param cols A named list where the names are the standardized column names
#'   and the values are the original column names
#' @returns A data.frame or tibble with columns cleaned of unlisted columns and
#'   with names set to the expected names.
#' @noRd
#' @keywords Internal
standardize_column_names <- function(x, cols, group_cols=NULL, insert_if_missing=list()) {
  stopifnot("cols must be a list"=is.list(cols))
  stopifnot("cols must be named"=!is.null(names(cols)))
  stopifnot("all cols must be named"=!any(names(cols) %in% ""))
  stopifnot("all original cols names must be names of x"=all(unlist(cols) %in% names(x)))
  stopifnot("group_cols must be NULL or a character vector"=is.null(group_cols) || is.character(group_cols))
  if (!is.null(group_cols) && (length(group_cols) > 0)) {
    # Give a clear error message if group columns overlap
    mask_overlap_colvalues <- group_cols %in% unlist(cols)
    mask_overlap_colnames <- group_cols %in% names(cols)
    if (any(mask_overlap_colvalues)) {
      stop(
        "group_cols must not overlap with other column names.  Change the name of the following groups: ",
        paste(group_cols[mask_overlap_colvalues], collapse=", ")
      )
    }
    if (any(mask_overlap_colnames)) {
      stop(
        "group_cols must not overlap with standardized column names.  Change the name of the following groups: ",
        paste(group_cols[mask_overlap_colnames], collapse=", ")
      )
    }
    new_group_cols <- paste0("group", seq_along(group_cols))
  } else {
    new_group_cols <- NULL
  }
  cols_clean <- cols[!vapply(X = cols, FUN = is.null, FUN.VALUE = TRUE)]
  ret <-
    stats::setNames(
      # Keep only columns of interest
      x[, c(group_cols, unlist(cols_clean)), drop=FALSE],
      nm=c(new_group_cols, names(cols_clean))
    )
  for (current_nm in names(insert_if_missing)) {
    if (!(current_nm %in% names(ret))) {
      ret[[current_nm]] <- insert_if_missing[[current_nm]]
    }
  }
  ret
}

restore_group_col_names <- function(x, group_cols=NULL) {
  if (is.null(group_cols) || (length(group_cols) == 0)) {
    return(x)
  }
  new_group_cols <- paste0("group", seq_along(group_cols))
  stopifnot("missing intermediate group_cols names"=all(new_group_cols %in% names(x)))
  stopifnot(
    "Intermediate group_cols are out of order"=
      all(names(x)[names(x) %in% new_group_cols] == new_group_cols)
  )
  names(x)[names(x) %in% new_group_cols] <- group_cols
  x
}

# prepare_* ####

test_that("prepare_*", {
  tmp_conc <- generate.conc(nsub=5, ntreat=2, time.points=0:24)
  tmp_dose <- generate.dose(tmp_conc)
  o_conc <- PKNCAconc(tmp_conc, formula=conc~time|treatment+ID)
  o_dose <- PKNCAdose(tmp_dose, formula=dose~time|treatment+ID)

  expect_equal(
    prepare_PKNCAconc(o_conc),
    tidyr::nest(
      dplyr::mutate(
        tmp_conc[, c("treatment", "ID", "conc", "time")],
        volume=NA_real_,
        duration=0
      ),
      data_conc=!c("treatment", "ID")
    )
  )
  expect_equal(
    prepare_PKNCAdose(o_dose, sparse=FALSE, subject_col=""),
    tidyr::nest(
      dplyr::mutate(
        tmp_dose,
        duration=0,
        route="extravascular"
      ),
      data_dose=!c("treatment", "ID")
    )
  )
  # No groups
  expect_equal(
    prepare_PKNCAintervals(.dat=PKNCA.options("single.dose.aucs")),
    tibble::tibble(
      data_intervals=list(tibble::as_tibble(PKNCA.options("single.dose.aucs")))
    )
  )
  # With groups
  tmp_intervals <- PKNCA.options("single.dose.aucs")
  tmp_intervals$g <- "A"
  expect_equal(
    prepare_PKNCAintervals(.dat=tmp_intervals, vars="g"),
    tibble::tibble(
      g="A",
      data_intervals=list(tibble::as_tibble(PKNCA.options("single.dose.aucs")))
    )
  )
})

# full_join for PKNCAconc, PKNCAdose, and PKNCAdata ####

test_that("full_join for PKNCAconc, PKNCAdose, and PKNCAdata", {
  tmp_conc <- generate.conc(nsub=5, ntreat=2, time.points=0:24)
  tmp_dose <- generate.dose(tmp_conc)
  o_conc <- PKNCAconc(tmp_conc, formula=conc~time|treatment+ID)
  o_dose <- PKNCAdose(tmp_dose, formula=dose~time|treatment+ID)
  o_data <- PKNCAdata(o_conc, o_dose)

  expect_equal(
    full_join_PKNCAconc_PKNCAdose(o_conc, o_dose),
    dplyr::full_join(
      prepare_PKNCAconc(o_conc),
      prepare_PKNCAdose(o_dose, sparse=FALSE, subject_col=""),
      by=c("treatment", "ID")
    )
  )
  expect_equal(
    full_join_PKNCAdata(o_data),
    tidyr::crossing(
      dplyr::full_join(
        prepare_PKNCAconc(o_conc),
        prepare_PKNCAdose(o_dose, sparse=FALSE, subject_col=""),
        by=c("treatment", "ID")
      ),
      data_intervals=list(tibble::as_tibble(PKNCA.options("single.dose.aucs")))
    )
  )
  # When intervals have no groups
  o_data_manual_interval <-
    PKNCAdata(
      o_conc,
      o_dose,
      intervals=PKNCA.options("single.dose.aucs")[1,]
    )
  expect_equal(
    full_join_PKNCAdata(o_data_manual_interval),
    tidyr::crossing(
      dplyr::full_join(
        prepare_PKNCAconc(o_conc),
        prepare_PKNCAdose(o_dose, sparse=FALSE, subject_col=""),
        by=c("treatment", "ID")
      ),
      data_intervals=list(tibble::as_tibble(PKNCA.options("single.dose.aucs")[1,]))
    )
  )
  # When dosing is not provided
  o_data_no_dose <- PKNCAdata(o_conc, intervals=PKNCA.options("single.dose.aucs")[1,])
  suppressMessages(
    expect_equal(
      full_join_PKNCAdata(o_data_no_dose),
      tidyr::crossing(
        prepare_PKNCAconc(o_conc),
        tibble::tibble(data_dose=list(NA)),
        data_intervals=list(tibble::as_tibble(PKNCA.options("single.dose.aucs")[1,]))
      )
    )
  )
})

test_that("check_reserved_column_names", {
  expect_null(
    check_reserved_column_names(data.frame())
  )
  expect_error(
    check_reserved_column_names(data.frame(data_dose=1)),
    regexp="The column 'data_dose' is reserved for internal use in PKNCA.  Change the name and retry.",
    fixed=TRUE
  )
  expect_error(
    check_reserved_column_names(data.frame(data_dose=1, data_conc=1, data_intervals=1)),
    regexp="The columns 'data_conc', 'data_dose', 'data_intervals' are reserved for internal use in PKNCA.  Change the names and retry.",
    fixed=TRUE
  )
})

# standardize_column_names ####

test_that("standardize_column_names", {
  # One column works
  expect_equal(
    standardize_column_names(data.frame(a=1), cols=list(b="a")),
    data.frame(b=1)
  )
  # Two columns work
  expect_equal(
    standardize_column_names(data.frame(a=1, b=2), cols=list(c="a", d="b")),
    data.frame(c=1, d=2)
  )
  # group_cols overlap with cols values fails
  expect_error(
    standardize_column_names(data.frame(a=1, b=2), cols=list(c="a", d="b"), group_cols="b"),
    regexp="group_cols must not overlap with other column names.  Change the name of the following groups: b"
  )
  # group_cols overlap with cols names fails
  expect_error(
    standardize_column_names(data.frame(a=1, b=2), cols=list(c="a", d="b"), group_cols="c"),
    regexp="group_cols must not overlap with standardized column names.  Change the name of the following groups: c"
  )
  # group_cols works
  expect_equal(
    standardize_column_names(data.frame(a=1, b=2), cols=list(d="b"), group_cols="a"),
    data.frame(group1=1, d=2)
  )
  # Missing values are inserted correctly
  expect_equal(
    standardize_column_names(
      data.frame(a=1, b=2),
      cols=list(d="b"),
      group_cols="a",
      insert_if_missing=list(dose=1)
    ),
    data.frame(group1=1, d=2, dose=1)
  )
  # Missing values are not inserted when something is already in the data for
  # that column
  expect_equal(
    standardize_column_names(
      data.frame(a=1, b=2, c=3),
      cols=list(d="b", dose="c"),
      group_cols="a",
      insert_if_missing=list(dose=2)
    ),
    data.frame(group1=1, d=2, dose=3)
  )
})

# restore_group_col_names ####

test_that("restore_group_col_names", {
  d <- data.frame(a=1, group1=2)
  # zero group cols
  expect_equal(
    restore_group_col_names(d),
    d
  )
  # One group col
  expect_equal(
    restore_group_col_names(d, group_cols="b"),
    data.frame(a=1, b=2)
  )
  # More than one group col
  expect_equal(
    restore_group_col_names(data.frame(a=1, group1=2, group2=3), c("b", "c")),
    data.frame(a=1, b=2, c=3)
  )
  expect_error(
    restore_group_col_names(d, group_cols=c("b", "c")),
    regexp="missing intermediate group_cols names"
  )
  expect_error(
    restore_group_col_names(data.frame(a=1, group2=3, group1=2), c("b", "c")),
    regexp="Intermediate group_cols are out of order"
  )
})

#' Compute the time to steady state using stepwise test of linear trend
#'
#' A linear slope is fit through the data to find when it becomes
#' non-significant.  Note that this is less preferred than the
#' `pk.tss.monoexponential` due to the fact that with more time or more subjects
#' the performance of the test changes (see reference).
#'
#' The model is fit with a different magnitude by treatment (as a factor, if
#' given) and a random slope by subject (if given).  A minimum of `min.points`
#' is required to fit the model.
#'
#' @param \dots See [pk.tss.data.prep()]
#' @param check See [pk.tss.data.prep()]
#' @param min.points The minimum number of points required for the fit
#' @param level The confidence level required for assessment of steady-state
#' @param verbose Describe models as they are run, show convergence of the model
#'   (passed to the nlme function), and additional details while running.
#' @returns A scalar float for the first time when steady-state is achieved or
#'   `NA` if it is not observed.
#' @family Time to steady-state calculations
#' @references Maganti L, Panebianco DL, Maes AL.  Evaluation of Methods for
#' Estimating Time to Steady State with Examples from Phase 1 Studies. AAPS
#' Journal 10(1):141-7. doi:10.1208/s12248-008-9014-y
#' @export
pk.tss.stepwise.linear <- function(...,
                                   min.points=3, level=0.95, verbose=FALSE,
                                   check=TRUE) {
  # Check inputs
  modeldata <- pk.tss.data.prep(..., check=check)
  if (is.factor(min.points) |
      !is.numeric(min.points))
    stop("min.points must be a number")
  if (!length(min.points) == 1) {
    warning("Only first value of min.points is used")
    min.points <- min.points[1]
  }
  if (min.points < 3)
    stop("min.points must be at least 3")
  if (is.factor(level) |
      !is.numeric(level)) {
    stop("level must be a number")
  }
  if (!length(level) == 1) {
    warning("Only first value of level is being used")
    level <- level[1]
  }
  if (level <= 0 | level >= 1) {
    stop("level must be between 0 and 1, exclusive")
  }
  # Confirm that we may have sufficient data to complete the
  # modeling.  Because of the variety of methods used for estimating
  # time to steady-state, assurance that we have enough data is more
  # simply determined by model convergence.
  if (length(unique(modeldata$time)) < min.points) {
    warning("After removing non-dosing time points, insufficient data remains for tss calculation")
    return(NA)
  }
  # Assign treatment if given and with multiple levels
  formula.to.fit <- stats::as.formula("conc~time")
  if ("treatment" %in% names(modeldata))
    formula.to.fit <- stats::as.formula("conc~time+treatment")
  # Ensure that the dosing times are in order to allow us to kick
  # them out in order.
  remaining.time <- sort(unique(modeldata$time))
  ret <- NA
  while (is.na(ret) &
         (length(remaining.time) >= min.points)) {
    if (verbose) {
      message("Trying ", min(remaining.time, na.rm=TRUE))
    }
    try({
      # Try to make the model
      current.interval <-
        if ("subject" %in% names(modeldata)) {
          # If we have a subject column, try to fit a linear
          # mixed-effects model.
          current.model <-
            nlme::lme(
              formula.to.fit,
              random=~time|subject,
              data=modeldata[modeldata$time >= min(remaining.time),,drop=FALSE])
          nlme::intervals(current.model, level=level, which="fixed")$fixed["time",]
        } else {
          # If we do not have a subject column, fit a linear model.
          current.model <-
            stats::glm(
              formula.to.fit,
              data=modeldata[modeldata$time >= min(remaining.time),,drop=FALSE])
          # There is no intervals function for glm, so build one
          ci <- as.vector(stats::confint(current.model, "time", level=level))
          c(ci[1], stats::coef(current.model)[["time"]], ci[2])
        }
      if (verbose) {
        message(
          sprintf("Current interval %g [%g, %g]",
                  current.interval[2],
                  current.interval[1],
                  current.interval[3])
        )
      }
      # If the signs of the upper and lower bounds of the slope of
      # the confidence interval for time are different, then we have
      # a non-significant slope.  A non-significant slope indicates steady-state.
      if (sign(current.interval[1]) != sign(current.interval[3]))
        ret <- min(remaining.time, na.rm=TRUE)
    }, silent=!verbose)
    remaining.time <- remaining.time[-1]
  }
  data.frame(
    tss.stepwise.linear=ret,
    stringsAsFactors=FALSE
  )
}

#' @importFrom dplyr inner_join
#' @export
dplyr::inner_join
#' @importFrom dplyr left_join
#' @export
dplyr::left_join
#' @importFrom dplyr right_join
#' @export
dplyr::right_join
#' @importFrom dplyr full_join
#' @export
dplyr::full_join

#' @importFrom dplyr filter
#' @export
dplyr::filter

#' @importFrom dplyr group_by
#' @export
dplyr::group_by

#' @importFrom dplyr ungroup
#' @export
dplyr::ungroup

#' @importFrom dplyr mutate
#' @export
dplyr::mutate

join_maker_PKNCA <- function(join_fun) {
  function(x, y, by = NULL, copy = FALSE, suffix = c(".x", ".y"), ..., keep = FALSE) { # nocov
    dataname <- getDataName(x)
    x[[dataname]] <- join_fun(x=x[[dataname]], y=y, by = by, copy = copy, suffix = suffix, ..., keep = keep)
    x
  } # nocov
}
filter_PKNCA <- function(.data, ..., .preserve=FALSE) {
  dataname <- getDataName(.data)
  .data[[dataname]] <- dplyr::filter(.data[[dataname]], ..., .preserve=.preserve)
  .data
}
mutate_PKNCA <- function(.data, ...) {
  dataname <- getDataName(.data)
  .data[[dataname]] <- dplyr::mutate(.data[[dataname]], ...)
  .data
}
group_by_PKNCA <- function(.data, ..., .add = FALSE, .drop = dplyr::group_by_drop_default(.data)) {
  dataname <- getDataName(.data)
  .data[[dataname]] <- dplyr::group_by(.data[[dataname]], ..., .add = FALSE, .drop = .drop)
  .data
}
ungroup_PKNCA <- function(x, ...) {
  dataname <- getDataName(x)
  x[[dataname]] <- dplyr::ungroup(x[[dataname]], ...)
  x
}

#' dplyr joins for PKNCA
#'
#' @inheritParams dplyr::inner_join
#' @family dplyr verbs
#' @export
inner_join.PKNCAresults <- join_maker_PKNCA(dplyr::inner_join)
#' @rdname inner_join.PKNCAresults
#' @export
left_join.PKNCAresults <- join_maker_PKNCA(dplyr::left_join)
#' @rdname inner_join.PKNCAresults
#' @export
right_join.PKNCAresults <- join_maker_PKNCA(dplyr::right_join)
#' @rdname inner_join.PKNCAresults
#' @export
full_join.PKNCAresults <- join_maker_PKNCA(dplyr::full_join)

#' @rdname inner_join.PKNCAresults
#' @export
inner_join.PKNCAconc <- join_maker_PKNCA(dplyr::inner_join)
#' @rdname inner_join.PKNCAresults
#' @export
left_join.PKNCAconc <- join_maker_PKNCA(dplyr::left_join)
#' @rdname inner_join.PKNCAresults
#' @export
right_join.PKNCAconc <- join_maker_PKNCA(dplyr::right_join)
#' @rdname inner_join.PKNCAresults
#' @export
full_join.PKNCAconc <- join_maker_PKNCA(dplyr::full_join)

#' @rdname inner_join.PKNCAresults
#' @export
inner_join.PKNCAdose <- join_maker_PKNCA(dplyr::inner_join)
#' @rdname inner_join.PKNCAresults
#' @export
left_join.PKNCAdose <- join_maker_PKNCA(dplyr::left_join)
#' @rdname inner_join.PKNCAresults
#' @export
right_join.PKNCAdose <- join_maker_PKNCA(dplyr::right_join)
#' @rdname inner_join.PKNCAresults
#' @export
full_join.PKNCAdose <- join_maker_PKNCA(dplyr::full_join)

#' dplyr filtering for PKNCA
#'
#' @inheritParams dplyr::filter
#' @family dplyr verbs
#' @export
filter.PKNCAresults <- filter_PKNCA
#' @rdname filter.PKNCAresults
#' @export
filter.PKNCAconc <- filter_PKNCA
#' @rdname filter.PKNCAresults
#' @export
filter.PKNCAdose <- filter_PKNCA

#' dplyr mutate-based modification for PKNCA
#'
#' @inheritParams dplyr::mutate
#' @family dplyr verbs
#' @export
mutate.PKNCAresults <- mutate_PKNCA
#' @rdname mutate.PKNCAresults
#' @export
mutate.PKNCAconc <- mutate_PKNCA
#' @rdname mutate.PKNCAresults
#' @export
mutate.PKNCAdose <- mutate_PKNCA

#' dplyr grouping for PKNCA
#'
#' @inheritParams dplyr::group_by
#' @family dplyr verbs
#' @export
group_by.PKNCAresults <- group_by_PKNCA
#' @rdname group_by.PKNCAresults
#' @export
group_by.PKNCAconc <- group_by_PKNCA
#' @rdname group_by.PKNCAresults
#' @export
group_by.PKNCAdose <- group_by_PKNCA
#' @rdname group_by.PKNCAresults
#' @inheritParams dplyr::ungroup
#' @export
ungroup.PKNCAresults <- ungroup_PKNCA
#' @rdname group_by.PKNCAresults
#' @export
ungroup.PKNCAconc <- ungroup_PKNCA
#' @rdname group_by.PKNCAresults
#' @export
ungroup.PKNCAdose <- ungroup_PKNCA

#' Create a PKNCAdata object.
#'
#' `PKNCAdata()` combines `PKNCAconc` and `PKNCAdose` objects and adds in the
#' intervals for PK calculations.
#'
#' @inheritParams PKNCA.choose.option
#' @param data.conc Concentration data as a `PKNCAconc` object or a data frame
#' @param data.dose Dosing data as a `PKNCAdose` object (see details)
#' @param impute Methods for imputation.  `NA` for to search for the column
#'   named "impute" in the intervals or no imputation if that column does not
#'   exist, a comma-or space-separated list of names, or the name of a column in
#'   the `intervals` data.frame.  See
#'   `vignette("v08-data-imputation", package="PKNCA")` for more details.
#' @param formula.conc Formula for making a `PKNCAconc` object with `data.conc`.
#'   This must be given if `data.conc` is a data.frame, and it must not be given
#'   if `data.conc` is a `PKNCAconc` object.
#' @param formula.dose Formula for making a `PKNCAdose` object with `data.dose`.
#'   This must be given if `data.dose` is a data.frame, and it must not be given
#'   if `data.dose` is a `PKNCAdose` object.
#' @param intervals A data frame with the AUC interval specifications as defined
#'   in [check.interval.specification()].  If missing, this will be
#'   automatically chosen by [choose.auc.intervals()]. (see details)
#' @param units A data.frame of unit assignments and conversions as created by
#'   [pknca_units_table()]
#' @param ... arguments passed to `PKNCAdata.default`
#' @returns A PKNCAdata object with concentration, dose, interval, and
#'   calculation options stored (note that PKNCAdata objects can also have
#'   results after a NCA calculations are done to the data).
#' @details If `data.dose` is not given or is `NA`, then the `intervals` must be
#'   given.  At least one of `data.dose` and `intervals` must be given.
#' @family PKNCA objects
#' @seealso [choose.auc.intervals()], [pk.nca()], [pknca_units_table()]
#' @export
PKNCAdata <- function(data.conc, data.dose, ...) {
  UseMethod("PKNCAdata", data.conc)
}

# Ensure that arguments are reversible
#' @rdname PKNCAdata
#' @export
PKNCAdata.PKNCAconc <- function(data.conc, data.dose, ...) {
  PKNCAdata.default(data.conc=data.conc, data.dose=data.dose, ...)
}

#' @rdname PKNCAdata
#' @export
PKNCAdata.PKNCAdose <- function(data.conc, data.dose, ...) {
  # Swap the arguments
  PKNCAdata.default(data.dose=data.conc, data.conc=data.dose, ...)
}

#' @rdname PKNCAdata
#' @export
PKNCAdata.default <- function(data.conc, data.dose, ...,
                              formula.conc, formula.dose,
                              impute = NA_character_,
                              intervals, units, options=list()) {
  if (length(list(...))) {
    stop("Unknown argument provided to PKNCAdata.  All arguments other than `data.conc` and `data.dose` must be named.")
  }
  ret <- list()
  # Generate the conc element
  if (inherits(data.conc, "PKNCAconc")) {
    if (!missing(formula.conc)) {
      rlang::warn(
        message = "data.conc was given as a PKNCAconc object.  Ignoring formula.conc",
        class = "pknca_dataconc_formulaconc"
      )
    }
    ret$conc <- data.conc
  } else {
    ret$conc <- PKNCAconc(data.conc, formula=formula.conc)
  }
  # Generate the dose element
  if (missing(data.dose)) {
    ret$dose <- NA
  } else if (identical(data.dose, NA)) {
    ret$dose <- NA
  } else if (inherits(data.dose, "PKNCAdose")) {
    if (!missing(formula.dose))
      rlang::warn(
        message = "data.dose was given as a PKNCAdose object.  Ignoring formula.dose",
        class = "pknca_dataconc_formuladose"
      )
    ret$dose <- data.dose
  } else {
    ret$dose <- PKNCAdose(data.dose, formula.dose)
  }
  # Check the options
  if (!is.list(options)) {
    stop("options must be a list.")
  }
  if (length(options) > 0) {
    if (is.null(names(options)))
      stop("options must have names.")
    for (n in names(options)) {
      tmp.opt <- list(options[[n]], TRUE)
      names(tmp.opt) <- c(n, "check")
      do.call(PKNCA.options, tmp.opt)
    }
  }
  ret$options <- options

  # Assign the class and give it all back to the user.
  class(ret) <- c("PKNCAdata", class(ret))

  # Check the intervals
  if (missing(intervals) & identical(ret$dose, NA)) {
    stop("If data.dose is not given, intervals must be given")
  } else if (missing(intervals)) {
    # Generate the intervals for each grouping of concentration and
    # dosing.
    if (length(ret$dose$columns$time) == 0) {
      stop("Dose times were not given, so intervals must be manually specified.")
    }
    n_conc_dose <-
      full_join_PKNCAconc_PKNCAdose(
        o_conc = ret$conc,
        o_dose = ret$dose
      )
    n_conc_dose$data_intervals <- rep(list(NULL), nrow(n_conc_dose))
    for (idx in seq_len(nrow(n_conc_dose))) {
      current_conc <- n_conc_dose$data_conc[[idx]]
      current_dose <- n_conc_dose$data_dose[[idx]]
      current_group <-
        n_conc_dose[
          idx,
          setdiff(names(n_conc_dose), c("data_conc", "data_dose")),
          drop=FALSE
        ]
      warning_prefix <-
        if (ncol(current_group) > 0) {
          paste0(
            paste(names(current_group), unlist(lapply(current_group, as.character)), sep="=", collapse="; "),
            ": "
          )
        } else {
          ""
        }
      if (!is.null(current_conc)) {
        generated_intervals <-
          choose.auc.intervals(
            current_conc$time,
            current_dose$time,
            options=options
          )
        if (nrow(generated_intervals) > 0) {
          n_conc_dose$data_intervals[[idx]] <- generated_intervals
        } else {
          warning(warning_prefix, "No intervals generated likely due to limited concentration data")
        }
      } else {
        rlang::warn(
          message = paste(warning_prefix, "No intervals generated due to no concentration data"),
          class = "pknca_no_intervals_generated"
        )
      }
    }
    intervals <-
      tidyr::unnest(
        n_conc_dose[, setdiff(names(n_conc_dose), c("data_conc", "data_dose")), drop=FALSE],
        cols="data_intervals"
      )
  }
  ret <- set_intervals(data = ret, intervals = intervals)
  ret$intervals <- check.interval.specification(intervals)
  # Verify that either everything or nothing is using units
  units_interval_start <- inherits(ret$intervals$start, "units")
  units_interval_end <- inherits(ret$intervals$end, "units")

  # Insert the unit conversion table
  if (missing(units)) {
    # What unit types are recognized?
    possible_units <-
      setdiff(
        grep(x = names(formals(pknca_units_table)), pattern = "_", invert = TRUE, value = TRUE),
        "conversions"
      )
    possible_units_pref <- paste0(possible_units, "_pref")
    # Accumulate available units
    conc_units_values <- ret$conc$units
    conc_units_cols <- ret$conc$columns[names(ret$conc$columns) %in% possible_units]

    unit_args <- conc_units_values
    for (nm in names(conc_units_cols)) {
      unit_args[[nm]] <- unique(stats::na.omit(ret$conc$data[[conc_units_cols[[nm]]]]))
    }

    if (!identical(ret$dose, NA)) {
      unit_args <- append(unit_args, ret$dose$units)
      dose_units_cols <- ret$dose$columns[names(ret$dose$columns) %in% possible_units]
      for (nm in names(dose_units_cols)) {
        unit_args[[nm]] <- unique(stats::na.omit(ret$dose$data[[dose_units_cols[[nm]]]]))
      }
    }
    # If there are any units to set, set them here
    if (length(unit_args) > 0) {
      unit_args <- lapply(X = unit_args, FUN = drop_attributes)
      ret$units <- do.call(pknca_units_table, args = unit_args)
    }
  } else {
    stopifnot("`units` must be a data.frame"=is.data.frame(units))
    stopifnot(
      "`units` data.frame must have at least names 'PPTESTCD' and 'PPORRESU'"=
        all(c("PPTESTCD", "PPORRESU") %in% names(units))
    )
    stopifnot("`units` must have at least one row"=nrow(units) > 0)
    ret$units <- units
  }

  # Insert the imputation methods, if applicable
  if (!identical(NA, impute)) {
    checkmate::assert_character(impute, len = 1)
    ret$impute <- impute
  }

  ret
}

drop_attributes <- function(x) {
  attributes(x) <- NULL
  x
}

#' @rdname is_sparse_pk
#' @export
is_sparse_pk.PKNCAdata <- function(object) {
  is_sparse_pk(object$conc)
}

#' Print a PKNCAdata object
#' @param x The object to print
#' @param ... Arguments passed on to [print.PKNCAconc()] and [print.PKNCAdose()]
#' @export
print.PKNCAdata <- function(x, ...) {
  print.PKNCAconc(x$conc, ...)
  if (identical(NA, x$dose)) {
    cat("No dosing information.\n")
  } else {
    print.PKNCAdose(x$dose, ...)
  }
  cat(sprintf("\nWith %d rows of interval specifications.\n",
              nrow(x$intervals)))
  if (!is.null(x$units)) {
    cat("With units\n")
  }
  if (!is.null(x$impute)) {
    cat(sprintf("With imputation: %s\n", x$impute))
  }
  if (length(x$options) == 0) {
    cat("No options are set differently than default.\n")
  } else {
    cat("Options changed from default are:\n")
    print(x$options)
  }
}

#' Summarize a PKNCAdata object showing important details about the
#' concentration, dosing, and interval information.
#' @param object The PKNCAdata object to summarize.
#' @param ... arguments passed on to [print.PKNCAdata()]
#' @export
summary.PKNCAdata <- function(object, ...) {
  print.PKNCAdata(object, summarize=TRUE, ...)
}

#' Get the groups (right hand side after the `|` from a PKNCA
#' object).
#'
#' @rdname getGroups.PKNCAconc
#' @param object The object to extract the data from
#' @param ... Arguments passed to other getGroups functions
#' @returns A data frame with the (selected) group columns.
#' @export
getGroups.PKNCAdata <- function(object, ...) {
  getGroups(as_PKNCAconc(object), ...)
}

#' @describeIn group_vars.PKNCAconc Get group_vars for a PKNCAdata object
#'   from the PKNCAconc object within
#' @exportS3Method dplyr::group_vars
group_vars.PKNCAdata <- function(x) {
  group_vars.PKNCAconc(as_PKNCAconc(x))
}

#' Create a PKNCAconc object
#'
#' @param data A data frame with concentration (or amount for urine/feces),
#'   time, and the groups defined in `formula`.
#' @param formula The formula defining the `concentration~time|groups` or
#'   `amount~time|groups` for urine/feces (In the remainder of the
#'   documentation, "concentration" will be used to describe concentration or
#'   amount.)  One special aspect of the `groups` part of the formula is that
#'   the last group is typically assumed to be the `subject`; see the
#'   documentation for the `subject` argument for exceptions to this assumption.
#' @param subject The column indicating the subject number.  If not provided,
#'   this defaults to the beginning of the inner groups: For example with
#'   `concentration~time|Study+Subject/Analyte`, the inner groups start with the
#'   first grouping variable before a `/`, `Subject`.  If there is only one
#'   grouping variable, it is assumed to be the subject (e.g.
#'   `concentration~time|Subject`), and if there are multiple grouping variables
#'   without a `/`, subject is assumed to be the last one.  For single-subject
#'   data, it is assigned as `NULL`.
#' @param time.nominal (optional) The name of the nominal time column (if the
#'   main time variable is actual time.  The `time.nominal` is not used during
#'   calculations; it is available to assist with data summary and checking.
#' @param exclude (optional) The name of a column with concentrations to exclude
#'   from calculations and summarization.  If given, the column should have
#'   values of `NA` or `""` for concentrations to include and non-empty text for
#'   concentrations to exclude.
#' @param volume (optional) The volume (or mass) of collection as is typically
#'   used for urine or feces measurements.
#' @param duration (optional) The duration of collection as is typically used
#'   for concentration measurements in urine or feces.
#' @param exclude_half.life,include_half.life A character scalar for the column
#'   name in the dataset of the points to exclude from the half-life calculation
#'   (still using normal curve-stripping selection rules for the other points)
#'   or to include for the half-life (using specifically those points and
#'   bypassing automatic curve-stripping point selection).  See the "Half-Life
#'   Calculation" vignette for more details on the use of these arguments.
#' @param sparse Are the concentration-time data sparse PK (commonly used in
#'   small nonclinical species or with terminal or difficult sampling) or dense
#'   PK (commonly used in clinical studies or larger nonclinical species)?
#' @param ... Ignored.
#' @returns A PKNCAconc object that can be used for automated NCA.
#' @family PKNCA objects
#' @export
PKNCAconc <- function(data, ...) {
  UseMethod("PKNCAconc")
}

#' @rdname PKNCAconc
#' @export
PKNCAconc.default <- function(data, ...) {
  PKNCAconc.data.frame(as.data.frame(data), ...)
}
#' @rdname PKNCAconc
#' @export
PKNCAconc.tbl_df <- function(data, ...) {
  PKNCAconc.data.frame(as.data.frame(data), ...)
}

#' @rdname PKNCAconc
#' @param concu,amountu,timeu Either unit values (e.g. "ng/mL") or column names
#'   within the data where units are provided.
#' @param concu_pref,amountu_pref,timeu_pref Preferred units for reporting (not
#'   column names)
#' @export
PKNCAconc.data.frame <- function(data, formula, subject,
                                 time.nominal, exclude = NULL, duration, volume,
                                 exclude_half.life, include_half.life, sparse = FALSE, ...,
                                 concu = NULL, amountu = NULL, timeu = NULL,
                                 concu_pref = NULL, amountu_pref = NULL, timeu_pref = NULL) {
  # The data must have... data
  if (nrow(data) == 0) {
    stop("data must have at least one row.")
  }
  # Verify that all the variables in the formula are columns in the data.
  missing_vars <- setdiff(all.vars(formula), names(data))
  if (length(missing_vars) > 0) {
    stop("All of the variables in the formula must be in the data.  Missing: ", paste(missing_vars))
  }
  parsed_form_raw <- parse_formula_to_cols(form = formula)
  parsed_form_groups <-
    if (length(parsed_form_raw$groups) > 0) {
      list(
        group_vars=parsed_form_raw$groups,
        group_analyte=character()
      )
    } else {
      list(
        group_vars=parsed_form_raw$groups_left_of_slash,
        group_analyte=parsed_form_raw$groups_right_of_slash
      )
    }
  parsed_form <-
    list(
      concentration = parsed_form_raw$lhs,
      time = parsed_form_raw$rhs,
      groups = parsed_form_groups
    )
  if (length(parsed_form$concentration) != 1) {
    stop("The left hand side of the formula must have exactly one variable")
  }
  if (length(parsed_form$time) != 1) {
    stop("The right hand side of the formula (excluding groups) must have exactly one variable")
  }
  # Do some general checking of the concentration and time data to give an early
  # error if the data are not correct.  Do not check monotonic.time because the
  # data may contain information for more than one subject.
  assert_conc_time(
    conc = data[[parsed_form$concentration]],
    time = data[[parsed_form$time]],
    sorted_time = FALSE
  )
  # Assign the subject
  if (missing(subject)) {
    subject <- parsed_form$groups$group_vars[length(parsed_form$groups$group_vars)]
  } else {
    # Ensure that the subject is part of the data definition and a scalar
    # character string.
    if (!is.character(subject))
      stop("subject must be a character string")
    if (!(length(subject) == 1))
      stop("subject must be a scalar")
    if (!(subject %in% names(data)))
      stop("The subject parameter must map to a name in the data")
  }
  parsed_form$subject <- subject
  if (sparse) {
    ret <-
      list(
        data_sparse = data,
        formula = formula,
        columns = parsed_form
      )
  } else {
    ret <-
      list(
        data = data,
        formula = formula,
        columns = parsed_form
      )
  }
  class(ret) <- c("PKNCAconc", class(ret))
  ret <- setExcludeColumn(ret, exclude = exclude, dataname = getDataName.PKNCAconc(ret))
  # Values must be unique (one value per measurement), check after the exclusion
  # column has been added to the object so that exclusions can be accounted for
  # in duplicate checking.
  duplicate_check(object = ret, data_type = "concentration")

  if (missing(volume)) {
    ret <- setAttributeColumn(ret, attr_name="volume", default_value=NA_real_)
  } else {
    ret <- setAttributeColumn(ret, attr_name="volume", col_or_value=volume)
    if (!is.numeric(getAttributeColumn(ret, attr_name="volume")[[1]])) {
      stop("Volume must be numeric")
    }
  }
  if (missing(duration)) {
    ret <- setDuration.PKNCAconc(ret)
  } else {
    ret <- setDuration.PKNCAconc(ret, duration=duration)
  }
  if (!missing(time.nominal)) {
    ret <-
      setAttributeColumn(object=ret,
                         attr_name="time.nominal",
                         col_name=time.nominal)
  }
  if (!missing(exclude_half.life)) {
    ret <-
      setAttributeColumn(object=ret,
                         attr_name="exclude_half.life",
                         col_name=exclude_half.life)
  }
  if (!missing(include_half.life)) {
    ret <-
      setAttributeColumn(object=ret,
                         attr_name="include_half.life",
                         col_name=include_half.life)
  }

  # Unit handling
  ret <-
    pknca_set_units(
      ret,
      units_orig = list(concu = concu, amountu = amountu, timeu = timeu),
      units_pref = list(concu_pref = concu_pref, amountu_pref = amountu_pref, timeu_pref = timeu_pref)
    )
  ret
}

#' Extract the formula from a PKNCAconc object.
#'
#' @param x The object to extract the formula from.
#' @param \dots Unused
#' @return A formula object
#' @export
#' @importFrom stats formula
formula.PKNCAconc <- function(x, ...) {
  x$formula
}

#' Extract the columns used in the formula (in order) from a PKNCAconc
#' or PKNCAdose object.
#'
#' @param formula The object to use (parameter name is `formula` to use the
#'   generic function)
#' @param \dots Unused
#' @returns A data frame with the columns from the object in formula order.
#' @export
#' @importFrom stats model.frame
#' @method model.frame PKNCAconc
model.frame.PKNCAconc <- function(formula, ...) {
  formula$data[, all.vars(formula$formula), drop=FALSE]
}

#' @export
getDepVar.PKNCAconc <- function(x, ...) {
  x$data[, x$columns$concentration]
}

#' @export
getIndepVar.PKNCAconc <- function(x, ...) {
  x$data[, x$columns$time]
}

#' Get the groups (right hand side after the `|` from a PKNCA
#' object).
#'
#' @param object The object to extract the data from
#' @param form The formula to extract the data from (defaults to the formula
#'   from `object`)
#' @param level optional.  If included, this specifies the level(s) of the
#'   groups to include.  If a numeric scalar, include the first `level` number
#'   of groups.  If a numeric vector, include each of the groups specified by
#'   the number.  If a character vector, include the named group levels.
#' @param data The data to extract the groups from (defaults to the data from
#'   `object`)
#' @param sep Unused (kept for compatibility with the nlme package)
#' @param ... Arguments passed to other getGroups functions
#' @returns A data frame with the (selected) group columns.
#' @export
getGroups.PKNCAconc <- function(object, form=stats::formula(object), level,
                                data=as.data.frame(object), sep) {
  grpnames <- unlist(object$columns$groups)
  if (!missing(level))
    if (is.factor(level) | is.character(level)) {
      level <- as.character(level)
      if (any(!(level %in% grpnames)))
        stop("Not all levels are listed in the group names.  Missing levels are: ",
             paste(setdiff(level, grpnames), collapse=", "))
      grpnames <- level
    } else if (is.numeric(level)) {
      if (length(level) == 1 &&
          level > 0) {
        grpnames <- grpnames[1:level]
      } else {
        grpnames <- grpnames[level]
      }
    }
  data[, grpnames, drop=FALSE]
}

#' Get grouping variables for a PKNCA object
#'
#' @param x The PKNCA object
#' @returns A character vector (possibly empty) of the grouping variables
#' @exportS3Method dplyr::group_vars
group_vars.PKNCAconc <- function(x) {
  unname(unlist(x$columns$groups))
}

#' @rdname getDataName
#' @export
getDataName.PKNCAconc <- function(object) {
  if (is_sparse_pk(object)) {
    "data_sparse"
  } else {
    "data"
  }
}

#' @rdname setDuration
#' @export
setDuration.PKNCAconc <- function(object, duration, ...) {
  if (missing(duration)) {
    object <-
      setAttributeColumn(
        object=object,
        attr_name="duration",
        default_value=0,
        message_if_default="Assuming point rather than interval concentration measurement"
      )
  } else {
    object <-
      setAttributeColumn(object=object, attr_name="duration", col_or_value=duration)
  }
  duration.val <- getAttributeColumn(object=object, attr_name="duration")[[1]]
  if (is.numeric(duration.val) &&
      !any(is.na(duration.val)) &&
      !any(is.infinite(duration.val)) &&
      all(duration.val >= 0)) {
    # It passes the test
  } else {
    stop("duration must be numeric without missing (NA) or infinite values, and all values must be >= 0")
  }
  object
}

#' Print and/or summarize a PKNCAconc or PKNCAdose object.
#'
#' @param x The object to print
#' @param object The object to summarize
#' @param n The number of rows of data to show (see
#' [head()])
#' @param summarize Summarize the nested number of groups
#' @param \dots Arguments passed to `print.formula` and
#' `print.data.frame`
#' @export
print.PKNCAconc <- function(x, n=6, summarize=FALSE, ...) {
  cat(sprintf("Formula for concentration:\n "))
  print(stats::formula(x), ...)
  if (is_sparse_pk(x)) {
    data_current <- x$data_sparse
    is_sparse <- TRUE
    cat("Data are sparse PK.\n")
  } else {
    data_current <- x$data
    is_sparse <- FALSE
    cat("Data are dense PK.\n")
  }
  single_subject <- is.na(x$columns$subject) || (length(x$columns$subject) == 0)
  if (single_subject) {
    cat("As a single-subject dataset.\n")
  } else {
    cat(sprintf("With %d subjects defined in the '%s' column.\n",
                length(unique(data_current[,x$columns$subject])),
                x$columns$subject))
  }
  if ("time.nominal" %in% names(x$columns)) {
    cat("Nominal time column is: ", x$columns$time.nominal, "\n", sep="")
  } else {
    cat("Nominal time column is not specified.\n")
  }
  if (summarize) {
    cat("\n")
    grp <- getGroups(x)
    if (ncol(grp) > 0) {
      tmp_summary <- data.frame(Group.Name=names(grp),
                                Count=0)
      for (i in seq_len(ncol(grp))) {
        tmp_summary$Count[i] <- nrow(unique(grp[,1:i,drop=FALSE]))
      }
      cat("Group summary:\n")
      names(tmp_summary) <- gsub("\\.", " ", names(tmp_summary))
      print.data.frame(tmp_summary, row.names=FALSE)
    } else {
      cat("No groups.\n")
    }
  }
  if (n != 0) {
    if (n >= nrow(data_current)) {
      cat("\nData for concentration:\n")
    } else if (n < 0) {
      cat(sprintf("\nFirst %d rows of concentration data:\n",
                  nrow(data_current)+n))
    } else {
      cat(sprintf("\nFirst %d rows of concentration data:\n",
                  n))
    }
    print.data.frame(utils::head(data_current, n=n), ..., row.names=FALSE)
  }
}

#' @rdname is_sparse_pk
#' @export
is_sparse_pk.PKNCAconc <- function(object) {
  "data_sparse" %in% names(object)
}

#' @rdname print.PKNCAconc
#' @export
summary.PKNCAconc <- function(object, n=0, summarize=TRUE, ...) {
  print.PKNCAconc(object, n=n, summarize=summarize)
}

#' @export
as.data.frame.PKNCAconc <- function(x, ...) {
  if (is_sparse_pk(x)) {
    x$data_sparse
  } else {
    x$data
  }
}

# The requirements for superposition are defined below to enable
# test-driven development.

#' Dosing must be single dose
#' * A single dose within the time period
#'   * Time period may be a subset of the data from a multi-dose study
#' * Concentrations must be BLQ at the beginning of the interval
#'   * User can over-ride the requirement of BLQ at the beginning of the interval
#' * Half-life
#'   * If the number of dosing intervals times the duration of the dosing interval is less than the observed data, half-life is not required.  (This will likely be rare.)
#'   * Half-life can be specified on input.
#'   * Half-life can be sufficiently determined from the data given (see the half-life cleaning functions for the definition of "sufficient")
#'   * Mixing methods of half-life determination are supported.
#' * Extrapolation can occur via either Clast,obs or Clast,pred.
#'   * If using Clast,pred, it must be given if the half-life is given.
#' * Dosing interval can be specified as either a simple number for dosing interval (tau) or multiple doses within the interval
#'   * Example: Daily dosing would be: 24
#'   * Example: Breakfast/dinner dosing would be: c(10, 24)
#' * How many times to repeat the dosing interval can either be given as a number of taus to repeat or "steady-state"
#'   * If steady-state, then a relative tolerance will be provided to confirm that no concentration changes by a greater ratio from the previous to the current dose.
#' * Doses amount may be provided
#'   * If doses are not given, they are assumed to be the same as the input dose.
#'   * If doses are given, linear dose-proportionality will be assumed.
#'   * Dose amount can be given either as a scalar which will apply to all intervals or a vector which will apply to each interval.
#'     * If dose amount is given as a vector, it must be the same length as the number of intervals input.
#' * User can specify either integration as either linear or linear-up/log-down
#' * Time points in the output are automatically selected as all observed points modulo the dosing interval.
#'   * More time points can be added by user specification, but points cannot be removed.  (Note that this will ensure the highest possible Cmax in the output data.)
#' * The functions work with either individual or grouped data.

test_that("superposition inputs", {
  # FIXME: Enforce single dose

  # Enforce BLQ checking for the first data point
  expect_error(superposition(conc=c(1, 2), time=c(0, 1), tau=24),
               regexp="The first concentration must be 0")
  expect_error(superposition(conc=c(NA, 2), time=c(0, 1), tau=24),
               regexp="The first concentration must be 0")

  # dose.input must be scalar
  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), dose.input=c(0, 1)),
    regexp="Assertion on 'dose.input' failed: Must have length 1."
  )
  # Ensure that dose.input must be a number
  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), dose.input=NA),
    regexp = "Assertion on 'dose.input' failed: Contains missing values (element 1).",
    fixed = TRUE
  )
  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), dose.input=factor("1")),
    regexp="Assertion on 'dose.input' failed: Must be of type 'numeric' (or 'NULL'), not 'factor'.",
    fixed = TRUE
  )
  # dose.input must be > 0
  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), dose.input=-1),
    regexp="Assertion on 'dose.input' failed: Element 1 is not > 0"
  )
  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), dose.input=0),
    regexp="Assertion on 'dose.input' failed: Element 1 is not > 0"
  )

  # tau must be scalar
  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), tau=c(0, 1)),
    regexp="Assertion on 'tau' failed: Must have length 1, but has length 2."
  )
  # Ensure that tau must be a number
  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), tau=NA),
    regexp="Assertion on 'tau' failed: Contains missing values (element 1).",
    fixed = TRUE
  )
  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), tau=factor("1")),
    regexp="Assertion on 'tau' failed: Must be of type 'numeric', not 'factor'."
  )
  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), tau="1"),
    regexp="Assertion on 'tau' failed: Must be of type 'numeric', not 'character'."
  )
  # tau must be > 0
  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), tau=-1),
    regexp="Assertion on 'tau' failed: Element 1 is not > 0"
  )
  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), tau=0),
    regexp="Assertion on 'tau' failed: Element 1 is not > 0"
  )

  # >= 1 dose time
  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), tau=24, dose.times=numeric()),
    regexp="Assertion on 'dose.times' failed: Must have length >= 1, but has length 0."
  )
  # Dose times must be a number
  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), tau=24, dose.times="1"),
    regexp = "Assertion on 'dose.times' failed: Must be of type 'numeric', not 'character'."
  )
  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), tau=24, dose.times=factor("1")),
    regexp="Assertion on 'dose.times' failed: Must be of type 'numeric', not 'factor'."
  )
  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), tau=24, dose.times=NA),
    regexp = "Assertion on 'dose.times' failed: Contains missing values (element 1).",
    fixed = TRUE
  )
  # Dose times nonnegative
  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), tau=24, dose.times=c(-1, 0)),
    regexp="Assertion on 'dose.times' failed: Element 1 is not >= 0."
  )
  # Dose times <= tau
  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), tau=24, dose.times=c(0, 25)),
    regexp = "Assertion on 'dose.times' failed: Element 2 is not < 24"
  )

  # dose.amount without dose.input
  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), tau=24, dose.times=0, dose.amount=1),
    regexp="must give dose.input to give dose.amount"
  )
  # dose.amount same length as dose.times
  expect_error(superposition(conc=c(0, 2), time=c(0, 1), dose.input=1,
                             tau=24, dose.times=0, dose.amount=c(1, 2)),
               regexp="dose.amount must either be a scalar or match the length of dose.times")
  expect_error(superposition(conc=c(0, 2), time=c(0, 1), dose.input=1,
                             tau=24, dose.times=c(0, 1), dose.amount=c(1, 2, 3)),
               regexp="dose.amount must either be a scalar or match the length of dose.times")
  # dose.amount numeric
  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), dose.input=1, tau=24, dose.times=0, dose.amount="1"),
    regexp = "Assertion on 'dose.amount' failed: Must be of type 'numeric', not 'character'."
  )
  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), dose.input=1, tau=24, dose.times=0, dose.amount=factor("1")),
    regexp = "Assertion on 'dose.amount' failed: Must be of type 'numeric', not 'factor'."
  )
  # dose.amount finite/NA
  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), dose.input=1, tau=24, dose.times=0, dose.amount=NA_real_),
    regexp = "Assertion on 'dose.amount' failed: Contains missing values (element 1).",
    fixed = TRUE
  )
  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), dose.input=1, tau=24, dose.times=0, dose.amount=Inf),
    regexp = "Assertion on 'dose.amount' failed: Must be finite."
  )
  # dose.amount non-negative
  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), dose.input=1, tau=24, dose.times=0, dose.amount=-1),
    regexp = "Assertion on 'dose.amount' failed: Element 1 is not > 0"
  )
  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), dose.input=1, tau=24, dose.times=0, dose.amount=0),
    regexp="Assertion on 'dose.amount' failed: Element 1 is not > 0"
  )

  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), tau=24, n.tau=c(1, 2)),
    regexp="Assertion on 'n.tau' failed: Must have length 1."
  )
  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), tau=24, n.tau="1"),
    regexp = "Assertion on 'n.tau' failed: Must be of type 'number', not 'character'."
  )
  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), tau=24, n.tau=NA),
    regexp="Assertion on 'n.tau' failed: May not be NA."
  )
  # n.tau >= 1
  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), tau=24, n.tau=0),
    regexp="Assertion on 'n.tau' failed: Element 1 is not >= 1."
  )
  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), tau=24, n.tau=-Inf),
    regexp = "Assertion on 'n.tau' failed: Element 1 is not >= 1."
  )
  # n.tau integer
  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), tau=24, n.tau=1.1),
    regexp = "Assertion on 'n.tau' failed: Must be of type 'integerish', but element 1 is not close to an integer."
  )
  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), tau=24, n.tau=1.0000001),
    regexp = "Assertion on 'n.tau' failed: Must be of type 'integerish', but element 1 is not close to an integer."
  )

  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), tau=24, n.tau=Inf, lambda.z=c(1, 2)),
    regexp="Assertion on 'lambda.z' failed: Must have length 1, but has length 2."
  )

  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), tau=24, lambda.z="1"),
    regexp="Assertion on 'lambda.z' failed: Must be of type 'numeric', not 'character'."
  )

  expect_equal(
    superposition(
      conc=c(4, 2, 1, 0.5),
      time=0:3,
      tau=24,
      n.tau=Inf,
      check.blq=FALSE
    ),
    superposition(
      conc=c(4, 2, 1, 0.5),
      time=0:3,
      tau=24,
      clast.pred=TRUE,
      n.tau=Inf,
      check.blq=FALSE
    ),
    info="clast.pred may be provided as 'TRUE'"
  )

  expect_equal(
    superposition(
      conc=c(4, 2, 1, 0.5),
      time=0:3,
      tau=24,
      clast.pred=NA,
      n.tau=Inf,
      check.blq=FALSE
    ),
    superposition(
      conc=c(4, 2, 1, 0.5),
      time=0:3,
      tau=24,
      clast.pred=FALSE,
      n.tau=Inf,
      check.blq=FALSE
    ),
    info="clast.pred NA is the same as FALSE"
  )

  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), tau=24, clast.pred=c(1, 2)),
    regexp = "Assertion on 'clast.pred' failed: One of the following must apply:.*Must have length 1"
  )
  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), tau=24, clast.pred=-1),
    regexp = "Assertion on 'clast.pred' failed: One of the following must apply:.*Element 1 is not >= 0"
  )

  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), tau=24, clast.pred="1"),
    regexp="Must be of type"
  )
  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), tau=24, clast.pred=factor("1")),
    regexp="Must be of type"
  )

  # tlast given and not a scalar
  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), tau=24, tlast=c(1, 2)),
    regexp="Assertion on 'tlast' failed: Must have length 1."
  )
  # tlast given and not a number
  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), tau=24, tlast="1"),
    regexp="Assertion on 'tlast' failed: Must be of type 'number', not 'character'."
  )
  expect_error(
    superposition(conc=c(0, 2), time=c(0, 1), tau=24, tlast=NA),
    regexp="Assertion on 'tlast' failed: May not be NA."
  )
  # TODO: test mixing clast.pred and lambda.z

  # additional.times NA
  expect_error(superposition(conc=c(0, 2), time=c(0, 1), tau=24,
                             additional.times=NA),
               regexp="No additional.times may be NA \\(to not include any additional.times, enter c\\(\\) as the function argument\\)")
  expect_error(superposition(conc=c(0, 2), time=c(0, 1), tau=24,
                             additional.times=c(2, NA)),
               regexp="No additional.times may be NA \\(to not include any additional.times, enter c\\(\\) as the function argument\\)")
  # additional.times nonnumeric
  expect_error(superposition(conc=c(0, 2), time=c(0, 1), tau=24,
                             additional.times="1"),
               regexp="additional.times must be a number")
  expect_error(superposition(conc=c(0, 2), time=c(0, 1), tau=24,
                             additional.times=factor("1")),
               regexp="additional.times must be a number")
  # additional times < 0
  expect_error(superposition(conc=c(0, 2), time=c(0, 1), tau=24,
                             additional.times=-1),
               regexp="All additional.times must be nonnegative")
  expect_error(superposition(conc=c(0, 2), time=c(0, 1), tau=24,
                             additional.times=c(-1, 0)),
               regexp="All additional.times must be nonnegative")
  # Additional times > tau
  expect_error(superposition(conc=c(0, 2), time=c(0, 1), tau=24,
                             additional.times=25),
               regexp="All additional.times must be <= tau")
  expect_error(superposition(conc=c(0, 2), time=c(0, 1), tau=24,
                             additional.times=c(0, 25)),
               regexp="All additional.times must be <= tau")

  # steady.state.tol scalar
  expect_error(superposition(conc=c(0, 2), time=c(0, 1), tau=24,
                             steady.state.tol=c(1, 2)),
               regexp="steady.state.tol must be a scalar")
  # steady.state.tol numeric
  expect_error(superposition(conc=c(0, 2), time=c(0, 1), tau=24,
                             steady.state.tol="1"),
               regexp="steady.state.tol must be a number")
  expect_error(superposition(conc=c(0, 2), time=c(0, 1), tau=24,
                             steady.state.tol="1"),
               regexp="steady.state.tol must be a number")
  expect_error(superposition(conc=c(0, 2), time=c(0, 1), tau=24,
                             steady.state.tol=factor("1")),
               regexp="steady.state.tol must be a number")
  expect_error(superposition(conc=c(0, 2), time=c(0, 1), tau=24,
                             steady.state.tol=NA),
               regexp="steady.state.tol must be a number")
  # steady.state.tol range
  expect_error(superposition(conc=c(0, 2), time=c(0, 1), tau=24,
                             steady.state.tol=0),
               regexp="steady.state.tol must be between 0 and 1, exclusive.")
  expect_error(superposition(conc=c(0, 2), time=c(0, 1), tau=24,
                             steady.state.tol=1),
               regexp="steady.state.tol must be between 0 and 1, exclusive.")
  expect_error(superposition(conc=c(0, 2), time=c(0, 1), tau=24,
                             steady.state.tol=-1),
               regexp="steady.state.tol must be between 0 and 1, exclusive.")
  expect_error(superposition(conc=c(0, 2), time=c(0, 1), tau=24,
                             steady.state.tol=2),
               regexp="steady.state.tol must be between 0 and 1, exclusive.")
  suppressWarnings(
    expect_warning(
      superposition(conc=c(0, 2), time=c(0, 1), tau=24, steady.state.tol=0.1),
      regexp="steady.state.tol is usually <= 0.01",
      fixed=TRUE
    )
  )

  # Combinations of lambda.z, clast.pred, tlast
  expect_error(superposition(conc=c(0, 2), time=c(0, 1), tau=24,
                             clast.pred=1),
               regexp="Either give all or none of the values for these arguments: lambda.z, clast.pred, and tlast")
  expect_error(superposition(conc=c(0, 2), time=c(0, 1), tau=24,
                             lambda.z=1),
               regexp="Either give all or none of the values for these arguments: lambda.z, clast.pred, and tlast")
  expect_error(superposition(conc=c(0, 2), time=c(0, 1), tau=24,
                             tlast=1),
               regexp="Either give all or none of the values for these arguments: lambda.z, clast.pred, and tlast")
  expect_error(superposition(conc=c(0, 2), time=c(0, 1), tau=24,
                             clast.pred=1, lambda.z=1),
               regexp="Either give all or none of the values for these arguments: lambda.z, clast.pred, and tlast")
  expect_error(superposition(conc=c(0, 2), time=c(0, 1), tau=24,
                             clast.pred=1, tlast=1),
               regexp="Either give all or none of the values for these arguments: lambda.z, clast.pred, and tlast")
  expect_error(superposition(conc=c(0, 2), time=c(0, 1), tau=24,
                             lambda.z=1),
               regexp="Either give all or none of the values for these arguments: lambda.z, clast.pred, and tlast")
  expect_error(superposition(conc=c(0, 2), time=c(0, 1), tau=24,
                             lambda.z=1, tlast=1),
               regexp="Either give all or none of the values for these arguments: lambda.z, clast.pred, and tlast")

})

test_that("superposition math", {
  # Superposition without half-life needed and no
  # interpolation/extrapolation
  c1 <- c(0, 1, 2, 3, 2, 1, 0.5)
  t1 <- 0:6
  expect_equal(superposition(conc=c1, time=t1, tau=3, n.tau=2),
               data.frame(conc=c(3, 3, 3, 3.5),
                          time=0:3))
  # With half-life extrapolation
  v1 <- superposition(conc=c1, time=t1, tau=3, n.tau=3)
  expect_equal(v1,
               data.frame(conc=c(3.5, 3.25, 3.125, 3.5625),
                          time=0:3))
  # To steady-state
  v2 <- superposition(conc=c1, time=t1, tau=3, n.tau=Inf)
  expect_equal(v2,
               data.frame(conc=c(3.571, 3.286, 3.143, 3.571),
                          time=0:3),
               tolerance=0.001)

  # all zeros input gives all zeros output
  expect_equal(superposition(conc=rep(0, 6), time=0:5, tau=24),
               data.frame(conc=0, time=c(0:5, 24)))

  # Times for output are a collation of 0, tau, dose.times, and
  # additional.times with time included and shifted for all the
  # dose.times.
  expect_equal(superposition(conc=rep(0, 6), time=0:5, tau=24,
                             additional.times=2.5),
               data.frame(conc=0, time=c(0, 1, 2, 2.5, 3, 4, 5, 24)))
  expect_equal(superposition(conc=rep(0, 6), time=0:5, tau=25,
                             additional.times=2.5),
               data.frame(conc=0, time=c(0, 1, 2, 2.5, 3, 4, 5, 25)))
  expect_equal(superposition(conc=rep(0, 6), time=0:5, tau=24,
                             additional.times=c(2.5, 3.5)),
               data.frame(conc=0, time=c(0, 1, 2, 2.5, 3, 3.5, 4, 5, 24)))
  expect_equal(superposition(conc=rep(0, 6), time=0:5, tau=24, dose.times=c(0, 1),
                             additional.times=c(2.5, 3.5)),
               data.frame(conc=0, time=c(0, 1, 2, 2.5, 3, 3.5, 4, 5, 6, 24)))
  expect_equal(superposition(conc=rep(0, 6), time=0:5, tau=24,
                             dose.times=c(0, 0.5),
                             additional.times=c(2.5, 3.5)),
               data.frame(conc=0,
                          time=c(0, 0.5, 1, 1.5, 2, 2.5, 3, 3.5, 4, 4.5,
                            5, 5.5, 24)))

  # Dose scaling
  v1 <-
    superposition(
      conc=c1, time=t1, dose.input=1, tau=24,
      dose.times=c(0, 0.5),
      dose.amount=1,
      additional.times=c(2.5, 3.5)
    )
  expect_equal(v1,
               data.frame(conc=c(
                            4.6047e-06,
                            0.5,
                            1.5,
                            2.5,
                            3.5,
                            4.5,
                            5.5,
                            5.4495,
                            4.4495,
                            3.4142,
                            2.4142,
                            1.7071,
                            1.2071,
                            0.85355,
                            4.6047e-06),
                          time=c(0, 0.5, 1, 1.5, 2, 2.5, 3, 3.5, 4, 4.5,
                            5, 5.5, 6, 6.5, 24)),
               tolerance=0.001,
               info="Dose scaling with matching input and output doses")

  v2 <-
    superposition(
      conc=c1, time=t1, dose.input=1, tau=24,
      dose.times=c(0, 0.5),
      dose.amount=c(0.5, 5),
      additional.times=c(2.5, 3.5)
    )
  expect_equal(v2,
               data.frame(conc=c(
                            1.444e-05,
                            0.25,
                            3,
                            5.75,
                            8.5,
                            11.25,
                            14,
                            16.22,
                            13.25,
                            10.71,
                            7.571,
                            5.354,
                            3.786,
                            2.677,
                            1.444e-05),
                          time=c(0, 0.5, 1, 1.5, 2, 2.5, 3, 3.5, 4, 4.5,
                            5, 5.5, 6, 6.5, 24)),
               tolerance=0.001,
               info="Dose scaling with different input and output doses")

  expect_warning(v3 <- superposition(conc=c(0, 2, 3, 5, 6, 3, 1, 0),
                                     time=c(0, 0.5, 1, 1.5, 2, 8, 12, 24),
                                     tau=24),
                 regexp="Too few points for half-life calculation \\(min.hl.points=3 with only 2 points\\)")
  expect_equal(v3,
               data.frame(conc=NA, time=c(0, 0.5, 1, 1.5, 2, 8, 12, 24)),
               info="Uncalculable lambda.z with extrapolation to steady-state gives NA conc")

  expect_equal(
    superposition(
      conc=c(0, 2, 3, 5, 6, 3, 1, 0),
      time=c(0, 0.5, 1, 1.5, 2, 8, 12, 24),
      tau=24,
      lambda.z=1, clast.pred=1, tlast=12
    ),
    data.frame(
      conc=c(6.144e-6, 2, 3, 5, 6, 3, 1, 6.144e-6),
      time=c(0, 0.5, 1, 1.5, 2, 8, 12, 24)
    ),
    tolerance=0.001,
    info="Uncalculable lambda.z with extrapolation to steady-state with lambda.z given, gives conc values"
  )
})

test_that("PKNCAconc superposition", {
  myconc <- PKNCAconc(conc~time|ID,
                      data=data.frame(ID=rep(1:2, each=7),
                                      conc=rep(c(0, 1, 2, 3, 2, 1, 0.5), 2),
                                      time=rep(0:6, 2)))
  expect_equal(
    superposition(myconc, tau=3, n.tau=2),
    tibble::tibble(
      ID=rep(1:2, each=4),
      conc=rep(c(3, 3, 3, 3.5), 2),
      time=rep(0:3, 2)
    )
  )
})

test_that("PKNCAconc expected errors", {
  tmp.conc <- generate.conc(nsub=1, ntreat=1, time.points=0:24)
  tmp.conc$foo <- "A"
  expect_error(
    PKNCAconc(conc~time, volume="foo", data=tmp.conc),
    regexp="Volume must be numeric"
  )
  expect_error(
    PKNCAconc(conc~time, duration="foo", data=tmp.conc),
    regexp="duration must be numeric without missing (NA) or infinite values, and all values must be >= 0",
    fixed=TRUE
  )
})

test_that("PKNCAconc", {
  tmp.conc <- generate.conc(nsub=5, ntreat=2, time.points=0:24)
  tmp.conc.analyte <- generate.conc(nsub=5, ntreat=2, time.points=0:24,
                                    nanalytes=2)
  tmp.conc.study <- generate.conc(nsub=5, ntreat=2, time.points=0:24,
                                  nstudies=2)
  tmp.conc.analyte.study <- generate.conc(nsub=5, ntreat=2, time.points=0:24,
                                          nanalytes=2, nstudies=2)
  # Data exists
  expect_error(PKNCAconc(data.frame()),
               regexp="data must have at least one row.",
               info="PKNCAconc requires data")

  # Variables present
  expect_error(
    PKNCAconc(tmp.conc, formula=XXX~time|treatment+ID),
    regexp="All of the variables in the formula must be in the data.  Missing: XXX",
    info="All formula parameters must be in the data (LHS)"
  )
  expect_error(
    PKNCAconc(tmp.conc, formula=conc~XXX|treatment+ID),
    regexp="All of the variables in the formula must be in the data.  Missing: XXX",
    info="All formula parameters must be in the data (RHS)"
  )
  expect_error(
    PKNCAconc(tmp.conc, formula=conc~time|XXX+ID),
    regexp="All of the variables in the formula must be in the data.  Missing: XXX",
    info="All formula parameters must be in the data (groups)"
  )

  # Number of variables
  expect_error(PKNCAconc(tmp.conc, formula=conc+ID~time|treatment+ID),
               regexp="The left hand side of the formula must have exactly one variable",
               info="The right number of parameters in the formula (LHS)")
  expect_error(PKNCAconc(tmp.conc, formula=conc~time+ID|treatment+ID),
               regexp="The right hand side of the formula \\(excluding groups\\) must have exactly one variable",
               info="The right number of parameters in the formula (RHS)")

  # Subject assignment
  expect_equal(PKNCAconc(tmp.conc.analyte, formula=conc~time|treatment+ID/analyte),
               PKNCAconc(tmp.conc.analyte, formula=conc~time|treatment+ID/analyte, subject="ID"))
  expect_error(PKNCAconc(tmp.conc.analyte, formula=conc~time|treatment+ID/analyte, subject=5),
               regexp="subject must be a character string")
  expect_error(PKNCAconc(tmp.conc.analyte, formula=conc~time|treatment+ID/analyte, subject=c("", "foo")),
               regexp="subject must be a scalar")
  expect_error(PKNCAconc(tmp.conc.analyte, formula=conc~time|treatment+ID/analyte, subject="foo"),
               regexp="The subject parameter must map to a name in the data")

  # Keys must be unique
  expect_error(PKNCAconc(tmp.conc.analyte, formula=conc~time|treatment+ID),
               regexp="Rows that are not unique per group and time",
               info="Duplicated key rows")

  expect_equal(
    PKNCAconc(
      tmp.conc.analyte,
      formula=conc~time|treatment+ID/analyte
    ),
    PKNCAconc(
      dplyr::as_tibble(tmp.conc.analyte),
      formula=conc~time|treatment+ID/analyte
    ),
    info="tibble and data.frame classes both work and create identical objects"
  )
})

test_that("PKNCAconc with input other than data.frames", {
  tmp <- structure(list(), class="foo")
  captured_message <- tryCatch(as.data.frame(tmp), error=function(e) e$message)
  expect_error(PKNCAconc(tmp, formula=conc~time|treatment+ID),
               regexp=captured_message,
               fixed=TRUE,
               info="Attempt to coerce into a data.frame.")
})

test_that("model frame and parameter extractions", {
  tmp.conc <- generate.conc(nsub=5, ntreat=2, time.points=0:24)
  tmp_conc_single <- generate.conc(nsub=1, ntreat=1, time.points=0:24)

  expect_equal(model.frame(PKNCAconc(tmp.conc, formula=conc~time|treatment+ID)),
               tmp.conc[,c("conc", "time", "treatment", "ID")],
               info="model.frame.PKNCAconc extracts the correct components")
  expect_equal(getDepVar(PKNCAconc(tmp.conc, formula=conc~time|treatment+ID)),
               tmp.conc$conc,
               info="getDepVar.PKNCAconc extracts the correct component")
  expect_equal(getIndepVar(PKNCAconc(tmp.conc, formula=conc~time|treatment+ID)),
               tmp.conc$time,
               info="getIndepVar.PKNCAconc extracts the correct component")
  expect_equal(getGroups.PKNCAconc(PKNCAconc(tmp.conc, formula=conc~time|treatment+ID)),
               tmp.conc[,c("treatment", "ID")],
               info="getGroups.PKNCAconc extracts the correct components")

  tmp.conc <- generate.conc(nsub=5, ntreat=1, time.points=0:24)
  expect_equal(getGroups.PKNCAconc(PKNCAconc(tmp.conc, formula=conc~time|ID)),
               tmp.conc[,"ID", drop=FALSE],
               info="getGroups.PKNCAconc returns a data.frame even with a single grouping level")
  # Levels
  tmp.conc <- generate.conc(nsub=5, ntreat=2, time.points=0:24)
  expect_equal(getGroups.PKNCAconc(PKNCAconc(tmp.conc, formula=conc~time|treatment+ID), level=1),
               tmp.conc[,"treatment", drop=FALSE],
               info="getGroups.PKNCAconc the correct level (numeric scalar)")
  expect_equal(getGroups.PKNCAconc(PKNCAconc(tmp.conc, formula=conc~time|treatment+ID), level=-1),
               tmp.conc[,"ID", drop=FALSE],
               info="getGroups.PKNCAconc the correct level (negative numeric scalar)")
  expect_equal(getGroups.PKNCAconc(PKNCAconc(tmp.conc, formula=conc~time|treatment+ID), level=1:2),
               tmp.conc[,c("treatment", "ID"), drop=FALSE],
               info="getGroups.PKNCAconc the correct level (numeric vector)")
  expect_equal(getGroups.PKNCAconc(PKNCAconc(tmp.conc, formula=conc~time|treatment+ID), level=2:1),
               tmp.conc[,c("ID", "treatment"), drop=FALSE],
               info="getGroups.PKNCAconc the correct level (numeric vector)")
  expect_equal(getGroups.PKNCAconc(PKNCAconc(tmp.conc, formula=conc~time|treatment+ID), level="ID"),
               tmp.conc[,"ID", drop=FALSE],
               info="getGroups.PKNCAconc the correct level (character string scalar)")
  expect_equal(getGroups.PKNCAconc(PKNCAconc(tmp.conc, formula=conc~time|treatment+ID), level=c("ID", "treatment")),
               tmp.conc[,c("ID", "treatment"), drop=FALSE],
               info="getGroups.PKNCAconc the correct level (character string vector)")
  expect_error(getGroups.PKNCAconc(PKNCAconc(tmp.conc, formula=conc~time|treatment+ID), level="foo"),
               regexp="Not all levels are listed in the group names",
               info="getGroups.PKNCAconc gives an error if a group name is not present")

  expect_equal(
    group_vars.PKNCAconc(PKNCAconc(tmp.conc, formula=conc~time|treatment+ID)),
    c("treatment", "ID")
  )
  expect_equal(
    group_vars.PKNCAconc(PKNCAconc(tmp_conc_single, formula=conc~time)),
    character(0),
    info="Ungrouped data works with group_vars"
  )
})

test_that("print.PKNCAconc", {
  tmp.conc <- generate.conc(nsub=2, ntreat=2, time.points=0:24)
  myconc <- PKNCAconc(tmp.conc, formula=conc~time|treatment+ID)

  expect_output(print.PKNCAconc(myconc),
                regexp="Formula for concentration:
 conc ~ time | treatment + ID
                With 2 subjects defined in the 'ID' column.
                Nominal time column is not specified.

                First 6 rows of concentration data:
                treatment ID time      conc
                Trt 1  1    0 0.0000000
                Trt 1  1    1 0.7052248
                Trt 1  1    2 0.7144320
                Trt 1  1    3 0.8596094
                Trt 1  1    4 0.9998126
                Trt 1  1    5 0.7651474",
                info="Generic print.PKNCAconc works")
  expect_output(print.PKNCAconc(myconc, n=0),
                regexp="Formula for concentration:
 conc ~ time | treatment + ID
With 2 subjects defined in the 'ID' column.
Nominal time column is not specified.",
                info="print.PKNCAconc respects the n argument.")
  expect_output(print.PKNCAconc(myconc, n=-98),
                regexp="Formula for concentration:
 conc ~ time | treatment + ID
                With 2 subjects defined in the 'ID' column.
                Nominal time column is not specified.

                First 2 rows of concentration data:
                treatment ID time      conc
                Trt 1  1    0 0.0000000
                Trt 1  1    1 0.7052248",
                info="print.PKNCAconc accurately uses negative n argument.")

  tmp.conc <- generate.conc(nsub=1, ntreat=1, time.points=0:24)
  myconc <- PKNCAconc(tmp.conc, formula=conc~time)
  expect_output(
    print(myconc),
    regexp="As a single-subject dataset"
  )
  expect_output(
    print(myconc, summarize=TRUE),
    regexp="No groups\\."
  )
  expect_output(
    print(myconc, n=1e6),
    regexp="Data for concentration"
  )

  myconc <- PKNCAconc(tmp.conc, formula=conc~time|treatment+ID, time.nominal="time")
  expect_output(
    print(myconc),
    regexp = "Nominal time column is: time"
  )
})

test_that("summary.PKNCAconc", {
  tmp.conc <- generate.conc(nsub=2, ntreat=2, time.points=0:24)
  myconc <- PKNCAconc(tmp.conc, formula=conc~time|treatment+ID)

  expect_output(summary(myconc),
                regexp="Formula for concentration:
 conc ~ time | treatment + ID
                With 2 subjects defined in the 'ID' column.
                Nominal time column is not specified.

                Group summary:
                Group Name Count
                treatment     2
                ID     4",
                info="Generic summary.PKNCAconc works.")
})

test_that("PKNCAconc with exclusions", {
  tmp.conc <- generate.conc(nsub=2, ntreat=2, time.points=0:24)
  tmp.conc$excl <- NA_character_
  myconc <- PKNCAconc(tmp.conc, formula=conc~time|treatment+ID, exclude="excl")
  expect_equal(
    myconc,
    structure(
      list(
        data=
          cbind(
            tmp.conc,
            volume=NA_real_,
            duration=0
          ),
        formula=conc~time|treatment+ID,
        columns=
          list(
            concentration="conc",
            time="time",
            groups=
              list(
                group_vars=c("treatment", "ID"),
                group_analyte=character()
              ),
            subject="ID",
            exclude="excl",
            volume="volume",
            duration="duration"
          ),
        units = list()
      ),
      class=c("PKNCAconc", "list")
    )
  )
})

test_that("PKNCAconc with duration", {
  tmp.conc <- generate.conc(nsub=2, ntreat=2, time.points=0:24)
  tmp.conc$duration_test <- 0.1
  myconc <- PKNCAconc(tmp.conc,
                      formula=conc~time|treatment+ID,
                      duration="duration_test")
  expect_equal(
    myconc,
    structure(
      list(
        data=
          cbind(
            tmp.conc,
            data.frame(
              exclude=NA_character_,
              volume=NA_real_,
              stringsAsFactors=FALSE
            )
          ),
        formula=conc~time|treatment+ID,
        columns=
          list(
            concentration="conc",
            time="time",
            groups=
              list(
                group_vars=c("treatment", "ID"),
                group_analyte=character()
              ),
            subject="ID",
            exclude="exclude",
            volume="volume",
            duration="duration_test"
          ),
        units = list()
      ),
      class=c("PKNCAconc", "list")
    )
  )
})

test_that("PKNCAconc with nominal time added", {
  tmp.conc <- generate.conc(nsub=2, ntreat=2, time.points=0:24)
  tmp.conc$tnom <- tmp.conc$time
  myconc <- PKNCAconc(tmp.conc, formula=conc~time|treatment+ID, time.nominal="tnom")
  expect_equal(
    myconc,
    structure(
      list(
        data=
          cbind(
            tmp.conc,
            data.frame(
              exclude=NA_character_,
              volume=NA_real_,
              duration=0,
              stringsAsFactors=FALSE
            )
          ),
        formula=conc~time|treatment+ID,
        columns=
          list(
            concentration="conc",
            time="time",
            groups=
              list(
                group_vars=c("treatment", "ID"),
                group_analyte=character()
              ),
            subject="ID",
            exclude="exclude",
            volume="volume",
            duration="duration",
            time.nominal="tnom"
          ),
        units = list()
      ),
      class=c("PKNCAconc", "list")
    )
  )
  expect_equal(
    PKNCAconc(tmp.conc, formula=conc~time|treatment+ID, time.nominal="foo"),
    structure(
      list(data=cbind(tmp.conc,
                      data.frame(exclude=NA_character_,
                                 volume=NA_real_,
                                 duration=0,
                                 foo=NA,
                                 stringsAsFactors=FALSE)),
           formula=conc~time|treatment+ID,
           columns=
             list(
               concentration="conc",
               time="time",
               groups=
                 list(
                   group_vars=c("treatment", "ID"),
                   group_analyte=character()
                 ),
               subject="ID",
               exclude="exclude",
               volume="volume",
               duration="duration",
               time.nominal="foo"
             ),
           units = list()
      ),
      class=c("PKNCAconc", "list"))
  )
})

test_that("PKNCAconc with volume added", {
  tmp.conc <- generate.conc(nsub=2, ntreat=2, time.points=0:24)
  tmp.conc$vol <- seq_len(nrow(tmp.conc))
  myconc <- PKNCAconc(tmp.conc, formula=conc~time|treatment+ID, volume="vol")
  expect_equal(
    myconc,
    structure(
      list(data=cbind(tmp.conc,
                      data.frame(exclude=NA_character_,
                                 duration=0,
                                 stringsAsFactors=FALSE)),
           formula=conc~time|treatment+ID,
           columns=list(
             concentration="conc",
             time="time",
             groups=
               list(
                 group_vars=c("treatment", "ID"),
                 group_analyte=character()
               ),
             subject="ID",
             exclude="exclude",
             volume="vol",
             duration="duration"
           ),
           units = list()
      ),
      class=c("PKNCAconc", "list"))
  )
  myconc_manual_vol <- PKNCAconc(tmp.conc, formula=conc~time|treatment+ID, volume=2)
  expect_equal(
    myconc_manual_vol,
    structure(
      list(data=cbind(tmp.conc,
                      data.frame(exclude=NA_character_,
                                 volume=2,
                                 duration=0,
                                 stringsAsFactors=FALSE)),
           formula=conc~time|treatment+ID,
           columns=
             list(
               concentration="conc",
               time="time",
               groups=
                 list(
                   group_vars=c("treatment", "ID"),
                   group_analyte=character()
                 ),
               subject="ID",
               exclude="exclude",
               volume="volume",
               duration="duration"
             ),
           units = list()
      ),
      class=c("PKNCAconc", "list")
    )
  )
  myconc_manual_vol_vector <- PKNCAconc(tmp.conc, formula=conc~time|treatment+ID, volume=seq_len(nrow(tmp.conc)))
  expect_equal(
    myconc_manual_vol_vector,
    structure(
      list(
        data=
          cbind(
            tmp.conc,
            data.frame(
              exclude=NA_character_,
              volume=seq_len(nrow(tmp.conc)),
              duration=0
            )
          ),
        formula=conc~time|treatment+ID,
        columns=
          list(
            concentration="conc",
            time="time",
            groups=
              list(
                group_vars=c("treatment", "ID"),
                group_analyte=character()
              ),
            subject="ID",
            exclude="exclude",
            volume="volume",
            duration="duration"
          ),
        units = list()
      ),
      class=c("PKNCAconc", "list")
    )
  )
})

test_that("as.data.frame.PKNCAconc", {
  tmp_conc <- generate.conc(nsub=1, ntreat=1, time.points=0:24)
  result_conc <- tmp_conc
  result_conc$exclude <- NA_character_
  result_conc$volume <- NA_real_
  result_conc$duration <- 0
  expect_equal(
    as.data.frame(PKNCAconc(conc~time, data=tmp_conc)),
    result_conc
  )
})

test_that("PKNCAconc with sparse data", {
  d_sparse <-
    data.frame(
      id = c(1L, 2L, 3L, 1L, 2L, 3L, 1L, 2L, 3L, 4L, 5L, 6L, 4L, 5L, 6L, 7L, 8L, 9L, 7L, 8L, 9L),
      conc = c(0, 0, 0,  1.75, 2.2, 1.58, 4.63, 2.99, 1.52, 3.03, 1.98, 2.22, 3.34, 1.3, 1.22, 3.54, 2.84, 2.55, 0.3, 0.0421, 0.231),
      time = c(0, 0, 0, 1, 1, 1, 6, 6, 6, 2, 2, 2, 10, 10, 10, 4, 4, 4, 24, 24, 24),
      dose = c(100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100)
    )
  o_conc_sparse <- PKNCAconc(d_sparse, conc~time|id, sparse=TRUE)
  expect_true("data_sparse" %in% names(o_conc_sparse))
  expect_false("data" %in% names(o_conc_sparse))

  d_sparse_aug <- d_sparse
  d_sparse_aug$exclude <- NA_character_
  d_sparse_aug$volume <- NA_real_
  d_sparse_aug$duration <- 0
  expect_equal(
    o_conc_sparse$data_sparse,
    d_sparse_aug
  )
})

test_that("print.PKNCAconc with sparse data", {
  d_sparse <-
    data.frame(
      id = c(1L, 2L, 3L, 1L, 2L, 3L, 1L, 2L, 3L, 4L, 5L, 6L, 4L, 5L, 6L, 7L, 8L, 9L, 7L, 8L, 9L),
      conc = c(0, 0, 0,  1.75, 2.2, 1.58, 4.63, 2.99, 1.52, 3.03, 1.98, 2.22, 3.34, 1.3, 1.22, 3.54, 2.84, 2.55, 0.3, 0.0421, 0.231),
      time = c(0, 0, 0, 1, 1, 1, 6, 6, 6, 2, 2, 2, 10, 10, 10, 4, 4, 4, 24, 24, 24),
      dose = c(100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100)
    )
  o_conc_sparse <- PKNCAconc(d_sparse, conc~time|id, sparse=TRUE)
  expect_output(print.PKNCAconc(o_conc_sparse),
                regexp="Formula for concentration:
 conc ~ time | id
Data are sparse PK.
With 9 subjects defined in the 'id' column.
Nominal time column is not specified.

First 6 rows of concentration data:
 id conc time dose exclude volume duration
  1 0.00    0  100    <NA>     NA        0
  2 0.00    0  100    <NA>     NA        0
  3 0.00    0  100    <NA>     NA        0
  1 1.75    1  100    <NA>     NA        0
  2 2.20    1  100    <NA>     NA        0
  3 1.58    1  100    <NA>     NA        0"
  )
})

test_that("Test uniqueness after excluding rows (#298)", {
  repeated_with_exclusion <-
    data.frame(
      conc = 1,
      time = c(0, 0),
      id = 1,
      exclude = c(NA, "duplicate")
    )

  expect_error(
    PKNCAconc(repeated_with_exclusion, conc~time|id),
    regexp="Rows that are not unique per group and time.*concentration"
  )
  expect_s3_class(
    PKNCAconc(repeated_with_exclusion, conc~time|id, exclude = "exclude"),
    class = "PKNCAconc"
  )
})

test_that("PKNCAconc units (#336)", {
  d <- data.frame(conc = 1, time = 0, concu_x = "A", timeu_x = "B", amountu_x = "C")

  # No units
  o_conc <- PKNCAconc(data = d, conc~time)
  expect_equal(o_conc$units, list())
  expect_null(o_conc$columns$concu)
  expect_null(o_conc$columns$timeu)
  expect_null(o_conc$columns$amountu)

  # Each unit column individually
  o_conc <- PKNCAconc(data = d, conc~time, concu = "concu_x")
  expect_equal(o_conc$units, list())
  expect_equal(o_conc$columns$concu, structure("concu_x", unit_type = "column"))
  expect_null(o_conc$columns$timeu)
  expect_null(o_conc$columns$amountu)

  o_conc <- PKNCAconc(data = d, conc~time, timeu = "timeu_x")
  expect_equal(o_conc$units, list())
  expect_null(o_conc$columns$concu)
  expect_equal(o_conc$columns$timeu, structure("timeu_x", unit_type = "column"))
  expect_null(o_conc$columns$amountu)

  o_conc <- PKNCAconc(data = d, conc~time, amountu = "amountu_x")
  expect_equal(o_conc$units, list())
  expect_null(o_conc$columns$concu)
  expect_null(o_conc$columns$timeu)
  expect_equal(o_conc$columns$amountu, structure("amountu_x", unit_type = "column"))

  # Each unit as a value, not a column
  o_conc <- PKNCAconc(data = d, conc~time, concu = "concu_y")
  expect_equal(o_conc$units, list(concu = structure("concu_y", unit_type = "value")))
  expect_null(o_conc$columns$concu)
  expect_null(o_conc$columns$timeu)
  expect_null(o_conc$columns$amountu)

  o_conc <- PKNCAconc(data = d, conc~time, timeu = "timeu_y")
  expect_equal(o_conc$units, list(timeu = structure("timeu_y", unit_type = "value")))
  expect_null(o_conc$columns$concu)
  expect_null(o_conc$columns$timeu)
  expect_null(o_conc$columns$amountu)

  o_conc <- PKNCAconc(data = d, conc~time, amountu = "amountu_y")
  expect_equal(o_conc$units, list(amountu = structure("amountu_y", unit_type = "value")))
  expect_null(o_conc$columns$concu)
  expect_null(o_conc$columns$timeu)
  expect_null(o_conc$columns$amountu)

  # Preferred units
  expect_error(
    PKNCAconc(data = d, conc~time, concu_pref = "concu_z"),
    regexp = "Preferred units may not be set unless original units are set: concu_pref"
  )
  o_conc <- PKNCAconc(data = d, conc~time, concu = "concu_y", concu_pref = "concu_z")
  expect_equal(
    o_conc$units,
    list(
      concu = structure("concu_y", unit_type = "value"),
      concu_pref = structure("concu_z", unit_type = "value")
    )
  )
  expect_null(o_conc$columns$concu)
  expect_null(o_conc$columns$timeu)
  expect_null(o_conc$columns$amountu)

  o_conc <- PKNCAconc(data = d, conc~time, concu = "concu_x", concu_pref = "concu_z")
  expect_equal(o_conc$units, list(concu_pref = structure("concu_z", unit_type = "value")))
  expect_equal(
    o_conc$columns$concu,
    structure("concu_x", unit_type = "column")
  )
  expect_null(o_conc$columns$timeu)
  expect_null(o_conc$columns$amountu)

  o_conc <- PKNCAconc(data = d, conc~time, timeu = "timeu_x", timeu_pref = "timeu_z")
  expect_equal(o_conc$units, list(timeu_pref = structure("timeu_z", unit_type = "value")))
  expect_null(o_conc$columns$concu)
  expect_equal(o_conc$columns$timeu, structure("timeu_x", unit_type = "column"))
  expect_null(o_conc$columns$amountu)

  o_conc <- PKNCAconc(data = d, conc~time, amountu = "amountu_x", amountu_pref = "amountu_z")
  expect_equal(o_conc$units, list(amountu_pref = structure("amountu_z", unit_type = "value")))
  expect_null(o_conc$columns$concu)
  expect_null(o_conc$columns$timeu)
  expect_equal(
    o_conc$columns$amountu,
    structure("amountu_x", unit_type = "column")
  )
})

# Setup the default options
.PKNCAEnv <- new.env(parent=emptyenv())
assign("options", NULL, envir=.PKNCAEnv)
assign("summary", list(), envir=.PKNCAEnv)
assign("interval.cols", list(), envir=.PKNCAEnv)

#' Add columns for calculations within PKNCA intervals
#'
#' @param name The column name as a character string
#' @param FUN The function to run (as a character string) or `NA` if the
#'   parameter is automatically calculated when calculating another parameter.
#' @param values Valid values for the column
#' @param depends Character vector of columns that must be run before this
#'   column.
#' @param desc A human-readable description of the parameter (<=40 characters to
#'   comply with SDTM)
#' @param sparse Is the calculation for sparse PK?
#' @param unit_type The type of units to use for assigning and converting units.
#' @param pretty_name The name of the parameter to use for printing in summary
#'   tables with units.  (If an analysis does not include units, then the normal
#'   name is used.)
#' @param formalsmap A named list mapping parameter names in the function call
#'   to NCA parameter names.  See the details for information on use of
#'   `formalsmap`.
#' @param datatype The type of data used for the calculation
#' @returns NULL (Calling this function has a side effect of changing the
#'   available intervals for calculations)
#'
#' @details The `formalsmap` argument enables mapping some alternate formal
#' argument names to parameters.  It is used to generalize functions that may
#' use multiple similar arguments (such as the variants of mean residence time).
#' The names of the list should correspond to function formal parameter names
#' and the values should be one of the following:
#'
#' \itemize{
#'   \item{For the current interval:}
#'   \describe{
#'     \item{character strings of NCA parameter name}{The value of the parameter calculated for the current interval.}
#'     \item{"conc"}{Concentration measurements for the current interval.}
#'     \item{"time"}{Times associated with concentration measurements for the current interval (values start at 0 at the beginning of the current interval).}
#'     \item{"volume"}{Volume associated with concentration measurements for the current interval (typically applies for excretion parameters like urine).}
#'     \item{"duration.conc"}{Durations associated with concentration measurements for the current interval.}
#'     \item{"dose"}{Dose amounts assocuated with the current interval.}
#'     \item{"time.dose"}{Time of dose start associated with the current interval (values start at 0 at the beginning of the current interval).}
#'     \item{"duration.dose"}{Duration of dose (typically infusion duration) for doses in the current interval.}
#'     \item{"route"}{Route of dosing for the current interval.}
#'     \item{"start"}{Time of interval start.}
#'     \item{"end"}{Time of interval end.}
#'     \item{"options"}{PKNCA.options governing calculations.}
#'   }
#'   \item{For the current group:}
#'   \describe{
#'     \item{"conc.group"}{Concentration measurements for the current group.}
#'     \item{"time.group"}{Times associated with concentration measurements for the current group (values start at 0 at the beginning of the current interval).}
#'     \item{"volume.group"}{Volume associated with concentration measurements for the current interval (typically applies for excretion parameters like urine).}
#'     \item{"duration.conc.group"}{Durations assocuated with concentration measurements for the current group.}
#'     \item{"dose.group"}{Dose amounts assocuated with the current group.}
#'     \item{"time.dose.group"}{Time of dose start associated with the current group (values start at 0 at the beginning of the current interval).}
#'     \item{"duration.dose.group"}{Duration of dose (typically infusion duration) for doses in the current group.}
#'     \item{"route.group"}{Route of dosing for the current group.}
#'   }
#' }
#' @examples
#' \dontrun{
#' add.interval.col("cmax",
#'                  FUN="pk.calc.cmax",
#'                  values=c(FALSE, TRUE),
#'                  unit_type="conc",
#'                  pretty_name="Cmax",
#'                  desc="Maximum observed concentration")
#' add.interval.col("cmax.dn",
#'                  FUN="pk.calc.dn",
#'                  values=c(FALSE, TRUE),
#'                  unit_type="conc_dosenorm",
#'                  pretty_name="Cmax (dose-normalized)",
#'                  desc="Maximum observed concentration, dose normalized",
#'                  formalsmap=list(parameter="cmax"),
#'                  depends="cmax")
#' }
#' @family Interval specifications
#' @export
add.interval.col <- function(name,
                             FUN,
                             values=c(FALSE, TRUE),
                             unit_type,
                             pretty_name,
                             depends=NULL,
                             desc="",
                             sparse=FALSE,
                             formalsmap=list(),
                             datatype=c("interval",
                               "individual",
                               "population")) {
  # Check inputs
  if (!is.character(name)) {
    stop("name must be a character string")
  } else if (length(name) != 1) {
    stop("name must have length == 1")
  }
  if (length(FUN) != 1) {
    stop("FUN must have length == 1")
  } else if (!(is.character(FUN) | is.na(FUN))) {
    stop("FUN must be a character string or NA")
  }
  if (!is.null(depends)) {
    if (!is.character(depends)) {
      stop("'depends' must be NULL or a character vector")
    }
  }
  checkmate::assert_logical(sparse, any.missing=FALSE, len=1)
  unit_type <-
    match.arg(
      unit_type,
      choices=c(
        "unitless", "fraction", "%", "count",
        "time", "inverse_time",
        "amount", "amount_dose",
        "conc", "conc_dosenorm",
        "dose",
        "volume",
        "auc", "aumc",
        "auc_dosenorm", "aumc_dosenorm",
        "clearance", "renal_clearance"
      )
    )
  stopifnot("pretty_name must be a scalar"=length(pretty_name) == 1)
  stopifnot("pretty_name must be a character"=is.character(pretty_name))
  stopifnot("pretty_name must not be an empty string"=nchar(pretty_name) > 0)
  datatype <- match.arg(datatype)
  if (!(datatype %in% "interval")) {
    stop("Only the 'interval' datatype is currently supported.")
  }
  if (length(desc) != 1) {
    stop("desc must have length == 1")
  } else if (!is.character(desc)) {
    stop("desc must be a character string")
  }
  if (!is.list(formalsmap)) {
    stop("formalsmap must be a list")
  } else if (length(formalsmap) > 0 &
             is.null(names(formalsmap))) {
    stop("formalsmap must be a named list")
  } else if (length(formalsmap) > 0 &
             is.na(FUN)) {
    stop("formalsmap may not be given when FUN is NA.")
  } else if (!all(nchar(names(formalsmap)) > 0)) {
    stop("All formalsmap elements must be named")
  }
  # Ensure that the function exists
  if (!is.na(FUN) &&
      length(utils::getAnywhere(FUN)$objs) == 0) {
    stop("The function named '", FUN, "' is not defined.  Please define the function before calling add.interval.col.")
  }
  if (!is.na(FUN) &
      length(formalsmap) > 0) {
    # Ensure that the formalsmap parameters are all in the list of
    # formal arguments to the function.
    if (!all(names(formalsmap) %in% names(formals(utils::getAnywhere(FUN)$objs[[1]])))) {
      stop("All names for the formalsmap list must be arguments to the function.")
    }
  }
  current <- get("interval.cols", envir=.PKNCAEnv)
  current[[name]] <-
    list(
      FUN=FUN,
      values=values,
      unit_type=unit_type,
      pretty_name=pretty_name,
      desc=desc,
      sparse=sparse,
      formalsmap=formalsmap,
      depends=depends,
      datatype=datatype
    )
  assign("interval.cols", current, envir=.PKNCAEnv)
}

#' Sort the interval columns by dependencies.
#'
#' Columns are always to the right of columns that they depend on.
sort.interval.cols <- function() {
  current <- get("interval.cols", envir=.PKNCAEnv)
  # Only sort if necessary
  sort_order <- get0("interval.cols_sorted", envir=.PKNCAEnv)
  if (identical(sort_order, names(current))) {
    # It is already sorted
    return(sort_order)
  }
  # Build a dependency tree
  myorder <- rep(NA, length(current))
  names(myorder) <- names(current)
  nextnum <- 1
  while (any(is.na(myorder))) {
    for (nextorder in seq_along(myorder)[is.na(myorder)]) {
      if (length(current[[nextorder]]$depends) == 0) {
        # If it doesn't depend on anything then it can go next in order.
        myorder[nextorder] <- nextnum
        nextnum <- nextnum + 1
      } else {
        # If all of its dependencies already have values, then it can be next.
        deps <- unique(unlist(current[[nextorder]]$depends))
        missing_deps <- deps[!(deps %in% names(myorder))]
        if (length(missing_deps) > 0) {
          stop(
            "Invalid dependencies for interval column (please report this as a bug): ",
            names(myorder)[nextorder],
            " The following dependencies are missing: ",
            paste(missing_deps, collapse=", ")
          )
        }
        if (!any(is.na(myorder[deps]))) {
          myorder[nextorder] <- nextnum
          nextnum <- nextnum + 1
        }
      }
    }
  }
  current <- current[names(sort(myorder))]
  assign("interval.cols_sorted", names(current), envir=.PKNCAEnv)
  assign("interval.cols", current, envir=.PKNCAEnv)
  invisible(myorder)
}

#' Get the columns that can be used in an interval specification
#'
#' @returns A list with named elements for each parameter.  Each list element
#'   contains the parameter definition.
#' @seealso [check.interval.specification()] and the vignette "Selection of
#'   Calculation Intervals"
#' @examples
#' get.interval.cols()
#' @family Interval specifications
#' @export
get.interval.cols <- function() {
  sort.interval.cols()
  get("interval.cols", envir=.PKNCAEnv)
}

# Add the start and end interval columns
add.interval.col(
  "start",
  FUN = NA,
  values = as.numeric,
  unit_type="time",
  pretty_name="Interval Start",
  desc = "Starting time of the interval"
)
add.interval.col(
  "end",
  FUN = NA,
  values = as.numeric,
  unit_type="time",
  pretty_name="Interval End",
  desc = "Ending time of the interval (potentially infinity)"
)

# Save the original state
original_state <- get("interval.cols", envir=PKNCA:::.PKNCAEnv)

test_that("add.interval.col", {
  # Invalid inputs fail
  expect_error(
    add.interval.col(name=1),
    regexp="name must be a character string",
    info="interval column name must be a character string"
  )
  expect_error(
    add.interval.col(name=c("a", "b")),
    regexp="name must have length",
    info="interval column name must be a scalar character string"
  )

  expect_error(
    add.interval.col(name="a", FUN=c("a", "b")),
    regexp="FUN must have length == 1",
    info="interval column function must be a scalar character string or NA"
  )
  expect_error(
    add.interval.col(name="a", FUN=1),
    regexp="FUN must be a character string or NA",
    info="interval column function must be a character string or NA"
  )
  
  expect_error(
    add.interval.col(name="a", FUN=NA, datatype="interval", desc="test addition"),
    regexp='argument "unit_type" is missing, with no default'
  )
  expect_error(
    add.interval.col(name="a", FUN=NA, unit_type="foo", datatype="interval", desc="test addition"),
    regexp="should be one of .*inverse_time"
  )
  
  # pretty_name checks
  expect_error(
    add.interval.col(name="a", FUN=NA, unit_type="conc", pretty_name=1:2, datatype="interval", desc=1),
    regexp="pretty_name must be a scalar"
  )
  expect_error(
    add.interval.col(name="a", FUN=NA, unit_type="conc", pretty_name=1, datatype="interval", desc=1),
    regexp="pretty_name must be a character"
  )
  expect_error(
    add.interval.col(name="a", FUN=NA, unit_type="conc", pretty_name="", datatype="interval", desc=1),
    regexp="pretty_name must not be an empty string"
  )
  
  expect_error(
    add.interval.col(name="a", FUN=NA, unit_type="conc", pretty_name="a", datatype="individual"),
    regexp="Only the 'interval' datatype is currently supported.",
    info="interval column datatype must be 'interval'"
  )
  
  expect_error(
    add.interval.col(name="a", FUN=NA, unit_type="conc", pretty_name="a", datatype="interval", desc=1:2),
    regexp="desc must have length == 1",
    info="interval column description must be a scalar"
  )
  expect_error(
    add.interval.col(name="a", FUN=NA, unit_type="conc", pretty_name="a", datatype="interval", desc=1),
    regexp="desc must be a character string",
    info="interval column description must be a character scalar"
  )
  expect_error(
    add.interval.col(name="a", FUN=NA, depends = 1, unit_type="conc", pretty_name="a", datatype="interval", desc=1),
    regexp="'depends' must be NULL or a character vector",
    info="depends column must be a NULL or a character string"
  )
  
  expect_error(
    add.interval.col(name="a", FUN="this function does not exist", unit_type="conc", pretty_name="foo", datatype="interval", desc="test addition"),
    regexp="The function named '.*' is not defined.  Please define the function before calling add.interval.col.",
    info="interval column function must exist (or be NA)"
  )

  # formalsmap
  expect_error(
    add.interval.col(name="a", FUN="mean", unit_type="conc", pretty_name="foo", formalsmap=NA),
    regexp="formalsmap must be a list"
  )
  expect_error(
    add.interval.col(name="a", FUN="mean", unit_type="conc", pretty_name="foo", formalsmap=list(1)),
    regexp="formalsmap must be a named list"
  )
  expect_error(
    add.interval.col(name="a", FUN=NA, unit_type="conc", pretty_name="foo", formalsmap=list(A="b")),
    regexp="formalsmap may not be given when FUN is NA",
    info="formalsmap cannot be used with FUN=NA"
  )
  expect_error(
    add.interval.col(name="a", FUN="mean", unit_type="conc", pretty_name="foo", formalsmap=list(A="a", "b")),
    regexp="All formalsmap elements must be named"
  )
  expect_error(
    add.interval.col(name="a", FUN="mean", unit_type="conc", pretty_name="a", formalsmap=list(y="a")),
    regexp="All names for the formalsmap list must be arguments to the function",
    info="formalsmap arguments must map to function arguments"
  )
  
  expect_equal(
    {
      add.interval.col(name="a", FUN=NA, unit_type="conc", pretty_name="a", datatype="interval", desc="test addition")
      get("interval.cols", PKNCA:::.PKNCAEnv)[["a"]]
    },
    list(
      FUN=NA,
      values=c(FALSE, TRUE),
      unit_type="conc",
      pretty_name="a",
      desc="test addition",
      sparse=FALSE,
      formalsmap=list(),
      depends=NULL,
      datatype="interval"
    ),
    info="interval column assignment works with FUN=NA"
  )
  expect_equal(
    {
      add.interval.col(name="a", FUN="mean", unit_type="conc", pretty_name="a", datatype="interval", desc="test addition")
      get("interval.cols", PKNCA:::.PKNCAEnv)[["a"]]
    },
    list(
      FUN="mean",
      values=c(FALSE, TRUE),
      unit_type="conc",
      pretty_name="a",
      desc="test addition",
      sparse=FALSE,
      formalsmap=list(),
      depends=NULL,
      datatype="interval"
    ),
    info="interval column assignment works with FUN=a character string"
  )
  expect_equal(
    {
      add.interval.col(name="a", FUN="mean", unit_type="conc", pretty_name="a", formalsmap=list(x="values"), desc="test addition")
      get("interval.cols", PKNCA:::.PKNCAEnv)[["a"]]
    },
    list(
      FUN="mean",
      values=c(FALSE, TRUE),
      unit_type="conc",
      pretty_name="a",
      desc="test addition",
      sparse=FALSE,
      formalsmap=list(x="values"),
      depends=NULL,
      datatype="interval"
    ),
    info="interval column assignment works with FUN=NA"
  )
})

# Reset the original state
assign("interval.cols", original_state, envir=PKNCA:::.PKNCAEnv)

test_that("fake parameters", {
  add.interval.col(
    name="fake_parameter",
    FUN="mean",
    unit_type="conc",
    pretty_name="a",
    formalsmap=list(x="values"),
    desc="test addition",
    depends="does_not_exist"
  )
  expect_error(
    sort.interval.cols(),
    regexp="Invalid dependencies for interval column (please report this as a bug): fake_parameter The following dependencies are missing: does_not_exist",
    fixed=TRUE
  )
})

# Reset the original state
assign("interval.cols", original_state, envir=PKNCA:::.PKNCAEnv)

# Calculate the simple parameters for PK.

#' Calculate the adjusted r-squared value
#'
#' @param r.sq The r-squared value
#' @param n The number of points
#' @return The numeric adjusted r-squared value
#' @export
adj.r.squared <- function(r.sq, n) {
  if (n <= 2) {
    rlang::warn(
      message = "n must be > 2 for adj.r.squared",
      class = "pknca_adjr2_2points"
    )
    structure(NA_real_, exclude="n must be > 2")
  } else {
    1-(1-r.sq)*(n-1)/(n-2)
  }
}

#' Determine maximum observed PK concentration
#'
#' @inheritParams assert_conc_time
#' @param check Run [assert_conc()]?
#' @return a number for the maximum concentration or NA if all concentrations
#'   are missing
#' @family NCA parameters for concentrations during the intervals
#' @export
pk.calc.cmax <- function(conc, check=TRUE) {
  if (check) {
    assert_conc(conc = conc)
  }
  if (length(conc) == 0 | all(is.na(conc))) {
    NA
  } else {
    max(conc, na.rm=TRUE)
  }
}
# Add the column to the interval specification
add.interval.col("cmax",
                 FUN="pk.calc.cmax",
                 values=c(FALSE, TRUE),
                 unit_type="conc",
                 pretty_name="Cmax",
                 desc="Maximum observed concentration",
                 depends=NULL)
PKNCA.set.summary(
  name="cmax",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)

#' @describeIn pk.calc.cmax Determine the minimum observed PK
#'   concentration
#' @family NCA parameters for concentrations during the intervals
#' @examples
#' conc_data <- Theoph[Theoph$Subject == 1,]
#' pk.calc.cmin(conc_data$conc)
#' @export
pk.calc.cmin <- function(conc, check=TRUE) {
  if (check) {
    assert_conc(conc=conc)
  }
  if (length(conc) == 0 | all(is.na(conc))) {
    NA
  } else {
    min(conc, na.rm=TRUE)
  }
}
# Add the column to the interval specification
add.interval.col("cmin",
                 FUN="pk.calc.cmin",
                 values=c(FALSE, TRUE),
                 unit_type="conc",
                 pretty_name="Cmin",
                 desc="Minimum observed concentration",
                 depends=NULL)
PKNCA.set.summary(
  name="cmin",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)

#' Determine time of maximum observed PK concentration
#'
#' Input restrictions are:
#' \enumerate{
#'   \item the `conc` and `time` must be the same length,
#'   \item the `time` may have no NAs,
#' }
#' `NA` will be returned if:
#' \enumerate{
#'   \item the length of `conc` and `time` is 0
#'   \item all `conc` is 0 or `NA`
#' }
#'
#' @inheritParams assert_conc_time
#' @inheritParams PKNCA.choose.option
#' @param first.tmax If there is more than time that matches the maximum
#'   concentration, should the first be considered as Tmax?  If not, then the
#'   last is considered Tmax.
#' @param check Run [assert_conc_time()]?
#' @returns The time of the maximum concentration
#' @examples
#' conc_data <- Theoph[Theoph$Subject == 1,]
#' pk.calc.tmax(conc = conc_data$conc, time = conc_data$Time)
#' @export
pk.calc.tmax <- function(conc, time,
                         options=list(),
                         first.tmax=NULL,
                         check=TRUE) {
  first.tmax <- PKNCA.choose.option(name="first.tmax", value=first.tmax, options=options)
  if (check) {
    assert_conc_time(conc = conc, time = time)
  }
  if (length(conc) == 0 | all(conc %in% c(NA, 0))) {
    NA
  } else {
    ret <- time[conc %in% pk.calc.cmax(conc, check=FALSE)]
    if (first.tmax) {
      ret[1]
    } else {
      ret[length(ret)]
    }
  }
}
# Add the column to the interval specification
add.interval.col("tmax",
                 FUN="pk.calc.tmax",
                 values=c(FALSE, TRUE),
                 unit_type="time",
                 pretty_name="Tmax",
                 desc="Time of the maximum observed concentration",
                 depends=NULL)
PKNCA.set.summary(
  name="tmax",
  description="median and range",
  point=business.median,
  spread=business.range
)

#' Determine time of last observed concentration above the limit of
#' quantification.
#'
#' `NA` will be returned if all `conc` are `NA` or 0.
#'
#' @inheritParams assert_conc_time
#' @param check Run [assert_conc_time()]?
#' @returns The time of the last observed concentration measurement
#' @export
pk.calc.tlast <- function(conc, time, check=TRUE) {
  if (check) {
    assert_conc_time(conc = conc, time = time)
  }
  if (all(conc %in% c(NA, 0))) {
    NA
  } else {
    max(time[!(conc %in% c(NA, 0))])
  }
}
# Add the column to the interval specification
add.interval.col("tlast",
                 FUN="pk.calc.tlast",
                 values=c(FALSE, TRUE),
                 unit_type="time",
                 pretty_name="Tlast",
                 desc="Time of the last concentration observed above the limit of quantification",
                 depends=NULL)
PKNCA.set.summary(
  name="tlast",
  description="median and range",
  point=business.median,
  spread=business.range
)

#' @describeIn pk.calc.tlast Determine the first concentration above
#'   the limit of quantification.
#' @export
pk.calc.tfirst <- function(conc, time, check=TRUE) {
  if (check) {
    assert_conc_time(conc, time)
  }
  if (all(conc %in% c(NA, 0))) {
    NA
  } else {
    min(time[!(conc %in% c(NA, 0))])
  }
}
# Add the column to the interval specification
add.interval.col("tfirst",
                 FUN="pk.calc.tfirst",
                 values=c(FALSE, TRUE),
                 unit_type="time",
                 pretty_name="Tfirst",
                 desc="Time of the first concentration above the limit of quantification",
                 depends=NULL)
PKNCA.set.summary(
  name="tfirst",
  description="median and range",
  point=business.median,
  spread=business.range
)

#' Determine the last observed concentration above the limit of quantification
#' (LOQ).
#'
#' If all concentrations are missing, `NA_real_` is returned.  If all
#' concentrations are zero (below the limit of quantification) or missing, zero
#' is returned.  If Tlast is NA (due to no non-missing above LOQ measurements),
#' this will return `NA_real_`.
#'
#' @inheritParams assert_conc_time
#' @param check Run [assert_conc_time()]?
#' @returns The last observed concentration above the LOQ
#' @family NCA parameters for concentrations during the intervals
#' @export
pk.calc.clast.obs <- function(conc, time, check=TRUE) {
  if (check) {
    assert_conc_time(conc = conc, time = time)
  }
  if (all(is.na(conc))) {
    NA_real_
  } else if (all(conc %in% c(0, NA))) {
    0
  } else {
    tlast <- pk.calc.tlast(conc, time, check = FALSE)
    if (!is.na(tlast)) {
      conc[time %in% tlast]
    } else {
      NA_real_
    }
  }
}
# Add the column to the interval specification
add.interval.col("clast.obs",
                 FUN="pk.calc.clast.obs",
                 values=c(FALSE, TRUE),
                 unit_type="conc",
                 pretty_name="Clast",
                 desc="The last concentration observed above the limit of quantification",
                 depends=NULL)
PKNCA.set.summary(
  name="clast.obs",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)

#' Calculate the effective half-life
#'
#' @details thalf.eff is `log(2)*mrt`.
#'
#' @param mrt the mean residence time to infinity
#' @return the numeric value of the effective half-life
#' @export
pk.calc.thalf.eff <- function(mrt) {
  log(2)*mrt
}
# Add the columns to the interval specification
add.interval.col("thalf.eff.obs",
                 FUN="pk.calc.thalf.eff",
                 values=c(FALSE, TRUE),
                 desc="The effective half-life (as determined from the MRTobs)",
                 unit_type="time",
                 pretty_name="Effective half-life (based on MRT,obs)",
                 formalsmap=list(mrt="mrt.obs"),
                 depends="mrt.obs")
PKNCA.set.summary(
  name="thalf.eff.obs",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)
add.interval.col("thalf.eff.pred",
                 FUN="pk.calc.thalf.eff",
                 values=c(FALSE, TRUE),
                 unit_type="time",
                 pretty_name="Effective half-life (based on MRT,pred)",
                 desc="The effective half-life (as determined from the MRTpred)",
                 formalsmap=list(mrt="mrt.pred"),
                 depends="mrt.pred")
PKNCA.set.summary(
  name="thalf.eff.pred",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)
add.interval.col("thalf.eff.last",
                 FUN="pk.calc.thalf.eff",
                 values=c(FALSE, TRUE),
                 unit_type="time",
                 pretty_name="Effective half-life (based on MRT,last)",
                 desc="The effective half-life (as determined from the MRTlast)",
                 formalsmap=list(mrt="mrt.last"),
                 depends="mrt.last")
PKNCA.set.summary(
  name="thalf.eff.last",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)
add.interval.col("thalf.eff.iv.obs",
                 FUN="pk.calc.thalf.eff",
                 values=c(FALSE, TRUE),
                 unit_type="time",
                 pretty_name="Effective half-life (for IV dosing, based on MRT,obs)",
                 desc="The effective half-life (as determined from the intravenous MRTobs)",
                 formalsmap=list(mrt="mrt.iv.obs"),
                 depends="mrt.iv.obs")
PKNCA.set.summary(
  name="thalf.eff.iv.obs",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)
add.interval.col("thalf.eff.iv.pred",
                 FUN="pk.calc.thalf.eff",
                 values=c(FALSE, TRUE),
                 unit_type="time",
                 pretty_name="Effective half-life (for IV dosing, based on MRT,pred)",
                 desc="The effective half-life (as determined from the intravenous MRTpred)",
                 formalsmap=list(mrt="mrt.iv.pred"),
                 depends="mrt.iv.pred")
PKNCA.set.summary(
  name="thalf.eff.iv.pred",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)
add.interval.col("thalf.eff.iv.last",
                 FUN="pk.calc.thalf.eff",
                 values=c(FALSE, TRUE),
                 unit_type="time",
                 pretty_name="Effective half-life (for IV dosing, based on MRTlast)",
                 desc="The effective half-life (as determined from the intravenous MRTlast)",
                 formalsmap=list(mrt="mrt.iv.last"),
                 depends="mrt.iv.last")
PKNCA.set.summary(
  name="thalf.eff.iv.last",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)

#' Calculate the AUC percent extrapolated
#'
#' @details aucpext is `100*(1-auclast/aucinf)`.
#'
#' @param auclast the area under the curve from time 0 to the last measurement
#'   above the limit of quantification
#' @param aucinf the area under the curve from time 0 to infinity
#' @returns The numeric value of the AUC percent extrapolated or `NA_real_` if
#'   any of the following are true `is.na(aucinf)`, `is.na(auclast)`,
#'   `aucinf <= 0`, or `auclast <= 0`.
#' @export
pk.calc.aucpext <- function(auclast, aucinf) {
  scalar_auclast <- length(auclast) == 1
  scalar_aucinf <- length(aucinf) == 1
  if (scalar_auclast | scalar_aucinf) {
    # no length checking needs to occur
  } else if ((!scalar_auclast & !scalar_aucinf) &
             length(auclast) != length(aucinf)) {
    stop("auclast and aucinf must either be a scalar or the same length.")
  }
  ret <- rep(NA_real_, max(c(length(auclast), length(aucinf))))
  mask_na <-
    is.na(auclast) |
    is.na(aucinf)
  mask_negative <-
    !mask_na &
    (aucinf <= 0 |
       auclast <= 0)
  mask_greater <-
    !mask_na &
    (auclast >= aucinf)
  mask_calc <- !mask_na & !(aucinf %in% 0)
  if (any(mask_greater))
    rlang::warn(
      message = "aucpext is typically only calculated when aucinf is greater than auclast.",
      class = "pknca_aucpext_aucinf_le_auclast"
    )
  if (any(mask_negative))
    rlang::warn(
      message = "aucpext is typically only calculated when both aucinf and auclast are positive.",
      class = "pknca_aucpext_aucinf_auclast_positive"
    )
  ret[mask_calc] <-
    100*(1-auclast[mask_calc]/aucinf[mask_calc])
  ret
}

# Add the columns to the interval specification
add.interval.col("aucpext.obs",
                 FUN="pk.calc.aucpext",
                 values=c(FALSE, TRUE),
                 unit_type="%",
                 pretty_name="AUCpext (based on AUCinf,obs)",
                 desc="Percent of the AUCinf that is extrapolated after Tlast calculated from the observed Clast",
                 formalsmap=list(aucinf="aucinf.obs"),
                 depends=c("auclast", "aucinf.obs"))
PKNCA.set.summary(
  name="aucpext.obs",
  description="arithmetic mean and standard deviation",
  point=business.mean,
  spread=business.sd
)
add.interval.col("aucpext.pred",
                 FUN="pk.calc.aucpext",
                 values=c(FALSE, TRUE),
                 unit_type="%",
                 pretty_name="AUCpext (based on AUCinf,pred)",
                 desc="Percent of the AUCinf that is extrapolated after Tlast calculated from the predicted Clast",
                 formalsmap=list(aucinf="aucinf.pred"),
                 depends=c("auclast", "aucinf.pred"))
PKNCA.set.summary(
  name="aucpext.pred",
  description="arithmetic mean and standard deviation",
  point=business.mean,
  spread=business.sd
)

#' Calculate the elimination rate (Kel)
#'
#' @param kel is `1/mrt`, not to be confused with lambda.z.
#'
#' @param mrt the mean residence time
#' @returns the numeric value of the elimination rate
#' @export
pk.calc.kel <- function(mrt) {
  1/mrt
}
# Add the columns to the interval specification
add.interval.col("kel.obs",
                 FUN="pk.calc.kel",
                 values=c(FALSE, TRUE),
                 unit_type="inverse_time",
                 pretty_name="Kel (based on AUCinf,obs)",
                 desc="Elimination rate (as calculated from the MRT with observed Clast)",
                 formalsmap=list(mrt="mrt.obs"),
                 depends="mrt.obs")
PKNCA.set.summary(
  name="kel.obs",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)
add.interval.col("kel.pred",
                 FUN="pk.calc.kel",
                 values=c(FALSE, TRUE),
                 unit_type="inverse_time",
                 pretty_name="Kel (based on AUCinf,pred)",
                 desc="Elimination rate (as calculated from the MRT with predicted Clast)",
                 formalsmap=list(mrt="mrt.pred"),
                 depends="mrt.pred")
PKNCA.set.summary(
  name="kel.pred",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)
add.interval.col("kel.last",
                 FUN="pk.calc.kel",
                 values=c(FALSE, TRUE),
                 unit_type="inverse_time",
                 pretty_name="Kel (based on AUClast)",
                 desc="Elimination rate (as calculated from the MRT using AUClast)",
                 formalsmap=list(mrt="mrt.last"),
                 depends="mrt.last")
PKNCA.set.summary(
  name="kel.last",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)
add.interval.col("kel.iv.obs",
                 FUN="pk.calc.kel",
                 values=c(FALSE, TRUE),
                 unit_type="inverse_time",
                 pretty_name="Kel (for IV dosing, based on AUCinf,obs)",
                 desc="Elimination rate (as calculated from the intravenous MRTobs)",
                 formalsmap=list(mrt="mrt.iv.obs"),
                 depends="mrt.iv.obs")
PKNCA.set.summary(
  name="kel.iv.obs",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)
add.interval.col("kel.iv.pred",
                 FUN="pk.calc.kel",
                 values=c(FALSE, TRUE),
                 unit_type="inverse_time",
                 pretty_name="Kel (for IV dosing, based on AUCinf,pred)",
                 desc="Elimination rate (as calculated from the intravenous MRTpred)",
                 formalsmap=list(mrt="mrt.iv.pred"),
                 depends="mrt.iv.pred")
PKNCA.set.summary(
  name="kel.iv.pred",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)
add.interval.col("kel.iv.last",
                 FUN="pk.calc.kel",
                 values=c(FALSE, TRUE),
                 unit_type="inverse_time",
                 pretty_name="Kel (for IV dosing, based on AUClast)",
                 desc="Elimination rate (as calculated from the intravenous MRTlast)",
                 formalsmap=list(mrt="mrt.iv.last"),
                 depends="mrt.iv.last")
PKNCA.set.summary(
  name="kel.iv.last",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)

#' Calculate the (observed oral) clearance
#'
#' @details cl is `dose/auc`.
#'
#' @param dose the dose administered
#' @param auc The area under the concentration-time curve.
#' @returns the numeric value of the total (CL) or observed oral clearance
#'   (CL/F)
#' @details If `dose` is the same length as the other inputs, then the output
#'   will be the same length as all of the inputs; the function assumes that you
#'   are calculating for multiple intervals simultaneously.  If the inputs other
#'   than `dose` are scalars and `dose` is a vector, then the function assumes
#'   multiple doses were given in a single interval, and the sum of the `dose`s
#'   will be used for the calculation.
#' @references Gabrielsson J, Weiner D. "Section 2.5.1 Derivation of clearance."
#'   Pharmacokinetic & Pharmacodynamic Data Analysis: Concepts and Applications,
#'   4th Edition.  Stockholm, Sweden: Swedish Pharmaceutical Press, 2000. 86-7.
#' @export
pk.calc.cl <- function(dose, auc) {
  if (length(auc) == 1) {
    dose <- sum(dose)
  }
  ret <- dose/auc
  mask_zero <- !is.na(auc) & (auc <= 0)
  if (any(mask_zero)) {
    ret[mask_zero] <- NA_real_
  }
  ret
}

# Add the columns to the interval specification
add.interval.col("cl.last",
                 FUN="pk.calc.cl",
                 values=c(FALSE, TRUE),
                 unit_type="clearance",
                 pretty_name="CL (based on AUClast)",
                 desc="Clearance or observed oral clearance calculated to Clast",
                 formalsmap=list(auc="auclast"),
                 depends="auclast")
PKNCA.set.summary(
  name="cl.last",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)
add.interval.col("cl.all",
                 FUN="pk.calc.cl",
                 values=c(FALSE, TRUE),
                 unit_type="clearance",
                 pretty_name="CL (based on AUCall)",
                 desc="Clearance or observed oral clearance calculated with AUCall",
                 formalsmap=list(auc="aucall"),
                 depends="aucall")
PKNCA.set.summary(
  name="cl.all",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)
add.interval.col("cl.obs",
                 FUN="pk.calc.cl",
                 values=c(FALSE, TRUE),
                 unit_type="clearance",
                 pretty_name="CL (based on AUCinf,obs)",
                 desc="Clearance or observed oral clearance calculated with observed Clast",
                 formalsmap=list(auc="aucinf.obs"),
                 depends="aucinf.obs")
PKNCA.set.summary(
  name="cl.obs",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)
add.interval.col("cl.pred",
                 FUN="pk.calc.cl",
                 values=c(FALSE, TRUE),
                 unit_type="clearance",
                 pretty_name="CL (based on AUCinf,pred)",
                 desc="Clearance or observed oral clearance calculated with predicted Clast",
                 formalsmap=list(auc="aucinf.pred"),
                 depends="aucinf.pred")
PKNCA.set.summary(
  name="cl.pred",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)

#' Calculate the absolute (or relative) bioavailability
#'
#' @details f is `(auc2/dose2)/(auc1/dose1)`.
#'
#' @param dose1 The dose administered in route or method 1
#' @param dose2 The dose administered in route or method 2
#' @param auc1 The AUC from 0 to infinity or 0 to tau administered in route or
#'   method 1
#' @param auc2 The AUC from 0 to infinity or 0 to tau administered in route or
#'   method 2
#' @export
pk.calc.f <- function(dose1, auc1, dose2, auc2) {
  ret <- (auc2/dose2)/(auc1/dose1)
  mask_zero <-
    is.na(auc1)  | (auc1 <= 0) |
    is.na(dose2) | (dose2 <= 0) |
    is.na(dose1) | (dose1 <= 0)
  if (any(mask_zero)) {
    ret[mask_zero] <- NA_real_
  }
  ret
}
add.interval.col("f",
                 FUN="pk.calc.f",
                 values=c(FALSE, TRUE),
                 unit_type="fraction",
                 pretty_name="Bioavailability",
                 desc="Bioavailability or relative bioavailability",
                 depends=NULL)
PKNCA.set.summary(
  name="f",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)

#' Calculate the mean residence time (MRT) for single-dose data or linear
#' multiple-dose data.
#'
#' @details mrt is `aumc/auc - duration.dose/2` where `duration.dose =
#'   0` for oral administration.
#'
#' @param auc the AUC from 0 to infinity or 0 to tau
#' @param aumc the AUMC from 0 to infinity or 0 to tau
#' @param duration.dose The duration of the dose (usually an infusion duration
#'   for an IV infusion)
#' @returns the numeric value of the mean residence time
#' @seealso [pk.calc.mrt.md()]
#' @export
pk.calc.mrt <- function(auc, aumc) {
  pk.calc.mrt.iv(auc, aumc, duration.dose=0)
}
# Add the columns to the interval specification
add.interval.col("mrt.obs",
                 FUN="pk.calc.mrt",
                 values=c(FALSE, TRUE),
                 unit_type="time",
                 pretty_name="MRT (based on AUCinf,obs)",
                 desc="The mean residence time to infinity using observed Clast",
                 formalsmap=list(auc="aucinf.obs", aumc="aumcinf.obs"),
                 depends=c("aucinf.obs", "aumcinf.obs"))
PKNCA.set.summary(
  name="mrt.obs",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)
add.interval.col("mrt.pred",
                 FUN="pk.calc.mrt",
                 values=c(FALSE, TRUE),
                 unit_type="time",
                 pretty_name="MRT (based on AUCinf,pred)",
                 desc="The mean residence time to infinity using predicted Clast",
                 formalsmap=list(auc="aucinf.pred", aumc="aumcinf.pred"),
                 depends=c("aucinf.pred", "aumcinf.pred"))
PKNCA.set.summary(
  name="mrt.pred",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)
add.interval.col("mrt.last",
                 FUN="pk.calc.mrt",
                 values=c(FALSE, TRUE),
                 unit_type="time",
                 pretty_name="MRT (based on AUClast)",
                 desc="The mean residence time to the last observed concentration above the LOQ",
                 formalsmap=list(auc="auclast", aumc="aumclast"),
                 depends=c("auclast", "aumclast"))
PKNCA.set.summary(
  name="mrt.last",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)

#' @describeIn pk.calc.mrt MRT for an IV infusion
#' @export
pk.calc.mrt.iv <- function(auc, aumc, duration.dose) {
  ret <- aumc/auc - duration.dose/2
  mask_zero <- is.na(auc) | auc <= 0
  if (any(mask_zero)) {
    ret[mask_zero] <- NA_real_
  }
  ret
}
# Add the columns to the interval specification
add.interval.col("mrt.iv.obs",
                 FUN="pk.calc.mrt.iv",
                 values=c(FALSE, TRUE),
                 unit_type="time",
                 pretty_name="MRT (for IV dosing, based on AUCinf,obs)",
                 desc="The mean residence time to infinity using observed Clast correcting for dosing duration",
                 formalsmap=list(auc="aucinf.obs", aumc="aumcinf.obs"),
                 depends=c("aucinf.obs", "aumcinf.obs"))
PKNCA.set.summary(
  name="mrt.iv.obs",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)
add.interval.col("mrt.iv.pred",
                 FUN="pk.calc.mrt.iv",
                 values=c(FALSE, TRUE),
                 unit_type="time",
                 pretty_name="MRT (for IV dosing, based on AUCinf,pred)",
                 desc="The mean residence time to infinity using predicted Clast correcting for dosing duration",
                 formalsmap=list(auc="aucinf.pred", aumc="aumcinf.pred"),
                 depends=c("aucinf.pred", "aumcinf.pred"))
PKNCA.set.summary(
  name="mrt.iv.pred",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)
add.interval.col("mrt.iv.last",
                 FUN="pk.calc.mrt.iv",
                 values=c(FALSE, TRUE),
                 unit_type="time",
                 pretty_name="MRT (for IV dosing, based on AUClast)",
                 desc="The mean residence time to the last observed concentration above the LOQ correcting for dosing duration",
                 formalsmap=list(auc="auclast", aumc="aumclast"),
                 depends=c("auclast", "aumclast"))
PKNCA.set.summary(
  name="mrt.iv.last",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)

#' Calculate the mean residence time (MRT) for multiple-dose data with nonlinear
#' kinetics.
#'
#' @details mrt.md is `aumctau/auctau + tau*(aucinf-auctau)/auctau` and should
#'   only be used for multiple dosing with equal intervals between doses.
#'
#' @param auctau the AUC from time 0 to the end of the dosing interval (tau).
#' @param aumctau the AUMC from time 0 to the end of the dosing interval (tau).
#' @param aucinf the AUC from time 0 to infinity (typically using single-dose
#'   data)
#' @inheritParams assert_dosetau
#' @details Note that if `aucinf == auctau` (as would be the assumption with
#'   linear kinetics), the equation becomes the same as the single-dose MRT.
#' @seealso [pk.calc.mrt()]
#' @export
pk.calc.mrt.md <- function(auctau, aumctau, aucinf, tau) {
  ret <- aumctau/auctau + tau*(aucinf-auctau)/auctau
  mask_zero <- is.na(auctau) | auctau <= 0
  if (any(mask_zero)) {
    ret[mask_zero] <- NA_real_
  }
  ret
}
add.interval.col("mrt.md.obs",
                 FUN="pk.calc.mrt.md",
                 values=c(FALSE, TRUE),
                 unit_type="time",
                 pretty_name="MRT (for multiple dosing, based on AUCinf,obs)",
                 desc="The mean residence time with multiple dosing and nonlinear kinetics using observed Clast",
                 formalsmap=list(auctau="auclast", aumctau="aumclast", aucinf="aucinf.obs"),
                 depends=c("auclast", "aumclast", "aucinf.obs"))
PKNCA.set.summary(
  name="mrt.md.obs",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)
add.interval.col("mrt.md.pred",
                 FUN="pk.calc.mrt.md",
                 values=c(FALSE, TRUE),
                 unit_type="time",
                 pretty_name="MRT (for multiple dosing, based on AUCinf,pred)",
                 desc="The mean residence time with multiple dosing and nonlinear kinetics using predicted Clast",
                 formalsmap=list(auctau="auclast", aumctau="aumclast", aucinf="aucinf.pred"),
                 depends=c("auclast", "aumclast", "aucinf.pred"))
PKNCA.set.summary(
  name="mrt.md.pred",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)

#' Calculate the terminal volume of distribution (Vz)
#'
#' @details vz is `cl/lambda.z`.
#'
#' @inheritParams assert_lambdaz
#' @param cl the clearance (or apparent observed clearance)
#' @export
pk.calc.vz <- function(cl, lambda.z) {
  assert_lambdaz(lambda.z)
  # Ensure that cl is either a scalar or the same length as AUC
  # (more complex repeating patterns while valid for general R are
  # likely errors here).
  if (!(length(cl) %in% c(1, length(lambda.z))) |
      !(length(lambda.z) %in% c(1, length(cl))))
    stop("'cl' and 'lambda.z' must be the same length")
  cl/lambda.z
}
# Add the columns to the interval specification
add.interval.col("vz.obs",
                 FUN="pk.calc.vz",
                 values=c(FALSE, TRUE),
                 unit_type="volume",
                 pretty_name="Vz (based on AUCinf,obs)",
                 desc="The terminal volume of distribution using observed Clast",
                 formalsmap=list(cl="cl.obs"),
                 depends=c("cl.obs", "lambda.z"))
PKNCA.set.summary(
  name="vz.obs",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)
add.interval.col("vz.pred",
                 FUN="pk.calc.vz",
                 values=c(FALSE, TRUE),
                 unit_type="volume",
                 pretty_name="Vz (based on AUCinf,pred)",
                 desc="The terminal volume of distribution using predicted Clast",
                 formalsmap=list(cl="cl.pred"),
                 depends=c("cl.pred", "lambda.z"))
PKNCA.set.summary(
  name="vz.pred",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)

#' Calculate the steady-state volume of distribution (Vss)
#'
#' @details vss is `cl*mrt`.
#' @param cl the clearance
#' @param mrt the mean residence time
#' @return the volume of distribution at steady-state
#' @export
pk.calc.vss <- function(cl, mrt) {
  cl*mrt
}
# Add the columns to the interval specification
add.interval.col("vss.obs",
                 FUN="pk.calc.vss",
                 values=c(FALSE, TRUE),
                 unit_type="volume",
                 pretty_name="Vss (based on AUCinf,obs)",
                 desc="The steady-state volume of distribution using observed Clast",
                 formalsmap=list(cl="cl.obs", mrt="mrt.obs"),
                 depends=c("cl.obs", "mrt.obs"))
PKNCA.set.summary(
  name="vss.obs",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)
add.interval.col("vss.pred",
                 FUN="pk.calc.vss",
                 values=c(FALSE, TRUE),
                 unit_type="volume",
                 pretty_name="Vss (based on AUCinf,pred)",
                 desc="The steady-state volume of distribution using predicted Clast",
                 formalsmap=list(cl="cl.pred", mrt="mrt.pred"),
                 depends=c("cl.pred", "mrt.pred"))
PKNCA.set.summary(
  name="vss.pred",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)
add.interval.col("vss.last",
                 FUN="pk.calc.vss",
                 values=c(FALSE, TRUE),
                 unit_type="volume",
                 pretty_name="Vss (based on AUClast)",
                 desc="The steady-state volume of distribution calculating through Tlast",
                 formalsmap=list(cl="cl.last", mrt="mrt.last"),
                 depends=c("cl.last", "mrt.last"))
PKNCA.set.summary(
  name="vss.last",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)
add.interval.col("vss.iv.obs",
                 FUN="pk.calc.vss",
                 values=c(FALSE, TRUE),
                 unit_type="volume",
                 pretty_name="Vss (for IV dosing, based on AUCinf,obs)",
                 desc="The steady-state volume of distribution with intravenous infusion using observed Clast",
                 formalsmap=list(cl="cl.obs", mrt="mrt.iv.obs"),
                 depends=c("cl.obs", "mrt.iv.obs"))
PKNCA.set.summary(
  name="vss.iv.obs",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)
add.interval.col("vss.iv.pred",
                 FUN="pk.calc.vss",
                 values=c(FALSE, TRUE),
                 unit_type="volume",
                 pretty_name="Vss (for IV dosing, based on AUCinf,pred)",
                 desc="The steady-state volume of distribution with intravenous infusion using predicted Clast",
                 formalsmap=list(cl="cl.pred", mrt="mrt.iv.pred"),
                 depends=c("cl.pred", "mrt.iv.pred"))
PKNCA.set.summary(
  name="vss.iv.pred",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)
add.interval.col("vss.iv.last",
                 FUN="pk.calc.vss",
                 values=c(FALSE, TRUE),
                 unit_type="volume",
                 pretty_name="Vss (for IV dosing, based on AUClast)",
                 desc="The steady-state volume of distribution with intravenous infusion calculating through Tlast",
                 formalsmap=list(cl="cl.last", mrt="mrt.iv.last"),
                 depends=c("cl.last", "mrt.iv.last"))
PKNCA.set.summary(
  name="vss.iv.last",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)

add.interval.col("vss.md.obs",
                 FUN="pk.calc.vss",
                 values=c(FALSE, TRUE),
                 unit_type="volume",
                 pretty_name="Vss (for multiple-dose, based on Clast,obs)",
                 desc="The steady-state volume of distribution for nonlinear multiple-dose data using observed Clast",
                 formalsmap=list(cl="cl.last", mrt="mrt.md.obs"),
                 depends=c("cl.last", "mrt.md.obs"))
PKNCA.set.summary(
  name="vss.md.obs",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)
add.interval.col("vss.md.pred",
                 FUN="pk.calc.vss",
                 values=c(FALSE, TRUE),
                 unit_type="volume",
                 pretty_name="Vss (for multiple-dose, based on Clast,pred)",
                 desc="The steady-state volume of distribution for nonlinear multiple-dose data using predicted Clast",
                 formalsmap=list(cl="cl.last", mrt="mrt.md.pred"),
                 depends=c("cl.last", "mrt.md.pred"))
PKNCA.set.summary(
  name="vss.md.pred",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)

#' Calculate the average concentration during an interval.
#'
#' @details cav is `auc/(end-start)`.
#'
#' @param auc The area under the curve during the interval
#' @inheritParams assert_intervaltime_single
#' @returns The Cav (average concentration during the interval)
#' @export
pk.calc.cav <- function(auc, start, end) {
  ret <- auc/(end-start)
  mask_zero <- is.na(end) | is.na(start) | end == start
  if (any(mask_zero)) {
    ret[mask_zero] <- NA_real_
  }
  ret
}
add.interval.col(
  "cav",
  FUN = "pk.calc.cav",
  values = c(FALSE, TRUE),
  unit_type = "conc",
  pretty_name = "Cav",
  desc = "The average concentration during an interval (calculated with AUClast)",
  depends = "auclast",
  formalsmap = list(auc = "auclast")
)
add.interval.col(
  "cav.int.last",
  FUN = "pk.calc.cav",
  values = c(FALSE, TRUE),
  unit_type = "conc",
  pretty_name = "Cav",
  desc = "The average concentration during an interval (calculated with AUCint.last)",
  depends = "aucint.last",
  formalsmap = list(auc = "aucint.last"),
)
add.interval.col(
  "cav.int.all",
  FUN = "pk.calc.cav",
  values = c(FALSE, TRUE),
  unit_type = "conc",
  pretty_name = "Cav",
  desc = "The average concentration during an interval (calculated with AUCint.all)",
  depends = "aucint.all",
  formalsmap = list(auc = "aucint.all"),
)
add.interval.col(
  "cav.int.inf.obs",
  FUN = "pk.calc.cav",
  values = c(FALSE, TRUE),
  unit_type = "conc",
  pretty_name = "Cav",
  desc = "The average concentration during an interval (calculated with AUCint.inf.obs)",
  depends = "aucint.inf.obs",
  formalsmap = list(auc = "aucint.inf.obs"),
)
add.interval.col(
  "cav.int.inf.pred",
  FUN = "pk.calc.cav",
  values = c(FALSE, TRUE),
  unit_type = "conc",
  pretty_name = "Cav",
  desc = "The average concentration during an interval (calculated with AUCint.inf.pred)",
  depends = "aucint.inf.pred",
  formalsmap = list(auc = "aucint.inf.pred"),
)

PKNCA.set.summary(
  name = c("cav", "cav.int.last", "cav.int.all", "cav.int.inf.obs", "cav.int.inf.pred"),
  description = "geometric mean and geometric coefficient of variation",
  point = business.geomean,
  spread = business.geocv
)

#' Determine the trough (end of interval) concentration
#'
#' @inheritParams assert_conc_time
#' @inheritParams assert_intervaltime_single
#' @returns The concentration when `time == end`.  If none match, then `NA`
#' @family NCA parameters for concentrations during the intervals
#' @export
pk.calc.ctrough <- function(conc, time, end) {
  assert_conc_time(conc = conc, time = time)
  mask_end <- time %in% end
  if (sum(mask_end) == 1) {
    conc[mask_end]
  } else if (sum(mask_end) == 0) {
    NA_real_
  } else {
    # This should be impossible as assert_conc_time should catch
    # duplicates.
    stop("More than one time matches the starting time.  Please report this as a bug with a reproducible example.") # nocov
  }
}
add.interval.col("ctrough",
                 FUN="pk.calc.ctrough",
                 values=c(FALSE, TRUE),
                 unit_type="conc",
                 pretty_name="Ctrough",
                 desc="The trough (end of interval) concentration",
                 depends=NULL)
PKNCA.set.summary(
  name="ctrough",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)

#' Determine the concentration at the beginning of the interval
#'
#' @inheritParams assert_conc_time
#' @inheritParams assert_intervaltime_single
#' @return The concentration when `time == end`.  If none match, then `NA`
#' @family NCA parameters for concentrations during the intervals
#' @export
pk.calc.cstart <- function(conc, time, start) {
  assert_conc_time(conc = conc, time = time)
  mask_start <- time %in% start
  if (sum(mask_start) == 1) {
    conc[mask_start]
  } else if (sum(mask_start) == 0) {
    NA_real_
  } else {
    # This should be impossible as assert_conc_time should catch
    # duplicates.
    stop("More than one time matches the starting time.  Please report this as a bug with a reproducible example.") # nocov
  }
}
add.interval.col("cstart",
                 FUN="pk.calc.cstart",
                 values=c(FALSE, TRUE),
                 unit_type="conc",
                 pretty_name="Cstart",
                 desc="The predose concentration",
                 depends=NULL)
PKNCA.set.summary(
  name="cstart",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)

#' Determine the peak-to-trough ratio
#'
#' @details ptr is `cmax/ctrough`.
#'
#' @param cmax The maximum observed concentration
#' @param ctrough The last concentration in an interval
#' @return The ratio of cmax to ctrough (if ctrough == 0, NA)
#' @export
pk.calc.ptr <- function(cmax, ctrough) {
  ret <- cmax/ctrough
  ret[ctrough %in% 0] <- NA_real_
  ret
}
add.interval.col("ptr",
                 FUN="pk.calc.ptr",
                 values=c(FALSE, TRUE),
                 unit_type="fraction",
                 pretty_name="Peak-to-trough ratio",
                 desc="Peak-to-Trough ratio (fraction)",
                 depends=c("cmax", "ctrough"))
PKNCA.set.summary(
  name="ptr",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)

#' Determine the observed lag time (time before the first
#' concentration above the limit of quantification or above the first
#' concentration in the interval)
#'
#' @inheritParams assert_conc_time
#' @returns The time associated with the first increasing concentration
#' @export
pk.calc.tlag <- function(conc, time) {
  assert_conc_time(conc = conc, time = time)
  mask.increase <- c(conc[-1] > conc[-length(conc)], FALSE)
  if (any(mask.increase)) {
    time[mask.increase][1]
  } else {
    NA_real_
  }
}
add.interval.col("tlag",
                 FUN="pk.calc.tlag",
                 values=c(FALSE, TRUE),
                 unit_type="time",
                 pretty_name="Tlag",
                 desc="Lag time",
                 depends=NULL)
PKNCA.set.summary(
  name="tlag",
  description="median and range",
  point=business.median,
  spread=business.range
)

#' Determine the degree of fluctuation
#'
#' @details deg.fluc is `100*(cmax - cmin)/cav`.
#'
#' @param cmax The maximum observed concentration
#' @param cmin The minimum observed concentration
#' @param cav The average concentration in the interval
#' @returns The degree of fluctuation around the average concentration.
#' @export
pk.calc.deg.fluc <- function(cmax, cmin, cav) {
  ret <- 100*(cmax - cmin)/cav
  mask_zero <- is.na(cav) | cav == 0
  if (any(mask_zero)) {
    ret[mask_zero] <- NA_real_
  }
  ret
}
add.interval.col("deg.fluc",
                 FUN="pk.calc.deg.fluc",
                 unit_type="%",
                 pretty_name="Degree of fluctuation",
                 desc="Degree of fluctuation",
                 depends=c("cmax", "cmin", "cav"))
PKNCA.set.summary(
  name="deg.fluc",
  description="arithmetic mean and standard deviation",
  point=business.mean,
  spread=business.sd
)

#' Determine the PK swing
#'
#' @details swing is `100*(cmax - cmin)/cmin`.
#'
#' @param cmax The maximum observed concentration
#' @param cmin The minimum observed concentration
#' @returns The swing above the minimum concentration.  If `cmin` is zero, then
#'   the result is infinity.
#' @export
pk.calc.swing <- function(cmax, cmin) {
  if (cmin > 0) {
    100*(cmax - cmin)/cmin
  } else {
    Inf
  }
}
add.interval.col("swing",
                 FUN="pk.calc.swing",
                 unit_type="%",
                 pretty_name="Swing",
                 desc="Swing relative to Cmin",
                 depends=c("cmax", "cmin"))
PKNCA.set.summary(
  name="swing",
  description="arithmetic mean and standard deviation",
  point=business.mean,
  spread=business.sd
)

#' Determine the concentration at the end of infusion
#'
#' @inheritParams assert_conc_time
#' @param duration.dose The duration for the dosing administration (typically
#'   from IV infusion)
#' @param check Run [assert_conc_time()]?
#' @returns The concentration at the end of the infusion, `NA` if
#'   `duration.dose` is `NA`, or `NA` if all `time != duration.dose`
#' @export
pk.calc.ceoi <- function(conc, time, duration.dose=NA, check=TRUE) {
  if (check) {
    assert_conc_time(conc = conc, time = time)
  }
  if (is.na(duration.dose)) {
    NA_real_
  } else if (all(time != duration.dose)) {
    NA_real_
  } else {
    conc[time == duration.dose][1]
  }
}
add.interval.col("ceoi",
                 FUN="pk.calc.ceoi",
                 unit_type="conc",
                 pretty_name="Ceoi",
                 desc="Concentration at the end of infusion",
                 depends=NULL)
PKNCA.set.summary(
  name="ceoi",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)

#' Calculate the AUC above a given concentration
#'
#' Concentrations below the given concentration (`conc_above`) will be set
#' to zero.
#' @inheritParams pk.calc.time_above
#' @returns The AUC of the concentration above the limit
#' @export
pk.calc.aucabove <- function(conc, time, conc_above = NA_real_, ..., options=list()) {
  stopifnot(length(conc_above) == 1)
  stopifnot(is.numeric(conc_above))
  if (is.na(conc_above)) {
    ret <- structure(NA_real_, exclude = "Missing concentration to be above")
  } else {
    ret <-
      pk.calc.auc(
        conc=pmax(conc - conc_above, 0), time=time, ..., options=options,
        auc.type="AUCall",
        lambda.z=NA
      )
  }
  ret
}
add.interval.col(
  "aucabove.predose.all",
  FUN="pk.calc.aucabove",
  unit_type="auc",
  pretty_name="AUC,above",
  desc="The area under the concentration time the beginning of the interval to the last concentration above the limit of quantification plus the triangle from that last concentration to 0 at the first concentration below the limit of quantification, with a concentration subtracted from all concentrations and values below zero after subtraction set to zero",
  depends="cstart",
  formalsmap = list(conc_above = "cstart")
)
PKNCA.set.summary(
  name="aucabove.predose.all",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)
add.interval.col(
  "aucabove.trough.all",
  FUN="pk.calc.aucabove",
  unit_type="auc",
  pretty_name="AUC,above",
  desc="The area under the concentration time the beginning of the interval to the last concentration above the limit of quantification plus the triangle from that last concentration to 0 at the first concentration below the limit of quantification, with a concentration subtracted from all concentrations and values below zero after subtraction set to zero",
  depends="ctrough",
  formalsmap = list(conc_above = "ctrough")
)
PKNCA.set.summary(
  name="aucabove.trough.all",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)

#' Count the number of concentration measurements in an interval
#'
#' `count_conc` and `count_conc_measured` are typically used for quality control
#' on the data to ensure that there are a sufficient number of non-missing
#' samples for a calculation and to ensure that data are consistent between
#' individuals.
#'
#' @inheritParams pk.calc.cmax
#' @returns a count of the non-missing concentrations (0 if all concentrations
#'   are missing)
#' @family NCA parameters for concentrations during the intervals
#' @export
pk.calc.count_conc <- function(conc, check=TRUE) {
  if (check) {
    assert_conc(conc)
  }
  sum(!is.na(conc))
}
# Add the column to the interval specification
add.interval.col(
  "count_conc",
  FUN = "pk.calc.count_conc",
  values = c(FALSE, TRUE),
  unit_type = "count",
  pretty_name = "Concentration count",
  desc = "Number of non-missing concentrations for an interval",
  depends = NULL
)

#' @describeIn pk.calc.count_conc Count the number of concentration measurements
#'   that are not missing, above, or below the limit of quantification in an
#'   interval
#'
#' @returns a count of the non-missing, measured (not below or above the limit
#'   of quantification) concentrations (0 if all concentrations are missing)
#' @export
pk.calc.count_conc_measured <- function(conc, check=TRUE) {
  if (check) {
    assert_conc(conc)
  }
  sum(!is.na(conc) & is.finite(conc) & conc > 0)
}
# Add the column to the interval specification
add.interval.col(
  "count_conc_measured",
  FUN="pk.calc.count_conc_measured",
  values=c(FALSE, TRUE),
  unit_type="count",
  pretty_name="Measured concentration count",
  desc="Number of measured and non BLQ/ALQ concentrations for an interval",
  depends=NULL
)
PKNCA.set.summary(
  name=c("count_conc", "count_conc_measured"),
  description="median and range",
  point=business.median,
  spread=business.range
)

#' Extract the dose used for calculations
#'
#' @inheritParams pk.calc.cl
#' @returns The total dose for an interval
#' @export
pk.calc.totdose <- function(dose) {
  sum(dose)
}
add.interval.col(
  "totdose",
  FUN="pk.calc.totdose",
  values=c(FALSE, TRUE),
  unit_type="dose",
  pretty_name="Total dose",
  desc="Total dose administered during an interval"
)
PKNCA.set.summary(
  name="totdose",
  description="median and range",
  point=business.median,
  spread=business.range
)

test_that("adj.r.squared", {
  # Ensure correct calculation
  expect_equal(adj.r.squared(1, 5), 1)
  expect_equal(adj.r.squared(0.5, 5), 1-0.5*4/3)

  # Ensure that N must be an integer > 2
  expect_warning(
    expect_equal(
      adj.r.squared(1, 2),
      structure(NA_real_, exclude="n must be > 2")
    ),
    regexp="n must be > 2"
  )
})

test_that("pk.calc.cmax", {
  # Confirm that all NAs give NA as an output
  expect_warning(v1 <- pk.calc.cmax(NA))
  expect_equal(v1, NA)
  expect_warning(v2 <- pk.calc.cmax(c(NA, NA)))
  expect_equal(v2, NA)

  # Confirm that no NAs give the max value
  expect_equal(pk.calc.cmax(c(1, 2)), 2)
  expect_equal(pk.calc.cmax(c(1, 2, 3)), 3)

  # Confirm that some NAs give the NA-removed maximum value
  expect_equal(pk.calc.cmax(c(1, NA, 3)), 3)
  expect_equal(pk.calc.cmax(c(NA, NA, 3)), 3)
  expect_equal(pk.calc.cmax(c(1, NA, NA)), 1)
  expect_equal(pk.calc.cmax(c(1, NA, 3, NA)), 3)

  # Confirm that no data gives NA with a warning
  expect_warning(v3 <- pk.calc.cmax(numeric()))
  expect_equal(v3, NA)
})

test_that("pk.calc.cmin", {
  # Confirm that all NAs give NA as an output
  expect_warning(v1 <- pk.calc.cmin(NA))
  expect_equal(v1, NA)
  expect_warning(v2 <- pk.calc.cmin(c(NA, NA)))
  expect_equal(v2, NA)

  # Confirm that no NAs give the min value
  expect_equal(pk.calc.cmin(c(1, 2)), 1)
  expect_equal(pk.calc.cmin(c(1, 2, 3)), 1)

  # Confirm that some NAs give the NA-removed minimum value
  expect_equal(pk.calc.cmin(c(1, NA, 3)), 1)
  expect_equal(pk.calc.cmin(c(NA, NA, 3)), 3)
  expect_equal(pk.calc.cmin(c(1, NA, NA)), 1)
  expect_equal(pk.calc.cmin(c(1, NA, 3, NA)), 1)

  # Confirm that no data gives NA with a warning
  expect_warning(v3 <- pk.calc.cmin(numeric()))
  expect_equal(v3, NA)
})

test_that("pk.calc.tmax", {
  # No data give a warning and NA
  expect_warning(expect_warning(
    v1 <- pk.calc.tmax(numeric(), numeric()),
    class = "pknca_conc_none"),
    class = "pknca_time_none"
  )
  expect_equal(v1, NA)

  # Either concentration or time is missing, give an error
  expect_error(
    suppressWarnings(pk.calc.tmax(conc = numeric())),
    regexp='argument "time" is missing, with no default'
  )
  expect_error(
    pk.calc.tmax(time=numeric()),
    regexp='argument "conc" is missing, with no default'
  )

  # It calculates tmax correctly based on the use.first option
  expect_equal(pk.calc.tmax(c(1, 2), c(0, 1), first.tmax=TRUE),
               1)
  expect_equal(pk.calc.tmax(c(1, 2), c(0, 1), first.tmax=FALSE),
               1)
  expect_equal(pk.calc.tmax(c(1, 1), c(0, 1), first.tmax=TRUE),
               0)
  expect_equal(pk.calc.tmax(c(1, 1), c(0, 1), first.tmax=FALSE),
               1)
})

test_that("pk.calc.tlast", {
  # Either concentration or time is missing, give an error
  expect_error(
    suppressWarnings(pk.calc.tlast(conc = numeric())),
    regexp='argument "time" is missing, with no default'
  )
  expect_error(
    pk.calc.tlast(time = numeric()),
    regexp = 'argument "conc" is missing, with no default'
  )

  # It calculates tlast correctly
  expect_equal(pk.calc.tlast(c(1, 2), c(0, 1)),
               1)
  expect_equal(pk.calc.tlast(c(0, 0), c(0, 1)),
               NA)
  expect_equal(pk.calc.tlast(c(0, 1), c(0, 1)),
               1)
  expect_equal(pk.calc.tlast(c(1, 0), c(0, 1)),
               0)
  expect_equal(pk.calc.tlast(c(1, 1), c(0, 1)),
               1)
})

test_that("pk.calc.tfirst", {
  expect_error(
    pk.calc.tfirst(),
    regexp='argument "conc" is missing, with no default'
  )
  expect_error(
    pk.calc.tfirst(conc=1),
    regexp='argument "time" is missing, with no default'
  )
  expect_equal(
    pk.calc.tfirst(conc=1, time=2, check=TRUE),
    2
  )
})

test_that("pk.calc.clast.obs", {
  # Ensure that it handles BLQ (0) values correctly
  c1 <- c(0, 1, 2, 0)
  t1 <- c(0, 1, 2, 3)
  expect_equal(pk.calc.clast.obs(c1, t1), 2)

  # Ensure that it handles all ALQ values correctly
  c1 <- c(0, 1, 2, 3)
  t1 <- c(0, 1, 2, 3)
  expect_equal(pk.calc.clast.obs(c1, t1), 3)

  # Ensure that it handles NA values correctly
  c1 <- c(0, 1, 2, NA)
  t1 <- c(0, 1, 2, 3)
  expect_equal(pk.calc.clast.obs(c1, t1), 2)

  c1 <- c(0, 1, NA, 3)
  t1 <- c(0, 1, 2, 3)
  expect_equal(pk.calc.clast.obs(c1, t1), 3)

  c1 <- c(NA, NA, NA, NA)
  t1 <- c(0, 1, 2, 3)
  expect_warning(
    v1 <- pk.calc.clast.obs(c1, t1),
    class = "pknca_conc_all_missing"
  )
  expect_equal(v1, NA_real_)

  c1 <- rep(0, 4)
  t1 <- c(0, 1, 2, 3)
  expect_equal(pk.calc.clast.obs(c1, t1), 0)
})

test_that("pk.calc.thalf.eff", {
  # No input gives equivalent no output
  expect_equal(
    pk.calc.thalf.eff(numeric()),
    numeric()
  )

  # NA input gives equivalent NA output
  expect_equal(pk.calc.thalf.eff(NA),
               as.numeric(NA))

  # Numbers mixed with NA give appropriate output
  d1 <- c(0, 1, NA, 3)
  r1 <- log(2)*d1
  expect_equal(pk.calc.thalf.eff(d1),
               r1)
})

test_that("pk.calc.kel", {
  # No input gives equivalent no output
  expect_equal(
    pk.calc.kel(numeric()),
    numeric()
  )

  # NA input gives equivalent NA output
  expect_equal(
    pk.calc.kel(NA),
    as.numeric(NA)
  )

  # Numbers mixed with NA give appropriate output
  d1 <- c(0, 1, NA, 3)
  r1 <- 1/d1
  expect_equal(pk.calc.kel(d1), r1)
})

test_that("pk.calc.cl", {
  # Ensure that dose and auc are required
  expect_error(pk.calc.cl(auc=NA),
               info="dose is required for clearance calculation")
  expect_error(pk.calc.cl(dose=NA),
               info="auc is required for clearance calculation")

  expect_equal(pk.calc.cl(dose=10, auc=100), 0.1,
               info="Normal clearance calculation works")
  expect_equal(pk.calc.cl(dose=c(10, 10, 10), auc=c(10, NA, 100)),
               c(1, NA, 0.1),
               info="Vectors for both dose and auc give vectors for clearance (with NAs)")
  expect_equal(pk.calc.cl(dose=c(50, 50), auc=100),
               1,
               info="Vector for dose and scalar for auc scalar output with the sum of doses")
  expect_equal(pk.calc.cl(dose=c(50, 50), auc=c(NA, 100)),
               c(NA_real_, 0.5),
               info="NA generation works with NA, including for a vector")
  expect_equal(pk.calc.cl(dose=c(50, 50), auc=c(0, 100)),
               c(NA_real_, 0.5),
               info="NA generation works with zero, including for a vector")
})

test_that("pk.calc.f", {
  expect_equal(pk.calc.f(1, 1, 1, 2), 2,
               info="Standard bioavailability calculation")
  expect_equal(pk.calc.f(NA, 1, 1, 1), NA_real_,
               info="NA handling per parameter in bioavailability (1)")
  expect_equal(pk.calc.f(1, NA, 1, 1), NA_real_,
               info="NA handling per parameter in bioavailability (2)")
  expect_equal(pk.calc.f(1, 1, NA, 1), NA_real_,
               info="NA handling per parameter in bioavailability (3)")
  expect_equal(pk.calc.f(1, 1, 1, NA), NA_real_,
               info="NA handling per parameter in bioavailability (4)")
  expect_equal(pk.calc.f(0, 1, 1, 1), NA_real_,
               info="Zero handling per parameter in bioavailability (1)")
  expect_equal(pk.calc.f(1, 0, 1, 1), NA_real_,
               info="Zero handling per parameter in bioavailability (2)")
  expect_equal(pk.calc.f(1, 1, 0, 1), NA_real_,
               info="Zero handling per parameter in bioavailability (3)")
  expect_equal(pk.calc.f(1, 1, 1, 0), 0,
               info="Zero handling per parameter in bioavailability (4)")
})

test_that("pk.calc.aucpext", {
  expect_equal(pk.calc.aucpext(1, 2), 50)
  expect_equal(pk.calc.aucpext(1.8, 2), 10)
  expect_warning(v1 <- pk.calc.aucpext(2, 1),
                 regexp="aucpext is typically only calculated when aucinf is greater than auclast.")
  expect_equal(v1, -100)
  expect_warning(expect_warning(
    v2 <- pk.calc.aucpext(auclast=0, aucinf=0),
    class = "pknca_aucpext_aucinf_le_auclast"),
    class = "pknca_aucpext_aucinf_auclast_positive"
  )
  expect_equal(v2, NA_real_,
               info="aucinf<=0 gives NA_real_ (not infinity)")
  expect_equal(pk.calc.aucpext(NA, NA),
               NA_real_,
               info="Percent extrapolated is NA when both inputs are NA.")
  expect_equal(pk.calc.aucpext(NA, 1),
               NA_real_,
               info="Percent extrapolated is NA when input auclast is NA.")
  expect_equal(pk.calc.aucpext(1, NA),
               NA_real_,
               info="Percent extrapolated is NA when input aucinf is NA.")
  expect_error(pk.calc.aucpext(1:2, 1:3),
               regexp="auclast and aucinf must either be a scalar or the same length.",
               info="aucpext input length checks require consistency.")
  expect_equal(pk.calc.aucpext(c(1, NA), 2),
               c(50, NA_real_),
               info="Percent extrapolated works with vector/scalar input.")
  expect_equal(pk.calc.aucpext(1, c(2, NA)),
               c(50, NA_real_),
               info="Percent extrapolated works with scalar/vector input.")
  expect_equal(pk.calc.aucpext(c(1, 1), c(2, NA)),
               c(50, NA_real_),
               info="Percent extrapolated works with vector/vector input.")
})

test_that("pk.calc.mrt", {
  expect_equal(pk.calc.mrt(auc=1, aumc=2),
               2,
               info="MRT is calculated correctly")
})

test_that("pk.calc.mrt.iv", {
  expect_equal(pk.calc.mrt.iv(auc=1, aumc=2, duration.dose=1),
               1.5,
               info="MRT.iv is calculated correctly")
  expect_equal(pk.calc.mrt.iv(auc=1, aumc=2, duration.dose=0),
               2,
               info="MRT.iv is calculated correctly when duration is 0")
  expect_equal(pk.calc.mrt.iv(auc=1, aumc=2, duration.dose=NA),
               NA_real_,
               info="MRT.iv is calculated correctly when duration is missing")
  expect_equal(pk.calc.mrt.iv(auc=0, aumc=2, duration.dose=NA),
               NA_real_,
               info="MRT.iv is calculated correctly when auc is zero")
})

test_that("pk.calc.mrt.md", {
  expect_equal(pk.calc.mrt.md(1, 2, 1.5, 24), 2 + 24*0.5)
  expect_equal(pk.calc.mrt.md(0, 2, 1.5, 24), NA_real_,
               info="auctau <= 0 becomes NA (not Inf)")
})

test_that("pk.calc.vz", {
  # Ensure that cl and lambda.z are required
  expect_equal(pk.calc.vz(cl=NA, lambda.z=NA), NA_integer_)
  expect_error(pk.calc.vz(cl=NA),
               info="lambda.z required for Vz calculation")
  expect_error(pk.calc.vz(lambda.z=NA),
               info="CL required for Vz calculation")

  # Ensure that length of cl and lambda.z are either 1 or the same length
  expect_error(
    pk.calc.vz(cl=1:2, lambda.z=1:3),
    regexp="'cl' and 'lambda.z' must be the same length",
    info="CL and lambda.z must be the same length (CL shorter)"
  )
  expect_error(
    pk.calc.vz(cl=1:3, lambda.z=1:2),
    regexp="'cl' and 'lambda.z' must be the same length",
    info="CL and lambda.z must be the same length (lambda.z shorter)"
  )

  # Estimate a single Vz (with permutations to ensure the right math
  # is happening)
  expect_equal(pk.calc.vz(cl=1, lambda.z=1), 1,
               info="vz math test 1")
  expect_equal(pk.calc.vz(cl=1, lambda.z=2), 0.5,
               info="vz math test 2")
  expect_equal(pk.calc.vz(cl=2, lambda.z=1), 2,
               info="vz math test 3")
  expect_equal(pk.calc.vz(cl=2, lambda.z=2), 1,
               info="vz math test 4")

  # Ensure that NA can go into either position or both
  expect_equal(pk.calc.vz(cl=NA, lambda.z=1), NA_integer_,
               info="Vz with missing (NA) cl")
  expect_equal(pk.calc.vz(cl=1, lambda.z=NA), NA_integer_,
               info="Vz with missing (NA) lambda.z")
  expect_equal(pk.calc.vz(cl=NA, lambda.z=NA), NA_integer_,
               info="Vz with missing (NA) lambda.z and cl")

  # vectorized vz calculation works
  expect_equal(pk.calc.vz(cl=
                            c(1, 1, 1, 2, 2, 2, NA, NA, NA),
                          lambda.z=
                            c(1, 2, NA, 1, 2, NA, 1, 2, NA)),
               c(1, 0.5, NA, 2, 1, NA, NA, NA, NA),
               info="Vz with vector inputs including missing values for both parameters (cl and lambda.z)")
})

test_that("pk.calc.vss and its wrappers", {
  expect_equal(pk.calc.vss(1, 1), 1)
  expect_equal(pk.calc.vss(2, 1), 2)
  expect_equal(pk.calc.vss(1, 2), 2)
})

test_that("pk.calc.cav", {
  expect_equal(pk.calc.cav(2, 0, 1), 2)
  expect_equal(pk.calc.cav(NA, 0, 1), NA_real_)
  expect_equal(pk.calc.cav(2, 1, 1), NA_real_,
               info="If end == start, return NA_real_")
})

test_that("pk.calc.ctrough", {
  expect_equal(pk.calc.ctrough(1:5, 0:4, 0), 1,
               info="Found and it's the first time")
  expect_equal(pk.calc.ctrough(1:5, 0:4, 1), 2,
               info="Found and it's not the first time")
  expect_equal(pk.calc.ctrough(1:5, 0:4, 1.5), NA_real_,
               info="Not found")
  expect_error(
    pk.calc.ctrough(1:5, c(0, 0:3), 0),
    regexp = "Assertion on 'time' failed: Contains duplicated values, position 2."
  )
})

test_that("pk.calc.ptr", {
  expect_equal(pk.calc.ptr(cmax=2, ctrough=1), 2,
               info="Confirm that the ratio goes the right way")
  expect_equal(pk.calc.ptr(2, 0), NA_real_,
               info="Division by zero returns NA")
})

test_that("pk.calc.tlag", {
  expect_equal(pk.calc.tlag(1:5, 0:4), 0,
               info="find the first point")
  expect_equal(pk.calc.tlag(c(0, 0, 0, 0, 1), 0:4), 3,
               info="find the next to last point")
  expect_equal(pk.calc.tlag(c(0, 0, 0, 0, 0), 0:4), NA_real_,
               info="No increase gives NA")
  expect_equal(pk.calc.tlag(5:1, 0:4), NA_real_,
               info="No increase gives NA")
})

test_that("pk.calc.deg.fluc", {
  expect_equal(pk.calc.deg.fluc(cmax=100, cmin=10, cav=45), 200,
               info="Degree of fluctuation math works")
  expect_equal(pk.calc.deg.fluc(cmax=100, cmin=10, cav=0), NA_real_,
               info="Degree of fluctuation returns NA when cav=0")
})

test_that("pk.calc.swing", {
  expect_equal(pk.calc.swing(100, 10), 900,
               info="Swing math works")
  expect_equal(pk.calc.swing(100, 0), Inf,
               info="Swing handle Ctrough=0")
  expect_equal(pk.calc.swing(100, -1), Inf,
               info="Swing handle Ctrough<0")
})

test_that("pk.calc.ceoi", {
  expect_equal(pk.calc.ceoi(conc=0:5, time=0:5, duration.dose=1),
               1,
               info="Ceoi returns the concentration at the end of the dosing duration")
  expect_equal(pk.calc.ceoi(conc=0:5, time=0:5, duration.dose=1.5),
               NA_real_,
               info="Ceoi returns NA if there is no measurement at the end of the dosing duration")
  expect_equal(pk.calc.ceoi(conc=0:5, time=0:5, duration.dose=NA),
               NA_real_,
               info="Ceoi returns NA if there is no dosing duration")
})

test_that("pk.calc.aucabove", {
  expect_equal(
    pk.calc.aucabove(conc = c(0:5, 1), time = 0:6, conc_above = 2),
    pk.calc.auc.all(conc = c(0, 0, 0, 1:3, 0), time = 0:6)
  )

  expect_equal(
  pk.calc.aucabove(conc = c(0:5, 1), time = 0:6, conc_above = NA_real_),
  structure(NA_real_, exclude = "Missing concentration to be above")
  )

  # Confirm that it works through NCA calculations
  d_conc <- data.frame(conc = c(2, 1:5, 3), time = 0:6)
  d_intervals <- data.frame(start = 0, end = 6, aucabove.trough.all = TRUE, aucabove.predose.all = TRUE)
  o_conc <- PKNCAconc(d_conc, conc~time)
  o_data <- PKNCAdata(o_conc, intervals = d_intervals)
  suppressMessages(
    o_nca <- pk.nca(o_data)
  )
  expect_equal(
    as.data.frame(o_nca),
    tibble::tibble(
      start = 0, end = 6,
      PPTESTCD = c("ctrough", "cstart", "aucabove.predose.all", "aucabove.trough.all"),
      PPORRES =
        c(
          3, 2,
          pk.calc.aucabove(conc = d_conc$conc, time = d_conc$time, conc_above = 2),
          pk.calc.aucabove(conc = d_conc$conc, time = d_conc$time, conc_above = 3)
        ),
      exclude = NA_character_
    )
  )
})

test_that("pk.calc.count_conc", {
  expect_equal(pk.calc.count_conc(1:5), 5)
  expect_equal(pk.calc.count_conc(c(1:2, NA)), 2)
  expect_equal(pk.calc.count_conc(c(1:2, NA, 0)), 3)
  expect_equal(suppressWarnings(pk.calc.count_conc(numeric())), 0)
  expect_equal(suppressWarnings(pk.calc.count_conc(NA)), 0)
})

test_that("pk.calc.count_conc_measured", {
  expect_equal(pk.calc.count_conc_measured(1:5), 5)
  expect_equal(pk.calc.count_conc_measured(c(1:2, NA)), 2)
  # including BLQ
  expect_equal(pk.calc.count_conc_measured(c(1:2, NA, 0)), 2)
  # Other
  expect_equal(suppressWarnings(pk.calc.count_conc_measured(numeric())), 0)
  expect_equal(suppressWarnings(pk.calc.count_conc_measured(NA)), 0)
})

test_that("pk.calc.totdose", {
  expect_equal(pk.calc.totdose(1), 1)
  expect_equal(pk.calc.totdose(c(1, 1)), 2)
})

test_that("pk.calc.cstart", {
  expect_equal(pk.calc.cstart(1:5, 0:4, 0), 1)
  expect_equal(pk.calc.cstart(1:5, 0:4, 1), 2)
  expect_equal(pk.calc.cstart(1:5, 0:4, 1.5), NA_real_)
  expect_error(
    pk.calc.cstart(1:5, c(0, 0:3), 0),
    regexp = "Assertion on 'time' failed: Contains duplicated values, position 2."
  )
})

# Generics ####

#' Convert an object into a PKNCAconc object
#'
#' @param x The object to convert
#' @param ... Passed to subsequent methods
#' @return A converted object
#' @export
as_PKNCAconc <- function(x, ...) {
  UseMethod("as_PKNCAconc")
}

#' @describeIn as_PKNCAconc Convert an object into a PKNCAdose object
#' @export
as_PKNCAdose <- function(x, ...) {
  UseMethod("as_PKNCAdose")
}

#' @describeIn as_PKNCAconc Convert an object into a PKNCAdata object
#' @export
as_PKNCAdata <- function(x, ...) {
  UseMethod("as_PKNCAdata")
}

#' @describeIn as_PKNCAconc Convert an object into a PKNCAresults object
#' @export
as_PKNCAresults <- function(x, ...) {
  UseMethod("as_PKNCAresults")
}

# as_PKNCAconc ####

#' @export
as_PKNCAconc.PKNCAresults <- function(x, ...) {
  as_PKNCAconc(as_PKNCAdata(x, ...), ...)
}

#' @export
as_PKNCAconc.PKNCAdata <- function(x, ...) {
  x$conc
}

#' @export
as_PKNCAconc.PKNCAconc <- function(x, ...) {
  x
}

# as_PKNCAdose ####

#' @export
as_PKNCAdose.PKNCAresults <- function(x, ...) {
  as_PKNCAdose(as_PKNCAdata(x, ...), ...)
}

#' @export
as_PKNCAdose.PKNCAdata <- function(x, ...) {
  x$dose
}

#' @export
as_PKNCAdose.PKNCAdose <- function(x, ...) {
  x
}

# as_PKNCAdata ####

#' @export
as_PKNCAdata.PKNCAresults <- function(x, ...) {
  x$data
}

#' @export
as_PKNCAdata.PKNCAdata <- function(x, ...) {
  x
}

# as_PKNCAresults ####

#' @export
as_PKNCAresults.PKNCAresults <- function(x, ...) {
  x
}

#' @export
as_PKNCAresults.PKNCAdata <- function(x, ...) {
  # the conversion from PKNCAdata to PKNCAresults is to do the calculation
  pk.nca(x)
}

test_that("as_PKNCA*", {
  tmpconc <- generate.conc(2, 1, 0:24)
  tmpdose <- generate.dose(tmpconc)
  myconc <- PKNCAconc(tmpconc, formula=conc~time|treatment+ID)
  mydose <- PKNCAdose(tmpdose, formula=dose~time|treatment+ID)
  mydata <- PKNCAdata(myconc, mydose)
  myresult <- pk.nca(mydata)

  # as_PKNCAconc
  expect_equal(as_PKNCAconc(myconc), myconc)
  expect_error(as_PKNCAconc(mydose))
  expect_equal(as_PKNCAconc(mydata), myconc)
  expect_equal(as_PKNCAconc(myresult), myconc)

  # as_PKNCAdose
  expect_error(as_PKNCAdose(myconc))
  expect_equal(as_PKNCAdose(mydose), mydose)
  expect_equal(as_PKNCAdose(mydata), mydose)
  expect_equal(as_PKNCAdose(myresult), mydose)

  # as_PKNCAdata
  expect_equal(as_PKNCAdata(mydata), mydata)
  # This is not the same as the input data because options have been applied
  expect_equal(as_PKNCAdata(myresult), myresult$data)

  # as_PKNCAresults
  # Ignore the provenance as that may differ due to recalculation and loaded packages
  expect_equal(as_PKNCAresults(mydata), myresult, ignore_attr = TRUE)
  expect_equal(as_PKNCAresults(myresult), myresult)
})

#' Create a unit assignment and conversion table
#'
#' This data.frame is typically used for the `units` argument for [PKNCAdata()].
#' If a unit is not given, then all of the units derived from that unit will be
#' `NA`.
#'
#' @param concu,doseu,amountu,timeu Units for concentration, dose, amount, and
#'   time in the source data
#' @param concu_pref,doseu_pref,amountu_pref,timeu_pref Preferred units for
#'   reporting; `conversions` will be automatically.
#' @param conversions An optional data.frame with columns of c("PPORRESU",
#'   "PPSTRESU", "conversion_factor") for the original calculation units, the
#'   standardized units, and a conversion factor to multiply the initial value
#'   by to get a standardized value.  This argument overrides any preferred unit
#'   conversions from `concu_pref`, `doseu_pref`, `amountu_pref`, or
#'   `timeu_pref`.
#' @returns A unit conversion table with columns for "PPTESTCD" and "PPORRESU"
#'   if `conversions` is not given, and adding "PPSTRESU" and
#'   "conversion_factor" if `conversions` is given.
#' @seealso The `units` argument for [PKNCAdata()]
#' @examples
#' pknca_units_table() # only parameters that are unitless
#' pknca_units_table(
#'   concu="ng/mL", doseu="mg/kg", amountu="mg", timeu="hr"
#' )
#' pknca_units_table(
#'   concu="ng/mL", doseu="mg/kg", amountu="mg", timeu="hr",
#'   # Convert clearance and volume units to more understandable units with
#'   # automatic unit conversion
#'   conversions=data.frame(
#'     PPORRESU=c("(mg/kg)/(hr*ng/mL)", "(mg/kg)/(ng/mL)"),
#'     PPSTRESU=c("mL/hr/kg", "mL/kg")
#'   )
#' )
#' pknca_units_table(
#'   concu="mg/L", doseu="mg/kg", amountu="mg", timeu="hr",
#'   # Convert clearance and volume units to molar units (assuming
#'   conversions=data.frame(
#'     PPORRESU=c("mg/L", "(mg/kg)/(hr*ng/mL)", "(mg/kg)/(ng/mL)"),
#'     PPSTRESU=c("mmol/L", "mL/hr/kg", "mL/kg"),
#'     # Manual conversion of concentration units from ng/mL to mmol/L (assuming
#'     # a molecular weight of 138.121 g/mol)
#'     conversion_factor=c(1/138.121, NA, NA)
#'   )
#' )
#'
#' # This will make all time-related parameters use "day" even though the
#' # original units are "hr"
#' pknca_units_table(
#'   concu = "ng/mL", doseu = "mg/kg", timeu = "hr", amountu = "mg",
#'   timeu_pref = "day"
#' )
#' @export
pknca_units_table <- function(concu, doseu, amountu, timeu,
                              concu_pref = NULL, doseu_pref = NULL, amountu_pref = NULL, timeu_pref = NULL,
                              conversions = data.frame()) {
  checkmate::assert_data_frame(conversions)
  if (nrow(conversions) > 0) {
    checkmate::assert_names(
      names(conversions),
      subset.of = c("PPORRESU", "PPSTRESU", "conversion_factor"),
      must.include = c("PPORRESU", "PPSTRESU")
    )
    if (!("conversion_factor" %in% names(conversions))) {
      conversions$conversion_factor <- NA_real_
    }
  }

  # The unit conversions are grouped by the type of inputs required
  ret <-
    rbind(
      pknca_units_table_unitless(),
      pknca_units_table_time(timeu=timeu),
      pknca_units_table_conc(concu=concu),
      pknca_units_table_amount(amountu=amountu),
      pknca_units_table_amount_dose(amountu = amountu, doseu = doseu),
      pknca_units_table_dose(doseu = doseu),
      pknca_units_table_conc_dose(concu=concu, doseu=doseu),
      pknca_units_table_conc_time(concu=concu, timeu=timeu),
      pknca_units_table_conc_time_dose(concu=concu, timeu=timeu, doseu=doseu),
      pknca_units_table_conc_time_amount(concu=concu, timeu=timeu, amountu=amountu)
    )

  # Generate preferred units and merge them into `conversions`
  if (any(!is.null(concu_pref), !is.null(doseu_pref), !is.null(amountu_pref), !is.null(timeu_pref))) {
    ret_pref <-
      pknca_units_table(
        concu = choose_first(concu_pref, concu),
        doseu = choose_first(doseu_pref, doseu),
        amountu = choose_first(amountu_pref, amountu),
        timeu = choose_first(timeu_pref, timeu)
      )
    ret_pref <- dplyr::rename(ret_pref, PPSTRESU = "PPORRESU")
    conversions_pref <- dplyr::left_join(ret, ret_pref, by = "PPTESTCD")
    conversions_pref$PPTESTCD <- NULL
    conversions_pref <- unique(conversions_pref)
    # Drop units that are not converted
    conversions_pref <- conversions_pref[conversions_pref$PPORRESU != conversions_pref$PPSTRESU, ]
    # Drop units that are not provided
    conversions_pref <- conversions_pref[!is.na(conversions_pref$PPORRESU), ]
    conversions_pref$conversion_factor <- NA_real_
    for (idx in seq_len(nrow(conversions))) {
      # Use the original conversions argument over `conversions_pref`
      mask_pref <- conversions_pref$PPORRESU %in% conversions$PPORRESU[idx]
      if (!any(mask_pref)) {
        stop("Cannot find PPORRESU match between conversions and preferred unit conversions.  Check PPORRESU values in 'conversions' argument.")
      }
      conversions_pref$PPSTRESU[mask_pref] <- conversions$PPSTRESU[idx]
      conversions_pref$conversion_factor[mask_pref] <- conversions$conversion_factor[idx]
    }
    conversions <- conversions_pref
  }

  extra_cols <- setdiff(ret$PPTESTCD, names(PKNCA::get.interval.cols()))
  if (length(extra_cols) > 0) {
    stop("Please report a bug.  Unknown NCA parameters have units defined: ", paste(extra_cols, collapse=", ")) # nocov
  }

  # Apply conversion factors
  if (nrow(conversions) > 0) {
    stopifnot(!duplicated(conversions$PPORRESU))
    # PPSTRESU may be duplicated because some differing original units may
    # converge (e.g. cmax.dn and vss)
    stopifnot(length(setdiff(names(conversions), c("PPORRESU", "PPSTRESU", "conversion_factor"))) == 0)
    if (any(is.na(conversions$conversion_factor)) && !requireNamespace("units", quietly=TRUE)) {
      stop("The units package is required for automatic unit conversion") # nocov
    }
    for (idx in which(is.na(conversions$conversion_factor))) {
      conversions$conversion_factor[idx] <-
        as.numeric(
          units::set_units(
            units::set_units(
              1,
              conversions$PPORRESU[idx], mode="standard"
            ),
            conversions$PPSTRESU[idx], mode="standard"
          )
        )
    }
    unexpected_conversions <- setdiff(conversions$PPORRESU, ret$PPORRESU)
    if (length(unexpected_conversions) > 0) {
      warning(
        "The following unit conversions were supplied but do not match any units to convert: ",
        paste0("'", unexpected_conversions, "'", collapse=", ")
      )
    }
    ret <-
      dplyr::left_join(
        ret, conversions,
        by=intersect(names(ret), names(conversions))
      )
    # anything that does not have a specific conversion factor is assumed to be
    # in the correct units, and a unit conversion factor is used to keep the
    # units the same.
    ret$PPSTRESU[is.na(ret$PPSTRESU)] <- ret$PPORRESU[is.na(ret$PPSTRESU)]
    ret$conversion_factor[is.na(ret$conversion_factor)] <- 1
  }
  ret
}

pknca_units_table_unitless <- function() {
  rbind(
    data.frame(
      PPORRESU="unitless",
      PPTESTCD=pknca_find_units_param(unit_type="unitless"),
      stringsAsFactors=FALSE
    ),
    data.frame(
      PPORRESU="fraction",
      PPTESTCD=pknca_find_units_param(unit_type="fraction"),
      stringsAsFactors=FALSE
    ),
    data.frame(
      PPORRESU="%",
      PPTESTCD=pknca_find_units_param(unit_type="%"),
      stringsAsFactors=FALSE
    ),
    data.frame(
      PPORRESU="count",
      PPTESTCD=pknca_find_units_param(unit_type="count"),
      stringsAsFactors=FALSE
    )
  )
}

choose_first <- function(x, y, .default = NA) {
  if (!useless(x)) {
    x
  } else if (!useless(y)) {
    y
  } else {
    .default
  }
}

useless <- function(x) {
  if (missing(x)) {
    return(TRUE)
  } else if (length(x) > 1) {
    stop("Only one unit may be provided at a time: ", paste(x, collapse = ", "))
  }
  is.null(x) || is.na(x)
}

pknca_units_table_time <- function(timeu) {
  if (useless(timeu)) {
    timeu <- NA_character_
  }
  inverse_timeu <-
    if (is.na(timeu)) {
      NA_character_
    } else {
      sprintf("1/%s", pknca_units_add_paren(timeu))
    }
  rbind(
    data.frame(
      PPORRESU=timeu,
      PPTESTCD=pknca_find_units_param(unit_type="time"),
      stringsAsFactors=FALSE
    ),
    data.frame(
      PPORRESU=inverse_timeu,
      PPTESTCD=pknca_find_units_param(unit_type="inverse_time"),
      stringsAsFactors=FALSE
    )
  )
}

pknca_units_table_conc <- function(concu) {
  if (useless(concu)) {
    concu <- NA_character_
  }
  data.frame(
    PPORRESU=concu,
    PPTESTCD=pknca_find_units_param(unit_type="conc"),
    stringsAsFactors=FALSE
  )
}

pknca_units_table_amount <- function(amountu) {
  if (useless(amountu)) {
    amountu <- NA_character_
  }
  data.frame(
    PPORRESU=amountu,
    PPTESTCD=pknca_find_units_param(unit_type="amount"),
    stringsAsFactors=FALSE
  )
}

pknca_units_table_amount_dose <- function(amountu, doseu) {
  if (useless(amountu) || useless(doseu)) {
    amount_doseu <- NA_character_
  } else {
    amount_doseu <- sprintf("%s/%s", pknca_units_add_paren(amountu), pknca_units_add_paren(doseu))
  }
  data.frame(
    PPORRESU=amount_doseu,
    PPTESTCD=pknca_find_units_param(unit_type="amount_dose"),
    stringsAsFactors=FALSE
  )
}

pknca_units_table_dose <- function(doseu) {
  if (useless(doseu)) {
    doseu <- NA_character_
  }
  data.frame(
    PPORRESU=doseu,
    PPTESTCD=pknca_find_units_param(unit_type="dose"),
    stringsAsFactors=FALSE
  )
}

pknca_units_table_conc_dose <- function(concu, doseu) {
  if (useless(concu) || useless(doseu)) {
    conc_dosenorm <- NA_character_
    volume <- NA_character_
  } else {
    conc_dosenorm <- sprintf("%s/%s", pknca_units_add_paren(concu), pknca_units_add_paren(doseu))
    volume <- sprintf("%s/%s", pknca_units_add_paren(doseu), pknca_units_add_paren(concu))
  }
  rbind(
    data.frame(
      PPORRESU=conc_dosenorm,
      PPTESTCD=pknca_find_units_param(unit_type="conc_dosenorm"),
      stringsAsFactors=FALSE
    ),
    data.frame(
      # Volume units
      PPORRESU=volume,
      PPTESTCD=pknca_find_units_param(unit_type="volume"),
      stringsAsFactors=FALSE
    )
  )
}

pknca_units_table_conc_time <- function(concu, timeu) {
  if (useless(concu) || useless(timeu)) {
    auc <- NA_character_
    aumc <- NA_character_
  } else {
    auc <- sprintf("%s*%s", timeu, concu)
    aumc <- sprintf("%s^2*%s", pknca_units_add_paren(timeu), concu)
  }
  rbind(
    data.frame(
      # AUC units
      PPORRESU=auc,
      PPTESTCD=pknca_find_units_param(unit_type="auc"),
      stringsAsFactors=FALSE
    ),
    data.frame(
      # AUMC units
      PPORRESU=aumc,
      PPTESTCD=pknca_find_units_param(unit_type="aumc"),
      stringsAsFactors=FALSE
    )
  )
}

pknca_units_table_conc_time_dose <- function(concu, timeu, doseu) {
  if (useless(concu) || useless(timeu) || useless(doseu)) {
    auc_dosenorm <-
      aumc_dosenorm <-
      clearance <-
      NA_character_
  } else {
    auc_dosenorm <- sprintf("(%s*%s)/%s", timeu, concu, pknca_units_add_paren(doseu))
    aumc_dosenorm <- sprintf("(%s^2*%s)/%s", pknca_units_add_paren(timeu), concu, pknca_units_add_paren(doseu))
    clearance <- sprintf("%s/(%s*%s)", pknca_units_add_paren(doseu), timeu, concu)
  }
  rbind(
    data.frame(
      # AUC units, dose-normalized
      PPORRESU=auc_dosenorm,
      PPTESTCD=pknca_find_units_param(unit_type="auc_dosenorm"),
      stringsAsFactors=FALSE
    ),
    data.frame(
      # AUMC units, dose-normalized
      PPORRESU=aumc_dosenorm,
      PPTESTCD=pknca_find_units_param(unit_type="aumc_dosenorm"),
      stringsAsFactors=FALSE
    ),
    data.frame(
      # Clearance units
      PPORRESU=clearance,
      PPTESTCD=pknca_find_units_param(unit_type="clearance"),
      stringsAsFactors=FALSE
    )
  )
}

pknca_units_table_conc_time_amount <- function(concu, timeu, amountu) {
  if (useless(concu) || useless(timeu) || useless(amountu)) {
    renal_clearance <- NA_character_
  } else {
    renal_clearance <- sprintf("%s/(%s*%s)", pknca_units_add_paren(amountu), timeu, concu)
  }
  data.frame(
    # Renal clearance units
    PPORRESU=renal_clearance,
    PPTESTCD=pknca_find_units_param(unit_type="renal_clearance"),
    stringsAsFactors=FALSE
  )
}

#' Find NCA parameters with a given unit type
#'
#' @param unit_type The type of unit as assigned with `add.interval.col`
#' @returns A character vector of parameters with a given unit type
#' @keywords Internal
pknca_find_units_param <- function(unit_type) {
  stopifnot(length(unit_type) == 1)
  stopifnot(is.character(unit_type))
  all_intervals <- get.interval.cols()
  ret <- character()
  for (nm in names(all_intervals)) {
    if (all_intervals[[nm]]$unit_type == unit_type) {
      ret <- c(ret, nm)
    }
  }
  if (length(ret) == 0) {
    stop("No parameters found for unit_type=", unit_type)
  }
  ret
}

#' Add parentheses to a unit value, if needed
#'
#' @param unit The text of the unit
#' @returns The unit with parentheses around it, if needed
#' @keywords Internal
pknca_units_add_paren <- function(unit) {
  mask_paren <- grepl(x=unit, pattern="[*/]")
  ifelse(mask_paren, yes=paste0("(", unit, ")"), no=unit)
}

#' Perform unit conversion (if possible) on PKNCA results
#'
#' @param result The results data.frame
#' @param units The unit conversion table
#' @param allow_partial_missing_units Should missing units be allowed for some
#'   but not all parameters?
#' @returns The result table with units converted
#' @keywords Internal
pknca_unit_conversion <- function(result, units, allow_partial_missing_units = FALSE) {
  ret <- result
  if (!is.null(units)) {
    ret <-
      dplyr::left_join(
        ret, units,
        by=intersect(names(ret), names(units))
      )
    mask_missing_units <- is.na(ret$PPORRESU)
    if (any(mask_missing_units)) {
      msg_missing <-
        paste(
          "Units are provided for some but not all parameters; missing for:",
          paste(sort(unique(ret$PPTESTCD[mask_missing_units])), collapse = ", ")
        )
      if (allow_partial_missing_units) {
        warning(msg_missing)
      } else {
        stop(msg_missing, "\nThis error can be converted to a warning using `PKNCA.options(allow_partial_missing_units = TRUE)`")
      }
    }
    if ("conversion_factor" %in% names(units)) {
      ret$PPSTRES <- ret$PPORRES * ret$conversion_factor
      # Drop the conversion factor column, since it shouldn't be in the output.
      ret <- ret[, setdiff(names(ret), "conversion_factor")]
    }
  }
  ret
}

test_that("all units have been defined", {
  all_units <- pknca_units_table(concu="conc", doseu="dose", amountu="amount", timeu="time")
  expect_equal(
    setdiff(names(get.interval.cols()), c(all_units$PPTESTCD, "start", "end")),
    character()
  )
})

test_that("pknca_find_units_param expected errors", {
  expect_error(
    pknca_find_units_param(unit_type=1:2)
  )
  expect_error(
    pknca_find_units_param(unit_type=1)
  )
  expect_error(
    pknca_find_units_param(unit_type="foo"),
    regexp="No parameters found for unit_type"
  )
})

test_that("pknca_find_units_param", {
  expect_true(
    "cmax" %in% pknca_find_units_param(unit_type="conc")
  )
})

test_that("unit conversion tables are created correctly", {
  expect_true(
    all(
      pknca_units_table(
        concu="ng/mL", doseu="mg/kg", amountu="mg", timeu="hr"
      )$conversion_factor == 1
    )
  )
  automatic_conversion <-
    pknca_units_table(
      concu="ng/mL", doseu="mg/kg", amountu="mg", timeu="hr",
      conversions=data.frame(
        PPORRESU=c("(mg/kg)/(hr*ng/mL)", "(mg/kg)/(ng/mL)"),
        PPSTRESU=c("mL/hr/kg", "mL/kg")
      )
    )
  expect_equal(
    unique(automatic_conversion$conversion_factor[automatic_conversion$PPSTRESU %in% "mL/hr/kg"]),
    1e6,
    tolerance=.Machine$double.eps
  )
  expect_equal(
    unique(automatic_conversion$conversion_factor[automatic_conversion$PPSTRESU %in% "mL/kg"]),
    1e6,
    tolerance=.Machine$double.eps
  )
  expect_equal(
    unique(automatic_conversion$conversion_factor[!(automatic_conversion$PPSTRESU %in% c("mL/hr/kg", "mL/kg"))]),
    1
  )

  mixed_conversion <-
    pknca_units_table(
      concu="mg/L", doseu="mg/kg", amountu="mg", timeu="hr",
      # Convert clearance and volume units to molar units (assuming
      conversions=data.frame(
        PPORRESU=c("mg/L", "(mg/kg)/(hr*mg/L)", "(mg/kg)/(mg/L)"),
        PPSTRESU=c("mmol/L", "mL/hr/kg", "mL/kg"),
        # Manual conversion of concentration units from ng/mL to mmol/L (assuming
        # a molecular weight of 138.121 g/mol)
        conversion_factor=c(1/138.121, NA, NA)
      )
    )
  expect_equal(
    unique(mixed_conversion$conversion_factor[mixed_conversion$PPSTRESU %in% "mL/hr/kg"]),
    1000,
    tolerance=1e-10
  )
  expect_equal(
    unique(mixed_conversion$conversion_factor[mixed_conversion$PPSTRESU %in% "mL/kg"]),
    1000,
    tolerance=1e-10
  )
  expect_equal(
    unique(mixed_conversion$conversion_factor[mixed_conversion$PPSTRESU %in% "mmol/L"]),
    1/138.121,
    tolerance=1e-10
  )
  expect_equal(
    unique(mixed_conversion$conversion_factor[!(mixed_conversion$PPSTRESU %in% c("mL/hr/kg", "mL/kg", "mmol/L"))]),
    1
  )
})

test_that("pknca_units_table", {
  expect_warning(
      pknca_units_table(
        concu="ng/mL", doseu="mg/kg", amountu="mg", timeu="hr",
        conversions=data.frame(PPORRESU="foo", PPSTRESU="bar", conversion_factor=1)
      ),
      regexp="The following unit conversions were supplied but do not match any units to convert"
  )
  # units library errors occur, if units are not convertible
  expect_error(
    pknca_units_table(
      concu="ng/mL", doseu="mg/kg", amountu="mg", timeu="hr",
      conversions=data.frame(PPORRESU="ng/mL", PPSTRESU="mol/L")
    ),
    regexp="cannot convert ng/mL into mol/L"
  )
  expect_error(
    pknca_units_table(
      concu=c("ng/mL", "umol/L"), doseu="mg/kg", amountu="mg", timeu="hr",
      conversions=data.frame(PPORRESU="ng/mL", PPSTRESU="mol/L")
    ),
    regexp = "Only one unit may be provided at a time: ng/mL, umol/L"
  )
  # fraction excreted now has units of amount/dose
  unit_fe_kg <-
    pknca_units_table(
      doseu="mg/kg", amountu="ng",
      conversions=data.frame(PPORRESU="ng/(mg/kg)", PPSTRESU="kg")
    )
  expect_equal(
    unit_fe_kg$conversion_factor[unit_fe_kg$PPTESTCD == "fe"],
    1e-6
  )
  unit_fe_fraction <-
    pknca_units_table(
      doseu="mg", amountu="ng",
      conversions=data.frame(PPORRESU="ng/mg", PPSTRESU="fraction")
    )
  expect_equal(
    unit_fe_fraction$conversion_factor[unit_fe_fraction$PPTESTCD == "fe"],
    1e-6
  )
})

test_that("pknca_units_add_paren", {
  expect_equal(pknca_units_add_paren("mg"), "mg")
  expect_equal(pknca_units_add_paren("mg/kg"), "(mg/kg)")
  expect_equal(pknca_units_add_paren("mg*kg"), "(mg*kg)")
})

test_that("pknca_units_table treats missing, NULL, and NA the same", {
  expect_equal(
    pknca_units_table(),
    pknca_units_table(concu = NULL, doseu = NULL, amountu = NULL, timeu = NULL)
  )
  expect_equal(
    pknca_units_table(),
    pknca_units_table(concu = NA, doseu = NULL, amountu = NULL, timeu = NULL)
  )
  expect_equal(
    pknca_units_table(),
    pknca_units_table(concu = NA, doseu = NA, amountu = NULL, timeu = NULL)
  )
  expect_equal(
    pknca_units_table(),
    pknca_units_table(concu = NA, doseu = NA, amountu = NA, timeu = NULL)
  )
  expect_equal(
    pknca_units_table(),
    pknca_units_table(concu = NA, doseu = NULL, amountu = NULL, timeu = NA)
  )
  expect_true(all(is.na(
    pknca_units_table(concu = "ng/mL", doseu = "mg", amountu = "umol", timeu = NULL) %>%
      dplyr::filter(PPTESTCD %in% c("start", "lambda.z", "auclast", "aumclast", "auclast.dn", "aumclast.dn", "cl.last", "clr.last")) %>%
      dplyr::pull("PPORRESU")
  )))
  expect_true(all(is.na(
    pknca_units_table(concu = "ng/mL", doseu = "mg", amountu = NULL, timeu = "hr") %>%
      dplyr::filter(PPTESTCD %in% c("ae", "clr.last")) %>%
      dplyr::pull("PPORRESU")
  )))
  expect_true(all(is.na(
    pknca_units_table(concu = "ng/mL", doseu = NULL, amountu = "umol", timeu = "hr") %>%
      dplyr::filter(PPTESTCD %in% c("cmax.dn", "vss.last", "cl.last", "auclast.dn", "aumclast.dn")) %>%
      dplyr::pull("PPORRESU")
  )))
  expect_true(all(is.na(
    pknca_units_table(concu = NULL, doseu = "mg", amountu = "umol", timeu = "hr") %>%
      dplyr::filter(PPTESTCD %in% c("cmax", "cmax.dn", "vss.last", "auclast", "aumclast", "auclast.dn", "aumclast.dn", "cl.last", "clr.last")) %>%
      dplyr::pull("PPORRESU")
  )))
})

test_that("allow duplicate PPSTRESU units", {
  d_conversion <-
    data.frame(
      PPORRESU = c("ng/mL", "(ng/mL)/(mg/kg)", "(mg/kg)/(hr*ng/mL)", "(mg/kg)/(ng/mL)"),
      PPSTRESU = c("mg/mL", "mL/kg", "mL/(h*kg)", "mL/kg")
    )
  # No error for consistent volume conversion
  expect_silent(
    pknca_units_table(concu = "ng/mL", doseu = "mg/kg", timeu = "hr", conversions = d_conversion)
  )
})

test_that("Use preferred units (#197)", {
  prep <-
    pknca_units_table(
      concu = "ng/mL", doseu = "mg/kg", timeu = "hr", amountu = "mg",
      concu_pref = "ug/mL"
    )
  expect_equal(prep$conversion_factor[prep$PPTESTCD == "cmax"], 0.001)
  prep <-
    pknca_units_table(
      concu = "ng/mL", doseu = "mg/kg", timeu = "hr", amountu = "mg",
      doseu_pref = "ug/kg"
    )
  expect_equal(prep$conversion_factor[prep$PPTESTCD == "cmax.dn"], 0.001)
  prep <-
    pknca_units_table(
      concu = "ng/mL", doseu = "mg/kg", timeu = "hr", amountu = "mg",
      timeu_pref = "day"
    )
  expect_equal(prep$conversion_factor[prep$PPTESTCD == "tmax"], 1/24)
  prep <-
    pknca_units_table(
      concu = "ng/mL", doseu = "mg/kg", timeu = "hr", amountu = "mg",
      amountu_pref = "kg"
    )
  expect_equal(prep$conversion_factor[prep$PPTESTCD == "clr.obs"], 1e-6)

  # conversions can override the preferred units parameter
  prep <-
    pknca_units_table(
      concu = "ng/mL", doseu = "mg/kg", timeu = "hr", amountu = "mg",
      timeu_pref = "day",
      conversions = data.frame(PPORRESU = "hr^2*ng/mL", PPSTRESU = "min^2*ng/mL")
    )
  prep_no_conversions <-
    pknca_units_table(
      concu = "ng/mL", doseu = "mg/kg", timeu = "hr", amountu = "mg",
      timeu_pref = "day"
    )
  expect_equal(prep$conversion_factor[prep$PPTESTCD == "tmax"], 1/24)
  expect_equal(prep$conversion_factor[prep$PPTESTCD == "aumcall"], 3600)
  expect_equal(prep_no_conversions$conversion_factor[prep_no_conversions$PPTESTCD == "aumcall"], 24^-2)

  expect_error(
    pknca_units_table(
      concu = "ng/mL", doseu = "mg/kg", timeu = "hr", amountu = "mg",
      timeu_pref = "day",
      conversions = data.frame(PPORRESU = "A", PPSTRESU = "B")
    ),
    regexp = "Cannot find PPORRESU match between conversions and preferred unit conversions.  Check PPORRESU values in 'conversions' argument.",
    fixed = TRUE
  )

  # Not all arguments are required
  expect_silent(
    pknca_units_table(
      concu = "ng/mL",
      doseu = "mg",
      timeu = "hr",
      timeu_pref = "day"
    )
  )
})

test_that("pknca_units_table expected errors", {
  expect_error(
    pknca_units_table(conversions = "A")
  )
  expect_error(
    pknca_units_table(conversions = data.frame(A = 1)),
    # Generate the error to match (in case its text changes)
    regexp =
      attr(
        try(
          checkmate::assert_names("A", subset.of = c("PPORRESU", "PPSTRESU", "conversion_factor"), .var.name = "names(conversions)"),
          silent = TRUE
        ),
        "condition"
      )$message,
    fixed = TRUE
  )
})

test_that("pknca_unit_conversion", {
  results <- data.frame(PPORRES = 1, PPTESTCD = "cmax")

  # No change when no unit conversion occurs
  expect_equal(pknca_unit_conversion(results, units = NULL), results)

  # Adding units with no conversion
  d_u <- data.frame(PPTESTCD = "cmax", PPORRESU = "ng/mL")
  results_u <- data.frame(PPORRES = 1, PPTESTCD = "cmax", PPORRESU = "ng/mL")
  expect_equal(pknca_unit_conversion(results, units = d_u), results_u)

  # Adding units with conversion
  d_u_conv <- data.frame(PPTESTCD = "cmax", PPORRESU = "ng/mL", conversion_factor = 0.001, PPSTRESU = "ug/mL")
  results_u_conv <- data.frame(PPORRES = 1, PPTESTCD = "cmax", PPORRESU = "ng/mL", PPSTRESU = "ug/mL", PPSTRES = 0.001)
  expect_equal(pknca_unit_conversion(results, units = d_u_conv), results_u_conv)

  # Adding units with some units missing gives an error
  d_u_missing <- data.frame(PPTESTCD = "cmax", PPORRESU = NA_character_)
  results_u_missing <- data.frame(PPORRES = 1, PPTESTCD = "cmax", PPORRESU = NA_character_)
  expect_error(
    pknca_unit_conversion(results, units = d_u_missing),
    regexp = "Units are provided for some but not all parameters; missing for: cmax\nThis error can be converted to a warning using `PKNCA.options(allow_partial_missing_units = TRUE)`",
    fixed = TRUE
  )

  # Adding units with some units missing gives an error; that error can be converted to a warning
  d_u_missing <- data.frame(PPTESTCD = "cmax", PPORRESU = NA_character_)
  results_u_missing <- data.frame(PPORRES = 1, PPTESTCD = "cmax", PPORRESU = NA_character_)
  expect_warning(
    pknca_unit_conversion(results, units = d_u_missing, allow_partial_missing_units = TRUE),
    regexp = "Units are provided for some but not all parameters; missing for: cmax",
    fixed = TRUE
  )

  # Fully-integrated test
  d_conc <- as.data.frame(Theoph[Theoph$Subject %in% Theoph$Subject[1], ])
  o_conc <- PKNCAconc(d_conc, conc~Time|Subject)
  o_dose <- PKNCAdose(d_conc[d_conc$Time == 0, ], Dose~Time|Subject)
  # Do not give dose units
  d_units <- pknca_units_table(concu = "mg/L", timeu = "hr")
  d_interval <- data.frame(start = 0, end = Inf, cmax = TRUE, cl.obs = TRUE)
  o_data <- PKNCAdata(o_conc, o_dose, intervals = d_interval, units = d_units)
  expect_error(
    pk.nca(o_data),
    regexp = "Units are provided for some but not all parameters; missing for: cl.obs\nThis error can be converted to a warning using `PKNCA.options(allow_partial_missing_units = TRUE)`",
    fixed = TRUE
  )
  o_data_warn <- PKNCAdata(o_conc, o_dose, intervals = d_interval, units = d_units, options = list(allow_partial_missing_units = TRUE))
  expect_warning(
    pk.nca(o_data_warn),
    regexp = "Units are provided for some but not all parameters; missing for: cl.obs",
    fixed = TRUE
  )
})

#' Determine time at or above a set value
#'
#' Interpolation is performed aligning with `PKNCA.options("auc.method")`.
#' Extrapolation outside of the measured times is not yet implemented.  The
#' `method` may be changed by giving a named `method` argument, as well.
#'
#' For `'lin up/log down'`, if `clast` is above `conc_above` and there are
#' concentrations BLQ after that, linear down is used to extrapolate to the BLQ
#' concentration (equivalent to AUCall).
#'
#' @inheritParams pk.calc.auxc
#' @param conc_above The concentration to be above
#' @param ... Extra arguments.  Currently, the only extra argument that is used
#'   is `method` as described in the details section.
#' @returns the time above the given concentration
#' @export
pk.calc.time_above <- function(conc, time,
                               conc_above,
                               #method=NULL,
                               ...,
                               options=list(),
                               check=TRUE) {
  arglist <- list(...)
  method <- PKNCA.choose.option(name="auc.method", value=arglist$method, options=options)
  if (missing(conc)) {
    stop("conc must be given")
  }
  if (missing(time)) {
    stop("time must be given")
  }
  stopifnot("conc_above must be a scalar"=length(conc_above) == 1)
  stopifnot("conc_above must not be NA"=!is.na(conc_above))
  stopifnot("conc_above must be numeric"=is.numeric(conc_above))
  if (check) {
    assert_conc_time(conc = conc, time = time)
  }
  # Only keep evaluable rows
  data <- data.frame(conc=conc, time=time)[!is.na(conc), , drop=FALSE]
  # Determine intervals where both values are above, the first value is above,
  # or the second value is above
  data$above <- data$conc >= conc_above
  mask_both <- data$above[-1] & data$above[-nrow(data)]
  mask_first <- !data$above[-1] & data$above[-nrow(data)]
  mask_second <- data$above[-1] & !data$above[-nrow(data)]
  conc1 <- data$conc[-nrow(data)]
  conc2 <- data$conc[-1]
  time1 <- data$time[-nrow(data)]
  time2 <- data$time[-1]
  # Calculations
  if (nrow(data) < 2) {
    ret <- structure(NA_real_, exclude="Too few measured concentrations to assess time_above")
  } else if (method %in% 'lin up/log down') {
    linear_up <- conc2 > conc1 | conc2 == conc1
    linear_down <- conc2 == 0
    log_down <- !linear_down & conc2 > 0
    ret <-
      sum((time2 - time1)[mask_both]) +
      # log down
      sum(
        ((log(conc_above) - log(conc1))/(log(conc2) - log(conc1))*(time2 - time1))[mask_first & log_down]
      ) +
      # linear down
      sum(
        ((conc_above - conc1)/(conc2 - conc1)*(time2 - time1))[mask_first & linear_down]
      ) +
      # linear up
      sum(
        ((conc2 - conc_above)/(conc2 - conc1)*(time2 - time1))[mask_second & linear_up]
      )
  } else if (method %in% 'linear') {
    ret <-
      sum((time2 - time1)[mask_both]) +
      # linear down
      sum(
        ((conc_above - conc1)/(conc2 - conc1)*(time2 - time1))[mask_first]
      ) +
      # linear up
      sum(
        ((conc2 - conc_above)/(conc2 - conc1)*(time2 - time1))[mask_second]
      )
  } else {
    # Should be caught by the method assignment above
    stop("Invalid 'method', please report this as a bug: ", method) # nocov
  }
  ret
}
# Add the column to the interval specification
add.interval.col("time_above",
                 FUN="pk.calc.time_above",
                 values=c(FALSE, TRUE),
                 unit_type="time",
                 pretty_name="Time above Concentration",
                 desc="Time above a given concentration")
PKNCA.set.summary(
  name="time_above",
  description="arithmetic mean and standard deviation",
  point=business.mean,
  spread=business.sd
)

test_that("time_above expected errors", {
  expect_error(
    pk.calc.time_above(conc=c(1, 1), time=c(1, 2), conc_above="X", method="linear"),
    regexp='conc_above must be numeric'
  )
  expect_error(
    pk.calc.time_above(conc=c(1, 1), time=c(1, 2), conc_above=1:2, method="linear"),
    regexp='conc_above must be a scalar'
  )
  expect_error(
    pk.calc.time_above(conc=c(1, 1), time=c(1, 2), conc_above=NA, method="linear"),
    regexp='conc_above must not be NA'
  )
  expect_error(
    pk.calc.time_above(time="X", conc_above=5, method="linear"),
    regexp='conc must be given'
  )
  expect_error(
    pk.calc.time_above(conc="X", conc_above=5, method="linear"),
    regexp='time must be given'
  )
  expect_error(
    pk.calc.time_above(conc="X", time="X", conc_above=5, method="linear"),
    regexp="Assertion on 'conc' failed: Must be of type 'numeric', not 'character'."
  )
  expect_error(
    pk.calc.time_above(conc=5, time="X", conc_above=5, method="linear"),
    regexp="Assertion on 'time' failed: Must be of type 'numeric', not 'character'."
  )
  expect_error(
    pk.calc.time_above(conc=5, time=5, conc_above=5, method="foo"),
    regexp='should be one of'
  )
})

test_that("time_above simple scenarios", {
  expect_equal(
    pk.calc.time_above(conc=5, time=5, conc_above=5, method="linear"),
    structure(NA_real_, exclude="Too few measured concentrations to assess time_above")
  )
})

test_that("time_above linear", {
  expect_equal(
    pk.calc.time_above(conc=c(5, 5), time=1:2, conc_above=5, method="linear"),
    1,
    info="All above, single period"
  )
  expect_equal(
    pk.calc.time_above(conc=c(5, 6, 5), time=1:3, conc_above=5, method="linear"),
    2,
    info="All above, multiple periods"
  )
  expect_equal(
    pk.calc.time_above(conc=c(4, 4), time=1:2, conc_above=5, method="linear"),
    0,
    info="All below, single period"
  )
  expect_equal(
    pk.calc.time_above(conc=c(5, 4, 5), time=1:3, conc_above=5, method="linear"),
    0,
    info="Some equal to conc_above, but all below"
  )
  expect_equal(
    pk.calc.time_above(conc=c(6, 4, 5), time=1:3, conc_above=5, method="linear"),
    0.5
  )
  expect_equal(
    pk.calc.time_above(conc=c(6, NA, 5), time=1:3, conc_above=5, method="linear"),
    2,
    info="NA is ignored"
  )
})

test_that("time_above with 'lin up/log down'", {
  expect_equal(
    pk.calc.time_above(conc=c(6, 4, 5), time=1:3, conc_above=5, method='lin up/log down'),
    (log(6) - log(5))/(log(6) - log(4))
  )
  expect_equal(
    pk.calc.time_above(conc=c(6, 4, 5), time=(1:3)*2, conc_above=5, method='lin up/log down'),
    (log(6) - log(5))/(log(6) - log(4))*2
  )

  expect_equal(
    pk.calc.time_above(conc=c(6, 4, 6), time=1:3, conc_above=5, method='lin up/log down'),
    (log(6) - log(5))/(log(6) - log(4)) + 0.5
  )
  expect_equal(
    pk.calc.time_above(conc=c(6, 4, 6, 0), time=1:4, conc_above=5, method='lin up/log down'),
    (log(6) - log(5))/(log(6) - log(4)) + 0.5 + 1/6
  )
  expect_equal(
    pk.calc.time_above(conc=c(6, 0, 6, 0), time=1:4, conc_above=5, method='lin up/log down'),
    1/6 + 1/6 + 1/6
  )
})

#' Choose intervals to compute AUCs from time and dosing information
#'
#' Intervals for AUC are selected by the following metrics:
#' \enumerate{
#'   \item If only one dose is administered, use the
#'         `PKNCA.options("single.dose.aucs")`
#'   \item If more than one dose is administered, estimate the AUC
#'         between any two doses that have PK taken at both of the
#'         dosing times and at least one time  between the doses.
#'   \item For the final dose of multiple doses, try to determine the
#'         dosing interval (\eqn{\tau}) and estimate the AUC in that
#'         interval if multiple samples are taken in the interval.
#'   \item If there are samples \eqn{> \tau} after the last dose,
#'         calculate the half life after the last dose.
#'  }
#'
#' @inheritParams PKNCA.choose.option
#' @param time.conc Time of concentration measurement
#' @param time.dosing Time of dosing
#' @param single.dose.aucs The AUC specification for single dosing.
#' @returns A data frame with columns for `start`, `end`, `auc.type`, and
#'   `half.life`.  See [check.interval.specification()] for column definitions.
#' The data frame may have zero rows if no intervals could be found.
#' @family Interval specifications
#' @family Interval determination
#' @seealso [pk.calc.auc()], [pk.calc.aumc()], [pk.calc.half.life()],
#'   [PKNCA.options()]
#' @export
choose.auc.intervals <- function(time.conc, time.dosing,
                                 options=list(),
                                 single.dose.aucs=NULL) {
  # Check inputs
  single.dose.aucs <- PKNCA.choose.option(name="single.dose.aucs", value=single.dose.aucs, options=options)
  if (any(is.na(time.conc)))
    stop("time.conc may not have any NA values")
  if (any(is.na(time.dosing)))
    stop("time.dosing may not have any NA values")
  if (length(unique(time.dosing)) == 1) {
    # If it is single-dose data, use the time of dosing and then offset it by
    # the dosing time (allowing the case where dosing time is not 0).
    ret <- check.interval.specification(single.dose.aucs)
    # If there is an offset from 0, use that offset
    ret$start <- ret$start + unique(time.dosing)
    ret$end <- ret$end + unique(time.dosing)
  } else {
    # Sort the times so sorting can be assumed farther down in the algorithm
    time.dosing <- sort(time.dosing)
    time.conc <- sort(time.conc)
    # Find the doses that have concentration measurements
    mask_dose_conc <- time.dosing %in% time.conc
    # Find indexes of pairs of doses that both have predose PK associated
    idx.paired.dose <-
      seq_len(length(time.dosing)-1)[
        mask_dose_conc[-1] &
          mask_dose_conc[-length(mask_dose_conc)]
        ]
    # A data frame with all the right columns and classes but no data
    ret <- check.interval.specification(data.frame(start=0, end=1, auclast=TRUE))[-1,]
    # Find the pairs that have at least one measurement between them
    for (n in idx.paired.dose) {
      if (any(time.dosing[n] < time.conc &
              time.conc < time.dosing[n+1])) {
        # If there are measurements between the doses, add it to the output.
        ret <-
          rbind(
            ret,
            check.interval.specification(
              data.frame(
                start=time.dosing[n],
                end=time.dosing[n+1],
                auclast=TRUE,
                cmax=TRUE,
                tmax=TRUE
              )
            )
          )
      }
    }
    # Find the repeating dosing interval if possible and add it to
    # the last dose if there is PK beyond the last dose to that time.
    tau <- find.tau(time.dosing)
    if (!is.na(tau)) {
      if ((max(time.dosing) + tau) %in% time.conc) {
        ret <-
          rbind(
            ret,
            check.interval.specification(
              data.frame(
                start=max(time.dosing),
                end=max(time.dosing) + tau,
                cmax=TRUE,
                tmax=TRUE,
                auclast=TRUE,
                stringsAsFactors=FALSE
              )
            )
          )
      }
      # If the maximum concentration measurement time is beyond the
      # max dosing time + tau, calculate a half-life.
      if ((max(time.dosing) + tau) < max(time.conc)) {
        ret <-
          rbind(
            ret,
            check.interval.specification(
              data.frame(
                start=max(time.dosing),
                end=Inf,
                half.life=TRUE
              )
            )
          )
      }
    }
  }
  ret
}

#' Find the repeating interval within a vector of doses
#'
#' This is intended to find the interval over which x repeats by the rule
#' `unique(mod(x, interval))` is minimized.
#'
#' @inheritParams PKNCA.choose.option
#' @param x the vector to find the interval within
#' @param na.action What to do with NAs in `x`
#' @param tau.choices the intervals to look for if the doses are not all equally
#'   spaced.
#' @returns A scalar indicating the repeating interval with the most repetition.
#' \enumerate{
#'   \item If all values are `NA` then NA is returned.
#'   \item If all values are the same, then 0 is returned.
#'   \item If all values are equally spaced, then that spacing is
#'         returned.
#'   \item If one of the `choices` can minimize the number of
#'         unique values, then that is returned.
#'   \item If none of the `choices` can minimize the number of
#'         unique values, then -1 is returned.
#' }
#' @family Interval determination
#' @export
find.tau <- function(x, na.action=stats::na.omit,
                     options=list(),
                     tau.choices=NULL) {
  # Check inputs
  tau.choices <- PKNCA.choose.option(name="tau.choices", value=tau.choices, options=options)
  ret <- NA
  x <- na.action(x)
  if (length(unique(x)) == 1) {
    # Single dose, no more effort needed
    ret <- 0
  } else if (identical(tau.choices, NA)) {
    all_deltas <-
      sort(unique(
        as.vector(vapply(
          X = x,
          FUN = function(x, y) x - y,
          y = x,
          FUN.VALUE = x
        ))
      ))
    tau.choices <- all_deltas[all_deltas > 0]
  }
  if (is.na(ret) &
      length(x) > 1) {
    delta_1 <- x[2] - x[1]
    if (all((x[-1] - x[-length(x)]) == delta_1)) {
      # Only one interval through the full data set
      ret <- delta_1
    } else {
      # Drop any tau.choices that are >= the difference in the range of x
      # because those are uninformative (i.e. if the maximum time is 12 hours,
      # don't test an interval of 12, 24, or ... hours because they will match
      # the x - tau < 0 test in a meaningless way.
      tau.choices <- tau.choices[tau.choices < (max(x) - min(x))]
      # Ensure that the choices are in order so that we find the minimum
      # interval.
      tau.choices <- sort(tau.choices)
      # Test all the tau.choices until we find the first (and thereby
      # smallest) usable one
      i <- 0
      while (is.na(ret) & i < length(tau.choices)) {
        i <- i+1
        tau <- tau.choices[i]
        # Is the dose either within the first tau or there is a dose
        # that far before it?
        dose_before <- ((x - tau < 0) | ((x - tau) %in% x))
        # And, is the dose either within the last tau or there is a dose that
        # far after it?
        dose_after <- ((x + tau > max(x)) | ((x + tau) %in% x))
        if (all(dose_before & dose_after)) {
          ret <- tau
        }
      }
    }
  }
  ret
}

test_that("find.tau", {
  # Regularly spaced intervals give the regular spacing
  expect_equal(
    find.tau(sort(unique(c(seq(0, 168, 12),
                           seq(4, 168, 12)))),
             tau.choices=NA),
    12)
  expect_equal(
    find.tau(sort(unique(c(seq(0, 168, 12),
                           seq(4, 168+12, 12)))),
             tau.choices=NA),
    12)
  expect_equal(find.tau(0:10, tau.choices=NA), 1)
  # It overrides tau.choices if everything is equally spaced.
  expect_equal(find.tau(0:10, tau.choices=c(24, 168)), 1)
  expect_equal(find.tau(seq(0, 100, by=10),
                        tau.choices=c(24, 168)), 10)
  expect_equal(find.tau(seq(0, 48, by=24),
                        tau.choices=c(24, 168)), 24)
  # Alternatively spaced intervals give the alternative spacing
  expect_equal(find.tau(c(seq(0, 48, by=24),
                          seq(10, 48, by=24)),
                        tau.choices=c(24, 168)),
               24)
  # Smaller interval spacing adheres to tau.choices with unequal
  # spacing.
  expect_equal(find.tau(c(seq(0, 48, by=12),
                          seq(10, 48, by=12)),
                        tau.choices=c(24, 168)),
               24)
  # It works with more complex spacing and many intervals
  expect_equal(find.tau(
    sort(unique(c(seq(0, 168, by=24),
                  seq(4, 168, by=24),
                  seq(10, 168, by=24)))),
    tau.choices=c(24, 168)),
    24)
  expect_equal(find.tau(c(0, 5, 19, 30) + rep((0:5)*168, each=4),
                        tau.choices=c(24, 168)),
               168)
  # If there is only one dosing time, return 0-- regardless of the
  # option.
  expect_equal(find.tau(rep(5, 5), tau.choices=c(24, 168)), 0)
  expect_equal(find.tau(rep(0, 5), tau.choices=c(24, 168)), 0)
  # If everything is NA, return NA
  expect_equal(find.tau(NA,
                        tau.choices=c(24, 168)),
               NA)
  expect_equal(find.tau(rep(NA, 10),
                        tau.choices=c(24, 168)),
               NA)
  # If there is no sequence, return NA
  expect_equal(find.tau(c(0, 1, 3, 5, 9),
                        tau.choices=c(24, 168)),
               NA)
  expect_equal(find.tau(c(0, 1, 3, 5, 9, 24),
                        tau.choices=NA),
               NA)
})

test_that("choose.auc.intervals", {
  tmp.single.dose.auc <-
    check.interval.specification(
      data.frame(start=0,
                 end=c(24, Inf),
                 auclast=c(TRUE, FALSE),
                 aucinf=c(FALSE, TRUE),
                 half.life=c(FALSE, TRUE),
                 stringsAsFactors=FALSE))

  # Check the inputs
  expect_error(choose.auc.intervals(NA, 1, tmp.single.dose.auc),
               regexp="time.conc may not have any NA values")
  expect_error(choose.auc.intervals(1, NA, tmp.single.dose.auc),
               regexp="time.dosing may not have any NA values")
  # The below test is a bit of a non-sequeter-- essentially, it just
  # needs to return a 0-row data frame.
  expect_error(choose.auc.intervals(1, 1,
                                    single.dose.aucs=data.frame()),
               regexp="interval specification has no rows")
  # It adjusts single dose AUCs by the starting time
  expect_equal(choose.auc.intervals(1, 1,
                                    single.dose.aucs=tmp.single.dose.auc),
               check.interval.specification(
                 data.frame(start=1,
                            end=c(25, Inf),
                            auclast=c(TRUE, FALSE),
                            aucinf=c(FALSE, TRUE),
                            half.life=c(FALSE, TRUE))))

  # Find intervals for two doses with PK at both points and one in
  # between.
  expect_equal(choose.auc.intervals(c(1, 2, 3), c(1, 3),
                                    single.dose.aucs=tmp.single.dose.auc),
               check.interval.specification(
                 data.frame(start=1,
                            end=3,
                            cmax=TRUE,
                            tmax=TRUE,
                            auclast=TRUE)))
  # Find intervals for two doses with PK at both points, one in
  # between, and one after asking for AUClast after the second dose
  # but no half-life.
  expect_equal(choose.auc.intervals(1:5, c(1, 3),
                                    single.dose.aucs=tmp.single.dose.auc),
               check.interval.specification(
                 data.frame(start=c(1, 3),
                            end=c(3, 5),
                            cmax=TRUE,
                            tmax=TRUE,
                            auclast=TRUE)))
  # Find intervals for two doses with PK at both points, one in
  # between, and one after asking for AUClast after the second dose
  # with half-life.
  expect_equal(choose.auc.intervals(1:6, c(1, 3),
                                    single.dose.aucs=tmp.single.dose.auc),
               check.interval.specification(
                 data.frame(start=c(1, 3, 3),
                            end=c(3, 5, Inf),
                            auclast=c(TRUE, TRUE, FALSE),
                            cmax=c(TRUE, TRUE, FALSE),
                            tmax=c(TRUE, TRUE, FALSE),
                            half.life=c(FALSE, FALSE, TRUE))))
  # Some doses have PK betwen them, some not.
  expect_equal(choose.auc.intervals(1:6, c(1, 3, 5, 7, 9),
                                    single.dose.aucs=tmp.single.dose.auc),
               check.interval.specification(
                 data.frame(start=c(1, 3),
                            end=c(3, 5),
                            auclast=TRUE,
                            cmax=TRUE,
                            tmax=TRUE)))
  # Find intervals when some doses do not have AUCs between them
  # (pairs of doses with trough but no PK between)  
  expect_equal(choose.auc.intervals(c(1, 2, 3, 5, 6, 7),
                                    c(1, 3, 5, 7, 9),
                                    single.dose.aucs=tmp.single.dose.auc),
               check.interval.specification(
                 data.frame(start=c(1, 5),
                            end=c(3, 7),
                            cmax=TRUE,
                            tmax=TRUE,
                            auclast=TRUE)))
  # Find intervals for two doses with PK at both points, one in
  # between, and one after asking for AUClast after the second dose
  # with half-life.  Since tau is not detectable, no half-life at the
  # end.
  expect_equal(choose.auc.intervals(1:6, c(1, 3, 5, 9),
                                    single.dose.aucs=tmp.single.dose.auc),
               check.interval.specification(
                 data.frame(start=c(1, 3),
                            end=c(3, 5),
                            auclast=TRUE,
                            cmax=TRUE,
                            tmax=TRUE)))
})

#' Check the formatting of a calculation interval specification data frame.
#'
#' Calculation interval specifications are data frames defining what
#' calculations will be required and summarized from all time intervals. Note:
#' parameters which are not requested may be calculated if it is required for
#' (or computed at the same time as) a requested parameter.
#'
#' `start` and `end` time must always be given as columns, and the `start` must
#' be before the `end`.  Other columns define the parameters to be calculated
#' and the groupings to apply the intervals to.
#'
#' @param x The data frame specifying what to calculate during each time
#'   interval
#' @returns x The potentially updated data frame with the interval calculation
#'   specification.
#'
#' @family Interval specifications
#' @seealso The vignette "Selection of Calculation Intervals"
#' @export
check.interval.specification <- function(x) {
  if (!is.data.frame(x)) {
    # Just a warning and let as.data.frame make it an error if it can't be
    # coerced.
    warning("Interval specification must be a data.frame")
    x <- as.data.frame(x, stringsAsFactors=FALSE)
  }
  if (nrow(x) == 0) {
    stop("interval specification has no rows")
  }
  # Confirm that the minimal columns (start and end) exist
  if (length(missing.required.cols <- setdiff(c("start", "end"), names(x))) > 0) {
    stop(sprintf("Column(s) %s missing from interval specification",
                 paste0("'", missing.required.cols, "'",
                        collapse=", ")))
  }
  interval_cols <- get.interval.cols()
  # Check the edit of each column
  for (n in names(interval_cols)) {
    if (!(n %in% names(x))) {
      if (is.vector(interval_cols[[n]]$values)) {
        # Set missing columns to the default value
        x[[n]] <- interval_cols[[n]]$values[1]
      } else {
        # It would probably take malicious code to get here (altering
        # the intervals without using add.interval.col
        stop("Cannot assign default value for interval column", n) # nocov
      }
    } else {
      # Confirm the edits of the given columns
      if (is.vector(interval_cols[[n]]$values)) {
        if (!all(x[[n]] %in% interval_cols[[n]]$values))
          stop(sprintf("Invalid value(s) in column %s:", n),
               paste(unique(setdiff(x[[n]], interval_cols[[n]]$values)),
                     collapse=", "))
      } else if (is.function(interval_cols[[n]]$values)) {
        if (is.factor(x[[n]])) {
          stop(sprintf("Interval column '%s' should not be a factor", n))
        }
        interval_cols[[n]]$values(x[[n]])
      } else {
        stop("Invalid 'values' for column specification ", n, " (please report this as a bug).") # nocov
      }
    }
  }
  # Now check specific columns
  # start and end
  if (any(x$start %in% NA)) {
    stop("Interval specification may not have NA for the starting time")
  }
  if (any(x$end %in% NA)) {
    stop("Interval specification may not have NA for the end time")
  }
  if (any(is.infinite(x$start))) {
    stop("start may not be infinite")
  }
  if (any(x$start >= x$end)) {
    stop("start must be < end")
  }
  # Confirm that something is being calculated for each interval (and warn if
  # not)
  mask_calculated <- rep(FALSE, nrow(x))
  for (n in setdiff(names(interval_cols), c("start", "end"))) {
    mask_calculated <-
      (mask_calculated |
       !(x[[n]] %in% c(NA, FALSE)))
  }
  if (any(!mask_calculated)) {
    warning("Nothing to be calculated in interval specification number(s): ",
            paste(seq_len(nrow(x))[!mask_calculated], collapse=", "))
  }
  # Put the columns in the right order and return the checked data frame
  x[,
    c(names(interval_cols), setdiff(names(x), names(interval_cols))),
    drop=FALSE
    ]
}

# Helper function to get.parameter.deps to determine the function map
get.parameter.deps_helper_funmap <- function(x, all_intervals) {
  if (is.na(x$FUN) &
      is.null(x$depends)) {
    # For columnns like "start" and "end"
    retfun <- NA
  } else if (is.na(x$FUN)) {
    if (length(x$depends) == 1) {
      # When the value is calculated by the same function as
      # another parameter.
      retfun <- all_intervals[[x$depends]]$FUN
    } else {
      # It would probably take malicious code to get here (an
      # example of malicious code could be altering the
      # intervals without using add.interval.col)
      stop("Invalid interval definition with no function and multiple dependencies.") # nocov
    }
  } else {
    retfun <- x$FUN
  }
  # Define a function call by its function name and the
  # changes to the formal arguments made.
  append(list(retfun), x$formalsmap)
}
# Helper function to get.parameter.deps to find all parameters that are defined
# by the same function
get.parameter.deps_helper_samefun <- function(n, funmap) {
  mask.funmap <- rep(FALSE, length(funmap))
  for (current in n) {
    for (i in seq_len(length(funmap))) {
      mask.funmap[i] <-
        mask.funmap[i] |
        !any(
          is.na(funmap[[current]][[1]]),
          is.na(funmap[[i]][[1]])
        ) &
        identical(funmap[[current]], funmap[[i]])
    }
  }
  names(funmap)[mask.funmap]
}
# Helper function to get.parameter.deps to search all dependencies
get.parameter.deps_helper_searchdeps <- function(current, funmap, all_intervals) {
  # Find any parameters using the same function
  start <- get.parameter.deps_helper_samefun(current, funmap)
  # Find any parameters that depend on the current parameter
  ret <-
    vapply(
      X = all_intervals,
      FUN = function(x) {
        any(x$depends %in% start)
      },
      FUN.VALUE = TRUE
    )
  # Extract their names
  added <- setdiff(names(ret)[ret], start)
  if (length(added) > 0) {
    # Find any parameters that depend on any of those parameters
    unique(
      c(start, added,
        get.parameter.deps_helper_searchdeps(added, funmap, all_intervals))
    )
  } else {
    c(start, added)
  }
}

#' Get all columns that depend on a parameter
#'
#' @param x The parameter name (as a character string)
#' @returns A character vector of parameter names that depend on the parameter
#'   `x`.  If none depend on `x`, then the result will be an empty vector.
#' @family Interval specifications
#' @export
get.parameter.deps <- function(x) {
  all_intervals <- get.interval.cols()
  if (!(x %in% names(all_intervals))) {
    stop("`x` must be the name of an NCA parameter listed by the function `get.interval.cols()`")
  }
  funmap <-
    lapply(
      X=all_intervals,
      FUN=get.parameter.deps_helper_funmap,
      all_intervals=all_intervals
    )
  sort(get.parameter.deps_helper_searchdeps(x, funmap, all_intervals))
}

#' Take in a single row of an interval specification and return that row updated
#' with any additional calculations that must be done to fulfill all
#' dependencies.
#'
#' @param x A data frame with one or more rows of the PKNCA interval
#' @returns The interval specification with additional calculations added where
#'   requested outputs require them.
#' @family Interval specifications
check.interval.deps <- function(x) {
  # Ensure that the input is a valid interval specification
  ret <- check.interval.specification(x)
  colspec <- get.interval.cols()
  for (n in names(colspec)) {
    if (is.logical(ret[,n])) {
      # This is a calculation to complete, otherwise it's something informative
      # but not caluclated.
      mask_calculated <- ret[,n]
      for (deps in colspec[[n]]$depends)
        ret[,deps] <- mask_calculated | ret[,deps]
    }
  }
  ret
}

test_that("check.interval.specification", {

  # Get the current name order of the expected results
  nameorder <- names(check.interval.deps(data.frame(start=0, end=1, cmax=TRUE)))

  d1 <- data.frame(start=0, end=1)
  r1 <- data.frame(start=0, end=1)
  r1[,setdiff(nameorder, names(r1))] <- FALSE
  expect_warning(expect_warning(
    check.interval.specification(as.matrix(d1)),
    regexp="Interval specification must be a data.frame",
    info="Interval must be a data.frame or coercable into a data frame"),
    regexp="Nothing to be calculated in interval specification number\\(s\\): 1"
  )
  expect_warning(
    d1.check <- check.interval.specification(d1),
    regexp="Nothing to be calculated in interval specification number\\(s\\): 1",
    info="Warn if nothing is to be calculated in an interval specification"
  )
  expect_equal(
    d1.check, r1[,nameorder],
    info="Expand a minimal data frame for interval specification"
  )

  # Giving one parameter will fill in everything else as false
  d2 <- data.frame(start=0, end=1, auclast=TRUE)
  r2 <- data.frame(start=0, end=1, auclast=TRUE)
  r2[,setdiff(nameorder, names(r2))] <- FALSE
  expect_equal(
    check.interval.specification(d2),
    r2[,nameorder],
    info="Expand a data frame interval specification with only one request given"
  )

  # start and end must both be specified
  d3 <- data.frame(start=0)
  expect_error(
    check.interval.specification(d3),
    regexp="Column\\(s\\) 'end' missing from interval specification",
    info="Confirm end column is in interval specification"
  )
  d4 <- data.frame(end=1)
  expect_error(
    check.interval.specification(d4),
    regexp="Column\\(s\\) 'start' missing from interval specification",
    info="Confirm start column is in interval specification"
  )
  d5 <- data.frame(blah=5)
  expect_error(
    check.interval.specification(d5),
    regexp="Column\\(s\\) 'start', 'end' missing from interval specification",
    info="Confirm start and end columns are in interval specification"
  )

  # Ensure that there are data
  d6 <- data.frame()
  expect_error(
    check.interval.specification(d6),
    regexp="interval specification has no rows",
    info="It is an error to have an interval specification with no rows"
  )

  # Confirm specific column values required
  d7 <- data.frame(start=as.numeric(NA), end=1)
  expect_error(
    check.interval.specification(d7),
    regexp="Interval specification may not have NA for the starting time",
    info="Interval specification may not have NA for the starting time"
  )
  d8 <- data.frame(start=0, end=as.numeric(NA))
  expect_error(
    check.interval.specification(d8),
    regexp="Interval specification may not have NA for the end time",
    info="Interval specification may not have NA for the end time"
  )
  d9 <- data.frame(start=1, end=1)
  expect_error(
    check.interval.specification(d9),
    regexp="start must be < end",
    info="In interval specification, start must be < end (they are equal)."
  )
  d10 <- data.frame(start=1, end=0)
  expect_error(
    check.interval.specification(d10),
    regexp="start must be < end",
    info="In interval specification, start must be < end (end is less)."
  )
  d11 <- data.frame(start=Inf, end=1)
  expect_error(
    check.interval.specification(d11),
    regexp="start may not be infinite",
    info="In interval specification, start may not be infinite (positive infinity)."
  )

  d12 <- data.frame(start=-Inf, end=1)
  expect_error(
    check.interval.specification(d12),
    regexp="start may not be infinite",
    info="In interval specification, start may not be infinite (negative infinity)."
  )

  # But it is OK to have an infinite end
  d13 <- data.frame(start=0, end=Inf)
  r13 <- data.frame(start=0, end=Inf)
  r13[,setdiff(nameorder, names(r13))] <- FALSE
  expect_warning(d13.check <- check.interval.specification(d13))
  expect_equal(
    d13.check, r13[,nameorder],
    info="In interval specification, end may be infinite (positive infinity)."
  )
  expect_error(
    check.interval.specification(data.frame(start=0, end=-Inf)),
    info="In interval specification, end may not be negative infinity (start is 0)."
  )
  expect_error(
    check.interval.specification(data.frame(start=-Inf, end=-Inf)),
    info="In interval specification, end may not be negative infinity (start is -Inf)."
  )

  # When the no-calculation interval specification is not the first,
  # ensure that is warned correctly
  d14 <- data.frame(start=0, end=24, auclast=c(rep(FALSE, 3), TRUE))
  expect_warning(
    check.interval.specification(d14),
    regexp="Nothing to be calculated in interval specification number\\(s\\): 1, 2, 3",
    info="Warn when nothing is to be calculated in all rows of the specification."
  )

  d14 <- data.frame(start=0, end=24, auclast=c(rep(TRUE, 3), FALSE))
  expect_warning(
    check.interval.specification(d14),
    regexp="Nothing to be calculated in interval specification number\\(s\\): 4",
    info="Warn when nothing is to be calculated in one but not all rows of the specification."
  )

  # Other information is passed through untouched after all the
  # calculation columns
  d15 <- data.frame(start=0, end=Inf, treatment="foo")
  r15 <- data.frame(start=0, end=Inf, treatment="foo")
  r15[,setdiff(nameorder, names(r15))] <- FALSE
  expect_warning(v15 <- check.interval.specification(d15))
  expect_equal(
    v15, r15[,c(nameorder, "treatment")],
    info="Extra information is maintained in the interval specification."
  )

  d16 <- data.frame(start=factor(0), end=1)
  expect_error(
    check.interval.specification(d16),
    regexp="Interval column 'start' should not be a factor",
    info="Start must be numeric and not a factor."
  )

  d17 <- data.frame(start=0, end=factor(1))
  expect_error(
    check.interval.specification(d17),
    regexp="Interval column 'end' should not be a factor",
    info="End must be numeric and not a factor."
  )
})

test_that("check.interval.deps", {
  # Get the current name order of the expected results
  nameorder <- names(check.interval.deps(data.frame(start=0, end=1, cmax=TRUE)))

  r1 <- data.frame(
    start = 0,
    end = 24,
    lambda.z = TRUE,
    clast.obs = TRUE,
    aucinf.obs = TRUE
  )
  r1[,setdiff(nameorder, names(r1))] <- FALSE
  expect_equal(
    check.interval.deps(data.frame(start=0, end=24, aucinf.obs=TRUE)),
    r1[,nameorder],
    info="Confirm that the interval dependencies are accurately added"
  )
})

test_that("get.parameter.deps", {
  expect_error(
    get.parameter.deps("foo"),
    regexp="`x` must be the name of an NCA parameter listed by the function `get.interval.cols()`",
    info="The argument must be a parameter.",
    fixed=TRUE
  )
  expect_equal(
    get.parameter.deps("kel.obs"),
    "kel.obs",
    info="Parameters that have nothing that depend on them return themselves only."
  )
  expect_equal(
    get.parameter.deps("ctrough"),
    c("aucabove.trough.all", "ctrough", "ctrough.dn", "ptr"),
    info="Parameters with formalsmap-related dependencies return themselves and the formalsmap-related dependencies."
  )
  expect_equal(
    get.parameter.deps("start"),
    character(0),
    info="Special columns that are not actually parameters have no dependencies (including themselves)."
  )
  expect_equal(
    get.parameter.deps("cl.obs"),
    c("cl.obs", "vss.iv.obs", "vss.obs", "vz.obs"),
    info="Parameters with dependencies return them."
  )
})

test_that("check.intervals requires a valid value", {
  expect_error(
    check.interval.specification(data.frame(start=0, end=1, cmax="A")),
    regexp="Invalid value(s) in column cmax:A",
    fixed=TRUE
  )
})

test_that("check.intervals works with tibble input (fix #141)", {
  d_dat <-
    data.frame(
      conc = c(
        0.5, 2, 5, 9.2, 12, 2, 1.85, 1.08, 0.5, 0.3, 2.4, 4.5, 10.2, 15, 2.6, 1.65, 1.1,
        0.5, 2, 5, 9.2, 12, 2, 1.85, 1.08, NA, 0.3, 2.4, 4.5, 10.2, 15, 2.6, 1.65, 1.1
      ),
      time = c(seq(264.2, 312.2, 3), seq(264, 312, 3)),
      ARM = rep(c(rep(1, 8), rep(2, 9)), 2),
      SUBJ = c(rep(1, 17), rep(2, 17)),
      Dose = c(rep(5, 17)), rep(5, 17)
    )

  intervals_manual_first <-
    d_dat %>%
    dplyr::group_by(SUBJ) %>%
    dplyr::summarize(
      start=time[dplyr::between(time, 264, 265)],
      end=time[dplyr::between(time, 288, 289)]
    )
  intervals_manual_second <-
    d_dat %>%
    dplyr::group_by(SUBJ) %>%
    dplyr::summarize(
      start=time[dplyr::between(time, 288, 289)],
      end=time[dplyr::between(time, 312, 313)]
    )
  intervals_manual <-
    dplyr::bind_rows(
      intervals_manual_first,
      intervals_manual_second
    ) %>%
    dplyr::mutate(
      auclast=TRUE,
      aucall=TRUE,
      tlast=TRUE
    )
  # There is some other issue here where intervals are having an issue being a tibble
  expect_equal(
    check.interval.specification(intervals_manual)$start,
    intervals_manual$start
  )
})

#' Add a hash and associated information to enable checking object provenance.
#'
#' @param object The object to add provenance
#' @param replace Replace provenance if the object already has a provenance
#'   attribute.  (If the object already has provenance and `replace` is `FALSE`,
#'   then an error will be raised.)
#' @returns The object with provenance as an added item
#' @seealso [checkProvenance()]
#' @export
addProvenance <- function(object, replace=FALSE) {
  if (replace) {
    attr(object, "provenance") <- NULL
  }
  if (is.null(attr(object, "provenance", exact=TRUE))) {
    # Get most of the provenance added
    tmp.prov <- list(
      sessionInfo=utils::sessionInfo(),
      datetime=Sys.time(),
      sysInfo=Sys.info(),
      hash=NA)
    class(tmp.prov) <- c("provenance", class(tmp.prov))
    attr(object, "provenance") <- tmp.prov
    attr(object, "provenance")$hash <-
      digest::digest(as.character(object), serialize=FALSE)
  } else {
    stop("object already has provenance and the option to replace it was not selected.")
  }
  object
}

#' Check the hash of an object to confirm its provenance.
#'
#' @param object The object to check provenance for
#' @returns `TRUE` if the provenance is confirmed to be consistent, `FALSE` if
#'   the provenance is not consistent, or `NA` if provenance is not present.
#' @seealso [addProvenance()]
#' @export
checkProvenance <- function(object) {
  tmp.prov <- attr(object, "provenance", exact=TRUE)
  if (is.null(tmp.prov)) {
    NA
  } else {
    hash <- tmp.prov$hash
    attr(object, "provenance")$hash <- NA
    (hash == digest::digest(as.character(object), serialize=FALSE))
  }
}

#' Print the summary of a provenance object
#'
#' @param x The object to be printed
#' @param ... Ignored
#' @returns invisible text of the printed information
#' @export
print.provenance <- function(x, ...) {
  ret <- sprintf("Provenance hash %s generated on %s with %s.",
                 x$hash,
                 x$datetime,
                 x$sessionInfo$R.version[["version.string"]])
  cat(ret, "\n", sep="")
  invisible(ret)
}

test_that("provenance", {
  a <- addProvenance("a")
  # It may take a bit of time between setting the value in addProvenance
  # and checking
  expect_true(as.numeric(difftime(Sys.time(), attr(a, "provenance", exact=TRUE)$datetime, units="secs")) < 0.1,
              info="A correct time is set in provenance")
  expect_equal(attr(a, "provenance", exact=TRUE)$sysInfo, Sys.info(),
               info="Correct system information is set in provenance")
  expect_equal(attr(a, "provenance", exact=TRUE)$sessionInfo, sessionInfo(),
               info="Correct session information is set in provenance")
  expect_error(addProvenance(a),
               regexp="object already has provenance and the option to replace it was not selected.",
               info="Adding provenance to an object that already has provenance is an error")
  expect_true(
    {
      # Sleep so that there is a difference and it confirms replacement
      Sys.sleep(0.2)
      as.numeric(difftime(Sys.time(),
                          attr(addProvenance(a, replace=TRUE), "provenance", exact=TRUE)$datetime, units="secs")) < 0.1
    },
    info="A correct time is reset in provenance")
  
  expect_true(checkProvenance(a),
              info="Provenance checking works to find correct provenance")
  b <- "b"
  expect_true(is.na(checkProvenance(b)),
              info="Provenance checking works to find missing provenance")
  attr(b, "provenance") <- attr(a, "provenance")
  expect_false(checkProvenance(b),
               info="Provenance checking works to find incorrect provenance")
  
  # Test printing with a fake provenance object
  fakeprov <- list(hash="a",
                   datetime="b",
                   sessionInfo=list(R.version=c(version.string="c")))
  expect_output(print.provenance(fakeprov),
                regexp="Provenance hash a generated on b with c.",
                info="Provenance prints correctly")
  expect_output(
    expect_equal(
      print.provenance(fakeprov),
      invisible("Provenance hash a generated on b with c."),
      info="Provenance printing returns an invisible string with the information in it."
    )
  )
})

#' Exclude data points or results from calculations or summarization.
#'
#' @param object The object to exclude data from.
#' @param reason The reason to add as a reason for exclusion.
#' @param mask A logical vector or numeric index of values to exclude (see
#'   details).
#' @param FUN A function to operate on the data (one group at a time) to select
#'   reasons for exclusions (see details).
#' @returns The object with updated information in the exclude column. The
#'   exclude column will contain the `reason` if `mask` or `FUN` indicate.  If a
#'   previous reason for exclusion was given, then subsequent reasons for
#'   exclusion will be added to the first with a semicolon space ("; ")
#'   separator.
#'
#' @details Only one of `mask` or `FUN` may be given.  If `FUN` is given, it
#'   will be called with two arguments:  a data.frame (or similar object) that
#'   consists of a single group of the data and the full object (e.g. the
#'   PKNCAconc object), `FUN(current_group, object)`, and it must return a
#'   logical vector equivalent to `mask` or a character vector with the reason
#'   text given when data should be excluded or `NA_character_` when the data
#'   should be included (for the current exclusion test).
#' @examples
#' myconc <- PKNCAconc(data.frame(subject=1,
#'                                time=0:6,
#'                                conc=c(1, 2, 3, 2, 1, 0.5, 0.25)),
#'                     conc~time|subject)
#' exclude(myconc,
#'         reason="Carryover",
#'         mask=c(TRUE, rep(FALSE, 6)))
#' @export
#' @family Result exclusions
#' @importFrom dplyr "%>%"
exclude <- function(object, reason, mask, FUN)
  UseMethod("exclude")

utils::globalVariables(c("exclude_current_group_XXX", "row_number_XXX", "exclude_current_group_XXX_row_num"))
#' @describeIn exclude The general case for data exclusion
#' @export
exclude.default <- function(object, reason, mask, FUN) {
  dataname <- getDataName(object)
  # Check inputs
  if (missing(mask) & !missing(FUN)) {
    # operate on one group at a time
    groupnames <-
      unique(c(
        names(getGroups(object)),
        intersect(names(object[[dataname]]),
                  c("start", "end"))
      ))
    mask_df <-
      object[[dataname]] %>%
      dplyr::mutate(row_number_XXX=seq_len(dplyr::n())) %>%
      dplyr::grouped_df(groupnames) %>%
      dplyr::mutate(
        exclude_current_group_XXX_row_num=row_number_XXX,
        exclude_current_group_XXX=
          do.call(
            FUN,
            list(
              as.data.frame(., stringsAsFactors=FALSE)[.$row_number_XXX %in% row_number_XXX,,drop=FALSE],
              object
            )
          )
      ) %>%
      dplyr::mutate(
        exclude_lengths_match=length(exclude_current_group_XXX) ==
          length(exclude_current_group_XXX_row_num)
      )
    # Extract the output and ensure that the output order equals the input order
    mask <- mask_df$exclude_current_group_XXX[order(mask_df$exclude_current_group_XXX_row_num)]
    if (is.character(mask)) {
      reason <- mask
      mask <- !is.na(reason)
    }
  } else if (!xor(missing(mask), missing(FUN))) {
    stop("Either mask for FUN must be given (but not both).")
  }
  if (!(length(reason) %in% c(1, nrow(object[[dataname]])))) {
    stop("reason must be a scalar or have the same length as the data.")
  } else if (!is.character(reason)) {
    stop("reason must be a character string.")
  }
  if (!("exclude" %in% names(object$columns))) {
    stop("object must have an exclude column specified.")
  } else if (!(object$columns$exclude %in% names(object[[dataname]]))) {
    stop("exclude column must exist in object[['", dataname, "']].")
  }
  # Make a scalar reason a vector
  if (length(reason) == 1)
    reason <- rep(reason, length(mask))
  # Find the original value of the 'exclude' column.
  orig <- object[[dataname]][[object$columns$exclude]]
  if (length(mask) != length(orig)) {
    stop("mask must match the length of the data.")
  }
  # No current value for exclude
  mask.none <- orig %in% c(NA, "")
  # Replace the empty value with the reason
  mask.one <- mask & mask.none
  # Add the new reason to an existing reason
  mask.multiple <- mask & (!mask.one)
  ret <- orig
  if (any(mask.one)) {
    ret[mask.one] <- reason[mask.one]
  }
  if (any(mask.multiple)) {
    ret[mask.multiple] <- paste(ret[mask.multiple], reason[mask.multiple], sep="; ")
  }
  object[[dataname]][,object$columns$exclude] <- ret
  object
}

#' Set the exclude parameter on an object
#'
#' This function adds the exclude column to an object.  To change the
#' exclude value, use the [exclude()] function.
#'
#' @param object The object to set the exclude column on.
#' @param exclude The column name to set as the exclude value.
#' @param dataname The name of the data.frame within the object to add the
#'   exclude column to.
#' @returns The object with an exclude column and attribute
setExcludeColumn <- function(object, exclude = NULL, dataname = "data") {
  add.exclude <- FALSE
  if (missing(exclude) || is.null(exclude)) {
    # Exclude is not provided.
    if ("exclude" %in% names(object$columns)) {
      # If exclude is already given, then do nothing.
    } else {
      add.exclude <- TRUE
    }
  } else if ("exclude" %in% names(object$columns)) {
    # If exclude is already in the object, then make sure it matches
    # (and do nothing).
    if (!(object$columns$exclude == exclude)) {
      stop("exclude is already set for the object.")
    }
  } else {
    # If exclude is not already in the object and it is given, then add
    # the column.
    add.exclude <- TRUE
  }
  if (add.exclude) {
    if (missing(exclude) | is.null(exclude)) {
      # Generate the column name
      exclude <-
        setdiff(c("exclude", paste0("exclude.", max(names(object[[dataname]])))),
                names(object[[dataname]]))[1]
      object[[dataname]][[exclude]] <- rep(NA_character_, nrow(object[[dataname]]))
    } else if (nrow(object[[dataname]]) == 0) {
      object[[dataname]][[exclude]] <- rep(NA_character_, nrow(object[[dataname]]))
    } else if (!(exclude %in% names(object[[dataname]]))) {
      stop("exclude, if given, must be a column name in the input data.")
    } else {
      if (is.factor(object[[dataname]][[exclude]])) {
        object[[dataname]][[exclude]] <- as.character(object[[dataname]][[exclude]])
      } else if (is.logical(object[[dataname]][[exclude]]) &
                 all(is.na(object[[dataname]][[exclude]]))) {
        object[[dataname]][[exclude]] <- rep(NA_character_, nrow(object[[dataname]]))
      } else if (!is.character(object[[dataname]][[exclude]])) {
        stop("exclude column must be character vector or something convertable to character without loss of information.")
      }
    }
    object$columns$exclude <- exclude
  }
  object
}

#' Normalize the exclude column by setting blanks to NA
#'
#' @param object The object to extract the exclude column from
#' @returns The exclude vector where `NA` indicates not to exclude and anything
#'   else indicates to exclude.
normalize_exclude <- function(object) {
  dataname <- getDataName(object)
  if (is.null(dataname)) {
    ret <- object
  } else {
    ret <- object[[dataname]][[object$columns$exclude]]
  }
  mask_blank <- ret %in% ""
  if (any(mask_blank)) {
    ret[mask_blank] <- NA
  }
  ret
}

test_that("setExcludeColumn", {
  # exclude argument not given
  expect_equal(
    setExcludeColumn(list(data=data.frame(a=1), columns = list(exclude="fake"))),
    list(data=data.frame(a=1), columns = list(exclude="fake")),
    info="setExcludeColumn does nothing if the exclude name is already given."
  )
  expect_equal(
    setExcludeColumn(list(data=data.frame(a=1))),
    list(data=data.frame(a=1, exclude=NA_character_, stringsAsFactors=FALSE),
         columns = list(exclude="exclude")),
    info="setExcludeColumn adds a column named exclude"
  )
  expect_equal(
    setExcludeColumn(list(data=data.frame(a=1, exclude=2))),
    list(data=data.frame(a=1, exclude=2, exclude.exclude=NA_character_, stringsAsFactors=FALSE),
         columns = list(exclude="exclude.exclude")),
    info="setExcludeColumn adds a column named exclude.exclude if 'exclude' is already present"
  )
  expect_equal(
    setExcludeColumn(list(results=data.frame(a=1)), dataname="results"),
    list(results=data.frame(a=1, exclude=NA_character_, stringsAsFactors=FALSE),
         columns = list(exclude="exclude")),
    info="setExcludeColumn works with an alternate dataname"
  )

  # exclude argument given
  expect_equal(
    setExcludeColumn(
      list(data=data.frame(a=1, exclude=2),
           columns = list(exclude="exclude")),
      exclude="exclude"
    ),
    list(
      data=data.frame(a=1, exclude=2),
      columns = list(exclude="exclude")
    ),
    info="setExcludeColumn does nothing if exclude is given and matching"
  )
  expect_error(setExcludeColumn(list(data=data.frame(a=1, exclude=2),
                                     columns = list(exclude="exclude")),
                                exclude="foo"),
               regexp="exclude is already set for the object.",
               info="setExcludeColumn gives an error if exclude is given and not matching")
  expect_error(setExcludeColumn(list(data=data.frame(a=1)),
                                exclude="exclude"),
               regexp="exclude, if given, must be a column name in the input data.",
               info="setExcludeColumn exclude column must be in the data.")
  expect_equal(
    setExcludeColumn(list(data=data.frame(a=1, exclude=factor("a"))),
                     exclude="exclude"),
    list(data=data.frame(a=1, exclude="a", stringsAsFactors=FALSE),
         columns = list(exclude="exclude")),
    info="setExcludeColumn converts factor column to character"
  )
  expect_equal(
    setExcludeColumn(list(data=data.frame(a=1, exclude=NA, stringsAsFactors=FALSE)),
                     exclude="exclude"),
    list(data=data.frame(a=1, exclude=NA_character_, stringsAsFactors=FALSE),
         columns = list(exclude="exclude")),
    info="setExcludeColumn converts logical NA column to character"
  )
  expect_error(setExcludeColumn(list(data=data.frame(a=1, exclude=FALSE, stringsAsFactors=FALSE)),
                                exclude="exclude"),
               regexp="exclude column must be character vector or something convertable to character without loss of information.",
               info="setExcludeColumn gives error on logical non-NA value")
  expect_error(setExcludeColumn(list(data=data.frame(a=1, exclude=5, stringsAsFactors=FALSE)),
                                exclude="exclude"),
               regexp="exclude column must be character vector or something convertable to character without loss of information.",
               info="setExcludeColumn gives error on non-character value")

  # Zero-row data works
  expect_warning(
    expect_equal(
      setExcludeColumn(list(data=data.frame())),
      list(
        data=data.frame(exclude=NA_character_, stringsAsFactors=FALSE)[-1,,drop=FALSE],
        columns = list(exclude="exclude")
      )
    ),
    info="setExcludeColumn works with zero-row data"
  )
  expect_equal(
    setExcludeColumn(list(data=data.frame()), exclude="foo"),
    list(
      data=data.frame(foo=NA_character_, stringsAsFactors=FALSE)[-1,,drop=FALSE],
      columns = list(exclude="foo")
    ),
    info="setExcludeColumn works with zero-row data"
  )
})

test_that("exclude.default", {
  # Check inputs
  my_conc <- generate.conc(nsub=5, ntreat=2, time.points=0:24)
  obj1 <- PKNCAconc(my_conc, formula=conc~time|treatment+ID)

  expect_error(exclude.default(obj1,
                              reason="Just because"),
               regexp="Either mask for FUN must be given \\(but not both\\).",
               info="One of mask and FUN must be given")
  expect_error(exclude.default(obj1,
                              reason="Just because",
                              mask=rep(TRUE, 5),
                              FUN=function(x) rep(TRUE, nrow(x$data))),
               regexp="Either mask for FUN must be given \\(but not both\\).",
               info="Both mask and FUN may not be given")
  obj2 <- obj1
  obj2$columns$exclude <- NULL
  expect_error(exclude.default(obj2,
                              reason="Just because",
                              mask=rep(TRUE, 5)),
               regexp="object must have an exclude column specified.",
               info="exclude column is required.")
  obj3 <- obj1
  obj3$columns$exclude <- "foo"
  expect_error(exclude.default(obj3,
                              reason="Just because",
                              mask=rep(TRUE, 5)),
               regexp="exclude column must exist in object\\[\\['data'\\]\\].",
               info="exclude column must exist in the data")
  expect_error(exclude.default(obj1,
                              reason="Just because",
                              mask=TRUE),
               regexp="mask must match the length of the data.",
               info="mask may not be a scalar")
  expect_error(exclude.default(obj1,
                              reason="Just because",
                              mask=rep(TRUE, 6)),
               regexp="mask must match the length of the data.",
               info="mask must match the length of the data.")
  expect_error(exclude.default(obj1,
                              reason=1:2,
                              FUN=function(x, ...) TRUE),
               regexp="reason must be a scalar or have the same length as the data",
               info="Interpretation of a non-scalar reason is unclear")
  expect_error(exclude.default(obj1,
                              reason=1,
                              FUN=function(x, ...) TRUE),
               regexp="reason must be a character string.",
               info="Interpretation of a non-character reason is unclear")

  # Check operation
  obj4 <- obj1
  obj4$data$exclude <- c(NA_character_, rep("Just because", nrow(obj4$data)-1))

  expect_equal(exclude.default(obj1,
                              reason="Just because",
                              mask=c(FALSE, rep(TRUE, nrow(obj1$data)-1))),
               obj4,
               info="Mask given as a vector works")

  obj5 <- obj1
  obj5$data$exclude <- ifelse(obj5$data$time == 0,
                              NA_character_, "Just because")
  expect_equal(exclude.default(obj1,
                              reason="Just because",
                              FUN=function(x, ...) c(FALSE, rep(TRUE, nrow(x)-1))),
               obj5,
               info="A function returning a vector works")

  obj7 <- obj1
  obj7$data <- obj7$data[rev(seq_len(nrow(obj7$data))),]
  exclude_1 <- function(x, ...) {
    ifelse(x$ID == 1,
           "Drop 1",
           NA_character_)
  }
  expect_equal(exclude.default(obj1,
                               FUN=exclude_1)$exclude,
               rev(
                 exclude.default(obj7,
                                 FUN=exclude_1)$exclude),
               info="Function application is order-invariant")

  expect_equal(exclude.default(obj1,
                               FUN=function(x, ...) c(NA_character_, rep("Just because", nrow(x)-1))),
               obj5,
               info="A function returning a character vector works")

  obj6 <- obj5
  obj6$data$exclude[1:2] <- c("really", "Just because; really")

  expect_equal(
    exclude.default(
      exclude.default(obj1,
                     reason="Just because",
                     FUN=function(x, ...) c(FALSE, rep(TRUE, nrow(x)-1))),
      reason="really",
      mask=c(TRUE, TRUE, rep(FALSE, nrow(obj1$data) - 2))),
    obj6,
    info="Multiple reasons are tracked.")

  # Check exclusion for PKNCAdose class
  my_dose <- generate.dose(my_conc)
  dose_obj <- PKNCAdose(my_dose, dose~time|treatment+ID)
  dose_obj_ex1 <- dose_obj
  dose_obj_ex1$data$exclude[dose_obj_ex1$data$ID == 1] <- "Not 1"
  expect_equal(exclude(dose_obj, reason="Not 1", FUN=function(x, ...) x$ID == 1),
               dose_obj_ex1,
               info="exclude works for PKNCAdose objects (with functions)")

  # Dose exclusion is respected
  data_obj <- PKNCAdata(obj1, dose_obj, intervals=data.frame(start=0, end=Inf, cl.last=TRUE))
  data_obj_ex1 <- PKNCAdata(obj1, dose_obj_ex1, intervals=data.frame(start=0, end=Inf, cl.last=TRUE))
  result_obj <- pk.nca(data_obj)
  result_obj_ex1 <- pk.nca(data_obj_ex1)

  expect_equal(result_obj_ex1$result$PPORRES[result_obj_ex1$result$ID == 1 &
                                               result_obj_ex1$result$PPTESTCD == "cl.last"],
               rep(NA_real_, 2),
               info="exclude of dose is respected")

  # Check exclusion for PKNCAresults class
  result_obj_not_1 <- result_obj
  result_obj_not_1$result$exclude[result_obj_not_1$result$ID == 1] <- "Not 1"
  expect_equal(
    exclude(result_obj, reason="Not 1", FUN=function(x, ...) x$ID == 1),
    result_obj_not_1,
    info="exclude works for PKNCAresults object"
  )

  expect_false(any(summary(result_obj)$cl.last == summary(result_obj_not_1)$cl.last),
               info="summary.PKNCAresults respects exclude")
})

test_that("normalize_exclude makes blanks into NA_character_ (#55)", {
  my_conc <- generate.conc(nsub=5, ntreat=2, time.points=0:24)
  my_conc$exclude <- c("", rep(NA_character_, nrow(my_conc) - 1))
  obj1 <- PKNCAconc(my_conc,
                    formula=conc~time|treatment+ID,
                    exclude="exclude")
  expect_equal(normalize_exclude(obj1),
               rep(NA_character_, nrow(my_conc)),
               info="normalize_exclude makes blanks into NA_character_")
  obj2 <- obj1
  obj2$data$exclude[2] <- "foo"
  expect_equal(normalize_exclude(obj2),
               c(NA_character_, "foo", rep(NA_character_, nrow(my_conc)-2)),
               info="normalize_exclude makes blanks into NA_character_ and leaves non-blank alone.")
  expect_equal(normalize_exclude(1:5), 1:5,
               info="normalize_exclude works with bare vectors (as opposed to PKNCA objects)")
})

test_that("multiple exclusions for the same row provide all the reasons (fix #113)", {
  my_conc <- generate.conc(nsub=5, ntreat=2, time.points=0:24)
  my_conc$exclude <- c("", rep(NA_character_, nrow(my_conc) - 1))
  suppressMessages(
    result_obj <-
      pk.nca(PKNCAdata(
        PKNCAconc(
          my_conc,
          formula=conc~time|treatment+ID,
          exclude="exclude"
        ),
        intervals=data.frame(start=0, end=Inf, cmax=TRUE)
      ))
  )
  result_excl1 <-
    exclude(
      result_obj,
      reason="test1",
      mask=c(TRUE, TRUE, rep(FALSE, nrow(as.data.frame(result_obj)) - 2))
    )
  result_excl2 <-
    exclude(
      result_excl1,
      reason="test2",
      mask=c(TRUE, FALSE, TRUE, rep(FALSE, nrow(as.data.frame(result_obj)) - 3))
    )
  expect_equal(
    as.data.frame(result_excl2)$exclude,
    c("test1; test2", "test1", "test2", rep(NA_character_, 7))
  )
})

#' Compute noncompartmental superposition for repeated dosing
#'
#' @inheritParams assert_conc_time
#' @inheritParams assert_lambdaz
#' @inheritParams PKNCA.choose.option
#' @inheritParams choose_interval_method
#' @param dose.input The dose given to generate the `conc` and `time` inputs. If
#'   missing, output doses will be assumed to be equal to the input dose.
#' @inheritParams assert_dosetau
#' @param dose.times The time of dosing within the dosing interval. The
#'   `min(dose.times)` must be >= 0, and the `max(dose.times)` must be < `tau`.
#'   There may be more than one dose times given as a vector.
#' @param dose.amount The doses given for the output.  Linear proportionality
#'   will be used from the input to output if they are not equal.  The length of
#'   dose.amount must be either 1 or matching the length of `dose.times`.
#' @param n.tau The number of tau dosing intervals to simulate or `Inf` for
#'   steady-state.
#' @param clast.pred To use predicted as opposed to observed Clast, either give
#'   the value for clast.pred here or set it to true (for automatic calculation
#'   from the half-life).
#' @param tlast The time of last observed concentration above the limit of
#'   quantification.  This is calculated if not provided.
#' @param additional.times Times to include in the final outputs in addition to
#'   the standard times (see details).  All `min(additional.times)` must be >=
#'   0, and the `max(additional.times)` must be <= `tau`.
#' @param check.blq Must the first concentration measurement be below the limit
#'   of quantification?
#' @param steady.state.tol The tolerance for assessing if steady-state has been
#'   achieved (between 0 and 1, exclusive).
#' @param ... Additional arguments passed to the `half.life` function if
#'   required to compute `lambda.z`.
#' @returns A data frame with columns named "conc" and "time".
#'
#' @details The returned superposition times will include all of the following
#' times: 0 (zero), `dose.times`, `time modulo tau` (shifting `time` for each
#' dose time as well), `additional.times`, and `tau`.
#'
#' @seealso [interp.extrap.conc()]
#' @export
superposition <- function(conc, ...) {
  UseMethod("superposition", conc)
}

#' @rdname superposition
#' @export
superposition.PKNCAconc <- function(conc, ...) {
  # Split the data by grouping and extract just the concentration and
  # time columns
  nested_data <- prepare_PKNCAconc(conc)
  tmp_results <-
    parallel::mclapply(
      X=seq_len(nrow(nested_data)),
      FUN=function(idx) {
        superposition.numeric(
          conc=nested_data$data_conc[[idx]]$conc,
          time=nested_data$data_conc[[idx]]$time,
          ...
        )
      }
    )
  # Replace the concentration data with the new results
  nested_data$data_conc <- tmp_results
  tidyr::unnest(nested_data, cols="data_conc")
}

#' @rdname superposition
#' @export
superposition.numeric <- function(conc, time, dose.input = NULL,
                                  tau, dose.times=0, dose.amount, n.tau=Inf,
                                  options=list(),
                                  lambda.z, clast.pred=FALSE, tlast,
                                  additional.times=numeric(),
                                  check.blq=TRUE,
                                  method = NULL,
                                  auc.type = "AUCinf",
                                  steady.state.tol=1e-3, ...) {
  # Check the inputs
  # Concentration and time
  assert_conc_time(conc = conc, time = time)
  if (check.blq) {
    if (!(conc[1] %in% 0)) {
      stop("The first concentration must be 0 (and not NA).  To change this set check.blq=FALSE.")
    }
  }
  assert_number_between(dose.input, na.ok = FALSE, null.ok = TRUE, lower = 0)
  assert_dosetau(tau)
  assert_numeric_between(x = dose.times, lower_eq = 0, min.len = 1, upper = tau)
  # dose.amount
  if (!missing(dose.amount)) {
    if (missing(dose.input)) {
      stop("must give dose.input to give dose.amount")
    }
    assert_numeric_between(x = dose.amount, lower = 0, finite = TRUE)
    if (!(length(dose.amount) %in% c(1, length(dose.times))))
      stop("dose.amount must either be a scalar or match the length of dose.times")
  }
  checkmate::assert_number(n.tau, lower = 1)
  if (is.finite(n.tau)) {
    n.tau <- checkmate::assert_integerish(n.tau, lower = 1)
  }
  # lambda.z
  if (!missing(lambda.z)) {
    lambda.z <- assert_number_between(x = lambda.z, lower = 0)
  }
  # clast.pred
  checkmate::assert(
    checkmate::check_number(clast.pred, lower = 0, na.ok = TRUE),
    checkmate::check_logical(clast.pred, any.missing = TRUE, len = 1),
    .var.name = "clast.pred"
  )
  if (is.na(clast.pred)) {
    clast.pred <- FALSE
  } else if (!is.numeric(clast.pred)) {
    checkmate::assert_logical(clast.pred, any.missing = FALSE, len = 1)
  }
  # tlast
  if (!missing(tlast)) {
    checkmate::assert_number(tlast)
  }
  # additional.times
  if (length(additional.times) > 0) {
    if (any(is.na(additional.times))) {
      stop("No additional.times may be NA (to not include any additional.times, enter c() as the function argument)")
    }
    if (!is.numeric(additional.times) | is.factor(additional.times))
      stop("additional.times must be a number")
    if (any(additional.times < 0))
      stop("All additional.times must be nonnegative")
    if (any(additional.times > tau))
      stop("All additional.times must be <= tau")
  }
  # steady.state.tol
  if (length(steady.state.tol) != 1)
    stop("steady.state.tol must be a scalar")
  if (!is.numeric(steady.state.tol) | is.factor(steady.state.tol) | is.na(steady.state.tol))
    stop("steady.state.tol must be a number")
  if (steady.state.tol <= 0 |
      steady.state.tol >= 1)
    stop("steady.state.tol must be between 0 and 1, exclusive.")
  if (steady.state.tol > 0.01)
    warning("steady.state.tol is usually <= 0.01")
  # We get all or none of lambda.z, clast, and tlast
  has.lambda.z <- !missing(lambda.z)
  has.clast.pred <- !is.logical(clast.pred)
  has.tlast <- !missing(tlast)
  if (any(c(has.lambda.z, has.clast.pred, has.tlast)) &
      !all(c(has.lambda.z, has.clast.pred, has.tlast)))
    stop("Either give all or none of the values for these arguments: lambda.z, clast.pred, and tlast")
  # combine dose.input and dose.amount as applicable to scale the
  # outputs.
  if (!missing(dose.amount)) {
    dose.scaling <- dose.amount / dose.input
    if (length(dose.scaling) != length(dose.times)) {
      if (length(dose.scaling) != 1)
        stop("bug in dose.amount, dose.times, and dose.input handling") # nocov
      # it is a scalar and there is more than one dose
      dose.scaling <- rep(dose.scaling, length(dose.times))
    }
  } else {
    dose.scaling <- rep(1, length(dose.times))
  }
  # Determine the output times
  tmp.times <- c(0, tau, dose.times, additional.times)
  # For the output, ensure that we have each of the input times
  # shifted for the dosing times.
  for (d in dose.times)
    tmp.times <- c(tmp.times, (d + time) %% tau)
  ret <- data.frame(conc=0, time=sort(unique(tmp.times)))
  # Check if all the input concentrations are 0, and if so, give that
  # simple answer.
  if (all(conc == 0)) {
    return(ret)
  } else {
    # Check if we will need to extrapolate
    tlast <- pk.calc.tlast(conc, time, check=FALSE)
    if ((tau*n.tau) > tlast) {
      if (missing(lambda.z)) {
        tmp <- pk.calc.half.life(conc, time, options=options,
                                 ..., check=FALSE)
        lambda.z <- tmp$lambda.z
      }
      if (identical(clast.pred, FALSE)) {
        clast <- pk.calc.clast.obs(conc, time, check=FALSE)
      } else if (identical(clast.pred, TRUE)) {
        clast <- tmp$clast.pred
      } else {
        # Use the provided clast.pred
        clast <- clast.pred
      }
    } else {
      # We don't need lambda.z for calculations, but it is simpler to
      # define it.
      lambda.z <- NA
    }
  }
  # cannot continue extrapolating due to missing data (likely due to
  # half-life not calculable)
  if ((n.tau * tau) > tlast & is.na(lambda.z)) {
    ret$conc <- NA
  } else {
    # Do the math! (Finally)
    current.tol <- steady.state.tol + 1
    tau.count <- 0
    # Stop either for reaching steady-state or for reaching the requested number of doses
    while (tau.count < n.tau &
           !is.na(current.tol) &
           current.tol >= steady.state.tol) {
      prev.conc <- ret$conc
      # Perform the dosing for a single dosing interval.
      for (i in seq_along(dose.times)) {
        tmp.time <- ret$time - dose.times[i] + (tau * tau.count)
        # For the first dosing interval, make sure that we do not
        # assign concentrations before the dose is given.
        mask.time <- tmp.time >= 0
        # Update the current concentration (previous concentration +
        # new concentration scaled by the relative dose)
        ret$conc[mask.time] <-
          (ret$conc[mask.time] +
           dose.scaling[i]*
           interp.extrap.conc(
             conc = conc, time = time,
             time.out=tmp.time[mask.time],
             lambda.z=lambda.z, clast=clast,
             options=options, check=FALSE,
             method = method, auc.type = auc.type
           )
          )
      }
      tau.count <- tau.count + 1
      if (any(ret$conc %in% 0)) {
        # prevent division by 0.  Since not all concentrations are 0,
        # all values will eventually be nonzero.
        current.tol <- steady.state.tol + 1
      } else {
        current.tol <- max(1-(prev.conc/ret$conc))
      }
    }
  }
  ret
}

#' Assert that an interval is accurately defined as an interval, and return the
#' interval
#'
#' @param interval Numeric vector of two numbers for the start and end time of
#'   integration
#' @param start The start time of the interval
#' @param end The end time of the interval
#' @returns `interval` (or `c(start, end)`)
#' @keywords Internal
assert_intervaltime_single <- function(interval = NULL, start = NULL, end = NULL) {
  if (is.null(interval) & is.null(start) & is.null(end)) {
    stop("One of `interval` or `start` and `end` must be given")
  }
  if (xor(is.null(start), is.null(end))) {
    stop("Both `start` and `end` or neither must be given")
  }
  if (!is.null(interval)) {
    checkmate::assert_numeric(x = interval, sorted = TRUE, unique = TRUE, any.missing = FALSE, len = 2)
    checkmate::assert_number(x = interval[1], na.ok = FALSE, finite = TRUE)
  }

  if (!is.null(start)) {
    # Check start and end
    checkmate::assert_number(start, na.ok = FALSE, finite = TRUE, null.ok = FALSE)
    checkmate::assert_number(end, na.ok = FALSE, finite = FALSE, lower = start, null.ok = FALSE)

    if (is.null(interval)) {
      interval <- c(start, end)
    } else if (start != interval[1]) {
      stop("`start` must be the same as the first value in the interval if both are given: ", start, "!=", interval[1])
    } else if (end != interval[2]) {
      stop("`end` must be the same as the second value in the interval if both are given: ", end, "!=", interval[2])
    }
  }

  interval
}

#' Verify that concentration measurements are valid
#'
#' @param conc Measured concentrations
#' @param any_missing_conc Are any concentration values allowed to be `NA`?
#' @returns `conc` or give an informative error
#' @rdname assert_conc_time
assert_conc <- function(conc, any_missing_conc = TRUE) {
  if (length(conc) == 0) {
    rlang::warn(
      message = "No concentration data given",
      class = "pknca_conc_none"
    )
  } else {
    checkmate::assert_numeric(conc, finite = TRUE, any.missing = any_missing_conc)
    if (all(is.na(conc))) {
      rlang::warn(
        message = "All concentration data are missing",
        class = "pknca_conc_all_missing"
      )
    } else if (any(!is.na(conc) & as.numeric(conc) < 0)) {
      # as.numeric(conc) is required for compatibility with units
      warning("Negative concentrations found")
    }
  }
  conc
}

#' Verify that time values are valid
#'
#' @param time Time of the measurement of the concentrations
#' @param sorted_time Must the time be unique and monotonically increasing?
#' @returns `time` or give an informative error
#' @rdname assert_conc_time
assert_time <- function(time, sorted_time = TRUE) {
  if (length(time) == 0) {
    rlang::warn(
      message = "No time data given",
      class = "pknca_time_none"
    )
  } else {
    checkmate::assert_numeric(time, any.missing = FALSE, sorted = sorted_time, unique = sorted_time)
  }
  time
}

#' Verify that the concentrations and times are valid
#'
#' If the concentrations or times are invalid, will provide an error.
#' Reasons for being invalid are
#' \itemize{
#'   \item `time` is not a number
#'   \item `conc` is not a number
#'   \item Any `time` value is NA
#'   \item `time` is not monotonically increasing
#'   \item `conc` and `time` are not the same length
#' }
#'
#' Some cases may generate warnings but allow the data to proceed.
#' \itemize{
#'   \item A negative concentration is often but not always an
#'     error; it will generate a warning.
#' }
#'
#' @returns A data.frame with columns named "conc" and "time" or an informative
#'   error
assert_conc_time <- function(conc, time, any_missing_conc = TRUE, sorted_time = TRUE) {
  assert_conc(conc, any_missing_conc = any_missing_conc)
  assert_time(time, sorted_time = sorted_time)
  checkmate::assert_numeric(conc, len = length(time))
  data.frame(conc = conc, time = time)
}

#' Confirm that a value is greater than another value
#'
#' @inheritParams checkmate::assert_numeric
#' @param lower_eq,upper_eq Values where equality is not allowed
#' @param ... Passed to `checkmate::assert_numeric()`
#' @returns `x`
assert_numeric_between <- function(x, any.missing = FALSE, null.ok = FALSE, lower_eq = -Inf, lower = -Inf, upper = Inf, upper_eq = Inf, ..., .var.name = checkmate::vname(x)) {
  checkmate::assert_numeric(x, any.missing = any.missing, null.ok = null.ok, lower = lower_eq, upper = upper_eq, ..., .var.name = .var.name)
  if (is.null(x) & null.ok) {
    # do nothing
  } else {
    # disallowed missing will have been previously caught
    mask_na <- is.na(x)
    mask_lower <- !mask_na & !is.infinite(lower) & x <= lower
    mask_upper <- !mask_na & !is.infinite(upper) & x >= upper
    msg <- NULL
    if (any(mask_lower)) {
      msg <-
        c(
          msg,
          sprintf("Assertion on '%s' failed: %s is not > %g", .var.name, element_find(mask_lower), lower)
        )
    }
    if (any(mask_upper)) {
      msg <-
        c(
          msg,
          sprintf("Assertion on '%s' failed: %s is not < %g", .var.name, element_find(mask_upper), upper)
        )
    }
    if (length(msg) > 0) {
      stop(paste(msg, collapse = "\n"))
    }
  }
  x
}

element_find <- function(x) {
  values <- which(x)
  ret_values <-
    if (length(values) > 6) {
      paste(values[1:5], collapse = ", ")
    } else {
      paste(values, collapse = ", ")
    }
  paste(
    ngettext(length(values), msg1 = "Element", msg2 = "Elements"),
    ret_values
  )
}

#' Confirm that a value is greater than another value
#'
#' @inheritParams checkmate::assert_number
#' @param len Ignored (must be 1)
#' @param ... Passed to `assert_numeric_between()`
#' @returns `x` or an informative error
assert_number_between <- function(x, ..., na.ok = FALSE, len = 1, .var.name = checkmate::vname(x)) {
  assert_numeric_between(x, len = 1, .var.name = .var.name, ..., any.missing = na.ok)
}

#' Assert that a value is a dosing interval
#'
#' @param tau The dosing interval
#' @returns `tau` or an informative error
assert_dosetau <- function(tau) {
  assert_number_between(x = tau, lower = 0, .var.name = checkmate::vname(tau), finite = TRUE)
}

#' Assert that a lambda.z value is valid
#'
#' @inheritParams assert_numeric_between
#' @param lambda.z The elimination rate (in units of inverse time) for
#'   extrapolation
#' @returns `lambda.z` or an informative error
assert_lambdaz <- function(lambda.z, any.missing = TRUE, .var.name = checkmate::vname(lambda.z)) {
  assert_numeric_between(x = lambda.z, lower = 0, any.missing = any.missing, .var.name = .var.name, finite = TRUE)
}

#' Assert that a value is a valid AUC method
#'
#' @param method The method for integration (one of 'lin up/log down',
#'   'lin-log', or 'linear')
#' @returns `method` or an informative error
assert_aucmethod <- function(method = c("lin up/log down", "linear", "lin-log")) {
  match.arg(method)
}

#' Assert that an object is a PKNCAdata object
#' @param object The PKNCAdata object
#' @returns The PKNCAdata object (confirmed to be usable)
assert_PKNCAdata <- function(object) {
  if (!inherits(object, "PKNCAdata")) {
    stop("Must be a PKNCAdata object")
  }
  if (nrow(object$intervals) == 0) {
    warning("No intervals given; no calculations will be done.")
  }
  object
}

#' @describeIn assert_unit Assert that a column name contains a character string
#'   (that could be a unit specification)
assert_unit_col <- function(unit, data) {
  if (length(unit) != 1) {
    stop("`unit` must be a single value")
  } else if (!is.character(unit)) {
    stop("`unit` must be a character string")
  } else if (!is.data.frame(data)) {
    stop("`data` must be a data.frame")
  } else if (!(unit %in% names(data))) {
    stop("`unit` (", unit, ") must be a column name in the data")
  } else if (!is.character(data[[unit]])) {
    stop("`unit` (", unit, ") must contain character data")
  }
  structure(unit, unit_type = "column")
}

#' @describeIn assert_unit Assert that a value may be a single unit
#'
#' The function does not verify that it is a real unit like "ng/mL" only that it
#' is a single character string.
assert_unit_value <- function(unit) {
  if (is.null(unit)) {
    return(unit)
  }

  if (length(unit) != 1) {
    stop("`unit` must be a single value")
  } else if (!is.character(unit)) {
    stop("`unit` must be a character string")
  }
  structure(unit, unit_type = "value")
}

#' Assert that a value may either be a column name in the data (first) or a
#' single unit value (second)
#'
#' @param unit The column name or unit value
#' @param data The data.frame that contains a column named `unit`
#' @returns `unit` with an attribute of "unit_type" that is either "column" or
#'   "value", or `NULL` if `is.null(unit)`
assert_unit <- function(unit, data) {
  unit_col <- try(assert_unit_col(unit = unit, data = data), silent = TRUE)
  unit_value <- try(assert_unit_value(unit = unit), silent = TRUE)
  if (!inherits(unit_col, "try-error")) {
    unit_col
  } else if (!inherits(unit_value, "try-error")) {
    unit_value
  } else {
    # Re-raise the unit_col error. That is better than unit_value since it is
    # stricter.
    stop(unit_col, call. = FALSE)
  }
}

test_that("assert_intervaltime_single", {
  expect_equal(
    assert_intervaltime_single(interval = c(1, 2)),
    c(1, 2)
  )
  expect_equal(
    assert_intervaltime_single(start = 1, end = 2),
    c(1, 2)
  )
  expect_equal(
    assert_intervaltime_single(interval = c(1, 2), start = 1, end = 2),
    c(1, 2)
  )
  expect_error(
    assert_intervaltime_single(interval = c(1, 2), start = 1.1, end = 2),
    regexp = "`start` must be the same as the first value in the interval if both are given: 1.1!=1",
    fixed = TRUE
  )
  expect_error(
    assert_intervaltime_single(interval = c(1, 2), start = 1, end = 2.1),
    regexp = "`end` must be the same as the second value in the interval if both are given: 2.1!=2",
    fixed = TRUE
  )
})

test_that("assert_conc", {
  expect_warning(
    assert_conc(conc = -1),
    regexp="Negative concentrations found"
  )
  expect_warning(
    assert_conc(conc = c(NA, -1)),
    regexp="Negative concentrations found"
  )
  expect_warning(
    assert_conc(conc = c(NA, -1, 1)),
    regexp="Negative concentrations found"
  )
  expect_warning(
    assert_conc(conc = NA),
    regexp="All concentration data are missing"
  )
  expect_error(
    assert_conc(conc="A"),
    regexp="Assertion on 'conc' failed: Must be of type 'numeric', not 'character'."
  )
  expect_error(
    assert_conc(conc=factor("A")),
    regexp="Assertion on 'conc' failed: Must be of type 'numeric', not 'factor'."
  )
})

test_that("assert_time", {
  expect_error(
    assert_time(time=NA),
    regexp="Assertion on 'time' failed: Contains missing values (element 1).",
    fixed = TRUE
  )
  expect_error(
    assert_time(time = c(0, 0)),
    regexp="Assertion on 'time' failed: Contains duplicated values, position 2."
  )
  expect_error(
    assert_time(time = c(1, 0)),
    regexp="Assertion on 'time' failed: Must be sorted."
  )
  expect_error(
    assert_time(time="A"),
    regexp="Assertion on 'time' failed: Must be of type 'numeric', not 'character'."
  )
  expect_error(
    assert_time(time=factor("A")),
    regexp="Assertion on 'time' failed: Must be of type 'numeric', not 'factor'."
  )
})

test_that("assert_conc_time", {
  expect_error(
    assert_conc_time(conc = 1, time = 1:2),
    regexp="Assertion on 'conc' failed: Must have length 2, but has length 1."
  )
  expect_error(
    assert_conc_time(conc = 1:2, time = 2),
    regexp="Assertion on 'conc' failed: Must have length 1, but has length 2."
  )
})

test_that("assert_lambdaz", {
  expect_equal(assert_lambdaz(1), 1)
  expect_equal(assert_lambdaz(NA), NA) # NA is allowed by default
  expect_error(
    assert_lambdaz(NA, any.missing = FALSE),
    regexp = "Assertion on 'NA' failed: Contains missing values (element 1).",
    fixed = TRUE
  )
  expect_error(
    assert_lambdaz(-1),
    regexp = "Assertion on '-1' failed: Element 1 is not > 0"
  )
})

test_that("assert_PKNCAdata", {
  expect_error(
    assert_PKNCAdata("A"),
    regexp = "Must be a PKNCAdata object"
  )
})

test_that("element_find", {
  values5<-c(TRUE, TRUE, TRUE, TRUE, TRUE, FALSE, TRUE, TRUE, TRUE, TRUE)
  values10<-c(TRUE, TRUE, TRUE, TRUE, TRUE, FALSE, TRUE, TRUE, TRUE, TRUE)
  expect_equal(
    element_find(values5),
    element_find(values10)
  )
  
  expect_equal(
    element_find(values5),
    "Elements 1, 2, 3, 4, 5"
  )
  
  expect_equal(
    element_find(values10),
    "Elements 1, 2, 3, 4, 5"
  )
})

test_that("assert_unit_col", {
  d <- data.frame(A = 1, B = "C")
  expect_equal(
    assert_unit_col(unit = "B", data = d),
    structure("B", unit_type = "column")
  )
  expect_error(
    assert_unit_col(unit = 1:2),
    regexp = "`unit` must be a single value"
  )
  expect_error(
    assert_unit_col(unit = 1),
    regexp = "`unit` must be a character string"
  )
  expect_error(
    assert_unit_col(unit = "D", data = "A"),
    regexp = "`data` must be a data.frame"
  )
  expect_error(
    assert_unit_col(unit = "D", data = d),
    regexp = "`unit` (D) must be a column name in the data",
    fixed = TRUE
  )
  expect_error(
    assert_unit_col(unit = "A", data = d),
    regexp = "`unit` (A) must contain character data",
    fixed = TRUE
  )
})

test_that("assert_unit_value", {
  expect_equal(assert_unit_value("A"), structure("A", unit_type = "value"))
  expect_null(assert_unit_value(NULL))
  expect_error(
    assert_unit_value(c("A", "B")),
    regexp = "`unit` must be a single value"
  )
  expect_error(
    assert_unit_value(1),
    regexp = "`unit` must be a character string"
  )
})

test_that("assert_unit", {
  d <- data.frame(A = 1, B = "C")

  expect_null(assert_unit(NULL, data = d))
  expect_equal(
    assert_unit(unit = "B", data = d),
    structure("B", unit_type = "column")
  )
  expect_equal(
    assert_unit(unit = "C", data = d),
    structure("C", unit_type = "value")
  )
  expect_error(
    assert_unit(unit = 1, data = d),
    regexp = "`unit` must be a character string"
  )
})

#' Check that the conversion to a data type does not change the number
#' of NA values
#'
#' @param x the value to convert
#' @param FUN the function to use for conversion
#' @param \dots arguments passed to `FUN`
#' @return `FUN(x, ...)` or an error if the set of NAs change.
#' @export
check.conversion <- function(x, FUN, ...) {
  ret <- FUN(x, ...)
  new.na <- sum(is.na(x) != is.na(ret))
  if (new.na != 0)
    # FIXME: It would be nice to have it give the function name as
    # part of the error
    stop(sprintf("%g new NA value(s) created during conversion",
                 new.na))
  ret
}

#' Round a value to a defined number of digits printing out trailing zeros, if
#' applicable.
#'
#' @param x The number to round
#' @param digits integer indicating the number of decimal places
#' @param sci_range See help for [signifString()] (and you likely want
#'   to round with `signifString` if you want to use this argument)
#' @param sci_sep The separator to use for scientific notation strings
#'   (typically this will be either "e" or "x10^" for computer- or
#'   human-readable output).
#' @param si_range Deprecated, please use `sci_range`
#' @return A string with the value
#' @details Values that are not standard numbers like `Inf`, `NA`, and
#'   `NaN` are returned as `"Inf"`, `"NA"`, and `NaN`.
#' @seealso [round()], [signifString()]
#' @export
roundString <- function(x, digits=0, sci_range=Inf, sci_sep="e", si_range) {
  if (!missing(si_range)) {
    .Deprecated(new="roundString with the sci_range argument",
                msg="The si_range argument is deprecated, please use sci_range")
    sci_range <- si_range
  }
  if (length(digits) == 1) {
    mask_na <- is.na(x)
    mask_aschar <- is.nan(x) | is.infinite(x)
    mask_manip <- !(mask_na | mask_aschar)
    ret <- rep(NA, length(x))
    # Put in the special values
    if (any(mask_na)) {
      ret[mask_na] <- "NA"
    }
    if (any(mask_aschar)) {
      ret[mask_aschar] <- as.character(x[mask_aschar])
    }
    if (any(mask_manip)) {
      xtmp <- round(x[mask_manip], digits)
      mask_sci <-
        xtmp != 0 &
        abs(log10(abs(xtmp))) >= sci_range
      mask_no_sci <- !mask_sci
      if (any(mask_sci)) {
        logval <- floor(log10(abs(xtmp[mask_sci])))
        ret[mask_manip][mask_sci] <-
          paste0(
            formatC(xtmp[mask_sci]/10^logval, format="f", digits=digits + logval),
            sci_sep,
            formatC(logval, format="d"))
      }
      if (any(mask_no_sci)) {
        if (digits < 0) {
          ret[mask_manip][mask_no_sci] <-
            formatC(xtmp[mask_no_sci], format='f', digits=0)
        } else {
          ret[mask_manip][mask_no_sci] <-
            formatC(xtmp[mask_no_sci], format='f', digits=digits)
        }
      }
    }
    ret
  } else if (length(x) == length(digits)) {
    mapply(roundString, x, digits=digits, sci_range=sci_range, sci_sep=sci_sep)
  } else {
    stop("digits must either be a scalar or the same length as x")
  }
}

#' Round a value to a defined number of significant digits printing out trailing
#' zeros, if applicable.
#'
#' @param x The number to round
#' @param digits integer indicating the number of significant digits
#' @param sci_range integer (or `Inf`) indicating when to switch to
#'   scientific notation instead of floating point. Zero indicates always use
#'   scientific; `Inf` indicates to never use scientific notation;
#'   otherwise, scientific notation is used when `abs(log10(x)) > si_range`.
#' @param sci_sep The separator to use for scientific notation strings
#'   (typically this will be either "e" or "x10^" for computer- or
#'   human-readable output).
#' @param si_range Deprecated, please use `sci_range`
#' @param ... Arguments passed to methods.
#' @return A string with the value
#' @details Values that are not standard numbers like `Inf`, `NA`, and
#'   `NaN` are returned as `"Inf"`, `"NA"`, and `NaN`.
#' @seealso [signif()], [roundString()]
#' @export
signifString <- function(x, ...)
  UseMethod("signifString")

#' @rdname signifString
#' @export
signifString.data.frame <- function(x, ...) {
  ret <- lapply(x,
                function(y) {
                  if (is.numeric(y) & !is.factor(y)) {
                    signifString(x=y, ...)
                  } else {
                    y
                  }
                })
  ret <- as.data.frame(ret,
                       stringsAsFactors=FALSE)
  rownames(ret) <- rownames(x)
  colnames(ret) <- colnames(x)
  ret
}

#' @rdname signifString
#' @export
signifString.default <- function(x, digits=6, sci_range=6, sci_sep="e", si_range, ...) {
  if (length(list(...))) {
    stop("Additional, unsupported arguments were passed")
  }
  if (!missing(si_range)) {
    .Deprecated(new="roundString with the sci_range argument",
                msg="The si_range argument is deprecated, please use sci_range")
    sci_range <- si_range
  }
  mask_na <- is.na(x)
  mask_aschar <- is.nan(x) | is.infinite(x)
  mask_manip <- !(mask_na | mask_aschar)
  ret <- rep(NA, length(x))
  # Put in the special values
  if (any(mask_na)) {
    ret[mask_na] <- "NA"
  }
  if (any(mask_aschar)) {
    ret[mask_aschar] <- as.character(x[mask_aschar])
  }
  if (any(mask_manip)) {
    xtmp <- x[mask_manip]
    toplog <- bottomlog <- rep(NA, length(xtmp))
    # When 0 give the digits as the output
    bottomlog[xtmp %in% 0] <- digits
    # Otherwise set it to digits orders of magnitude lower than the
    # current value
    toplog <- log10(abs(xtmp))
    # When the order of magnitude is an exact log 10, move up one so
    # that the math works for determing the lower log.
    mask.exact.log <- (toplog %% 1) %in% 0
    toplog[mask.exact.log] <- toplog[mask.exact.log] + 1
    toplog <- ceiling(toplog)
    bottomlog[is.na(bottomlog)] <- digits-toplog[is.na(bottomlog)]
    # Find times when rounding increases the toplog and shift up the
    # bottomlog to a corresponding degree. e.g. x=0.9999 and digits=2
    # should be 1.0 not 1.00.
    newtoplog <- log10(abs(round(xtmp, digits=bottomlog)))
    mask.exact.log <- (newtoplog %% 1) %in% 0
    newtoplog[mask.exact.log] <- newtoplog[mask.exact.log] + 1
    newtoplog <- ceiling(newtoplog)
    mask.move.up <- toplog < newtoplog
    bottomlog[mask.move.up] <- bottomlog[mask.move.up] - 1
    # Do the rounding
    ret[mask_manip] <- roundString(xtmp, digits=bottomlog,
                                   sci_range=sci_range, sci_sep=sci_sep)
  }
  ret
}

#' Find the summary statistic value with a different value if the input is
#' zero-length.
#'
#' @details If `na.rm` is `TRUE`, then `NA` values will be removed prior to the
#'   check if `length(c(...)) == 0`.
#'
#' @param ... objects to find the summary_statistic for (combined with `c()`
#'   prior to calculation)
#' @param na.rm a logical indicating whether missing values should be removed
#'   (see Details).
#' @param zero_length The value to return if `length(x) == 0`
#' @param FUN the summary statistic function (such as `max` or `min`)
#' @return Either `zero_length` or `FUN(...)`
#' @noRd
zero_len_summary <- function(FUN) {
  function(..., na.rm=FALSE, zero_length=NA) { #nocov
    x <- c(...)
    if (na.rm) {
      x <- stats::na.omit(x)
    }
    if (length(x) == 0) {
      zero_length
    } else {
      FUN(x)
    }
  } #nocov
}

#' @describeIn zero_len_summary Find the maximum value with a different value if
#'   the input is zero-length
#' @noRd
max_zero_len <- zero_len_summary(FUN=max)
#' @describeIn zero_len_summary Find the minimum value with a different value if
#'   the input is zero-length
#' @noRd
min_zero_len <- zero_len_summary(FUN=min)

test_that("check.conversion", {
  good <- LETTERS
  expect_equal(check.conversion(good, as.character), good)
  expect_error(
    expect_warning(
      check.conversion(good, as.numeric)
    ),
    regexp="26 new NA value\\(s\\) created during conversion"
  )
  good <- 1:5
  expect_equal(check.conversion(good, as.character),
               as.character(good))
})

test_that("Rounding", {
  expect_error(roundString(1, c(2, 3)),
               regexp="digits must either be a scalar or the same length as x")
  expect_equal(roundString(11), "11")
  expect_equal(roundString(5), "5")
  expect_equal(roundString(0.05), "0")
  expect_equal(roundString(NA), "NA")
  expect_equal(roundString(NaN), "NaN")
  expect_equal(roundString(Inf), "Inf")
  expect_equal(roundString(-Inf), "-Inf")
  # Respecting the digits
  expect_equal(roundString(0.05, 3), "0.050")
  expect_equal(roundString(123.05, 3), "123.050")
  expect_equal(roundString(c(100, 0.1), 3), c("100.000", "0.100"),
               info="Vectors work with different orders of magnitude work")
  expect_equal(roundString(c(100, 0.1), c(0, 3)), c("100", "0.100"),
               info="Vectors of digits work")
  expect_equal(roundString(c(0.1, NA), digits=3), c("0.100", "NA"),
               info="Mixed inputs (NA, NaN, Inf or numeric), NA")
  expect_equal(roundString(c(0.1, NA, NaN, Inf, -Inf), digits=3),
               c("0.100", "NA", "NaN", "Inf", "-Inf"),
               info="Mixed inputs (NA, NaN, Inf or numeric)")
  # All zeros
  expect_equal(roundString(0, digits=3), "0.000")
  expect_equal(roundString(c(0, NA), digits=3), c("0.000", "NA"))
  # scientific notation
  expect_equal(roundString(1234567, digits=3, sci_range=5), "1.234567000e6",
               info="sci_range works with roundString (even if it looks odd)")
  expect_warning(roundString(1234567, digits=3, si_range=5),
                 regexp="The si_range argument is deprecated, please use sci_range")
  expect_equal(roundString(1234567, digits=3, sci_range=5),
               roundString(1234567, digits=3, sci_range=5),
               info="sci_range works with roundString (even if it looks odd)")
  expect_equal(roundString(1234567, digits=3, sci_range=5, sci_sep="x10^"),
               "1.234567000x10^6",
               info="sci_sep is respected.")
  expect_equal(roundString(c(1e7, 1e10), digits=c(-3, -9), sci_range=5),
               c("1.0000e7", "1.0e10"),
               info="Different numbers of digits for rounding work with roundString")
})

test_that("Significance", {
  expect_equal(signifString(11), "11.0000")
  expect_equal(signifString(5), "5.00000")
  expect_equal(signifString(0.05), "0.0500000")
  expect_equal(signifString(NA), "NA")
  expect_equal(signifString(NaN), "NaN")
  expect_equal(signifString(Inf), "Inf")
  expect_equal(signifString(-Inf), "-Inf")
  # Respecting the digits
  expect_equal(signifString(0.05, 3), "0.0500")
  expect_equal(signifString(123.05, 3), "123")
  expect_equal(signifString(123456.05, 3), "123000")
  expect_warning(signifString(123456.05, 3, si_range=6),
                regexp="The si_range argument is deprecated, please use sci_range")
  expect_warning(
    expect_equal(
      signifString(123456.05, 3, sci_range=6),
      signifString(123456.05, 3, si_range=6),
      info="si_range and sci_range arguments are treated equally."
    )
  )
  expect_equal(signifString(123456.05, 3, sci_range=6), "123000")
  expect_equal(signifString(123456.05, 3, sci_range=5), "1.23e5")
  expect_equal(signifString(-123000.05, 3, sci_range=5), "-1.23e5")
  expect_equal(signifString(999999, 3, sci_range=6), "1.00e6",
               info="Rounding around the edge of the sci_range works correctly (going up)")
  expect_equal(signifString(999999, 7, sci_range=6), "999999.0",
               info="Rounding around the edge of the sci_range works correctly (going staying the same)")
  expect_equal(signifString(-.05, 3), "-0.0500")
  # Exact orders of magnitude work on both sides of 0
  expect_equal(signifString(0.01, 3), "0.0100")
  expect_equal(signifString(1, 3), "1.00")
  expect_equal(signifString(100, 3), "100")
  # Vectors work with different orders of magnitude work
  expect_equal(signifString(c(100, 0.1), 3), c("100", "0.100"))
  # Rounding to a higher number of significant digits works correctly
  expect_equal(signifString(0.9999999, 3), "1.00")
  # Mixed inputs (NA, NaN, Inf or numeric)
  expect_equal(signifString(NA), "NA")
  expect_equal(signifString(c(0.1, NA), digits=3), c("0.100", "NA"))
  expect_equal(signifString(c(0.1, NA, NaN, Inf, -Inf), digits=3),
               c("0.100", "NA", "NaN", "Inf", "-Inf"))
  # All zeros
  expect_equal(signifString(0, digits=3), "0.000")
  expect_equal(signifString(c(0, NA), digits=3), c("0.000", "NA"))

  expect_equal(signifString(1234567, digits=3, sci_range=5, sci_sep="x10^"),
               "1.23x10^6",
               info="sci_sep is respected.")
  expect_equal(signifString(c(1e7, 1e10), digits=3),
               c("1.00e7", "1.00e10"),
               info="Different numbers of digits for rounding work with signifString")

  # Data Frames
  expect_equal(
    signifString(data.frame(A=c(0, 1.111111),
                            B=factor(LETTERS[1:2]),
                            C=LETTERS[1:2],
                            stringsAsFactors=FALSE),
                 digits=3),
    data.frame(A=c("0.000", "1.11"),
               B=factor(LETTERS[1:2]),
               C=LETTERS[1:2],
               stringsAsFactors=FALSE),
    ignore_attr=TRUE,
    info="Data frame significance is calculated correctly"
  )
  expect_equal(
    signifString(data.frame(A=c(0, 1.111111),
                            B=factor(LETTERS[1:2]),
                            C=LETTERS[1:2],
                            stringsAsFactors=FALSE),
                 digits=4),
    data.frame(A=c("0.0000", "1.111"),
               B=factor(LETTERS[1:2]),
               C=LETTERS[1:2],
               stringsAsFactors=FALSE),
    ignore_attr=TRUE,
    info="Data frame digits are respected"
  )
})

test_that("signifString stops when bad arguments are passed", {
  expect_error(
    signifString(1, foo=1),
    regexp="Additional, unsupported arguments were passed",
    fixed=TRUE
  )
})

test_that("max_na and min_na", {
  expect_equal(max_zero_len(1), max(1))
  expect_equal(max_zero_len(c(1, NA)), max(c(1, NA)))
  expect_equal(max_zero_len(1, NA), max(1, NA))
  expect_equal(max_zero_len(1, NA, na.rm=TRUE), max(1, NA, na.rm=TRUE))
  expect_equal(
    max_zero_len(numeric(), na.rm=TRUE),
    NA
  )
  expect_equal(
    max_zero_len(NA_integer_, na.rm=TRUE),
    NA
  )
  expect_equal(
    max_zero_len(NA_integer_, na.rm=TRUE, zero_length=NA_real_),
    NA_real_,
    info="zero_length/na.rm combinations are respected (made zero length)"
  )
  expect_equal(
    max_zero_len(NA_integer_, na.rm=FALSE, zero_length=NA_real_),
    NA_integer_,
    info="zero_length/na.rm combinations are respected (NOT made zero length)"
  )
})

test_that("zero_len_summary", {
  testing_function <- zero_len_summary(mean)
  expect_equal(testing_function(1:5), 3)
  expect_equal(testing_function(1:5, na.rm = TRUE), 3)
  expect_equal(testing_function(), NA)
})

#' Compute noncompartmental pharmacokinetics
#'
#' Compute pharmacokinetic (PK) noncompartmental analysis (NCA)
#' parameters.
#'
#' PKNCA has been cross-validated with both Phoenix WinNonlin(R) and Pumas
#' (click here for the
#' [cross-validation
#' article](http://www.humanpredictions.com/wp-content/uploads/2020/01/ACOP_2019_T102_NCA_performance_evaluation_Yingbo_revised.pdf))
#'
#' A common workflow would load data from a file or database into a
#' data.frame then run the following code.
#' @examples
#' \dontrun{
#' # Load concentration-time data into a data.frame called d.conc
#' # with columns named "conc", "time", and "subject".
#' my.conc <- PKNCAconc(d.conc, conc~time|subject)
#' # Load dose-time data into a data.frame called d.dose
#' # with columns named "dose", "time", and "subject".
#' my.dose <- PKNCAdose(d.dose, dose~time|subject)
#' # Combine the concentration-time and dose-time data into an object
#' # ready for calculations.
#' my.data <- PKNCAdata(my.conc, my.dose)
#' # Perform the calculations
#' my.results <- pk.nca(my.data)
#' # Look at summary results
#' summary(my.results)
#' # Look at a listing of results
#' as.data.frame(my.results)
#' }
#' @docType package
#' @name PKNCA
"_PACKAGE"

# To work with the use of dplyr's pipe within the exclude function
utils::globalVariables(".")

.onLoad <- function(...) {
  if (requireNamespace("units", quietly = TRUE)) {
    # Allow "fraction" to be used as a unit for fraction excreted
    units::install_unit(symbol = "fraction", def = "1")
  }
}

#' Interpolate concentrations between measurements or extrapolate concentrations
#' after the last measurement.
#'
#' `interpolate.conc()` and `extrapolate.conc()` returns an interpolated (or
#' extrapolated) concentration. `interp.extrap.conc()` will choose whether
#' interpolation or extrapolation is required and will also operate on many
#' concentrations.  These will typically be used to estimate the concentration
#' between two measured concentrations or after the last measured concentration.
#' Of note, these functions will not extrapolate prior to the first point.
#'
#' An `NA` value for the `lambda.z` parameter will prevent extrapolation.
#'
#' @inheritParams assert_conc_time
#' @inheritParams assert_lambdaz
#' @inheritParams PKNCA.choose.option
#' @inheritParams choose_interval_method
#' @param interp.method,extrap.method deprecated in favor of method and auc.type
#' @param time.dose Time of the dose
#' @param time.out Time when interpolation is requested (vector for
#'   `interp.extrap.conc()`, scalar otherwise)
#' @param clast The last observed concentration above the limit of
#'   quantification.  If not given, `clast` is calculated from
#'   [pk.calc.clast.obs()]
#' @param conc.origin The concentration before the first measurement.
#'   `conc.origin` is typically used to set predose values to zero (default),
#'   set a predose concentration for endogenous compounds, or set predose
#'   concentrations to `NA` if otherwise unknown.
#' @param conc.blq How to handle BLQ values. (See [clean.conc.blq()] for usage
#'   instructions.)
#' @param conc.na How to handle NA concentrations.  (See [clean.conc.na()])
#' @param route.dose What is the route of administration ("intravascular" or
#'   "extravascular").  See the details for how this parameter is used.
#' @param duration.dose What is the duration of administration? See the details
#'   for how this parameter is used.
#' @param out.after Should interpolation occur from the data before (`FALSE`) or
#'   after (`TRUE`) the interpolated point?  See the details for how this
#'   parameter is used.  It only has a meaningful effect at the instant of an IV
#'   bolus dose.
#' @param check Run [assert_conc_time()], [clean.conc.blq()], and
#'   [clean.conc.na()]?
#' @param ... Additional arguments passed to `interpolate.conc()` or
#'   `extrapolate.conc()`.
#' @returns The interpolated or extrapolated concentration value as a scalar
#'   double (or vector for `interp.extrap.conc()`).
#'
#' @details
#' \describe{
#'   \item{extrap.method}{
#'     \describe{
#'       \item{'AUCinf'}{Use lambda.z to extrapolate beyond the last point with the half-life.}
#'       \item{'AUCall'}{If the last point is above the limit of quantification or missing, this is identical to 'AUCinf'. If the last point is below the limit of quantification, then linear interpolation between the Clast and the next BLQ is used for that  interval and all additional points are extrapolated as 0.}
#'       \item{'AUClast'}{Extrapolates all points after the last above the  limit of quantification as 0.}
#'     }
#'   }
#' }
#'
#' `duration.dose` and `direction.out` are ignored if `route.dose ==
#' "extravascular"`.  `direction.out` is ignored if `duration.dose > 0`.
#'
#' `route.dose` and `duration.dose` affect how interpolation/extrapolation of
#' the concentration occurs at the time of dosing.  If `route.dose ==
#' "intravascular"` and `duration.dose == 0` then extrapolation occurs for an IV
#' bolus using [pk.calc.c0()] with the data after dosing.  Otherwise (either
#' `route.dose == "extravascular"` or `duration.dose > 0`), extrapolation occurs
#' using the concentrations before dosing and estimating the half-life (or more
#' precisely, estimating `lambda.z`).  Finally, `direction.out` can change the
#' direction of interpolation in cases with `route.dose == "intravascular"` and
#' `duration.dose == 0`.  When `direction.out == "before"` interpolation occurs
#' only with data before the dose (as is the case for `route.dose ==
#' "extravascular"`), but if `direction.out == "after"` interpolation occurs
#' from the data after dosing.
#'
#' @seealso [pk.calc.clast.obs()], [pk.calc.half.life()], [pk.calc.c0()]
#' @export
interp.extrap.conc <- function(conc, time, time.out,
                               lambda.z = NA,
                               clast = pk.calc.clast.obs(conc, time),
                               options = list(),
                               method = NULL,
                               auc.type = "AUCinf",
                               interp.method,
                               extrap.method,
                               ...,
                               conc.blq = NULL,
                               conc.na = NULL,
                               check = TRUE) {
  # Defunct inputs
  if (!missing(interp.method)) {
    .Defunct(msg = "the `interp.method` has been replaced by the `method` argument for consistency with the rest of PKNCA") # nocov
  } else if (!missing(extrap.method)) {
    .Defunct(msg = "the `extrap.method` has been replaced by the `auc.type` argument for consistency with the rest of PKNCA") # nocov
  }

  # Check inputs
  method <- PKNCA.choose.option(name="auc.method", value=method, options=options)
  if (check) {
    assert_conc_time(conc = conc, time = time)
    data <-
      clean.conc.blq(
        conc = conc, time = time,
        conc.blq = conc.blq, conc.na = conc.na,
        options = options,
        check=FALSE
      )
  } else {
    data <- data.frame(conc, time)
  }
  if (length(time.out) < 1) {
    stop("time.out must be a vector with at least one element")
  }
  if (all(data$conc %in% 0)) {
    # tlast would be NA in this case, but if everything input is zero, then all
    # interpolated and extrapolated times will be zero, too.
    ret <- rep(0, length(time.out))
  } else {
    tlast <- pk.calc.tlast(data$conc, data$time, check=FALSE)
    ret <- rep(NA, length(time.out))
    for (i in seq_len(length(time.out)))
      if (is.na(tlast)) {
        stop("Please report a bug:  tlast is NA; cannot interpolate/extrapolate") # nocov
      } else if (is.na(time.out[i])) {
        warning("An interpolation/extrapolation time is NA")
      } else if (time.out[i] <= tlast) {
        ret[i] <-
          interpolate.conc(
            conc=data$conc, time=data$time,
            time.out=time.out[i],
            method=method,
            conc.blq=conc.blq,
            conc.na=conc.na,
            check=FALSE
          )
      } else {
        ret[i] <-
          extrapolate.conc(
            conc=data$conc, time=data$time,
            time.out=time.out[i],
            lambda.z=lambda.z,
            clast=clast,
            auc.type=auc.type,
            check=FALSE
          )
      }
  }
  ret
}

#' @describeIn interp.extrap.conc Interpolate concentrations through Tlast (inclusive)
#' @export
interpolate.conc <- function(conc, time, time.out,
                             options=list(),
                             method = NULL,
                             interp.method,
                             conc.blq=NULL,
                             conc.na=NULL,
                             conc.origin=0,
                             ...,
                             check=TRUE) {
  # Defunct inputs
  if (!missing(interp.method)) {
    .Defunct(msg = "the `interp.method` has been replaced by the `method` argument for consistency with the rest of PKNCA") # nocov
  }
  # Check the inputs
  method <-
    tolower(PKNCA.choose.option(
      name="auc.method", value=method, options=options
    ))
  if (check) {
    assert_conc_time(conc, time)
    data <-
      clean.conc.blq(
        conc=conc, time=time,
        conc.blq=conc.blq,
        conc.na=conc.na,
        options = options,
        check=FALSE
      )
  } else {
    data <- data.frame(conc, time)
  }
  checkmate::assert_number(x=conc.origin, na.ok=TRUE)
  checkmate::assert_number(x=time.out, na.ok=FALSE)
  if (time.out > max(data$time)) {
    stop("`interpolate.conc()` does not extrapolate, use `interp.extrap.conc()`")
  }
  # Verify that we are interpolating between the first concentration
  # and the last above LOQ concentration
  tlast <- pk.calc.tlast(conc=data$conc, time=data$time, check=FALSE)
  if (time.out < min(data$time)) {
    ret <- conc.origin
  } else if (all(data$conc == 0)) {
    ret <- 0
  } else if (time.out > tlast) {
    stop("`interpolate.conc()` can only works through Tlast, please use `interp.extrap.conc()` to combine both interpolation and extrapolation.")
  } else if (time.out %in% data$time) {
    # See if there is an exact time match and return that if it
    # exists.
    ret <- data$conc[time.out == data$time]
  } else {
    interp_methods_all <-
      choose_interval_method(
        conc = data$conc,
        time = data$time,
        method = method,
        # AUClast because it doesn't affect the output for interpolation
        auc.type = "AUClast",
        options = options
      )
    # Find the last time before and the first time after the output
    # time, then interpolate.
    idx_before <- rev(which(data$time <= time.out))[1]
    idx_after <- which(time.out <= data$time)[1]
    time_1 <- data$time[idx_before]
    time_2 <- data$time[idx_after]
    conc_1 <- data$conc[idx_before]
    conc_2 <- data$conc[idx_after]
    interp_method <- interp_methods_all[idx_before]
    ret <-
      if (interp_method == "linear") {
        interpolate_conc_linear(conc_1=conc_1, conc_2=conc_2, time_1=time_1, time_2=time_2, time_out=time.out)
      } else if (interp_method == "log") {
        interpolate_conc_log(conc_1=conc_1, conc_2=conc_2, time_1=time_1, time_2=time_2, time_out=time.out)
      } else if (interp_method == "zero") {
        0
      } else {
        stop("Please report a bug: invalid interp_method") # nocov
      }
  }
  ret
}

#' @describeIn interp.extrap.conc Extrapolate concentrations after Tlast
#' @export
extrapolate.conc <- function(conc, time, time.out,
                             lambda.z=NA, clast=pk.calc.clast.obs(conc, time),
                             auc.type = "AUCinf",
                             extrap.method,
                             options=list(),
                             conc.na=NULL,
                             conc.blq=NULL,
                             ...,
                             check=TRUE) {
  if (!missing(extrap.method)) {
    .Defunct(msg = "the `extrap.method` has been replaced by the `auc.type` argument for consistency with the rest of PKNCA") # nocov
  }
  assert_lambdaz(lambda.z)
  if (check) {
    assert_conc_time(conc, time)
    data <-
      clean.conc.blq(
        conc = conc, time = time,
        conc.blq = conc.blq, conc.na = conc.na,
        options = options,
        check = FALSE
      )
  } else {
    data <- data.frame(conc, time)
  }
  auc.type <- tolower(auc.type)
  if (!(auc.type %in% c("aucinf", "aucall", "auclast")))
    stop("`auc.type` must be one of 'AUCinf', 'AUClast', or 'AUCall'")
  if (length(time.out) != 1)
    stop("Only one time.out value may be estimated at once.")
  tlast <- pk.calc.tlast(conc=data$conc, time=data$time, check=FALSE)
  if (is.na(tlast)) {
    # If there are no observed concentrations, return NA
    ret <- NA
  } else if (time.out <= tlast) {
    stop("extrapolate.conc can only work beyond Tlast, please use interp.extrap.conc to combine both interpolation and extrapolation.")
  } else {
    # Start the interpolation
    if (auc.type %in% "aucinf") {
      # If AUCinf is requested, extrapolate using the half-life
      ret <- extrapolate_conc_lambdaz(clast=clast, lambda.z=lambda.z, tlast=tlast, time_out=time.out)
    } else if (auc.type %in% "auclast" ||
                 (auc.type %in% "aucall" &&
                    tlast == max(data$time))) {
      # If AUClast is requested or AUCall is requested and there are
      # no BLQ at the end, we are already certain that we are after
      # Tlast, so the answer is 0.
      ret <- 0
    } else if (auc.type %in% "aucall") {
      # If the last non-missing concentration is below the limit of
      # quantification, extrapolate with the triangle method of
      # AUCall.
      time_prev <- max(data$time[data$time <= time.out])
      conc_prev <- data$conc[data$time %in% time_prev]
      if (conc_prev %in% 0) {
        # If we are already BLQ, then we have confirmed that there
        # are no more ALQ measurements (because we are beyond
        # Tlast) and therefore we can extrapolate as 0.
        ret <- 0
      } else {
        if (time_prev != max(data$time)) {
          time_next <- min(data$time[data$time >= time.out])
        }
        # If we are not already BLQ, then we have confirmed that we
        # are in the triangle extrapolation region and need to draw
        # a line.
        ret <-
          interpolate_conc_linear(
            conc_1 = conc_prev, conc_2 = 0,
            time_1 = time_prev, time_2 = time_next,
            time_out = time.out
          )
      }
    } else {
      stop("Invalid auc.type caught too late (seeing this error indicates a software bug)") # nocov
    }
  }
  ret
}

# Choices for events in interp.extrap.conc.dose.  This is included here to
# assist with testing later.
event_choices_interp.extrap.conc.dose <-
  list(conc_dose_iv_bolus_after="conc_dose_iv_bolus_after",
       conc_dose="conc_dose",
       dose_iv_bolus_after="dose_iv_bolus_after",
       dose="dose",
       conc="conc",
       output_only="output_only",
       none="none")

#' @describeIn interp.extrap.conc Interpolate and extrapolate concentrations
#'   without interpolating or extrapolating beyond doses.
#' @export
interp.extrap.conc.dose <- function(conc, time,
                                    time.dose, route.dose = "extravascular", duration.dose = NA,
                                    time.out, out.after = FALSE,
                                    options = list(),
                                    conc.blq = NULL,
                                    conc.na = NULL,
                                    ...,
                                    check = TRUE) {
  # Check inputs
  if (check) {
    assert_conc_time(conc = conc, time = time)
    data_conc <-
      clean.conc.blq(
        conc = conc, time = time,
        conc.blq = conc.blq, conc.na = conc.na,
        options = options,
        check=FALSE
      )
  } else {
    data_conc <- data.frame(conc, time)
  }
  # Check other inputs
  if (!is.character(route.dose)) {
    route.dose <- as.character(route.dose)
  }
  if (!(all(route.dose %in% c("extravascular", "intravascular")))) {
    stop("route.dose must be either 'extravascular' or 'intravascular'")
  }
  if (!(length(route.dose) %in% c(1, length(time.dose)))) {
    stop("route.dose must either be a scalar or the same length as time.dose")
  }
  if (!all(is.na(duration.dose) | (is.numeric(duration.dose) & !is.factor(duration.dose)))) {
    stop("duration.dose must be NA or a number.")
  }
  if (!(length(duration.dose) %in% c(1, length(time.dose)))) {
    stop("duration.dose must either be a scalar or the same length as time.dose")
  }

  # Generate a single timeline

  # Concentrations are assumed to occur before dosing
  data_conc$out_after <- FALSE
  data_dose <-
    merge(
      data.frame(dose=TRUE,
                 time=time.dose,
                 route=route.dose,
                 duration=duration.dose,
                 iv_bolus=route.dose %in% "intravascular" & duration.dose %in% 0,
                 stringsAsFactors=FALSE),
      # Expand IV bolus dosing to have a before and after concentration
      data.frame(iv_bolus=c(FALSE, TRUE, TRUE),
                 out_after=c(FALSE, FALSE, TRUE)),
      all.x=TRUE)
  data_out <-
    data.frame(out=TRUE,
               out_after=out.after,
               out_order=seq_along(time.out),
               time=time.out)
  data_all <-
    merge(merge(data_conc,
                data_dose,
                all=TRUE),
          data_out,
          all=TRUE)
  data_all$dose_event <- data_all$dose %in% TRUE
  data_all$conc_event <- !is.na(data_all$conc)
  data_all$iv_bolus <- data_all$iv_bolus %in% TRUE
  data_all$out <- data_all$out %in% TRUE
  data_all$dose_count <- cumsum(data_all$dose_event)
  mask_include_before <- data_all$dose_event & data_all$conc_event & !data_all$out_after
  data_all$dose_count_prev <- data_all$dose_count - mask_include_before
  data_all$event <- dplyr::case_when(
    data_all$dose_event & data_all$conc_event & data_all$iv_bolus & data_all$out_after~event_choices_interp.extrap.conc.dose$conc_dose_iv_bolus_after,
    data_all$dose_event & data_all$conc_event~event_choices_interp.extrap.conc.dose$conc_dose,
    data_all$dose_event & data_all$iv_bolus & data_all$out_after~event_choices_interp.extrap.conc.dose$dose_iv_bolus_after,
    data_all$dose_event~event_choices_interp.extrap.conc.dose$dose,
    data_all$conc_event~event_choices_interp.extrap.conc.dose$conc,
    data_all$out~event_choices_interp.extrap.conc.dose$output_only, # interpolation/extrapolation-only row
    TRUE~"unknown") # should never happen
  if (any(mask_unknown <- data_all$event %in% "unknown")) {
    # All events should be accounted for already
    stop( # nocov
      "Unknown event in interp.extrap.conc.dose at time(s): ", # nocov
      paste(unique(data_all$time[mask_unknown]), collapse=", "), # nocov
      " (Please report this as a bug)" # nocov
    ) # nocov
  }
  # Remove "output_only" from event_before and event_after
  simple_locf <- function(x, missing_val) {
    mask_found <- !(x %in% missing_val)
    ret <- x[mask_found]
    ret[cumsum(mask_found)]
  }
  data_all$event_before <- simple_locf(c(event_choices_interp.extrap.conc.dose$none, data_all$event[-nrow(data_all)]),
                                       "output_only")
  data_all$event_after <- rev(simple_locf(rev(c(data_all$event[-1], event_choices_interp.extrap.conc.dose$none)),
                                          "output_only"))

  # Loop through the methods until all have been tested or no missing
  # values remain.
  data_all$method <- NA_character_
  for (nm in names(interp.extrap.conc.dose.select)) {
    mask <- is.na(data_all$method) &
      do.call(interp.extrap.conc.dose.select[[nm]]$select, list(x=data_all))
    if (any(mask)) {
      if ("warning" %in% names(interp.extrap.conc.dose.select[[nm]])) {
        warning(sprintf("%s: %d data points",
                        interp.extrap.conc.dose.select[[nm]]$warning,
                        sum(mask)))
        data_all$method[mask] <- nm
      } else {
        for (current_idx in which(mask)) {
          data_all$conc[current_idx] <-
            do.call(interp.extrap.conc.dose.select[[nm]]$value,
                    list(data_all=data_all,
                         current_idx=current_idx,
                         options=options,
                         ...))
          data_all$method[current_idx] <- nm
        }
      }
    }
  }
  if (any(mask_no_method <- is.na(data_all$method))) {
    # This should never happen, all eventualities should be covered
    stop("No method for imputing concentration at time(s): ", # nocov
         paste(unique(data_all$time[mask_no_method]), collapse=", ")) # nocov
  }
  # Filter to the requested time points and output
  data_out <- data_all[data_all$out,,drop=FALSE]
  data_out <- data_out[order(data_out$out_order),,drop=FALSE]
  ret <- data_out$conc
  attr(ret, "Method") <- data_out$method
  ret
}

# Dose-aware interpolation/extrapolation functions ####

# Impossible combinations ####
iecd_impossible_select <- function(x) {
  x$event_before %in% "output_only" | # Restricted in code
    x$event %in% "none" | # "none" events do not occur, they are before or after the timeline
    x$event_after %in% "output_only" | # Restricted in code
    (x$event %in% "output_only" &
       x$event_after %in% c("conc_dose_iv_bolus_after", "dose_iv_bolus_after")) |
    (x$event_before %in% c("conc_dose_iv_bolus_after", "dose_iv_bolus_after") &
       x$event %in% c("conc_dose_iv_bolus_after", "dose_iv_bolus_after")) |
    (x$event %in% c("conc_dose_iv_bolus_after", "dose_iv_bolus_after") &
       x$event_after %in% c("conc_dose_iv_bolus_after", "dose_iv_bolus_after")) |
    (x$event_before %in% c("conc_dose_iv_bolus_after", "dose_iv_bolus_after") &
       x$event %in% c("none", "output_only") &
       x$event_after %in% c("conc_dose_iv_bolus_after", "dose_iv_bolus_after"))
}
iecd_impossible_value <- function(data_all, current_idx, ...) {
  stop(sprintf( # nocov
    "Impossible combination requested for interp.extrap.conc.dose (please report this as a bug).  event_before: %s, event: %s, event_after: %s", # nocov
    data_all$event_before[current_idx], # nocov
    data_all$event[current_idx], # nocov
    data_all$event_after[current_idx] # nocov
  )) # nocov
}

# Observed concentration ####
iecd_observed_select <- function(x) {
  x$event %in% c("conc_dose_iv_bolus_after", "conc_dose", "conc") &
    !do.call(interp.extrap.conc.dose.select[["Impossible combinations"]]$select, list(x=x))
}
iecd_observed_value <- function(data_all, current_idx, ...) {
  data_all$conc[current_idx]
}

# Before all events ####
iecd_before_select <- function(x) {
  x$event_before %in% "none" &
    !(x$event %in% c("conc_dose_iv_bolus_after", "conc_dose", "conc", "dose_iv_bolus_after", "none")) &
    !(x$event_after %in% "output_only" |
        (x$event %in% "output_only" &
           x$event_after %in% c("conc_dose_iv_bolus_after", "dose_iv_bolus_after"))) # because these are impossible
}
iecd_before_value <- function(data_all, current_idx, conc.origin=0, ...) {
  conc.origin
}

# Interpolation ####
iecd_interp_select <- function(x) {
  x$event_before %in% c("conc_dose_iv_bolus_after", "conc_dose", "conc") &
    x$event %in% "output_only" &
    x$event_after %in% c("conc_dose", "conc") &
    !(x$event_before %in% "conc_dose" &
        x$event_after %in% "conc_dose")
}
iecd_interp_value <- function(data_all, current_idx, ...) {
  tmp_conc <- data_all[!is.na(data_all$conc) &
                         data_all$dose_count %in% data_all$dose_count[current_idx],]
  interpolate.conc(conc=tmp_conc$conc, time=tmp_conc$time,
                   time.out=data_all$time[current_idx],
                   check=FALSE, ...)
}

# Extrapolation ####
iecd_extrap_select <- function(x) {
  extrap_output_only <-
    x$event_before %in% c("conc_dose_iv_bolus_after", "conc") &
    x$event %in% "output_only" &
    x$event_after %in% c("dose", "none")
  extrap_dose <-
    x$event_before %in% c("conc_dose_iv_bolus_after", "conc") &
    x$event %in% "dose" &
    !(x$event_after %in% "output_only")
  extrap_output_only | extrap_dose
}
iecd_extrap_value <- function(data_all, current_idx, lambda.z, ...) {
  last_conc <- data_all[data_all$time < data_all$time[current_idx] &
                          !is.na(data_all$conc),]
  last_conc <- last_conc[nrow(last_conc),]
  if (last_conc$conc %in% 0) {
    # BLQ continues to be BLQ
    0
  } else {
    if (missing(lambda.z)) {
      lambda.z <- NA_real_
    }
    args <- list(conc=last_conc$conc[nrow(last_conc)],
                 time=last_conc$time[nrow(last_conc)],
                 time.out=data_all$time[current_idx], lambda.z=lambda.z,
                 ...)
    if (!("clast" %in% names(args))) {
      args$clast <- last_conc$conc[nrow(last_conc)]
    }
    do.call(extrapolate.conc, args)
  }
}

# Immediately after an IV bolus with a concentration next ####
iecd_iv_conc_select <- function(x) {
  !(x$event_before %in% c("conc_dose_iv_bolus_after", "dose_iv_bolus_after", "output_only")) &
    x$event %in% "dose_iv_bolus_after" &
    x$event_after %in% c("conc_dose", "conc")
}
iecd_iv_conc_value <- function(data_all, current_idx, ...) {
  tmp_conc <- data_all[data_all$conc_event &
                         (data_all$dose_count %in% data_all$dose_count[current_idx] |
                            data_all$dose_count_prev %in%  data_all$dose_count[current_idx]),]
  tmp_dose <- data_all[data_all$dose_event &
                         data_all$dose_count %in% data_all$dose_count[current_idx],]
  pk.calc.c0(conc=tmp_conc$conc, time=tmp_conc$time,
             time.dose=tmp_dose$time,
             method=c("logslope", "c1"))
}

# Immediately after an IV bolus without a concentration next ####
iecd_iv_noconc_select <- function(x) {
  bolus_is_current <-
    x$event_before %in% c("conc", "conc_dose", "dose", "none") &
    x$event %in% "dose_iv_bolus_after" &
    x$event_after %in% c("dose", "none")
  bolus_is_previous <-
    x$event_before %in% "dose_iv_bolus_after" &
    x$event %in% "dose" &
    x$event_after %in% c("conc_dose_iv_bolus_after", "conc_dose", "dose_iv_bolus_after", "dose", "conc", "none")
  bolus_is_current | bolus_is_previous
}

# After an IV bolus with a concentration next ####
iecd_afteriv_conc_select <- function(x) {
  x$event_before %in% "dose_iv_bolus_after" &
    x$event %in% "output_only" &
    x$event_after %in% c("conc_dose", "conc")
}
iecd_afteriv_conc_value <- function(data_all, current_idx, ...) {
  tmp_conc <- data_all[data_all$conc_event &
                         (data_all$dose_count %in% data_all$dose_count[current_idx] |
                            data_all$dose_count_prev %in%  data_all$dose_count[current_idx]),
                       c("conc", "time")]
  tmp_dose <- data_all[data_all$dose_event &
                         data_all$dose_count %in% data_all$dose_count[current_idx],]
  tmp_conc <- rbind(
    data.frame(
      time=tmp_dose$time,
      conc=
        pk.calc.c0(conc=tmp_conc$conc, time=tmp_conc$time,
                   time.dose=tmp_dose$time,
                   method=c("logslope", "c1"))),
    tmp_conc)
  interpolate.conc(conc=tmp_conc$conc,
                   time=tmp_conc$time,
                   time.out=data_all$time[current_idx],
                   check=FALSE, ...)
}

# After an IV bolus without a concentration next ####
iecd_afteriv_noconc_select <- function(x) {
  x$event_before %in% "dose_iv_bolus_after" &
    x$event %in% "output_only" &
    x$event_after %in% c("dose", "none")
}

# Doses with no concentrations between ####
iecd_dose_noconc_select <- function(x) {
  dose_current <-
    x$event_before %in% c("conc_dose", "dose") &
    x$event %in% "dose" &
    !(x$event_after %in% "output_only")
  dose_around <-
    x$event_before %in% c("dose", "conc_dose") &
    x$event %in% "output_only" &
    x$event_after %in% c("dose", "conc_dose")
  dose_current | dose_around
}

# Dose as the last event in the timeline and requesting a concentration after ####
iecd_dose_last_select <- function(x) {
  x$event_before %in% c("conc_dose", "dose") &
    x$event %in% "output_only" &
    x$event_after %in% "none"
}

# Dose before, concentration after without a dose ####
iecd_dose_conc_select <- function(x) {
  x$event_before %in% "dose" &
    x$event %in% "output_only" &
    x$event_after %in% "conc"
}
iecd_dose_conc_value <- function(data_all, current_idx, ...) {
  data_tmp <- data_all[data_all$dose_event | data_all$conc_event,]
  interpolate.conc(conc=data_tmp$conc, time=data_tmp$time,
                   time.out=data_all$time[current_idx], ...,
                   check=FALSE)
}

interp.extrap.conc.dose.select <-
  list(
    "Impossible combinations"=
      list(
        select="iecd_impossible_select",
        value="iecd_impossible_value",
        description="The event combination cannot exist."),
    "Observed concentration"=
      list(
        select="iecd_observed_select",
        value="iecd_observed_value",
        description="Copy the input concentration at the given time to the output."),
    "Before all events"=
      list(
        select="iecd_before_select",
        value="iecd_before_value",
        description=paste("Interpolation before any events is NA or zero (0)",
                          "depending on the value of conc.origin.  conc.origin",
                          "defaults to zero which is the implicit assumption",
                          "that a complete washout occurred and there is no",
                          "endogenous source of the analyte.")),
    "Interpolation"=
      list(
        select="iecd_interp_select",
        value="iecd_interp_value",
        description=paste("With concentrations before and after and not an IV",
                          "bolus before, interpolate between observed concentrations.")),
    "Extrapolation"=
      list(
        select="iecd_extrap_select",
        value="iecd_extrap_value",
        description="Extrapolate from a concentration to a dose"),
    "Immediately after an IV bolus with a concentration next"=
      list(
        select="iecd_iv_conc_select",
        value="iecd_iv_conc_value",
        description=paste("Calculate C0 for the time immediately after an IV",
                          "bolus.  First, attempt using log slope",
                          "back-extrapolation.  If that fails, use the first",
                          "concentration after the dose as C0.")),
    "Immediately after an IV bolus without a concentration next"=
      list(
        select="iecd_iv_noconc_select",
        warning="Cannot interpolate immediately after an IV bolus without a concentration next.",
        description="Cannot calculate C0 without a concentration after an IV bolus; return NA."),
    "After an IV bolus with a concentration next"=
      list(
        select="iecd_afteriv_conc_select",
        value="iecd_afteriv_conc_value",
        description=paste("First, calculate C0 using log slope back-extrapolation",
                          "(falling back to the first post-dose concentration",
                          "if that fails).  Then, interpolate between C0 and",
                          "the first post-dose concentration.")),
    "After an IV bolus without a concentration next"=
      list(
        select="iecd_afteriv_noconc_select",
        warning="Cannot interpolate after an IV bolus without a concentration next.",
        description=paste("Between an IV bolus and anything other than a",
                          "concentration, interpolation cannot occur.  Return NA")),
    "Doses with no concentrations between"=
      list(
        select="iecd_dose_noconc_select",
        warning="Cannot interpolate between two doses or after a dose without a concentration after the first dose.",
        description="Two doses with no concentrations between them, return NA."),
    "Dose as the last event in the timeline and requesting a concentration after"=
      list(
        select="iecd_dose_last_select",
        warning="Cannot extrapolate from a dose without any concentrations after it.",
        description=paste("Cannot estimate the concentration after a dose",
                          "without concentrations after the dose, return NA.")),
    "Dose before, concentration after without a dose"=list(
      select="iecd_dose_conc_select",
      value="iecd_dose_conc_value",
      description="If the concentration at the dose is estimable, interpolate.  Otherwise, NA.")
  )

# expected errors ####

## interpolate.conc expected errors ####

test_that("interpolate.conc expected errors", {
  # Ensure that it is an error to extrapolate past the last point.
  expect_error(
    interpolate.conc(
      conc=c(0, 1, 0.5, 1, 0),
      time=0:4,
      time.out=5,
      method="lin up/log down"
    ),
    regexp="`interpolate.conc()` does not extrapolate, use `interp.extrap.conc()`",
    fixed=TRUE
  )
  # Change documentation when this is not true
  expect_error(
    interpolate.conc(
      conc=0:1,
      time=0:1,
      time.out=0:1
    ),
    regexp="Assertion on 'time.out' failed: Must have length 1."
  )

  expect_error(
    interpolate.conc(
      conc=0:1,
      time=0:1,
      time.out=0.5,
      method="this doesn't work"
    ),
    regexp=tryCatch(expr={
      match.arg("foo", choices=c("lin up/log down", "linear"))
    },
    error=function(e) e)$message,
    fixed=TRUE,
    info="Confirm that invalid interpolation methods are an error."
  )

  expect_error(
    interpolate.conc(
      conc=0:1,
      time=0:1,
      time.out=0.5,
      conc.origin=1:2
    ),
    regexp="Assertion on 'conc.origin' failed: Must have length 1."
  )
  expect_error(
    interpolate.conc(
      conc=0:1,
      time=0:1,
      time.out=0.5,
      conc.origin="A"
    ),
    regexp="Assertion on 'conc.origin' failed: Must be of type 'number', not 'character'."
  )
  expect_error(
    interpolate.conc(
      conc=0:1,
      time=0:1,
      time.out=0.5,
      conc.origin=factor("A")
    ),
    regexp="Assertion on 'conc.origin' failed: Must be of type 'number', not 'factor'."
  )

  expect_error(
    interpolate.conc(
      conc=c(NA, 1),
      time=c(0, 1),
      time.out=0.5,
      check=FALSE
    )
  )

  # When tlast is before the end of the data, do not extrapolate
  expect_error(
    interpolate.conc(
      conc=c(1, 2, 0),
      time=0:2,
      time.out=1.5
    ),
    regexp="`interpolate.conc()` can only works through Tlast, please use `interp.extrap.conc()` to combine both interpolation and extrapolation.",
    fixed=TRUE
  )
})

## extrapolate.conc expected errors ####

test_that("extrapolate.conc expected errors", {
  # Confirm that the auc.type must be AUCinf, AUClast, AUCall
  expect_error(
    extrapolate.conc(
      conc=1,
      time=1,
      time.out=2,
      auc.type="wrong"
    ),
    regexp="`auc.type` must be one of 'AUCinf', 'AUClast', or 'AUCall'"
  )

  # Confirm that time.out may only be a scalar
  expect_error(
    extrapolate.conc(
      conc=1,
      time=1,
      time.out=c(2, 3),
      auc.type="AUCinf"
    ),
    regexp="Only one time.out value may be estimated at once."
  )
  # Confirm that it is an error to extrapolate at or before Tlast
  expect_error(
    extrapolate.conc(
      conc=1,
      time=1,
      time.out=0.5,
      auc.type="AUCinf"
    ),
    regexp="extrapolate.conc can only work beyond Tlast, please use interp.extrap.conc to combine both interpolation and extrapolation."
  )
  expect_error(
    extrapolate.conc(
      conc=1,
      time=1,
      time.out=1,
      auc.type="AUCinf"
    ),
    regexp="extrapolate.conc can only work beyond Tlast, please use interp.extrap.conc to combine both interpolation and extrapolation."
  )
})

## interp.extrap.conc expected errors ####

test_that("interp.extrap.conc expected errors", {
  # Check inputs
  expect_error(
    interp.extrap.conc(
      conc=c(0, 1, 0.5, 1, 0),
      time=0:4,
      time.out=numeric(),
      method="lin up/log down"
    ),
    regexp="time.out must be a vector with at least one element"
  )
})

## interp.extrap.conc.dose expected errors ####

test_that("interp.extrap.conc.dose expected errors", {
  # Check inputs
  expect_error(
    interp.extrap.conc.dose(
      conc=c(0, 1, 2, 1, 0.5, 0.25),
      time=c(-1, 1:5),
      time.dose=0,
      route.dose="foo",
      duration.dose=NA,
      time.out=c(-1, -0.1, 0, 0.1, 7),
      out.after=FALSE
    ),
    regexp="route.dose must be either 'extravascular' or 'intravascular'",
    info="Route must be valid"
  )
  expect_error(
    interp.extrap.conc.dose(
      conc=c(0, 1, 2, 1, 0.5, 0.25),
      time=c(-1, 1:5),
      time.dose=0,
      route.dose=factor("foo"),
      duration.dose=NA,
      time.out=c(-1, -0.1, 0, 0.1, 7),
      out.after=FALSE
    ),
    regexp="route.dose must be either 'extravascular' or 'intravascular'",
    info="Route must be valid (as a factor)"
  )
  expect_error(
    interp.extrap.conc.dose(
      conc=c(0, 1, 2, 1, 0.5, 0.25),
      time=c(-1, 1:5),
      time.dose=0,
      route.dose=c("extravascular", "extravascular"),
      duration.dose=NA,
      time.out=c(-1, -0.1, 0, 0.1, 7),
      out.after=FALSE
    ),
    regexp="route.dose must either be a scalar or the same length as time.dose",
    info="Route must have the correct length"
  )
  expect_error(
    interp.extrap.conc.dose(
      conc=c(0, 1, 2, 1, 0.5, 0.25),
      time=c(-1, 1:5),
      time.dose=0,
      route.dose="extravascular",
      duration.dose="A",
      time.out=c(-1, -0.1, 0, 0.1, 7),
      out.after=FALSE
    ),
    regexp="duration.dose must be NA or a number.",
    info="duration.dose must be NA or a number (character)."
  )
  expect_error(
    interp.extrap.conc.dose(
      conc=c(0, 1, 2, 1, 0.5, 0.25),
      time=c(-1, 1:5),
      time.dose=0,
      route.dose="extravascular",
      duration.dose=factor("A"),
      time.out=c(-1, -0.1, 0, 0.1, 7),
      out.after=FALSE
    ),
    regexp="duration.dose must be NA or a number.",
    info="duration.dose must be NA or a number (factor)."
  )
  expect_error(
    interp.extrap.conc.dose(
      conc=c(0, 1, 2, 1, 0.5, 0.25),
      time=c(-1, 1:5),
      time.dose=0,
      route.dose="extravascular",
      duration.dose=c(1, NA),
      time.out=c(-1, -0.1, 0, 0.1, 7),
      out.after=FALSE
    ),
    regexp="duration.dose must either be a scalar or the same length as time.dose",
    info="duration.dose must match the length of time.dose or be a scalar."
  )
})

# interpolate.conc ####

test_that("interpolate.conc", {
  # Confirm that interpolating between all zeros results in zero
  interp_method <- c("linear", "lin up/log down")
  for (n in interp_method) {
    expect_equal(
      interpolate.conc(
        conc=c(0, 0),
        time=0:1,
        time.out=0,
        method=n
      ),
      0,
      info=paste("all zero", n)
    )
  }

  # Confirm that interpolating to a given point results in the point itself
  interpolations <-
    list(
      linear=0,
      "lin up/log down"=0
    )
  for (n in names(interpolations)) {
    expect_equal(
      interpolate.conc(
        conc=c(0, 1),
        time=0:1,
        time.out=0,
        method=n
      ),
      interpolations[[n]],
      info=n
    )
  }

  interpolations <-
    list(
      linear=1,
      "lin up/log down"=1
    )
  for (n in names(interpolations)) {
    expect_equal(
      interpolate.conc(
        conc=c(0, 1),
        time=0:1,
        time.out=1,
        method=n
      ),
      interpolations[[n]],
      info=n
    )
  }

  # Confirm that interpolating to a given point when that point is NA gives a
  # interpolates as expected.
  interpolations <-
    list(
      linear=1,
      "lin up/log down"=1
    )
  for (n in names(interpolations)) {
    expect_equal(
      interpolate.conc(
        conc=c(0, 1, NA, 1),
        time=0:3,
        time.out=2,
        method=n
      ),
      interpolations[[n]],
      info=n
    )
  }

  # Slightly less trivial tests
  interpolations <-
    list(
      linear=0.5,
      "lin up/log down"=0.5
    )
  for (n in names(interpolations)) {
    expect_equal(
      interpolate.conc(
        conc=c(0, 1, 0),
        time=0:2,
        time.out=0.5,
        method=n
      ),
      interpolations[[n]],
      info=n
    )
  }

  interpolations <-
    list(
      linear=1.75,
      "lin up/log down"=2^0.75
    )
  for (n in names(interpolations)) {
    expect_equal(
      interpolate.conc(
        conc=c(0, 2, 1),
        time=0:2,
        time.out=1.25,
        method=n
      ),
      interpolations[[n]],
      info=n
    )
  }
  # Relative time is the only thing that matters.  Negative time
  # values are fine.
  interpolations <-
    list(
      linear=1.75,
      "lin up/log down"=2^0.75
    )
  for (n in names(interpolations)) {
    expect_equal(
      interpolate.conc(
        conc=c(0, 2, 1),
        time=seq(-10, -8, by=1),
        time.out=-8.75,
        method=n
      ),
      interpolations[[n]],
      info=n
    )
  }
  # Ensure that linear-up/log down interpolation works linearly for the "up"
  # even if it is after a down.
  interpolations <-
    list(
      linear=0.25,
      "lin up/log down"=0.25
    )
  for (n in names(interpolations)) {
    expect_equal(
      interpolate.conc(
        conc=c(0, 1, 0, 1, 0),
        time=0:4,
        time.out=2.25,
        conc.blq="keep",
        method=n
      ),
      interpolations[[n]],
      info=n
    )
  }

  # Ensure that interpolation with linear works with a BLQ in the middle: going
  # down and up from that BLQ
  interpolations <-
    list(
      linear=0.5,
      "lin up/log down"=0.5
    )
  for (n in names(interpolations)) {
    expect_equal(
      interpolate.conc(
        conc=c(0, 1, 0, 1, 0),
        time=0:4,
        time.out=1.5,
        conc.blq="keep",
        method=n
      ),
      interpolations[[n]],
      info=n
    )
  }

  interpolations <-
    list(
      linear=0.5,
      "lin up/log down"=0.5
    )
  for (n in names(interpolations)) {
    expect_equal(
      interpolate.conc(
        conc=c(0, 1, 0, 1, 0),
        time=0:4,
        time.out=2.5,
        conc.blq="keep",
        method=n
      ),
      interpolations[[n]],
      info=n
    )
  }

  # Ensure that conc.blq is appropriately observed
  interpolations <-
    list(
      linear=1,
      "lin up/log down"=1
    )
  for (n in names(interpolations)) {
    expect_equal(
      interpolate.conc(
        conc=c(0, 1, 0, 1, 0),
        time=0:4,
        time.out=2.25,
        conc.blq="drop",
        method=n
      ),
      interpolations[[n]],
      info=n
    )
  }
  # Ensure that the "log down" part works, too.
  expect_equal(
    interpolate.conc(
      conc=c(0, 1, 0.5, 1, 0),
      time=0:4,
      time.out=1.5,
      method="lin up/log down"),
    exp(mean(log(c(1, 0.5))))
  )

  # Confirm that extrapolating before the first time uses conc.origin
  expect_equal(
    interpolate.conc(
      conc=0:1,
      time=0:1,
      time.out=-1),
    0,
    info="conc.origin defaults to zero"
  )
  expect_equal(
    interpolate.conc(
      conc=0:1,
      time=0:1,
      time.out=-1,
      conc.origin=NA),
    NA,
    info="conc.origin is honored as NA"
  )
  expect_equal(
    interpolate.conc(
      conc=0:1,
      time=0:1,
      time.out=-1,
      conc.origin=5
    ),
    5,
    info="conc.origin is honored as a number"
  )

  # expect_equal(
  #   interpolate.conc(
  #     conc=c(NA, 1, 2),
  #     time=c(0, 1, 2),
  #     time.out=1.5,
  #     check=FALSE
  #   ),
  #   1.5,
  #   info="Skipping the checks with an NA, but not bounding the interpolation gives the expected value."
  # )

  # by default, zeros in the middle are omitted
  expect_equal(
    interpolate.conc(
      conc=c(0, 1, 2, 0, 1, 0),
      time=0:5,
      time.out=3,
      conc.blq=NULL
    ),
    exp(mean(log(c(2, 1))))
  )
  # Zeros in the middle are accurately interpolated
  expect_equal(
    interpolate.conc(
      conc=c(0, 1, 2, 0, 1, 0),
      time=0:5,
      time.out=2.5,
      conc.blq=0
    ),
    1
  )
  # interpolation works with the "zero" method of interpolation
  expect_equal(
    interpolate.conc(
      conc=c(0, 1, 2, 0, 0, 1, 0),
      time=0:6,
      time.out=3.5,
      conc.blq=0
    ),
    0
  )
})

# extrapolate.conc ####

test_that("extrapolate.conc", {
  # Ensure that if Clast is NA that the extrapolated concentration is NA.
  expect_warning(
    v1 <-
      extrapolate.conc(
        conc=NA,
        time=1,
        time.out=2,
        auc.type="AUCinf"
      ),
    regexp="All concentration data are missing"
  )
  expect_equal(v1, NA)

  # Extrapolating AUClast is always 0
  expect_equal(
    extrapolate.conc(
      conc=c(0, 1, 0),
      time=1:3,
      time.out=4,
      auc.type="AUClast"
      ),
    0
  )
  expect_equal(
    extrapolate.conc(
      conc=c(0, 1, 0),
      time=1:3,
      time.out=2.5,
      auc.type="AUClast"
    ),
    0
  )

  # Extrapolating AUCall after the last value and it is 0 is 0
  expect_equal(
    extrapolate.conc(
      conc=c(0, 1, 0),
      time=1:3,
      time.out=4,
      auc.type="AUCall"
    ),
    0
  )

  # Extrapolating AUCall after the last value and it is nonzero is 0
  expect_equal(
    extrapolate.conc(
      conc=c(0, 1, 1),
      time=1:3,
      time.out=4,
      lambda.z=2,
      auc.type="AUCall"
    ),
    0
  )

  # Extrapolating AUCall between Tlast and the first BLQ is a linear
  # interpolation.
  expect_equal(
    extrapolate.conc(
      conc=c(0, 1, 0),
      time=1:3,
      time.out=2.5,
      auc.type="AUCall"
    ),
    0.5
  )

  # Extrapolating AUCinf between the Clast value and a 0 ignores the
  # 0.
  expect_equal(
    extrapolate.conc(
      conc=c(0, 1, 0),
      time=1:3,
      time.out=2.5,
      lambda.z=1,
      auc.type="AUCinf"
    ),
    1*exp(-1*0.5)
  )

  # A slightly less trivial example
  expect_equal(
    extrapolate.conc(
      conc=c(0, 5, 0),
      time=1:3,
      time.out=2.5,
      lambda.z=3,
      auc.type="AUCinf"
    ),
    5*exp(-3*0.5)
  )

  # Extrapolating AUCinf with lambda.z=NA is NA.
  expect_equal(
    extrapolate.conc(
      conc=c(0, 5, 0),
      time=1:3,
      time.out=2.5,
      lambda.z=NA,
      auc.type="AUCinf"
    ),
    as.numeric(NA)
  )

  # Extrapolating with all NA is NA.
  expect_warning(
    expect_equal(
      extrapolate.conc(
        conc=rep(NA, 3),
        time=1:3,
        time.out=2.5,
        lambda.z=NA,
        auc.type="AUCinf"
      ),
      NA
    ),
    class = "pknca_conc_all_missing"
  )

  # Ensure that extrapolation beyond the last point works if the last point is 0
  extrapolations <-
    list(
      AUCinf=exp(-2),
      AUCall=0,
      AUClast=0
    )
  for (n in names(extrapolations)) {
    expect_equal(
      extrapolate.conc(
        conc=c(0, 1, 0.5, 1, 0),
        time=0:4, 5,
        lambda.z=1,
        auc.type=n
      ),
      extrapolations[[n]],
      info=n
    )
  }

  # Ensure that extrapolation to the last point of 0 works
  extrapolations <-
    list(
      AUCinf=exp(-1),
      AUCall=0,
      AUClast=0
    )
  for (n in names(extrapolations)) {
    expect_equal(
      extrapolate.conc(
        conc=c(0, 1, 0.5, 1, 0),
        time=0:4, 4,
        lambda.z=1,
        auc.type=n
      ),
      extrapolations[[n]],
      info=n
    )
  }

  # Ensure that extrapolation between the last points with the last point 0 works
  extrapolations <-
    list(
      AUCinf=exp(-0.5),
      AUCall=0.5,
      AUClast=0
    )
  for (n in names(extrapolations)) {
    expect_equal(
      extrapolate.conc(
        conc=c(0, 1, 0.5, 1, 0),
        time=0:4, 3.5,
        lambda.z=1,
        auc.type=n
      ),
      extrapolations[[n]],
      info=n
    )
  }
})

test_that("extrapolate.conc with AUCall goes down over time (#249)", {
  expect_equal(
    extrapolate.conc(
      conc = c(1, 0), time = 0:1,
      time.out = 0.75,
      clast = 1,
      auc.type = "AUCall"
    ),
    0.25
  )
})

# interp.extrap.conc ####

test_that("interp.extrap.conc", {
  # All zeros always interpolates to all zeros
  expect_equal(
    interp.extrap.conc(
      conc=c(0, 0, 0),
      time=0:2,
      time.out=c(0, 0.5, 2.5),
      method="lin up/log down"
    ),
    rep(0, 3)
  )

  # Ensure that data checking works correctly
  expect_equal(
    interp.extrap.conc(
      conc=c(0, 1, 0.5, 1, 0),
      time=0:4,
      time.out=1.5,
      method="lin up/log down"
    ),
    exp(mean(log(c(1, 0.5))))
  )

  # Ensure that NA for time.out results in NA output
  expect_warning(
    expect_equal(
      interp.extrap.conc(
        conc=c(0, 1, 0.5, 1, 0),
        time=0:4,
        time.out=c(1.5, NA),
        method="lin up/log down"
      ),
      c(exp(mean(log(c(1, 0.5)))), NA)
    )
  )

  # Ensure a warning with NA for time.out
  expect_warning(
    interp.extrap.conc(
      conc=c(0, 1, 0.5, 1, 0),
      time=0:4,
      time.out=c(1.5, NA),
      method="lin up/log down"
    ),
    regexp="An interpolation/extrapolation time is NA"
  )
})

# interp.extrap.conc.dose ####

test_that("interp.extrap.conc.dose handles all eventualities", {
  event_choices <- unlist(event_choices_interp.extrap.conc.dose, use.names=FALSE)
  eventualities <-
    expand.grid(
      event_before=setdiff(event_choices, "output_only"),
      event=setdiff(event_choices, "none"),
      event_after=setdiff(event_choices, "output_only")
    )
  eventualities$method <- NA_character_
  for (nm in names(interp.extrap.conc.dose.select)) {
    mask_selected <-
      do.call(
        interp.extrap.conc.dose.select[[nm]]$select,
        list(x=eventualities)
      )
    expect_true(
      any(mask_selected),
      info=sprintf("interp.extrap.conc.dose.select[[%s]] matched at least one eventuality", nm)
    )
    expect_true(
      !any(mask_selected & !is.na(eventualities$method)),
      info=sprintf("interp.extrap.conc.dose.select[[%s]] overlapped with another method.", nm)
    )
    eventualities$method[mask_selected] <- nm
    # if (interactive()) {
    #   cat(sprintf("%d/%d rows filled at %s\n", sum(!is.na(eventualities$method)), nrow(eventualities), nm))
    # }
  }
  expect_false(
    any(is.na(eventualities$method)),
    info="interp.extrap.conc.dose.select matched all eventualities"
  )
})

test_that("interp.extrap.conc.dose", {
  expect_equal(
    interp.extrap.conc.dose(
      conc=c(0, 1, 2, 1, 0.5, 0.25),
      time=c(-1, 1:5),
      time.dose=0,
      time.out=-2,
      check=FALSE
    ),
    interp.extrap.conc.dose(
      conc=c(0, 1, 2, 1, 0.5, 0.25),
      time=c(-1, 1:5),
      time.dose=0,
      time.out=-2
    ),
    info="Check is respected"
  )

  expect_equal(
    interp.extrap.conc.dose(
      conc=c(0, 1, 2, 1, 0.5, 0.25),
      time=c(-1, 1:5),
      time.dose=0,
      time.out=-2
    ),
    structure(0, Method="Before all events"),
    info="Interpolation before all events yields conc.origin which defaults to zero."
  )

  expect_equal(
    interp.extrap.conc.dose(
      conc=c(0, 1, 2, 1, 0.5, 0.25),
      time=c(-1, 1:5),
      time.dose=0,
      conc.origin=NA,
      time.out=-2),
    structure(NA_real_, Method="Before all events"),
    info="Interpolation before all events yields conc.origin respecting its input."
  )

  expect_equal(
    interp.extrap.conc.dose(
      conc=c(0, 1, 2, 1, 0.5, 0.25),
      time=c(-1, 1:5),
      time.dose=0,
      time.out=-1
    ),
    structure(0, Method="Observed concentration"),
    info="When there is a concentration measurement at a time point, it is returned."
  )

  expect_equal(
    interp.extrap.conc.dose(
      conc=c(0, 1, 2, 1, 0.5, 0.25),
      time=c(-1, 1:5),
      time.dose=0,
      time.out=-0.1
    ),
    structure(0, Method="Extrapolation"),
    info="When the previous measurement is zero and there is no dose between, it is returned."
  )

  expect_equal(
    interp.extrap.conc.dose(
      conc=c(0, 1, 2, 1, 0.5, 0.25),
      time=c(-1, 1:5),
      time.dose=0,
      time.out=0),
    structure(0, Method="Extrapolation"),
    info="When the previous measurement is zero it is at the time of the dose, zero is returned."
  )

  expect_equal(
    interp.extrap.conc.dose(
      conc=c(0, 1, 2, 1, 0.5, 0.25),
      time=c(-1, 1:5),
      time.dose=0,
      time.out=0.1
    ),
    structure(0.1, Method="Dose before, concentration after without a dose"),
    info="Extrapolation to a dose then interpolation between the dose and the next time works."
  )

  # expect_equal(
  #   expect_warning(
  #       interp.extrap.conc.dose(
  #         conc=c(0, 1, 2, 1, 0.5, 0.25),
  #         time=c(-1, 1:5),
  #         time.dose=c(0, 0.1),
  #         time.out=0.2
  #       ),
  #       regexp="Cannot interpolate between two doses or after a dose without a concentration after the first dose.",
  #       fixed=TRUE,
  #       info="Two doses in a row generates a warning"
  #   ),
  #   structure(NA_real_, Method="Dose before, concentration after without a dose"),
  #   info="Extrapolation to a dose then interpolation between the dose and the next time gives NA when the dose is NA."
  # )

  expect_equal(
    interp.extrap.conc.dose(
      conc=c(0, 1, 2, 1, 0.5, 0.25),
      time=c(-1, 1:5),
      time.dose=0,
      time.out=5
    ),
    structure(0.25, Method="Observed concentration"),
    info="Copy from after the dose."
  )

  expect_equal(
    interp.extrap.conc.dose(
      conc=c(0, 1, 2, 1, 0.5, 0.25),
      time=c(-1, 1:5),
      time.dose=0,
      time.out=7
    ),
    structure(NA_real_, Method="Extrapolation"),
    info="Extrapolation without lambda.z gives NA result"
  )

  expect_equal(
    interp.extrap.conc.dose(
      conc=c(0, 1, 2, 1, 0.5, 0.25),
      time=c(-1, 1:5),
      time.dose=0,
      time.out=7,
      lambda.z=log(2)
    ),
    structure(0.0625, Method="Extrapolation"),
    info="Extrapolation with lambda.z gives result"
  )

  expect_equal(
    interp.extrap.conc.dose(
      conc=0:2,
      time=0:2,
      time.dose=0,
      time.out=0.5
    ),
    structure(0.5, Method="Interpolation"),
    info="Interpolation works"
  )

  expect_equal(
    interp.extrap.conc.dose(
      conc=c(0:2, 1),
      time=0:3,
      time.dose=0,
      time.out=2.5,
      method="linear"
    ),
    structure(1.5, Method="Interpolation"),
    info="Interpolation respects method"
  )

  expect_equal(
    interp.extrap.conc.dose(
      conc=0:2,
      time=0:2,
      time.dose=0,
      route.dose="intravascular",
      time.out=0,
      duration.dose=0,
      out.after=FALSE
    ),
    structure(0, Method="Observed concentration"),
    info="Observed before IV bolus"
  )

  expect_equal(
    interp.extrap.conc.dose(
      conc=0:2,
      time=0:2,
      time.dose=0,
      route.dose="intravascular",
      time.out=0,
      duration.dose=0,
      out.after=TRUE
    ),
    structure(1, Method="Immediately after an IV bolus with a concentration next"),
    info="Observed after IV bolus, one concentration"
  )

  expect_equal(
    interp.extrap.conc.dose(
      conc=c(0, 2, 1),
      time=0:2,
      time.dose=0,
      route.dose="intravascular",
      time.out=0,
      duration.dose=0,
      out.after=TRUE
    ),
    structure(4, Method="Immediately after an IV bolus with a concentration next"),
    info="Observed after IV bolus, two concentrations"
  )

  expect_equal(
    interp.extrap.conc.dose(
      conc=c(2, 1),
      time=1:2,
      time.dose=0,
      route.dose="intravascular",
      time.out=0.5,
      duration.dose=0
    ),
    structure(2*sqrt(2), Method="After an IV bolus with a concentration next"),
    info="After IV bolus, two concentrations"
  )

  expect_equal(
    interp.extrap.conc.dose(
      conc=2,
      time=1,
      time.dose=0,
      route.dose="intravascular",
      time.out=0.5,
      duration.dose=0
    ),
    structure(2, Method="After an IV bolus with a concentration next"),
    info="After IV bolus, one concentration"
  )

  expect_equal(
    interp.extrap.conc.dose(
      conc=c(0, 1, 2, 1, 0.5, 0.25),
      time=c(-1, 1:5),
      time.dose=0,
      time.out=c(-2, 2)
      ),
    structure(
      c(0, 2),
      Method=c("Before all events", "Observed concentration")
    ),
    info="Outputs are in the same order as inputs (initially sorted)"
  )

  expect_equal(
    interp.extrap.conc.dose(
      conc=c(0, 1, 2, 1, 0.5, 0.25),
      time=c(-1, 1:5),
      time.dose=0,
      time.out=c(2, -2)
    ),
    structure(
      c(2, 0),
      Method=c("Observed concentration", "Before all events")
    ),
    info="Outputs are in the same order as inputs (reverse sorted time.out)"
  )
})

#' Compute the time to steady state using nonlinear, mixed-effects modeling of
#' trough concentrations.
#'
#' Trough concentrations are selected as concentrations at the time of dosing.
#' An exponential curve is then fit through the data with a different magnitude
#' by treatment (as a factor) and a random steady-state concentration and time
#' to stead-state by subject (see `random.effects` argument).
#'
#' @param \dots See [pk.tss.data.prep()]
#' @param tss.fraction The fraction of steady-state required for calling
#'   steady-state
#' @param output Which types of outputs should be produced? `population` is the
#'   population estimate for time to steady-state (from an nlme model), `popind`
#'   is the individual estimate (from an nlme model), `individual` fits each
#'   individual separately with a gnls model (requires more than one individual;
#'   use `single` for one individual), and `single` fits all the data to a
#'   single gnls model.
#' @param check See [pk.tss.data.prep()].
#' @param verbose Describe models as they are run, show convergence of the model
#'   (passed to the nlme function), and additional details while running.
#'
#' @returns A scalar float for the first time when steady-state is achieved or
#'   `NA` if it is not observed.
#' @family Time to steady-state calculations
#' @references
#' Maganti, L., Panebianco, D.L. & Maes, A.L. Evaluation of Methods for
#' Estimating Time to Steady State with Examples from Phase 1 Studies. AAPS J
#' 10, 141–147 (2008). https://doi.org/10.1208/s12248-008-9014-y
#' @export
pk.tss.monoexponential <- function(...,
                                   tss.fraction=0.9,
                                   output=c(
                                     "population",
                                     "popind",
                                     "individual",
                                     "single"),
                                   check=TRUE,
                                   verbose=FALSE) {
  # Check inputs
  modeldata <- pk.tss.data.prep(..., check=check)
  if (is.factor(tss.fraction) |
      !is.numeric(tss.fraction))
    stop("tss.fraction must be a number")
  if (!length(tss.fraction) == 1) {
    warning("Only first value of tss.fraction is being used")
    tss.fraction <- tss.fraction[1]
  }
  if (tss.fraction <= 0 | tss.fraction >= 1) {
    stop("tss.fraction must be between 0 and 1, exclusive")
  } else if (tss.fraction < 0.8) {
    warning("tss.fraction is usually >= 0.8")
  }
  # Note that this will by default choose "population" if nothing is
  # requested.
  output <- match.arg(output, several.ok=TRUE)
  if (!("subject" %in% names(modeldata))) {
    if (any(c("population", "popind", "individual") %in% output)) {
      warning("Cannot give 'population', 'popind', or 'individual' ",
              "output without multiple subjects of data")
      output <- setdiff(output, c("population", "popind", "individual"))
    }
  }
  # Set the tss.constant so that exp(tss.constant) == tss.fraction so
  # that the model below solves for the requested tss.
  modeldata$tss.constant <- log(1-tss.fraction)
  ret_population <-
    if (any(c("population", "popind") %in% output)) {
      pk.tss.monoexponential.population(
        modeldata,
        output=intersect(c("population", "popind"), output),
        verbose=verbose)
    } else {
      NA
    }
  ret_individual <-
    if (any(c("individual", "single") %in% output)) {
      pk.tss.monoexponential.individual(
        modeldata,
        output=intersect(c("individual", "single"), output),
        verbose=verbose)
    } else {
      NA
    }
  ret <-
    if (!identical(NA, ret_population) & !identical(NA, ret_individual)) {
      merge(ret_population, ret_individual)
    } else if (!identical(NA, ret_population)) {
      ret_population
    } else if (!identical(NA, ret_individual)) {
      ret_individual
    } else {
      stop("Error in selection of return values for pk.tss.monoexponential.  This is likely a bug.") # nocov
    }
  ret
}

#' A helper function to generate the formula and starting values for the
#' parameters in monoexponential models.
#'
#' @param data The data used for the model
#' @returns a list with elements for each of the variables
tss.monoexponential.generate.formula <- function(data) {
  # Setup the correct ctrough.ss by treatment or not.
  if ("treatment" %in% names(data)) {
    ctrough_by <-
      list(
        "Ctrough.ss by treatment"=
          list(
            formula=ctrough.ss~treatment-1,
            # Set the starting values for the ctrough.ss as the mean
            # concentration by treatment
            start=dplyr::summarize(
              dplyr::grouped_df(data, vars = "treatment"),
              dplyr::across(.cols=dplyr::all_of("conc"), .fns = mean)
            )$conc
          )
      )
  } else {
    ctrough_by <-
      list("Single Ctrough.ss"=list(
        formula=ctrough.ss~1,
        # Set the starting values for the ctrough.ss as the mean
        # concentration.
        start=mean(data$conc)))
  }
  # Try all combinations of Tss by treatment and single value
  tss.by <- list(
    "Single value for Tss"=list(
      formula=tss~1,
      start=stats::median(unique(data$time))))
  if ("treatment" %in% names(data))
    tss.by[["Tss by treatment"]] <- list(
      formula=tss~treatment,
      start=rep(stats::median(unique(data$time)),
        length(unique(data$treatment))))
  # Try combinations of random effects with Ctrough.ss and
  # Tss. (These are returned even if they aren't always used)
  ranef.by <-
    list("Ctrough.ss and Tss"=list(formula=ctrough.ss+tss~1|subject),
         "Tss"=list(formula=tss~1|subject),
         "Ctrough.ss"=list(formula=ctrough.ss~1|subject))
  # Return the list of parameters to test
  list(
    ctrough.by=ctrough_by,
    tss.by=tss.by,
    ranef.by=ranef.by
  )
}

#' A helper function to estimate population and popind outputs for
#' monoexponential time to steady-state.
#'
#' This function is not intended to be called directly.  Please use
#' `pk.tss.monoexponential`.
#'
#' If no model converges, then the `tss.monoexponential.population` column will
#' be set to NA. If the best model does not include a random effect for subject
#' on Tss then the `tss.monoexponential.popind` column of the output will be set
#' to NA.
#'
#' @param data a data frame as prepared by [pk.tss.data.prep()].  It must
#'   contain at least columns for `subject`, `time`, `conc`, and `tss.constant`.
#' @param output a character vector requesting the output types.
#' @param verbose Show verbose output.
#' @returns A data frame with either one row (if `population` output is
#'   provided) or one row per subject (if `popind` is provided).  The columns
#'   will be named `tss.monoexponential.population` and/or
#'   `tss.monoexponential.popind`.
pk.tss.monoexponential.population <- function(data,
                                              output=c(
                                                "population",
                                                "popind"),
                                              verbose=FALSE) {
  output <- match.arg(output, several.ok=TRUE)
  test.formula <- tss.monoexponential.generate.formula(data)
  # Loop over creating all models
  models <- list()
  for (myctrough.ss in names(test.formula$ctrough.by)) {
    for (mytss in names(test.formula$tss.by)) {
      for (myranef in names(test.formula$ranef.by)) {
        # Describe the current model
        current.desc <-
          sprintf("Fixed effects of %s and %s; random effects of %s",
                  myctrough.ss, mytss, myranef)
        if (verbose)
          print(current.desc)
        # Be prepared for failure
        current.model <- NA
        current.aic <- NA
        current.model.summary <- "Did not converge"
        try({
          # Test the current model
          current.model <-
            nlme::nlme(conc~ctrough.ss*(1-exp(tss.constant*time/tss)),
                       fixed=list(
                         test.formula$ctrough.by[[myctrough.ss]]$formula,
                         test.formula$tss.by[[mytss]]$formula),
                       random=test.formula$ranef.by[[myranef]]$formula,
                       start=c(
                         test.formula$ctrough.by[[myctrough.ss]]$start,
                         test.formula$tss.by[[mytss]]$start),
                       data=data,
                       verbose=verbose)
          # If the model converges, get the summary and AIC out.
          if (!is.null(current.model)) {
            current.model.summary <- summary(current.model)
            current.aic <- stats::AIC(current.model)
          }
        }, silent=!verbose)
        # Put the current model (or model attempt) into the list of
        # models.
        models <-
          append(
            models,
            list(
              list(
                desc=current.desc,
                model=current.model,
                summary=current.model.summary,
                AIC=current.aic
              )
            )
          )
      }
    }
  }
  # Find the best model of the set and return the output from that one.
  all.model.summary <- AIC.list(lapply(models, function(x) x$model))
  rownames(all.model.summary) <-
    vapply(
      X = models,
      FUN = function(x) x$desc,
      FUN.VALUE = ""
    )
  if (verbose)
    print(all.model.summary)
  if (all(is.na(all.model.summary$AIC)) |
      length(all.model.summary) == 0) {
    warning("No population model for monoexponential Tss converged, no results given")
    ret <-
      data.frame(
        tss.monoexponential.population=NA,
        tss.monoexponential.popind=NA,
        subject=unique(data[["subject"]]),
        stringsAsFactors=FALSE
      )
  } else {
    best.model <-
      models[all.model.summary$AIC %in%
             min(all.model.summary$AIC, na.rm=TRUE)][[1]]$model
    ret <-
      data.frame(
        tss.monoexponential.population=nlme::fixef(best.model)[["tss"]],
        stringsAsFactors=FALSE
      )
    best.ranef <- nlme::ranef(best.model)
    if ("tss" %in% names(best.ranef)) {
      ret <-
        merge(
          ret,
          data.frame(
            tss.monoexponential.popind=(best.ranef$tss +
                                          ret$tss.monoexponential.population),
            subject=factor(rownames(best.ranef)),
            stringsAsFactors=FALSE
          ),
          all=TRUE
        )
    } else if ("popind" %in% output) {
      warning("tss.monoexponential.popind was requested, but the best model did not include a random effect for tss.  Set to NA.")
      ret <-
        merge(
          ret,
          data.frame(
            tss.monoexponential.popind=NA,
            subject=unique(data$subject),
            stringsAsFactors=FALSE
          ),
          all=TRUE
        )
    }
  }
  # Return the requested columns
  ret[,intersect(c("subject", "treatment",
                   paste("tss.monoexponential", output, sep=".")),
                 names(ret)),
      drop=FALSE]
}

#' A helper function to estimate individual and single outputs for
#' monoexponential time to steady-state.
#'
#' This function is not intended to be called directly.  Please use
#' `pk.tss.monoexponential`.
#'
#' If no model converges, then the `tss.monoexponential.single` and/or
#' `tss.monoexponential.individual` column will be set to NA.
#'
#' @param data a data frame as prepared by [pk.tss.data.prep()].  It must
#'   contain at least columns for `subject`, `time`, `conc`, and `tss.constant`.
#' @param output a character vector requesting the output types.
#' @param verbose Show verbose output.
#' @returns A data frame with either one row (if `population` output is
#'   provided) or one row per subject (if `popind` is provided).  The columns
#'   will be named `tss.monoexponential.population` and/or
#'   `tss.monoexponential.popind`.
#' @importFrom rlang .data
pk.tss.monoexponential.individual <- function(data,
                                              output=c(
                                                "individual",
                                                "single"),
                                              verbose=FALSE) {
  fit_tss <- function(d) {
    current_model <-
      try({
        nlme::gnls(
          conc~ctrough.ss*(1-exp(tss.constant*time/tss)),
          params=list(
            ctrough.ss~1,
            tss~1),
          start=c(
            mean(d$conc),
            stats::median(unique(d$time))),
          data=d,
          verbose=verbose
        )
      }, silent=!verbose)
    tss <-
      if (inherits(current_model, "try-error")) {
        NA_real_
      } else if (is.null(current_model)) {
        NA_real_
      } else {
        stats::coef(current_model)[["tss"]]
      }
    tss
  }
  output <- match.arg(output, several.ok=TRUE)
  # Run by treatment or a single value for the full data frame
  data_maybe_grouped <-
    if ("treatment" %in% names(data)) {
      dplyr::grouped_df(data, vars = "treatment")
    } else {
      data
    }
  ret <-
    data_maybe_grouped %>%
    dplyr::summarize(
      tss.monoexponential.single=
        fit_tss(
          data.frame(
            time=.$time,
            tss.constant=.$tss.constant,
            conc=.$conc,
            stringsAsFactors=FALSE
          )
        )
    )
  if ("subject" %in% names(data) &
      "individual" %in% output) {
    data_grouped <-
      if (all(c("treatment", "subject") %in% names(data))) {
        dplyr::grouped_df(data, vars = c("treatment", "subject"))
      } else if ("subject" %in% names(data)) {
        dplyr::grouped_df(data, vars="subject")
      } else {
        stop("Please report a bug. Subject must be specified to have subject-level fitting") # nocov
      }
    ret_sub <-
      dplyr::summarize(
        data_grouped,
        tss.monoexponential.individual=
          fit_tss(
            data.frame(
              time=.data$time,
              tss.constant=.data$tss.constant,
              conc=.data$conc,
              stringsAsFactors=FALSE
            )
          )
      )
    ret <- merge(ret, ret_sub, all=TRUE)
  }
  # Return the requested columns
  as.data.frame(
    ret[,c(intersect(names(ret), c("subject", "treatment")),
           paste("tss.monoexponential", output, sep=".")),
        drop=FALSE],
    stringsAsFactors=FALSE
  )
}

test_that("pk.tss.data.prep", {
  conc.test <- 1:5
  time.test <- 0:4
  subject.test <- letters[1:5]
  treatment.test <- LETTERS[1:5]
  time.dosing.test <- 0
  # Confirm that any NAs in time.dosing are an error
  expect_error(
    pk.tss.data.prep(
      conc=conc.test,
      time=time.test,
      subject=subject.test,
      treatment=treatment.test,
      time.dosing=NA
    ),
    regexp="time.dosing may not contain any NA values"
  )
  expect_error(
    pk.tss.data.prep(
      conc=conc.test,
      time=time.test,
      subject.dosing=subject.test,
      treatment=treatment.test,
      time.dosing=NA
    ),
    regexp="Cannot give subject.dosing without subject"
  )
  expect_equal(
    pk.tss.data.prep(
      conc=conc.test,
      time=time.test,
      time.dosing=0
    ),
    data.frame(conc=1, time=0)
  )
  
  expect_error(pk.tss.data.prep(conc=conc.test,
                                time=time.test,
                                subject=subject.test,
                                treatment=treatment.test,
                                time.dosing=c(0, NA)),
               regexp="time.dosing may not contain any NA values")

  # Confirm that conc and subject must be the same length
  expect_error(pk.tss.data.prep(conc=conc.test,
                                time=time.test,
                                subject=subject.test[-1],
                                treatment=treatment.test,
                                time.dosing=time.dosing.test),
               regexp="arguments imply differing number of rows: 5, 4")

  # Confirm that conc and treatment must be the same length
  expect_error(pk.tss.data.prep(conc=conc.test,
                                time=time.test,
                                subject=subject.test,
                                treatment=treatment.test[-1],
                                time.dosing=time.dosing.test),
               regexp="arguments imply differing number of rows: 5, 4")

  # If removed down to one treatment, treatment is not a column of
  # the output.
  expect_equal(pk.tss.data.prep(conc=conc.test,
                                time=time.test,
                                subject=subject.test,
                                treatment=treatment.test,
                                time.dosing=time.dosing.test),
               data.frame(conc=1, time=0))

  # If no treatment is given, it still works.
  expect_equal(pk.tss.data.prep(conc=conc.test,
                                time=time.test,
                                subject=subject.test,
                                time.dosing=time.dosing.test),
               data.frame(conc=1, time=0))

  # If no subject is given, it still works
  expect_equal(pk.tss.data.prep(conc=conc.test,
                                time=time.test,
                                treatment=treatment.test,
                                time.dosing=time.dosing.test),
               data.frame(conc=1, time=0))
  
  # What do we actually expect to get out?
  # Check a single row output dropping treatment
  expect_equal(pk.tss.data.prep(conc=conc.test,
                                time=time.test,
                                subject=subject.test,
                                treatment=treatment.test,
                                time.dosing=time.dosing.test,
                                subject.dosing=subject.test),
               data.frame(time=0, conc=1))

  # Check a single row output with no treatment given
  expect_equal(pk.tss.data.prep(conc=conc.test,
                                time=time.test,
                                subject=subject.test,
                                time.dosing=time.dosing.test),
               data.frame(conc=1, time=0))

  # Check a multi-row output with treatments kept
  conc.test <- 1:10
  time.test <- rep(0:4, 2)
  subject.test <- letters[1:10]
  treatment.test <- LETTERS[1:10]
  time.dosing.test <- 0
  expect_equal(
    pk.tss.data.prep(
      conc=conc.test,
      time=time.test,
      subject=subject.test,
      treatment=treatment.test,
      time.dosing=time.dosing.test
    ),
    data.frame(
      conc=c(1, 6), time=c(0, 0),
      subject=factor(c("a", "f")),
      treatment=factor(c("A", "F")),
      stringsAsFactors=FALSE
    ),
    ignore_attr = TRUE
  )

  # Check a multi-row output with treatments dropped
  conc.test <- 1:10
  time.test <- rep(0:4, 2)
  subject.test <- letters[1:10]
  treatment.test <- rep(LETTERS[1:5], 2)
  time.dosing.test <- 0
  expect_equal(
    pk.tss.data.prep(
      conc=conc.test,
      time=time.test,
      subject=subject.test,
      treatment=treatment.test,
      time.dosing=time.dosing.test
    ),
    data.frame(
      conc=c(1, 6),
      time=c(0, 0),
      subject=factor(c("a", "f")),
      stringsAsFactors=FALSE
    ),
    ignore_attr = TRUE
  )
})

# This data will be used multiple times in testing, and it is
# nontrivial to create.
generate.data <- function() {
  set.seed(5)
  tmpdata <-
    merge(
      data.frame(
        subject=factor(1:10),
        css.re=rnorm(10, sd=0.2),
        tss.re=rnorm(10, sd=0.2),
        treatment=rep(c("A", "B"), each=5),
        stringsAsFactors=FALSE
      ),
      data.frame(
        treatment=c("A", "B"),
        css.mean=c(5, 10),
        tss.mean=5,
        stringsAsFactors=FALSE
      )
    )
  tmpdata <- merge(tmpdata, data.frame(time=0:14))
  tmpdata$conc.resid <- rnorm(nrow(tmpdata), sd=0.05)
  tmpdata$conc <- with(tmpdata,
                       css.mean*exp(css.re+conc.resid)*
                       (1-exp(log(1-0.9)*time/(tss.mean*exp(tss.re)))))
  tmpdata
}

# Note that this graphically represents the test
# library(latticeExtra)
# (xyplot(conc~time|treatment,
#         groups=subject,
#         data=generate.data(),
#         type="l") +
#  layer(panel.abline(h=5, col="gray", lty=2), packets=1) +
#  layer(panel.abline(h=10, col="gray", lty=2), packets=2))

test_that("pk.tss.stepwise.linear", {
  tmpdata <- generate.data()
  expect_equal(
    pk.tss.stepwise.linear(conc=tmpdata$conc,
                           time=tmpdata$time,
                           subject=tmpdata$subject,
                           treatment=tmpdata$treatment,
                           time.dosing=0:14,
                           verbose=FALSE),
    data.frame(tss.stepwise.linear=7)
  )

  expect_warning(
    pk.tss.stepwise.linear(conc=tmpdata$conc,
                           time=tmpdata$time,
                           subject=tmpdata$subject,
                           treatment=tmpdata$treatment,
                           min.points=c(3, 4),
                           time.dosing=0:14,
                           verbose=FALSE),
    regexp="Only first value of min.points is used"
  )

  expect_error(
    pk.tss.stepwise.linear(
      conc=tmpdata$conc,
      time=tmpdata$time,
      subject=tmpdata$subject,
      treatment=tmpdata$treatment,
      time.dosing=0:14,
      min.points="A",
      level="A",
      verbose=FALSE
    ),
    regexp="min.points must be a number"
  )
  expect_error(
    pk.tss.stepwise.linear(
      conc=tmpdata$conc,
      time=tmpdata$time,
      subject=tmpdata$subject,
      treatment=tmpdata$treatment,
      time.dosing=0:14,
      min.points=1,
      level="A",
      verbose=FALSE
    ),
    regexp="min.points must be at least 3"
  )

  expect_error(
    pk.tss.stepwise.linear(conc=tmpdata$conc,
                           time=tmpdata$time,
                           subject=tmpdata$subject,
                           treatment=tmpdata$treatment,
                           time.dosing=0:14,
                           level="A",
                           verbose=FALSE),
    regexp="level must be a number"
  )

  expect_error(
    pk.tss.stepwise.linear(conc=tmpdata$conc,
                           time=tmpdata$time,
                           subject=tmpdata$subject,
                           treatment=tmpdata$treatment,
                           time.dosing=0:14,
                           level=2,
                           verbose=FALSE),
    regexp="level must be between 0 and 1, exclusive"
  )
  
  expect_error(
    pk.tss.stepwise.linear(conc=tmpdata$conc,
                           time=tmpdata$time,
                           subject=tmpdata$subject,
                           treatment=tmpdata$treatment,
                           time.dosing=0:14,
                           level=-1,
                           verbose=FALSE),
    regexp="level must be between 0 and 1, exclusive"
  )
  expect_error(
    pk.tss.stepwise.linear(conc=tmpdata$conc,
                           time=tmpdata$time,
                           subject=tmpdata$subject,
                           treatment=tmpdata$treatment,
                           time.dosing=0:14,
                           level=0,
                           verbose=FALSE),
    regexp="level must be between 0 and 1, exclusive"
  )
  expect_error(
    pk.tss.stepwise.linear(conc=tmpdata$conc,
                           time=tmpdata$time,
                           subject=tmpdata$subject,
                           treatment=tmpdata$treatment,
                           time.dosing=0:14,
                           level=1,
                           verbose=FALSE),
    regexp="level must be between 0 and 1, exclusive"
  )

  expect_warning(
    pk.tss.stepwise.linear(conc=tmpdata$conc,
                           time=tmpdata$time,
                           subject=tmpdata$subject,
                           treatment=tmpdata$treatment,
                           time.dosing=0:14,
                           level=c(0.95, 0.99),
                           verbose=FALSE),
    regexp="Only first value of level is being used"
  )

  # This is mainly to test verbosity
  expect_warning(
    pk.tss.stepwise.linear(conc=tmpdata$conc,
                           time=tmpdata$time,
                           subject=tmpdata$subject,
                           treatment=tmpdata$treatment,
                           time.dosing=0:14,
                           level=c(0.95, 0.99),
                           verbose=FALSE),
    regexp="Only first value of level is being used"
  )

  # Check outputs
  suppressMessages(
    withr::with_options(
      list(try.outFile=nullfile()),
      expect_message(
        pk.tss.stepwise.linear(
          conc=tmpdata$conc,
          time=tmpdata$time,
          subject=tmpdata$subject,
          treatment=tmpdata$treatment,
          time.dosing=0:14,
          level=0.8,
          verbose=TRUE
        ),
        regexp="Trying 0"
      )
    )
  )
  suppressMessages(
    withr::with_options(
      list(try.outFile=nullfile()),
      expect_message(
        pk.tss.stepwise.linear(
          conc=tmpdata$conc,
          time=tmpdata$time,
          subject=tmpdata$subject,
          treatment=tmpdata$treatment,
          time.dosing=0:14,
          level=0.8,
          verbose=TRUE
        ),
        regexp="Current interval"
      )
    )
  )
  
  # Ensure that the first value really is used
  expect_warning(v1 <-
    pk.tss.stepwise.linear(conc=tmpdata$conc,
                           time=tmpdata$time,
                           subject=tmpdata$subject,
                           treatment=tmpdata$treatment,
                           time.dosing=0:14,
                           level=c(0.8, 0.99),
                           verbose=FALSE))
  expect_equal(
    v1,
    pk.tss.stepwise.linear(
      conc=tmpdata$conc,
      time=tmpdata$time,
      subject=tmpdata$subject,
      treatment=tmpdata$treatment,
      time.dosing=0:14,
      level=0.8,
      verbose=FALSE
    )
  )

  # Confirm testing for minimum number of data points
  expect_warning(
    v1 <- pk.tss.stepwise.linear(
      conc=tmpdata$conc,
      time=tmpdata$time,
      subject=tmpdata$subject,
      treatment=tmpdata$treatment,
      time.dosing=0:1,
      min.points=3,
      level=0.99,
      verbose=FALSE
    ),
    regexp="After removing non-dosing time points, insufficient data remains for tss calculation"
  )
  expect_equal(v1, NA)
  expect_warning(
    pk.tss.stepwise.linear(conc=tmpdata$conc,
                           time=tmpdata$time,
                           subject=tmpdata$subject,
                           treatment=tmpdata$treatment,
                           time.dosing=0:1,
                           min.points=3,
                           level=0.99,
                           verbose=FALSE),
    regexp="After removing non-dosing time points, insufficient data remains for tss calculation")

  # Confirm the glm model works when subject-level data is not given.
  suppressMessages(
    expect_equal(
      pk.tss.stepwise.linear(conc=tmpdata$conc,
                             time=tmpdata$time,
                             treatment=tmpdata$treatment,
                             time.dosing=0:14,
                             verbose=FALSE),
      data.frame(tss.stepwise.linear=5),
      info="pk.tss.stepwise.linear no subject"
    )
  )
})

test_that("pk.tss.monoexponential", {
  tmpdata <- generate.data()
  expect_warning(
    expect_equal(
      pk.tss.monoexponential(
        conc=tmpdata$conc,
        time=tmpdata$time,
        subject=tmpdata$subject,
        treatment=tmpdata$treatment,
        time.dosing=0:14,
        verbose=FALSE
      ),
      data.frame(
        subject=factor(as.character(c(1, 10, 2:9))),
        tss.monoexponential.population=4.57618156812974,
        tss.monoexponential.popind=c(
          5.14156352865421, 4.64862524830397, 4.45956707917941,
          4.41492203844343, 4.6782583033301, 4.0823047621517,
          4.96242115751172, 4.52424147509819, 3.70338406668837,
          5.1465280219363),
        treatment=
          factor(c("A", "B", "A", "A", "A", "A", "B", "B", "B", "B")),
        tss.monoexponential.individual=c(
          5.87784329336254, 4.71066285661623, 4.51882509145954,
          3.91269286106442, 4.74475071729459, 3.99341726779716,
          5.08737230904342, 4.50068650719192, 3.4876172020751,
          5.35051537086801),
        tss.monoexponential.single=4.56067603534,
        stringsAsFactors=FALSE
      ),
      tolerance=1e-4
    )
  )
  expect_output(
    expect_equal(
      pk.tss.monoexponential(
        conc=c(0, 1000),
        time=0:1,
        subject=c(1, 1),
        treatment=c("A", "A"),
        time.dosing=0:1,
        tss.fraction=0.9,
        output="single"
      ),
      data.frame(tss.monoexponential.single=NA_real_),
      info="Single-subject data fitting works when it does not converge."
    ),
    regexp = "approximate covariance matrix for parameter estimates not of full rank"
  )
})

test_that("pk.tss.monoexponential without treatment", {
  tmpdata <- generate.data()
    expect_equal(
      pk.tss.monoexponential(
        conc=tmpdata$conc,
        time=tmpdata$time,
        subject=tmpdata$subject,
        time.dosing=0:14,
        verbose=FALSE
      ),
      data.frame(
        subject=factor(as.character(c(1, 10, 2:9))),
        tss.monoexponential.population=4.561580,
        tss.monoexponential.popind=c(
          5.028357, 4.672555, 4.536178,
          4.295758, 4.660561, 4.158700,
          4.940674, 4.507900, 3.676531,
          5.138583),
        tss.monoexponential.individual=c(
          5.87784329336254, 4.71066285661623, 4.51882509145954,
          3.91269286106442, 4.74475071729459, 3.99341726779716,
          5.08737230904342, 4.50068650719192, 3.4876172020751,
          5.35051537086801),
        tss.monoexponential.single=4.56067603534,
        stringsAsFactors=FALSE
      ),
      tolerance=1e-4
    )
})

test_that("pk.tss.monoexponential corner case tests", {
  tmpdata <- generate.data()
  # population output, only
  expect_warning(
    expect_equal(
      pk.tss.monoexponential(
        conc=tmpdata$conc,
        time=tmpdata$time,
        subject=tmpdata$subject,
        treatment=tmpdata$treatment,
        time.dosing=0:14,
        output = "population",
        verbose=FALSE
      ),
      data.frame(
        subject=factor(as.character(seq_len(10))),
        tss.monoexponential.population=4.57618156812974,
        stringsAsFactors=FALSE
      ),
      tolerance=1e-4
    )
  )
  # (Pseudo) single treatment, only
  expect_equal(
    pk.tss.monoexponential(
      conc=tmpdata$conc,
      time=tmpdata$time,
      subject=tmpdata$subject,
      time.dosing=0:14,
      output = "population",
      verbose=FALSE
    ),
    data.frame(
      subject=factor(as.character(seq_len(10))),
      tss.monoexponential.population=4.56157960341961,
      stringsAsFactors=FALSE
    ),
    tolerance=1e-4
  )
})

test_that("pk.tss.monoexponential expected warnings and errors", {
  tmpdata <- generate.data()
  expect_error(
    pk.tss.monoexponential(conc=tmpdata$conc,
                           time=tmpdata$time,
                           subject=tmpdata$subject,
                           treatment=tmpdata$treatment,
                           time.dosing=0:14,
                           tss.fraction=factor(1)),
    regexp="tss.fraction must be a number"
  )
  suppressWarnings(
    expect_warning(
      pk.tss.monoexponential(
        conc=tmpdata$conc,
        time=tmpdata$time,
        subject=tmpdata$subject,
        treatment=tmpdata$treatment,
        time.dosing=0:14,
        tss.fraction=c(0.5, 0.8)
      ),
      regexp="Only first value of tss.fraction is being used"
    )
  )
  expect_error(
    pk.tss.monoexponential(conc=tmpdata$conc,
                           time=tmpdata$time,
                           subject=tmpdata$subject,
                           treatment=tmpdata$treatment,
                           time.dosing=0:14,
                           tss.fraction=0),
    regexp="tss.fraction must be between 0 and 1, exclusive")
  expect_error(
    pk.tss.monoexponential(conc=tmpdata$conc,
                           time=tmpdata$time,
                           subject=tmpdata$subject,
                           treatment=tmpdata$treatment,
                           time.dosing=0:14,
                           tss.fraction=1),
    regexp="tss.fraction must be between 0 and 1, exclusive")
  expect_error(
    pk.tss.monoexponential(conc=tmpdata$conc,
                           time=tmpdata$time,
                           subject=tmpdata$subject,
                           treatment=tmpdata$treatment,
                           time.dosing=0:14,
                           tss.fraction=-1),
    regexp="tss.fraction must be between 0 and 1, exclusive")
  expect_error(
    pk.tss.monoexponential(conc=tmpdata$conc,
                           time=tmpdata$time,
                           subject=tmpdata$subject,
                           treatment=tmpdata$treatment,
                           time.dosing=0:14,
                           tss.fraction=2),
    regexp="tss.fraction must be between 0 and 1, exclusive")
  suppressWarnings(
    expect_warning(
      pk.tss.monoexponential(conc=tmpdata$conc,
                             time=tmpdata$time,
                             subject=tmpdata$subject,
                             treatment=tmpdata$treatment,
                             time.dosing=0:14,
                             tss.fraction=0.5),
      regexp="tss.fraction is usually >= 0.8"
    )
  )
  suppressWarnings(
    expect_warning(
      pk.tss.monoexponential(conc=tmpdata$conc,
                             time=tmpdata$time,
                             subject=tmpdata$subject,
                             treatment=tmpdata$treatment,
                             time.dosing=0:14,
                             tss.fraction=c(0.5, 0.8)),
      regexp="Only first value of tss.fraction is being used"
    )
  )
  tmpdata_single_subject <- tmpdata[tmpdata$subject == 1 & tmpdata$treatment == "A", ]
  expect_warning(
    tss_single_subject <-
      pk.tss.monoexponential(
        conc=tmpdata_single_subject$conc,
        time=tmpdata_single_subject$time,
        subject=tmpdata_single_subject$subject,
        treatment=tmpdata_single_subject$treatment,
        time.dosing=0:14,
        tss.fraction=0.8
      ),
    regexp="Cannot give 'population', 'popind', or 'individual' output without multiple subjects of data",
    fixed=TRUE
  )
  expect_equal(
    tss_single_subject,
    data.frame(tss.monoexponential.single=4.108541),
    tolerance=1e-5
  )
})

test_that("pk.tss", {
  # Ensure that pk.tss will go to the correct type of model
  tmpdata <- generate.data()
  suppressWarnings(
    expect_equal(
      pk.tss(
        conc=tmpdata$conc,
        time=tmpdata$time,
        subject=tmpdata$subject,
        treatment=tmpdata$treatment,
        time.dosing=0:14,
        verbose=FALSE,
        type="monoexponential"
      ),
      pk.tss.monoexponential(
        conc=tmpdata$conc,
        time=tmpdata$time,
        subject=tmpdata$subject,
        treatment=tmpdata$treatment,
        time.dosing=0:14,
        verbose=FALSE
      )
    )
  )

  expect_equal(
    pk.tss(conc=tmpdata$conc,
           time=tmpdata$time,
           subject=tmpdata$subject,
           treatment=tmpdata$treatment,
           time.dosing=0:14,
           verbose=FALSE,
           type="stepwise.linear"),
    pk.tss.stepwise.linear(conc=tmpdata$conc,
                           time=tmpdata$time,
                           subject=tmpdata$subject,
                           treatment=tmpdata$treatment,
                           time.dosing=0:14,
                           verbose=FALSE))

  # pk.tss will calculate both if requested
  suppressWarnings(
    expect_equal(
      pk.tss(
        conc=tmpdata$conc,
        time=tmpdata$time,
        subject=tmpdata$subject,
        treatment=tmpdata$treatment,
        time.dosing=0:14,
        verbose=FALSE,
        type=c("monoexponential", "stepwise.linear")
      ),
      merge(
        pk.tss.monoexponential(
          conc=tmpdata$conc,
          time=tmpdata$time,
          subject=tmpdata$subject,
          treatment=tmpdata$treatment,
          time.dosing=0:14,
          verbose=FALSE
        ),
        pk.tss.stepwise.linear(
          conc=tmpdata$conc,
          time=tmpdata$time,
          subject=tmpdata$subject,
          treatment=tmpdata$treatment,
          time.dosing=0:14,
          verbose=FALSE
        ),
        all=TRUE
      )
    )
  )
})

test_that("pk.tss.monoexponential with single-subject data", {
  d_prep <- datasets::Theoph[datasets::Theoph$Subject %in% 2, ]
  dose_times <- seq(0, 96-1, by=6)
  d_multidose <-
    superposition(
      conc=d_prep$conc,
      time=d_prep$Time,
      tau=96, # 48 hours
      n.tau=1, # One tau interval (0 to 48 hours)
      dose.times=dose_times
    )
  expect_equal(
    pk.tss.monoexponential(
      conc=d_multidose$conc, time=d_multidose$time, subject=rep(1, nrow(d_multidose)),
      time.dosing=dose_times, subject.dosing=rep(1, length(dose_times)),
      output="single"
    ),
    data.frame(tss.monoexponential.single=22.53),
    tolerance=0.001
  )
})

test_that("verbose being TRUE", {
  d_prep <- datasets::Theoph[datasets::Theoph$Subject %in% 2, ]
  dose_times <- seq(0, 96-1, by=6)
  d_multidose <-
    superposition(
      conc=d_prep$conc,
      time=d_prep$Time,
      tau=96, # 48 hours
      n.tau=1, # One tau interval (0 to 48 hours)
      dose.times=dose_times
    )
  expect_equal(
    pk.tss.monoexponential(
      conc=d_multidose$conc, time=d_multidose$time, subject=rep(1, nrow(d_multidose)),
      time.dosing=dose_times, subject.dosing=rep(1, length(dose_times)),
      output="single", verbose = TRUE
    ),
    data.frame(tss.monoexponential.single=22.53),
    tolerance=0.001
  )
})

test_that("tss Monoexponential no models converged", {
  bad_data <- data.frame(subject=rep(1:2, each = 4), time=0, conc=0, tss.constant=0.9)
  expect_warning(pk.tss.monoexponential.population(data = bad_data, output = "population"),
  regexp = "No population model for monoexponential Tss converged, no results given")
})

test_that("tss Monoexponential no models converged with verbose = TRUE", {
  bad_data <- data.frame(subject=rep(1:2, each = 4), time=0, conc=0, tss.constant=0.9)
  expect_warning(pk.tss.monoexponential.population(data = bad_data, output = "population", verbose = TRUE),
                 regexp = "No population model for monoexponential Tss converged, no results given")
})

# Options for use within the code for setting and getting PKNCA default options. ####

.PKNCA.option.check <- list(
  adj.r.squared.factor=function(x, default=FALSE, description=FALSE) {
    if (description)
      return(paste(
        "The adjusted r^2 for the calculation of lambda.z has this factor",
        "times the number of data points added to it.  It allows for more",
        "data points to be preferred in the calculation of half-life."))
    if (default)
      return(0.0001)
    if (length(x) != 1)
      stop("adj.r.squared.factor must be a scalar")
    if (is.factor(x) |
        !is.numeric(x))
      stop("adj.r.squared.factor must be numeric (and not a factor)")
    # Must be between 0 and 1, exclusive
    if (x <= 0 | x >= 1)
      stop("adj.r.squared.factor must be between 0 and 1, exclusive")
    if (x > 0.01)
      warning("adj.r.squared.factor is usually <0.01")
    x
  },
  max.missing=function(x, default=FALSE, description=FALSE) {
    if (description)
      return(paste(
        "The maximum fraction of the data that may be missing ('NA') to",
        "calculate summary statistics with the business.* functions."))
    if (default)
      return(0.5)
    if (length(x) != 1)
      stop("max.missing must be a scalar")
    if (is.factor(x) | !is.numeric(x))
      stop("max.missing must be numeric (and not a factor)")
    # Must be between 0 and 1, inclusive
    if (x < 0 | x >= 1)
      stop("max.missing must be between 0 and 1")
    if (x > 0.5)
      warning("max.missing is usually <= 0.5")
    x
  },
  auc.method=function(x, default=FALSE, description=FALSE) {
    choices <- eval(formals(assert_aucmethod)$method)
    if (description)
      return(paste(
        "The method used to calculate the AUC and related statistics.",
        "Options are:",
        paste0('"', choices, '"', collapse = ", ")
      ))
    if (default)
      return(choices[1])
    assert_aucmethod(x)
  },
  conc.na=function(x, default=FALSE, description=FALSE) {
    if (description)
      return(paste(
        "How should missing ('NA') concentration values be handled?  See",
        "help for 'clean.conc.na' for how to use this option."))
    if (default)
      return("drop")
    if (is.na(x))
      stop("conc.na must not be NA")
    if (is.factor(x)) {
      warning("conc.na may not be a factor; attempting conversion")
      x <- as.character(x)
    }
    if (tolower(x) %in% "drop") {
      x <- tolower(x)
    } else if (is.numeric(x)) {
      if (is.infinite(x)) {
        stop("When a number, conc.na must be finite")
      } else if (x < 0) {
        warning("conc.na is usually not < 0")
      }
    } else {
      stop("conc.na must either be a finite number or the text 'drop'")
    }
    x
  },
  conc.blq=function(x, default=FALSE, description=FALSE) {
    if (description)
      return(paste(
        "How should below the limit of quantification (zero, 0) concentration",
        "values be handled?  See help for 'clean.conc.blq' for how to use",
        "this option."))
    if (default)
      return(list(first="keep",
                  middle="drop",
                  last="keep"))
    check.element <- function(x) {
      if (length(x) != 1)
        stop("conc.blq must be a scalar")
      if (is.na(x))
        stop("conc.blq must not be NA")
      if (is.factor(x)) {
        warning("conc.blq may not be a factor; attempting conversion")
        x <- as.character(x)
      }
      if (tolower(x) %in% c("drop", "keep")) {
        x <- tolower(x)
      } else if (is.numeric(x)) {
        if (is.infinite(x)) {
          stop("When a number, conc.blq must be finite")
        } else if (x < 0) {
          warning("conc.blq is usually not < 0")
        }
      } else {
        stop("conc.blq must either be a finite number or the text 'drop' or 'keep'")
      }
      x
    }
    if (is.list(x)) {
      tfirst_names <- c("first", "last", "middle")
      tmax_names <- c("before.tmax", "after.tmax")

      are.names.mixed <- any(names(x) %in% tfirst_names) & any(names(x) %in% tmax_names)
      extra.names <- setdiff(names(x), c(tfirst_names, tmax_names))
      missing.names <- if (any(names(x) %in% tfirst_names)) setdiff(tfirst_names, names(x)) else setdiff(tmax_names, names(x))
      duplicated.names <- names(x)[duplicated(names(x))]
      if (are.names.mixed)
        stop("When given as a list, prevent mixing arguments of different BLQ strategies.
             Either define 'first', 'middle' and 'last' or 'before.tmax' and 'after.tmax'.")
      if (length(extra.names) != 0)
        stop("When given as a list, conc.blq must only have elements named 'first', 'middle' and 'last' or 'before.tmax' and 'after.tmax'.")
      if (length(missing.names) != 0)
        stop("When given as a list, conc.blq must include all elements named 'first', 'middle' and 'last' or 'before.tmax' and 'after.tmax'.")
      if (length(duplicated.names) != 0)
        stop("When given as a list, conc.blq should not have duplicated names")
      # After the names are confirmed, confirm each value.
      x <- lapply(x, check.element)
    } else {
      x <- check.element(x)
    }
    x
  },
  debug = function(x, default = FALSE, description = FALSE) {
    if (description) {
      return("Enable PKNCA debugging mode (not for production use)")
    }
    if (default) {
      return(NULL)
    }
    x
  },
  first.tmax=function(x, default=FALSE, description=FALSE) {
    if (description)
      return(paste(
        "If there is more than one concentration equal to Cmax, which time",
        "should be selected for Tmax?  If 'TRUE', the first will be selected.",
        "If 'FALSE', the last will be selected."))
    if (default)
      return(TRUE)
    if (length(x) != 1)
      stop("first.tmax must be a scalar")
    if (is.na(x))
      stop("first.tmax may not be NA")
    if (!is.logical(x)) {
      x <- as.logical(x)
      if (is.na(x)) {
        stop("Could not convert first.tmax to a logical value")
      } else {
        warning("Converting first.tmax to a logical value: ", x)
      }
    }
    x
  },
  allow.tmax.in.half.life=function(x, default=FALSE, description=FALSE) {
    if (description)
      return(paste(
        "Should the concentration and time at Tmax be allowed in the",
        "half-life calculation?  'TRUE' is yes and 'FALSE' is no."))
    if (default)
      return(FALSE)
    if (length(x) != 1)
      stop("allow.tmax.in.half.life must be a scalar")
    if (is.na(x))
      stop("allow.tmax.in.half.life may not be NA")
    if (!is.logical(x)) {
      x <- as.logical(x)
      if (is.na(x)) {
        stop("Could not convert allow.tmax.in.half.life to a logical value")
      } else {
        warning("Converting allow.tmax.in.half.life to a logical value: ", ret)
      }
    }
    x
  },

  keep_interval_cols = function(x, default = FALSE, description = FALSE) {
    if (description)
      return("What additional columns from the intervals should be kept in the results?")
    if (default)
      return(NULL)
    checkmate::assert_names(x)
    x
  },

  min.hl.points=function(x, default=FALSE, description=FALSE) {
    if (description)
      return("What is the minimum number of points required to calculate half-life?")
    if (default)
      return(3)
    if (length(x) != 1)
      stop("min.hl.points must be a scalar")
    if (is.factor(x))
      stop("min.hl.points cannot be a factor")
    if (!is.numeric(x))
      stop("min.hl.points must be a number")
    if (x < 2)
      stop("min.hl.points must be >=2")
    if (min(x %% 1, 1 - (x %% 1)) >
        100*.Machine$double.eps) {
      warning("Non-integer given for min.hl.points; rounding to nearest integer")
      x <- round(x)
    }
    x
  },
  min.span.ratio=function(x, default=FALSE, description=FALSE) {
    if (description)
      return("What is the minimum span ratio required to consider a half-life valid?")
    if (default)
      return(2)
    if (length(x) != 1)
      stop("min.span.ratio must be a scalar")
    if (is.factor(x))
      stop("min.span.ratio cannot be a factor")
    if (!is.numeric(x))
      stop("min.span.ratio must be a number")
    if (x <= 0)
      stop("min.span.ratio must be > 0")
    if (x < 2)
      warning("min.span.ratio is usually >= 2")
    x
  },
  max.aucinf.pext=function(x, default=FALSE, description=FALSE) {
    if (description)
      return("What is the maximum percent extrapolation to consider an AUCinf valid?")
    if (default)
      return(20)
    if (length(x) != 1)
      stop("max.aucinf.pext must be a scalar")
    if (is.factor(x))
      stop("max.aucinf.pext cannot be a factor")
    if (!is.numeric(x))
      stop("max.aucinf.pext must be a number")
    if (x <= 0)
      stop("max.aucinf.pext must be > 0")
    if (x > 25)
      warning("max.aucinf.pext is usually <=25")
    if (x < 1)
      warning("max.aucinf.pext is on the percent not ratio scale, value given is <1%")
    x
  },
  min.hl.r.squared=function(x, default=FALSE, description=FALSE) {
    if (description)
      return("What is the minimum r-squared value to consider a half-life calculation valid?")
    if (default)
      return(0.9)
    if (length(x) != 1)
      stop("min.hl.r.squared must be a scalar")
    if (is.factor(x))
      stop("min.hl.r.squared cannot be a factor")
    if (!is.numeric(x))
      stop("min.hl.r.squared must be a number")
    if (x <= 0 | x >= 1)
      stop("min.hl.r.squared must be between 0 and 1, exclusive")
    if (x < 0.9)
      warning("min.hl.r.squared is usually >= 0.9")
    x
  },

  progress = function(x, default = FALSE, description = FALSE) {
    if (description)
      return("A value to pass to purrr::pmap(.progress = ) to create a progress bar while running")
    if (default) {
      return(TRUE)
    }
    x
  },

  tau.choices=function(x, default=FALSE, description=FALSE) {
    if (description)
      return(paste(
        "What values for tau (repeating interdose interval) should be",
        "considered when attempting to automatically determine the intervals",
        "for multiple dosing?  See 'choose.auc.intervals' and 'find.tau' for",
        "more information.  'NA' means automatically look at any potential",
        "interval."))
    if (default)
      return(NA)
    if (is.factor(x))
      stop("tau.choices cannot be a factor")
    if (length(x) > 1 & any(is.na(x)))
      stop("tau.choices may not include NA and be a vector")
    if (!identical(x, NA))
      if (!is.numeric(x))
        stop("tau.choices must be a number")
      if (!is.vector(x)) {
        warning("tau.choices must be a vector, converting")
        x <- as.vector(x)
      }
    x
  },
  single.dose.aucs=function(x, default=FALSE, description=FALSE) {
    if (description)
      return("When data is single-dose, what intervals should be used?")
    if (default) {
      # It is good to put this through the specification checker in case they
      # get out of sync during development.  (A free test case!)
      x <- data.frame(
        start=0,
        end=c(24, Inf),
        auclast=c(TRUE, FALSE),
        aucinf.obs=c(FALSE, TRUE),
        half.life=c(FALSE, TRUE),
        tmax=c(FALSE, TRUE),
        cmax=c(FALSE, TRUE))
    }
    check.interval.specification(x)
  },
  allow_partial_missing_units = function(x, default = FALSE, description = FALSE) {
    if (description)
      return("When using unit assignment and conversions, should some units be allowed to be missing?")
    if (default) {
      return(FALSE)
    }
    checkmate::assert_logical(x, any.missing = FALSE, len = 1)
    x
  }
)

# Functions controlling and modifying options ####

#' Set default options for PKNCA functions
#'
#' This function will set the default PKNCA options.  If given no inputs, it
#' will provide the current option set.  If given name/value pairs, it will set
#' the option (as in the [options()] function).  If given a name, it will return
#' the value for the parameter.  If given the `default` option as true, it will
#' provide the default options.
#'
#' Options are either for calculation or summary functions. Calculation options
#' are required for a calculation function to report a result (otherwise the
#' reported value will be `NA`). Summary options are used during summarization
#' and are used for assessing what values are included in the summary.
#'
#' See the vignette 'Options for Controlling PKNCA' for a current list of
#' options (`vignette("Options-for-Controlling-PKNCA", package="PKNCA")`).
#'
#' @param \dots options to set or get the value for
#' @param default (re)sets all default options
#' @param check check a single option given, but do not set it (for validation
#'   of the values when used in another function)
#' @param name An option name to use with the `value`.
#' @param value An option value (paired with the `name`) to set or check (if
#'   `NULL`, ).
#' @returns If...
#' \describe{
#'   \item{no arguments are given}{returns the current options.}
#'   \item{a value is set (including the defaults)}{returns `NULL`}
#'   \item{a single value is requested}{the current value of that option is returned as a scalar}
#'   \item{multiple values are requested}{the current values of those options are returned as a list}
#' }
#' @family PKNCA calculation and summary settings
#' @seealso [PKNCA.options.describe()]
#' @examples
#'
#' PKNCA.options()
#' PKNCA.options(default=TRUE)
#' PKNCA.options("auc.method")
#' PKNCA.options(name="auc.method")
#' PKNCA.options(auc.method="lin up/log down", min.hl.points=3)
#' @export
PKNCA.options <- function(..., default=FALSE, check=FALSE, name, value) {
  current <- get("options", envir=.PKNCAEnv)
  # If the options have not been initialized, initialize them and then proceed.
  if (is.null(current) & !default) {
    PKNCA.options(default=TRUE)
    current <- get("options", envir=.PKNCAEnv)
  }
  args <- list(...)
  # Put the name/value pair into the args as if they were specified
  # like another argument.
  if (missing(name)) {
    if (!missing(value))
      stop("Cannot have a value without a name")
  } else {
    if (name %in% names(args))
      stop("Cannot give an option name both with the name argument and as a named argument.")
    if (!missing(value)) {
      args[[name]] <- value
    } else {
      args <- append(args, name)
    }
  }
  if (default & check)
    stop("Cannot request both default and check")
  if (default) {
    if (length(args) > 0)
      stop("Cannot set default and set new options at the same time.")
    # Extract all the default values
    defaults <- lapply(.PKNCA.option.check,
                       FUN=function(x) x(default=TRUE))
    # Set the default options
    assign("options", defaults, envir=.PKNCAEnv)
  } else if (check) {
    # Check an option for accuracy, but don't set it
    if (length(args) != 1)
      stop("Must give exactly one option to check")
    n <- names(args)
    if (!(n %in% names(.PKNCA.option.check)))
      stop(paste("Invalid setting for PKNCA:", n))
    # Verify the option, and return the sanitized version
    return(.PKNCA.option.check[[n]](args[[n]]))
  } else if (length(args) > 0) {
    if (is.null(names(args))) {
      # Confirm that the settings exist
      if (length(bad.args <- setdiff(unlist(args), names(current))) > 0)
        stop(sprintf("PKNCA.options does not have value(s) for %s.",
                     paste(bad.args, collapse=", ")))
      # Get the setting(s)
      if (length(args) == 1) {
        ret <- current[[args[[1]]]]
      } else {
        ret <- list()
        for (i in seq_len(length(args))) {
          ret[[ args[[i]] ]] <- current[[ args[[i]] ]]
        }
      }
      return(ret)
    } else {
      # Set a value
      # Verify values are viable and then set them.
      for (n in names(args)) {
        if (!(n %in% names(.PKNCA.option.check)))
          stop(paste("Invalid setting for PKNCA:", n))
        # Verify and set the option value
        current[[n]] <- .PKNCA.option.check[[n]](args[[n]])
      }
      # Assign current into the setting environment
      assign("options", current, envir=.PKNCAEnv)
    }
  } else {
    return(current)
  }
}

#' Choose either the value from an option list or the current set value for an
#' option.
#'
#' @param name The option name requested.
#' @param value A value to check for the option (`NULL` to choose not to check
#'   the value).
#' @param options List of changes to the default PKNCA options (see
#'   `PKNCA.options()`)
#' @returns The value of the option first from the `options` list and if it is
#'   not there then from the current settings.
#' @family PKNCA calculation and summary settings
#' @export
PKNCA.choose.option <- function(name, value=NULL, options=list()) {
  if (!is.null(value)) {
    PKNCA.options(name=name, value=value, check=TRUE)
  } else if (name %in% names(options)) {
    PKNCA.options(name=name, value=options[[name]], check=TRUE)
  } else {
    PKNCA.options(name)
  }
}

#' Describe a PKNCA.options option by name.
#'
#' @param name The option name requested.
#' @return A character string of the description.
#' @seealso [PKNCA.options()]
#' @export
PKNCA.options.describe <- function(name) {
  .PKNCA.option.check[[name]](description=TRUE)
}

#' Define how NCA parameters are summarized.
#'
#' @param name The parameter name or a vector of parameter names.  It must have
#'   already been defined (see [add.interval.col()]).
#' @param description A single-line description of the summary
#' @param point The function to calculate the point estimate for the summary.
#'   The function will be called as `point(x)` and must return a scalar value
#'   (typically a number, NA, or a string).
#' @param spread Optional.  The function to calculate the spread (or
#'   variability).  The function will be called as `spread(x)` and must return a
#'   scalar or two-long vector (typically a number, NA, or a string).
#' @param rounding Instructions for how to round the value of point and spread.
#'   It may either be a list or a function.  If it is a list, then it must have
#'   a single entry with a name of either "signif" or "round" and a value of the
#'   digits to round.  If a function, it is expected to return a scalar number
#'   or character string with the correct results for an input of either a
#'   scalar or a two-long vector.
#' @param reset Reset all the summary instructions
#' @returns All current summary settings (invisibly)
#' @seealso [summary.PKNCAresults()]
#' @family PKNCA calculation and summary settings
#' @examples
#' \dontrun{
#' PKNCA.set.summary(
#'   name="half.life",
#'   description="arithmetic mean and standard deviation",
#'   point=business.mean,
#'   spread=business.sd,
#'   rounding=list(signif=3)
#' )
#' }
#' @export
PKNCA.set.summary <- function(name, description, point, spread,
                              rounding=list(signif=3), reset=FALSE) {
  if (reset) {
    current <- list()
  } else {
    current <- get("summary", envir=.PKNCAEnv)
  }
  if (missing(name) & missing(point) & missing(spread)) {
    if (reset)
      assign("summary", current, envir=.PKNCAEnv)
    return(invisible(current))
  }
  # Confirm that the name exists
  if (!all(found_names <- name %in% names(get("interval.cols", envir=.PKNCAEnv)))) {
    stop(paste("You must first define the parameter name with add.interval.col.  Parameters not yet defined are:",
               paste(name[!found_names], collapse=", ")))
  }
  # Reset all names to prep for settings below
  for (current_name in name) {
    current[[current_name]] <- list()
  }
  # Confirm that description is a scalar character string
  if (!is.character(description)) {
    stop("`description` must be a character string.")
  } else if (length(description) != 1) {
    stop("`description` must be a scalar.")
  }
  for (current_name in name) {
    current[[current_name]]$description <- description
  }
  # Confirm that point is a function
  if (!is.function(point)) {
    stop("`point` must be a function")
  }
  for (current_name in name) {
    current[[current_name]]$point <- point
  }
  # Confirm that spread is a function (if given)
  if (!missing(spread)) {
    if (!is.function(spread)) {
      stop("spread must be a function")
    }
    for (current_name in name) {
      current[[current_name]]$spread <- spread
    }
  }
  # Confirm that rounding is either a single-entry list or a function
  if (is.list(rounding)) {
    if (length(rounding) != 1) {
      stop("rounding must have a single value in the list")
    }
    if (!(names(rounding) %in% c("signif", "round"))) {
      stop("When a list, rounding must have a name of either 'signif' or 'round'")
    }
    for (current_name in name) {
      current[[current_name]]$rounding <- rounding
    }
  } else if (is.function(rounding)) {
    for (current_name in name) {
      current[[current_name]]$rounding <- rounding
    }
  } else {
    stop("rounding must be either a list or a function")
  }
  # Set the summary parameters
  assign("summary", current, envir=.PKNCAEnv)
  invisible(current)
}

test_that("PKNCA.options", {
  # Missing/incorrect option names give an error indicating all the
  # options that are missing.
  expect_error(PKNCA.options("foo"),
               regexp="PKNCA.options does not have value\\(s\\) for foo.")
  expect_error(PKNCA.options("foo", "bar"),
               regexp="PKNCA.options does not have value\\(s\\) for foo, bar.")
  # A mix of mixxing and extant options only give the missing ones
  expect_error(PKNCA.options("foo", "bar", "tau.choices"),
               regexp="PKNCA.options does not have value\\(s\\) for foo, bar.")

  # Single extant options give their default value
  expect_equal(PKNCA.options("min.hl.points"), 3)

  # Multiple extant options give a list of their default values
  expect_equal(PKNCA.options("min.hl.points", "min.hl.r.squared"),
               list(min.hl.points=3,
                    min.hl.r.squared=0.9))
  # The returned options are in the order they were requested
  expect_equal(PKNCA.options("min.hl.points", "min.hl.r.squared"),
               list(min.hl.points=3,
                    min.hl.r.squared=0.9))

  # Asking for an option using the name argument works the same as
  # asking for one with a string.
  expect_equal(PKNCA.options(name="single.dose.aucs"),
               PKNCA.options("single.dose.aucs"))
  # You can request more than one option by using string inputs
  # combined with a single "name" argument.
  expect_equal(PKNCA.options("first.tmax", name="single.dose.aucs"),
               PKNCA.options("first.tmax", name="single.dose.aucs"))
  # You cannot both set an option and give it with a name.
  expect_error(PKNCA.options(foo="var", name="foo", value="bar"),
               regexp="Cannot give an option name both with the name argument and as a named argument.")
  # You cannot both set an option (with a value) and request an
  # option
  expect_error(PKNCA.options("first.tmax", name="min.span.ratio", value=2),
               regexp="Invalid setting for PKNCA")
  # You cannot give a value without a name.
  expect_error(PKNCA.options(value=5),
               regexp="Cannot have a value without a name")

  expect_error(PKNCA.options("adj.r.squared.factor", default=TRUE, check=TRUE),
               regexp="Cannot request both default and check")

  expect_error(PKNCA.options(adj.r.squared.factor=0.1, default=TRUE, check=TRUE),
               regexp="Cannot request both default and check")

  expect_error(PKNCA.options(adj.r.squared.factor=0.1, max.aucinf.pext=15, check=TRUE),
               regexp="Must give exactly one option to check")

  # Confirm that the default state is as expected (setting it first
  # in case the tests are run in a non-default state)
  PKNCA.options(default=TRUE)
  expect_equal(
    PKNCA.options(),
    list(
      adj.r.squared.factor = 0.0001,
      max.missing = 0.5,
      auc.method = "lin up/log down",
      conc.na = "drop",
      conc.blq = list(
        first = "keep",
        middle = "drop",
        last = "keep"
      ),
      debug = NULL,
      first.tmax = TRUE,
      allow.tmax.in.half.life = FALSE,
      keep_interval_cols = NULL,
      min.hl.points = 3,
      min.span.ratio = 2,
      max.aucinf.pext = 20,
      min.hl.r.squared = 0.9,
      progress = TRUE,
      tau.choices = NA,
      single.dose.aucs = check.interval.specification(
        data.frame(
          start = 0,
          end = c(24, Inf),
          auclast = c(TRUE, FALSE),
          aucinf.obs = c(FALSE, TRUE),
          half.life = c(FALSE, TRUE),
          tmax = c(FALSE, TRUE),
          cmax = c(FALSE, TRUE)
        )
      ),
      allow_partial_missing_units = FALSE
    )
  )

  # Check all the checks on options

  # adj.r.squared.factor
  expect_error(PKNCA.options(adj.r.squared.factor=c(0.1, 0.9), check=TRUE),
               regexp="adj.r.squared.factor must be a scalar")
  expect_error(PKNCA.options(adj.r.squared.factor=1, check=TRUE),
               regexp="adj.r.squared.factor must be between 0 and 1, exclusive")
  expect_error(PKNCA.options(adj.r.squared.factor=0, check=TRUE),
               regexp="adj.r.squared.factor must be between 0 and 1, exclusive")
  expect_error(PKNCA.options(adj.r.squared.factor="A", check=TRUE),
               regexp="adj.r.squared.factor must be numeric \\(and not a factor\\)")
  expect_warning(v1 <- PKNCA.options(adj.r.squared.factor=0.9, check=TRUE))
  expect_equal(v1, 0.9)
  expect_warning(PKNCA.options(adj.r.squared.factor=0.9, check=TRUE),
                 regexp="adj.r.squared.factor is usually <0.01")

  # max.missing
  expect_error(PKNCA.options(max.missing=c(1, 2), check=TRUE),
               regexp="max.missing must be a scalar")
  expect_error(PKNCA.options(max.missing="A", check=TRUE),
               regexp="max.missing must be numeric \\(and not a factor\\)")
  expect_error(PKNCA.options(max.missing=-1, check=TRUE),
               regexp="max.missing must be between 0 and 1")
  expect_error(PKNCA.options(max.missing=1, check=TRUE),
               regexp="max.missing must be between 0 and 1")
  expect_equal(PKNCA.options(max.missing=0, check=TRUE),
               0)
  expect_equal(PKNCA.options(max.missing=0.2, check=TRUE),
               0.2)
  expect_warning(PKNCA.options(max.missing=0.6, check=TRUE),
                 regexp="max.missing is usually <= 0.5")

  # auc.method
  # All possible methods
  expect_equal(PKNCA.options(auc.method="linear", check=TRUE),
               "linear",
               info="auc.method selection works for linear")
  expect_equal(PKNCA.options(auc.method="lin up/log down", check=TRUE),
               "lin up/log down",
               info="auc.method selection works for lin up/log down")
  expect_error(PKNCA.options(auc.method="foo", check=TRUE),
               regexp="should be one of",
               info="auc.method is a valid method")

  # conc.na
  expect_equal(PKNCA.options(conc.na="drop", check=TRUE),
               "drop")
  expect_warning(v1 <- PKNCA.options(conc.na=factor("drop"), check=TRUE),
                 regexp="conc.na may not be a factor; attempting conversion")
  expect_equal(v1, "drop")
  expect_warning(PKNCA.options(conc.na=factor("drop"), check=TRUE),
                 regexp="conc.na may not be a factor; attempting conversion")
  expect_equal(PKNCA.options(conc.na=1, check=TRUE),
               1)
  expect_warning(v1 <- PKNCA.options(conc.na=-1, check=TRUE),
                 regexp="conc.na is usually not < 0")
  expect_equal(v1, -1)
  expect_warning(PKNCA.options(conc.na=-1, check=TRUE),
                 regexp="conc.na is usually not < 0")
  expect_error(PKNCA.options(conc.na=Inf, check=TRUE),
               regexp="When a number, conc.na must be finite")
  expect_error(PKNCA.options(conc.na="foo", check=TRUE),
               regexp="conc.na must either be a finite number or the text 'drop'")

  # conc.blq
  # Confirm all types of single-style inputs
  expect_equal(PKNCA.options(conc.blq="drop", check=TRUE),
               "drop")
  expect_equal(PKNCA.options(conc.blq="keep", check=TRUE),
               "keep")
  expect_equal(PKNCA.options(conc.blq=0, check=TRUE),
               0)
  expect_equal(PKNCA.options(conc.blq=1, check=TRUE),
               1)
  expect_warning(v1 <- PKNCA.options(conc.blq=factor("drop"), check=TRUE),
                 "conc.blq may not be a factor; attempting conversion")
  expect_equal(v1, "drop")
  expect_error(PKNCA.options(conc.blq="foo", check=TRUE),
               regexp="conc.blq must either be a finite number or the text 'drop' or 'keep'")
  expect_error(PKNCA.options(conc.blq=c(1, 2), check=TRUE),
               regexp="conc.blq must be a scalar")
  expect_error(PKNCA.options(conc.blq=NA, check=TRUE),
               regexp="conc.blq must not be NA")

  # Confirm that list-style input also works
  expect_equal(PKNCA.options(conc.blq=list(first="drop", middle=5, last="keep"),
                             check=TRUE),
               list(first="drop", middle=5, last="keep"))
  expect_error(PKNCA.options(conc.blq=list(first="drop", middle=5, last="keep",
                                           foo=5),
                             check=TRUE),
               regexp="When given as a list, conc.blq must only have elements named 'first', 'middle' and 'last' or 'before.tmax' and 'after.tmax'.")
  expect_error(PKNCA.options(conc.blq=list(first="drop", middle=5),
                             check=TRUE),
               regexp="When given as a list, conc.blq must include all elements named 'first', 'middle' and 'last' or 'before.tmax' and 'after.tmax'.")

  # Confirm that before.tmax and after.tmax must be specified together
  expect_error(PKNCA.options(conc.blq=list(after.tmax="drop"), check=TRUE),
               regexp="When given as a list, conc.blq must include all elements named 'first', 'middle' and 'last' or 'before.tmax' and 'after.tmax'.")

  # Confirm that before.tmax and after.tmax work correctly
  expect_equal(PKNCA.options(conc.blq=list(before.tmax="drop", after.tmax="keep"),
                             check=TRUE),
               list(before.tmax="drop", after.tmax="keep"))

  # Confirm that first/middle/last and before.tmax/after.tmax cannot be mixed and need to be complete
  names_tmax <- c("before.tmax", "after.tmax")
  names_tlast <- c("first", "middle", "last")
  all_combinations <- unlist(lapply(seq_along(c(names_tlast, names_tmax)), function(x) combn(c(names_tlast, names_tmax), x, simplify = FALSE)), recursive = FALSE)
  for (i in seq_along(all_combinations)) {
    conc.blq.i <- as.list(setNames(rep(0, length(all_combinations[[i]])), all_combinations[[i]]))
    are.names.mixed <- any(names_tmax %in% all_combinations[[i]]) && any(names_tlast %in% all_combinations[[i]])
    are.names.incomplete_tmax <- any(names_tmax %in% all_combinations[[i]]) && !all(names_tmax %in% all_combinations[[i]])
    are.names.incomplete_tlast <- any(names_tlast %in% all_combinations[[i]]) && !all(names_tlast %in% all_combinations[[i]])

    if (are.names.mixed) {
      # Lists with mixed names for both BLQ strategies should provide an error
      expect_error(PKNCA.options(conc.blq=conc.blq.i, check=TRUE),
                   regexp="When given as a list, prevent mixing arguments of different BLQ strategies\\.")
    } else if (are.names.incomplete_tmax || are.names.incomplete_tlast) {
      # Lists with missing names for either BLQ strategy should provide an error
      expect_error(PKNCA.options(conc.blq=conc.blq.i, check=TRUE),
                   regexp="When given as a list, conc.blq must include all elements named 'first', 'middle' and 'last' or 'before.tmax' and 'after.tmax'.")
    }
  }

  # first.tmax
  expect_equal(PKNCA.options(first.tmax=FALSE, check=TRUE),
               FALSE)
  expect_error(PKNCA.options(first.tmax=c(FALSE, TRUE), check=TRUE),
               regexp="first.tmax must be a scalar")
  # Conversion works
  expect_warning(v1 <- PKNCA.options(first.tmax="T", check=TRUE),
                 regexp="Converting first.tmax to a logical value: TRUE")
  expect_equal(v1, TRUE)
  expect_warning(v1 <- PKNCA.options(first.tmax=1, check=TRUE),
                 regexp="Converting first.tmax to a logical value: TRUE")
  expect_equal(v1, TRUE)
  expect_error(PKNCA.options(first.tmax=NA, check=TRUE),
               regexp="first.tmax may not be NA")
  expect_error(PKNCA.options(first.tmax="x", check=TRUE),
               regexp="Could not convert first.tmax to a logical value")

  # min.hl.points
  expect_equal(PKNCA.options(min.hl.points=3, check=TRUE),
               3)
  expect_error(PKNCA.options(min.hl.points=c(3, 4), check=TRUE),
               regexp="min.hl.points must be a scalar")
  expect_error(PKNCA.options(min.hl.points=factor(3), check=TRUE),
               regexp="min.hl.points cannot be a factor")
  expect_error(PKNCA.options(min.hl.points="a", check=TRUE),
               regexp="min.hl.points must be a number")
  expect_error(PKNCA.options(min.hl.points=1.5, check=TRUE),
               regexp="min.hl.points must be >=2")
  expect_warning(v1 <- PKNCA.options(min.hl.points=2.5, check=TRUE),
                 regexp="Non-integer given for min.hl.points; rounding to nearest integer")
  # Note that R uses the engineer's rule of rounding
  expect_equal(v1, 2)

  # min.span.ratio
  expect_equal(PKNCA.options(min.span.ratio=2, check=TRUE),
               2)
  expect_error(PKNCA.options(min.span.ratio=0, check=TRUE),
               regexp="min.span.ratio must be > 0")
  expect_error(PKNCA.options(min.span.ratio=c(2, 1), check=TRUE),
               regexp="min.span.ratio must be a scalar")
  expect_error(PKNCA.options(min.span.ratio=factor(1), check=TRUE),
               regexp="min.span.ratio cannot be a factor")
  expect_error(PKNCA.options(min.span.ratio="a", check=TRUE),
               regexp="min.span.ratio must be a number")
  expect_warning(PKNCA.options(min.span.ratio=1, check=TRUE),
                 regexp="min.span.ratio is usually >= 2")

  # max.aucinf.pext
  expect_equal(PKNCA.options(max.aucinf.pext=20, check=TRUE),
               20)
  expect_error(PKNCA.options(max.aucinf.pext=0, check=TRUE),
               regexp="max.aucinf.pext must be > 0")
  expect_error(PKNCA.options(max.aucinf.pext=c(2, 1), check=TRUE),
               regexp="max.aucinf.pext must be a scalar")
  expect_error(PKNCA.options(max.aucinf.pext=factor(1), check=TRUE),
               regexp="max.aucinf.pext cannot be a factor")
  expect_error(PKNCA.options(max.aucinf.pext="a", check=TRUE),
               regexp="max.aucinf.pext must be a number")
  expect_warning(PKNCA.options(max.aucinf.pext=25.1, check=TRUE),
                 regexp="max.aucinf.pext is usually <=25")
  expect_warning(PKNCA.options(max.aucinf.pext=0.1, check=TRUE),
                 regexp="max.aucinf.pext is on the percent not ratio scale, value given is <1%")

  # min.hl.r.squared
  expect_equal(PKNCA.options(min.hl.r.squared=0.9, check=TRUE),
               0.9)
  expect_error(PKNCA.options(min.hl.r.squared=0, check=TRUE),
               regexp="min.hl.r.squared must be between 0 and 1, exclusive")
  expect_error(PKNCA.options(min.hl.r.squared=c(2, 1), check=TRUE),
               regexp="min.hl.r.squared must be a scalar")
  expect_error(PKNCA.options(min.hl.r.squared=factor(1), check=TRUE),
               regexp="min.hl.r.squared cannot be a factor")
  expect_error(PKNCA.options(min.hl.r.squared="a", check=TRUE),
               regexp="min.hl.r.squared must be a number")
  expect_warning(PKNCA.options(min.hl.r.squared=0.89, check=TRUE),
                 regexp="min.hl.r.squared is usually >= 0.9")

  # tau.choices
  expect_equal(PKNCA.options(tau.choices=NA, check=TRUE),
               NA)
  expect_equal(PKNCA.options(tau.choices=c(1, 2), check=TRUE),
               c(1, 2))
  expect_error(PKNCA.options(tau.choices=c(NA, 1), check=TRUE),
               regexp="tau.choices may not include NA and be a vector")
  expect_error(PKNCA.options(tau.choices="x", check=TRUE),
               regexp="tau.choices must be a number")

  # Reset all options to their default to ensure that any subsequent
  # tests work correctly.
  PKNCA.options(default=TRUE)
})

test_that("PKNCA.choose.option", {
  current.options <- PKNCA.options()
  # If nothing is given for the non-default options, the default
  # option is returned.
  expect_equal(PKNCA.choose.option("conc.na"),
               current.options[["conc.na"]])
  # If an invalid option is requested, it gives an error
  expect_error(PKNCA.choose.option("foo"),
               regexp="PKNCA.options does not have value\\(s\\) for foo.")
  # It gives an error even if there is a value in the option list.
  # Note that the error is different because it checks the passed-in
  # options while it just extracts the default options.
  expect_error(PKNCA.choose.option("foo", options=list(foo="bar")),
               regexp="Invalid setting for PKNCA: foo")
  # When given in the options list, it will choose that instead of
  # the default value.
  expect_equal(PKNCA.choose.option("max.aucinf.pext",
                                   options=list(max.aucinf.pext=10)),
               10)
  # When multiple values are given in the options, it chooses the
  # right one and ignores all the others (so invalid options can be
  # listed as long as they are not used).
  expect_equal(PKNCA.choose.option("max.aucinf.pext",
                                   options=list(
                                     foo="bar",
                                     max.aucinf.pext=10)),
               10)

  # Manage NULL and the "value" argument
  expect_equal(PKNCA.choose.option("single.dose.aucs"),
               PKNCA.options(name="single.dose.aucs"),
               info="PKNCA.choose.option gives the default option when nothing else is given")
  expect_equal(PKNCA.choose.option("single.dose.aucs",
                                   options=list(single.dose.aucs=data.frame(start=0, end=1, cmax=TRUE))),
               check.interval.specification(data.frame(start=0, end=1, cmax=TRUE)),
               info="PKNCA.choose.option gives the default option when nothing else is given")
  expect_equal(PKNCA.choose.option("single.dose.aucs",
                                   value=data.frame(start=0, end=1, cmax=TRUE, tmax=TRUE),
                                   options=list(single.dose.aucs=data.frame(start=0, end=1, cmax=TRUE))),
               check.interval.specification(data.frame(start=0, end=1, cmax=TRUE, tmax=TRUE)),
               info="PKNCA.choose.option gives the default option when nothing else is given")
  expect_equal(PKNCA.choose.option("single.dose.aucs",
                                   value=NULL,
                                   options=list(single.dose.aucs=data.frame(start=0, end=1, cmax=TRUE))),
               check.interval.specification(data.frame(start=0, end=1, cmax=TRUE)),
               info="PKNCA.choose.option gives the default option when nothing else is given")
})

test_that("PKNCA.set.summary input checking", {
  # Get the current state to reset it at the end
  initial.summary.set <- PKNCA.set.summary()
  PKNCA.set.summary(reset=TRUE)
  # Confirm that reset actually resets the summary settings
  expect_equal(PKNCA.set.summary(), list())

  # name must already be defined
  expect_error(PKNCA.set.summary("blah"),
               regexp="You must first define the parameter name with add.interval.col")
  # point must be a function
  expect_error(PKNCA.set.summary("auclast", description="A", point="a"),
               regexp="`point` must be a function")
  # description is required and must be a scalar character string
  expect_error(
    PKNCA.set.summary("auclast", description=1),
    regexp="`description` must be a character string",
    fixed=TRUE
  )
  expect_error(
    PKNCA.set.summary("auclast", description=c("A", "B")),
    regexp="`description` must be a scalar.",
    fixed=TRUE
  )
  expect_error(PKNCA.set.summary("auclast", description=1))
  # spread must be a function
  expect_error(PKNCA.set.summary("auclast", description="A", point=mean, spread="a"),
               regexp="spread must be a function")
  # Rounding must either be a function or a list
  expect_error(PKNCA.set.summary("auclast", description="A", point=mean, spread=sd,
                                 rounding="a"),
               regexp="rounding must be either a list or a function")
  expect_error(PKNCA.set.summary("auclast", description="A", point=mean, spread=sd,
                                 rounding=list(foo=3, bar=4)),
               regexp="rounding must have a single value in the list")
  expect_error(PKNCA.set.summary("auclast", description="A", point=mean, spread=sd,
                                 rounding=list(foo=3)),
               regexp="When a list, rounding must have a name of either 'signif' or 'round'")
  # An initial setting works
  expect_equal(
    PKNCA.set.summary(
      "auclast", description="A", point=mean, spread=sd,
      rounding=round),
    list(auclast=
           list(
             description = "A",
             point=mean,
             spread=sd,
             rounding=round
           )
    )
  )
  # Changing a setting works
  expect_equal(PKNCA.set.summary("auclast", description="A", point=mean, spread=sd,
                                 rounding=list(round=2)),
               list(auclast=list(description="A", point=mean, spread=sd,
                                 rounding=list(round=2))))
  # Changing a vector of settings works
  PKNCA.set.summary(reset=TRUE)
  expect_equal(
    PKNCA.set.summary(
      name=c("cmax", "auclast"),
      description="A",
      point=mean, spread=sd,
      rounding=list(round=2)
    ),
    list(cmax=
           list(
             description="A",
             point=mean, spread=sd,
             rounding=list(round=2)
           ),
         auclast=
           list(
             description="A",
             point=mean, spread=sd,
             rounding=list(round=2)
           )
    )
  )

  # Reset all the values to the defaults
  PKNCA.set.summary(reset=TRUE)
  for (n in names(initial.summary.set)) {
    tmp <- initial.summary.set[[n]]
    tmp$name <- n
    do.call(PKNCA.set.summary, tmp)
  }
})

test_that("PKNCA.set.summary exists for all paramters", {
  missing_summaries <-
    setdiff(setdiff(names(get.interval.cols()),
                    names(PKNCA.set.summary())),
            c("start", "end"))
  expect_true(length(missing_summaries) == 0,
              info="All parameters have summary functions")
})

test_that("PKNCA.options.describe", {
  expect_equal(PKNCA:::PKNCA.options.describe("adj.r.squared.factor"),
               PKNCA:::.PKNCA.option.check[["adj.r.squared.factor"]](description=TRUE),
               info="Option descriptions are provided accurately.")
})

test_that("PKNCA.options fails when setting defaults and another option simultaneously", {
  expect_error(
    PKNCA.options(default=TRUE, tau.choices=24),
    regexp="Cannot set default and set new options at the same time.",
    fixed=TRUE
  )
})

#' Handle NA values in the concentration measurements as requested by the user.
#'
#' NA concentrations (and their associated times) will be removed then the BLQ
#' values in the middle
#'
#' @inheritParams assert_conc_time
#' @inheritParams PKNCA.choose.option
#' @param \dots Additional items to add to the data frame
#' @param conc.na How to handle NA concentrations?  Either 'drop' or a number to
#'   impute.
#' @param check Run [assert_conc_time()]?
#' @returns The concentration and time measurements (data frame) filtered
#'   and cleaned as requested relative to NA in the concentration.
#' @family Data cleaners
#' @export
clean.conc.na <- function(conc, time, ...,
                          options=list(),
                          conc.na=NULL,
                          check=TRUE) {
  conc.na <- PKNCA.choose.option(name="conc.na", value=conc.na, options=options)
  if (check)
    assert_conc_time(conc, time)
  # Prep it as a data frame
  ret <- data.frame(conc, time, ..., stringsAsFactors=FALSE)
  if (conc.na %in% "drop") {
    # If it is set to "drop" then omit the NA concentrations
    ret <- ret[!is.na(conc),]
  } else if (is.numeric(conc.na)) {
    ret$conc[is.na(conc)] <- conc.na
  } else {
    # This case should already have been captured by the PKNCA.options
    # call above.
    stop("Unknown how to handle conc.na") # nocov
  }
  ret
}

#' Handle BLQ values in the concentration measurements as requested by the user.
#'
#' @inheritParams assert_conc_time
#' @inheritParams PKNCA.choose.option
#' @param \dots Additional arguments passed to clean.conc.na
#' @param conc.blq How to handle a BLQ value that is between above LOQ values?
#'   See details for description.
#' @param conc.na How to handle NA concentrations.  (See [clean.conc.na()])
#' @param check Run [assert_conc_time()]?
#' @returns The concentration and time measurements (data frame) filtered and
#'   cleaned as requested relative to BLQ in the middle.
#'
#' @details `NA` concentrations (and their associated times) will be handled as
#'   described in [clean.conc.na()] before working with the BLQ values.  The
#'   method for handling NA concentrations can affect the output of which points
#'   are considered BLQ and which are considered "middle".  Values are
#'   considered BLQ if they are 0.
#'
#'   `conc.blq` can be set either a scalar indicating what should be done for
#'   all BLQ values or a list with elements either named "first", "middle" and "last" or "before.tmax" and "after.tmax"
#'   each set to a scalar.
#'
#' The meaning of each of the list elements is:
#' \describe{
#'   \item{first}{Values up to the first non-BLQ value.  Note
#'     that if all values are BLQ, this includes all values.}
#'   \item{middle}{Values that are BLQ between the first and last
#'     non-BLQ values.}
#'   \item{last}{Values that are BLQ after the last non-BLQ value}
#'   \item{before.tmax}{Values that are BLQ before the time at first maximum concentration}
#'   \item{after.tmax}{Values that are BLQ after the time at first maximum concentration}
#' }
#'
#' The valid settings for each are:
#' \describe{
#'   \item{"drop"}{Drop the BLQ values}
#'   \item{"keep"}{Keep the BLQ values}
#'   \item{a number}{Set the BLQ values to that number}
#' }
#'
#' @family Data cleaners
#' @export
clean.conc.blq <- function(conc, time,
                           ...,
                           options=list(),
                           conc.blq=NULL,
                           conc.na=NULL,
                           check=TRUE) {
  conc.blq <- PKNCA.choose.option(name="conc.blq", value=conc.blq, options=options)
  conc.na <- PKNCA.choose.option(name="conc.na", value=conc.na, options=options)
  if (check) {
    assert_conc_time(conc, time)
  }
  # Handle NA concentrations and make the data frame
  ret <- clean.conc.na(conc, time, ..., conc.na=conc.na, check=FALSE)
  # If all data has been excluded, then don't do anything
  if (nrow(ret) > 0) {
    tfirst <- pk.calc.tfirst(ret$conc, ret$time, check=FALSE)
    tlast <- pk.calc.tlast(ret$conc, ret$time, check=FALSE)
    tmax <- pk.calc.tmax(ret$conc, ret$time, check=FALSE)

    # If all measurements are BLQ
    if (all(ret$conc == 0)){
      # Apply "first" BLQ rule to everything for tfirst/tlast
      tfirst <- max(ret$time)
      tlast <- tfirst + 1

      # Apply "before.tmax" BLQ rule to everything for tmax
      tmax <- max(ret$time)
    }

    # Depending on the specified argument perform the corresponding action
    for (i in seq_len(length(conc.blq))) {
      # Set the mask to apply the rule to
      time_type <- names(conc.blq)[i]
      if (is.null(time_type) & length(conc.blq) == 1) {
        mask <- ret$conc %in% 0
      } else if (time_type == "first") {
        mask <- ret$time <= tfirst & ret$conc %in% 0
      } else if (time_type == "middle") {
        mask <- tfirst < ret$time & ret$time < tlast & ret$conc %in% 0
      } else if (time_type == "last") {
        mask <- tlast <= ret$time & ret$conc %in% 0
      } else if (time_type == "before.tmax") {
        mask <- ret$time < tmax & ret$conc %in% 0
      } else if (time_type == "after.tmax") {
        mask <- tmax <= ret$time & ret$conc %in% 0
      } else {
        stop("There is a bug in cleaning the conc.blq with position names") # nocov
      }
      # Choose the rule to apply
      this_rule <- unname(conc.blq)[[i]]

      if (this_rule %in% "keep") {
        # Do nothing
      } else if (this_rule %in% "drop") {
        ret <- ret[!mask,]
      } else if (is.numeric(this_rule)) {
        ret$conc[mask] <- this_rule
      } else {
        # This case should already have been captured by the PKNCA.options
        # call above.
        stop(sprintf("Unknown how to handle conc.blq rule %s", # nocov
                     as.character(this_rule)))                 # nocov
      }
    }
  }
  ret
}

test_that("clean.conc.na", {
  # Check that it gives errors for invalid conc.na values
  expect_error(clean.conc.na(conc=1, time=1, conc.na="foo"),
               regexp="conc.na must either be a finite number or the text 'drop'")

  # It drops NA values if requested (even if they are the only value)
  expect_warning(v1 <-
    clean.conc.na(conc=as.numeric(NA), time=1, conc.na="drop"))
  expect_equal(v1,
               data.frame(conc=numeric(), time=numeric()))
  expect_warning(v2 <-
    clean.conc.na(conc=as.numeric(c(NA, NA)), time=1:2,
                  conc.na="drop"))
  expect_equal(v2,
               data.frame(conc=numeric(), time=numeric()))
  expect_equal(clean.conc.na(conc=c(1, NA), time=1:2, conc.na="drop"),
               data.frame(conc=1, time=1))
  expect_equal(clean.conc.na(conc=c(1, NA, 2), time=1:3, conc.na="drop"),
               data.frame(conc=1:2, time=c(1, 3)),
               ignore_attr=TRUE)

  # It also works with a number as the conc.na value
  expect_warning(v3 <-
    clean.conc.na(conc=as.numeric(NA), time=1, conc.na=5))
  expect_equal(v3,
               data.frame(conc=5, time=1))
  expect_warning(v4 <-
    clean.conc.na(conc=c(NA, NA), time=1:2, conc.na=5))
  expect_equal(v4,
               data.frame(conc=c(5, 5), time=1:2))
  expect_equal(clean.conc.na(conc=c(1, NA), time=1:2, conc.na=5),
               data.frame(conc=c(1, 5), time=1:2))
  expect_equal(clean.conc.na(conc=c(1, NA, 2), time=1:3, conc.na=5),
               data.frame(conc=c(1, 5, 2), time=1:3))

  # It doesn't touch BLQ
  expect_equal(clean.conc.na(conc=0, time=1, conc.na="drop"),
               data.frame(conc=0, time=1))
  expect_equal(clean.conc.na(conc=c(0, 0), time=1:2, conc.na="drop"),
               data.frame(conc=c(0, 0), time=1:2))
  expect_equal(clean.conc.na(conc=c(1, 0), time=1:2, conc.na="drop"),
               data.frame(conc=c(1, 0), time=1:2))
  expect_equal(clean.conc.na(conc=c(1, 0, 2), time=1:3, conc.na="drop"),
               data.frame(conc=c(1, 0, 2), time=1:3))

  # Additional arguments are returned into the data frame and
  # filtered if NA conc goes with them.
  expect_equal(
    clean.conc.na(
      conc=1:3,
      time=1:3,
      extra=c("a", "b", "c"),
      conc.na="drop"
    ),
    data.frame(
      conc=1:3,
      time=1:3,
      extra=c("a", "b", "c"),
      stringsAsFactors=FALSE
    )
  )
  expect_equal(
    clean.conc.na(
      conc=c(1, NA, 2),
      time=1:3,
      extra=c("a", "b", "c"),
      conc.na="drop"
    ),
    data.frame(
      conc=1:2,
      time=c(1, 3),
      extra=c("a", "c"),
      stringsAsFactors=FALSE
    ),
    ignore_attr=TRUE
  )
})

test_that("clean.conc.blq", {
  # If there are no BLQ values or NA values, it does nothing
  d.test <- data.frame(conc=1, time=1)
  expect_equal(clean.conc.blq(d.test$conc, d.test$time,
                                conc.blq="drop", conc.na="drop"),
               d.test)

  # If there are NA values, it runs conc.na on those and returns that
  d.test <- data.frame(conc=c(1, NA), time=1:2)
  expect_equal(clean.conc.blq(d.test$conc, d.test$time,
                                conc.blq="drop", conc.na="drop"),
               clean.conc.na(d.test$conc, d.test$time, conc.na="drop"))

  # If there are BLQ values at the beginning, drops those if given a
  # simple drop
  d.test <- data.frame(conc=c(0, 1), time=1:2)
  expect_equal(clean.conc.blq(d.test$conc, d.test$time,
                                conc.blq="drop", conc.na="drop"),
               d.test[2,])
  # It leaves beginning BLQs if instructed with a list.
  d.test <- data.frame(conc=c(0, 1), time=1:2)
  expect_equal(clean.conc.blq(d.test$conc, d.test$time,
                              conc.blq=list(
                                first="keep",
                                middle="drop",
                                last="drop"),
                              conc.na="drop"),
               d.test)

  # Errors in how to handle first/middle/last rules are caught.
  d.test <- data.frame(conc=c(0, 1), time=1:2)
  expect_error(clean.conc.blq(d.test$conc, d.test$time,
                              conc.blq=list(
                                first="foo",
                                middle="drop",
                                last="drop"),
                              conc.na="drop"),
               regexp="conc.blq must either be a finite number or the text 'drop' or 'keep'")
  expect_error(clean.conc.blq(d.test$conc, d.test$time,
                              conc.blq=list(
                                first="keep",
                                middle="foo",
                                last="drop"),
                              conc.na="drop"),
               regexp="conc.blq must either be a finite number or the text 'drop' or 'keep'")
  expect_error(clean.conc.blq(d.test$conc, d.test$time,
                              conc.blq=list(
                                first="keep",
                                middle="drop",
                                last="foo"),
                              conc.na="drop"),
               regexp="conc.blq must either be a finite number or the text 'drop' or 'keep'")

  # If there are BLQ values at the end, it drops them if the
  # instructions are generic drop.
  d.test <- data.frame(conc=c(1, 0), time=1:2)
  expect_equal(clean.conc.blq(d.test$conc, d.test$time,
                                conc.blq="drop", conc.na="drop"),
               d.test[1,])
  # And it keeps them if instructed with a list.
  d.test <- data.frame(conc=c(1, 0), time=1:2)
  expect_equal(clean.conc.blq(d.test$conc, d.test$time,
                              conc.blq=list(
                                first="drop",
                                middle="drop",
                                last="keep"),
                              conc.na="drop"),
               d.test)

  # If there are BLQ values at the beginning and end, it drops those
  # if given a single instruction.
  d.test <- data.frame(conc=c(0, 1, 0), time=1:3)
  expect_equal(clean.conc.blq(d.test$conc, d.test$time,
                                conc.blq="drop", conc.na="drop"),
               d.test[2,])

  # If all values are BLQ, drops all rows
  d.test <- data.frame(conc=0, time=1:3)
  expect_equal(
    clean.conc.blq(d.test$conc, d.test$time,
                   conc.blq="drop", conc.na="drop"),
    d.test[logical(),]
  )

  # If there are BLQ values in the middle, it drops or keeps those or
  # sets them to a number
  d.test <- data.frame(conc=c(1, 0, 2), time=1:3)
  expect_equal(clean.conc.blq(d.test$conc, d.test$time,
                                conc.blq="drop", conc.na="drop"),
               d.test[-2,])

  d.test <- data.frame(conc=c(1, 0, 2), time=1:3)
  expect_equal(clean.conc.blq(d.test$conc, d.test$time,
                                conc.blq="keep", conc.na="drop"),
               d.test)

  d.test <- data.frame(conc=c(1, 0, 2), time=1:3)
  d.result <- data.frame(conc=c(1, 0.5, 2), time=1:3)
  expect_equal(
    clean.conc.blq(
      d.test$conc, d.test$time,
      conc.blq=0.5, conc.na="drop"
    ),
    d.result
  )
  expect_equal(
    clean.conc.blq(
      d.test$conc, d.test$time,
      conc.blq=list(first=0, middle=0.5, last=0), conc.na="drop"
    ),
    d.result,
    info="fix related item to #145"
  )

  # If there are BLQ values at the beginning, middle, and end, it
  # only drops all of them or drops them selectively as instructed.
  d.test <- data.frame(conc=c(0, 1, 0, 2, 0), time=1:5)
  expect_equal(clean.conc.blq(d.test$conc, d.test$time,
                                conc.blq="drop", conc.na="drop"),
               d.test[c(2, 4),])
  for (first in c("drop", "keep")) {
    for (middle in c("drop", "keep")) {
      for (last in c("drop", "keep")) {
        expect_equal(clean.conc.blq(d.test$conc, d.test$time,
                                    conc.blq=list(
                                      first=first,
                                      middle=middle,
                                      last=last),
                                    conc.na=0),
                     d.test[c(first %in% "keep",
                              TRUE,
                              middle %in% "keep",
                              TRUE,
                              last %in% "keep"),],
                     info=paste(first, middle, last))
      }
    }
  }

  # When conc.na is 0, it drops those.
  d.test <- data.frame(conc=c(0, 1, NA, 2, 0), time=1:5)
  expect_equal(clean.conc.blq(d.test$conc, d.test$time,
                                conc.blq="drop", conc.na=0),
               d.test[c(2, 4),])

  # When conc.na is a number, it keeps those
  d.test <- data.frame(conc=c(0, 1, NA, 2, 0), time=1:5)
  d.result <- data.frame(conc=c(0, 1, 0.5, 2, 0), time=1:5)
  expect_equal(clean.conc.blq(d.test$conc, d.test$time,
                                conc.blq="drop", conc.na=0.5),
               d.result[2:4,])

  # It passes additional to be part of the output data frame
  d.test <- data.frame(conc=c(0, 1, NA, 2, 0), time=1:5, more=6:10)
  d.result <- data.frame(conc=c(0, 1, 0.5, 2, 0), time=1:5, more=6:10)
  expect_equal(clean.conc.blq(d.test$conc, d.test$time,
                              more=d.test$more,
                              conc.blq="drop", conc.na=0.5),
               d.result[2:4,])
  d.test <- data.frame(conc=c(0, 1, NA, 2, 0), time=1:5, more=6:10)
  expect_equal(clean.conc.blq(d.test$conc, d.test$time,
                              more=d.test$more,
                              conc.blq="drop", conc.na="drop"),
               d.test[c(2,4),])


    # If there are BLQ values before tmax, drops those if given a simple drop
    d.test <- data.frame(conc=c(0, 1), time=1:2)
    expect_equal(
      clean.conc.blq(
        d.test$conc, d.test$time,
        conc.blq=list(before.tmax="drop", after.tmax="drop"), conc.na="drop"),
      d.test[2, ]
    )

    # If there are BLQ values after tmax, drops those if given a simple drop
    d.test <- data.frame(conc=c(1, 0), time=1:2)
    expect_equal(
      clean.conc.blq(
        d.test$conc, d.test$time,
        conc.blq=list(before.tmax="drop", after.tmax="drop"), conc.na="drop"),
      d.test[1,]
    )

    # If there are BLQ values before and after tmax, drops those if given a single instruction
    d.test <- data.frame(conc=c(0, 1, 0), time=1:3)
    expect_equal(
      clean.conc.blq(
        d.test$conc, d.test$time,
        conc.blq=list(before.tmax="drop", after.tmax="drop"), conc.na="drop"
      ),
      d.test[2,]
    )

    # If all values are BLQ, drops all rows
    d.test <- data.frame(conc=0, time=1:3)
    expect_equal(
      clean.conc.blq(
        d.test$conc, d.test$time,
        conc.blq=list(before.tmax="drop", after.tmax="drop"), conc.na="drop"
      ),
      d.test[logical(),]
    )

    # If there are BLQ values in the middle, it drops or keeps those or sets them to a number
    d.test <- data.frame(conc=c(1, 0, 2), time=1:3)
    expect_equal(
      clean.conc.blq(
        d.test$conc, d.test$time,
        conc.blq=list(before.tmax="drop", after.tmax="drop"), conc.na="drop"
      ),
      d.test[-2,]
    )

    d.test <- data.frame(conc=c(1, 0, 2), time=1:3)
    expect_equal(
      clean.conc.blq(
        d.test$conc, d.test$time,
        conc.blq=list(before.tmax="keep", after.tmax="keep"), conc.na="drop"
      ),
      d.test
    )

    d.test <- data.frame(conc=c(1, 0, 2), time=1:3)
    d.result <- data.frame(conc=c(1, 0.5, 2), time=1:3)
    expect_equal(
      clean.conc.blq(
        d.test$conc, d.test$time,
        conc.blq=list(before.tmax=0.5, after.tmax=0.5), conc.na="drop"
      ),
      d.result
    )

    # If there are BLQ values before and after tmax, it only drops all of them
    # or drops them selectively as instructed
    d.test <- data.frame(conc=c(0, 1, 0, 2, 0), time=1:5)
    expect_equal(
      clean.conc.blq(
        d.test$conc, d.test$time,
        conc.blq=list(before.tmax="drop", after.tmax="drop"), conc.na="drop"
      ),
      d.test[c(2, 4),]
    )
    for (before.tmax in c("drop", "keep")) {
      for (after.tmax in c("drop", "keep")) {
        expect_equal(
          clean.conc.blq(
            d.test$conc, d.test$time,
            conc.blq=list(before.tmax=before.tmax, after.tmax=after.tmax),
            conc.na=0
          ),
          d.test[
            c(before.tmax %in% "keep",
              TRUE,
              before.tmax %in% "keep",
              TRUE,
              after.tmax %in% "keep"),],
          info=paste(before.tmax, after.tmax)
        )
      }
    }

    # When conc.na is 0, it drops those
    d.test <- data.frame(conc=c(0, 1, NA, 2, 0), time=1:5)
    expect_equal(
      clean.conc.blq(
        d.test$conc, d.test$time,
        conc.blq=list(before.tmax="drop", after.tmax="drop"), conc.na=0
      ),
      d.test[c(2, 4),]
    )

    # When conc.na is a number, it keeps those
    d.test <- data.frame(conc=c(0, 1, NA, 2, 0), time=1:5)
    d.result <- data.frame(conc=c(0, 1, 0.5, 2, 0), time=1:5)
    expect_equal(
      clean.conc.blq(
        d.test$conc, d.test$time,
        conc.blq=list(before.tmax="drop", after.tmax="drop"), conc.na=0.5
      ),
      d.result[2:4,]
    )

    # It passes additional to be part of the output data frame
    d.test <- data.frame(conc=c(0, 1, NA, 2, 0), time=1:5, more=6:10)
    d.result <- data.frame(conc=c(0, 1, 0.5, 2, 0), time=1:5, more=6:10)
    expect_equal(
      clean.conc.blq(
        d.test$conc, d.test$time,
        more=d.test$more,
        conc.blq=list(before.tmax="drop", after.tmax="drop"), conc.na=0.5
      ),
      d.result[2:4,]
    )
    d.test <- data.frame(conc=c(0, 1, NA, 2, 0), time=1:5, more=6:10)
    expect_equal(
      clean.conc.blq(
        d.test$conc, d.test$time,
        more=d.test$more,
        conc.blq=list(before.tmax="drop", after.tmax="drop"), conc.na="drop"
      ),
      d.test[c(2,4),]
    )
})

#' Get the impute function from either the intervals column or from the method
#'
#' @param intervals the data.frame of intervals
#' @param impute the imputation definition
#' @return The imputation function vector
get_impute_method <- function(intervals, impute) {
  stopifnot(length(impute) == 1)
  checkmate::assert_data_frame(intervals)
  if (impute %in% names(intervals)) {
    impute_funs <- intervals[[impute]]
  } else if (is.na(impute) && "impute" %in% names(intervals)) {
    impute_funs <- intervals$impute
  } else {
    impute_funs <- impute
  }
  checkmate::assert_character(impute_funs)
  impute_funs
}

#' Add the imputation column to the intervals, if it is not already there
#'
#' @param object The PKNCAdata object to impute data within
#' @return The PKNCAdata object with an impute column added to the intervals (if
#'   it is not already there) and the object$impute set to that column name
#' @keywords internal
add_impute_to_intervals <- function(object) {
  if (is.na(object$impute)) {
    rlang::abort(
      message = "add_impute_to_intervals cannot have an NA imputation defined",
      class = "pknca_add_impute_to_intervals_NA"
    )
  }
  # Do nothing if the impute column is already in the data
  if (!(object$impute %in% names(object$intervals))) {
    impute_col <-
      if (!("impute" %in% names(object$intervals))) {
        "impute"
      } else {
        paste0(max(grep(x = names(object$intervals), "^impute", value = TRUE)), "X")
      }
    object$intervals[[impute_col]] <- object$impute
    object$impute <- impute_col
  }
  object
}

#' Methods for imputation of data with PKNCA
#' @name PKNCA_impute_method
#' @return A data.frame with one column named conc with imputed concentrations
#'   and one column named time with the times.
NULL

#' @describeIn PKNCA_impute_method Add a new concentration of 0 at the start
#'   time, even if a nonzero concentration exists at that time (usually used
#'   with single-dose data)
#' @inheritParams pk.calc.auxc
#' @inheritParams assert_intervaltime_single
#' @param ... ignored
#' @export
PKNCA_impute_method_start_conc0 <- function(conc, time, start=0, ..., options = list()) {
  ret <- data.frame(conc = conc, time = time)
  mask_start <- time %in% start
  if (any(mask_start)) {
    ret$conc[mask_start] <- 0
  } else {
    ret <- rbind(ret, data.frame(time = start, conc = 0))
    ret <- ret[order(ret$time), ]
  }
  ret
}

#' @describeIn PKNCA_impute_method Add a new concentration of the minimum during
#'   the interval at the start time (usually used with multiple-dose data)
#' @export
PKNCA_impute_method_start_cmin <- function(conc, time, start, end, ..., options = list()) {
  ret <- data.frame(conc = conc, time = time)
  mask_start <- time %in% start
  if (!any(mask_start)) {
    all_concs <- conc[start <= time & time <= end]
    if (!all(is.na(all_concs))) {
      cmin <- min(all_concs, na.rm = TRUE)
      ret <- rbind(ret, data.frame(time = start, conc = cmin))
      ret <- ret[order(ret$time), ]
    }
  }
  ret
}

#' @describeIn PKNCA_impute_method Shift a predose concentration to become the
#'   time zero concentration (only if a time zero concentration does not exist)
#' @param max_shift The maximum amount of time to shift a concentration forward
#'   (defaults to 5% of the interval duration, i.e. `0.05*(end - start)`, if
#'   `is.finite(end)`, and when `is.infinite(end)`, defaults to 5% of the time
#'   from start to `max(time)`)
#' @inheritParams pk.nca.interval
#' @export
PKNCA_impute_method_start_predose <- function(conc, time, start, end, conc.group, time.group, ..., max_shift = NA_real_, options = list()) {
  ret <- data.frame(conc = conc, time = time)
  if (is.na(max_shift)) {
    if (is.infinite(end)) {
      shift_end <- max(time)
    } else {
      shift_end <- end
    }
    max_shift <- 0.05 * (shift_end - start)
  }
  # determine if the start time is already in the
  mask_start <- time %in% start
  if (!any(mask_start)) {
    mask_predose <- time.group < start
    if (any(mask_predose)) {
      time_predose <- max(time.group[mask_predose])
      if ((-time_predose) <= max_shift) {
        mask_predose_change <- time.group == time_predose
        ret_predose <- data.frame(conc = conc.group[mask_predose_change], time = start)
        ret <- dplyr::bind_rows(ret_predose, ret)
      }
    }
  }
  ret
}

#' Separate out a vector of PKNCA imputation methods into a list of functions
#'
#' An error will be raised if the functions are not found.
#'
#' This function is not for use by users of PKNCA.
#'
#' @param x The character vector of PKNCA imputation method functions (without
#'   the `PKNCA_impute_method_` part)
#' @return A list of character vectors of functions to run.
#' @keywords Internal
PKNCA_impute_fun_list <- function(x) {
  if (all(is.na(x))) {
    x <- rep(NA_character_, length(x))
  }
  ret <- strsplit(x = x, split = "[, ]+", perl = TRUE)
  mask_none <- vapply(X = ret, FUN = length, FUN.VALUE = 1L) == 0
  ret[mask_none] <- NA_character_
  ret <- lapply(X = ret, FUN = PKNCA_impute_fun_list_paste)
  # Confirm that the functions exist and are functions
  # Sort will ensure that the results are not NA
  all_funs <- sort(unlist(ret))
  bad_fun <- character()
  for (idx in seq_along(all_funs)) {
    found_fun <- utils::getAnywhere(all_funs[[idx]])
    if (length(found_fun$objs) == 0) {
      bad_fun <- c(bad_fun, all_funs[[idx]])
    } else if (!is.function(found_fun$objs[[1]])) {
      bad_fun <- c(bad_fun, all_funs[[idx]])
    }
  }
  if (length(bad_fun) > 0) {
    stop(
      "The following imputation functions were not found: ",
      paste(bad_fun, collapse = ", ")
    )
  }
  ret
}

# A helper for PKNCA_impute_fun_list that pastes PKNCA_impute_method_ to the
# beginning of everything but NA
PKNCA_impute_fun_list_paste <- function(x) {
  mask_paste <- !is.na(x) & !startsWith(x, "PKNCA_impute_method_")
  if (any(mask_paste)) {
    x[mask_paste] <- paste0("PKNCA_impute_method_", x[mask_paste])
  }
  x
}

test_that("PKNCA_impute_method_start_conc0", {
  # Time 0 is replaced
  expect_equal(
    PKNCA_impute_method_start_conc0(conc = 1:3, time = 0:2),
    data.frame(conc = c(0, 2:3), time = 0:2)
  )
  # Time 0 is added
  expect_equal(
    PKNCA_impute_method_start_conc0(conc = 2:3, time = 1:2),
    data.frame(conc = c(0, 2:3), time = 0:2),
    ignore_attr = TRUE
  )
  # Time 0 is inserted
  expect_equal(
    PKNCA_impute_method_start_conc0(conc = 1:3, time = c(-1, 1:2)),
    data.frame(conc = c(1, 0, 2:3), time = -1:2),
    ignore_attr = TRUE
  )
})

test_that("PKNCA_impute_method_start_predose", {
  # No modification if no predose samples
  expect_equal(
    PKNCA_impute_method_start_predose(conc = 1:3, time = 1:3, conc.group = 1:3, time.group = 1:3, start = 0, end = 24),
    data.frame(conc = 1:3, time = 1:3)
  )
  # No modification if time 0 is already present
  expect_equal(
    PKNCA_impute_method_start_predose(conc = 1:3, time = 0:2, conc.group = 1:3, time.group = 0:2, start = 0, end = 24),
    data.frame(conc = 1:3, time = 0:2)
  )
  # Shift happens when time is within max_shift
  expect_equal(
    PKNCA_impute_method_start_predose(conc = 2:3, time = 1:2, conc.group = 1:3, time.group = c(-1, 1:2), start = 0, end = 24),
    data.frame(conc = 1:3, time = 0:2)
  )
  # Shift happens when time is equal to max_shift
  expect_equal(
    PKNCA_impute_method_start_predose(conc = 2:3, time = 1:2, conc.group = 1:3, time.group = c(-1.2, 1:2), start = 0, end = 24),
    data.frame(conc = 1:3, time = 0:2)
  )
  # Shift occurs to a new start
  expect_equal(
    PKNCA_impute_method_start_predose(conc = 2:3, time = 1:2, conc.group = 1:3, time.group = c(-0.3, 1:2), start = 0.5, end = 24),
    data.frame(conc = 1:3, time = c(0.5, 1:2))
  )
  # Shift does not when time is more than max_shift
  expect_equal(
    PKNCA_impute_method_start_predose(conc = 2:3, time = 1:2, conc.group = 1:3, time.group =c(-3, 1:2), start = 0, end = 24),
    data.frame(conc = 2:3, time = 1:2)
  )
  # max_shift overrides the start/end automation
  expect_equal(
    PKNCA_impute_method_start_predose(conc = 2:3, time = 1:2, conc.group = 1:3, time.group = c(-3, 1:2), max_shift = 3, start = 0, end = 24),
    data.frame(conc = 1:3, time = 0:2)
  )
  # shift automation works reasonably even when the end is infinite
  expect_equal(
    PKNCA_impute_method_start_predose(conc = 2:4, time = c(1:2, 24), conc.group = 1:4, time.group = c(-1.25, 1:2, 24), start = 0, end = Inf),
    data.frame(conc = 2:4, time = c(1:2, 24))
  )
  expect_equal(
    PKNCA_impute_method_start_predose(conc = 2:4, time = c(1:2, 24), conc.group = 1:4, time.group = c(-1.2, 1:2, 24), start = 0, end = Inf),
    data.frame(conc = 1:4, time = c(0:2, 24))
  )
})

test_that("PKNCA_impute_method_start_cmin", {
  # No imputation when start is in the data
  expect_equal(
    PKNCA_impute_method_start_cmin(conc = 1:3, time = 0:2, start = 0, end = 24),
    data.frame(conc = 1:3, time = 0:2)
  )
  # impute when start is not in the data
  expect_equal(
    PKNCA_impute_method_start_cmin(conc = 1:3, time = 1:3, start = 0, end = 24),
    data.frame(conc = c(1, 1:3), time = 0:3),
    ignore_attr = TRUE
  )
  # data outside the interval are ignored (before interval)
  expect_equal(
    PKNCA_impute_method_start_cmin(conc = 1:3, time = 1:3, start = 1.5, end = 24),
    data.frame(conc = c(1, 2, 2:3), time = c(1, 1.5, 2:3)),
    ignore_attr = TRUE
  )
  # data outside the interval are ignored (after interval)
  expect_equal(
    PKNCA_impute_method_start_cmin(conc = c(1:3, 0.5), time = c(1:3, 25), start = 1.5, end = 24),
    data.frame(conc = c(1, 2, 2:3, 0.5), time = c(1, 1.5, 2:3, 25)),
    ignore_attr = TRUE
  )

})

test_that("PKNCA_impute_fun_list", {
  expect_equal(
    PKNCA_impute_fun_list(NA_character_),
    list(NA_character_)
  )
  # an empty string is the same as NA
  expect_equal(
    PKNCA_impute_fun_list(NA_character_),
    PKNCA_impute_fun_list("")
  )
  # logical NA works the same as character NA
  expect_equal(
    PKNCA_impute_fun_list(NA_character_),
    PKNCA_impute_fun_list(NA)
  )
  # Non-character input (other than all NA) is an error
  expect_error(
    PKNCA_impute_fun_list(1),
    regexp = "non-character argument"
  )
  # One imputation method works
  expect_equal(
    PKNCA_impute_fun_list("start_conc0"),
    list("PKNCA_impute_method_start_conc0")
  )
  # Two imputation methods detect errors correctly
  expect_error(
    PKNCA_impute_fun_list("start_conc0,foo"),
    regexp = "The following imputation functions were not found: PKNCA_impute_method_foo"
  )
  # Two imputation methods work
  expect_equal(
    PKNCA_impute_fun_list("start_conc0,start_predose"),
    list(c("PKNCA_impute_method_start_conc0", "PKNCA_impute_method_start_predose"))
  )
  # A vector of different imputation methods works
  expect_equal(
    PKNCA_impute_fun_list(c(NA, NA_character_, "", "start_conc0,start_predose", "start_conc0")),
    list(
      NA_character_,
      NA_character_,
      NA_character_,
      c("PKNCA_impute_method_start_conc0", "PKNCA_impute_method_start_predose"),
      "PKNCA_impute_method_start_conc0"
    )
  )
})

test_that("PKNCA_impute_fun_list_paste", {
  expect_equal(
    PKNCA_impute_fun_list_paste("A"),
    "PKNCA_impute_method_A"
  )
  expect_equal(
    PKNCA_impute_fun_list_paste("PKNCA_impute_method_A"),
    "PKNCA_impute_method_A"
  )
  expect_equal(
    PKNCA_impute_fun_list_paste(NA_character_),
    NA_character_
  )
  expect_equal(
    PKNCA_impute_fun_list_paste(c("PKNCA_impute_method_A", "A", NA)),
    c("PKNCA_impute_method_A", "PKNCA_impute_method_A", NA_character_)
  )
})

test_that("add_impute_to_intervals", {
  d_conc <- generate.conc(nsub=5, ntreat=2, time.points=0:24)
  d_dose <- generate.dose(d_conc)
  o_conc <- PKNCAconc(d_conc, formula=conc~time|treatment+ID)
  o_dose <- PKNCAdose(d_dose, formula=dose~time|treatment+ID)
  o_data <- PKNCAdata(o_conc, o_dose)
  expect_error(
    add_impute_to_intervals(o_data),
    class = "pknca_add_impute_to_intervals_NA"
  )
  # Nothing if impute is in the data
  o_data_start <- o_data
  o_data_start$impute <- "start"
  expect_equal(
    add_impute_to_intervals(PKNCAdata(o_conc, o_dose, impute = "start")),
    o_data_start
  )
  # Impute added to the intervals and the column is modified, if it is not in
  # the data
  o_data <- PKNCAdata(o_conc, o_dose, impute = "start_conc0")
  expect_equal(add_impute_to_intervals(o_data)$impute, "impute")
  expect_true(all(
    add_impute_to_intervals(o_data)$intervals$impute %in% "start_conc0"
  ))
  # If impute is already in the names of the intervals, add something else as
  # the impute column name
  o_data <- PKNCAdata(o_conc, o_dose, impute = "start_conc0")
  o_data$intervals$impute <- "foo"
  expect_equal(add_impute_to_intervals(o_data)$impute, "imputeX")
  expect_true(all(
    add_impute_to_intervals(o_data)$intervals$imputeX %in% "start_conc0"
  ))
})

test_that("PKNCAdata moves imputation to the intervals column, as applicable", {
  d_conc <- generate.conc(nsub = 1, ntreat = 1, time.points = 1:3, nstudies = 1)
  o_conc <- PKNCAconc(conc~time, data = d_conc)
  o_data <- PKNCAdata(o_conc, intervals = data.frame(start = 0, end = 3, auclast = TRUE))
  # No imputation creates no imputation instructions
  expect_equal(o_data$impute, NA_character_)

  o_data <- PKNCAdata(o_conc, intervals = data.frame(start = 0, end = 3, auclast = TRUE), impute = "PKNCA_impute_method_start_conc0")
  expect_equal(o_data$impute, "PKNCA_impute_method_start_conc0")
  pk.nca(o_data)

  o_data <- PKNCAdata(o_conc, intervals = data.frame(start = 0, end = 3, auclast = TRUE), impute = "start_conc0")
  expect_equal(o_data$impute, "start_conc0")
})

# Putting it all together
test_that("pk.nca with imputation", {
  d_conc <- as.data.frame(datasets::Theoph)[!datasets::Theoph$Time == 0, ]
  conc_obj <- PKNCAconc(d_conc, conc~Time|Subject)
  d_dose <- unique(datasets::Theoph[datasets::Theoph$Time == 0,
                                    c("Dose", "Time", "Subject")])
  dose_obj <- PKNCAdose(d_dose, Dose~Time|Subject)
  data_obj_noimpute <- PKNCAdata(conc_obj, dose_obj)
  data_obj_impute <- PKNCAdata(conc_obj, dose_obj, impute = "start_predose,start_conc0")
  suppressWarnings(nca_obj_noimpute <- pk.nca(data_obj_noimpute))
  nca_obj_impute <- pk.nca(data_obj_impute)

  # By interval imputation works
  d_intervals <-
    data.frame(
      start=0, end=c(24, 24.1),
      auclast=TRUE,
      impute=c(NA, "start_conc0")
    )

  data_obj_manualimpute <- PKNCAdata(conc_obj, dose_obj, intervals = d_intervals, impute = "impute")
  suppressWarnings(nca_obj_manualimpute <- pk.nca(data_obj_manualimpute))

  auclast_noimpute <- as.data.frame(nca_obj_noimpute)
  auclast_noimpute <- auclast_noimpute$PPORRES[auclast_noimpute$PPTESTCD %in% "auclast"]
  expect_true(all(is.na(auclast_noimpute)))
  auclast_impute <- as.data.frame(nca_obj_impute)
  auclast_impute <- auclast_impute$PPORRES[auclast_impute$PPTESTCD %in% "auclast"]
  expect_true(!any(is.na(auclast_impute)))

  auclast_manualimpute <- as.data.frame(nca_obj_manualimpute)
  auclast_manualimpute_24 <- auclast_manualimpute$PPORRES[auclast_manualimpute$PPTESTCD %in% "auclast" & auclast_manualimpute$end %in% 24]
  auclast_manualimpute_24.1 <- auclast_manualimpute$PPORRES[auclast_manualimpute$PPTESTCD %in% "auclast" & auclast_manualimpute$end %in% 24.1]
  expect_true(all(is.na(auclast_manualimpute_24)))
  expect_true(!any(is.na(auclast_manualimpute_24.1)))
})

test_that("start_conc0 imputation works (fix #257)", {
  d <- data.frame(time=c(0.08, 1, 2, 3, 4),
                  conc=c(1, 0.5, 0.4, 0.3, 0.2),
                  subject=1)
  myconc <- PKNCAconc(data=d, conc~time|subject)
  mydata <-
    PKNCAdata(
      myconc,
      intervals=
        data.frame(
          start=0, end=5,
          impute="start_conc0",
          cmax        = TRUE,
          tmax        = TRUE,
          aucinf.obs  = TRUE,
          aucint.last = TRUE,
          auclast     = TRUE
        )
    )
  suppressMessages(suppressWarnings(
    myres <- pk.nca(mydata)
  ))
  df_myres <- as.data.frame(myres)
  expect_false(is.na(df_myres$PPORRES[df_myres$PPTESTCD == "auclast"]))
})

test_that("PKNCA_impute_fun_list", {
  PKNCA_impute_method_character <- "A"
  expect_error(PKNCA_impute_fun_list("character"),
               regexp = "The following imputation functions were not found: PKNCA_impute_method_character"
  )
})

#' Run any function with a maximum missing fraction of X and 0s possibly
#' counting as missing.  The maximum fraction missing comes from
#' `PKNCA.options("max.missing")`.
#'
#' Note that all missing values are removed prior to calling the function.
#'
#' @param FUN function to run.  The function is called as `FUN(x, ...)` with
#'   missing values removed.
#' @param zero.missing Are zeros counted as missing?  If `TRUE` then include
#'   them in the missing count.
#' @param max.missing The maximum fraction of the data allowed to be missing (a
#'   number between 0 and 1, inclusive).
#' @returns A version of FUN that can be called with parameters that are checked
#'   for missingness (and zeros) with missing (and zeros) removed before the
#'   call.  If `max.missing` is exceeded, then NA is returned.
#' @examples
#' my_mean <- pk.business(FUN=mean)
#' mean(c(1:3, NA))
#' # Less than half missing results in the summary statistic of the available
#' # values.
#' my_mean(c(1:3, NA))
#' # More than half missing results in a missing value
#' my_mean(c(1:3, rep(NA, 4)))
#' @export
pk.business <- function(FUN,
                        zero.missing=FALSE,
                        max.missing)
  function(x, ...) {
    # Allow max.missing to be specified at either function initiation or for it
    # to use PKNCA.Options("max.missing")
    max.missing <- PKNCA.options("max.missing")
    mask.missing <- is.na(x) | (zero.missing & (x %in% 0))
    if (sum(mask.missing)/length(x) > max.missing)
      return(NA)
    ret <- FUN(x[!mask.missing], ...)
    attr(ret, "n") <- sum(!mask.missing)
    ret
  }

#' Compute the geometric mean, sd, and CV
#'
#' @param x A vector to compute the geometric mean of
#' @param na.rm Should missing values be removed?
#' @return The scalar value of the geometric mean, geometric standard deviation,
#'   or geometric coefficient of variation.
#' @references Kirkwood T. B.L. Geometric means and measures of dispersion.
#'   Biometrics 1979; 35: 908-909
#' @examples
#' geomean(1:3)
#' @export
geomean <- function(x, na.rm=FALSE) {
  if (na.rm)
    x <- stats::na.omit(x)
  if (any(is.na(x))) {
    as.numeric(NA)
  } else if (any(x == 0)) {
    0
  } else if (any(x < 0)) {
    # Protect from overflows by using the logarithm
    prod(sign(x))*exp(sum(log(abs(x)))/length(x))
  } else {
    exp(sum(log(x))/length(x))
  }
}

#' @describeIn geomean
#' Compute the geometric standard deviation, `exp(sd(log(x)))`.
#' @examples
#' geosd(1:3)
#' @export
geosd <- function(x, na.rm=FALSE) {
  exp(stats::sd(log(x), na.rm=na.rm))
}

#' @describeIn geomean
#' Compute the geometric coefficient of variation,
#' `sqrt(exp(sd(log(x))^2)-1)*100`.
#' @examples
#' geocv(1:3)
#' @export
geocv <- function(x, na.rm=FALSE) {
  sqrt(exp(stats::sd(log(x), na.rm=na.rm)^2)-1)*100
}

#' Generate functions to do the named function (e.g. mean) applying
#' the business rules.
#'
#' @param x vector to be passed to the various functions
#' @param ... Additional arguments to be passed to the underlying function.
#' @return The value of the various functions or NA if too many values are
#'   missing
#' @seealso [pk.business()]
#' @export
business.mean <- pk.business(mean, max.missing=~PKNCA::PKNCA.Options("max.missing"))

#' @describeIn business.mean
#' Compute the standard deviation with business rules.
#' @export
business.sd <- pk.business(stats::sd, max.missing=~PKNCA::PKNCA.Options("max.missing"))

#' @describeIn business.mean
#' Compute the coefficient of variation with business rules.
#' @export
business.cv <-
  pk.business(
    function(x, ...) {
      100*stats::sd(x, ...)/mean(x, ...)
    },
    max.missing=~PKNCA::PKNCA.Options("max.missing")
  )

#' @describeIn business.mean
#' Compute the geometric mean with business rules.
#' @export
business.geomean <-
  pk.business(
    geomean, zero.missing=TRUE,
    max.missing=~PKNCA::PKNCA.Options("max.missing")
  )
#' @describeIn business.mean
#' Compute the geometric coefficient of variation with business rules.
#' @export
business.geocv <-
  pk.business(
    geocv, zero.missing=TRUE,
    max.missing=~PKNCA::PKNCA.Options("max.missing")
  )
#' @describeIn business.mean
#' Compute the minimum with business rules.
#' @export
business.min <-
  pk.business(min, max.missing=~PKNCA::PKNCA.Options("max.missing"))
#' @describeIn business.mean
#' Compute the maximum with business rules.
#' @export
business.max <-
  pk.business(max, max.missing=~PKNCA::PKNCA.Options("max.missing"))
#' @describeIn business.mean
#' Compute the median with business rules.
#' @export
business.median <-
  pk.business(stats::median, max.missing=~PKNCA::PKNCA.Options("max.missing"))
#' @describeIn business.mean
#' Compute the range with business rules.
#' @export
business.range <-
  pk.business(range, max.missing=~PKNCA::PKNCA.Options("max.missing"))

test_that("geomean", {
  # Test normal, nonzero scalars
  expect_equal(geomean(5), 5)
  expect_equal(geomean(c(5, 5)), 5)
  expect_equal(geomean(c(10, 1000)), 100)
  # Test zeros
  expect_equal(geomean(0), 0)
  expect_equal(geomean(c(0, 1)), 0)
  # Test negative numbers
  expect_equal(geomean(c(-1, 2)), -geomean(c(1, 2)))
  # Test NAs
  expect_equal(geomean(NA), as.numeric(NA))
  expect_equal(geomean(c(NA, 0)), as.numeric(NA))
  expect_equal(geomean(c(NA, 5), na.rm=TRUE), 5)
  expect_equal(geomean(c(NA, NA), na.rm=TRUE), NaN)
})

test_that("geosd", {
  expect_equal(geosd(c(1, 2)), exp(sd(log(c(1, 2)))))
  expect_equal(geosd(c(NA, 1, 2)), as.numeric(NA))
})

test_that("geocv", {
  expect_equal(geocv(c(1, 2)),
               sqrt(exp(sd(log(c(1, 2)))^2)-1)*100)
  expect_equal(geocv(c(NA, 1, 2)), as.numeric(NA))
})

test_that("business.mean", {
  PKNCA.options(default=TRUE)
  PKNCA.options(max.missing=0.5)
  # Test a normal mean of a scalar and vector
  expect_equal(business.mean(1), structure(1, n = 1))
  expect_equal(business.mean(c(1, 2)), structure(1.5, n = 2))
  # Ensure that at the max.missing fraction a value is reported and
  # above that, NA is returned.
  expect_equal(business.mean(c(NA, NA, 1, 2)), structure(1.5, n = 2))
  expect_equal(business.mean(c(NA, NA, NA, 2)), NA)
  # Ensure that it uses the current value of max.missing
  PKNCA.options(max.missing=0.3)
  expect_equal(business.mean(c(NA, 1, 2, 3)), structure(2, n = 3))
  expect_equal(business.mean(c(NA, NA, 1, 2)), NA)
})

test_that("pk.business", {
  PKNCA.options(default=TRUE)
  PKNCA.options(max.missing=0.5)
  b.mean <- pk.business(mean)
  expect_equal(b.mean(c(1, 2)), structure(1.5, n = 2))
  # Right at the border, it still reports
  expect_equal(b.mean(c(1, NA)), structure(1, n = 1))
  # When too much data is missing, NA is returned
  expect_equal(b.mean(c(1, NA, NA)), NA)
  # It respects zero.missing
  b.mean.2 <- pk.business(mean, zero.missing=TRUE)
  expect_equal(b.mean(c(0, 1)), structure(0.5, n = 2))
  expect_equal(b.mean.2(c(0, 1)), structure(1, n = 1))
})

#' Generate an example concentration data set for testing
#'
#' @param nsub The number of subjects to generate
#' @param ntreat The number of treatments per subject
#' @param time.points The times points to generate for each subject (a
#' vector)
#' @param nstudies The number of studies (\code{NA} gives one study
#' without a study column)
#' @param nanalytes The number of analytes to generate (\code{NA}
#' gives one analyte without an analyte column)
#' @param resid The normally distributed random error
#' @return A data frame with columns for study, treatment, subject,
#' time, analyte, and concentration.
generate.conc <- function(nsub, ntreat, time.points, nstudies=NA, nanalytes=NA, resid=0.1) {
  one.cmt.oral <- function(time, dose, v, ka, kel, f=1)
    f*dose*ka/(v*(ka-kel))*(exp(-kel*time) - exp(-ka*time))
  # one.cmt.iv <- function(time, dose, v, kel)
  #   dose/v*exp(-kel*time)
  set.seed(5)
  ret <- data.frame()
  nanalytes.include <- !is.na(nanalytes)
  if (is.na(nanalytes))
    nanalytes <- 1
  nstudies.include <- !is.na(nstudies)
  if (is.na(nstudies))
    nstudies <- 1
  for (i in seq_len(ntreat))
    for (j in seq_len(nsub))
      for (k in seq_len(nanalytes))
        for (m in seq_len(nstudies)) {
          ret <- rbind(
            ret,
            data.frame(
              study=paste("Study", m),
              treatment=paste("Trt", i),
              ID=j,
              time=time.points,
              conc=exp(
                rnorm(length(time.points), mean=0, sd=resid))*
                one.cmt.oral(time.points,
                             dose=i,
                             v=1,
                             ka=1,
                             kel=0.05),
              analyte=paste("Analyte", k),
              stringsAsFactors=FALSE
            )
          )
        }
  if (!nanalytes.include)
    ret$analyte <- NULL
  if (!nstudies.include)
    ret$study <- NULL
  ret
}

#' Generate dose information to go with concentration data.
#'
#' @param concdata Concentration-time data generated by
#' \code{\link{generate.conc}}
#' @return A data frame of dosing data
generate.dose <- function(concdata) {
  ret <-
    unique(
      concdata[
        ,
        setdiff(names(concdata), c("time", "conc", "analyte"))
        ]
    )
  ret$dose <- as.numeric(factor(ret$treatment))
  ret$time <- 0
  ret
}

# based on purrr:::capture_output
msg_grabber <- function(code) {
  messages <- character()
  handler <- function(m) {
    messages <<- c(messages, m$message)
    invokeRestart("muffleMessage")
  }
  withCallingHandlers(code, message = handler)
  messages
}

test_that("dplyr filter", {
  tmpconc <- generate.conc(2, 1, 0:24)
  tmpdose <- generate.dose(tmpconc)
  myconc <- PKNCAconc(tmpconc, formula=conc~time|treatment+ID)
  mydose <- PKNCAdose(tmpdose, formula=dose~time|treatment+ID)
  mydata <- PKNCAdata(myconc, mydose)
  myresult <- pk.nca(mydata)

  filtered <- filter(myresult, PPTESTCD == "auclast")
  filtered_manual <- myresult
  filtered_manual$result <- filtered_manual$result[filtered_manual$result$PPTESTCD == "auclast", ]
  expect_equal(filtered, filtered_manual)

  filtered <- filter(myconc, ID == 1)
  filtered_manual <- myconc
  filtered_manual$data <- myconc$data[myconc$data$ID == 1, ]
  expect_equal(filtered, filtered_manual)
})

test_that("dplyr left_join", {
  tmpconc <- generate.conc(2, 1, 0:24)
  tmpdose <- generate.dose(tmpconc)
  myconc <- PKNCAconc(tmpconc, formula=conc~time|treatment+ID)
  mydose <- PKNCAdose(tmpdose, formula=dose~time|treatment+ID)
  mydata <- PKNCAdata(myconc, mydose)
  myresult <- pk.nca(mydata)

  joindf <- data.frame(ID=1, foo="bar")
  msg_join_id <- msg_grabber(left_join(data.frame(ID = 1), data.frame(ID = 1)))
  expect_message(
    joined <- left_join(myresult, joindf),
    msg_join_id,
    fixed = TRUE
  )
  joined_manual <- myresult
  expect_message(
    joined_manual$result <- dplyr::left_join(joined_manual$result, joindf),
    msg_join_id,
    fixed = TRUE
  )
  expect_equal(joined, joined_manual)

  expect_message(
    joined <- left_join(myconc, joindf),
    msg_join_id,
    fixed = TRUE
  )
  joined_manual <- myconc
  expect_message(
    joined_manual$data <- left_join(joined_manual$data, joindf),
    msg_join_id,
    fixed = TRUE
  )
  expect_equal(joined, joined_manual)
})

test_that("dplyr mutate", {
  tmpconc <- generate.conc(2, 1, 0:24)
  tmpdose <- generate.dose(tmpconc)
  myconc <- PKNCAconc(tmpconc, formula=conc~time|treatment+ID)
  mydose <- PKNCAdose(tmpdose, formula=dose~time|treatment+ID)
  mydata <- PKNCAdata(myconc, mydose)
  myresult <- pk.nca(mydata)

  mutated <- mutate(myresult, foo="bar")
  mutated_manual <- myresult
  mutated_manual$result <- mutate(mutated_manual$result, foo="bar")
  expect_equal(mutated, mutated_manual)

  mutated <- mutate(myconc, foo="bar")
  mutated_manual <- myconc
  mutated_manual$data <- mutate(mutated_manual$data, foo="bar")
  expect_equal(mutated, mutated_manual)
})

test_that("dplyr group_by and ungroup", {
  tmpconc <- generate.conc(2, 1, 0:24)
  tmpdose <- generate.dose(tmpconc)
  myconc <- PKNCAconc(tmpconc, formula=conc~time|treatment+ID)
  mydose <- PKNCAdose(tmpdose, formula=dose~time|treatment+ID)
  mydata <- PKNCAdata(myconc, mydose)
  myresult <- pk.nca(mydata)

  grouped <- group_by(myconc, treatment)
  expect_s3_class(grouped$data, "grouped_df")
  ungrouped <- ungroup(grouped)
  expect_false("grouped_df" %in% class(ungrouped$data))
  expect_s3_class(ungrouped$data, "data.frame")
})

test_that("PKNCAresults summary", {
  # Note that generate.conc sets the random seed, so it doesn't have
  # to happen here.
  tmpconc <- generate.conc(2, 1, 0:24)
  tmpdose <- generate.dose(tmpconc)
  myconc <- PKNCAconc(tmpconc, formula = conc ~ time | treatment + ID)
  mydose <- PKNCAdose(tmpdose, formula = dose ~ time | treatment + ID)
  mydata <- PKNCAdata(myconc, mydose)
  myresult <- pk.nca(mydata)

  # Testing the summarization
  mysummary <- summary(myresult)
  expect_true(is.data.frame(mysummary))
  expect_equal(
    mysummary,
    as_summary_PKNCAresults(
      data.frame(
        start = 0,
        end = c(24, Inf),
        treatment = "Trt 1",
        N = "2",
        auclast = c("13.8 [2.51]", "."),
        cmax = c(".", "0.970 [4.29]"),
        tmax = c(".", "3.00 [2.00, 4.00]"),
        half.life = c(".", "14.2 [2.79]"),
        aucinf.obs = c(".", "20.5 [6.84]"),
        stringsAsFactors = FALSE
      ),
      caption = "auclast, cmax, aucinf.obs: geometric mean and geometric coefficient of variation; tmax: median and range; half.life: arithmetic mean and standard deviation; N: number of subjects"
    ),
    info = "simple summary of PKNCAresults performs as expected"
  )

  tmpconc <- generate.conc(2, 1, 0:24)
  tmpconc$conc[tmpconc$ID %in% 2] <- 0
  tmpdose <- generate.dose(tmpconc)
  myconc <- PKNCAconc(tmpconc, conc ~ time | treatment + ID)
  mydose <- PKNCAdose(tmpdose, dose ~ time | treatment + ID)
  mydata <- PKNCAdata(myconc, mydose)
  # Not capturing the warning due to R bug
  # https://bugs.r-project.org/bugzilla3/show_bug.cgi?id=17122
  # expect_warning(myresult <- pk.nca(mydata),
  #               regexp="Too few points for half-life calculation")
  suppressWarnings(myresult <- pk.nca(mydata))
  mysummary <- summary(myresult)
  expect_equal(
    mysummary,
    as_summary_PKNCAresults(
      data.frame(
        start = 0,
        end = c(24, Inf),
        treatment = "Trt 1",
        N = "2",
        auclast = c("13.5 [NC]", "."),
        cmax = c(".", "1.00 [NC]"),
        tmax = c(".", "4.00, n=1"),
        half.life = c(".", "16.1, n=1"),
        aucinf.obs = c(".", "21.5 [NC]"),
        stringsAsFactors = FALSE
      ),
      caption = "auclast, cmax, aucinf.obs: geometric mean and geometric coefficient of variation; tmax: median and range; half.life: arithmetic mean and standard deviation; N: number of subjects; n: number of measurements included in summary"
    ),
    info = "summary of PKNCAresults with some missing values results in NA for spread"
  )

  tmpconc <- generate.conc(2, 1, 0:24)
  tmpconc$conc <- 0
  tmpdose <- generate.dose(tmpconc)
  myconc <- PKNCAconc(tmpconc, conc ~ time | treatment + ID)
  mydose <- PKNCAdose(tmpdose, dose ~ time | treatment + ID)
  mydata <- PKNCAdata(myconc, mydose)
  # Not capturing the warning due to R bug
  # https://bugs.r-project.org/bugzilla3/show_bug.cgi?id=17122
  # expect_warning(myresult <- pk.nca(mydata),
  #               regexp="Too few points for half-life calculation")
  suppressWarnings(myresult <- pk.nca(mydata))
  mysummary <- summary(myresult)
  expect_equal(
    mysummary,
    as_summary_PKNCAresults(
      data.frame(
        start = 0,
        end = c(24, Inf),
        treatment = "Trt 1",
        N = "2",
        auclast = c("NC", "."),
        cmax = c(".", "NC"),
        tmax = c(".", "NC"),
        half.life = c(".", "NC"),
        aucinf.obs = c(".", "NC"),
        stringsAsFactors = FALSE
      ),
      caption = "auclast, cmax, aucinf.obs: geometric mean and geometric coefficient of variation; tmax: median and range; half.life: arithmetic mean and standard deviation; N: number of subjects"
    ),
    info = "summary of PKNCAresults without most results gives NC"
  )

  mysummary <- summary(myresult,
    not_requested = "NR",
    not_calculated = "NoCalc"
  )
  expect_equal(
    mysummary,
    as_summary_PKNCAresults(
      data.frame(
        start = 0,
        end = c(24, Inf),
        treatment = "Trt 1",
        N = "2",
        auclast = c("NoCalc", "NR"),
        cmax = c("NR", "NoCalc"),
        tmax = c("NR", "NoCalc"),
        half.life = c("NR", "NoCalc"),
        aucinf.obs = c("NR", "NoCalc"),
        stringsAsFactors = FALSE
      ),
      caption = "auclast, cmax, aucinf.obs: geometric mean and geometric coefficient of variation; tmax: median and range; half.life: arithmetic mean and standard deviation; N: number of subjects"
    ),
    info = "Summary respects the not.requested.string and not.calculated.string"
  )

  mysummary <- summary(myresult,
    summarize_n = FALSE,
    not_requested = "NR",
    not_calculated = "NoCalc"
  )
  expect_equal(
    mysummary,
    as_summary_PKNCAresults(
      data.frame(
        start = 0,
        end = c(24, Inf),
        treatment = "Trt 1",
        auclast = c("NoCalc", "NR"),
        cmax = c("NR", "NoCalc"),
        tmax = c("NR", "NoCalc"),
        half.life = c("NR", "NoCalc"),
        aucinf.obs = c("NR", "NoCalc"),
        stringsAsFactors = FALSE
      ),
      caption = "auclast, cmax, aucinf.obs: geometric mean and geometric coefficient of variation; tmax: median and range; half.life: arithmetic mean and standard deviation"
    ),
    info = "N is optionally omitted"
  )
})

test_that("PKNCAresults summary counts N correctly", {
  tmpconc <- generate.conc(2, 1, 0:24)
  tmpdose <- generate.dose(tmpconc)
  myconc <- PKNCAconc(tmpconc, formula = conc ~ time | treatment + ID)
  mydose <- PKNCAdose(tmpdose, formula = dose ~ time | treatment + ID)
  mydata <- PKNCAdata(myconc, mydose, intervals = data.frame(start = 0, end = c(24, Inf), cmax = TRUE))
  myresult <- pk.nca(mydata)

  # Testing the summarization
  mysummary_two_row <- summary(myresult)
  expect_warning(
    expect_warning(
      mysummary_one_row <- summary(myresult, drop_group = c("ID", "end")),
      "Some subjects may have more than one result for cmax"
    ),
    "drop.group including start or end may result in incorrect groupings"
  )
  expect_equal(mysummary_two_row$N, c("2", "2"))
  expect_equal(mysummary_one_row$N, "2")

  # No subject identifier
  tmpconc <- generate.conc(1, 1, 0:24)
  tmpdose <- generate.dose(tmpconc)
  myconc <- PKNCAconc(tmpconc, formula = conc ~ time)
  mydose <- PKNCAdose(tmpdose, formula = dose ~ time)
  mydata <- PKNCAdata(myconc, mydose, intervals = data.frame(start = 0, end = c(24, Inf), cmax = TRUE))
  myresult <- pk.nca(mydata)

  mysummary_one_subject <- summary(myresult)
  expect_false("N" %in% names(mysummary_one_subject))
  expect_warning(
    mysummary_one_subject_askn <- summary(myresult, summarize_n = TRUE),
    "summarize_n was requested, but no subject column exists"
  )
  expect_false("N" %in% names(mysummary_one_subject_askn))
})

test_that("dropping `start` and `end` from groups is allowed with a warning.", {
  tmpconc <- generate.conc(2, 1, 0:24)
  tmpdose <- generate.dose(tmpconc)
  myconc <- PKNCAconc(tmpconc, formula = conc ~ time | treatment + ID)
  mydose <- PKNCAdose(tmpdose, formula = dose ~ time | treatment + ID)
  mydata <- PKNCAdata(myconc, mydose)
  myresult <- pk.nca(mydata)

  expect_warning(
    current_summary <- summary(myresult, drop_group = c("ID", "start")),
    regexp = "drop.group including start or end may result", fixed = TRUE
  )
  expect_false("start" %in% names(current_summary))
})

test_that("summary.PKNCAresults manages exclusions as missing not as non-existent.", {
  # Note that generate.conc sets the random seed, so it doesn't have
  # to happen here.
  tmpconc <- generate.conc(2, 1, 0:24)
  tmpdose <- generate.dose(tmpconc)
  myconc <- PKNCAconc(tmpconc, formula = conc ~ time | treatment + ID)
  mydose <- PKNCAdose(tmpdose, formula = dose ~ time | treatment + ID)
  mydata <- PKNCAdata(myconc, mydose)
  myresult <- pk.nca(mydata)
  myresult_excluded <-
    exclude(
      myresult,
      reason = "testing",
      mask = with(
        as.data.frame(myresult),
        PPTESTCD %in% "auclast" & ID %in% 1
      )
    )
  myresult_excluded2 <-
    exclude(
      myresult,
      reason = "testing",
      mask = with(
        as.data.frame(myresult),
        PPTESTCD %in% "auclast"
      )
    )
  # Testing the summarization
  mysummary <- summary(myresult)
  mysummary_excluded <- summary(myresult_excluded)
  mysummary_excluded2 <- summary(myresult_excluded2)
  expect_equal(
    mysummary,
    as_summary_PKNCAresults(
      data.frame(
        start = 0,
        end = c(24, Inf),
        treatment = "Trt 1",
        N = "2",
        auclast = c("13.8 [2.51]", "."),
        cmax = c(".", "0.970 [4.29]"),
        tmax = c(".", "3.00 [2.00, 4.00]"),
        half.life = c(".", "14.2 [2.79]"),
        aucinf.obs = c(".", "20.5 [6.84]"),
        stringsAsFactors = FALSE
      ),
      caption = "auclast, cmax, aucinf.obs: geometric mean and geometric coefficient of variation; tmax: median and range; half.life: arithmetic mean and standard deviation; N: number of subjects"
    ),
    info = "simple summary of PKNCAresults performs as expected"
  )
  expect_equal(
    mysummary_excluded,
    as_summary_PKNCAresults(
      data.frame(
        start = 0,
        end = c(24, Inf),
        treatment = "Trt 1",
        N = "2",
        auclast = c("14.0, n=1", "."),
        cmax = c(".", "0.970 [4.29]"),
        tmax = c(".", "3.00 [2.00, 4.00]"),
        half.life = c(".", "14.2 [2.79]"),
        aucinf.obs = c(".", "20.5 [6.84]"),
        stringsAsFactors = FALSE
      ),
      caption = "auclast, cmax, aucinf.obs: geometric mean and geometric coefficient of variation; tmax: median and range; half.life: arithmetic mean and standard deviation; N: number of subjects; n: number of measurements included in summary"
    ),
    info = "summary of PKNCAresults correctly excludes auclast when requested"
  )
  expect_equal(
    mysummary_excluded2,
    as_summary_PKNCAresults(
      data.frame(
        start = 0,
        end = c(24, Inf),
        treatment = "Trt 1",
        N = "2",
        auclast = c("NC", "."),
        cmax = c(".", "0.970 [4.29]"),
        tmax = c(".", "3.00 [2.00, 4.00]"),
        half.life = c(".", "14.2 [2.79]"),
        aucinf.obs = c(".", "20.5 [6.84]"),
        stringsAsFactors = FALSE
      ),
      caption = "auclast, cmax, aucinf.obs: geometric mean and geometric coefficient of variation; tmax: median and range; half.life: arithmetic mean and standard deviation; N: number of subjects"
    ),
    info = "summary of PKNCAresults correctly excludes all of auclast when requested"
  )
})

test_that("print.summary_PKNCAresults works", {
  tmpconc <- generate.conc(2, 1, 0:24)
  tmpdose <- generate.dose(tmpconc)
  myconc <- PKNCAconc(tmpconc, formula = conc ~ time | treatment + ID)
  mydose <- PKNCAdose(tmpdose, formula = dose ~ time | treatment + ID)
  mydata <- PKNCAdata(myconc, mydose)
  myresult <- pk.nca(mydata)

  expect_output(
    print(summary(myresult)),
    paste(
      " start end treatment N     auclast         cmax              tmax   half.life.*",
      "     0  24     Trt 1 2 13.8 \\[2.51\\]            .                 .           ..*",
      "     0 Inf     Trt 1 2           . 0.970 \\[4.29\\] 3.00 \\[2.00, 4.00\\] 14.2 \\[2.79\\].*",
      "",
      "Caption: auclast, cmax, aucinf.obs: geometric mean and geometric coefficient of variation; tmax: median and range; half.life: arithmetic mean and standard deviation",
      sep = "\n"
    )
  )
})

test_that("summary pretty_name control", {
  tmpconc <- generate.conc(2, 1, 0:24)
  tmpdose <- generate.dose(tmpconc)
  myconc <- PKNCAconc(tmpconc, formula = conc ~ time | treatment + ID)
  mydose <- PKNCAdose(tmpdose, formula = dose ~ time | treatment + ID)
  mydata <- PKNCAdata(myconc, mydose)
  myresult <- pk.nca(mydata)

  d_units_orig <- pknca_units_table(concu = "ng/mL", doseu = "mg", amountu = "mg", timeu = "hr")
  d_units_std <-
    pknca_units_table(
      concu = "ng/mL", doseu = "mg", amountu = "mg", timeu = "hr",
      conversions = data.frame(PPORRESU = "ng/mL", PPSTRESU = "mg/mL")
    )
  mydata_orig <- PKNCAdata(myconc, mydose, units = d_units_orig)
  myresult_units_orig <- pk.nca(mydata_orig)

  s_plain <- summary(myresult)
  s_pretty <- summary(myresult, pretty_names = TRUE)
  s_plain_units <- summary(myresult_units_orig, pretty_names = FALSE)
  s_pretty_units <- summary(myresult_units_orig)
  expect_equal(
    names(s_plain),
    c("start", "end", "treatment", "N", "auclast", "cmax", "tmax", "half.life", "aucinf.obs")
  )
  expect_equal(
    names(s_pretty),
    c(
      "Interval Start", "Interval End", "treatment", "N", "AUClast",
      "Cmax", "Tmax", "Half-life", "AUCinf,obs"
    )
  )
  expect_equal(
    names(s_plain_units),
    c(
      "start", "end", "treatment", "N", "auclast (hr*ng/mL)", "cmax (ng/mL)",
      "tmax (hr)", "half.life (hr)", "aucinf.obs (hr*ng/mL)"
    )
  )
  expect_equal(
    names(s_pretty_units),
    c(
      "Interval Start", "Interval End", "treatment", "N", "AUClast (hr*ng/mL)",
      "Cmax (ng/mL)", "Tmax (hr)", "Half-life (hr)", "AUCinf,obs (hr*ng/mL)"
    )
  )
  # Captions use the pretty_names, if requested
  expect_equal(
    attr(s_plain, "caption"),
    "auclast, cmax, aucinf.obs: geometric mean and geometric coefficient of variation; tmax: median and range; half.life: arithmetic mean and standard deviation; N: number of subjects"
  )
  expect_equal(
    attr(s_pretty, "caption"),
    "AUClast, Cmax, AUCinf,obs: geometric mean and geometric coefficient of variation; Tmax: median and range; Half-life: arithmetic mean and standard deviation; N: number of subjects"
  )
  # Default for pretty_names are kept
  expect_equal(
    names(s_plain),
    names(summary(myresult, pretty_names = FALSE))
  )
  expect_equal(
    names(s_pretty_units),
    names(summary(myresult_units_orig, pretty_names = TRUE))
  )
})

test_that("roundingSummarize", {
  expect_error(
    roundingSummarize(1, "foo"),
    regexp = "foo is not in the summarization instructions from PKNCA.set.summary"
  )

  PKNCA.set.summary(name = "lambda.z.n.points", description = "not a real parameter", point = mean, spread = sd, rounding = function(x) round(x, 1))
  expect_equal(roundingSummarize(1.2345, "lambda.z.n.points"), "1.2")
  PKNCA.set.summary(name = "lambda.z.n.points", description = "not a real parameter", point = mean, spread = sd, rounding = list(round = 1))
  expect_equal(roundingSummarize(1.2345, "lambda.z.n.points"), "1.2")

  # reset it
  PKNCA.set.summary(
    name = "lambda.z.n.points",
    description = "median and range",
    point = business.median,
    spread = business.range
  )
})

test_that("PKNCAresults summary counts N and n", {
  tmpconc <- generate.conc(2, 1, 0:6)
  tmpdose <- generate.dose(tmpconc)
  myconc <- PKNCAconc(tmpconc, formula = conc ~ time | treatment + ID)
  mydose <- PKNCAdose(tmpdose, formula = dose ~ time | treatment + ID)
  mydata <- PKNCAdata(myconc, mydose)
  suppressWarnings(
    myresult <- pk.nca(mydata)
  )
  o_summary <- summary(myresult)
  expect_equal(o_summary$half.life, c(".", "85.0, n=1"))
  expect_equal(
    attr(o_summary, "caption"),
    "auclast, cmax, aucinf.obs: geometric mean and geometric coefficient of variation; tmax: median and range; half.life: arithmetic mean and standard deviation; N: number of subjects; n: number of measurements included in summary"
  )
})

test_that("summary.PKNCAresults drop_param argument works", {
  tmpconc <- generate.conc(2, 1, 0:6)
  tmpdose <- generate.dose(tmpconc)
  myconc <- PKNCAconc(tmpconc, formula = conc ~ time | treatment + ID)
  mydose <- PKNCAdose(tmpdose, formula = dose ~ time | treatment + ID)
  mydata <- PKNCAdata(myconc, mydose)
  suppressWarnings(
    myresult <- pk.nca(mydata)
  )
  o_summary <- summary(myresult)
  expect_true("auclast" %in% names(o_summary))
  o_summary_noauclast <- summary(myresult, drop_param = "auclast")
  expect_false("auclast" %in% names(o_summary_noauclast))
})

test_that("pk.calc.dn", {
  # Ensure correct calculation
  expect_equal(pk.calc.dn(1, 5), 0.2)
  expect_equal(pk.calc.dn(NA, 5), NA_real_)
  expect_equal(pk.calc.dn(1, NA), NA_real_)
})

test_that("pk.calc.cmax", {
  # Ensure that the formalsmap functionality works within pk.nca
  tmpconc <- generate.conc(2, 2, 0:24)
  tmpdose <- generate.dose(tmpconc)
  tmpconc <- merge(tmpconc, tmpdose[,c("ID", "treatment", "dose")])
  myconc <- PKNCAconc(tmpconc,
                      conc~time|treatment+dose+ID)
  mydose <- PKNCAdose(tmpdose,
                      dose~time|treatment+ID)
  mydata <- PKNCAdata(myconc, mydose,
                      intervals=data.frame(start=0, end=24, auclast.dn=TRUE))
  myres <- pk.nca(mydata)
  expect_equal(myres$result$PPORRES[myres$result$PPTESTCD %in% "auclast"]/
                 myres$result$PPORRES[myres$result$PPTESTCD %in% "auclast.dn"],
               myres$result$dose[myres$result$PPTESTCD %in% "auclast"],
               info="Dose normalization works when requested as a parameter in pk.nca")
})

test_that("PKNCAresults object creation", {
  minimal_result <- PKNCAresults(data.frame(a=1), data=list())
  expect_equal(minimal_result$columns$exclude, "exclude")
  result_with_exclude_col <- PKNCAresults(data.frame(exclude=1), data=list())
  expect_equal(result_with_exclude_col$columns$exclude, "exclude.exclude")
})

test_that("PKNCAresults generation", {
  # Note that generate.conc sets the random seed, so it doesn't have
  # to happen here.
  tmpconc <- generate.conc(2, 1, 0:24)
  tmpdose <- generate.dose(tmpconc)
  myconc <- PKNCAconc(tmpconc, formula=conc~time|treatment+ID)
  mydose <- PKNCAdose(tmpdose, formula=dose~time|treatment+ID)
  mydata <- PKNCAdata(myconc, mydose)
  myresult <- pk.nca(mydata)

  expect_equal(
    names(myresult),
    c("result", "data", "columns"),
    info="Make sure that the result has the expected names (and only the expected names) in it."
  )
  expect_true(
    checkProvenance(myresult),
    info="Provenance exists and can be confirmed on results"
  )

  # Test each of the pieces for myresult for accuracy
  expect_equal(
    myresult$data, {
      tmp <- mydata
      # The options should be the default options after the
      # calculations are done.
      tmp$options <- PKNCA.options()
      tmp
    }, info="The data is just a copy of the input data plus an instantiation of the PKNCA.options"
  )

  verify.result <-
    tibble::tibble(
      treatment="Trt 1",
      ID=as.integer(rep(c(1, 2), each=15)),
      start=0,
      end=c(24, rep(Inf, 14),
            24, rep(Inf, 14)),
      PPTESTCD=rep(c("auclast", "cmax", "tmax", "tlast", "clast.obs",
                     "lambda.z", "r.squared", "adj.r.squared",
                     "lambda.z.time.first", "lambda.z.time.last", "lambda.z.n.points",
                     "clast.pred", "half.life", "span.ratio", "aucinf.obs"),
                   times=2),
      PPORRES=c(13.54, 0.9998, 4.000, 24.00, 0.3441,
                0.04297, 0.9072, 0.9021, 5.000, 24.00,
                20.00, 0.3356, 16.13, 1.178,
                21.55, 14.03, 0.9410, 2.000,
                24.00, 0.3148, 0.05689, 0.9000,
                0.8944, 5.000, 24.00, 20.00, 0.3011,
                12.18, 1.560, 19.56),
      exclude=NA_character_
    )
  expect_equal(
    myresult$result,
    verify.result,
    tolerance=0.001,
    info="The specific order of the levels isn't important-- the fact that they are factors and that the set doesn't change is important."
  )

  # Test conversion to a data.frame
  expect_equal(
    as.data.frame(myresult),
    verify.result,
    tolerance=0.001,
    info="Conversion of PKNCAresults to a data.frame in long format (default long format)"
  )
  expect_equal(
    as.data.frame(myresult, out_format="long"),
    verify.result,
    tolerance=0.001,
    info="Conversion of PKNCAresults to a data.frame in long format (specifying long format)"
  )
  expect_equal(
    as.data.frame(myresult, out_format="wide"),
    tidyr::spread(verify.result, key="PPTESTCD", value="PPORRES"),
    tolerance=0.001,
    info="Conversion of PKNCAresults to a data.frame in wide format (specifying wide format)"
  )

  tmpconc <- generate.conc(2, 1, 0:24)
  tmpdose <- generate.dose(tmpconc)
  myconc <- PKNCAconc(tmpconc, formula=conc~time|treatment+ID)
  mydose <- PKNCAdose(tmpdose, formula=dose~time|treatment+ID)
  mydata <- PKNCAdata(myconc, mydose, intervals=data.frame(start=0, end=12, aucint.inf.obs=TRUE))
  myresult <- pk.nca(mydata)

  tmpconc12 <- tmpconc
  tmpconc12$time <- tmpconc$time + 12
  tmpdose12 <- generate.dose(tmpconc12)
  myconc12 <- PKNCAconc(tmpconc12, formula=conc~time|treatment+ID)
  mydose12 <- PKNCAdose(tmpdose12, formula=dose~time|treatment+ID)
  mydata12 <- PKNCAdata(myconc12, mydose12, intervals=data.frame(start=12, end=24, aucint.inf.obs=TRUE))
  myresult12 <- pk.nca(mydata12)
  comparison_orig <- as.data.frame(myresult)
  comparison_12 <- as.data.frame(myresult12)
  expect_equal(
    comparison_orig$PPORRES[comparison_orig$PPTESTCD %in% "aucint.inf.obs"],
    comparison_12$PPORRES[comparison_12$PPTESTCD %in% "aucint.inf.obs"],
    info="Time shift does not affect aucint calculations."
  )
})

test_that("PKNCAresults has exclude, when applicable", {
  tmpconc <- generate.conc(2, 1, 0:24)
  tmpconc$conc[tmpconc$ID %in% 2] <- 0
  tmpdose <- generate.dose(tmpconc)
  myconc <- PKNCAconc(tmpconc, conc~time|treatment+ID)
  mydose <- PKNCAdose(tmpdose, dose~time|treatment+ID)
  mydata <- PKNCAdata(myconc, mydose)
  # Not capturing the warning due to R bug
  # https://bugs.r-project.org/bugzilla3/show_bug.cgi?id=17122
  #expect_warning(myresult <- pk.nca(mydata),
  #               regexp="Too few points for half-life calculation")
  suppressWarnings(myresult <- pk.nca(mydata))
  myresult_df <- as.data.frame(myresult)
  expect_true(
    all(myresult_df$PPTESTCD %in%
          c(
            "adj.r.squared", "aucinf.obs", "auclast", "clast.obs",
            "clast.pred", "cmax", "half.life", "lambda.z", "lambda.z.n.points",
            "lambda.z.time.first", "lambda.z.time.last", "r.squared",
            "span.ratio", "tlast", "tmax"
          )
    ),
    info="verify that only expected results are present"
  )
  expect_equal(
    unique(
      myresult_df$exclude[
        myresult_df$ID == 2 &
          myresult_df$PPTESTCD %in%
          c(
            "lambda.z", "r.squared", "adj.r.squared",
            "lambda.z.time.first", "lambda.z.time.last",
            "lambda.z.n.points", "clast.pred", "half.life", "span.ratio"
          )
        ]
    ),
    "Too few points for half-life calculation (min.hl.points=3 with only 0 points)",
    info="exclusions are propogated to results"
  )
  expect_equal(
    unique(
      myresult_df$exclude[
        !(myresult_df$ID == 2 &
            myresult_df$PPTESTCD %in%
            c("lambda.z", "r.squared", "adj.r.squared", "lambda.z.time.first",
              "lambda.z.time.last", "lambda.z.n.points", "clast.pred", "half.life", "span.ratio")
        )
        ]
    ),
    NA_character_,
    info="exclusions are propogated to results only when applicable"
  )
})

test_that("ptr works as a parameter", {
  tmpconc <- generate.conc(2, 1, 0:24)
  tmpdose <- generate.dose(tmpconc)
  myconc <- PKNCAconc(tmpconc, formula=conc~time|treatment+ID)
  mydose <- PKNCAdose(tmpdose, formula=dose~time|treatment+ID)
  myinterval <- data.frame(start=0, end=24, ptr=TRUE)
  mydata <- PKNCAdata(myconc, mydose, intervals=myinterval)
  myresult <- pk.nca(mydata)
  ptr_result <- as.data.frame(myresult)
  expect_equal(
    ptr_result$PPORRES[ptr_result$PPTESTCD %in% "ptr"],
    c(2.9055, 2.9885),
    tolerance=0.0001
  )
})

test_that("exclude values are maintained in derived parameters during automatic calculation (#112)", {
  my_conc <-
    data.frame(
      conc = c(0, 2.5, 3, 2.7, 2.3),
      time = 0:4,
      subject = 1
    )

  conc_obj <- PKNCAconc(my_conc, conc~time|subject)
  data_obj <-
    PKNCAdata(
      data.conc=conc_obj,
      intervals=
        data.frame(
          start=0,
          end=Inf,
          aucinf.obs=TRUE
        )
    )
  expect_message(
    expect_warning(
      results_obj <- pk.nca(data_obj),
      regexp="Too few points for half-life"
    ),
    regexp="No dose information provided"
  )
  d_results <- as.data.frame(results_obj)
  expect_equal(
    d_results$exclude[d_results$PPTESTCD == "aucinf.obs"],
    d_results$exclude[d_results$PPTESTCD == "half.life"]
  )
})

test_that("ctrough is correctly calculated", {
  my_conc <- data.frame(time=0:6, conc=2^(0:-6), subject=1)
  conc_obj <- PKNCAconc(my_conc, conc~time|subject)
  data_obj <-
    PKNCAdata(
      data.conc=conc_obj,
      intervals=
        data.frame(
          start=0,
          end=c(6, Inf),
          ctrough=TRUE
        )
    )
  expect_message(
    expect_equal(
      as.data.frame(pk.nca(data_obj))$PPORRES,
      c(2^-6, NA_real_)
    ),
    regexp="No dose information provided"
  )
})

test_that("single subject, ungrouped data works (#74)", {
  my_conc <- data.frame(time=0:6, conc=2^(0:-6))
  conc_obj <- PKNCAconc(my_conc, conc~time)
  data_obj <-
    PKNCAdata(
      data.conc=conc_obj,
      intervals=
        data.frame(
          start=0,
          end=Inf,
          cmax=TRUE
        )
    )
  expect_message(
    expect_equal(
      as.data.frame(pk.nca(data_obj))$PPORRES,
      1
    ),
    regexp="No dose information provided",
  )
})

test_that("units work for calculations and summaries with one set of units across all analytes", {
  tmpconc <- generate.conc(2, 1, 0:24)
  tmpdose <- generate.dose(tmpconc)
  myconc <- PKNCAconc(tmpconc, formula=conc~time|treatment+ID)
  mydose <- PKNCAdose(tmpdose, formula=dose~time|treatment+ID)
  mydata <- PKNCAdata(myconc, mydose)
  myresult <- pk.nca(mydata)

  d_units_orig <- pknca_units_table(concu="ng/mL", doseu="mg", amountu="mg", timeu="hr")
  d_units_std <-
    pknca_units_table(
      concu="ng/mL", doseu="mg", amountu="mg", timeu="hr",
      conversions=data.frame(PPORRESU="ng/mL", PPSTRESU="mg/mL")
    )
  mydata_orig <- PKNCAdata(myconc, mydose, units=d_units_orig)
  myresult_units_orig <- pk.nca(mydata_orig)
  mydata_std <- PKNCAdata(myconc, mydose, units=d_units_std)
  myresult_units_std <- pk.nca(mydata_std)

  # Summaries are the same except for the column names
  expect_equal(
    unname(summary(myresult)),
    unname(summary(myresult_units_orig)),
    # The caption attribute will differ
    ignore_attr = TRUE
  )
  expect_equal(
    summary(myresult_units_orig) %>% dplyr::select(-`Cmax (ng/mL)`),
    summary(myresult_units_std) %>% dplyr::select(-`Cmax (mg/mL)`)
  )
  # The units are converted to standard units, if requested
  expect_equal(
    summary(myresult_units_orig)$`Cmax (ng/mL)`,
    c(".", "0.970 [4.29]")
  )
  expect_equal(
    summary(myresult_units_std)$`Cmax (mg/mL)`,
    c(".", "9.70e-7 [4.29]")
  )
  # Wide conversion works for original and standardized units
  df_wide_orig <- as.data.frame(myresult_units_orig, out_format="wide")
  df_wide_std <- as.data.frame(myresult_units_std, out_format="wide")
  expect_equal(
    as.data.frame(myresult, out_format="wide"),
    # The difference is the addition of units to the column names
    df_wide_orig %>%
      dplyr::rename_with(.fn=gsub, pattern=" \\(.*$", replacement="")
  )
  expect_true(
    all(
      names(df_wide_orig) %in% c("treatment", "ID", "start", "end", "exclude") |
        grepl(x=names(df_wide_orig), pattern=" (", fixed=TRUE)
    )
  )
  # Everything is the same unless it is a concentration which has been converted
  expect_equal(
    df_wide_orig %>% dplyr::select(-`cmax (ng/mL)`, -`clast.obs (ng/mL)`, -`clast.pred (ng/mL)`),
    df_wide_std %>% dplyr::select(-`cmax (mg/mL)`, -`clast.obs (mg/mL)`, -`clast.pred (mg/mL)`)
  )
  # Concentration conversion works correctly
  expect_equal(
    df_wide_orig$`cmax (ng/mL)`,
    df_wide_std$`cmax (mg/mL)`*1e6
  )
})

test_that("units work for calculations and summaries with one set of units across all analytes", {
  tmpconc1 <- generate.conc(2, 1, 0:24)
  tmpconc1$analyte <- "drug1"
  tmpconc2 <- tmpconc1
  tmpconc2$analyte <- "drug2"
  tmpconc <- rbind(tmpconc1, tmpconc2)

  tmpdose <- generate.dose(tmpconc)
  myconc <- PKNCAconc(tmpconc, formula=conc~time|treatment+ID/analyte)
  mydose <- PKNCAdose(tmpdose, formula=dose~time|treatment+ID)
  mydata <- PKNCAdata(myconc, mydose)
  myresult <- pk.nca(mydata)

  d_units_std1 <-
    pknca_units_table(
      concu="ng/mL", doseu="mg", amountu="mg", timeu="hr",
      conversions=data.frame(PPORRESU="ng/mL", PPSTRESU="mg/mL")
    )
  d_units_std1$analyte <- "drug1"
  d_units_std2 <-
    pknca_units_table(
      concu="ng/mL", doseu="mg", amountu="mg", timeu="hr",
      conversions=data.frame(PPORRESU="ng/mL", PPSTRESU="mmol/L", conversion_factor=2)
    )
  d_units_std2$analyte <- "drug2"
  d_units_std <- rbind(d_units_std1, d_units_std2)
  mydata_std <- PKNCAdata(myconc, mydose, units=d_units_std)
  myresult_units_std <- pk.nca(mydata_std)
  summary_myresult_units_std <- summary(myresult_units_std)
  # Everything is the same between analytes except for "cmax"
  for (nm in setdiff(names(summary_myresult_units_std), c("analyte", "Cmax"))) {
    expect_equal(
      summary_myresult_units_std[[nm]][1:2],
      summary_myresult_units_std[[nm]][3:4]
    )
  }
  # Different units in the same column are shown in the cell
  expect_equal(
    summary_myresult_units_std$Cmax,
    c(".", "9.70e-7 [4.29] mg/mL", ".", "1.94 [4.29] mmol/L")
  )

  # I can't think of a way to trigger this error without explicit manipulation.
  myresult_units_manipulated <- myresult_units_std
  myresult_units_manipulated$result$PPSTRESU[myresult_units_manipulated$result$PPTESTCD %in% "auclast"][1] <- "foo"
  expect_error(
    summary(myresult_units_manipulated),
    regexp="Multiple units cannot be summarized together.  For auclast, trying to combine: foo, hr*ng/mL",
    fixed=TRUE
  )
})

test_that("getGroups.PKNCAresults", {
  tmpconc <- generate.conc(2, 1, 0:24)
  tmpdose <- generate.dose(tmpconc)
  myconc <- PKNCAconc(tmpconc, formula=conc~time|treatment+ID)
  mydose <- PKNCAdose(tmpdose, formula=dose~time|treatment+ID)
  mydata <- PKNCAdata(myconc, mydose)
  myresult <- pk.nca(mydata)

  expect_equal(
    getGroups(myresult, level="treatment"),
    myresult$result[, "treatment", drop=FALSE]
  )
  expect_equal(
    getGroups(myresult, level=factor("treatment")),
    myresult$result[, "treatment", drop=FALSE]
  )
  expect_error(
    getGroups(myresult, level="foo"),
    regexp="Not all levels are listed in the group names.  Missing levels are: foo"
  )
  expect_equal(
    getGroups(myresult, level=2),
    myresult$result[, c("treatment", "ID")]
  )
  expect_equal(
    getGroups(myresult, level=2:3),
    myresult$result[, c("ID", "start")]
  )
  expect_equal(
    getGroups(myresult, level=3:4),
    myresult$result[, c("start", "end")]
  )
})

test_that("group_vars.PKNCAresult", {
  o_conc_group <- PKNCAconc(as.data.frame(datasets::Theoph), conc~Time|Subject)
  o_data_group <- PKNCAdata(o_conc_group, intervals = data.frame(start = 0, end = 1, cmax = TRUE))
  suppressMessages(o_nca_group <- pk.nca(o_data_group))

  expect_equal(dplyr::group_vars(o_nca_group), "Subject")

  # Check that it works without groupings as expected [empty]
  o_conc_nongroup <- PKNCAconc(as.data.frame(datasets::Theoph)[datasets::Theoph$Subject == 1,], conc~Time)
  o_data_nogroup <- PKNCAdata(o_conc_nongroup, intervals = data.frame(start = 0, end = 1, cmax = TRUE))
  suppressMessages(o_nca_nogroup <- pk.nca(o_data_nogroup))

  expect_equal(dplyr::group_vars(o_nca_nogroup), character(0))
})

test_that("as.data.frame.PKNCAresults can filter for only requested parameters", {
  tmpconc <- generate.conc(2, 1, 0:24)
  tmpdose <- generate.dose(tmpconc)
  myconc <- PKNCAconc(tmpconc, formula=conc~time|treatment+ID)
  mydose <- PKNCAdose(tmpdose, formula=dose~time|treatment+ID)
  mydata <- PKNCAdata(myconc, mydose, intervals = data.frame(start = 0, end = Inf, half.life = TRUE))
  myresult <- pk.nca(mydata)

  expect_equal(nrow(as.data.frame(myresult)), 22)
  expect_equal(nrow(as.data.frame(myresult, filter_requested = TRUE)), 2)
})

test_that("as.data.frame.PKNCAresults can filter to remove excluded parameters", {
  tmpconc <- generate.conc(2, 1, c(0, 2, 6, 12, 24))
  tmpdose <- generate.dose(tmpconc)
  myconc <- PKNCAconc(tmpconc, formula=conc~time|treatment+ID)
  mydose <- PKNCAdose(tmpdose, formula=dose~time|treatment+ID)
  mydata <- PKNCAdata(myconc, mydose, intervals = data.frame(start = 0, end = Inf, half.life = TRUE))
  myresult <- exclude(pk.nca(mydata), FUN = exclude_nca_span.ratio(1))

  expect_equal(nrow(as.data.frame(myresult)), 22)
  expect_equal(nrow(as.data.frame(myresult, filter_excluded = TRUE)), 13)
})

#' Determine dose normalized NCA parameter
#'
#' @param parameter Parameter to dose normalize
#' @param dose Dose in units compatible with the area under the curve
#' @returns a number for dose normalized AUC
#' @examples
#' pk.calc.dn(90, 10)
#' @export
pk.calc.dn <- function(parameter, dose) {
  parameter/dose
}

local({
  for (n in c("auclast", "aucall", "aucinf.obs", "aucinf.pred",
              "aumclast", "aumcall", "aumcinf.obs", "aumcinf.pred",
              "cmax", "cmin", "clast.obs", "clast.pred", "cav", "ctrough")) {
    current_unit_type <- get.interval.cols()[[n]]$unit_type
    current_pretty_name <- get.interval.cols()[[n]]$pretty_name
    # Add the column to the interval specification
    add.interval.col(
      name=paste(n, "dn", sep="."),
      FUN="pk.calc.dn",
      values=c(FALSE, TRUE),
      unit_type=paste0(current_unit_type, "_dose"),
      pretty_name=paste(current_pretty_name, "(dose-normalized)"),
      desc=paste("Dose normalized", n),
      formalsmap=list(parameter=n),
      depends=c(n)
    )
    PKNCA.set.summary(
      name=paste(n, "dn", sep="."),
      description="geometric mean and geometric coefficient of variation",
      point=business.geomean,
      spread=business.geocv)
  }
})

#' Calculate AUC for intravenous dosing
#'
#' @details The AUC for intravenous (IV) dosing extrapolates the AUC back from
#'   the first measurement to time 0 using c0 and the AUC calculated by another
#'   method (for example the auclast).
#'
#' The calculation method takes the following steps:
#'
#' \itemize{
#'   \item{`time = 0` must be present in the data with a measured concentration.}
#'   \item{The AUC between `time = 0` and the next time point is calculated (`auc_first`).}
#'   \item{The AUC between `time = 0` with `c0` and the next time point is calculated (`auc_second`).}
#'   \item{The final AUC is the initial AUC plus the difference between the two AUCs (`auc_final <- auc + auc_second - auc_first`).}
#' }
#' @inheritParams pk.calc.auxc
#' @inheritParams PKNCA.choose.option
#' @param c0 The concentration at time 0, typically calculated using
#'   `pk.calc.c0()`
#' @param auc The AUC calculated using `conc` and `time` without `c0` (it may be
#'   calculated using any method)
#' @return `pk.calc.auciv`: The AUC calculated using `c0`
#' @export
pk.calc.auciv <- function(conc, time, c0, auc, ..., options = list(), check=TRUE) {
  if (check) {
    assert_conc_time(conc = conc, time = time)
    data <-
      clean.conc.blq(
        conc, time,
        options = options,
        check=FALSE
      )
  } else {
    data <- data.frame(conc = conc, time = time)
  }
  if (!(0 %in% time)) {
    return(structure(NA_real_, exclude="No time 0 in data"))
  } else if (is.na(c0)) {
    return(structure(NA_real_, exclude="c0 is not calculated"))
  }
  auc_first <- pk.calc.auc.last(conc = data$conc[1:2], time = data$time[1:2], ..., check=FALSE)
  auc_second <- pk.calc.auc.last(conc = c(c0, data$conc[2]), time = data$time[1:2], ..., check=FALSE)
  auc_final <- auc + auc_second - auc_first
  auc_final
}

add.interval.col(
  name = "aucivlast",
  FUN = "pk.calc.auciv",
  unit_type = "auc",
  pretty_name = "AUClast (IV dosing)",
  depends = c("auclast", "c0"),
  desc = "The AUClast calculated with back-extrapolation for intravenous dosing using extrapolated C0",
  sparse = FALSE,
  formalsmap = list(auc="auclast")
)

add.interval.col(
  name = "aucivall",
  FUN = "pk.calc.auciv",
  unit_type = "auc",
  pretty_name = "AUCall (IV dosing)",
  depends = c("aucall", "c0"),
  desc = "The AUCall calculated with back-extrapolation for intravenous dosing using extrapolated C0",
  sparse = FALSE,
  formalsmap = list(auc="aucall")
)

add.interval.col(
  name = "aucivint.last",
  FUN = "pk.calc.auciv",
  unit_type = "auc",
  pretty_name = "AUCint,last (IV dosing)",
  depends = c("aucint.last", "c0"),
  desc = "The AUCint,last calculated with back-extrapolation for intravenous dosing using extrapolated C0",
  sparse = FALSE,
  formalsmap = list(auc="aucint.last")
)

add.interval.col(
  name = "aucivint.all",
  FUN = "pk.calc.auciv",
  unit_type = "auc",
  pretty_name = "AUCint,all (IV dosing)",
  depends = c("aucint.all", "c0"),
  desc = "The AUCint,all calculated with back-extrapolation for intravenous dosing using extrapolated C0",
  sparse = FALSE,
  formalsmap = list(auc="aucint.all")
)

add.interval.col(
  name = "aucivinf.obs",
  FUN = "pk.calc.auciv",
  unit_type = "auc",
  pretty_name = "AUCinf,obs (IV dosing)",
  depends = c("aucinf.obs", "c0"),
  desc = "The AUCinf,obs calculated with back-extrapolation for intravenous dosing using extrapolated C0",
  sparse = FALSE,
  formalsmap = list(auc="aucinf.obs")
)

add.interval.col(
  name = "aucivinf.pred",
  FUN = "pk.calc.auciv",
  unit_type = "auc",
  pretty_name = "AUCinf,pred (IV dosing)",
  depends = c("aucinf.pred", "c0"),
  desc = "The  calculated with back-extrapolation for intravenous dosing using extrapolated C0",
  sparse = FALSE,
  formalsmap = list(auc="aucinf.pred")
)

PKNCA.set.summary(
  name=c("aucivlast", "aucivall", "aucivint.last", "aucivint.all", "aucivinf.obs", "aucivinf.pred"),
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)

#' @describeIn pk.calc.auciv Calculate the percent back-extrapolated AUC for IV
#'   administration
#' @details The calculation for back-extrapolation is `100*(1 - auc/auciv)`.
#' @param auciv The AUC calculated using `c0`
#' @returns `pk.calc.auciv_pctbackextrap`: The AUC percent back-extrapolated
#' @export
pk.calc.auciv_pbext <- function(auc, auciv) {
  100*(1 - auc/auciv)
}

add.interval.col(
  name = "aucivpbextlast",
  FUN = "pk.calc.auciv_pbext",
  unit_type = "%",
  pretty_name = "AUCbext (based on AUClast)",
  depends = c("auclast", "aucivlast"),
  desc = "The back-extrapolation percent for intravenous dosing based on AUClast",
  sparse = FALSE,
  formalsmap = list(auc="auclast", auciv="aucivlast")
)

add.interval.col(
  name = "aucivpbextall",
  FUN = "pk.calc.auciv_pbext",
  unit_type = "%",
  pretty_name = "AUCbext (based on AUCall)",
  depends = c("aucall", "aucivall"),
  desc = "The back-extrapolation percent for intravenous dosing based on AUCall",
  sparse = FALSE,
  formalsmap = list(auc="aucall", auciv="aucivall")
)

add.interval.col(
  name = "aucivpbextint.last",
  FUN = "pk.calc.auciv_pbext",
  unit_type = "%",
  pretty_name = "AUCbext (based on AUCint,last)",
  depends = c("aucint.last", "aucivint.last"),
  desc = "The back-extrapolation percent for intravenous dosing based on AUCint,last",
  sparse = FALSE,
  formalsmap = list(auc="aucint.last", auciv="aucivint.last")
)

add.interval.col(
  name = "aucivpbextint.all",
  FUN = "pk.calc.auciv_pbext",
  unit_type = "%",
  pretty_name = "AUCbext (based on AUCint,all)",
  depends = c("aucint.all", "aucivint.all"),
  desc = "The back-extrapolation percent for intravenous dosing based on AUCint,all",
  sparse = FALSE,
  formalsmap = list(auc="aucint.all", auciv="aucivint.all")
)

add.interval.col(
  name = "aucivpbextinf.obs",
  FUN = "pk.calc.auciv_pbext",
  unit_type = "%",
  pretty_name = "AUCbext (based on AUCinf,obs)",
  depends = c("aucinf.obs", "aucivinf.obs"),
  desc = "The back-extrapolation percent for intravenous dosing based on AUCinf,obs",
  sparse = FALSE,
  formalsmap = list(auc="aucinf.obs", auciv="aucivinf.obs")
)

add.interval.col(
  name = "aucivpbextinf.pred",
  FUN = "pk.calc.auciv_pbext",
  unit_type = "%",
  pretty_name = "AUCbext (based on AUCinf,pred)",
  depends = c("aucinf.pred", "aucivinf.pred"),
  desc = "The back-extrapolation percent for intravenous dosing based on AUCinf,pred",
  sparse = FALSE,
  formalsmap = list(auc="aucinf.pred", auciv="aucivinf.pred")
)

PKNCA.set.summary(
  name = c("aucivpbextlast", "aucivpbextall", "aucivpbextint.last", "aucivpbextint.all", "aucivpbextinf.obs", "aucivpbextinf.pred"),
  description="arithmetic mean and standard deviation",
  point=business.mean,
  spread=business.sd
)

test_that("pk.calc.auciv", {
  expect_equal(
    pk.calc.auciv(conc = 1:2, time = 1:2),
    structure(NA_real_, exclude="No time 0 in data")
  )
  expect_equal(
    # No check is done to confirm that the auc argument matches the data
    pk.calc.auciv(conc = 0:5, time = 0:5, c0 = 1, auc = 2.75),
    2.75 + 1 - 0.5
  )
  expect_equal(
    # No verifications are made on the data
    pk.calc.auciv(conc = 0:5, time = 0:5, c0 = 1, auc = 2.75, check=FALSE),
    2.75 + 1 - 0.5
  )
})

test_that("pk.calc.auciv_pbext", {
  expect_equal(
    pk.calc.auciv_pbext(auc = 1, auciv = 2.1),
    100 * (1 - 1/2.1)
  )
})

test_that("NA data are removed from concentrations for calculation of AUCiv (#353)", {
  d_iv_353alt <- data.frame(conc = c(NA, 4, 2, 1, 0.45), time = c(0, 5, 15, 30, 60))
  d_intervals <- data.frame(start = 0, end = Inf, aucivinf.obs = TRUE)
  o_conc_353alt <- PKNCAconc(data = d_iv_353alt, conc~time)
  o_dose <- PKNCAdose(data = data.frame(time = 0), ~time)
  o_data_353alt <- PKNCAdata(o_conc_353alt, o_dose, intervals = d_intervals)
  # The same warning is expected three times
  expect_warning(expect_warning(expect_warning(
    o_nca <- pk.nca(o_data_353alt),
    regexp = "Requesting an AUC range starting (0) before the first measurement (5) is not allowed",
    fixed = TRUE),
    regexp = "Requesting an AUC range starting (0) before the first measurement (5) is not allowed",
    fixed = TRUE),
    regexp = "Requesting an AUC range starting (0) before the first measurement (5) is not allowed",
    fixed = TRUE
  )
  expect_s3_class(o_nca, "PKNCAresults")
})

test_that("missing dose information does not cause NA time (#353)", {
  d_iv_nodose <- data.frame(conc = c(4, 2, 1, 0.45), time = c(5, 15, 30, 60))
  d_intervals <- data.frame(start = 0, end = Inf, aucivinf.obs = TRUE)
  o_conc_nodose <- PKNCAconc(data = d_iv_nodose, conc~time)
  o_data_nodose <- PKNCAdata(o_conc_nodose, intervals = d_intervals, impute = "start_conc0")
  expect_warning(
    o_nca <- pk.nca(o_data_nodose),
    regexp = "time.dose is NA"
  )
  expect_s3_class(o_nca, "PKNCAresults")
})

#' Calculate amount excreted (typically in urine or feces)
#'
#' @details ae is `sum(conc*volume)`.
#'
#' @inheritParams assert_conc_time
#' @param volume The volume (or mass) of the sample
#' @param check Should the concentration and volume data be checked?
#' @return The amount excreted during the interval
#' @details The units for the concentration and volume should match such that
#'   `sum(conc*volume)` has units of mass or moles.
#' @seealso [pk.calc.clr()], [pk.calc.fe()]
#' @export
pk.calc.ae <- function(conc, volume, check=TRUE) {
  mask_missing_conc <- is.na(conc)
  mask_missing_vol <- is.na(volume)
  mask_missing_both <- mask_missing_conc & mask_missing_vol
  mask_missing_conc <- mask_missing_conc & !mask_missing_both
  mask_missing_vol <- mask_missing_vol & !mask_missing_both
  message_both <- message_conc <- message_vol <- NA_character_
  if (all(mask_missing_both)) {
    message_both <- "All concentrations and volumes are missing"
  } else if (any(mask_missing_both)) {
    message_both <- sprintf("%g of %g concentrations and volumes are missing", sum(mask_missing_both), length(conc))
  }
  if (all(mask_missing_conc)) {
    message_conc <- "All concentrations are missing"
  } else if (any(mask_missing_conc)) {
    message_conc <- sprintf("%g of %g concentrations are missing", sum(mask_missing_conc), length(conc))
  }
  if (all(mask_missing_vol)) {
    message_vol <- "All volumes are missing"
  } else if (any(mask_missing_vol)) {
    message_vol <- sprintf("%g of %g volumes are missing", sum(mask_missing_vol), length(conc))
  }
  message_all <- stats::na.omit(c(message_both, message_conc, message_vol))
  ret <- sum(conc*volume)
  if (length(message_all) != 0) {
    message <- paste(message_all, collapse = "; ")
    ret <- structure(ret, exclude = message)
  }
  ret
}
add.interval.col("ae",
                 FUN="pk.calc.ae",
                 values=c(FALSE, TRUE),
                 unit_type="amount",
                 pretty_name="Amount excreted",
                 desc="The amount excreted (typically into urine or feces)")
PKNCA.set.summary(
  name="ae",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)

#' Calculate renal clearance
#'
#' @details clr is `sum(ae)/auc`.
#'
#' @param ae The amount excreted in urine (as a numeric scalar or vector)
#' @param auc The area under the curve (as a numeric scalar or vector)
#' @returns The renal clearance as a number
#' @details The units for the `ae` and `auc` should match such that `ae/auc` has
#'   units of volume/time.
#' @seealso [pk.calc.ae()], [pk.calc.fe()]
#' @export
pk.calc.clr <- function(ae, auc) {
  sum(ae)/auc
}
add.interval.col("clr.last",
                 FUN="pk.calc.clr",
                 values=c(FALSE, TRUE),
                 unit_type="renal_clearance",
                 pretty_name="Renal clearance (from AUClast)",
                 formalsmap=list(auc="auclast"),
                 desc="The renal clearance calculated using AUClast")
PKNCA.set.summary(
  name="clr.last",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)
add.interval.col("clr.obs",
                 FUN="pk.calc.clr",
                 values=c(FALSE, TRUE),
                 unit_type="renal_clearance",
                 pretty_name="Renal clearance (from AUCinf,obs)",
                 formalsmap=list(auc="aucinf.obs"),
                 desc="The renal clearance calculated using AUCinf,obs")
PKNCA.set.summary(
  name="clr.obs",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)
add.interval.col("clr.pred",
                 FUN="pk.calc.clr",
                 values=c(FALSE, TRUE),
                 unit_type="renal_clearance",
                 pretty_name="Renal clearance (from AUCinf,pred)",
                 formalsmap=list(auc="aucinf.pred"),
                 desc="The renal clearance calculated using AUCinf,pred")
PKNCA.set.summary(
  name="clr.pred",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)

#' Calculate fraction excreted (typically in urine or feces)
#'
#' @details fe is `sum(ae)/dose`
#'
#' @param ae The amount excreted (as a numeric scalar or vector)
#' @param dose The dose (as a numeric scalar or vector)
#' @returns The fraction of dose excreted
#' @details   The units for `ae` and `dose` should be the same so that `ae/dose`
#'   is a unitless fraction.
#' @seealso [pk.calc.ae()], [pk.calc.clr()]
#' @export
pk.calc.fe <- function(ae, dose) {
  sum(ae)/dose
}
add.interval.col("fe",
                 FUN="pk.calc.fe",
                 unit_type="amount_dose",
                 pretty_name="Fraction excreted",
                 values=c(FALSE, TRUE),
                 desc="The fraction of the dose excreted")
PKNCA.set.summary(
  name="fe",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)

test_that("pk.calc.ae", {
  expect_equal(
    pk.calc.ae(conc=1:5, volume=1:5),
    sum((1:5)^2)
  )
  expect_equal(
    pk.calc.ae(conc = NA, volume = NA),
    structure(NA_real_, exclude = "All concentrations and volumes are missing")
  )
  expect_equal(
    pk.calc.ae(conc = c(1, NA), volume = c(1, NA)),
    structure(NA_real_, exclude = "1 of 2 concentrations and volumes are missing")
  )
  expect_equal(
    pk.calc.ae(conc = c(NA, NA), volume = c(1, 1)),
    structure(NA_real_, exclude = "All concentrations are missing")
  )
  expect_equal(
    pk.calc.ae(conc = c(1, NA), volume = c(1, 1)),
    structure(NA_real_, exclude = "1 of 2 concentrations are missing")
  )
  expect_equal(
    pk.calc.ae(conc = c(1, 1), volume = c(NA, NA)),
    structure(NA_real_, exclude = "All volumes are missing")
  )
  expect_equal(
    pk.calc.ae(conc = c(1, 1), volume = c(NA, 1)),
    structure(NA_real_, exclude = "1 of 2 volumes are missing")
  )
  expect_equal(
    pk.calc.ae(conc = c(NA, NA, 1, 1), volume = c(NA, 1, NA, 1)),
    structure(NA_real_, exclude = "1 of 4 concentrations and volumes are missing; 1 of 4 concentrations are missing; 1 of 4 volumes are missing")
  )
})

test_that("pk.calc.clr", {
  expect_equal(pk.calc.clr(ae=1, auc=10),
               0.1,
               info="CLr is calculated correctly with both scalars")
  expect_equal(pk.calc.clr(ae=c(1, 2), auc=10),
               0.3,
               info="CLr is calculated correctly with a vector Ae and a scalar AUC")
  expect_equal(pk.calc.clr(ae=c(1, 2), auc=c(1, 10)),
               c(3, 0.3),
               info="CLr is calculated correctly with both vectors (but that is not the likely calculation method)")
})

test_that("pk.calc.fe", {
  expect_equal(pk.calc.fe(1, 10),
               0.1,
               info="fe is calculated correctly with both scalars")
  expect_equal(pk.calc.fe(c(1, 2), 10),
               0.3,
               info="fe is calculated correctly with both vector/scalar")
})

#' Exclude NCA parameters based on examining the parameter set.
#'
#' @param min.span.ratio The minimum acceptable span ratio (uses
#'   `PKNCA.options("min.span.ratio")` if not provided).
#' @param max.aucinf.pext The maximum acceptable percent AUC extrapolation (uses
#'   `PKNCA.options("max.aucinf.pext")` if not provided).
#' @param min.hl.r.squared The minimum acceptable r-squared value for half-life
#'   (uses `PKNCA.options("min.hl.r.squared")` if not provided).
#' @param min.hl.adj.r.squared The minimum acceptable adjusted r-squared for half-life
#'   (uses 0.9 if not provided).
#' @examples
#' my_conc <- PKNCAconc(data.frame(conc=1.1^(3:0),
#'                                 time=0:3,
#'                                 subject=1),
#'                      conc~time|subject)
#' my_data <- PKNCAdata(my_conc,
#'                      intervals=data.frame(start=0, end=Inf,
#'                                           aucinf.obs=TRUE,
#'                                           aucpext.obs=TRUE))
#' my_result <- pk.nca(my_data)
#' my_result_excluded <- exclude(my_result,
#'                               FUN=exclude_nca_max.aucinf.pext())
#' as.data.frame(my_result_excluded)
#' @name exclude_nca
#' @family Result exclusions
NULL

#' @describeIn exclude_nca Exclude based on span.ratio
#' @export
exclude_nca_span.ratio <- function(min.span.ratio) {
  missing_min.span.ratio <- missing(min.span.ratio)
  if (missing_min.span.ratio) {
    min.span.ratio <- PKNCA.options("min.span.ratio")
  }
  exclude_nca_by_param(
    parameter = "span.ratio",
    min_thr = min.span.ratio,
    affected_parameters = get.parameter.deps("half.life")
  )
}

#' @describeIn exclude_nca Exclude based on AUC percent extrapolated (both observed and predicted)
#' @export
exclude_nca_max.aucinf.pext <- function(max.aucinf.pext) {
  missing_max.aucinf.pext <- missing(max.aucinf.pext)
  if (missing_max.aucinf.pext) {
    max.aucinf.pext <- PKNCA.options("max.aucinf.pext")
  }
  # Exclude for both obs and pred
  function(x, ...) {
    res_obs <- exclude_nca_by_param(
      parameter = "aucpext.obs",
      max_thr = max.aucinf.pext,
      affected_parameters = get.parameter.deps("aucinf.obs")
    )(x, ...)
    res_pred <- exclude_nca_by_param(
      parameter = "aucpext.pred",
      max_thr = max.aucinf.pext,
      affected_parameters = get.parameter.deps("aucinf.pred")
    )(x, ...)
    # Combine results, prioritizing non-NA from either
    is.obs <- grepl(paste0("aucpext.obs > ", max.aucinf.pext), res_obs)
    is.pred <- grepl(paste0("aucpext.pred > ", max.aucinf.pext), res_pred)
    ret <- ifelse(
      is.obs | is.pred,
      gsub(
        pattern = paste0("aucpext...+ > ", max.aucinf.pext),
        replacement = paste0("aucpext > ", max.aucinf.pext),
        x = dplyr::coalesce(res_obs, res_pred)
      ),
      res_obs
    )
    ret
  }
}

#' @describeIn exclude_nca Exclude AUC measurements based on count of concentrations measured and not below the lower limit of quantification
#' @param min_count Minimum number of measured concentrations
#' @param exclude_param_pattern Character vector of regular expression patterns to exclude
#' @export
exclude_nca_count_conc_measured <- function(min_count, exclude_param_pattern = c("^aucall", "^aucinf", "^aucint", "^auciv", "^auclast", "^aumc", "^sparse_auc")) {
  all_parameters <- names(PKNCA::get.interval.cols())
  affected_parameters_base <-
    sort(unique(unlist(
      lapply(
        X = exclude_param_pattern,
        FUN = grep,
        x = all_parameters,
        value = TRUE
      )
    )))
  affected_parameters <-
    sort(unique(unlist(
      lapply(
        X = affected_parameters_base,
        FUN = get.parameter.deps
      )
    )))
  exclude_nca_by_param(
    parameter = "count_conc_measured",
    min_thr = min_count,
    affected_parameters = affected_parameters
  )
}

#' @describeIn exclude_nca Exclude based on half-life r-squared
#' @export
exclude_nca_min.hl.r.squared <- function(min.hl.r.squared) {
  missing_min.hl.r.squared <- missing(min.hl.r.squared)
  if (missing_min.hl.r.squared) {
    min.hl.r.squared <- PKNCA.options("min.hl.r.squared")
  }
  exclude_nca_by_param(
    parameter = "r.squared",
    min_thr = min.hl.r.squared,
    affected_parameters = get.parameter.deps("half.life")
  )
}

#' @describeIn exclude_nca Exclude based on half-life adjusted r-squared
#' @export
exclude_nca_min.hl.adj.r.squared <- function(min.hl.adj.r.squared = 0.9) {
  exclude_nca_by_param(
    parameter = "adj.r.squared",
    min_thr = min.hl.adj.r.squared,
    affected_parameters = get.parameter.deps("half.life")
  )
}

#' @describeIn exclude_nca Exclude based on implausibly early Tmax (often used for extravascular dosing with a Tmax value of 0)
#' @param tmax_early The time for Tmax which is considered too early to be a valid NCA result
#' @export
exclude_nca_tmax_early <- function(tmax_early = 0) {
    
    exc_fun <- exclude_nca_by_param(
      parameter = "tmax",
      min_thr = tmax_early,
      affected_parameters = names(get.interval.cols())
    )

    function(x, ...) {
    # Get the exclusion messages
    ret <- exc_fun(x, ...)
    # Add the special annotation for tmax_early cases (if not already annotated)
    tmax_cases <- grepl(pattern = "^tmax < ", x = ret)
    tmax_already_annotated <- grepl("(likely missed dose, insufficient PK samples, or PK sample swap)", x = ret, fixed = TRUE)
    ret <- ifelse(
        test = tmax_cases & !tmax_already_annotated & !is.na(ret),
        yes = gsub(
          pattern = paste0("^tmax < ", tmax_early),
          replacement = paste0("tmax < ", tmax_early, " (likely missed dose, insufficient PK samples, or PK sample swap)"),
          x = ret
        ),
        no = ret
      )
    ret
    }
}

#' @describeIn exclude_nca Exclude based on implausibly early Tmax (special case for `tmax_early = 0`)
#' @export
exclude_nca_tmax_0 <- function() {
  exc_fun <- exclude_nca_tmax_early(1e-99)
  function(x, ...) {
    ret <- exc_fun(x, ...)

    # Replace the messages
    ret <- gsub(
      pattern = "^tmax < 1e-99",
      replacement = "tmax <= 0",
      x = ret
    )
    ret
  }
}


#' Exclude NCA Results Based on Parameter Thresholds
#'
#' Exclude rows from NCA results based on specified thresholds for a given parameter.
#' This function allows users to define minimum and/or maximum acceptable values
#' for a parameter and excludes rows that fall outside these thresholds.
#'
#' @param parameter The name of the PKNCA parameter to evaluate (e.g., "span.ratio").
#' @param min_thr The minimum acceptable value for the parameter. If not provided, is not applied.
#' @param max_thr The maximum acceptable value for the parameter. If not provided, is not applied.
#' @param affected_parameters Character vector of PKNCA parameters that will be marked as excluded.
#'                            By default is the defined parameter.
#' @returns A function that can be used with `PKNCA::exclude` to mark through the 'exclude'  column
#'          the rows in the PKNCA results based on the specified thresholds for a parameter.
#' @examples
#' # Example dataset
#' my_data <- PKNCA::PKNCAdata(
#'   PKNCA::PKNCAconc(data.frame(conc = 5:1,
#'                               time = 0:4,
#'                               subject = 1),
#'                    conc ~ time | subject),
#'   PKNCA::PKNCAdose(data.frame(subject = 1, dose = 100, time = 0),
#'                    dose ~ time | subject)
#' )
#' my_result <- PKNCA::pk.nca(my_data)
#'
#' # Exclude rows where span.ratio is less than 2
#' excluded_result <- PKNCA::exclude(
#'   my_result,
#'   FUN = exclude_nca_by_param("span.ratio", min_thr = 2)
#' )
#' as.data.frame(excluded_result)
#'
#' @export

exclude_nca_by_param <- function(
  parameter,
  min_thr = NULL,
  max_thr = NULL,
  affected_parameters = parameter
) {
  # Check that defined thresholds are single numeric objects
  checkmate::expect_number(min_thr, finite = TRUE, null.ok = TRUE)
  checkmate::expect_number(max_thr, finite = TRUE, null.ok = TRUE)

  if (isTRUE(min_thr > max_thr))
    stop("if both defined min_thr must be less than max_thr")

  function(x, ...) {
    ret <- rep(NA_character_, nrow(x))
    idx_param <- which(x$PPTESTCD == parameter)
    idx_aff_params <- which(x$PPTESTCD %in% affected_parameters)

    if (length(idx_param) > 1)
      stop("Should not see more than one ", parameter, " (please report this as a bug)")

    if (length(idx_param) == 1 && !is.na(x$PPORRES[idx_param]) && length(idx_aff_params) > 0) {
      current_value <- x$PPORRES[idx_param]
      pretty_name <- parameter # Pretty name did not convince me for some parameters (e.g, "r.squared")
      if (isTRUE(current_value < min_thr)) {
        ret[idx_aff_params] <- sprintf("%s < %g", pretty_name, min_thr)
      } else if (isTRUE(current_value > max_thr)) {
        ret[idx_aff_params] <- sprintf("%s > %g", pretty_name, max_thr)
      }
    }
    ret
  }
}

test_that("exclude_nca", {
  my_conc <- PKNCAconc(data.frame(conc=c(1.1^(3:0), 1.1), time=0:4, subject=1), conc~time|subject)
  my_data <- PKNCAdata(my_conc, intervals=data.frame(start=0, end=Inf, aucinf.obs=TRUE, aucpext.obs=TRUE))
  suppressMessages(
    my_result <- pk.nca(my_data)
  )

  my_result_excluded <- exclude(my_result, FUN=exclude_nca_max.aucinf.pext())
  expect_equal(as.data.frame(my_result_excluded)$exclude,
               c(rep(NA_character_, nrow(my_result_excluded$result)-2),
                 rep("aucpext > 20", 2)))

  my_result_excluded <- exclude(my_result, FUN=exclude_nca_max.aucinf.pext(50))
  expect_equal(as.data.frame(my_result_excluded)$exclude,
               c(rep(NA_character_, nrow(my_result_excluded$result)-2),
                 rep("aucpext > 50", 2)))

  my_result_excluded <- exclude(my_result, FUN=exclude_nca_span.ratio())
  expect_equal(as.data.frame(my_result_excluded)$exclude,
               c(rep(NA_character_, 4),
                 rep("span.ratio < 2", 11)))
  my_result_excluded <- exclude(my_result, FUN=exclude_nca_span.ratio(1))
  expect_equal(as.data.frame(my_result_excluded)$exclude,
               c(rep(NA_character_, 4),
                 rep("span.ratio < 1", 11)))

  my_result_excluded <- exclude(my_result, FUN=exclude_nca_min.hl.r.squared())
  expect_equal(as.data.frame(my_result_excluded)$exclude,
               c(rep(NA_character_, 4),
                 rep("r.squared < 0.9", 11)))
  my_result_excluded <- exclude(my_result, FUN=exclude_nca_min.hl.r.squared(0.95))
  expect_equal(as.data.frame(my_result_excluded)$exclude,
               c(rep(NA_character_, 4),
                 rep("r.squared < 0.95", 11)))

  my_result_excluded <- exclude(my_result, FUN=exclude_nca_min.hl.adj.r.squared())
  expect_equal(as.data.frame(my_result_excluded)$exclude,
               c(rep(NA_character_, 4),
                 rep("adj.r.squared < 0.9", 11)))
  my_result_excluded <- exclude(my_result, FUN=exclude_nca_min.hl.adj.r.squared(0.95))
  expect_equal(as.data.frame(my_result_excluded)$exclude,
               c(rep(NA_character_, 4),
                 rep("adj.r.squared < 0.95", 11)))

  my_data <- PKNCAdata(my_conc, intervals=data.frame(start=0, end=Inf, cmax=TRUE))
  suppressMessages(
    my_result <- pk.nca(my_data)
  )
  expect_equal(my_result,
               exclude(my_result, FUN=exclude_nca_max.aucinf.pext()),
               info="Result is ignored when not calculated")
  expect_equal(my_result,
               exclude(my_result, FUN=exclude_nca_span.ratio()),
               info="Result is ignored when not calculated")
  expect_equal(my_result,
               exclude(my_result, FUN=exclude_nca_min.hl.r.squared()),
               info="Result is ignored when not calculated")
  expect_equal(my_result,
               exclude(my_result, FUN=exclude_nca_min.hl.adj.r.squared()),
               info="Result is ignored when not calculated")
})

test_that("exclude_nca_max.aucinf.pext", {
  my_conc <- PKNCAconc(data.frame(conc=c(1.1^(3:0), 1.1), time=0:4, subject=1), conc~time|subject)
  my_data <- PKNCAdata(my_conc, intervals=data.frame(start=0, end=Inf, aucpext.pred=TRUE, aucpext.obs=TRUE))
  suppressMessages(
    my_result <- pk.nca(my_data)
  )
  expect_equal(
    as.data.frame(my_result)$exclude,
    rep(NA_character_, nrow(as.data.frame(my_result)))
  )
  my_result_exclude_20 <- exclude(my_result, FUN = exclude_nca_max.aucinf.pext(max.aucinf.pext = 20))
  expect_equal(
    as.data.frame(my_result_exclude_20)$exclude,
    c(rep(NA_character_, nrow(as.data.frame(my_result_exclude_20))-4), rep("aucpext > 20", 4))
  )
  my_result_exclude_50 <- exclude(my_result, FUN = exclude_nca_max.aucinf.pext(max.aucinf.pext = 50))
  expect_equal(
    as.data.frame(my_result_exclude_50)$exclude,
    c(rep(NA_character_, nrow(as.data.frame(my_result_exclude_50))-4), rep("aucpext > 50", 4))
  )
})

test_that("exclude_nca_count_conc_measured", {
  my_conc <- PKNCAconc(data.frame(conc=c(1.1^(c(3:0, -Inf)), 1.1), time=0:5, subject = 1), conc~time|subject)
  my_data <- PKNCAdata(my_conc, intervals=data.frame(start=0, end=Inf, aucinf.obs=TRUE, aucpext.obs=TRUE, count_conc_measured = TRUE))
  suppressMessages(
    my_result <- pk.nca(my_data)
  )
  expect_equal(
    as.data.frame(my_result)$exclude,
    rep(NA_character_, 16)
  )
  my_result_exclude5 <- exclude(my_result, FUN = exclude_nca_count_conc_measured(min_count = 5))
  expect_equal(
    as.data.frame(my_result_exclude5)$exclude,
    rep(NA_character_, 16)
  )
  my_result_exclude10 <- exclude(my_result, FUN = exclude_nca_count_conc_measured(min_count = 10))
  expect_equal(
    as.data.frame(my_result_exclude10)$exclude,
    c("count_conc_measured < 10", rep(NA_character_, 13), rep("count_conc_measured < 10", 2))
  )
})

test_that("exclude_nca_tmax_early", {
  my_conc <- PKNCAconc(data.frame(conc=c(1.1^(c(3:0, -Inf)), 1.1), time=0:5, subject = 1), conc~time|subject)
  my_data <- PKNCAdata(my_conc, intervals=data.frame(start=0, end=Inf, auclast = TRUE, half.life = TRUE))
  suppressMessages(
    my_result <- pk.nca(my_data)
  )
  expect_equal(
    as.data.frame(my_result)$exclude,
    rep(NA_character_, nrow(as.data.frame(my_result)))
  )
  my_result_exclude_1 <- exclude(my_result, FUN = exclude_nca_tmax_early(tmax_early = 1))
  expect_equal(
    as.data.frame(my_result_exclude_1)$exclude,
    rep("tmax < 1 (likely missed dose, insufficient PK samples, or PK sample swap)", nrow(as.data.frame(my_result_exclude_1)))
  )
  my_result_exclude_0 <- exclude(my_result, FUN = exclude_nca_tmax_0())
  expect_equal(
    as.data.frame(my_result_exclude_0)$exclude,
    rep("tmax <= 0 (likely missed dose, insufficient PK samples, or PK sample swap)", nrow(as.data.frame(my_result_exclude_0)))
  )
  # This should never happen in real code
  expect_error(
    exclude_nca_tmax_early()(data.frame(PPTESTCD = "tmax", PPORRES = 1:2)),
    regexp = "Should not see more than one tmax (please report this as a bug)",
    fixed = TRUE
  )
})

test_that("exclude_nca_by_param works as expected", {
  # Define the input
  my_conc <- PKNCAconc(data.frame(conc=c(1.1^(3:0), 1.1), time=0:4, subject=1), conc~time|subject)
  my_data <- PKNCAdata(my_conc, intervals=data.frame(start=0, end=Inf, span.ratio=TRUE))
  suppressMessages(
    my_result <- pk.nca(my_data)
  )

  # excludes rows based on min_thr
  res_min_excluded <- PKNCA::exclude(
    my_result,
    FUN = exclude_nca_by_param("span.ratio", min_thr = 100)
  )
  expect_equal(
    as.data.frame(res_min_excluded)$exclude,
    c(rep(NA, 10), "span.ratio < 100")
  )

  # does not exclude rows when min_thr is not met
  res_min_not_excluded <- PKNCA::exclude(
    my_result,
    FUN = exclude_nca_by_param("span.ratio", min_thr = 0.01)
  )
  expect_equal(
    as.data.frame(res_min_not_excluded)$exclude,
    rep(NA_character_, 11)
  )

  # excludes rows based on max_thr
  res_max_excluded <- PKNCA::exclude(
    my_result,
    FUN = exclude_nca_by_param("span.ratio", max_thr = 0.01)
  )
  expect_equal(
    as.data.frame(res_max_excluded)$exclude,
    c(rep(NA, 10), "span.ratio > 0.01")
  )

  # does not exclude rows when max_thr is not exceeded
  res_max_not_excluded <- PKNCA::exclude(
    my_result,
    FUN = exclude_nca_by_param("span.ratio", max_thr = 100)
  )
  expect_equal(
    as.data.frame(res_max_not_excluded)$exclude,
    rep(NA_character_, 11)
  )

  # throws an error for invalid min_thr
  expect_error(
    exclude_nca_by_param("span.ratio", min_thr = "invalid"),
    "Check on 'min_thr' failed: Must be of type 'number'"
  )

  # throws an error for invalid max_thr
  expect_error(
    exclude_nca_by_param(parameter = "span.ratio", max_thr = c(1, 2)),
    "Check on 'max_thr' failed: Must have length 1"
  )

  # throws an error when min_thr is greater than max_thr
  expect_error(
    exclude_nca_by_param("span.ratio", min_thr = 10, max_thr = 5),
    "if both defined min_thr must be less than max_thr"
  )

  # returns the original object when the parameter is not found
  res <- PKNCA::exclude(my_result, FUN = exclude_nca_by_param("nonexistent", min_thr = 0))
  expect_true(all(is.na(as.data.frame(res)$exclude)))

  # returns the object when the parameter's value is NA
  my_result_na <- my_result
  my_result_na$result$PPORRES <- NA
  res <- PKNCA::exclude(
    my_result_na,
    FUN = exclude_nca_by_param("span.ratio", min_thr = 0)
  )
  expect_true(all(is.na(as.data.frame(res)$exclude)))

  # marks records associated with the affected_parameters
  res <- PKNCA::exclude(
    my_result,
    FUN = exclude_nca_by_param(
      "span.ratio", min_thr = 0.01, affected_parameters = c("lambda.z", "span.ratio")
    )
  )
  # All span.ratio records should be NA (not excluded)
  expect_true(all(is.na(as.data.frame(res)$exclude[res$result$PPTESTCD == "span.ratio"])))
  expect_true(all(is.na(as.data.frame(res)$exclude[res$result$PPTESTCD == "lambda.z"])))

  # produces an error when more than 1 PPORRES is per parameter (should never happen in real code)
  expect_error(
    exclude_nca_by_param(
      param = "r.squared",
      min_thr = 0.7
    )(data.frame(PPTESTCD = "r.squared", PPORRES = c(1, 1))),
    regexp = "Should not see more than one r.squared (please report this as a bug)",
    fixed = TRUE
  )
})

#' The following functions are defunct
#' @name defunct
#' @param ... Ignored
NULL

#' @describeIn defunct Defunct as of version 0.11
#' @export
check.conc.time <- function(...) {
  .Defunct("assert_conc_time") # nocov
}

# Notes for deprecated items that should be made defunct:
# * interp.extrap.conc arguments interp.method and extrap.method

#' Find the first occurrence of an operator in a formula and return
#' the left, right, or both sides of the operator.
#'
#' @param x The formula to parse
#' @param op The operator to search for (e.g. `+`, `-`, `*`, `/`, ...)
#' @param side Which side of the operator would you like to see: 'left',
#'   'right', or 'both'.
#' @returns The side of the operator requested, NA if requesting the left side
#'   of a unary operator, and NULL if the operator is not found.
#' @family Formula parsing
findOperator <- function(x, op, side) {
  side <- match.arg(tolower(side),
                    choices=c("left", "right", "both"))
  if (inherits(x, "name")) {
    # This is a specific variable, we never found the operator going
    # down this branch of the tree.
    return(NULL)
  } else if (is.null(x)) {
    return(NULL)
  } else if (inherits(x, "call") |
             inherits(x, "formula") |
             inherits(x, "(")) {
    # This is all or part of a formula
    op <- as.name(op)
    if (identical(x[[1]], op)) {
      # We found the operator
      if (length(x) == 1) {
        stop("call or formula with length 1 found after finding the operator, unknown how to proceed") # nocov
      } else if (length(x) == 2) {
        # Unary operators have a right hand side only
        if (side == "left") {
          return(NA)
        } else if (side == "right") {
          return(x[[2]])
        } else if (side == "both") {
          return(x)
        }
        stop("Unknown side with a found unary operator") # nocov
      } else if (length(x) == 3) {
        # Binary operator
        if (side == "left") {
          return(x[[2]])
        } else if (side == "right") {
          return(x[[3]])
        } else if (side == "both") {
          return(x)
        }
        stop("Unknown side with a found binary operator") # nocov
      }
    } else {
      # Go down the left then right side of the tree
      if (length(x) == 1)
        stop("call or formula with length 1 found without finding the operator, unknown how to proceed")
      # First search the left side
      ret <- findOperator(x[[2]], op, side)
      if ((identical(ret, NA) |
           is.null(ret)) &
          length(x) == 3)
        ret <- findOperator(x[[3]], op, side)
    }
  } else {
    # This should not happen-- find the class that the object is
    stop(sprintf("Cannot handle class %s",
         paste(class(x), sep=", ")))
  }
  ret
}

#' Convert a formula representation to the columns for input data
#'
#' @param form the formula (or something coercible into a formula) to extract
#'   into its parts
#' @returns A list of column names for various formula parts
#' @keywords Internal
#' @family Formula parsing
parse_formula_to_cols <- function(form) {
  if (!inherits(form, "formula")) {
    form <- try({stats::as.formula(form)}, silent = TRUE)
  }
  if (!inherits(form, "formula")) {
    stop("form must be a formula or coercable into one")
  }
  rhs_raw <- findOperator(form, "~", "right")
  groups_raw <- findOperator(rhs_raw, "|", "right")
  if (!is.null(groups_raw)) {
    rhs <- findOperator(rhs_raw, "|", "left")
  } else {
    rhs <- rhs_raw
  }
  # groups_los_raw becomes the last variable to the left of the slash
  groups_raw_c <- all.vars(groups_raw)
  groups_los_raw <- all.vars(findOperator(groups_raw, "/", "left"))
  if (length(groups_los_raw) > 0) {
    groups <- character()
    groups_los <- groups_raw_c[1:which(groups_raw_c == groups_los_raw)]
    groups_ros <- setdiff(groups_raw_c, groups_los)
  } else {
    groups <- groups_raw_c
    groups_los <- character()
    groups_ros <- character()
  }
  ret <-
    list(
      lhs=setdiff(all.vars(findOperator(form, "~", "left")), "."),
      rhs=setdiff(all.vars(rhs), "."),
      groups=groups,
      groups_left_of_slash=groups_los,
      groups_right_of_slash=groups_ros
    )
  ret
}

test_that("findOperator", {
  # Check the left, right, and both sides of various operators
  f1 <- a~b
  expect_equal(findOperator(f1, "~", "both"), f1)
  expect_equal(findOperator(f1, "~", "left"), as.name("a"))
  expect_equal(findOperator(f1, "~", "right"), as.name("b"))

  f2 <- a~b+c
  expect_equal(findOperator(f2, "+", "both"),
               call("+", as.name("b"), as.name("c")))
  expect_equal(findOperator(f2, "+", "left"),
               as.name("b"))
  expect_equal(findOperator(f2, "+", "right"),
               as.name("c"))

  f3 <- a+b~c
  expect_equal(findOperator(f3, "+", "both"),
               call("+", as.name("a"), as.name("b")))
  expect_equal(findOperator(f3, "+", "left"),
               as.name("a"))
  expect_equal(findOperator(f3, "+", "right"),
               as.name("b"))

  f4 <- ~b
  expect_equal(findOperator(f4, "~", "both"), f4)
  expect_equal(findOperator(f4, "~", "left"), NA)
  expect_equal(findOperator(f4, "~", "right"), as.name("b"))

  # Parentheses are handled unusually-- check that they work
  f5 <- a+b~(c+d)
  expect_equal(findOperator(f5, "+", "left"), as.name("a"))
  expect_equal(findOperator(f5, "+", "right"), as.name("b"))

  f6 <- a~(c+d)
  expect_equal(findOperator(f6, "~", "both"), f6)
  expect_equal(findOperator(f6, "+", "left"), as.name("c"))
  expect_equal(findOperator(f6, "+", "right"), as.name("d"))
  expect_equal(findOperator(f6, "(", "left"), NA)
  expect_equal(findOperator(f6, "(", "right"),
               call("+", as.name("c"), as.name("d")))

  # Grouping is handled correctly
  f7 <- a+b~c+d|e
  expect_equal(findOperator(f7, "~", "both"), f7)
  expect_equal(findOperator(f7, "|", "both"),
               call("|",
                    call("+", as.name("c"),
                         as.name("d")),
                    as.name("e")))

  # Grouping with parentheses is handled correctly (as used in
  # lme4-type formulas)
  f8 <- a+b~c+(d|e)+(f|g/h)
  expect_equal(findOperator(f8, "~", "both"), f8)
  expect_equal(findOperator(f8, "|", "both"),
               call("|",
                    as.name("d"),
                    as.name("e")))
  # Sub-searching.  Note: that order of operations for formula group
  # from right to left first.
  expect_equal(
    findOperator(
      findOperator(
        findOperator(f8, "~", "right"),
        "+", "right"),
      "|", "both"),
    call("|",
         as.name("f"),
         call("/", as.name("g"), as.name("h"))))

  # When things aren't found, we get a NULL back
  f9 <- a~b
  expect_equal(findOperator(f9, "+", "left"), NULL)
  # When getting the left side of a unary operator, we get an NA back
  f10 <- a~-b
  expect_equal(findOperator(f10, "-", "left"), NA)
  f11 <- a~!b
  expect_equal(findOperator(f11, "!", "left"), NA)
  f12 <- a~+b
  expect_equal(findOperator(f12, "+", "left"), NA)
  f13 <- a~(b)
  expect_equal(findOperator(f13, "(", "left"), NA)

  # The side can only be "left", "right", or "both"
  expect_error(findOperator(f1, "~", "neither"))

  # When given an invalid class, return an error
  expect_error(findOperator(list(), side="both"),
               regexp="Cannot handle class list")

  expect_error(
    findOperator(x=a~b(), op="+", side="left"),
    regexp="call or formula with length 1 found without finding the operator, unknown how to proceed"
  )
})

test_that("parse_formula_to_cols", {
  expect_equal(
    parse_formula_to_cols(form = "~foo"),
    parse_formula_to_cols(form = ~foo)
  )
  expect_equal(
    parse_formula_to_cols(form = ~foo),
    list(
      lhs = character(),
      rhs = "foo",
      groups = character(),
      groups_left_of_slash = character(),
      groups_right_of_slash = character()
    )
  )
  expect_equal(
    parse_formula_to_cols(form = .~foo),
    list(
      lhs = character(),
      rhs = "foo",
      groups = character(),
      groups_left_of_slash = character(),
      groups_right_of_slash = character()
    )
  )
  expect_equal(
    parse_formula_to_cols(form = foo~.),
    list(
      lhs = "foo",
      rhs = character(),
      groups = character(),
      groups_left_of_slash = character(),
      groups_right_of_slash = character()
    )
  )
  expect_equal(
    parse_formula_to_cols(form = foo~bar),
    list(
      lhs = "foo",
      rhs = "bar",
      groups = character(),
      groups_left_of_slash = character(),
      groups_right_of_slash = character()
    )
  )
  expect_equal(
    parse_formula_to_cols(form = foo~.|bar),
    list(
      lhs = "foo",
      rhs = character(),
      groups = "bar",
      groups_left_of_slash = character(),
      groups_right_of_slash = character()
    )
  )
  expect_equal(
    parse_formula_to_cols(form = foo~baz|bar),
    list(
      lhs = "foo",
      rhs = "baz",
      groups = "bar",
      groups_left_of_slash = character(),
      groups_right_of_slash = character()
    )
  )
  expect_equal(
    parse_formula_to_cols(form = foo~baz|bar+quk),
    list(
      lhs = "foo",
      rhs = "baz",
      groups = c("bar", "quk"),
      groups_left_of_slash = character(),
      groups_right_of_slash = character()
    )
  )
  expect_equal(
    parse_formula_to_cols(form = foo~baz|bar+quk/zot),
    list(
      lhs = "foo",
      rhs = "baz",
      groups = character(),
      groups_left_of_slash = c("bar", "quk"),
      groups_right_of_slash = "zot"
    )
  )
  expect_equal(
    parse_formula_to_cols(form = foo~baz|bar+quk/zot+yap),
    list(
      lhs = "foo",
      rhs = "baz",
      groups = character(),
      groups_left_of_slash = c("bar", "quk"),
      groups_right_of_slash = c("zot", "yap")
    )
  )
})

test_that("parse_formula_to_cols expected errors", {
  expect_error(
    parse_formula_to_cols(form = "foo"),
    regexp = "form must be a formula or coercable into one"
  )
})

#' Calculate the AUC over an interval with interpolation and/or
#' extrapolation of concentrations for the beginning and end of the
#' interval.
#'
#' @details
#' When `pk.calc.aucint()` needs to extrapolate using `lambda.z` (in other
#' words, using the half-life), it will always extrapolate using the logarithmic
#' trapezoidal rule to align with using a half-life calculation for the
#' extrapolation.
#'
#'
#' @inheritParams pk.calc.auxc
#' @inheritParams assert_intervaltime_single
#' @inheritParams assert_lambdaz
#' @param clast,clast.obs,clast.pred The last concentration above the limit of
#'   quantification; this is used for AUCinf calculations.  If provided as
#'   clast.obs (observed clast value, default), AUCinf is AUCinf,obs. If
#'   provided as clast.pred, AUCinf is AUCinf,pred.
#' @param time.dose,route,duration.dose The time of doses, route of
#'   administration, and duration of dose used with interpolation and
#'   extrapolation of concentration data (see [interp.extrap.conc.dose()]).  If
#'   `NULL`, [interp.extrap.conc()] will be used instead (assuming that no doses
#'   affecting concentrations are in the interval).
#' @param ... Additional arguments passed to `pk.calc.auxc` and
#'   `interp.extrap.conc`
#' @family AUC calculations
#' @seealso [PKNCA.options()], [interp.extrap.conc.dose()]
#' @returns The AUC for an interval of time as a number
#' @export
pk.calc.aucint <- function(conc, time,
                           interval=NULL, start=NULL, end=NULL,
                           clast=pk.calc.clast.obs(conc, time),
                           lambda.z=NA,
                           time.dose=NULL,
                           route="extravascular",
                           duration.dose=0,
                           method=NULL,
                           auc.type="AUClast",
                           conc.blq=NULL,
                           conc.na=NULL,
                           check=TRUE,
                           ...,
                           options=list()) {
  # Check inputs
  method <- PKNCA.choose.option(name="auc.method", value=method, options=options)
  if (check) {
    assert_conc_time(conc, time)
    data <-
      clean.conc.blq(
        conc = conc, time = time,
        conc.blq = conc.blq, conc.na = conc.na, options = options,
        check = FALSE
      )
  } else {
    data <- data.frame(conc, time)
  }
  if (all(data$conc %in% 0)) {
    return(structure(0, exclude = "DO NOT EXCLUDE"))
  }
  interval <- assert_intervaltime_single(interval = interval, start = start, end = end)
  missing_times <-
    if (is.infinite(interval[2])) {
      setdiff(c(interval[1], time.dose), data$time)
    } else {
      setdiff(c(interval, time.dose), data$time)
    }
  # Handle the potential double-calculation (before/after tlast) with AUCinf
  conc_clast <- NULL
  time_clast <- NULL
  if (auc.type %in% "AUCinf") {
    tlast <- pk.calc.tlast(conc=data$conc, time=data$time)
    clast_obs <- pk.calc.clast.obs(conc=data$conc, time=data$time)
    if (is.na(clast) && is.na(lambda.z)) {
      # clast.pred is NA likely because the half-life was not calculable
      return(structure(NA_real_, exclude = "clast.pred is NA because the half-life is NA"))
    } else if (is.na(clast)) {
      stop("Please report a bug. clast is NA and the half-life is not NA") # nocov
    } else if (clast != clast_obs & interval[2] > tlast) {
      # If using clast.pred, we need to doubly calculate at tlast.
      conc_clast <- clast
      time_clast <- tlast
    }
  }
  extrap_times <- numeric()
  if (length(missing_times) > 0) {
    if (is.null(time.dose)) {
      missing_conc <-
        interp.extrap.conc(
          conc = data$conc, time = data$time,
          time.out = missing_times,
          method = method,
          auc.type = auc.type,
          clast = clast,
          lambda.z = lambda.z,
          options = options,
          ...
        )
    } else {
      missing_conc <-
        interp.extrap.conc.dose(
          conc = data$conc, time = data$time,
          time.out = missing_times,
          method = method,
          auc.type = auc.type,
          clast = clast, lambda.z = lambda.z,
          options = options,
          # arguments specific to interp.extrap.conc.dose
          time.dose = time.dose,
          route.dose = route,
          duration.dose = duration.dose,
          out.after = FALSE,
          ...
        )
    }
    new_data <- data.frame(conc=c(data$conc, conc_clast, missing_conc),
                           time=c(data$time, time_clast, missing_times))
    tlast <- pk.calc.tlast(conc = data$conc, time = data$time, check = FALSE)
    extrap_times <- missing_times[missing_times > tlast]
    new_data <- new_data[new_data$time >= interval[1] &
                           new_data$time <= interval[2],]
    new_data <- new_data[order(new_data$time),]
    conc_interp <- new_data$conc
    time_interp <- new_data$time
    if (any(mask_na_conc <- is.na(conc_interp))) {
      missing_times <- time_interp[mask_na_conc]
      warning_message <-
        if (any(is.na(lambda.z))) {
          paste("Some interpolated/extrapolated concentration values are missing",
                "(may be due to interpolating or extrapolating over a dose with lambda.z=NA).",
                "Time points with missing data are: ",
                paste(missing_times, collapse=", "))
        } else {
          paste("Some interpolated/extrapolated concentration values are missing",
                "Time points with missing data are: ",
                paste(missing_times, collapse=", "))
        }
      warning(warning_message)
      return(NA_real_)
    }
  } else {
    mask_time <- data$time >= interval[1] & data$time <= interval[2]
    conc_interp <- data$conc[mask_time]
    time_interp <- data$time[mask_time]
  }
  # AUCinf traces an AUClast curve if the interval is finite (because
  # the interval doesn't go to infinity) while AUCall and AUClast trace
  # their own curves.  Or, they all trace their own curves.
  auc.type_map <-
    if (is.infinite(interval[2])) {
      list(
        AUClast="AUClast",
        AUCall="AUCall",
        AUCinf="AUCinf"
      )[[auc.type]]
    } else {
      list(
        AUClast="AUClast",
        AUCall="AUCall",
        AUCinf="AUClast"
      )[[auc.type]]
    }

  interval_method <-
    choose_interval_method(
      conc = conc_interp,
      time = time_interp,
      tlast = max(time_interp),
      method = method,
      auc.type = auc.type,
      options = options
    )
  if (is.finite(interval[2])) {
    interval_method[length(interval_method)] <- "zero"
  }
  if (length(extrap_times) > 0) {
    interval_method[which(time_interp == extrap_times) - 1] <- "log"
  }
  ret <-
    auc_integrate(
      conc = conc_interp, time = time_interp,
      clast = clast, tlast = tlast, lambda.z = lambda.z,
      interval_method = interval_method,
      fun_linear = aucintegrate_linear,
      fun_log = aucintegrate_log,
      fun_inf = aucintegrate_inf
    )
  ret
}

#' @describeIn pk.calc.aucint Interpolate or extrapolate concentrations for
#'   AUClast
#' @export
pk.calc.aucint.last <- function(conc, time, start=NULL, end=NULL, time.dose, ..., options=list()) {
  if (missing(time.dose))
    time.dose <- NULL
  pk.calc.aucint(conc=conc, time=time,
                 start=start, end=end,
                 options=options,
                 time.dose=time.dose,
                 ...,
                 auc.type="AUClast")
}
#' @describeIn pk.calc.aucint Interpolate or extrapolate concentrations for
#'   AUCall
#' @export
pk.calc.aucint.all <- function(conc, time, start=NULL, end=NULL, time.dose, ..., options=list()) {
  if (missing(time.dose))
    time.dose <- NULL
  pk.calc.aucint(conc=conc, time=time,
                 start=start, end=end,
                 options=options,
                 time.dose=time.dose,
                 ...,
                 auc.type="AUCall")
}
#' @describeIn pk.calc.aucint Interpolate or extrapolate concentrations for
#'   AUCinf.obs
#' @export
pk.calc.aucint.inf.obs <- function(conc, time, start=NULL, end=NULL, time.dose, lambda.z, clast.obs, ..., options=list()) {
  if (missing(time.dose))
    time.dose <- NULL
  pk.calc.aucint(conc=conc, time=time,
                 start=start, end=end,
                 time.dose=time.dose,
                 lambda.z=lambda.z, clast=clast.obs,
                 options=options, ...,
                 auc.type="AUCinf")
}
#' @describeIn pk.calc.aucint Interpolate or extrapolate concentrations for
#'   AUCinf.pred
#' @export
pk.calc.aucint.inf.pred <- function(conc, time, start=NULL, end=NULL, time.dose, lambda.z, clast.pred, ..., options=list()) {
  if (missing(time.dose)) {
    time.dose <- NULL
  }
  pk.calc.aucint(conc=conc, time=time,
                 start=start, end=end,
                 time.dose=time.dose,
                 lambda.z=lambda.z, clast=clast.pred,
                 options=options, ...,
                 auc.type="AUCinf")
}

add.interval.col("aucint.last",
                 FUN="pk.calc.aucint.last",
                 values=c(FALSE, TRUE),
                 unit_type="auc",
                 pretty_name="AUCint (based on AUClast extrapolation)",
                 desc="The area under the concentration time curve in the interval extrapolating from Tlast to infinity with zeros (matching AUClast)",
                 formalsmap=list(conc="conc.group", time="time.group", time.dose=NULL))
PKNCA.set.summary(
  name="aucint.last",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)

add.interval.col("aucint.last.dose",
                 FUN="pk.calc.aucint.last",
                 values=c(FALSE, TRUE),
                 unit_type="auc",
                 pretty_name="AUCint (based on AUClast extrapolation, dose-aware)",
                 desc="The area under the concentration time curve in the interval extrapolating from Tlast to infinity with zeros (matching AUClast) with dose-aware interpolation/extrapolation of concentrations",
                 formalsmap=list(conc="conc.group", time="time.group", time.dose="time.dose.group"))
PKNCA.set.summary(
  name="aucint.last.dose",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)

add.interval.col("aucint.all",
                 FUN="pk.calc.aucint.all",
                 values=c(FALSE, TRUE),
                 unit_type="auc",
                 pretty_name="AUCint (based on AUCall extrapolation)",
                 desc="The area under the concentration time curve in the interval extrapolating from Tlast to infinity with the triangle from Tlast to the next point and zero thereafter (matching AUCall)",
                 formalsmap=list(conc="conc.group", time="time.group", time.dose=NULL))
PKNCA.set.summary(
  name="aucint.all",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)

add.interval.col("aucint.all.dose",
                 FUN="pk.calc.aucint.all",
                 values=c(FALSE, TRUE),
                 unit_type="auc",
                 pretty_name="AUCint (based on AUCall extrapolation, dose-aware)",
                 desc="The area under the concentration time curve in the interval extrapolating from Tlast to infinity with the triangle from Tlast to the next point and zero thereafter (matching AUCall) with dose-aware interpolation/extrapolation of concentrations",
                 formalsmap=list(conc="conc.group", time="time.group", time.dose="time.dose.group"))
PKNCA.set.summary(
  name="aucint.all.dose",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)

add.interval.col("aucint.inf.obs",
                 FUN="pk.calc.aucint.inf.obs",
                 values=c(FALSE, TRUE),
                 unit_type="auc",
                 pretty_name="AUCint (based on AUCinf,obs extrapolation)",
                 desc="The area under the concentration time curve in the interval extrapolating from Tlast to infinity with zeros (matching AUClast)",
                 formalsmap=list(conc="conc.group", time="time.group", time.dose=NULL),
                 depends=c("lambda.z", "clast.obs"))
PKNCA.set.summary(
  name="aucint.inf.obs",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)

add.interval.col("aucint.inf.obs.dose",
                 FUN="pk.calc.aucint.inf.obs",
                 values=c(FALSE, TRUE),
                 unit_type="auc",
                 pretty_name="AUCint (based on AUCinf,obs extrapolation, dose-aware)",
                 desc="The area under the concentration time curve in the interval extrapolating from Tlast to infinity with zeros (matching AUClast) with dose-aware interpolation/extrapolation of concentrations",
                 formalsmap=list(conc="conc.group", time="time.group", time.dose="time.dose.group"),
                 depends=c("lambda.z", "clast.obs"))
PKNCA.set.summary(
  name="aucint.inf.obs.dose",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)

add.interval.col("aucint.inf.pred",
                 FUN="pk.calc.aucint.inf.pred",
                 values=c(FALSE, TRUE),
                 unit_type="auc",
                 pretty_name="AUCint (based on AUCinf,pred extrapolation)",
                 desc="The area under the concentration time curve in the interval extrapolating from Tlast to infinity with the triangle from Tlast to the next point and zero thereafter (matching AUCall)",
                 formalsmap=list(conc="conc.group", time="time.group", time.dose=NULL),
                 depends=c("lambda.z", "clast.pred"))
PKNCA.set.summary(
  name="aucint.inf.pred",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)

add.interval.col("aucint.inf.pred.dose",
                 FUN="pk.calc.aucint.inf.pred",
                 values=c(FALSE, TRUE),
                 unit_type="auc",
                 pretty_name="AUCint (based on AUCinf,pred extrapolation, dose-aware)",
                 desc="The area under the concentration time curve in the interval extrapolating from Tlast to infinity with the triangle from Tlast to the next point and zero thereafter (matching AUCall) with dose-aware interpolation/extrapolation of concentrations",
                 formalsmap=list(conc="conc.group", time="time.group", time.dose="time.dose.group"),
                 depends=c("lambda.z", "clast.pred"))
PKNCA.set.summary(
  name="aucint.inf.pred.dose",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)

test_that("AUCint gives errors appropriately", {
  expect_error(
    pk.calc.aucint(conc = 1, time = 1),
    regexp = "One of `interval` or `start` and `end` must be given"
  )
  expect_error(
    pk.calc.aucint(conc = 1, time = 1, end = 1),
    regexp = "Both `start` and `end` or neither must be given"
  )
  expect_error(
    pk.calc.aucint(conc = 1, time = 1, start = 1),
    regexp = "Both `start` and `end` or neither must be given"
  )
  expect_error(
    pk.calc.aucint(conc = 1, time = 1, start = 1:2, end = 1),
    regexp = "Assertion on 'start' failed: Must have length 1."
  )
  expect_error(
    pk.calc.aucint(conc = 1, time = 1, start = 1, end = 1:2),
    regexp = "Assertion on 'end' failed: Must have length 1."
  )

  expect_error(
    pk.calc.aucint(conc = 1, time = 1, start = "A", end = 1),
    regexp = "Assertion on 'start' failed: Must be of type 'number', not 'character'."
  )
  expect_error(
    pk.calc.aucint(conc = 1, time = 1, start = factor("A"), end = 1),
    regexp = "Assertion on 'start' failed: Must be of type 'number', not 'factor'."
  )
  expect_error(
    pk.calc.aucint(conc = 1, time = 1, start = Inf, end = 1),
    regexp = "Assertion on 'start' failed: Must be finite."
  )

  expect_error(
    pk.calc.aucint(conc = 1, time = 1, end = "A", start = 1),
    regexp = "Assertion on 'end' failed: Must be of type 'number', not 'character'."
  )
  expect_error(
    pk.calc.aucint(conc = 1, time = 1, end = factor("A"), start = 1),
    regexp = "Assertion on 'end' failed: Must be of type 'number', not 'factor'."
  )

  expect_error(
    pk.calc.aucint(conc = 1, time = 1, interval = 1:3),
    regexp = "Assertion on 'interval' failed: Must have length 2, but has length 3."
  )
  expect_error(
    pk.calc.aucint(conc = 1, time = 1, interval = c("A", "B")),
    regexp = "Assertion on 'interval' failed: Must be of type 'numeric', not 'character'."
  )
  expect_error(
    pk.calc.aucint(conc = 1, time = 1, interval = factor(c("A", "B"))),
    regexp = "Assertion on 'interval' failed: Must be of type 'numeric', not 'factor'."
  )
  expect_error(
    pk.calc.aucint(conc = 1, time = 1, interval = c(Inf, 1)),
    regexp = "Assertion on 'interval' failed: Must be sorted."
  )
  expect_error(
    pk.calc.aucint(conc = 1, time = 1, interval = c(-Inf, 1)),
    regexp = "Assertion on 'interval[1]' failed: Must be finite.",
    fixed = TRUE
  )
  expect_error(
    pk.calc.aucint(conc = 1, time = 1, interval = c(1, 1)),
    regexp = "Assertion on 'interval' failed: Contains duplicated values, position 2."
  )
  expect_error(
    pk.calc.aucint(conc = 1, time = 1, interval = c(1, 0)),
    regexp = "Assertion on 'interval' failed: Must be sorted."
  )
})

test_that("AUCint gives the same value when no interpolation/extrapolation is required", {
  concdata <- data.frame(conc = c(8, 4, 2, 1), time = 0:3)
  expect_equal(
    pk.calc.aucint(
      conc = concdata$conc, time = concdata$time,
      interval = c(0, 3)
    ),
    pk.calc.auc(
      conc = concdata$conc, time = concdata$time,
      interval = c(0, 3)
    ),
    info = "No interpolation/extrapolation is equivalent to normal AUC"
  )
  expect_equal(
    pk.calc.aucint(
      conc = concdata$conc, time = concdata$time,
      start = 0, end = 3
    ),
    pk.calc.auc(
      conc = concdata$conc, time = concdata$time,
      interval = c(0, 3)
    ),
    info = "Giving interval and start+end are the same, no interp/extrap (test 1)"
  )
  expect_equal(
    pk.calc.aucint(
      conc = concdata$conc, time = concdata$time,
      start = 0, end = 2
    ),
    pk.calc.auc(
      conc = concdata$conc, time = concdata$time,
      interval = c(0, 2)
    ),
    info = "Giving interval and start+end are the same, no interp/extrap (test 2)"
  )
})

test_that("AUCint gives a warning and NA when it cannot interpolate or extrapolate a value", {
  concdata <- data.frame(conc = c(8, 4, 2, 1), time = 0:3)
  expect_warning(
    over_dose <-
      pk.calc.aucint(
        conc = concdata$conc,
        time = concdata$time,
        interval = c(0, 4),
        time.dose = 1.5,
        lambda.z = NA,
        auc.type = "AUCinf"
      ),
    regexp = "Some interpolated/extrapolated concentration values are missing (may be due to interpolating or extrapolating over a dose with lambda.z=NA). Time points with missing data are:  1.5, 4",
    fixed = TRUE,
    info = "warned when integrating over a dose with lambda.z=NA"
  )
  expect_equal(over_dose, NA_real_,
    info = "When you cannot integrate over a dose, you get NA"
  )

  expect_warning(
    expect_warning(
      before_time <-
        pk.calc.aucint(
          conc = concdata$conc,
          time = concdata$time,
          interval = c(-1, 4),
          time.dose = c(-1.5, -0.5),
          lambda.z = 1,
          auc.type = "AUCinf"
        ),
      regexp = "Some interpolated/extrapolated concentration values are missing Time points with missing data are:  -1, -0.5",
      fixed = TRUE,
      info = "warned when integrating over a dose with lambda.z=NA"
    ),
    regexp = "Cannot interpolate between two doses or after a dose without a concentration after the first dose"
  )
  expect_equal(
    before_time, NA_real_,
    info = "When you cannot interpolate a point, you get NA"
  )
})

test_that("AUCint respects auc.type and does the correct calculations for each AUC type", {
  concdata <- data.frame(conc = c(8, 4, 2, 1), time = 0:3)
  concdata_blq <- data.frame(conc = c(8, 4, 2, 1, 0), time = 0:4)
  expect_equal(
    pk.calc.aucint(
      conc = concdata$conc, time = concdata$time,
      interval = c(0, 3), auc.type = "AUClast"
    ),
    pk.calc.aucint(
      conc = concdata$conc, time = concdata$time,
      interval = c(0, 3)
    ),
    info = "Default AUC type is AUClast"
  )

  expect_equal(
    pk.calc.aucint(
      conc = concdata$conc, time = concdata$time,
      interval = c(0, 4), auc.type = "AUCinf", clast = 1, lambda.z = log(2)
    ),
    pk.calc.auc(
      conc = c(concdata$conc, 0.5), time = c(concdata$time, 4),
      interval = c(0, 4)
    ),
    info = "AUCinf is traced"
  )
  expect_equal(
    pk.calc.aucint(
      conc = concdata$conc, time = concdata$time,
      interval = c(0, 4), auc.type = "AUCinf", clast = 2, lambda.z = log(2)
    ),
    pk.calc.auc(
      conc = concdata$conc, time = concdata$time,
      interval = c(0, 3)
    ) +
      pk.calc.auc(
        conc = c(2, 1), time = c(3, 4),
        interval = c(3, 4)
      ),
    info = "AUCinf is traced with clast respected"
  )
  expect_equal(
    pk.calc.aucint(
      conc = concdata$conc, time = concdata$time,
      interval = c(0, 4), auc.type = "AUCinf", clast = 1, lambda.z = log(2) * 2
    ),
    pk.calc.auc(
      conc = c(concdata$conc, 0.25), time = c(concdata$time, 4),
      interval = c(0, 4)
    ),
    info = "AUCinf is traced with lambda.z respected"
  )
  expect_equal(
    pk.calc.aucint(
      conc = concdata$conc, time = concdata$time,
      interval = c(0, 4), auc.type = "AUCinf", clast = 2, lambda.z = 2 * log(2)
    ),
    pk.calc.auc(
      conc = concdata$conc, time = concdata$time,
      interval = c(0, 3)
    ) +
      pk.calc.auc(
        conc = c(2, 0.5), time = c(3, 4),
        interval = c(3, 4)
      ),
    info = "AUCinf is traced with clast and lambda.z respected"
  )

  expect_equal(
    pk.calc.aucint(
      conc = concdata$conc, time = concdata$time,
      interval = c(0, 3), auc.type = "AUCall"
    ),
    pk.calc.auc(
      conc = concdata$conc, time = concdata$time,
      interval = c(0, 3), auc.type = "AUCall"
    ),
    info = "AUCall is the same as AUClast when no BLQ follow tlast (both AUCall)"
  )
  expect_equal(
    pk.calc.aucint(
      conc = concdata$conc, time = concdata$time,
      interval = c(0, 3), auc.type = "AUCall"
    ),
    pk.calc.auc(
      conc = concdata$conc, time = concdata$time,
      interval = c(0, 3), auc.type = "AUClast"
    ),
    info = "AUCall is the same as AUClast when no BLQ follow tlast (test AUClast)"
  )
  expect_equal(
    pk.calc.aucint(
      conc = concdata_blq$conc, time = concdata_blq$time,
      interval = c(0, 4), auc.type = "AUCall"
    ),
    pk.calc.auc(
      conc = concdata_blq$conc, time = concdata_blq$time,
      interval = c(0, 4), auc.type = "AUCall"
    ),
    info = "AUCall is the same the normal calculation when no interpolation/extrapolation happens"
  )
  expect_equal(
    pk.calc.aucint(
      conc = concdata_blq$conc, time = concdata_blq$time,
      interval = c(0, 3.5), auc.type = "AUCall"
    ),
    pk.calc.auc(
      conc = c(concdata$conc, 0.5), time = c(concdata$time, 3.5),
      interval = c(0, 4), auc.type = "AUClast"
    ),
    info = "AUCall traces correctly"
  )
})

test_that("aucint respects doses", {
  concdata <- data.frame(conc = c(8, 4, 2, 1), time = 0:3)
  concdata_blq <- data.frame(conc = c(8, 4, 2, 1, 0), time = 0:4)
  time.dose_at <- 1
  time.dose_between <- 1.5
  time.dose_after_tlast <- 3.5
  time.dose_after_all <- 4.5
  expect_equal(
    pk.calc.aucint(
      conc = concdata_blq$conc, time = concdata_blq$time,
      interval = c(0, 4), auc.type = "AUCall", time.dose = time.dose_at
    ),
    pk.calc.aucint(
      conc = concdata_blq$conc, time = concdata_blq$time,
      interval = c(0, 4), auc.type = "AUCall"
    ),
    info = "Calculation with dosing at the same time as an observation causes no change."
  )
  expect_equal(
    pk.calc.aucint(
      conc = concdata_blq$conc, time = concdata_blq$time,
      interval = c(0, 4), auc.type = "AUCall", time.dose = time.dose_after_all
    ),
    pk.calc.aucint(
      conc = concdata_blq$conc, time = concdata_blq$time,
      interval = c(0, 4), auc.type = "AUCall"
    ),
    info = "Calculation with dosing at a time after all observations causes no change."
  )
  # expect_warning(no_lambda.z_dose <-
  #                  pk.calc.aucint(conc=concdata_blq$conc, time=concdata_blq$time,
  #                                 interval=c(0, 4), auc.type="AUCall", time.dose=time.dose_between),
  #                regexp="Cannot interpolate/extrapolate to dosing times in pk.calc.aucint without lambda.z",
  #                fixed=TRUE)
  # expect_equal(no_lambda.z_dose, NA_real_,
  #              info="Calculating with dosing times, without observations, and without lambda.z")
  expect_equal(
    pk.calc.aucint(
      conc = concdata_blq$conc, time = concdata_blq$time,
      time.dose = 5,
      interval = c(0, 4), auc.type = "AUCall"
    ),
    pk.calc.aucint(
      conc = concdata_blq$conc, time = concdata_blq$time,
      interval = c(0, 4), auc.type = "AUCall"
    ),
    info = "Calculation with dosing at a time after all observations causes no change."
  )
})

test_that("aucint works with infinite intervals", {
  concdata <- data.frame(conc = c(8, 4, 2, 1), time = 0:3)
  expect_equal(pk.calc.aucint.last(conc = concdata$conc, time = concdata$time, start = 0, end = Inf),
    pk.calc.auc.last(conc = concdata$conc, time = concdata$time),
    info = "Simple AUClast = aucint.last"
  )
  expect_equal(pk.calc.aucint.all(conc = concdata$conc, time = concdata$time, start = 0, end = Inf),
    pk.calc.auc.all(conc = concdata$conc, time = concdata$time),
    info = "Simple AUCall = aucint.all"
  )
  expect_equal(
    pk.calc.aucint.inf.obs(
      conc = concdata$conc, time = concdata$time,
      start = 0, end = Inf,
      clast.obs = 1, lambda.z = log(2)
    ),
    pk.calc.auc.inf.obs(
      conc = concdata$conc, time = concdata$time,
      clast.obs = 1, lambda.z = log(2)
    ),
    info = "Simple AUCinf.obs = aucint.inf.obs"
  )
  expect_equal(
    pk.calc.aucint.inf.pred(
      conc = concdata$conc, time = concdata$time,
      start = 0, end = Inf,
      clast.pred = 2, lambda.z = log(2)
    ),
    pk.calc.auc.inf.pred(
      conc = concdata$conc, time = concdata$time,
      clast.pred = 2, lambda.z = log(2)
    ),
    info = "Simple AUCinf.pred = aucint.inf.pred"
  )
})

test_that("aucint respects the check argument", {
  concdata <- data.frame(conc = c(8, 4, 2, 1), time = 0:3)
  expect_equal(
    pk.calc.aucint.last(conc = concdata$conc, time = concdata$time, start = 0, end = Inf),
    pk.calc.aucint.last(conc = concdata$conc, time = concdata$time, start = 0, end = Inf, check = FALSE)
  )
  baddata <- data.frame(
    conc = c(8, 4, 2, NA),
    time = c(0:2, NA)
  )
  expect_error(
    pk.calc.aucint.last(conc = baddata$conc, time = baddata$time, start = 0, end = Inf, check = FALSE)
  )
})

test_that("aucint works for all zero concentrations with interpolated or extrapolated concentrations", {
  expect_equal(
    pk.calc.aucint(conc = c(0, 0, 0, 0), time = 0:3, interval = c(0, 4)),
    structure(0, exclude = "DO NOT EXCLUDE")
  )
  expect_equal(
    pk.calc.aucint(conc = c(0, 0, 0, 0), time = 0:3, interval = c(0, 2.5)),
    structure(0, exclude = "DO NOT EXCLUDE")
  )
})

test_that("aucint uses log extrapolation regardless of the interpolation method (#203)", {
  d_conc <-
    data.frame(
      conc = c(0, 1, 2, 0.75, 0.5, 0.2),
      time = c(0, 1, 2, 3, 4, 5)
    )
  lambda_z <- 0.661
  clast <- d_conc$conc[nrow(d_conc)]
  ctau_extrap <- clast*exp(-lambda_z*(6-5))

  aucinf_obs5_lin <-
    pk.calc.aucint.inf.obs(
      conc = d_conc$conc,
      time = d_conc$time,
      start = 0, end = 5,
      clast.obs = clast,
      lambda.z = lambda_z,
      options = list(auc.method="linear")
    )
  aucinf_obs6_lin <-
    pk.calc.aucint.inf.obs(
      conc = d_conc$conc,
      time = d_conc$time,
      start = 0, end = 6,
      clast.obs = clast,
      lambda.z = lambda_z,
      options = list(auc.method="linear")
    )
  aucinf_obs5_log <-
    pk.calc.aucint.inf.obs(
      conc = d_conc$conc,
      time = d_conc$time,
      start = 0, end = 5,
      clast.obs = clast,
      lambda.z = lambda_z,
      options = list(auc.method="lin up/log down")
    )
  aucinf_obs6_log <-
    pk.calc.aucint.inf.obs(
      conc = d_conc$conc,
      time = d_conc$time,
      start = 0, end = 6,
      clast.obs = clast,
      lambda.z = lambda_z,
      options = list(auc.method="lin up/log down")
    )
  # These are two ways of saying the same thing, the first is simpler logically,
  # the second is more directly mathematical.
  expect_equal(
    aucinf_obs6_lin - aucinf_obs5_lin,
    aucinf_obs6_log - aucinf_obs5_log
  )
  expect_equal(
    aucinf_obs6_lin,
    aucinf_obs5_lin + (6-5)*(clast-ctau_extrap)/log(clast/ctau_extrap)
  )
})

test_that("aucint.inf.pred returns NA when half-life is not estimable (#450)", {
  d_conc <-
    data.frame(
      time = c(0.00, 1.08, 2.08, 4.08, 6.08, 8.08, 13.90),
      conc = c(0.4, 0.8, 1.6, 1.9, 1.7, 1.2, 0.3)
    )
  intervals <- data.frame(start = 0, end = 12, aucint.inf.pred = TRUE)
  o_conc <- PKNCAconc(d_conc, conc ~ time)
  o_data <- PKNCAdata(o_conc, intervals = intervals)
  expect_warning(
    o_nca <- suppressMessages(pk.nca(o_data)),
    regexp = "Too few points for half-life calculation"
  )
  aucint_inf_pred_prep <- as.data.frame(o_nca)
  aucint_inf_pred <- aucint_inf_pred_prep$PPORRES[aucint_inf_pred_prep$PPTESTCD %in% "aucint.inf.pred"]
  expect_equal(aucint_inf_pred, NA_real_)
})

#' Generate a PKNCAresults object
#'
#' This function should not be run directly.  The object is created for
#' summarization.
#'
#' @param result a data frame with NCA calculation results and groups. Each row
#'   is one interval and each column is a group name or the name of an NCA
#'   parameter.
#' @param data The PKNCAdata used to generate the result
#' @param exclude (optional) The name of a column with concentrations to exclude
#'   from calculations and summarization.  If given, the column should have
#'   values of `NA` or `""` for concentrations to include and non-empty text for
#'   concentrations to exclude.
#' @returns A PKNCAresults object with each of the above within.
#' @family PKNCA objects
#' @export
PKNCAresults <- function(result, data, exclude = NULL) {
  result <-
    pknca_unit_conversion(
      result = result,
      units = data$units,
      allow_partial_missing_units = data$options$allow_partial_missing_units
    )
  # Add all the parts into the object
  ret <- list(result=result,
              data=data)
  ret <- setExcludeColumn(ret, exclude = exclude, dataname = "result")
  class(ret) <- c("PKNCAresults", class(ret))
  addProvenance(ret)
}

#' Extract the parameter results from a PKNCAresults and return them as a
#' data.frame.
#'
#' @param x The object to extract results from
#' @param ... Ignored (for compatibility with generic [as.data.frame()])
#' @param out_format Should the output be 'long' (default) or 'wide'?
#' @param filter_requested Only return rows with parameters that were
#'   specifically requested?
#' @param filter_excluded Should excluded values be removed?
#' @param out.format Deprecated in favor of `out_format`
#' @returns A data.frame (or usually a tibble) of results
#' @export
as.data.frame.PKNCAresults <- function(x, ..., out_format = c('long', 'wide'), filter_requested = FALSE, filter_excluded = FALSE, out.format = deprecated()) {
  if (!filter_excluded) {
    ret <- x$result
  } else {
    ret <- summarize_PKNCAresults_clean_exclude(x)
    ret <- ret[is.na(ret[[x$columns$exclude]]), ]
  }
  # nocov start
  if (lifecycle::is_present(out.format)) {
    lifecycle::deprecate_warn(
      when = "0.11.0",
      what = "PKNCA::as.data.frame.PKNCAresults(out.format = )",
      with = "PKNCA::as.data.frame.PKNCAresults(out_format = )"
    )
    out_format <- out.format
  }
  # nocov end
  out_format <- match.arg(out_format)

  if (filter_requested) {
    intervals_long <-
      tidyr::pivot_longer(
        x$data$intervals,
        cols = setdiff(names(get.interval.cols()), c("start", "end")),
        names_to = "PPTESTCD",
        values_to = "keep_interval"
      )
    intervals_long_filtered <- intervals_long[intervals_long$keep_interval, , drop = FALSE]
    intervals_long_filtered$keep_interval <- NULL
    ret <-
      dplyr::inner_join(
        ret, intervals_long_filtered,
        by = intersect(names(ret), names(intervals_long_filtered))
      )
  }

  if (out_format %in% 'wide') {
    if ("PPSTRESU" %in% names(ret)) {
      # Use standardized results
      ret$PPTESTCD <- sprintf("%s (%s)", ret$PPTESTCD, ret$PPSTRESU)
      ret$PPORRES <- ret$PPSTRES
    } else if ("PPORRESU" %in% names(ret)) {
      # Use original results
      ret$PPTESTCD <- sprintf("%s (%s)", ret$PPTESTCD, ret$PPORRESU)
    }
    # Since we moved the results into PPTESTCD and PPORRES regardless of what
    # they really are in the source data, remove the extra units and unit
    # conversion columns to allow spread to work.
    ret <- ret[, setdiff(names(ret), c("PPSTRES", "PPSTRESU", "PPORRESU"))]
    ret <- tidyr::spread(ret, key="PPTESTCD", value="PPORRES")
  }
  ret
}

#' @rdname getDataName
#' @export
getDataName.PKNCAresults <- function(object) {
  "result"
}

#' @rdname is_sparse_pk
#' @export
is_sparse_pk.PKNCAresults <- function(object) {
  is_sparse_pk(object$data)
}

#' @rdname getGroups.PKNCAconc
#' @export
getGroups.PKNCAresults <- function(object,
                                   form=formula(object$data$conc), level,
                                   data=object$result, sep) {
  # Include the start time as a group; this may be dropped later
  grpnames <- c(unlist(object$data$conc$columns$groups), "start", "end")
  if (is_sparse_pk(object)) {
    grpnames <- setdiff(grpnames, object$data$conc$columns$subject)
  }
  if (!missing(level))
    if (is.factor(level) | is.character(level)) {
      level <- as.character(level)
      if (any(!(level %in% grpnames)))
        stop("Not all levels are listed in the group names.  Missing levels are: ",
             paste(setdiff(level, grpnames), collapse=", "))
      grpnames <- level
    } else if (is.numeric(level)) {
      if (length(level) == 1) {
        grpnames <- grpnames[1:level]
      } else {
        grpnames <- grpnames[level]
      }
    }
  data[, grpnames, drop=FALSE]
}

#' @describeIn group_vars.PKNCAconc Get group_vars for a PKNCAresults object
#'   from the PKNCAconc object within
#' @exportS3Method dplyr::group_vars
group_vars.PKNCAresults <- function(x) {
  group_vars.PKNCAconc(as_PKNCAconc(x))
}

#' Summarize PKNCA results
#'
#' @details Excluded results will not be included in the summary.
#'
#' @param object The results to summarize
#' @param drop_group Which group(s) should be dropped from the formula?
#' @param drop_param Which parameters should be excluded from the summary?
#' @param not_requested A character string to use when a parameter summary was
#'   not requested for a parameter within an interval.
#' @param not_calculated A character string to use when a parameter summary was
#'   requested, but the point estimate AND spread calculations (if applicable)
#'   returned `NA`.
#' @param summarize_n Should a column for `N` be added (`TRUE` or `FALSE`)?
#'   `NA` means to automatically detect adding `N` if the data has a subject
#'   column indicated.  Note that `N` is maximum number of parameter results for
#'   any parameter; if no parameters are requested for a group, then `N` will be
#'   `NA`.
#' @param pretty_names Should pretty names (easier to understand in a report) be
#'   used?  `TRUE` is yes, `FALSE` is no, and `NULL` is yes if units are used
#'   and no if units are not used.
#' @param ... Ignored.
#' @param drop.group,summarize.n.per.group,not.requested.string,not.calculated.string
#' Deprecated use `drop_group`, `not_requested`, `not_calculated`, or
#' `summarize_n`, instead
#' @returns A data frame of NCA parameter results summarized according to the
#'   summarization settings.
#' @seealso [PKNCA.set.summary()], [print.summary_PKNCAresults()]
#' @examples
#' conc_obj <- PKNCAconc(as.data.frame(datasets::Theoph), conc ~ Time | Subject)
#' d_dose <-
#'   unique(datasets::Theoph[
#'     datasets::Theoph$Time == 0,
#'     c("Dose", "Time", "Subject")
#'   ])
#' dose_obj <- PKNCAdose(d_dose, Dose ~ Time | Subject)
#' data_obj_automatic <- PKNCAdata(conc_obj, dose_obj)
#' results_obj_automatic <- pk.nca(data_obj_automatic)
#' # To get standard results run summary
#' summary(results_obj_automatic)
#' # To enable numeric conversion and extraction, do not give a spread function
#' # and subsequently run as.numeric on the result columns.
#' PKNCA.set.summary(
#'   name = c("auclast", "cmax", "half.life", "aucinf.obs"),
#'   point = business.geomean,
#'   description = "geometric mean"
#' )
#' PKNCA.set.summary(
#'   name = c("tmax"),
#'   point = business.median,
#'   description = "median"
#' )
#' summary(results_obj_automatic, not_requested = "NA")
#' @export
#' @importFrom lifecycle deprecated
summary.PKNCAresults <- function(object, ...,
                                 drop_group = object$data$conc$columns$subject,
                                 drop_param = character(),
                                 summarize_n = NA,
                                 not_requested = ".",
                                 not_calculated = "NC",

                                 drop.group = deprecated(),
                                 summarize.n.per.group = deprecated(),
                                 not.requested.string = deprecated(),
                                 not.calculated.string = deprecated(),
                                 pretty_names = NULL) {
  # Process inputs ####

  ## Deprecated inputs ####
  # nocov start
  if (lifecycle::is_present(drop.group)) {
    lifecycle::deprecate_warn(
      when = "0.11.0",
      what = "PKNCA::summary.PKNCAresults(drop.group = )",
      with = "PKNCA::summary.PKNCAresults(drop_group = )"
    )
    drop_group <- drop.group
  }
  if (lifecycle::is_present(summarize.n.per.group)) {
    lifecycle::deprecate_warn(
      when = "0.11.0",
      what = "PKNCA::summary.PKNCAresults(summarize.n.per.group = )",
      with = "PKNCA::summary.PKNCAresults(summarize_n = )"
    )
    summarize_n <- summarize.n.per.group
  }
  if (lifecycle::is_present(not.requested.string)) {
    lifecycle::deprecate_warn(
      when = "0.11.0",
      what = "PKNCA::summary.PKNCAresults(not.requested.string = )",
      with = "PKNCA::summary.PKNCAresults(not_requested = )"
    )
    not_requested <- not.requested.string
  }
  if (lifecycle::is_present(not.calculated.string)) {
    lifecycle::deprecate_warn(
      when = "0.11.0",
      what = "PKNCA::summary.PKNCAresults(not.calculated.string = )",
      with = "PKNCA::summary.PKNCAresults(not_calculated = )"
    )
    not_calculated <- not.calculated.string
  }
  # nocov end

  ## Simple inputs ####
  group_cols <- get_summary_PKNCAresults_drop_group(object = object, drop_group = drop_group)
  subject_col <- object$data$conc$columns$subject
  has_subject_col <- length(subject_col) > 0
  if (is.na(summarize_n)) {
    summarize_n <- has_subject_col
  } else if (summarize_n & !has_subject_col) {
    warning("summarize_n was requested, but no subject column exists")
    summarize_n <- FALSE
  }

  # Preparation ####

  # Set excluded rows to NA, give the cleaned data.frame
  raw_results <- summarize_PKNCAresults_clean_exclude(object)

  # Find any parameters that request any summaries, and exclude ones that are
  # not requested
  parameter_cols <-
    setdiff(
      intersect(
        names(object$data$intervals),
        names(get.interval.cols())
      ),
      c(c("start", "end"), drop_param)
    )

  # Extract columns that have been requested by the user for summary in any
  # intervals
  result_data_cols_list <-
    lapply(
      X = object$data$intervals[, parameter_cols, drop = FALSE],
      FUN = any
    )
  # Then, filter them the the ones that have any "TRUE" values
  result_data_cols_list <- result_data_cols_list[unlist(result_data_cols_list)]

  # Prepare for unit management
  unit_col <- get_summary_PKNCAresults_result_unit_col(object = object)
  unit_list <- get_summary_PKNCAresults_result_unit_list(data = raw_results, unit_col = unit_col)
  if (is.null(pretty_names)) {
    pretty_names <- !is.null(unit_list)
  }

  ## Prepare the result data.frame ####
  # Populate the groups on the left and the results on the right.
  result_values <- as.data.frame(result_data_cols_list)
  # If no other value is filled in, then the default is that it was not
  # requested.
  result_values[, names(result_values)] <- not_requested
  # Rows that will have results
  result_groups <- unique(raw_results[, group_cols, drop = FALSE])
  # result_n will be the same as result_groups if summarize_n is FALSE and have
  # an "N" column added if summarize_n is TRUE.
  result_n <-
    get_summary_PKNCAresults_count_N(
      data = raw_results,
      result_group = result_groups,
      subject_col = subject_col,
      summarize_n = summarize_n,
      not_calculated = not_calculated
    )

  # Calculation ####
  ret <-
    summarize_PKNCAresults_object(
      data = raw_results,
      result_group = result_n,
      subject_col = subject_col,
      result_value_template = result_values,
      result_units = unit_list,
      intervals = object$data$intervals,
      not_calculated = not_calculated
    )

  # Decoration ####
  caption <-
    get_summary_PKNCAresults_caption(
      param_names = names(result_values),
      pretty_names = pretty_names,
      footnote_N = "N" %in% names(ret),
      footnote_n = attr(ret, "footnote_n", exact = TRUE)
    )
  attr(ret, "footnote_n") <- NULL
  ret_pretty <- rename_summary_PKNCAresults(data = ret, unit_list = unit_list, pretty_names = pretty_names)
  as_summary_PKNCAresults(
    ret_pretty,
    caption = caption
  )
}

# A helper function for summary.PKNCAresults to find the groups to drop
get_summary_PKNCAresults_drop_group <- function(object, drop_group) {
  all_group_cols <- getGroups(object)
  if (any(c("start", "end") %in% drop_group)) {
    warning("drop.group including start or end may result in incorrect groupings (such as inaccurate comparison of intervals).  Drop these with care.")
  }
  ret <-
    unique(
      setdiff(
        c("start", "end", object$data$options$keep_interval_cols, names(all_group_cols)),
        drop_group
      )
    )
  ret
}

# Get the column name with the results to use for summarization
get_summary_PKNCAresults_result_number_col <- function(object) {
  if (is.data.frame(object)) {
    data <- object
  } else {
    data <- object$result
  }
  intersect(c("PPSTRES", "PPORRES"), names(data))[1]
}

# Get the column name with the result unitss to use for summarization
get_summary_PKNCAresults_result_unit_col <- function(object) {
  if (is.data.frame(object)) {
    data <- object
  } else {
    data <- object$result
  }
  # This will return NULL if neither column is present
  ret <- intersect(c("PPSTRESU", "PPORRESU"), names(data))[1]
  if (is.na(ret)) {
    ret <- NULL
  }
  ret
}

# Get the list of units for each parameter to use for summarization
get_summary_PKNCAresults_result_unit_list <- function(data, unit_col) {
  unit_list <- NULL
  if (!is.null(unit_col)) {
    all_params <- unique(data$PPTESTCD)
    unit_list <-
      stats::setNames(
        as.list(rep(NA_character_, length(all_params))),
        nm = all_params
      )
    for (nm in names(unit_list)) {
      # Get all of the units for a given parameter that will be summarized
      unit_list[[nm]] <- unique(data[[unit_col]][data$PPTESTCD %in% nm])
    }
  }
  unit_list
}

# Count the number of subjects in each group, always return a data.frame with
# the right number of rows so that it can go into cbind
get_summary_PKNCAresults_count_N <- function(data, result_group, subject_col, summarize_n, not_calculated) {
  # subject_col %in% names(data) handles sparse data where the subject column
  # may be imputed rather than present in the data.  And, adding the `any()`
  # outside that ensures that single-subject data (with no subject column) also
  # returns FALSE.
  if (summarize_n && any(subject_col %in% names(data))) {
    # R CMD Check hack
    N <- NULL
    ret <-
      data %>%
      dplyr::grouped_df(vars = names(result_group)) %>%
      dplyr::summarize(
        N = length(unique(.data[[subject_col]]))
      ) %>%
      dplyr::ungroup()
    # Reorder the return value to be in the same order as the original groups
    key_col <- paste0(max(names(ret)), "X")
    # This could fail if there are no groups.  But also, `summarize_n` should be
    # FALSE in that case.  So, it shouldn't be a problem.
    group_key <- do.call(paste, result_group)
    ret[[key_col]] <-
      factor(
        do.call(paste, ret[, names(result_group)]),
        levels = group_key,
        ordered = TRUE
      )
    ret <- ret[order(ret[[key_col]]), ]
    ret[[key_col]] <- NULL

    ret$N <- as.character(ret$N)
    if (any(is.na(ret$N))) {
      stop("Please report a bug. If N is requested, but it is not provided, then it should be set to not calculated.") # nocov
    }
  } else {
    ret <- result_group
  }
  ret
}

# Provide a clean caption for summarized parameters
get_summary_PKNCAresults_caption <- function(param_names, pretty_names, footnote_N, footnote_n) {
  # Extract the summarization descriptions for the caption
  summary_descriptions <-
    unlist(
      lapply(
        X = PKNCA.set.summary()[param_names],
        FUN = `[[`,
        i = "description"
      )
    )

  if (pretty_names) {
    # Make the caption use pretty names if they're used in the header
    all_intervals <- get.interval.cols()
    for (idx in seq_along(summary_descriptions)) {
      names(summary_descriptions)[idx] <- all_intervals[[names(summary_descriptions)[idx]]]$pretty_name
    }
  }
  # Remove duplicate descriptions and then find all the ones that are duplicates
  simplified_summary_descriptions <- summary_descriptions[!duplicated(summary_descriptions)]
  for (idx in seq_along(simplified_summary_descriptions)) {
    names(simplified_summary_descriptions)[idx] <-
      paste(names(summary_descriptions)[summary_descriptions %in% simplified_summary_descriptions[idx]],
            collapse = ", "
      )
  }
  ret <-
    paste(
      names(simplified_summary_descriptions),
      simplified_summary_descriptions,
      sep = ": ",
      collapse = "; "
    )
  if (footnote_N) {
    ret <- c(ret, "N: number of subjects")
  }
  if (footnote_n) {
    ret <- c(ret, "n: number of measurements included in summary")
  }
  paste(ret, collapse = "; ")
}

#' Clean up the exclusions in the object
#'
#' @param object The PKNCAresults object to clean
#' @returns The results data.frame with exclusions set to `NA`
#' @noRd
summarize_PKNCAresults_clean_exclude <- function(object) {
  result_number_col <- get_summary_PKNCAresults_result_number_col(object)
  exclude_col <- object$columns$exclude
  # Ensure that the exclude_col is NA instead of "" for subsequent processing.
  raw_results <- object$result
  raw_results[[exclude_col]] <- normalize_exclude(raw_results[[exclude_col]])
  raw_results[[result_number_col]][!is.na(raw_results[[exclude_col]])] <- NA
  raw_results
}

# A helper function for summary.PKNCAresults to summarize everything
summarize_PKNCAresults_object <- function(data, result_group, subject_col, result_value_template, result_units, intervals, not_calculated) {
  ret_values_list <- list()
  footnote_n <- FALSE
  for (idx in seq_len(nrow(result_group))) {
    ret_idx <-
      summarize_PKNCAresults_group(
        data = data,
        current_group = result_group[idx,],
        subject_col = subject_col,
        result_value_template = result_value_template,
        result_units = result_units,
        intervals = intervals,
        not_calculated = not_calculated
      )
    ret_values_list <- append(ret_values_list, list(ret_idx))
    if (attr(ret_idx, "footnote_n", exact = TRUE)) {
      footnote_n <- TRUE
    }
  }
  ret <- cbind(result_group, dplyr::bind_rows(ret_values_list))
  attr(ret, "footnote_n") <- footnote_n
  ret
}

# A helper function for summary.PKNCAresults to summarize one group
summarize_PKNCAresults_group <- function(data, current_group, subject_col, result_value_template, result_units, intervals, not_calculated) {
  ret <- result_value_template

  current_data <- dplyr::inner_join(data, current_group, by = intersect(names(data), names(current_group)))
  if (nrow(current_data) == 0) {
    # I don't think that a user can get here
    warning("No results to summarize for result row, please report a bug") # nocov
    return(ret) # nocov
  }
  current_interval <- dplyr::inner_join(intervals, current_group, by = intersect(names(intervals), names(current_group)))
  current_param_prep <-
    vapply(
      X = current_interval[, setdiff(names(get.interval.cols()), c("start", "end"))],
      FUN = any,
      FUN.VALUE = TRUE
    )
  current_param_all <-
    intersect(
      names(current_param_prep[current_param_prep]),
      # This ensures that parameters that were dropped with drop_param
      # previously are not summarized
      names(ret)
    )

  footnote_n <- FALSE
  for (current_param in current_param_all) {
    current_summary <-
      summarize_PKNCAresults_parameter(
        data = current_data,
        subject_col = subject_col,
        parameter = current_param,
        include_units = length(result_units[[current_param]]) > 1,
        not_calculated = not_calculated
      )
    # summarize N, if requested and there is a value for calculation
    if (("N" %in% names(current_group)) && (current_summary != not_calculated)) {
      N_group <- as.integer(current_group$N)
      n_summary <- attr(current_summary, "n", exact = TRUE)
      if (N_group != n_summary) {
        current_summary <- sprintf("%s, n=%d", current_summary, n_summary)
        footnote_n <- TRUE
      }
    }
    # use `as.character()` to ensure that the output stays a data.frame rather
    # than a complicated list.
    ret[[current_param]] <- as.character(current_summary)
  }
  attr(ret, "footnote_n") <- footnote_n
  ret
}

# A helper function for summary.PKNCAresults to summarize one parameter within a
# group
summarize_PKNCAresults_parameter <- function(data, parameter, subject_col, include_units, not_calculated) {
  current_data <- data[data$PPTESTCD %in% parameter, , drop = FALSE]
  number_col <- get_summary_PKNCAresults_result_number_col(data)
  unit_col <- get_summary_PKNCAresults_result_unit_col(data)

  units <- NULL
  if (!is.null(unit_col)) {
    units <- unique(current_data[[unit_col]])
    if (length(units) > 1) {
      stop(
        "Multiple units cannot be summarized together.  For ",
        parameter, ", trying to combine: ",
        paste(units, collapse = ", ")
      )
    }
  }

  if (length(subject_col) == 1) {
    N <- length(unique(current_data[[subject_col]]))
    if (any(duplicated(current_data[[subject_col]]))) {
      warning("Some subjects may have more than one result for ", parameter)
    }
  } else {
    N <- NULL
  }
  n <- sum(!is.na(current_data[[number_col]]))

  current_summary_instructions <- PKNCA.set.summary()[[parameter]]
  if (is.null(current_summary_instructions)) {
    stop("No summary function is set for parameter ", parameter, ".  Please set it with PKNCA.set.summary and report this as a bug in PKNCA.") # nocov
  }

  point <- current_summary_instructions$point(current_data[[number_col]])
  # We could count only the number of measurements included in the point
  # estimate which may differ when zeros are excluded, but that would likely be
  # less clear.  So, we are counting all points with available measurements.
  # point_n <- attr(point, which = "n", exact = TRUE)
  # if (!is.null(point_n)) {
  #   n <- point_n
  # }
  point_txt <- roundingSummarize(point, parameter)
  na_point <- is.na(point)
  result_txt <- point_txt

  spread <- NULL
  spread_txt <- NULL
  na_spread <- TRUE
  if ("spread" %in% names(current_summary_instructions) && n > 1) {
    spread <- current_summary_instructions$spread(current_data[[number_col]])
    na_spread <- all(is.na(spread))
    if (na_spread) {
      # The spread couldn't be calculated, so show that
      spread_txt <- not_calculated
    } else {
      # Round the spread
      spread_txt <- roundingSummarize(spread, parameter)
    }
    # Collapse the spread into a usable form if it is longer than one
    # (e.g. a range or a confidence interval) and put brackets around
    # it.
    spread_txt <- paste0(" [", paste(spread_txt, collapse = ", "), "]")
    result_txt <- paste0(point_txt, spread_txt)
  }

  if (na_point & na_spread) {
    result_txt <- not_calculated
  } else if (include_units) {
    result_txt <- paste(result_txt, units)
  }

  structure(
    result_txt,
    parameter = parameter,
    point = point,
    spread = spread,
    N = N,
    n = n,
    units = units,
    class = "summarize_PKNCAresults_parameter"
  )
}

rename_summary_PKNCAresults <- function(data, unit_list, pretty_names) {
  units_to_use <-
    stats::setNames(rep(NA_character_, ncol(data)), names(data))
  if (!is.null(unit_list)) {
    # add the units to the column header, if applicable
    for (nm in names(unit_list)) {
      if (length(unit_list[[nm]]) == 1) {
        units_to_use[nm] <- unit_list[[nm]]
      }
    }
  }
  pretty_names_to_use <-
    stats::setNames(rep(NA_character_, ncol(data)), names(data))
  if (pretty_names) {
    all_intervals <- get.interval.cols()
    for (nm in names(pretty_names_to_use)) {
      if (!is.null(all_intervals[[nm]]$pretty_name)) {
        pretty_names_to_use[nm] <- all_intervals[[nm]]$pretty_name
      }
    }
  }
  for (idx in seq_len(ncol(data))) {
    current_col <- names(data)[idx]
    first_part <- stats::na.omit(c(pretty_names_to_use[current_col], current_col))[1]
    unit_part <- units_to_use[current_col]
    names(data)[idx] <-
      if (is.na(unit_part)) {
        first_part
      } else {
        sprintf("%s (%s)", first_part, unit_part)
      }
  }
  data
}

as_summary_PKNCAresults <- function(data, caption) {
  structure(
    data,
    caption = caption,
    class = c("summary_PKNCAresults", "data.frame")
  )
}

#' Print the results summary
#' @param x A summary_PKNCAresults object
#' @param ... passed to print.data.frame (`row.names` is always set to `FALSE`)
#' @returns `x` invisibly
#' @seealso [summary.PKNCAresults()]
#' @export
print.summary_PKNCAresults <- function(x, ...) {
  print.data.frame(x, row.names = FALSE, ...)
  cat(paste0("\nCaption: ", attr(x, "caption"), "\n"), fill = TRUE)
  invisible(x)
}

#' During the summarization of PKNCAresults, do the rounding of values
#' based on the instructions given.
#'
#' @param x The values to summarize
#' @param name The NCA parameter name (matching a parameter name in
#'   [PKNCA.set.summary()])
#' @returns A string of the rounded value
#' @export
roundingSummarize <- function(x, name) {
  summary_instructions <- PKNCA.set.summary()
  if (!(name %in% names(summary_instructions))) {
    stop(name, " is not in the summarization instructions from PKNCA.set.summary")
  }
  roundingInstructions <- summary_instructions[[name]]$rounding
  if (is.function(roundingInstructions)) {
    ret <- roundingInstructions(x)
  } else if (is.list(roundingInstructions)) {
    if (length(roundingInstructions) != 1) {
      stop("Cannot interpret rounding instructions for ", name, " (please report this as a bug)") # nocov
    }
    if ("signif" == names(roundingInstructions)) {
      ret <- signifString(x, roundingInstructions$signif)
    } else if ("round" == names(roundingInstructions)) {
      ret <- roundString(x, roundingInstructions$round)
    } else {
      stop("Invalid rounding instruction list name for ", name, " (please report this as a bug)") # nocov
    }
  }
  if (!is.character(ret)) {
    ret <- as.character(ret)
  }
  ret
}

#' Generate a sparse_pk object
#'
#' @inheritParams assert_conc_time
#' @param subject Subject identifiers (may be any class; may not be null)
#' @returns A sparse_pk object which is a list of lists.  The inner lists have
#'   elements named: "time", The time of measurement; "conc", The concentration
#'   measured; "subject", The subject identifiers.  The object will usually be
#'   modified by future functions to add more named elements to the inner list.
#' @family Sparse Methods
#' @export
as_sparse_pk <- function(conc, time, subject) {
  if (is.data.frame(conc) && missing(time) && missing(subject)) {
    time <- conc$time
    subject <- conc$subject
    conc <- conc$conc
  }
  assert_conc_time(conc = conc, time = time, any_missing_conc = FALSE, sorted_time = FALSE)
  checkmate::check_vector(subject, any.missing=FALSE, len=length(conc), null.ok=FALSE)
  unique_times <- sort(unique(time))
  ret <- list()
  for (current_time in unique_times) {
    current_mask <- time %in% current_time
    ret <-
      append(
        ret,
        list(list(
          time=current_time,
          conc=conc[current_mask],
          subject=subject[current_mask]
        ))
      )
  }
  class(ret) <- "sparse_pk"
  ret
}

#' Set or get a sparse_pk object attribute
#'
#' @param sparse_pk A sparse_pk object from [as_sparse_pk()]
#' @param ... Either a character string (to get that value) or a named vector
#'   the same length as `sparse_pk` to set the value.
#' @returns Either the attribute value or an updated `sparse_pk` object
#' @keywords Internal
sparse_pk_attribute <- function(sparse_pk, ...) {
  args <- list(...)
  stopifnot(length(args) == 1)
  if (is.null(names(args))) {
    vapply(X=sparse_pk, FUN="[[", args[[1]], FUN.VALUE = 1)
  } else {
    stopifnot(length(args[[1]]) == length(sparse_pk))
    for (idx in seq_along(sparse_pk)) {
      sparse_pk[[idx]][names(args)[1]] <- args[[1]][idx]
    }
    sparse_pk
  }
}

#' Calculate the weight for sparse AUC calculation with the linear-trapezoidal
#' rule
#'
#' The weight is used as the \eqn{w_i}{w_i} parameter in [pk.calc.sparse_auc()]
#'
#' \deqn{w_i = \frac{\delta_{time,i-1,i} + \delta_{time,i,i+1}}{2}}{w_i = (d_time[i-1,i] + d_time[i,i+1])/2}
#' \deqn{\delta_{time,i,i+1} = t_{i+1} - t_i}{d_time = t_[i+1] - t_i, and zero if i < 1 or i > K}
#'
#' Where:
#'
#' \describe{
#'   \item{\eqn{w_i}{w_i}}{is the weight at time i}
#'   \item{\eqn{\delta_{time,i-1,i}}{d_time[i-1,i]} and \eqn{\delta_{time,i,i+1}}{d_time[i,i+1]}}{are the changes between time i-1 and i or i and i+1 (zero outside of the time range)}
#'   \item{\eqn{t_i}{t_i}}{is the time at time i}
#' }
#'
#' @inheritParams sparse_pk_attribute
#' @returns A numeric vector of weights for sparse AUC calculations the same
#'   length as `sparse_pk`
#' @family Sparse Methods
#' @export
sparse_auc_weight_linear <- function(sparse_pk) {
  times <- vapply(X=sparse_pk, FUN="[[", "time", FUN.VALUE = 1)
  half_diff_times <- diff(times)/2
  weights <- c(0, half_diff_times) + c(half_diff_times, 0)
  sparse_pk_attribute(sparse_pk=sparse_pk, weight=weights)
}

#' Calculate the mean concentration at all time points for use in sparse NCA
#' calculations
#'
#' Choices for the method of calculation (the argument `sparse_mean_method`)
#' are:
#'
#' \describe{
#'   \item{"arithmetic mean"}{Arithmetic mean (ignoring number of BLQ samples)}
#'   \item{"arithmetic mean, <=50% BLQ"}{If >= 50% of the measurements are BLQ, zero.  Otherwise, the arithmetic mean of all samples (including the BLQ as zero).}
#' }
#'
#' @inheritParams sparse_pk_attribute
#' @param sparse_mean_method The method used to calculate the sparse mean (see
#'   details)
#' @returns A vector the same length as `sparse_pk` with the mean concentration
#'   at each of those times.
#' @family Sparse Methods
#' @export
sparse_mean <- function(sparse_pk, sparse_mean_method=c("arithmetic mean, <=50% BLQ", "arithmetic mean")) {
  sparse_mean_method <- match.arg(sparse_mean_method)
  ret <-
    vapply(
      X = sparse_pk,
      FUN = function(current_time) mean(current_time$conc),
      FUN.VALUE = 1
    )
  if (sparse_mean_method == "arithmetic mean, <=50% BLQ") {
    numerator <-
      vapply(
        X=sparse_pk,
        FUN=function(current_time) sum(current_time$conc == 0),
        FUN.VALUE = 1
      )
    denominator <-
      vapply(
        X = sparse_pk,
        FUN = function(current_time) length(current_time$conc),
        FUN.VALUE = 1
      )
    frac_blq <- numerator/denominator
    ret[frac_blq > 0.5] <- 0
  } else if (sparse_mean_method == "arithmetic mean") {
    # do nothing
  } else {
    stop("Invalid sparse_mean_method: ", sparse_mean_method) # nocov
  }
  sparse_pk <- sparse_pk_attribute(sparse_pk, mean=ret)
  sparse_pk <- sparse_pk_attribute(sparse_pk, mean_method=rep(sparse_mean_method, length(ret)))
  sparse_pk
}

#' Calculate the variance for the AUC of sparsely sampled PK
#'
#' Equation 7.vii in Nedelman and Jia, 1998 is used for this calculation:
#'
#' \deqn{var\left(\hat{AUC}\right) = \sum\limits_{i=0}^m\left(\frac{w_i^2 s_i^2}{r_i}\right) + 2\sum\limits_{i<j}\left(\frac{w_i w_j r_{ij} s_{ij}}{r_i r_j}\right)}{var(AUC) = sum_(i=0)^(m) ((w_i^2 * s_i^2)/(r_i) + + 2*sum_(i<j)((w_i * w_j * r_ij * s_ij)/(r_i * r_j))}
#'
#' The degrees of freedom are calculated as described in equation 6 of the same
#' paper.
#'
#' @inheritParams sparse_pk_attribute
#' @references
#' Nedelman JR, Jia X. An extension of Satterthwaite’s approximation applied to
#' pharmacokinetics. Journal of Biopharmaceutical Statistics. 1998;8(2):317-328.
#' doi:10.1080/10543409808835241
#' @export
var_sparse_auc <- function(sparse_pk) {
  covariance <- cov_holder(sparse_pk)
  var_auc <- 0
  weights <- sparse_pk_attribute(sparse_pk, "weight")
  # number of subjects at a given time point
  n <- rep(0, length(sparse_pk))
  df <- 0
  for (idx1 in seq_along(sparse_pk)) {
    n_idx1 <- length(unique(sparse_pk[[idx1]]$subject))
    n[idx1] <- n_idx1
    var_auc <-
      var_auc +
      weights[idx1]^2*covariance[idx1, idx1]/n_idx1
    for (idx2 in seq_len(idx1 - 1)) {
      n_idx2 <- length(unique(sparse_pk[[idx2]]$subject))
      n_both <- length(unique(intersect(sparse_pk[[idx1]]$subject, sparse_pk[[idx2]]$subject)))
      var_auc <-
        var_auc +
        2*weights[idx1]*weights[idx2]*n_both*covariance[idx1, idx2]/(n_idx1*n_idx2)
    }
  }
  # df based on equation 6 of Nedelman and Jia 1998
  # df_e <- sum(diag(covariance))
  # df_v <- 2*sum(diag(covariance %*% covariance))
  # df <- 2*df_e^2/df_v
  # df based on equation 6a of Nedelman et al 1995
  df <-
    sum(weights^2 * diag(covariance)/n)^2 /
    sum(weights^4 * diag(covariance)^2/(n^2*(n-1)))
  if (sum(covariance[lower.tri(covariance)] != 0) > 0) {
    rlang::warn(
      message = "Cannot yet calculate sparse degrees of freedom for multiple samples per subject",
      class = "pknca_sparse_df_multi"
    )
    df <- NA_real_
  }
  attr(var_auc, "df") <- df
  var_auc
}

#' Calculate the covariance for two time points with sparse sampling
#'
#' The calculation follows equation A3 in Holder 2001 (see references below):
#'
#' \deqn{\hat{\sigma}_{ij} = \sum\limits_{k=1}^{r_{ij}}{\frac{\left(x_{ik} - \bar{x}_i\right)\left(x_{jk} - \bar{x}_j\right)}{\left(r_{ij} - 1\right) + \left(1 - \frac{r_{ij}}{r_i}\right)\left(1 - \frac{r_{ij}}{r_j}\right)}}}{sigma_ij = sum_(k=1)^(r_ij)((x_ik-xbar_i)(x_jk-xbar_j)/((r_ij-1)+(1-r_ij/r_i)*(1-r_ij/r_j)))}
#'
#' If \eqn{r_{ij} = 0}{r_ij = 0}, then \eqn{\hat{\sigma}_{ij}}{sigma_ij} is
#' defined as zero (rather than dividing by zero).
#'
#' Where:
#' \describe{
#'   \item{\eqn{\hat{\sigma}_{ij}}{sigma_ij}}{The covariance of times i and j}
#'   \item{\eqn{r_i}{r_i} and \eqn{r_j}{r_j}}{The number of subjects (usually animals) at times i and j, respectively}
#'   \item{\eqn{r_{ij}{r_ij}}}{The number of subjects (usually animals) at both times i and j}
#'   \item{\eqn{x_{ik}}{x_ik} and \eqn{x_{jk}}{x_jk}}{The concentration measured for animal k at times i and j, respectively}
#'   \item{\eqn{\bar{x}_i}{xbar_i} and \eqn{\bar{x}_j}{xbar_j}}{The mean of the concentrations at times i and j, respectively}
#' }
#'
#' The Cauchy-Schwartz inequality is enforced for covariances to keep
#' correlation coefficients between -1 and 1, inclusive, as described in
#' equations 8 and 9 of Nedelman and Jia 1998.
#'
#' @inheritParams sparse_pk_attribute
#' @returns A matrix with one row and one column for each element of
#'   `sparse_pk_attribute`.  The covariances are on the off diagonals, and for
#'   simplicity of use, it also calculates the variance on the diagonal
#'   elements.
#' @keywords Internal
#' @references
#' Holder DJ. Comments on Nedelman and Jia’s Extension of Satterthwaite’s
#' Approximation Applied to Pharmacokinetics. Journal of Biopharmaceutical
#' Statistics. 2001;11(1-2):75-79. doi:10.1081/BIP-100104199
#'
#' Nedelman JR, Jia X. An extension of Satterthwaite’s approximation applied to
#' pharmacokinetics. Journal of Biopharmaceutical Statistics. 1998;8(2):317-328.
#' doi:10.1080/10543409808835241
#' @export
cov_holder <- function(sparse_pk) {
  ret <-
    matrix(
      data=0,
      nrow=length(sparse_pk),
      ncol=length(sparse_pk)
    )

  time_means <- sparse_pk_attribute(sparse_pk, "mean")

  for (idx1 in seq_along(sparse_pk)) {
    # Variance on the diagonal
    ret[idx1, idx1] <- stats::var(sparse_pk[[idx1]]$conc)
    for (idx2 in seq_len(idx1 - 1)) {
      subject_idx1 <- sparse_pk[[idx1]]$subject
      subject_idx2 <- sparse_pk[[idx2]]$subject
      subject_both <- intersect(subject_idx1, subject_idx2)
      if (length(subject_both) > 1) {
        # Holder covariance on the off-diagonals when there is more than one
        # subject in both times
        cov_ij <- 0
        for (current_subject in subject_both) {
          cov_ij <-
            cov_ij +
            (sparse_pk[[idx1]]$conc[sparse_pk[[idx1]]$subject %in% current_subject] - time_means[[idx1]]) *
            (sparse_pk[[idx2]]$conc[sparse_pk[[idx2]]$subject %in% current_subject] - time_means[[idx2]])
        }
        # Apply the common denominator
        cov_ij <-
          cov_ij /
          (
            (length(subject_both) - 1) + (1 - length(subject_both)/length(subject_idx1))*(1 - length(subject_both)/length(subject_idx2))
          )
        # Enforce the Cauchy-Schwartz inequality
        cov_cs <- sqrt(ret[idx1, idx1] * ret[idx2, idx2])
        if (abs(cov_ij) > cov_cs) {
          cov_ij <- sign(cov_ij)*cov_cs
        }
        # The matrix is symmetric
        ret[idx1, idx2] <- ret[idx2, idx1] <- cov_ij
      }
    }
  }
  ret
}

#' Extract the mean concentration-time profile as a data.frame
#'
#' @inheritParams sparse_pk_attribute
#' @return A data.frame with names of "conc" and "time"
#' @keywords Internal
sparse_to_dense_pk <- function(sparse_pk) {
  data.frame(
    conc=sparse_pk_attribute(sparse_pk, "mean"),
    time=sparse_pk_attribute(sparse_pk, "time")
  )
}

#' Calculate AUC and related parameters using sparse NCA methods
#'
#' The AUC is calculated as:
#'
#' \deqn{AUC=\sum\limits_{i} w_i \bar{C}_i}{AUC = sum(w_i * Cbar_i)}
#'
#' Where:
#'
#' \describe{
#'   \item{\eqn{AUC}{AUC}}{is the estimated area under the concentration-time curve}
#'   \item{\eqn{w_i}{w_i}}{is the weight applied to the concentration at time i (related to the time which it affects, see [sparse_auc_weight_linear()])}
#'   \item{\eqn{\bar{C}_i}{Cbar_i}}{is the average concentration at time i}
#' }
#' @inheritParams pk.calc.auc
#' @inheritParams as_sparse_pk
#' @family Sparse Methods
#' @export
pk.calc.sparse_auc <- function(conc, time, subject,
                               method=NULL,
                               auc.type="AUClast",
                               ...,
                               options=list()) {
  sparse_pk <- as_sparse_pk(conc=conc, time=time, subject=subject)
  sparse_pk_wt <- sparse_auc_weight_linear(sparse_pk)
  sparse_pk_mean <- sparse_mean(sparse_pk=sparse_pk_wt, sparse_mean_method="arithmetic mean, <=50% BLQ")
  auc <-
    pk.calc.auc(
      conc=sparse_pk_attribute(sparse_pk_mean, "mean"),
      time=sparse_pk_attribute(sparse_pk_mean, "time"),
      auc.type=auc.type,
      method="linear"
    )
  var_auc <- var_sparse_auc(sparse_pk_mean)
  data.frame(
    sparse_auc=auc,
    # as.numeric() drops the "df" attribute
    sparse_auc_se=sqrt(as.numeric(var_auc)),
    sparse_auc_df=attr(var_auc, "df")
  )
}

#' @describeIn pk.calc.sparse_auc Compute the AUClast for sparse PK
#' @export
pk.calc.sparse_auclast <- function(conc, time, subject, ..., options=list()) {
  if ("auc.type" %in% names(list(...))) {
    rlang::abort(
      message = "auc.type cannot be changed when calling pk.calc.sparse_auclast, please use pk.calc.sparse_auc",
      class = "pknca_sparse_auclast_change_auclast"
    )
  }
  ret <-
    pk.calc.sparse_auc(
      conc=conc, time=time, subject=subject, ...,
      options=options,
      auc.type="AUClast",
      lambda.z=NA
    )
  names(ret)[names(ret) == "sparse_auc"] <- "sparse_auclast"
  ret
}

add.interval.col(
  "sparse_auclast",
  sparse=TRUE,
  FUN="pk.calc.sparse_auclast",
  values=c(FALSE, TRUE),
  unit_type="auc",
  pretty_name="Sparse AUClast",
  desc="For sparse PK sampling, the area under the concentration time curve from the beginning of the interval to the last concentration above the limit of quantification"
)
PKNCA.set.summary(
  name="sparse_auclast",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)

add.interval.col(
  "sparse_auc_se",
  FUN=NA,
  values=c(FALSE, TRUE),
  unit_type="auc",
  pretty_name="Sparse AUClast standard error",
  desc="For sparse PK sampling, the standard error of the area under the concentration time curve from the beginning of the interval to the last concentration above the limit of quantification",
  depends="sparse_auclast"
)
PKNCA.set.summary(
  name="sparse_auc_se",
  description="arithmetic mean and standard deviation",
  point=business.mean,
  spread=business.sd
)

add.interval.col(
  "sparse_auc_df",
  FUN=NA,
  values=c(FALSE, TRUE),
  unit_type="count",
  pretty_name="Sparse AUClast degrees of freedom",
  desc="For sparse PK sampling, the standard error degrees of freedom of the area under the concentration time curve from the beginning of the interval to the last concentration above the limit of quantification",
  depends="sparse_auclast"
)
PKNCA.set.summary(
  name="sparse_auc_df",
  description="arithmetic mean and standard deviation",
  point=business.mean,
  spread=business.sd
)

#' Is a PKNCA object used for sparse PK?
#'
#' @param object The object to see if it includes sparse PK
#' @returns `TRUE` if sparse and `FALSE` if dense (not sparse)
#' @export
is_sparse_pk <- function(object) {
  UseMethod("is_sparse_pk")
}

test_that("sparse_auc", {
  d_sparse <-
    data.frame(
      id = c(1L, 2L, 3L, 1L, 2L, 3L, 1L, 2L, 3L, 4L, 5L, 6L, 4L, 5L, 6L, 7L, 8L, 9L, 7L, 8L, 9L),
      conc = c(0, 0, 0,  1.75, 2.2, 1.58, 4.63, 2.99, 1.52, 3.03, 1.98, 2.22, 3.34, 1.3, 1.22, 3.54, 2.84, 2.55, 0.3, 0.0421, 0.231),
      time = c(0, 0, 0, 1, 1, 1, 6, 6, 6, 2, 2, 2, 10, 10, 10, 4, 4, 4, 24, 24, 24),
      dose = c(100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100)
    )

  # calculated using the PK library with
  # v_batch <- PK::auc(data=d_sparse, method="t", design="batch")
  # v_serial <- PK::auc(data=d_sparse, method="t", design="ssd")
  auclast <- 39.4689 # using linear trapezoidal rule
  auclast_se_batch <- 7.30997787038754 # for a batch design (with multiple measures from the same animal taken into account)
  auclast_df_batch <- 2.74598236184576
  auclast_se_serial <- 6.86584835083522 # for a serial design (with multiple measures from the same animal not taken into account)
  auclast_df_serial <- 2.82631153092225

  expect_warning(
    sparse_batch <- pk.calc.sparse_auc(conc=d_sparse$conc, time=d_sparse$time, subject=d_sparse$id),
    regexp="Cannot yet calculate sparse degrees of freedom for multiple samples per subject"
  )
  expect_equal(sparse_batch$sparse_auc, auclast)
  expect_equal(sparse_batch$sparse_auc_se, auclast_se_batch)
  expect_equal(sparse_batch$sparse_auc_df, NA_real_)

  sparse_serial <- pk.calc.sparse_auc(conc=d_sparse$conc, time=d_sparse$time, subject=seq_len(nrow(d_sparse)))
  expect_equal(sparse_serial$sparse_auc, auclast)
  expect_equal(as.numeric(sparse_serial$sparse_auc_se), auclast_se_serial)
  expect_equal(sparse_serial$sparse_auc_df, auclast_df_serial)
})

test_that("sparse_auclast expected errors", {
  expect_error(
    pk.calc.sparse_auclast(auc.type = "foo"),
    class = "pknca_sparse_auclast_change_auclast"
  )
})

test_that("sparse_auc_df and sparse_auc_se are in the parameter list (#292)", {
  expect_true(
    all(c("sparse_auc_df", "sparse_auc_se") %in% names(get.interval.cols()))
  )
})

test_that("sparse_mean", {
  d_sparse <-
    data.frame(
      id = c(1L, 2L, 3L, 1L, 2L, 3L, 1L, 2L, 3L, 4L, 5L, 6L, 4L, 5L, 6L, 7L, 8L, 9L, 7L, 8L, 9L),
      conc = c(0, 0, 0,  1.75, 2.2, 1.58, 4.63, 2.99, 1.52, 3.03, 1.98, 2.22, 3.34, 1.3, 1.22, 3.54, 2.84, 2.55, 0.3, 0.0421, 0.231),
      time = c(0, 0, 0, 1, 1, 1, 6, 6, 6, 2, 2, 2, 10, 10, 10, 4, 4, 4, 24, 24, 24),
      dose = c(100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100)
    )
  
  sparse_pk <- as_sparse_pk(d_sparse)
  sparse_pk_wt <- sparse_auc_weight_linear(sparse_pk)
  sparse_pk_mean <- sparse_mean(sparse_pk_wt, sparse_mean_method="arithmetic mean")
  
  expect_equal(
    sparse_pk_mean[[7]]$mean_method,
    "arithmetic mean"
  )
})

#' @importFrom nlme getGroups
#' @export
nlme::getGroups

#' Get the dependent variable (left hand side of the formula) from a
#' PKNCA object.
#'
#' @param x The object to extract the formula from
#' @param \dots Unused
#' @returns The vector of the dependent variable from the object.
#' @family PKNCA object extractors
#' @export
getDepVar <- function(x, ...) {
  UseMethod("getDepVar", x)
}

#' Get the independent variable (right hand side of the formula) from
#' a PKNCA object.
#'
#' @param x The object to extract the formula from
#' @param \dots Unused
#' @returns The vector of the independent variable from the object.
#' @family PKNCA object extractors
#' @export
getIndepVar <- function(x, ...) {
  UseMethod("getIndepVar", x)
}

#' Get the value from a column in a data frame if the value is a column
#' there, otherwise, the value should be a scalar or the length of the
#' data.
#'
#' @param data A data.frame or similar object
#' @param value A character string giving the name of a column in the `data`, a
#'   scalar, or a vector the same length as the `data`
#' @param prefix The prefix to use if a column must be added (it will be used as
#'   the full column name if it is not already in the dataset or it will be
#'   prepended to the maximum column name if not.)
#' @returns A list with elements named "data", "name" giving the `data` with a
#'   column named "name" with the value in that column.
getColumnValueOrNot <- function(data, value, prefix="X") {
  col.name <- setdiff(c(prefix, paste(prefix, max(names(data)), sep=".")), names(data))[1]
  if (is.character(value) && length(value) == 1 && (value %in% names(data))) {
    # It was a column from the data.frame
    ret <- list(data=data, name=value)
  } else if (length(value) %in% c(1, nrow(data))) {
    data[[col.name]] <- value
    ret <- list(data=data, name=col.name)
  } else {
    stop("value was not a column name nor was it a scalar or a vector matching the length of the data.")
  }
  ret
}

#' Get the name of the element containing the data for the current
#' object.
#'
#' @param object The object to get the data name from.
#' @family PKNCA object extractors
#' @return A character scalar with the name of the data object (or `NULL` if the
#'   method does not apply).
#' @keywords Internal
getDataName <- function(object) {
  UseMethod("getDataName")
}

#' @describeIn getDataName If no data name exists, returns NULL.
getDataName.default <- function(object) {
  NULL
}

#' Add an attribute to an object where the attribute is added as a name
#' to the names of the object.
#'
#' @param object The object to set the attribute column on.
#' @param attr_name The attribute name to set
#' @param col_or_value If this exists as a column in the data, it is used as the
#'   `col_name`.  If not, this becomes the `default_value`.
#' @param col_name The name of the column within the dataset to use (if missing,
#'   uses `attr_name`)
#' @param default_value The value to fill in the column if the column does not
#'   exist (the column is filled with `NA` if it does not exist and no value is
#'   provided).
#' @param stop_if_default,warn_if_default,message_if_default A character string
#'   to provide as an error, a warning, or a message to the user if the
#'   `default_value` is used.  They are tested in order (if stop, the code
#'   stops; if warning, the message is ignored; and message last).
#' @returns The object with the attribute column added to the data.
#' @seealso [getAttributeColumn()]
setAttributeColumn <- function(object, attr_name, col_or_value, col_name, default_value,
                               stop_if_default, warn_if_default, message_if_default) {
  dataname <- getDataName(object)
  # Check inputs
  if (!is.character(attr_name) | (length(attr_name) != 1)) {
    stop("attr_name must be a character scalar.")
  }
  if (!missing(col_or_value) &
      any(!c(missing(col_name), missing(default_value)))) {
    stop("Cannot provide col_or_value and col_name or default_value")
  }
  # Apply col_or_value to col_name or to default_value
  if (!missing(col_or_value)) {
    if (all(col_or_value %in% names(object[[dataname]]))) {
      col_name <- col_or_value
    } else {
      default_value <- col_or_value
    }
  }
  # Set the column name
  if (missing(col_name)) {
    col_name <- attr_name
    if (attr_name %in% names(object[[dataname]])) {
      rlang::inform(
        message = paste0("Found column named ", attr_name, ", using it for the attribute of the same name."),
        class = paste0("pknca_foundcolumn_", attr_name)
      )
    }
  } else if (!is.character(col_name) | (length(col_name) != 1)) {
    stop("col_name must be a character scalar.")
  }
  # Set the default value
  if (missing(default_value)) {
    if (col_name %in% names(object[[dataname]])) {
      default_value <- object[[dataname]][[col_name]]
    } else {
      default_value <- NA
      # React to using the default value, if requested
      if (!missing(stop_if_default)) {
        stop(stop_if_default)
      } else if (!missing(warn_if_default)) {
        warning(warn_if_default)
      } else if (!missing(message_if_default)) {
        message(message_if_default)
      }
    }
  }
  # Check that the default_value can work
  if (!(length(default_value) %in% c(1, nrow(object[[dataname]])))) {
    stop("default_value must be a scalar or the same length as the rows in the data.")
  }
  object[[dataname]][[col_name]] <- default_value
  # Inform the object that the column exists
  if (!("columns" %in% names(object))) {
    object$columns <- list()
  }
  object$columns[[attr_name]] <- col_name
  object
}

#' Retrieve the value of an attribute column.
#'
#' @param object The object to extract the attribute value from.
#' @param attr_name The name of the attribute to extract
#' @param warn_missing Give a warning if the "attr"ibute or "column" is missing.
#'   Character vector with zero, one, or both of "attr" and "column".
#' @returns The value of the attribute (or `NULL` if the attribute is not set or
#'   the column does not exist)
getAttributeColumn <- function(object, attr_name, warn_missing=c("attr", "column")) {
  if (length(setdiff(warn_missing, c("attr", "column")))) {
    stop("warn_missing must have a valid value or be empty")
  }
  warn_missing <- warn_missing[warn_missing %in% c("attr", "column")]
  columns <- object$columns[[attr_name]]
  dataname <- getDataName(object)
  if (is.null(columns)) {
    if ("attr" %in% warn_missing)
      warning(attr_name, " is not set.")
    NULL
  } else if (length(missing_cols <- setdiff(columns, names(object[[dataname]])))) {
    if ("column" %in% warn_missing)
      warning("Columns ", paste(missing_cols, collapse=", "), " are not present.")
    NULL
  } else {
    object[[dataname]][, columns, drop=FALSE]
  }
}

#' Check for duplicate values in a dataset
#'
#' @param object A PKNCAconc or PKNCAdose object to check for duplicates
#' @param data_type The name of the type of data for error reporting
#' @returns `object` unmodified, or an error
#'
#' @keywords Internal
#' @noRd
duplicate_check <- function(object, data_type) {
  mask_excluded <- !is.na(object$data[[object$columns$exclude]])
  mask_dup <- rep(FALSE, nrow(object$data))
  key_cols <- unique(c(object$columns$time, unlist(object$columns$groups)))
  if (length(key_cols) == 0) {
    # If there are no key columns, then there can only be one data row that is
    # not excluded.
    mask_dup[!mask_excluded] <- duplicated(mask_dup[!mask_excluded])
  } else {
    # In case an excluded row is the first row of the duplicated set, do not
    # report duplication.
    mask_dup[!mask_excluded] <- duplicated(object$data[!mask_excluded, key_cols])
  }
  if (any(mask_dup)) {
    stop(
      "Rows that are not unique per group and time (column names: ",
      paste(key_cols, collapse=", "),
      ") found within ", data_type, " data.  Row numbers: ",
      paste(which(mask_dup), collapse=", ")
    )
  }
  object
}

#' Set units for a PKNCAconc or PKNCAdose object
#'
#' @param object a PKNCAconc or PKNCAdose object
#' @param units_orig unit specification that may be columns or values
#' @param units_pref unit specification that must be values
#' @returns The object with the units columns definitions optionally added an a
#'   "units" value list added.
#' @noRd
pknca_set_units <- function(object, units_orig = list(), units_pref = list()) {
  all_units <-
    list(
      orig = lapply(X = units_orig, FUN = assert_unit, data = object$data),
      pref = lapply(X = units_pref, FUN = assert_unit_value)
    )

  object$units <- list()
  for (col_units in names(all_units$orig)) {
    current_unit_type <- attr(all_units$orig[[col_units]], "unit_type")
    if (is.null(current_unit_type)) {
      # do nothing
    } else if (current_unit_type %in% "column") {
      object <-
        setAttributeColumn(
          object = object,
          attr_name = col_units,
          col_name = all_units$orig[[col_units]]
        )
    } else if (current_unit_type %in% "value") {
      object$units[[col_units]] <- all_units$orig[[col_units]]
    } else {
      stop(paste("Please report a bug. Unit setting for", col_units)) # nocov
    }
  }
  for (pref_units in names(all_units$pref)) {
    current_unit_type <- attr(all_units$pref[[pref_units]], "unit_type")
    if (is.null(current_unit_type)) {
      # do nothing
    } else if (current_unit_type %in% "value") {
      # you can only set preferred units if you set original units
      original_unit_col <- gsub(x = pref_units, pattern = "_pref", replacement = "")
      if (!(original_unit_col %in% c(names(object$columns), names(object$units)))) {
        stop("Preferred units may not be set unless original units are set: ", pref_units)
      }
      object$units[[pref_units]] <- all_units$pref[[pref_units]]
    } else {
      stop(paste("Please report a bug. Preferred unit setting for", pref_units)) # nocov
    }
  }

  object
}

getDataName.list <- function(object) "data"

test_that("getColumnValueorNot", {
  tmpdata <- data.frame(a=1:2, b=3:4)
  expect_equal(getColumnValueOrNot(tmpdata, "a", "d"),
               list(data=tmpdata, name="a"))
  expect_equal(getColumnValueOrNot(tmpdata, "d", "d"),
               list(data=cbind(tmpdata,
                               data.frame(d="d", stringsAsFactors=FALSE)),
                    name="d"))
  expect_error(getColumnValueOrNot(tmpdata, 1:3, "d"),
               regexp="value was not a column name nor was it a scalar or a vector matching the length of the data")
})

test_that("setAttributeColumn", {
  #skip("Issue #226 in testthat prevents these tests from succeeding")
  obj1 <- structure(list(data=data.frame(A=1:3)),
                    class="PKNCAconc")
  # Validation of inputs
  expect_error(setAttributeColumn(object=obj1,
                                  attr_name=c("A", "B")),
               regexp="attr_name must be a character scalar",
               info="attr_name must be a scalar")
  expect_error(setAttributeColumn(object=obj1,
                                  attr_name=1),
               regexp="attr_name must be a character scalar",
               info="attr_name must be a character")
  expect_error(setAttributeColumn(object=obj1,
                                  attr_name="A",
                                  col_name=c("foo", "A"),
                                  default_value=4),
               regexp="col_name must be a character scalar")
  expect_error(setAttributeColumn(object=obj1,
                                  attr_name="A",
                                  col_name=1,
                                  default_value=4),
               regexp="col_name must be a character scalar")
  expect_error(setAttributeColumn(object=obj1,
                                  attr_name="B",
                                  default_value=1:2),
               regexp="default_value must be a scalar or the same length as the rows in the data")
  # Settings
  expect_message(val1 <- setAttributeColumn(object=obj1,
                                            attr_name="A",
                                            default_value=4),
                 regexp="Found column named A, using it for the attribute of the same name")
  expect_equal(val1,
               structure(list(data=data.frame(A=rep(4, 3)),
                              columns=list(A="A")),
                         class="PKNCAconc"),
               info="col_name defaults to attr_name, column values are not automatically replaced")
  
  # Info provided back with *_if_default
  expect_error(setAttributeColumn(object=obj1,
                                  attr_name="foo",
                                  stop_if_default="bar"),
               regexp="bar",
               info="Error is triggered if default value is used and stop_if_default")
  expect_warning(setAttributeColumn(object=obj1,
                                    attr_name="foo",
                                    warn_if_default="bar"),
                 regexp="bar",
                 info="Error is triggered if default value is used and stop_if_default")
  expect_message(setAttributeColumn(object=obj1,
                                    attr_name="foo",
                                    message_if_default="bar"),
                 regexp="bar",
                 info="Message is triggered if default value is used and stop_if_default")

  expect_equal(setAttributeColumn(object=obj1,
                                  attr_name="foo",
                                  default_value=4),
               structure(list(data=data.frame(A=1:3,
                                              foo=4),
                              columns=list(foo="foo")),
                         class="PKNCAconc"),
               info="col_name defaults to attr_name, column values are added if the column doesn't exist")
  expect_equal(setAttributeColumn(object=obj1,
                                  attr_name="foo",
                                  col_name="bar",
                                  default_value=4),
               structure(list(data=data.frame(A=1:3,
                                              bar=4),
                              columns=list(foo="bar")),
                         class="PKNCAconc"),
               info="attr_name is set to col_name, column values are added if the column exists")
  expect_equal(setAttributeColumn(object=obj1,
                                  attr_name="foo",
                                  col_name="A",
                                  default_value=4),
               structure(list(data=data.frame(A=rep(4, 3)),
                              columns=list(foo="A")),
                         class="PKNCAconc"),
               info="attr_name is set to col_name, column values are not added if the column exists")
  obj2 <- setAttributeColumn(object=obj1,
                             attr_name="foo",
                             col_name="A",
                             default_value=4)
  expect_equal(setAttributeColumn(object=obj2,
                                  attr_name="bar",
                                  col_name="B",
                                  default_value=5),
               structure(list(data=data.frame(A=rep(4, 3),
                                              B=5),
                              columns=list(foo="A",
                                           bar="B")),
                         class="PKNCAconc"),
               info="Adding a second attribute works")
  expect_equal(setAttributeColumn(object=obj2,
                                  attr_name="foo",
                                  col_name="B",
                                  default_value=5),
               structure(list(data=data.frame(A=rep(4, 3),
                                              B=5),
                              columns=list(foo="B")),
                         class="PKNCAconc"),
               info="Overwriting an attribute works and is non-destructive to the existing data")

  # col_or_value testing
  expect_equal(setAttributeColumn(object=obj1,
                                  attr_name="foo",
                                  col_name="A"),
               setAttributeColumn(object=obj1,
                                  attr_name="foo",
                                  col_or_value="A"),
               info="col_or_value assigns to col_name when present")
  expect_equal(setAttributeColumn(object=obj1,
                                  attr_name="foo",
                                  default_value=5),
               setAttributeColumn(object=obj1,
                                  attr_name="foo",
                                  col_or_value=5),
               info="col_or_value assigns to default_value when not present as a col_name")
  expect_error(setAttributeColumn(object=obj1,
                                  attr_name="foo",
                                  col_or_value=5,
                                  default_value=1),
               regexp="Cannot provide col_or_value and col_name or default_value")
  expect_error(setAttributeColumn(object=obj1,
                                  attr_name="foo",
                                  col_or_value=5,
                                  col_name="B"),
               regexp="Cannot provide col_or_value and col_name or default_value")
})

test_that("getAttributeColumn", {
  #skip("Issue #226 in testthat prevents these tests from succeeding")
  obj1 <- structure(list(data=data.frame(A=1:3,
                                         B=4:6),
                         columns=list(foo="A")),
                    class="PKNCAconc")
  obj2 <- structure(list(data=data.frame(A=1:3,
                                         B=4:6),
                         columns=list(foo="A",
                                      bar=c("A", "B"))),
                    class="PKNCAconc")
  obj3 <- structure(list(data=data.frame(A=1:3),
                         columns=list(foo="C")),
                    class="PKNCAconc")
  expect_equal(getAttributeColumn(object=obj1, attr_name="foo"),
               data.frame(A=1:3),
               info="A data frame (not a vector) is returned")
  expect_equal(getAttributeColumn(object=obj2, attr_name="bar"),
               data.frame(A=1:3,
                          B=4:6),
               info="A data frame (not a vector) is returned, and that data frame may have multiple columns")
  expect_warning(val1 <- getAttributeColumn(object=obj1, attr_name="bar"),
                 regexp="bar is not set",
                 info="The user is warned for missing attribute")
  expect_true(is.null(val1),
              info="missing attributes return NULL")
  expect_warning(val1 <- getAttributeColumn(object=obj3, attr_name="foo"),
                 regexp="Columns C are not present",
                 info="The user is warned for missing column")
  # when warnings aren't requested, they should not be given
  expect_silent(getAttributeColumn(object=obj1, attr_name="bar", warn_missing="column"))
  expect_silent(getAttributeColumn(object=obj3, attr_name="foo", warn_missing="attr"))
  expect_silent(getAttributeColumn(object=obj3, attr_name="foo", warn_missing=character()))
  expect_silent(getAttributeColumn(object=obj3, attr_name="bar", warn_missing=character()))
  expect_error(getAttributeColumn(object=obj3, attr_name="foo", warn_missing="foo"),
               info="warn_missing must have a valid value")
})

test_that("getDataName.default returns NULL", {
  expect_null(getDataName(1:5),
              info="getDataName.default returns NULL (numeric)")
  expect_null(getDataName("a"),
              info="getDataName.default returns NULL (character)")
  expect_null(getDataName(factor("A")),
              info="getDataName.default returns NULL (factor)")
  expect_null(getDataName(TRUE),
              info="getDataName.default returns NULL (logical)")
})

#' Estimate the concentration at dosing time for an IV bolus dose.
#'
#' @inheritParams assert_conc_time
#' @param time.dose The time when dosing occurred
#' @param method The order of methods to test (see details)
#' @param check Check the `conc` and `time` inputs
#' @returns The estimated concentration at time 0.
#'
#' @details Methods available for interpolation are below, and each
#' has its own specific function.
#'
#' \describe{
#'   \item{`c0`}{If the observed `conc` at `time.dose` is nonzero, return that.  This method should usually be used first for single-dose IV bolus data in case nominal time zero is measured.}
#'   \item{`logslope`}{Compute the semilog line between the first two measured times, and use that line to extrapolate backward to `time.dose`}
#'   \item{`c1`}{Use the first point after `time.dose`}
#'   \item{`cmin`}{Set c0 to cmin during the interval.  This method should usually be used for multiple-dose oral data and IV infusion data.}
#'   \item{`set0`}{Set c0 to zero (regardless of any other data).  This method should usually be used first for single-dose oral data.}
#' }
#' @export
pk.calc.c0 <- function(conc, time, time.dose=0,
                       method=c("c0", "logslope", "c1", "cmin", "set0"),
                       check=TRUE) {
  # Check the inputs
  if (check) {
    assert_conc_time(conc = conc, time = time)
  }
  if (length(time.dose) != 1) {
    stop("time.dose must be a scalar")
  } else if (!is.numeric(time.dose) | is.factor(time.dose)) {
    stop("time.dose must be a number")
  }
  if (is.na(time.dose)) {
    warning("time.dose is NA")
    return(structure(NA_real_, exclude = "dose time is missing"))
  } else if (time.dose > max(time)) {
    warning("time.dose is after all available data")
    return(structure(NA_real_, exclude = "dose time is after all available concentration data"))
  }
  method <- match.arg(method, several.ok=TRUE)
  # Find the value
  ret <- NA
  while (is.na(ret) &
         length(method) > 0) {
    current.method <- method[1]
    method <- method[-1]
    ret <- do.call(
      paste("pk.calc.c0.method", current.method, sep="."),
      args=list(
        conc=conc,
        time=time,
        time.dose=time.dose,
        check=FALSE
      )
    )
  }
  ret
}

#' @describeIn pk.calc.c0 Semilog regress the first and second points
#' after time.dose.  This method will return `NA` if the second
#' `conc` after `time.dose` is 0 or greater than the first.
pk.calc.c0.method.logslope <- function(conc, time, time.dose=0,
                                       check=TRUE) {
  if (check) {
    assert_conc_time(conc = conc, time = time)
  }
  mask.positive.time <- (time > time.dose &
                         !(is.na(conc)))
  positive.time <- time[mask.positive.time]
  if (length(positive.time) < 2)
    return(NA)
  # If there is enough data, proceed to calculate
  mask.1 <- time %in% positive.time[1]
  mask.2 <- time %in% positive.time[2]
  c1 <- conc[mask.1]
  c2 <- conc[mask.2]
  t1 <- time[mask.1]
  t2 <- time[mask.2]
  if (c2 < c1 &
      c2 != 0) {
    exp(log(c1) - (log(c2)-log(c1))/(t2-t1)*(t1 - time.dose))
  } else {
    NA
  }
}

#' @describeIn pk.calc.c0 Use `C0` = `conc[time %in% time.dose]` if it is
#'   nonzero.
pk.calc.c0.method.c0 <- function(conc, time, time.dose=0, check=TRUE) {
  if (check) {
    assert_conc_time(conc = conc, time = time)
  }
  # If there is a non-missing and nonzero concentration measurement
  # at time.dose, that's our answer.
  mask.dose <- (time %in% time.dose &
                !(conc %in% c(NA, 0)))
  if (any(mask.dose)) {
    conc[mask.dose]
  } else {
    NA
  }
}

#' @describeIn pk.calc.c0 Use `C0` = `C1`.
pk.calc.c0.method.c1 <- function(conc, time, time.dose=0, check=TRUE) {
  if (check) {
    assert_conc_time(conc = conc, time = time)
  }
  mask.post.dose <- (time > time.dose &
                     !is.na(conc))
  if (any(mask.post.dose)) {
    conc[mask.post.dose][1]
  } else {
    NA
  }
}

#' @describeIn pk.calc.c0 Use `C0` = 0 (typically used for single dose oral and
#'   IV infusion)
pk.calc.c0.method.set0 <- function(conc, time, time.dose=0, check=TRUE) {
  0
}

#' @describeIn pk.calc.c0 Use `C0` = Cmin (typically used for multiple dose oral
#'   and IV infusion but not IV bolus)
pk.calc.c0.method.cmin <- function(conc, time, time.dose=0, check=TRUE) {
  pk.calc.cmin(conc, check=check)
}

# Add the column to the interval specification
add.interval.col("c0",
                 FUN="pk.calc.c0",
                 values=c(FALSE, TRUE),
                 unit_type="conc",
                 pretty_name="C0",
                 desc="Initial concentration after an IV bolus",
                 depends=NULL)
PKNCA.set.summary(
  name="c0",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)

test_that("pk.calc.c0", {
  # Input checks
  expect_error(
    pk.calc.c0(5:1, 4:0),
    regexp="Assertion on 'time' failed: Must be sorted."
  )
  expect_error(pk.calc.c0(5:1, 0:4, time.dose=1:2),
               regexp="time.dose must be a scalar")
  expect_error(pk.calc.c0(5:1, 0:4, time.dose="1"),
               regexp="time.dose must be a number")
  expect_error(pk.calc.c0(5:1, 0:4, method="blah"),
               regexp="should be one of",
               info="method must be valid")
  expect_warning(
    c0_na_time_dose <- pk.calc.c0(conc = 0:3, time = 0:3, time.dose = NA_real_),
    regexp = "time.dose is NA"
  )
  expect_equal(
    c0_na_time_dose,
    structure(NA_real_, exclude = "dose time is missing")
  )
  expect_warning(
    c0_late_time_dose <- pk.calc.c0(5:1, 0:4, time.dose=30),
    regexp="time.dose is after all available data"
  )
  expect_equal(
    c0_late_time_dose,
    structure(NA_real_, exclude = "dose time is after all available concentration data")
  )

  # Simple calculations
  expect_equal(pk.calc.c0(1:5, 0:4), 1,
               info="When there is a nonzero concentration at the default time of dosing")
  expect_equal(pk.calc.c0(1:5, 0:4, time.dose=1), 2,
               info="When there is a nonzero concentration at the nondefault time of dosing")
  # It will fall through to subsequent methods if applicable
  expect_equal(pk.calc.c0(c(0, 2, 1, 0.5), 0:3),
               pk.calc.c0.method.logslope(c(0, 2, 1, 0.5), 0:3),
               info="Falls through to second method")
  expect_equal(pk.calc.c0(c(0, 2, 0, 0), 0:3),
               pk.calc.c0.method.c1(c(0, 2, 0, 0), 0:3),
               info="Falls through to third method")
  expect_equal(pk.calc.c0(c(0, 2, 1, 0.5), 0:3, method="c1"),
               pk.calc.c0.method.c1(c(0, 2, 1, 0.5), 0:3),
               info="Respects method order")

})

test_that("pk.calc.c0.method.logslope", {
  expect_equal(pk.calc.c0.method.logslope(c(0, 2, 1, 0.5), 0:3), 4)
  expect_equal(pk.calc.c0.method.logslope(c(0, 2, 0, 0), 0:3), NA,
               info="returns NA when C2 is 0")
  expect_equal(pk.calc.c0.method.logslope(c(0, 2, 0, 0.5), 0:3), NA,
               info="returns NA when C2 is 0, even when subsequent are > 0")
  expect_equal(pk.calc.c0.method.logslope(c(0, 2, NA, 0.5), 0:3), 4,
               info="ignores missing concentrations")
  expect_equal(pk.calc.c0.method.logslope(c(0, 2, NA, 0.5), 2:5, time.dose=2), 4,
               info="handles nonzero time.dose")
  expect_equal(pk.calc.c0.method.logslope(c(2, NA, 0.5), 3:5, time.dose=2), 4,
               info="works when time.dose isn't an observed time")
})

test_that("pk.calc.c0.method.c0", {
  expect_equal(pk.calc.c0.method.c0(c(0, 2, 1, 0.5), 0:3), NA)
  expect_equal(pk.calc.c0.method.c0(c(2, NA, 0.5), 3:5, time.dose=3), 2,
               info="works when time.dose matches a nonzero conc")
})

test_that("pk.calc.c0.method.c1", {
  expect_equal(pk.calc.c0.method.c1(c(0, 2, 1, 0.5), 0:3), 2)
  expect_equal(pk.calc.c0.method.c1(c(2, NA, 0.5), 3:5, time.dose=3), 0.5,
               info="works when time.dose matches a nonzero conc and skips over NA")
  expect_equal(pk.calc.c0.method.c1(c(2, NA, 0.5), 3:5, time.dose=30), NA,
               info="returns NA when time.dose is after the last measurement")
  expect_warning(
    v1 <- pk.calc.c0.method.c1(rep(NA, 3), 3:5, time.dose=1),
    regexp="All concentration data are missing"
  )
  expect_equal(v1, NA,
               info="returns NA when all inputs are NA")
})

test_that("pk.calc.c0.method.set0", {
  expect_equal(pk.calc.c0.method.set0(), 0)
  expect_equal(pk.calc.c0.method.set0(1, 1), 0)
})

test_that("pk.calc.c0.method.cmin", {
  expect_equal(pk.calc.c0.method.cmin(c(0, 2, 1, 0.5), 0:3), 0)
  expect_equal(pk.calc.c0.method.cmin(c(2, NA, 0.5), 3:5, time.dose=3), 0.5)
})

#' Assess the AIC for all models in a list of models
#'
#' @param object the list of models
#' @param \dots parameters passed to the underlying AIC function (typically the
#'   parameter k)
#' @param assess.best determine which model is the best (by lowest AIC)
#' @seealso [get.best.model()]
#' @returns a data frame with row names matching the names of the list `x` and
#'   columns for degrees of freedom (`df`) and `AIC`.  If `assess.best` is true,
#'   then there will be another column `isBest`.
#' @noRd
AIC.list <- function(object, ..., assess.best=TRUE) {
  allAICs <-
    lapply(object, FUN=function(subobject, ...) {
      # Return the AIC of the new model relative to the reference model
      if (identical(NA, subobject)) {
        ret <- data.frame(AIC=NA, df=NA, indentation=0)
      } else {
        ret <- stats::AIC(subobject, ...)
        if (is.numeric(ret)) {
          ret <- data.frame(AIC=ret,
                            df=attr(stats::logLik(subobject), "df"),
                            indentation=0)
        } else if (is.data.frame(ret)) {
          if ("indentation" %in% names(ret)) {
            ret$indentation <- ret$indentation + 1
          } else {
            stop("Unknown way to get a data.frame without indentation set.  This is likely a bug.") # nocov
          }
        }
      }
      ret
    })
  retnames <- names(allAICs)
  if (is.null(retnames))
    retnames <- rep("", length(allAICs))
  ret <- data.frame()
  for (i in seq_len(length(allAICs))) {
    tmpAICs <- allAICs[[i]]
    # If the best value has already been established, drop it for assessment
    # later.
    tmpAICs$isBest <- NULL
    # Assign the correct rownames to tmpAICs
    if (!(retnames[i] %in% ""))
      if (nrow(tmpAICs) > 1 |
          !identical(rownames(tmpAICs), as.character(seq_len(nrow(tmpAICs))))) {
        rownames(tmpAICs) <- paste(retnames[i], rownames(tmpAICs))
      } else {
        rownames(tmpAICs) <- retnames[i]
      }
    # Add tmpAICs to the data frame to return
    ret <- rbind(ret, tmpAICs)
  }
  if (assess.best) {
    ret$isBest <- ""
    # The next row prevents warnings about no data when na.rm=TRUE
    if (any(!is.na(ret$AIC)))
      ret$isBest[ret$AIC %in% min(ret$AIC, na.rm=TRUE)] <- "Best Model"
  }
  ret
}

#' Extract the best model from a list of models using the AIC.
#'
#' @param object the list of models
#' @param \dots Parameters passed to AIC.list
#' @returns The model which is assessed as best.  If more than one are equal,
#'   the first is chosen.
#' @export
get.best.model <- function(object, ...) {
  object[stats::AIC(object, ...)$isBest %in% "Best Model"][[1]]
}

#' Get the first model from a list of models
#'
#' @param object the list of (lists of, ...) models
#' @returns The first item in the `object` that is not a list or `NA`.  If `NA`
#'   is passed in or the list (of lists) is all `NA`, then `NA` is returned.
get.first.model <- function(object) {
  ret <- NA
  if (inherits(object, "list")) {
    idx <- 0
    while (identical(NA, ret) & idx < length(object)) {
      idx <- idx + 1
      if (identical(NA, object[[idx]])) {
        # Do nothing
      } else if (inherits(object[[idx]], "list")) {
        ret <- get.first.model(object[[idx]])
      } else {
        # It is neither NA or a list, it's our first usable object;
        # return it.
        ret <- object[[idx]]
      }
    }
  } else {
    ret <- object
  }
  ret
}

test_that("AIC.list", {
  tmpdat <- data.frame(x=c(1, 2, 3), y=c(1, 2.02, 3), z=c(0, 0.01, 0))
  mod1 <- glm(y~x, data=tmpdat)
  mod2 <- glm(y~x+z, data=tmpdat)
  d1 <- list(list(mod1), mod2)
  result1 <- data.frame(AIC=c(as.numeric(AIC(mod1)), as.numeric(AIC(mod2))),
                        df=c(3, 4),
                        indentation=c(1, 0),
                        isBest=c("", "Best Model"),
                        stringsAsFactors=FALSE)
  expect_equal(AIC.list(d1), result1)

  # Check that simple names apply correctly
  d2 <- list(list("A"=mod1), "B"=mod2)
  result2 <- data.frame(AIC=c(as.numeric(AIC(mod1)), as.numeric(AIC(mod2))),
                        df=c(3, 4),
                        indentation=c(1, 0),
                        isBest=c("", "Best Model"),
                        stringsAsFactors=FALSE,
                        row.names=c("A", "B"))
  expect_equal(AIC.list(d2), result2)

  d3 <- list("A"=mod1, "B"=mod2)
  result3 <- data.frame(AIC=c(as.numeric(AIC(mod1)), as.numeric(AIC(mod2))),
                        df=c(3, 4),
                        indentation=c(0, 0),
                        isBest=c("", "Best Model"),
                        stringsAsFactors=FALSE,
                        row.names=c("A", "B"))
  expect_equal(AIC.list(d3), result3)
  
  d4 <- list("C"=list("A"=mod1), "B"=mod2)
  result4 <- data.frame(AIC=c(as.numeric(AIC(mod1)), as.numeric(AIC(mod2))),
                        df=c(3, 4),
                        indentation=c(1, 0),
                        isBest=c("", "Best Model"),
                        stringsAsFactors=FALSE,
                        row.names=c("C A", "B"))
  expect_equal(AIC.list(d4), result4)

  d5 <- list("C"=list("A"=mod1), "B"=mod2, C=NA)
  result5 <- data.frame(AIC=c(as.numeric(AIC(mod1)), as.numeric(AIC(mod2)), NA),
                        df=c(3, 4, NA),
                        indentation=c(1, 0, 0),
                        isBest=c("", "Best Model", ""),
                        stringsAsFactors=FALSE,
                        row.names=c("C A", "B", "C"))
  expect_equal(AIC.list(d5), result5)

})

test_that("get.first.model", {
  # If only given NA, return NA.
  expect_equal(get.first.model(NA), NA)
  # If given a list of NAs, return NA
  expect_equal(get.first.model(list(NA)), NA)
  expect_equal(get.first.model(list(NA, NA)), NA)
  # Named lists are treated the same way
  expect_equal(get.first.model(list(A=NA, B=NA)), NA)
  # If given a recursive list of all NAs, return NA
  expect_equal(get.first.model(list(list(NA), list(NA))), NA)
  expect_equal(get.first.model(list(list(NA, NA))), NA)
  expect_equal(get.first.model(list(list(list(NA, NA)))), NA)

  # If given something usable, return that
  expect_equal(get.first.model(1), 1)
  expect_equal(get.first.model("B"), "B")

  # Return the first usable object found
  expect_equal(get.first.model(list("A", "B")), "A")
  # If the first usable object is deep within a list, return that
  expect_equal(get.first.model(list(list(list("A")), "B")), "A")
  # If there are NAs before the first usable object, return the first
  # thing that is not NA.
  expect_equal(get.first.model(list(NA, "A")), "A")
  expect_equal(get.first.model(list(NA, list("A"))), "A")
  expect_equal(get.first.model(list(list(NA, "A"), "B")), "A")

  # Ensure that we do not recurse into items that are not of the list
  # class (but are lists internally)
  tmpdat <- data.frame(x=c(1, 2, 3), y=c(1, 2.02, 3), z=c(0, 0.01, 0))
  mod1 <- glm(y~x, data=tmpdat)
  expect_equal(get.first.model(list(list(NA, mod1), "B")), mod1)
})

test_that("get.best.model", {
  tmpdat <- data.frame(x=c(1, 2, 3), y=c(1, 2.02, 3), z=c(0, 0.01, 0))
  mod1 <- glm(y~x, data=tmpdat)
  mod2 <- glm(y~x+z, data=tmpdat)
  d1 <- list(list(mod1), mod2)
  expect_equal(get.best.model(d1), mod2)
})

#' A compute the Area Under the (Moment) Curve
#'
#' Compute the area under the curve (AUC) and the area under the moment curve
#' (AUMC) for pharmacokinetic (PK) data.  AUC and AUMC are used for many
#' purposes when analyzing PK in drug development.
#'
#' `pk.calc.auc.last` is simply a shortcut setting the `interval`
#' parameter to `c(0, "last")`.
#'
#' Extrapolation beyond Clast occurs using the half-life and Clast,obs;
#' Clast,pred is not yet supported.
#'
#' If all conc input are zero, then the AU(M)C is zero.
#'
#' You probably do not want to call `pk.calc.auxc()`.  Usually, you will call
#' one of the other functions for calculating AUC like `pk.calc.auc.last()`,
#' `pk.calc.auc.inf.obs()`, etc.
#'
#' @inheritParams assert_conc_time
#' @inheritParams assert_intervaltime_single
#' @inheritParams choose_interval_method
#' @inheritParams assert_lambdaz
#' @inheritParams PKNCA.choose.option
#' @param clast,clast.obs,clast.pred The last concentration above the limit of
#'   quantification; this is used for AUCinf calculations.  If provided as
#'   clast.obs (observed clast value, default), AUCinf is AUCinf,obs. If
#'   provided as clast.pred, AUCinf is AUCinf,pred.
#' @param conc.blq How to handle BLQ values in between the first and last above
#'   LOQ concentrations. (See [clean.conc.blq()] for usage instructions.)
#' @param conc.na How to handle missing concentration values.  (See
#'   [clean.conc.na()] for usage instructions.)
#' @param check Run [assert_conc_time()], [clean.conc.blq()], and
#'   [clean.conc.na()]?
#' @param fun_linear The function to use for integration of the linear part of
#'   the curve (not required for AUC or AUMC functions)
#' @param fun_log The function to use for integration of the logarithmic part of
#'   the curve (if log integration is used; not required for AUC or AUMC
#'   functions)
#' @param fun_inf The function to use for extrapolation from the final
#'   measurement to infinite time (not required for AUC or AUMC functions.
#' @param ... For functions other than `pk.calc.auxc`, these values are passed
#'   to `pk.calc.auxc`
#' @returns A numeric value for the AU(M)C.
#' @aliases pk.calc.auc pk.calc.aumc pk.calc.auc.last
#' @family AUC calculations
#' @seealso [clean.conc.blq()]
#' @references
#'
#' Gabrielsson J, Weiner D.  "Section 2.8.1 Computation methods - Linear
#' trapezoidal rule."  Pharmacokinetic & Pharmacodynamic Data Analysis: Concepts
#' and Applications, 4th Edition.  Stockholm, Sweden: Swedish Pharmaceutical
#' Press, 2000.  162-4.
#'
#' Gabrielsson J, Weiner D.  "Section 2.8.3 Computation methods - Log-linear
#' trapezoidal rule."  Pharmacokinetic & Pharmacodynamic Data Analysis: Concepts
#' and Applications, 4th Edition.  Stockholm, Sweden: Swedish Pharmaceutical
#' Press, 2000.  164-7.
#' @examples
#' myconc <- c(0, 1, 2, 1, 0.5, 0.25, 0)
#' mytime <- c(0, 1, 2, 3, 4,   5,    6)
#' pk.calc.auc(myconc, mytime, interval=c(0, 6))
#' pk.calc.auc(myconc, mytime, interval=c(0, Inf))
#' @export
pk.calc.auxc <- function(conc, time, interval=c(0, Inf),
                         clast=pk.calc.clast.obs(conc, time, check=FALSE),
                         lambda.z=NA,
                         auc.type=c("AUClast", "AUCinf", "AUCall"),
                         options=list(),
                         method=NULL,
                         conc.blq=NULL,
                         conc.na=NULL,
                         check=TRUE,
                         fun_linear, fun_log, fun_inf) {
  # Check the inputs
  method <- PKNCA.choose.option(name="auc.method", value=method, options=options)
  if (check) {
    assert_conc_time(conc = conc, time = time)
    data <-
      clean.conc.blq(
        conc = conc, time = time,
        conc.blq = conc.blq, conc.na = conc.na, options = options,
        check = FALSE
      )
  } else {
    data <- data.frame(conc = conc, time = time)
  }
  if (nrow(data) == 0) {
    # All the data were missing
    return(
      structure(
        NA_real_,
        exclude="No data for AUC calculation"
      )
    )
  } else if (nrow(data) == 1) {
    return(
      structure(
        NA_real_,
        exclude="AUC cannot be calculated with only one measured concentration"
      )
    )
  } else if (all(data$conc %in% c(0, NA))) {
    # All the data were missing or 0 before excluding points
    return(structure(0, exclude="DO NOT EXCLUDE"))
  }
  auc.type <- match.arg(auc.type)
  interval <- assert_intervaltime_single(interval = interval)
  if (auc.type %in% "AUCinf" & is.finite(interval[2])) {
    warning("Requesting AUCinf when the end of the interval is not Inf")
  }

  # Subset the data to the range of interest ####
  interval_start <- interval[1]
  interval_end <- interval[2]
  # Find the first time point
  if (interval_start < min(data$time)) {
    warn_message <-
      sprintf(
        "Requesting an AUC range starting (%g) before the first measurement (%g) is not allowed",
        interval_start, min(data$time)
      )
    rlang::warn(message = warn_message, class = "pknca_warn_auc_before_first")
    return(structure(NA_real_, exclude=warn_message))
  } else if (interval_start > max(data$time)) {
    # Give this as a warning, but allow it to continue
    warning(sprintf("AUC start time (%g) is after the maximum observed time (%g)",
                    interval_start, max(data$time)))
  }
  # Ensure that we have clean concentration and time data.  This means that we
  # need to make sure that we have our starting point. Interpolation ensures
  # that (and will give the same answer if it already exists in the right form).
  conc_start <-
    interp.extrap.conc(
      data$conc, data$time,
      time.out = interval_start,
      lambda.z = lambda.z,
      method = method,
      auc.type = auc.type,
      check = FALSE
    )
  # Add that concentration and time to the vectors removing the
  # original version if it was there.
  data <- rbind(data.frame(conc=conc_start,
                           time=interval_start),
                data[!(data$time %in% interval_start), ])
  # * either have our ending point or the ending point is Inf
  if (is.finite(interval_end)) {
    conc_end <-
      interp.extrap.conc(
        data$conc, data$time,
        interval_end,
        method = method,
        auc.type = auc.type,
        check = FALSE
      )
    # !mask.end because we may be replacing an entry that is a 0.
    if (!is.na(conc_end)) {
      data <-
        rbind(
          data[!(data$time %in% interval_end), ],
          data.frame(conc=conc_end, time=interval_end)
        )
    }
  }
  # Subset the conc and time data to the interval of interest
  data <- data[interval_start <= data$time & data$time <= interval_end, , drop=FALSE]
  # Set the overall tlast
  tlast <- pk.calc.tlast(data$conc, data$time, check=FALSE)
  if (all(data$conc %in% 0)) {
    # All the data were missing or 0 after excluding points
    ret <- structure(0, exclude="DO NOT EXCLUDE")
  } else if (is.na(tlast)) {
    # All concentrations are BLQ (note that this has to be checked
    # after full subsetting and interpolation to ensure that it is
    # still true)
    stop("Unknown error with NA tlast but non-BLQ concentrations") # nocov
  } else {
    interval_method <- choose_interval_method(conc = data$conc, time = data$time, tlast = tlast, method = method, auc.type = auc.type, options = options)
    ret <-
      auc_integrate(
        conc = data$conc, time = data$time,
        clast = clast, tlast = tlast, lambda.z = lambda.z,
        interval_method = interval_method,
        fun_linear = fun_linear, fun_log = fun_log, fun_inf = fun_inf
      )
  }
  ret
}

#' @describeIn pk.calc.auxc Compute the area under the curve
#' @export
pk.calc.auc <- function(conc, time, ..., options=list()) {
  pk.calc.auxc(
    conc=conc, time=time, ...,
    options=options,
    fun_linear=aucintegrate_linear,
    fun_log=aucintegrate_log,
    fun_inf=aucintegrate_inf
  )
}

# Note that lambda.z is set to NA for both auc.last and auc.all because all
# interpolation should happen within given points. lambda.z should not be used,
# and if it is used, that should be caught as an error.
#' @describeIn pk.calc.auxc Compute the AUClast.
#' @export
pk.calc.auc.last <- function(conc, time, ..., options=list()) {
  if ("auc.type" %in% names(list(...)))
    stop("auc.type cannot be changed when calling pk.calc.auc.last, please use pk.calc.auc")
  pk.calc.auc(conc=conc, time=time, ...,
              options=options,
              auc.type="AUClast",
              lambda.z=NA)
}

#' @describeIn pk.calc.auxc Compute the AUCinf
#' @export
pk.calc.auc.inf <- function(conc, time, ..., options=list(), lambda.z) {
  if ("auc.type" %in% names(list(...)))
    stop("auc.type cannot be changed when calling pk.calc.auc.inf, please use pk.calc.auc")
  pk.calc.auc(conc=conc, time=time, ...,
              options=options,
              auc.type="AUCinf",
              lambda.z=lambda.z)
}

#' @describeIn pk.calc.auxc Compute the AUCinf with the observed Clast.
#' @export
pk.calc.auc.inf.obs <- function(conc, time, clast.obs, ..., options=list(),
                                lambda.z) {
  pk.calc.auc.inf(conc=conc, time=time, clast=clast.obs, ...,
                  options=options,
                  lambda.z=lambda.z)
}

#' @describeIn pk.calc.auxc Compute the AUCinf with the predicted Clast.
#' @export
pk.calc.auc.inf.pred <- function(conc, time, clast.pred, ..., options=list(),
                                 lambda.z) {
  pk.calc.auc.inf(conc=conc, time=time, clast=clast.pred, ...,
                  options=options,
                  lambda.z=lambda.z)
}

#' @describeIn pk.calc.auxc Compute the AUCall.
#' @export
pk.calc.auc.all <- function(conc, time, ..., options=list()) {
  if ("auc.type" %in% names(list(...)))
    stop("auc.type cannot be changed when calling pk.calc.auc.all, please use pk.calc.auc")
  pk.calc.auc(conc=conc, time=time, ..., options=options,
              auc.type="AUCall",
              lambda.z=NA)
}


#' @describeIn pk.calc.auxc Compute the area under the moment curve
#' @export
pk.calc.aumc <- function(conc, time, ..., options=list()) {
  pk.calc.auxc(conc=conc, time=time, ..., options=options,
    fun_linear = aumcintegrate_linear,
    fun_log = aumcintegrate_log,
    fun_inf = aumcintegrate_inf
  )
}

#' @describeIn pk.calc.auxc Compute the AUMClast.
#' @export
pk.calc.aumc.last <- function(conc, time, ..., options=list()) {
  if ("auc.type" %in% names(list(...)))
    stop("auc.type cannot be changed when calling pk.calc.aumc.last, please use pk.calc.aumc")
  pk.calc.aumc(conc=conc, time=time, ..., options=options,
               auc.type="AUClast",
               lambda.z=NA)
}

#' @describeIn pk.calc.auxc Compute the AUMCinf
#' @export
pk.calc.aumc.inf <- function(conc, time, ..., options=list(),
                             lambda.z) {
  if ("auc.type" %in% names(list(...))) {
    stop("auc.type cannot be changed when calling pk.calc.aumc.inf, please use pk.calc.aumc")
  }
  pk.calc.aumc(conc=conc, time=time, ..., options=options,
               auc.type="AUCinf",
               lambda.z=lambda.z)
}

#' @describeIn pk.calc.auxc Compute the AUMCinf with the observed Clast.
#' @export
pk.calc.aumc.inf.obs <- function(conc, time, clast.obs, ..., options=list(),
                                 lambda.z) {
  pk.calc.aumc.inf(conc=conc, time=time, clast=clast.obs, ..., options=options,
                   lambda.z=lambda.z)
}

#' @describeIn pk.calc.auxc Compute the AUMCinf with the predicted Clast.
#' @export
pk.calc.aumc.inf.pred <- function(conc, time, clast.pred, ..., options=list(),
                             lambda.z) {
  pk.calc.aumc.inf(conc=conc, time=time, clast=clast.pred, ..., options=options,
                   lambda.z=lambda.z)
}

#' @describeIn pk.calc.auxc Compute the AUMCall.
#' @export
pk.calc.aumc.all <- function(conc, time, ..., options=list()) {
  if ("auc.type" %in% names(list(...)))
    stop("auc.type cannot be changed when calling pk.calc.aumc.all, please use pk.calc.aumc")
  pk.calc.aumc(conc=conc, time=time, ..., options=options,
               auc.type="AUCall",
               lambda.z=NA)
}

# Add the columns to the interval specification
add.interval.col("aucinf.obs",
                 FUN="pk.calc.auc.inf.obs",
                 values=c(FALSE, TRUE),
                 unit_type="auc",
                 pretty_name="AUCinf,obs",
                 desc="The area under the concentration time curve from the beginning of the interval to infinity with extrapolation to infinity from the observed Clast",
                 depends=c("lambda.z", "clast.obs"))

add.interval.col("aucinf.pred",
                 FUN="pk.calc.auc.inf.pred",
                 values=c(FALSE, TRUE),
                 unit_type="auc",
                 pretty_name="AUCinf,pred",
                 desc="The area under the concentration time curve from the beginning of the interval to infinity with extrapolation to infinity from the predicted Clast",
                 depends=c("lambda.z", "clast.pred"))

add.interval.col("auclast",
                 FUN="pk.calc.auc.last",
                 values=c(FALSE, TRUE),
                 unit_type="auc",
                 pretty_name="AUClast",
                 desc="The area under the concentration time curve from the beginning of the interval to the last concentration above the limit of quantification")

add.interval.col("aucall",
                 FUN="pk.calc.auc.all",
                 values=c(FALSE, TRUE),
                 unit_type="auc",
                 pretty_name="AUCall",
                 desc="The area under the concentration time curve from the beginning of the interval to the last concentration above the limit of quantification plus the triangle from that last concentration to 0 at the first concentration below the limit of quantification"
)

add.interval.col("aumcinf.obs",
                 FUN="pk.calc.aumc.inf.obs",
                 values=c(FALSE, TRUE),
                 unit_type="aumc",
                 pretty_name="AUMC,inf,obs",
                 desc="The area under the concentration time moment curve from the beginning of the interval to infinity with extrapolation to infinity from the observed Clast",
                 depends=c("lambda.z", "clast.obs"))

add.interval.col("aumcinf.pred",
                 FUN="pk.calc.aumc.inf.pred",
                 values=c(FALSE, TRUE),
                 unit_type="aumc",
                 pretty_name="AUMC,inf,pred",
                 desc="The area under the concentration time moment curve from the beginning of the interval to infinity with extrapolation to infinity from the predicted Clast",
                 depends=c("lambda.z", "clast.pred"))

add.interval.col("aumclast",
                 FUN="pk.calc.aumc.last",
                 values=c(FALSE, TRUE),
                 unit_type="aumc",
                 pretty_name="AUMC,last",
                 desc="The area under the concentration time moment curve from the beginning of the interval to the last concentration above the limit of quantification")

add.interval.col("aumcall",
                 FUN="pk.calc.aumc.all",
                 values=c(FALSE, TRUE),
                 unit_type="aumc",
                 pretty_name="AUMC,all",
                 desc="The area under the concentration time moment curve from the beginning of the interval to the last concentration above the limit of quantification plus the moment of the triangle from that last concentration to 0 at the first concentration below the limit of quantification")

PKNCA.set.summary(
  name=
    c(
      "aucinf.obs", "aucinf.pred", "auclast", "aucall",
      "aumcinf.obs", "aumcinf.pred", "aumclast", "aumcall"
    ),
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)

test_that("pk.calc.auxc", {
  # Verify input checks

  expect_error(
    pk.calc.auxc(conc=1:2, time=0:1, interval=2:1, method="linear"),
    regexp="Assertion on 'interval' failed: Must be sorted."
  )
  expect_warning(pk.calc.auxc(conc=1:2, time=2:3, interval=c(1, 3),
                              method="linear"),
                 regexp="Requesting an AUC range starting \\(1\\) before the first measurement \\(2\\) is not allowed",
                 info="AUC should start at or after the first measurement and should be before the last measurement")
  expect_warning(v1 <- pk.calc.auxc(conc=1:2, time=2:3, interval=c(1, 3),
                                    method="linear"),
                 info="Starting before the beginning time returns NA (not an error)")
  expect_equal(
    v1,
    structure(NA_real_, exclude = 'Requesting an AUC range starting (1) before the first measurement (2) is not allowed'),
    info="Starting before the beginning time returns NA (not an error)"
  )

  # All concentrations are NA, return NA
  expect_warning(
    v2 <- pk.calc.auxc(conc=c(NA, NA), time=2:3, interval=c(1, 3), method="linear"),
    regexp = "All concentration data are missing"
  )
  expect_equal(
    v2,
    structure(NA_real_, exclude="No data for AUC calculation"),
    info="All concentrations NA is NA"
  )
  # All concentrations are 0, return 0
  expect_equal(
    pk.calc.auxc(
      conc=c(0, 0), time=2:3, interval=c(2, 3),
      method="linear"
    ),
    structure(0, exclude="DO NOT EXCLUDE"),
    info="All zeros is zero"
  )
  # Concentrations mix 0 and NA, return 0
  expect_equal(
    pk.calc.auxc(
      conc=c(NA, 0, NA, 0, NA), time=2:6, interval=c(1, 3),
      method="linear"
    ),
    structure(0, exclude="DO NOT EXCLUDE"),
    info="Mixed zeros and NA is still zero."
  )
  # Invalid integration method
  expect_error(pk.calc.auxc(conc=c(NA, 0, NA), time=2:4, interval=c(1, 3),
                            method="foo"),
               info="Invalid integration methods are caught.")
})

test_that("pk.calc.auc: Linear AUC when the conc at the end of the interval is above LOQ", {
  # lambda.z is unused
  tests <- list(AUCinf=as.numeric(NA),
                AUClast=1.5,
                AUCall=1.5)
  for (t in names(tests)) {
    # Note: using this structure ensures that there will not be
    # excessive warnings during testing.
    if (t == "AUCinf") {
      runner <- expect_warning
    } else {
      runner <- identity
    }
    runner(v1 <- pk.calc.auc(conc=c(0, 1, 1),
                             time=0:2,
                             interval=c(0, 2),
                             lambda.z=NA,
                             auc.type=t,
                             method="linear"))
    expect_equal(v1, tests[[t]], info=t)
  }
})

test_that("pk.calc.auc: Linear AUC when the conc at the end of the interval is BLQ, lambda.z missing", {
  # lambda.z is used to extrapolate to the end of the interval.
  # Since lambda.z is NA, the result is NA.
  tests <- list(AUCinf=as.numeric(NA),
                AUClast=0.5,
                AUCall=1)
  for (t in names(tests)) {
    # Note: using this structure ensures that there will not be
    # excessive warnings during testing.
    if (t == "AUCinf") {
      runner <- expect_warning
    } else {
      runner <- identity
    }
    runner(v1 <- pk.calc.auc(conc=c(0, 1, 0),
                             time=0:2,
                             interval=c(0, 2),
                             lambda.z=NA,
                             auc.type=t,
                             method="linear"))
    expect_equal(v1, tests[[t]], info=t)
  }
})

test_that("pk.calc.auc: Linear AUC when the conc at the end of the interval is BLQ, lambda.z given", {
  # The same when lambda.z is given
  tests <- list(AUCinf=1.5,
                AUClast=0.5,
                AUCall=1)
  for (t in names(tests)) {
    # Note: using this structure ensures that there will not be
    # excessive warnings during testing.
    if (t == "AUCinf") {
      runner <- expect_warning
    } else {
      runner <- identity
    }
    runner(v1 <- pk.calc.auc(conc=c(0, 1, 0),
                             time=0:2,
                             interval=c(0, 2),
                             lambda.z=1,
                             auc.type=t,
                             method="linear"))
    expect_equal(v1, tests[[t]], info=t)
  }
})

test_that("pk.calc.auc: Linear AUC when when there are multiple BLQ values at the end, lambda.z given", {
  tests <- list(AUCinf=1.5,
                AUClast=0.5,
                AUCall=1)
  for (t in names(tests)) {
    # Note: using this structure ensures that there will not be
    # excessive warnings during testing.
    if (t == "AUCinf") {
      runner <- expect_warning
    } else {
      runner <- identity
    }
    runner(v1 <- pk.calc.auc(conc=c(0, 1, rep(0, 5)),
                             time=0:6,
                             interval=c(0, 6),
                             lambda.z=1,
                             auc.type=t,
                             method="linear"))
    expect_equal(v1, tests[[t]], info=t)
  }
})

test_that("pk.calc.auc: Confirm that center BLQ points are dropped, kept, or imputed correctly", {
  # Do this with both "linear" and "lin up/log down"
  tests <- list(
    "linear"=list(
      AUCinf=1+1+0.5+1.5+1.5+1,
      AUClast=1+1+0.5+1.5+1.5,
      AUCall=1+1+0.5+1.5+1.5),
    "lin up/log down"=list(
      AUCinf=1+1+0.5+1.5+1/log(2)+1,
      AUClast=1+1+0.5+1.5+1/log(2),
      AUCall=1+1+0.5+1.5+1/log(2)))
  for (t in names(tests)) {
    for (n in names(tests[[t]])) {
      # Note: using this structure ensures that there will not be
      # excessive warnings during testing.
      if (n == "AUCinf") {
        runner <- expect_warning
      } else {
        runner <- identity
      }
      runner(v1 <- pk.calc.auc(conc=c(0, 2, 0, 1, 2, 1),
                               time=0:5,
                               interval=c(0, 5),
                               lambda.z=1,
                               auc.type=n,
                               conc.blq="keep",
                               method=t))
      expect_equal(v1, tests[[t]][[n]], info=paste(t, n))
    }
  }
})

test_that("pk.calc.auc: Confirm BLQ in the middle or end are calculated correctly", {
  # AUCall looks different when there are BLQs at the end
  tests <- list(
    "linear"=list(
      AUCinf=1+1+0.5+1.5+1.5+1,
      AUClast=1+1+0.5+1.5+1.5,
      AUCall=1+1+0.5+1.5+1.5+0.5),
    "lin up/log down"=list(
      AUCinf=1+1+0.5+1.5+1/log(2)+1,
      AUClast=1+1+0.5+1.5+1/log(2),
      AUCall=1+1+0.5+1.5+1/log(2)+0.5))
  for (t in names(tests)) {
    for (n in names(tests[[t]])) {
      # Note: using this structure ensures that there will not be
      # excessive warnings during testing.
      if (n == "AUCinf") {
        runner <- expect_warning
      } else {
        runner <- identity
      }
      runner(v1 <- pk.calc.auc(conc=c(0, 2, 0, 1, 2, 1, 0, 0),
                               time=0:7,
                               interval=c(0, 7),
                               lambda.z=1,
                               auc.type=n,
                               conc.blq="keep",
                               method=t))
      expect_equal(v1, tests[[t]][[n]], info=paste(t, n))
    }
  }

  # When BLQ in the middle and end are dropped, you get different
  # answers.  (Note that first BLQ dropping would cause errors due to
  # starting times differing, so not tested here.)
  tests <- list(
    "linear"=list(
      AUCinf=1+3+1.5+1.5+1,
      AUClast=1+3+1.5+1.5,
      AUCall=1+3+1.5+1.5+0.5),
    "lin up/log down"=list(
      AUCinf=1+2/log(2)+1.5+1/log(2)+1,
      AUClast=1+2/log(2)+1.5+1/log(2),
      AUCall=1+2/log(2)+1.5+1/log(2)+0.5))
  for (t in names(tests)) {
    for (n in names(tests[[t]])) {
      # Note: using this structure ensures that there will not be
      # excessive warnings during testing.
      if (n == "AUCinf") {
        runner <- expect_warning
      } else {
        runner <- identity
      }
      runner(v1 <- pk.calc.auc(conc=c(0, 2, 0, 1, 2, 1, 0, 0),
                               time=0:7,
                               interval=c(0, 7),
                               lambda.z=1,
                               auc.type=n,
                               conc.blq=list(
                                 first="keep",
                                 middle="drop",
                                 last="keep"),
                               method=t))
      expect_equal(v1,
                   tests[[t]][[n]],
                   info=paste(t, n))
    }
  }
})

test_that("pk.calc.auc: When AUCinf is requested with NA for lambda.z, the result is NA", {
  tests <- list(
    "linear"=list(
      AUCinf=as.numeric(NA),
      AUClast=1+3+1.5+1.5,
      AUCall=1+3+1.5+1.5+0.5),
    "lin up/log down"=list(
      AUCinf=as.numeric(NA),
      AUClast=1+2/log(2)+1.5+1/log(2),
      AUCall=1+2/log(2)+1.5+1/log(2)+0.5))
  for (t in names(tests)) {
    for (n in names(tests[[t]])) {
      # Note: using this structure ensures that there will not be
      # excessive warnings during testing.
      if (n == "AUCinf") {
        runner <- expect_warning
      } else {
        runner <- identity
      }
      runner(v1 <- pk.calc.auc(conc=c(0, 2, 0, 1, 2, 1, 0, 0),
                               time=0:7,
                               interval=c(0, 7),
                               lambda.z=NA,
                               auc.type=n,
                               conc.blq=list(
                                 first="keep",
                                 middle="drop",
                                 last="keep"),
                               method=t))
      expect_equal(v1,
                   tests[[t]][[n]],
                   info=paste(t, n))
    }
  }
})

# Test NA at the beginning is in "Confirm error with NA at the same time as
# the beginning of the interval" below

test_that("pk.calc.auc: Test NA at the end", {
  # Test NA at the end
  tests <- list(
    "linear"=list(
      AUCinf=as.numeric(NA),
      AUClast=1+3+1.5+1.5,
      AUCall=1+3+1.5+1.5+1),
    "lin up/log down"=list(
      AUCinf=as.numeric(NA),
      AUClast=1+2/log(2)+1.5+1/log(2),
      AUCall=1+2/log(2)+1.5+1/log(2)+1))
  for (t in names(tests))
    for (n in names(tests[[t]])) {
      # Note: using this structure ensures that there will not be
      # excessive warnings during testing.
      if (n == "AUCinf") {
        runner <- expect_warning
      } else {
        runner <- identity
      }
      runner(v1 <- pk.calc.auc(conc=c(0, 2, 0, 1, 2, 1, NA, 0),
                               time=0:7,
                               interval=c(0, 7),
                               lambda.z=NA,
                               auc.type=n,
                               conc.blq=list(
                                 first="keep",
                                 middle="drop",
                                 last="keep"),
                               method=t))
      expect_equal(v1,
                   tests[[t]][[n]],
                   info=paste(t, n))
    }
})

test_that("pk.calc.auc: interpolation of times within the time interval", {
  tests <- list(
    "linear"=list(
      AUCinf=1+3+1.5+1.5+1,
      AUClast=1+3+1.5+1.5,
      AUCall=1+3+1.5+1.5+0.75),
    "lin up/log down"=list(
      AUCinf=1+2/log(2)+1.5+1/log(2)+1,
      AUClast=1+2/log(2)+1.5+1/log(2),
      AUCall=1+2/log(2)+1.5+1/log(2)+0.5/log(2)))
  for (t in names(tests)) {
    for (n in names(tests[[t]])) {
      # Note: using this structure ensures that there will not be
      # excessive warnings during testing.
      if (n == "AUCinf") {
        runner <- expect_warning
      } else {
        runner <- identity
      }
      runner(v1 <- pk.calc.auc(conc=c(0, 2, 0, 1, 2, 1, 0),
                               time=c(0:5, 7),
                               interval=c(0, 6),
                               lambda.z=1,
                               auc.type=n,
                               conc.blq=list(
                                 first="keep",
                                 middle="drop",
                                 last="keep"),
                               method=t))
      expect_equal(v1,
                   tests[[t]][[n]],
                   info=paste(t, n))
    }
  }
})

test_that("pk.calc.auc: warning with beginning of interval before the beginning of time", {
  na_before_start <-
    structure(
      NA_real_,
      exclude = 'Requesting an AUC range starting (-1) before the first measurement (0) is not allowed'
    )
  tests <-
    list(
      "linear"=
        list(
          AUCinf=na_before_start,
          AUClast=na_before_start,
          AUCall=na_before_start
        ),
      "lin up/log down"=
        list(
          AUCinf=na_before_start,
          AUClast=na_before_start,
          AUCall=na_before_start
        )
    )
  for (t in names(tests)) {
    for (n in names(tests[[t]])) {
      # Note: using this structure ensures that there will not be
      # excessive warnings during testing.
      expect_warning(
        v1 <- pk.calc.auc(conc=c(0, 2, 0, 1, 2, 1, 0),
                          time=c(0:5, 7),
                          interval=c(-1, Inf),
                          lambda.z=1,
                          auc.type=n,
                          conc.blq=list(
                            first="keep",
                            middle="drop",
                            last="keep"),
                          method=t),
        class="pknca_warn_auc_before_first"
      )
      expect_equal(
        v1,
        tests[[t]][[n]],
        info=paste(t, n)
      )
    }
  }
})

test_that("pk.calc.auc: warning with beginning of interval before the beginning of time", {
  # Requesting an AUC interval that starts after the last measurement
  # results in a warning, but will provide an answer (to be tested
  # elsewhere).
  expect_warning(pk.calc.auc(conc=1:2, time=0:1, interval=2:3,
                             method="linear"),
                 regexp="AUC start time \\(2\\) is after the maximum observed time \\(1\\)")

  # Confirm error with NA at the same time as the beginning of the interval
  na_before_start <-
    structure(
      NA_real_,
      exclude = 'Requesting an AUC range starting (0) before the first measurement (1) is not allowed'
    )
  tests <-
    list(
      "linear"=
        list(
          AUCinf=na_before_start,
          AUClast=na_before_start,
          AUCall=na_before_start
        ),
      "lin up/log down"=
        list(
          AUCinf=na_before_start,
          AUClast=na_before_start,
          AUCall=na_before_start
        )
    )
  for (t in names(tests))
    for (n in names(tests[[t]])) {
      expect_warning(
        v1 <- pk.calc.auc(conc=c(NA, 2, 0, 1, 2, 1, 0),
                          time=c(0:5, 7),
                          interval=c(0, Inf),
                          lambda.z=1,
                          auc.type=n,
                          conc.blq=list(
                            first="keep",
                            middle="drop",
                            last="keep"),
                          method=t),
        class = "pknca_warn_auc_before_first"
      )
      expect_equal(v1,
                   tests[[t]][[n]],
                   info=paste(t, n))
    }

  # Confirm error with concentration and time not of equal lengths
  expect_error(
    pk.calc.auc(conc=c(1, 2, 3), time=c(1, 2)),
    regexp="Assertion on 'conc' failed: Must have length 2, but has length 3."
  )

  # Confirm error with time not monotonically increasing (less than)
  expect_error(
    pk.calc.auc(conc = c(1, 2, 3), time = c(1, 2, 1)),
    regexp = "Assertion on 'time' failed: Contains duplicated values, position 3."
  )

  # Confirm error with time not monotonically increasing (equal)
  expect_error(
    pk.calc.auc(conc = c(1, 2, 3), time = c(1, 2, 2)),
    regexp = "Assertion on 'time' failed: Contains duplicated values, position 3."
  )

  # Confirm that AUC method checking works
  expect_error(pk.calc.auc(conc=c(1, 2, 3), time=c(1, 2, 3), method="wrong"),
               regexp='should be one of',
               info="Method names are confirmed for pk.calc.auc")

  # Confirm that everything works even when check is FALSE
  expect_equal(
    pk.calc.auc.inf(conc=c(0, 1, 1, 0),
                    time=0:3,
                    interval=c(0, Inf),
                    lambda.z=1,
                    check=FALSE,
                    method="linear"),
    pk.calc.auc.inf(conc=c(0, 1, 1, 0),
                    time=0:3,
                    interval=c(0, Inf),
                    lambda.z=1,
                    method="linear"))
})

test_that("pk.calc.auc.last", {
  expect_equal(
    pk.calc.auc.last(conc=c(0, 1, 1, 0),
                     time=0:3,
                     interval=c(0, 3),
                     method="linear"),
    pk.calc.auc(conc=c(0, 1, 1, 0),
                time=0:3,
                interval=c(0, 3),
                auc.type="AUClast",
                method="linear"))
  expect_error(
    pk.calc.auc.last(conc=c(0, 1, 1, 0),
                     time=0:3,
                     interval=c(0, 3),
                     auc.type="foo",
                     method="linear"),
    regexp="auc.type cannot be changed when calling pk.calc.auc.last, please use pk.calc.auc")
})

test_that("pk.calc.auc.inf", {
  # Just ensuring that it is a simple wrapper.  Computation testing
  # is done elsewhere.
  expect_equal(
    pk.calc.auc.inf(conc=c(0, 1, 1, 0),
                    time=0:3,
                    interval=c(0, Inf),
                    lambda.z=1,
                    method="linear"),
    pk.calc.auc(conc=c(0, 1, 1, 0),
                time=0:3,
                interval=c(0, Inf),
                lambda.z=1,
                auc.type="AUCinf",
                method="linear"))
  expect_error(
    pk.calc.auc.inf(conc=c(0, 1, 1, 0),
                    time=0:3,
                    interval=c(0, 3),
                    auc.type="foo",
                    method="linear"),
    regexp="auc.type cannot be changed when calling pk.calc.auc.inf, please use pk.calc.auc")
})

test_that("pk.calc.auc.all", {
  # Just ensuring that it is a simple wrapper.  Computation testing
  # is done elsewhere.
  expect_equal(
    pk.calc.auc.all(conc=c(0, 1, 1, 0),
                    time=0:3,
                    interval=c(0, 3),
                    method="linear"),
    pk.calc.auc(conc=c(0, 1, 1, 0),
                time=0:3,
                interval=c(0, 3),
                auc.type="AUCall",
                method="linear"))
  expect_error(
    pk.calc.auc.all(conc=c(0, 1, 1, 0),
                    time=0:3,
                    interval=c(0, 3),
                    auc.type="foo",
                    method="linear"),
    regexp="auc.type cannot be changed when calling pk.calc.auc.all, please use pk.calc.auc")
})

test_that("pk.calc.aumc", {
  expect_equal(
    pk.calc.aumc(
      conc=c(0, 1, 1, 0.5),
      time=0:3,
      interval=c(0, 3),
      method="linear"),
    3.75)
  expect_equal(
    pk.calc.aumc(
      conc=c(0, 1, 1, 0.5),
      time=0:3,
      interval=c(0, 3),
      method="lin up/log down"),
    2-0.5/log(0.5)+0.5/(log(0.5)^2))
  expect_equal(
    pk.calc.aumc(
      conc=c(0, 1, 1, 0.5),
      time=0:3,
      interval=c(0, Inf),
      auc.type="AUCinf",
      lambda.z=1,
      method="lin up/log down"),
    2 - 0.5/log(0.5) + 0.5/(log(0.5)^2) + 1.5 + 0.5)
})


test_that("pk.calc.aumc.inf", {
  # Just ensuring that it is a simple wrapper.  Computation testing
  # is done elsewhere.
  expect_equal(
    pk.calc.aumc.inf(conc=c(0, 1, 1, 0),
                     time=0:3,
                     interval=c(0, Inf),
                     lambda.z=1,
                     method="linear"),
    pk.calc.aumc(conc=c(0, 1, 1, 0),
                 time=0:3,
                 interval=c(0, Inf),
                 lambda.z=1,
                 auc.type="AUCinf",
                 method="linear"))
  expect_error(
    pk.calc.aumc.inf(conc=c(0, 1, 1, 0),
                    time=0:3,
                    interval=c(0, 3),
                    auc.type="foo",
                    method="linear"),
    regexp="auc.type cannot be changed when calling pk.calc.aumc.inf, please use pk.calc.aumc")
})

test_that("pk.calc.aumc.all", {
  # Just ensuring that it is a simple wrapper.  Computation testing
  # is done elsewhere.
  expect_equal(
    pk.calc.aumc.all(conc=c(0, 1, 1, 0),
                    time=0:3,
                    interval=c(0, 3),
                    method="linear"),
    pk.calc.aumc(conc=c(0, 1, 1, 0),
                 time=0:3,
                 interval=c(0, 3),
                 auc.type="AUCall",
                 method="linear"))
  expect_error(
    pk.calc.aumc.all(conc=c(0, 1, 1, 0),
                    time=0:3,
                    interval=c(0, 3),
                    auc.type="foo",
                    method="linear"),
    regexp="auc.type cannot be changed when calling pk.calc.aumc.all, please use pk.calc.aumc")
})

test_that("pk.calc.aumc.last", {
  # Just ensuring that it is a simple wrapper.  Computation testing
  # is done elsewhere.
  expect_equal(
    pk.calc.aumc.last(conc=c(0, 1, 1, 0),
                    time=0:3,
                    interval=c(0, 3),
                    method="linear"),
    pk.calc.aumc(conc=c(0, 1, 1, 0),
                 time=0:3,
                 interval=c(0, 3),
                 auc.type="AUClast",
                 method="linear"))
  expect_error(
    pk.calc.aumc.last(conc=c(0, 1, 1, 0),
                    time=0:3,
                    interval=c(0, 3),
                    auc.type="foo",
                    method="linear"),
    regexp="auc.type cannot be changed when calling pk.calc.aumc.last, please use pk.calc.aumc")
})

test_that("pk.calc.auc.inf.pred", {
  expect_equal(
    pk.calc.auc.inf.pred(
      conc=c(0, 1, 1, 0),
      time=0:3,
      clast.pred=1,
      lambda.z=1,
      interval=c(0, Inf),
      method="linear"),
    pk.calc.auc.inf(
      conc=c(0, 1, 1, 0),
      time=0:3,
      clast=1,
      lambda.z=1,
      interval=c(0, Inf),
      method="linear"),
    info="pk.calc.auc.inf.pred is a simple wrapper around pk.calc.auc.inf")
})

test_that("pk.calc.aumc.inf.obs", {
  expect_equal(
    pk.calc.aumc.inf.obs(
      conc=c(0, 1, 1, 0),
      time=0:3,
      clast.obs=1,
      lambda.z=1,
      interval=c(0, Inf),
      method="linear"),
    pk.calc.aumc.inf(
      conc=c(0, 1, 1, 0),
      time=0:3,
      clast=1,
      lambda.z=1,
      interval=c(0, Inf),
      method="linear"),
    info="pk.calc.aumc.inf.obs is a simple wrapper around pk.calc.aumc.inf")
})

test_that("pk.calc.aumc.inf.pred", {
  expect_equal(
    pk.calc.aumc.inf.pred(
      conc=c(0, 1, 1, 0),
      time=0:3,
      clast.pred=1,
      lambda.z=1,
      interval=c(0, Inf),
      method="linear"),
    pk.calc.aumc.inf(
      conc=c(0, 1, 1, 0),
      time=0:3,
      clast=1,
      lambda.z=1,
      interval=c(0, Inf),
      method="linear"),
    info="pk.calc.aumc.inf.pred is a simple wrapper around pk.calc.aumc.inf")
})

test_that("AUC with a single concentration measured should return NA (fix #176)", {
  expect_equal(
    pk.calc.auc.last(conc = 1, time = 0),
    structure(
      NA_real_,
      exclude="AUC cannot be calculated with only one measured concentration"
    )
  )
  # One concentration measured with others missing is sill excluded
  expect_equal(
    pk.calc.auxc(
      conc=c(NA, 0, NA), time=2:4, interval=c(1, 3),
      method="linear"
    ),
    structure(
      NA_real_,
      exclude="AUC cannot be calculated with only one measured concentration"
    )
  )
})

#'  Set Intervals
#'
#'  Takes in two objects, the PKNCAdata object and the proposed intervals.
#'  It will then check that the intervals are valid, given the data object.
#'  If the intervals are valid, it will set them in the object.
#'  It will return the data object with the intervals set.
#'  
#' @param data PKNCAdata object
#' @param intervals Proposed intervals
#' @returns The data object with the intervals set.
#'   
#' @export
set_intervals <- function(data, intervals) {
  valid_intervals <- assert_intervals(intervals, data)
  
  data$intervals <- valid_intervals
  
  data
}

#'  Assert Intervals
#'
#'  Verifies that an interval definition is valid for a PKNCAdata object.
#'  Valid means that intervals are a data.frame (or data.frame-like object), 
#'  that the column names are either the groupings of the PKNCAconc part of 
#'  the PKNCAdata object or that they are one of the NCA parameters allowed 
#'  (i.e. names(get.interval.cols())). 
#'  It will return the intervals argument unchanged, or it will raise an error.
#'  
#' @param intervals Proposed intervals
#' @param data PKNCAdata object
#' @returns The intervals argument unchanged, or it will raise an error.
#'   
#' @export
assert_intervals <- function(intervals, data) {
  if (!is.data.frame(intervals)) {
    stop("The 'intervals' argument must be a data frame or a data frame-like object.")
  }
  
  if (!inherits(data, "PKNCAdata")) {
    stop("The 'data' argument must be a PKNCAdata object.")
  }
  
  allowed_columns <-
    c(
      names(getGroups.PKNCAdata(data)),
      names(get.interval.cols()),
      "conc_above",
      "time_above",
      "impute",
      # If not used, data$options$keep_interval_cols will be NULL
      data$options$keep_interval_cols
    )
  
  invalid_columns <- setdiff(names(intervals), allowed_columns)
  
  if (length(invalid_columns) > 0) {
    stop("The following columns in 'intervals' are not allowed: ", paste(invalid_columns, collapse = ", "))
  }
  
  intervals
}

test_that("assert_intervals works with valid intervals", {
  o_conc <- PKNCAconc(as.data.frame(datasets::Theoph), conc~Time|Subject)
  o_data <- PKNCAdata(o_conc, intervals = data.frame(start = 0, end = 1, cmax = TRUE))
  
  result <- assert_intervals(intervals = data.frame(start = 0, end = 1, cmax = TRUE), data = o_data)
  expect_equal(result, expected = data.frame(start = 0, end = 1, cmax = TRUE))
})

test_that("assert_intervals works with valid intervals (ungrouped)", {
  o_conc <- PKNCAconc(as.data.frame(datasets::Theoph)[datasets::Theoph$Subject == 1,], conc~Time)
  o_data <- PKNCAdata(o_conc, intervals = data.frame(start = 0, end = 1, cmax = TRUE))
  
  result <- assert_intervals(intervals = data.frame(start = 0, end = 1, cmax = TRUE), data = o_data)
  expect_equal(result, expected = data.frame(start = 0, end = 1, cmax = TRUE))
})


test_that("assert_intervals errors with non-data frame intervals", {
  o_conc <- PKNCAconc(as.data.frame(datasets::Theoph), conc~Time|Subject)
  o_data <- PKNCAdata(o_conc, intervals = data.frame(start = 0, end = 1, cmax = TRUE))
  non_df_intervals <- list(a = 1, b = 2)
  
  expect_error(assert_intervals(non_df_intervals, data = o_data), 
               regex = "The 'intervals' argument must be a data frame or a data frame-like object.",
               fixed = TRUE)
})

test_that("assert_intervals errors with non-data frame intervals (ungrouped)", {
  o_conc <- PKNCAconc(as.data.frame(datasets::Theoph)[datasets::Theoph$Subject == 1,], conc~Time)
  o_data <- PKNCAdata(o_conc, intervals = data.frame(start = 0, end = 1, cmax = TRUE))
  non_df_intervals <- list(a = 1, b = 2)
  
  expect_error(assert_intervals(intervals = non_df_intervals, data = o_data), 
               regex = "The 'intervals' argument must be a data frame or a data frame-like object.",
               fixed = TRUE)
})

test_that("assert_intervals errors with non-PKNCAdata data object", {
  expect_error(assert_intervals(intervals = data.frame(start = 0, end = 1, cmax = TRUE), 
               data = data.frame(a = 1, b = 2)),
               regex = "The 'data' argument must be a PKNCAdata object.",
               fixed = TRUE)
})

test_that("assert_intervals errors with invalid columns", {
  o_conc <- PKNCAconc(as.data.frame(datasets::Theoph), conc~Time|Subject)
  o_data <- PKNCAdata(o_conc, intervals = data.frame(start = 0, end = 1, cmax = TRUE))
  
  invalid_intervals <- data.frame(
    mean = TRUE,  # Not allowed NCA params
    median = TRUE
  )
  
  expect_error(assert_intervals(intervals = invalid_intervals, data = o_data), 
               regex = "The following columns in 'intervals' are not allowed:",
               fixed = TRUE)
})

test_that("assert_intervals errors with invalid columns", {
  o_conc <- PKNCAconc(as.data.frame(datasets::Theoph)[datasets::Theoph$Subject == 1,], conc~Time)
  o_data <- PKNCAdata(o_conc, intervals = data.frame(start = 0, end = 1, cmax = TRUE))
  
  invalid_intervals <- data.frame(
    mean = TRUE,  # Not allowed NCA params
    median = TRUE
  )
  
  expect_error(assert_intervals(intervals = invalid_intervals, data = o_data), 
               regex = "The following columns in 'intervals' are not allowed:",
               fixed = TRUE)
})

test_that("set_intervals works with valid intervals", {
  o_conc <- PKNCAconc(as.data.frame(datasets::Theoph), conc~Time|Subject)
  o_data <- PKNCAdata(o_conc, intervals = data.frame(start = 0, end = 1, cmax = TRUE))
  
  result <- set_intervals(data = o_data, intervals = data.frame(start = 0, end = 1, cmin = TRUE))
  
  expect_equal(result$intervals, data.frame(start = 0, end = 1, cmin = TRUE))
})

test_that("set_intervals works with valid intervals (ungrouped)", {
  o_conc <- PKNCAconc(as.data.frame(datasets::Theoph)[datasets::Theoph$Subject == 1,], conc~Time)
  o_data <- PKNCAdata(o_conc, intervals = data.frame(start = 0, end = 1, cmax = TRUE))
  
  result <- set_intervals(data = o_data, intervals = data.frame(start = 0, end = 1, cmin = TRUE))
  
  expect_equal(result$intervals, data.frame(start = 0, end = 1, cmin = TRUE))
})

test_that("set_intervals fails with invalid intervals", {
  o_conc <- PKNCAconc(as.data.frame(datasets::Theoph), conc~Time|Subject)
  o_data <- PKNCAdata(o_conc, intervals = data.frame(start = 0, end = 1, cmax = TRUE))
  
  expect_error(set_intervals(data = o_data, intervals = data.frame(start = 0, end = 1, cmedian = TRUE)), 
               regex = "The following columns in 'intervals' are not allowed:",
               fixed = TRUE)
})

test_that("set_intervals fails when not using PKNCAdata", {
  o_conc <- PKNCAconc(as.data.frame(datasets::Theoph), conc~Time|Subject)
  o_data <- PKNCAdata(o_conc, intervals = data.frame(start = 0, end = 1, cmax = TRUE))
  
  expect_error(set_intervals(data = o_conc, intervals = data.frame(start = 0, end = 1, cmin = TRUE)), 
               regex = "The 'data' argument must be a PKNCAdata object.",
               fixed = TRUE)
})

#' Times relative to an event (typically dosing)
#'
#' @param time_event A vector of times for events
#' @param time_obs A vector of times for observations
#' @param units Passed to `base::as.numeric.difftime()`
#' @returns A data.frame with columns for:
#' \describe{
#'   \item{event_number_before}{The index of `time_event` that is the last one before `time_obs` or `NA` if none are before.}
#'   \item{event_number_after}{The index of `time_event` that is the first one after `time_obs` or `NA` if none are after.}
#'   \item{time_before}{The minimum time that the current `time_obs` is before a `time_event`, 0 if at least one `time_obs == time_event`.}
#'   \item{time_after}{The minimum time that the current `time_obs` is after a `time_event`, 0 if at least one `time_obs == time_event`.}
#'   \item{time_after_first}{The time after the first event (may be negative or positive).}
#' }
#'
#' `time_after` and `time_before` are calculated if they are at the same time as
#' a dose, they equal zero, and otherwise, they are calculated relative to the
#' dose number in the `event_number_*` columns.
#' @export
time_calc <- function(time_event, time_obs, units=NULL) {
  UseMethod("time_calc")
}

#' @export
time_calc.numeric <- function(time_event, time_obs, units=NULL) {
  if (length(time_event) == 0) {
    warning("No events provided")
    time_event <- NA_real_
  } else if (any(order(stats::na.omit(time_event)) != seq_along(stats::na.omit(time_event)))) {
    stop("`time_event` must be sorted.")
  }
  if (!is.numeric(time_obs)) {
    stop("Both `time_event` and `time_obs` must be the same class (numeric).")
  }
  ret <-
    data.frame(
      event_number_before=
        vapply(
          X = time_obs,
          FUN = function(x)
            max_zero_len(
              which(time_event <= x),
              zero_length=NA_integer_
            ),
          FUN.VALUE = 1L
        ),
      event_number_after=
        vapply(
          X = time_obs,
          FUN = function(x)
            min_zero_len(
              which(time_event >= x),
              zero_length = NA_integer_
            ),
          FUN.VALUE = 1L
        )
    )
  ret$time_after_event <-
    time_obs - time_event[ret$event_number_before]
  ret$time_before_event <-
    time_obs - time_event[ret$event_number_after]
  ret$time_after_first <-
    if (all(is.na(time_event))) {
      NA_real_
    } else {
      time_obs - min(time_event, na.rm=TRUE)
    }
  ret
}

#' @export
time_calc.POSIXt <- function(time_event, time_obs, units=NULL) {
  if (is.null(units)) {
    stop("`units` must be provided.")
  }
  if (!("POSIXt" %in% class(time_obs))) {
    stop("Both `time_event` and `time_obs` must be the same class (POSIXt).")
  }
  first_event <- min(time_event, na.rm=TRUE)
  time_calc(
    time_event=difftime(time_event, first_event, units=units),
    time_obs=difftime(time_obs, first_event, units=units),
    units=units
  )
}

#' @export
time_calc.difftime <- function(time_event, time_obs, units=NULL) {
  if (is.null(units)) {
    stop("`units` must be provided.")
  }
  if (!("difftime" %in% class(time_obs))) {
    stop("Both `time_event` and `time_obs` must be the same class (difftime).")
  }
  time_calc(
    time_event=as.numeric(time_event, units=units),
    time_obs=as.numeric(time_obs, units=units)
  )
}

test_that("Numeric time_calc() is successful with all scalar combinations of number and NA and numeric order.", {
  expect_equal(
    time_calc(time_event=0, time_obs=0),
    data.frame(
      event_number_before=1,
      event_number_after=1,
      time_after_event=0,
      time_before_event=0,
      time_after_first=0
    )
  )
  expect_equal(
    time_calc(time_event=1, time_obs=0),
    data.frame(
      event_number_before=NA_integer_,
      event_number_after=1,
      time_after_event=NA_real_,
      time_before_event=-1,
      time_after_first=-1
    )
  )
  expect_equal(
    time_calc(time_event=0, time_obs=1),
    data.frame(
      event_number_before=1,
      event_number_after=NA_integer_,
      time_after_event=1,
      time_before_event=NA_real_,
      time_after_first=1
    )
  )
  expect_equal(
    time_calc(time_event=NA_real_, time_obs=1),
    data.frame(
      event_number_before=NA_integer_,
      event_number_after=NA_integer_,
      time_after_event=NA_real_,
      time_before_event=NA_real_,
      time_after_first=NA_real_
    )
  )
  expect_equal(
    time_calc(time_event=0, time_obs=NA_real_),
    data.frame(
      event_number_before=NA_integer_,
      event_number_after=NA_integer_,
      time_after_event=NA_real_,
      time_before_event=NA_real_,
      time_after_first=NA_real_
    )
  )
  expect_equal(
    time_calc(time_event=NA_real_, time_obs=NA_real_),
    data.frame(
      event_number_before=NA_integer_,
      event_number_after=NA_integer_,
      time_after_event=NA_real_,
      time_before_event=NA_real_,
      time_after_first=NA_real_
    )
  )
  expect_error(
    time_calc(
      time_event=5,
      time_obs=as.difftime(5, units="hours")
    ),
    regexp="Both `time_event` and `time_obs` must be the same class (numeric).",
    fixed=TRUE
  )
  expect_warning(
    expect_equal(
      time_calc(time_event=numeric(0), time_obs=1),
      time_calc(time_event=NA_real_, time_obs=1)
    ),
    regexp="No events provided",
    fixed=TRUE
  )
})

test_that("Numeric time_calc() is successful with all vector combinations of number and NA and numeric order.", {
  expect_equal(
    time_calc(time_event=c(0, 10, 20), time_obs=c(-1, 0, 1, 9, 10, 11, 19, 20, 21)),
    data.frame(
      event_number_before=c(NA_integer_, 1, 1, 1, 2, 2, 2, 3, 3),
      event_number_after=c(1, 1, 2, 2, 2, 3, 3, 3, NA_integer_),
      time_after_event=c(NA_real_, 0, 1, 9, 0, 1, 9, 0, 1),
      time_before_event=c(-1, 0, -9, -1, 0, -9, -1, 0, NA_real_),
      time_after_first=c(-1, 0, 1, 9, 10, 11, 19, 20, 21)
    )
  )
  expect_equal(
    time_calc(time_event=c(0, 10, 10, 20), time_obs=c(-1, 0, 1, 9, 10, 11, 19, 20, 21)),
    data.frame(
      event_number_before=c(NA_integer_, 1, 1, 1, 3, 3, 3, 4, 4),
      event_number_after=c(1, 1, 2, 2, 2, 4, 4, 4, NA_integer_),
      time_after_event=c(NA_real_, 0, 1, 9, 0, 1, 9, 0, 1),
      time_before_event=c(-1, 0, -9, -1, 0, -9, -1, 0, NA_real_),
      time_after_first=c(-1, 0, 1, 9, 10, 11, 19, 20, 21)
    ),
    info="Duplicated events are counted as the first for event_number_before and the last for event_number_after."
  )
  expect_equal(
    time_calc(time_event=c(0, 10, 10, 20, NA_real_), time_obs=c(-1, 0, 1, 9, 10, 11, 19, 20, 21)),
    data.frame(
      event_number_before=c(NA_integer_, 1, 1, 1, 3, 3, 3, 4, 4),
      event_number_after=c(1, 1, 2, 2, 2, 4, 4, 4, NA_integer_),
      time_after_event=c(NA_real_, 0, 1, 9, 0, 1, 9, 0, 1),
      time_before_event=c(-1, 0, -9, -1, 0, -9, -1, 0, NA_real_),
      time_after_first=c(-1, 0, 1, 9, 10, 11, 19, 20, 21)
    ),
    info="NA time_event is ignored."
  )
  expect_equal(
    time_calc(time_event=c(0, 10, NA_real_, 10, 20), time_obs=c(-1, 0, 1, 9, 10, 11, 19, 20, 21)),
    data.frame(
      event_number_before=c(NA_integer_, 1, 1, 1, 4, 4, 4, 5, 5),
      event_number_after=c(1, 1, 2, 2, 2, 5, 5, 5, NA_integer_),
      time_after_event=c(NA_real_, 0, 1, 9, 0, 1, 9, 0, 1),
      time_before_event=c(-1, 0, -9, -1, 0, -9, -1, 0, NA_real_),
      time_after_first=c(-1, 0, 1, 9, 10, 11, 19, 20, 21)
    ),
    info="NA time_event is counted correctly in the middle."
  )
  expect_error(
    time_calc(time_event=c(10, 0), time_obs=c(-1, 0, 1, 9, 10, 11, 19, 20, 21)),
    regexp="`time_event` must be sorted."
  )
})

test_that("POSIXt objects work", {
  expect_equal(
    time_calc(
      time_event=as.POSIXlt("2019-07-30 06:16"),
      time_obs=as.POSIXct("2019-07-30 06:00"),
      units="hours"
    ),
    data.frame(
      event_number_before=NA_integer_,
      event_number_after=1L,
      time_after_event=NA_real_,
      time_before_event=-16/60,
      time_after_first=-16/60
    )
  )
  expect_equal(
    time_calc(
      time_event=as.POSIXlt("2019-07-30 06:16"),
      time_obs=as.POSIXct("2019-07-30 06:00"),
      units="days"
    ),
    data.frame(
      event_number_before=NA_integer_,
      event_number_after=1L,
      time_after_event=NA_real_,
      time_before_event=-16/1440,
      time_after_first=-16/1440
    ),
    info="Units are respected"
  )
  expect_equal(
    time_calc(
      time_event=as.POSIXlt("2019-07-30 06:16"),
      time_obs=as.POSIXct("2019-07-30 06:00"),
      units="hours"
    ),
    data.frame(
      event_number_before=NA_integer_,
      event_number_after=1L,
      time_after_event=NA_real_,
      time_before_event=-16/60,
      time_after_first=-16/60
    ),
    info="Mixed POSIXlt and POSIXct inputs work"
  )
  expect_error(
    time_calc(
      time_event=as.POSIXlt("2019-07-30 06:16"),
      time_obs=5,
      units="hours"
    ),
    regexp="Both `time_event` and `time_obs` must be the same class (POSIXt).",
    fixed=TRUE
  )
  expect_error(
    time_calc(
      time_event=as.POSIXlt("2019-07-30 06:16"),
      time_obs=as.POSIXct("2019-07-30 06:00")
    ),
    regexp="`units` must be provided.",
    fixed=TRUE
  )
})

test_that("difftime", {
  expect_error(
    time_calc(
      time_event=as.difftime(5, units="hours"),
      time_obs=as.difftime(5, units="hours")
    ),
    regexp="`units` must be provided.",
    fixed=TRUE
  )
  expect_error(
    time_calc(
      time_event=as.difftime(5, units="hours"),
      time_obs=5,
      units="hours"
    ),
    regexp="Both `time_event` and `time_obs` must be the same class (difftime).",
    fixed=TRUE
  )
  expect_equal(
    time_calc(
      time_event=as.difftime(5, units="hours"),
      time_obs=as.difftime(5, units="hours"),
      units="hours"
    ),
    time_calc(
      time_event=5,
      time_obs=5
    )
  )
  expect_equal(
    time_calc(
      time_event=as.difftime(5, units="hours"),
      time_obs=as.difftime(5, units="hours"),
      units="days"
    ),
    time_calc(
      time_event=5/24,
      time_obs=5/24
    ),
    info="Units are not inherited from the difftime object."
    # because they could differ, and the default automatic selection may not
    # match the user's desire.
  )
})

#' Create a PKNCAdose object
#'
#' @param data A data frame with time and the groups defined in `formula`.
#' @param formula The formula defining the `dose.amount~time|groups` where
#'   `time` is the time of the dosing and `dose.amount` is the amount
#'   administered at that time (see Details).
#' @param route Define the route of administration.  The value may be either a
#'   column name from the `data` (checked first) or a character string of either
#'   `"extravascular"` or `"intravascular"` (checked second).  If given as a
#'   column name, then every value of the column must be either
#'   `"extravascular"` or `"intravascular"`.
#' @param rate,duration (optional) for `"intravascular"` dosing, the rate or
#'   duration of dosing.  If given as a character string, it is the name of a
#'   column from the `data`, and if given as a number, it is the value for all
#'   doses.  Only one may be given, and if neither is given, then the dose is
#'   assumed to be a bolus (`duration=0`).  If `rate` is given, then the dose
#'   amount must be given (the left hand side of the `formula`).
#' @param time.nominal (optional) The name of the nominal time column (if the
#'   main time variable is actual time.  The `time.nominal` is not used during
#'   calculations; it is available to assist with data summary and checking.
#' @param exclude (optional) The name of a column with concentrations to exclude
#'   from calculations and summarization.  If given, the column should have
#'   values of `NA` or `""` for concentrations to include and non-empty text for
#'   concentrations to exclude.
#' @param ... Ignored.
#' @returns A PKNCAconc object that can be used for automated NCA.
#' @details The `formula` for a `PKNCAdose` object can be
#'   given three ways: one-sided (missing left side), one-sided (missing
#'   right side), or two-sided.  Each of the three ways can be given
#'   with or without groups.  When given one-sided missing the left
#'   side, the left side can either be omitted or can be given as a
#'   period (`.`): `~time|treatment+subject` and
#'   `.~time|treatment+subject` are identical, and dose-related NCA
#'   parameters will all be reported as not calculable (for example,
#'   clearance).  When given one-sided missing the right side, the right
#'   side must be specified as a period (`.`):
#'   `dose~.|treatment+subject`, and only a single row may be given
#'   per group.  When the right side is missing, PKNCA assumes that the
#'   same dose is given in every interval.  When given as a two-sided
#'   formula
#' @family PKNCA objects
#' @export
PKNCAdose <- function(data, ...)
  UseMethod("PKNCAdose")

#' @rdname PKNCAdose
#' @export
PKNCAdose.default <- function(data, ...) {
  PKNCAdose.data.frame(as.data.frame(data), ...)
}
#' @rdname PKNCAdose
#' @export
PKNCAdose.tbl_df <- function(data, ...) {
  PKNCAdose.data.frame(as.data.frame(data), ...)
}

#' @rdname PKNCAdose
#' @param doseu Either unit values (e.g. "mg") or column names within the data
#'   where units are provided.
#' @param doseu_pref Preferred units for reporting (not column names)
#' @export
PKNCAdose.data.frame <- function(data, formula, route, rate, duration,
                                 time.nominal, exclude = NULL, ...,
                                 doseu = NULL, doseu_pref = NULL) {
  # The data must have... data
  if (nrow(data) == 0) {
    stop("data must have at least one row.")
  }
  # Check inputs
  if (!missing(time.nominal)) {
    if (!(time.nominal %in% names(data))) {
      stop("time.nominal, if given, must be a column name in the input data.")
    }
  }
  # Verify that all the variables in the formula are columns in the data.
  parsed_form_raw <- parse_formula_to_cols(form = formula)
  if (length(parsed_form_raw$groups_left_of_slash) > 0) {
    stop("formula for PKNCAdose may not include a slash")
  }
  parsed_form_groups <-
    if (length(parsed_form_raw$groups) > 0) {
      list(
        group_vars=parsed_form_raw$groups,
        group_analyte=character()
      )
    } else {
      list(
        group_vars=parsed_form_raw$groups_left_of_slash,
        group_analyte=parsed_form_raw$groups_right_of_slash
      )
    }
  parsed_form <-
    list(
      dose = parsed_form_raw$lhs,
      time = parsed_form_raw$rhs,
      groups = parsed_form_groups
    )
  # Check for variable existence and length
  if (!(length(parsed_form$dose) %in% c(0, 1))) {
    stop("The left side of the formula must have zero or one variable")
  } else if (length(parsed_form$dose) == 1 &&
             !(parsed_form$dose %in% names(data))) {
    # the "." is handled in parse_formula_to_cols
    stop("The left side formula must be a variable in the data, empty, or '.'.")
  }
  if (!(length(parsed_form$time) %in% c(0, 1))) {
    stop("The right side of the formula (excluding groups) must have exactly one variable")
  } else if (length(parsed_form$time) == 1 &&
             !(parsed_form$time %in% names(data))) {
    stop("The right side formula must be a variable in the data or '.'.")
  }
  if (!all(unlist(parsed_form$groups) %in% names(data))) {
    stop("All of the variables in the groups must be in the data")
  }
  ret <-
    list(
      data = data,
      formula = formula,
      columns = parsed_form
    )
  class(ret) <- c("PKNCAdose", class(ret))
  ret <- setExcludeColumn(ret, exclude = exclude)
  # Values must be unique (one value per measurement), check after the exclusion
  # column has been added to the object so that exclusions can be accounted for
  # in duplicate checking.
  duplicate_check(object = ret, data_type = "dosing")

  mask.indep <- is.na(getIndepVar.PKNCAdose(ret))
  if (any(mask.indep) & !all(mask.indep)) {
    stop("Some but not all values are missing for the independent variable, please see the help for PKNCAdose for how to specify the formula and confirm that your data has dose times for all doses.")
  }
  if (missing(route)) {
    ret <- setRoute(ret)
  } else {
    ret <- setRoute(ret, route)
  }
  ret <- setDuration.PKNCAdose(ret, duration=duration,
                               rate=rate, dose=getDepVar.PKNCAdose(ret))
  if (!missing(time.nominal)) {
    ret <-
      setAttributeColumn(object=ret,
                         attr_name="time.nominal",
                         col_name=time.nominal)
  }

  # Unit handling
  ret <-
    pknca_set_units(
      ret,
      units_orig = list(doseu = doseu),
      units_pref = list(doseu_pref = doseu_pref)
    )

  ret
}

#' Set the dosing route
#'
#' @param object A PKNCAdose object
#' @param route A character string indicating one of the following:  the column
#'   from the data which indicates the route of administration, a scalar
#'   indicating the route of administration for all subjects, or a vector
#'   indicating the route of administration for each dose in the dataset.
#' @param ... Arguments passed to another setRoute function
#' @returns The object with an updated route
#' @export
setRoute <- function(object, ...) {
  UseMethod("setRoute")
}
#' @rdname setRoute
#' @export
setRoute.PKNCAdose <- function(object, route, ...) {
  if (missing(route)) {
    object <-
      setAttributeColumn(object=object,
                         attr_name="route",
                         default_value="extravascular",
                         message_if_default="Assuming route of administration is extravascular")
  } else {
    object <-
      setAttributeColumn(object=object,
                         attr_name="route",
                         col_or_value=route)
  }
  if (!all(tolower(getAttributeColumn(object=object, attr_name="route")[[1]]) %in%
           c("extravascular", "intravascular"))) {
    stop("route must have values of either 'extravascular' or 'intravascular'.  Please set to one of those values and retry.")
  }
  object
}

#' Set the duration of dosing or measurement
#'
#' @param object An object to set a duration on
#' @param duration The value to set for the duration or the name of the column
#'   in the data to use for the duration.
#' @param rate (for PKNCAdose objects only) The rate of infusion
#' @param dose (for PKNCAdose objects only) The dose amount
#' @param ... Arguments passed to another setDuration function
#' @returns The object with duration set
#' @export
setDuration <- function(object, ...)
  UseMethod("setDuration")
#' @rdname setDuration
#' @export
setDuration.PKNCAdose <- function(object, duration, rate, dose, ...) {
  if (missing(dose)) {
    dose <- object$columns$dose
  }
  if (missing(duration) & missing(rate)) {
    object <- setAttributeColumn(object=object, attr_name="duration", default_value=0,
                                 message_if_default="Assuming instant dosing (duration=0)")

  } else if (!missing(duration) & !missing(rate)) {
    stop("Both duration and rate cannot be given at the same time")
    # TODO: A consistency check could be done, but that would get into
    # requiring near-equal checks for floating point error.
  } else if (!missing(duration)) {
    object <- setAttributeColumn(object=object, attr_name="duration", col_or_value=duration)
  } else if (!missing(rate) & !missing(dose) && !is.na(dose)) {
    tmprate <- getColumnValueOrNot(object$data, rate, "rate")
    tmpdose <- getColumnValueOrNot(object$data, dose, "dose")
    duration <- tmpdose$data[[tmpdose$name]]/tmprate$data[[tmprate$name]]
    object <- setAttributeColumn(object=object, attr_name="duration", col_or_value=duration)
  }
  duration.val <- getAttributeColumn(object=object, attr_name="duration")[[1]]
  if (is.numeric(duration.val) &&
      !any(is.na(duration.val)) &&
      !any(is.infinite(duration.val)) &&
      all(duration.val >= 0)) {
    # It passes
  } else {
    stop("duration must be numeric without missing (NA) or infinite values, and all values must be >= 0")
  }
  object
}

#' @rdname formula.PKNCAconc
#' @export
#' @importFrom stats formula
formula.PKNCAdose <-  function(x, ...) {
  x$formula
}

#' @rdname model.frame.PKNCAconc
#' @export
#' @importFrom stats model.frame
#' @method model.frame PKNCAdose
model.frame.PKNCAdose <- function(formula, ...) {
  cbind(getDepVar.PKNCAdose(formula),
        getIndepVar.PKNCAdose(formula),
        getGroups.PKNCAdose(formula))
}

#' @export
getDepVar.PKNCAdose <- function(x, ...) {
  dose_col <- x$columns$dose
  if (length(dose_col) == 0) {
    rep(NA_integer_, nrow(x$data))
  } else {
    x$data[, dose_col]
  }
}

#' @export
getIndepVar.PKNCAdose <- function(x, ...) {
  time_col <- x$columns$time
  if (length(time_col) == 0) {
    rep(NA_integer_, nrow(x$data))
  } else {
    x$data[, time_col]
  }
}

#' @rdname getGroups.PKNCAconc
#' @export
getGroups.PKNCAdose <- function(...) {
  getGroups.PKNCAconc(...)
}

#' @describeIn group_vars.PKNCAconc Get group_vars for a PKNCAdose object
#' @exportS3Method dplyr::group_vars
group_vars.PKNCAdose <- function(x) {
  group_vars.PKNCAconc(x)
}

#' @rdname getDataName
#' @export
getDataName.PKNCAdose <- function(object) {
  "data"
}

#' @rdname print.PKNCAconc
#' @export
print.PKNCAdose <- function(x, n=6, summarize=FALSE, ...) {
  cat("Formula for dosing:\n ")
  print(stats::formula(x), showEnv=FALSE, ...)
  if (!is.null(time_nom_data <- getAttributeColumn(x, attr_name="time.nominal", warn_missing=character()))) {
    cat("Nominal time column is: ", names(time_nom_data), "\n", sep="")
  } else {
    cat("Nominal time column is not specified.\n")
  }
  if (summarize) {
    cat("\n")
    grp <- getGroups.PKNCAdose(x)
    if (ncol(grp) > 0) {
      tmp.summary <- as.data.frame(
        lapply(grp, FUN=function(y) length(unique(y))))
      cat("Number unique entries in each group:\n")
      print.data.frame(tmp.summary, row.names=FALSE)
    } else {
      cat("No groups.\n")
    }
  } else {
    if (n != 0) {
      if (n >= nrow(x$data)) {
        cat("\nData for dosing:\n")
      } else if (n < 0) {
        cat(sprintf("\nFirst %d rows of dosing data:\n",
                    nrow(x$data)+n))
      } else {
        cat(sprintf("\nFirst %d rows of dosing data:\n",
                    n))
      }
      print.data.frame(utils::head(x$data, n=n), ..., row.names=FALSE)
    }
  }
}

#' @rdname print.PKNCAconc
#' @export
summary.PKNCAdose <- summary.PKNCAconc

#' @export
as.data.frame.PKNCAdose <- function(x, ...) {
  x$data
}

test_that("PKNCAdose", {
  tmp.conc <- generate.conc(nsub=5, ntreat=2, time.points=0:24)
  tmp.conc.analyte <- generate.conc(nsub=5, ntreat=2, time.points=0:24,
                                    nanalytes=2)
  tmp.conc.study <- generate.conc(nsub=5, ntreat=2, time.points=0:24,
                                  nstudies=2)
  tmp.conc.analyte.study <- generate.conc(nsub=5, ntreat=2, time.points=0:24,
                                          nanalytes=2, nstudies=2)
  tmp.dose <- generate.dose(tmp.conc)
  rownames(tmp.dose) <- NULL
  tmp.dose.analyte <- generate.dose(tmp.conc.analyte)
  tmp.dose.study <- generate.dose(tmp.conc.study)
  tmp.dose.analyte.study <- generate.dose(tmp.conc.analyte.study)

  # Data exists
  expect_error(PKNCAdose(data.frame()),
               regexp="data must have at least one row.",
               info="PKNCAconc requires data")

  # Variables present
  expect_error(PKNCAdose(tmp.dose, formula=dosea~time|treatment+ID),
               regexp="The left side formula must be a variable in the data, empty, or '.'.",
               info="All formula parameters must be in the data (LHS)")
  expect_error(PKNCAdose(tmp.dose, formula=dose~timea|treatment+ID),
               regexp="The right side formula must be a variable in the data or '.'.",
               info="All formula parameters must be in the data (RHS)")
  expect_error(PKNCAdose(tmp.dose, formula=dose~time|treatmenta+ID),
               regexp="All of the variables in the groups must be in the data",
               info="All formula parameters must be in the data (groups)")
  expect_error(PKNCAdose(tmp.dose, formula=dosea~time|treatment/ID),
               regexp="formula for PKNCAdose may not include a slash",
               info="The formula string must not contain any slashes")

  # Number of variables
  expect_error(PKNCAdose(tmp.dose, formula=dose+ID~time|treatment+ID),
               regexp="The left side of the formula must have zero or one variable",
               info="The right number of parameters in the formula (LHS)")
  expect_error(PKNCAdose(tmp.dose, formula=dose~time+ID|treatment+ID),
               regexp="The right side of the formula \\(excluding groups\\) must have exactly one variable",
               info="The right number of parameters in the formula (RHS)")
  expect_error(PKNCAdose(tmp.dose, formula=~time+ID|treatment+ID),
               regexp="The right side of the formula \\(excluding groups\\) must have exactly one variable",
               info="The right number of parameters in the formula (RHS)")

  # Accept "." on either side of the ~
  expect_equal(
    PKNCAdose(tmp.dose, formula=.~time|treatment+ID),
    structure(
      list(
        data =
          cbind(
            tmp.dose,
            data.frame(
              exclude = NA_character_,
              route = "extravascular",
              duration = 0,
              stringsAsFactors = FALSE
            )
          ),
        formula = . ~ time | treatment + ID,
        columns =
          list(
            dose = character(),
            time = "time",
            groups =
              list(
                group_vars = c("treatment", "ID"),
                group_analyte = character()
              ),
            exclude = "exclude",
            route = "route",
            duration = "duration"
          ),
        units = list()
      ),
      class = c("PKNCAdose", "list")
    ),
    info="PKNCAdose accepts . on the left side of the formula"
  )
  expect_equal(
    PKNCAdose(tmp.dose, formula = dose~.|treatment+ID),
    structure(
      list(
        data =
          cbind(
            tmp.dose,
            data.frame(
              exclude = NA_character_,
              route = "extravascular",
              duration = 0,
              stringsAsFactors = FALSE
            )
          ),
        formula = dose ~ . | treatment + ID,
        columns =
          list(
            dose = "dose",
            time = character(),
            groups =
              list(
                group_vars = c("treatment", "ID"),
                group_analyte = character()
              ),
            exclude = "exclude",
            route = "route",
            duration = "duration"
          ),
        units = list()
      ),
      class = c("PKNCAdose", "list")
    ),
    info = "PKNCAdose accepts . on the right side of the formula"
  )

  tmp.dose.na <- tmp.dose
  tmp.dose.na$time[1] <- NA
  expect_error(PKNCAdose(tmp.dose.na, formula=dose~time|treatment+ID),
               regexp="Some but not all values are missing for the independent variable",
               info="Dose time must either all or none be NA.")
  # Keys must be unique
  bad.dose.analyte <- unique(tmp.conc.analyte[,c("treatment", "ID", "analyte")])
  bad.dose.analyte$dose <- 1
  bad.dose.analyte$time <- 0
  expect_error(
    PKNCAdose(bad.dose.analyte, formula=dose~time|treatment+ID),
    regexp="Rows that are not unique per group and time"
  )

  expect_equal(
    PKNCAdose(
      tmp.dose,
      formula=dose~time|treatment+ID
    ),
    PKNCAdose(
      dplyr::as_tibble(tmp.dose),
      formula=dose~time|treatment+ID
    ),
    info="tibble and data.frame classes both work and create identical objects"
  )
})

test_that("PKNCAdose without a data.frame as input", {
  tmp <- structure(list(), class="foo")
  captured_message <- tryCatch(as.data.frame(tmp), error=function(e) e$message)
  expect_error(PKNCAdose(tmp, formula=.~time|treatment+ID),
               regexp=captured_message,
               fixed=TRUE,
               info="PKNCAdose tries to make arbitrary data into a data.frame")
})

test_that("PKNCAdose model.frame", {
  tmp.conc <- generate.conc(nsub=5, ntreat=2, time.points=0:24)
  tmp_conc_single <- generate.conc(nsub=1, ntreat=1, time.points=0:24)
  tmp.conc.analyte <- generate.conc(nsub=5, ntreat=2, time.points=0:24,
                                    nanalytes=2)
  tmp.conc.study <- generate.conc(nsub=5, ntreat=2, time.points=0:24,
                                  nstudies=2)
  tmp.conc.analyte.study <- generate.conc(nsub=5, ntreat=2, time.points=0:24,
                                          nanalytes=2, nstudies=2)
  tmp.dose <- generate.dose(tmp.conc)
  tmp_dose_single <- generate.dose(tmp_conc_single)

  mydose1 <- PKNCAdose(formula=dose~time|treatment+ID, data=tmp.dose)
  expect_equal(getDepVar.PKNCAdose(mydose1),
               rep(1:2, each=5),
               info="getDepVar.PKNCAdose works with two-sided formula")
  expect_equal(getIndepVar.PKNCAdose(mydose1),
               rep(0, 10),
               info="getIndepVar.PKNCAdose works with two-sided formula")
  expect_equal(model.frame.PKNCAdose(mydose1),
               data.frame("getDepVar.PKNCAdose(formula)"=rep(1:2, each=5),
                          "getIndepVar.PKNCAdose(formula)"=0,
                          treatment=rep(c("Trt 1", "Trt 2"), each=5),
                          ID=rep(1:5, 2),
                          stringsAsFactors=FALSE),
               ignore_attr=TRUE,
               info="model.frame.PKNCAdose works with two-sided formula")

  mydose2 <- PKNCAdose(formula=~time|treatment+ID, data=tmp.dose)
  expect_equal(getDepVar.PKNCAdose(mydose2),
               rep(NA_integer_, 10),
               info="getDepVar.PKNCAdose works with one-sided formula")
  expect_equal(getIndepVar.PKNCAdose(mydose2),
               rep(0, 10),
               info="getIndepVar.PKNCAdose works with one-sided formula")
  expect_equal(model.frame.PKNCAdose(mydose2),
               data.frame("getDepVar.PKNCAdose(formula)"=NA_integer_,
                          "getIndepVar.PKNCAdose(formula)"=0,
                          treatment=rep(c("Trt 1", "Trt 2"), each=5),
                          ID=rep(1:5, 2),
                          stringsAsFactors=FALSE),
               ignore_attr=TRUE,
               info="model.frame.PKNCAdose works with one-sided formula")

  mydose3 <- PKNCAdose(formula=.~time|treatment+ID, data=tmp.dose)
  expect_equal(getDepVar.PKNCAdose(mydose3),
               rep(NA_integer_, 10),
               info="getDepVar.PKNCAdose works with one-sided formula ('.' on LHS)")
  expect_equal(getIndepVar.PKNCAdose(mydose3),
               rep(0, 10),
               info="getIndepVar.PKNCAdose works with one-sided formula ('.' on LHS)")
  expect_equal(model.frame.PKNCAdose(mydose3),
               data.frame("getDepVar.PKNCAdose(formula)"=NA_integer_,
                          "getIndepVar.PKNCAdose(formula)"=0,
                          treatment=rep(c("Trt 1", "Trt 2"), each=5),
                          ID=rep(1:5, 2),
                          stringsAsFactors=FALSE),
               ignore_attr=TRUE,
               info="model.frame.PKNCAdose works with one-sided formula ('.' on LHS)")

  mydose4 <- PKNCAdose(formula=dose~.|treatment+ID, data=tmp.dose)
  expect_equal(getDepVar.PKNCAdose(mydose4),
               rep(1:2, each=5),
               info="getDepVar.PKNCAdose works with one-sided formula ('.' on RHS)")
  expect_equal(getIndepVar.PKNCAdose(mydose4),
               rep(NA_integer_, 10),
               info="getIndepVar.PKNCAdose works with one-sided formula ('.' on RHS)")
  expect_equal(model.frame.PKNCAdose(mydose4),
               data.frame("getDepVar.PKNCAdose(formula)"=rep(1:2, each=5),
                          "getIndepVar.PKNCAdose(formula)"=NA_integer_,
                          treatment=rep(c("Trt 1", "Trt 2"), each=5),
                          ID=rep(1:5, 2),
                          stringsAsFactors=FALSE),
               ignore_attr=TRUE,
               info="model.frame.PKNCAdose works with one-sided formula ('.' on RHS)")

  # You can't give multiple rows per group if you don't give time.
  expect_error(PKNCAdose(formula=dose~.|treatment+ID, data=rbind(tmp.dose, tmp.dose)),
               regexp="Rows that are not unique per group and time.*found within dosing data",
               info="Dosing must have unique values with time and group")

  expect_equal(
    group_vars.PKNCAdose(PKNCAdose(tmp.dose, formula=dose~time|treatment+ID)),
    c("treatment", "ID")
  )
  expect_equal(
    group_vars.PKNCAdose(PKNCAdose(tmp_dose_single, formula=dose~time)),
    character(0),
    info="Ungrouped data works with group_vars"
  )
})

test_that("print.PKNCAdose", {
  tmp.conc <- generate.conc(nsub=5, ntreat=2, time.points=0:24)
  tmp.dose <- generate.dose(tmp.conc)
  mydose <- PKNCAdose(tmp.dose, formula=dose~time|treatment+ID)
  tmp.conc.nogroup <- generate.conc(nsub=1, ntreat=1, time.points=0:24)
  tmp.dose.nogroup <- generate.dose(tmp.conc.nogroup)
  mydose.nogroup <- PKNCAdose(tmp.dose.nogroup, formula=dose~time)

  expect_output(print(mydose),
                regexp="Formula for dosing:
 dose ~ time | treatment + ID
Nominal time column is not specified.

First 6 rows of dosing data:
 treatment ID dose time exclude         route duration
     Trt 1  1    1    0    <NA> extravascular        0
     Trt 1  2    1    0    <NA> extravascular        0
     Trt 1  3    1    0    <NA> extravascular        0
     Trt 1  4    1    0    <NA> extravascular        0
     Trt 1  5    1    0    <NA> extravascular        0
     Trt 2  1    2    0    <NA> extravascular        0",
                fixed=TRUE,
                info="Generic print.PKNCAdose works")
  expect_output(print(mydose.nogroup),
                regexp="Formula for dosing:
 dose ~ time
Nominal time column is not specified.

Data for dosing:
 treatment ID dose time exclude         route duration
     Trt 1  1    1    0    <NA> extravascular        0",
                fixed=TRUE,
                info="Generic print.PKNCAdose works with no groups")

  expect_output(print(mydose, n=-5),
                regexp="Formula for dosing:
 dose ~ time | treatment + ID
Nominal time column is not specified.

First 5 rows of dosing data:
 treatment ID dose time exclude         route duration
     Trt 1  1    1    0    <NA> extravascular        0
     Trt 1  2    1    0    <NA> extravascular        0
     Trt 1  3    1    0    <NA> extravascular        0
     Trt 1  4    1    0    <NA> extravascular        0
     Trt 1  5    1    0    <NA> extravascular        0",
                fixed=TRUE,
                info="Generic print.PKNCAdose works")

  expect_output(print(mydose, summarize=TRUE),
                regexp="Formula for dosing:
 dose ~ time | treatment + ID
Nominal time column is not specified.

Number unique entries in each group:
 treatment ID
         2  5",
                fixed=TRUE,
                info="Summary print.PKNCAdose works")

  expect_output(
    print(mydose.nogroup, summarize=TRUE),
    regexp="No groups"
  )
})

test_that("PKNCAdose with exclusions", {
  tmp.conc <- generate.conc(nsub=2, ntreat=2, time.points=0:24)
  tmp.dose <- generate.dose(tmp.conc)
  tmp.dose$excl <- NA_character_
  mydose <- PKNCAdose(tmp.dose, formula=dose~time|treatment+ID, exclude="excl")
  expect_equal(
    mydose,
    structure(
      list(
        data =
          cbind(
            tmp.dose,
            data.frame(
              route = "extravascular",
              duration = 0,
              stringsAsFactors = FALSE
            )
          ),
        formula = dose~time|treatment+ID,
        columns =
          list(
            dose = "dose",
            time = "time",
            groups =
              list(
                group_vars = c("treatment", "ID"),
                group_analyte = character()
              ),
            exclude = "excl",
            route = "route",
            duration = "duration"
          ),
        units = list()
      ),
      class=c("PKNCAdose", "list")
    )
  )
})

test_that("PKNCAdose route and duration", {
  tmp.conc <- generate.conc(nsub=2, ntreat=2, time.points=0:24)
  tmp.dose <- generate.dose(tmp.conc)
  expect_equal(PKNCAdose(tmp.dose, formula=dose~time|treatment+ID),
               PKNCAdose(tmp.dose, formula=dose~time|treatment+ID, route="extravascular"),
               info="route is assumed as extravascular")
  expect_equal(PKNCAdose(tmp.dose, formula=dose~time|treatment+ID),
               PKNCAdose(tmp.dose, formula=dose~time|treatment+ID, duration=0),
               info="duration is assumed as 0")
  expect_equal(PKNCAdose(tmp.dose, formula=dose~time|treatment+ID),
               PKNCAdose(tmp.dose, formula=dose~time|treatment+ID, duration=0, route="extravascular"),
               info="route and duration are correctly assumed")
  dose.iv <- PKNCAdose(tmp.dose, formula=dose~time|treatment+ID, duration=0, route="intravascular")
  dose.ev <- PKNCAdose(tmp.dose, formula=dose~time|treatment+ID, duration=0, route="extravascular")
  expect_equal(dose.iv,
               {
                 tmp <- dose.ev
                 tmp$data$route <- "intravascular"
                 tmp
               },
               info="Intravascular route works")
  expect_error(PKNCAdose(tmp.dose, formula=dose~time|treatment+ID, duration=0, route="foo"),
               regexp="route must have values of either 'extravascular' or 'intravascular'.  Please set to one of those values and retry.",
               info="route must be an accepted value")
  tmp.dose.iv <- tmp.dose
  tmp.dose.iv$route <- "intravascular"
  # Note that the column names are in a different order when specified
  # in the input data.frame or not.
  expect_equal(
    {
      tmp <- PKNCAdose(tmp.dose, formula=dose~time|treatment+ID, duration=0, route="intravascular")
      tmp$data <- tmp$data[,sort(names(tmp$data))]
      tmp
    },
    {
      tmp <- PKNCAdose(tmp.dose.iv, formula=dose~time|treatment+ID, duration=0, route="route")
      tmp$data <- tmp$data[,sort(names(tmp$data))]
      tmp
    },
    ignore_attr=TRUE
  )

})

test_that("time.nominal within PKNCAdose", {
  tmp.conc <- generate.conc(nsub=5, ntreat=2, time.points=0:24)
  tmp.dose <- generate.dose(tmp.conc)
  tmp.dose$nom_time <- tmp.dose$time
  rownames(tmp.dose) <- NULL
  expect_equal(
    PKNCAdose(tmp.dose, formula=dose~time|treatment+ID,
              time.nominal="nom_time"),
    structure(list(
      data=
        cbind(
          tmp.dose,
          data.frame(
            exclude=NA_character_,
            route="extravascular",
            duration=0,
            stringsAsFactors=FALSE
          )
        ),
      formula = dose ~ time | treatment + ID,
      columns =
        list(
          dose = "dose",
          time = "time",
          groups = list(group_vars = c("treatment", "ID"), group_analyte = character()),
          exclude = "exclude",
          route = "route",
          duration = "duration",
          time.nominal = "nom_time"
        ),
      units = list()
    ),
    class = c("PKNCAdose", "list")),
    info="PKNCAdose accepts time.nominal"
  )
  expect_error(PKNCAdose(tmp.dose, formula=dose~time|treatment+ID,
                         time.nominal="foo"),
               regexp="time.nominal, if given, must be a column name in the input data",
               info="PKNCAdose time.nominal must be a column in the data")
  # Test printing
  dose_with_nom_time <- PKNCAdose(tmp.dose, formula=dose~time|treatment+ID,
                                  time.nominal="nom_time")
  expect_output(print(dose_with_nom_time),
                regexp="Formula for dosing:
 dose ~ time | treatment + ID
Nominal time column is: nom_time

First 6 rows of dosing data:
 treatment ID dose time nom_time exclude         route duration
     Trt 1  1    1    0        0    <NA> extravascular        0
     Trt 1  2    1    0        0    <NA> extravascular        0
     Trt 1  3    1    0        0    <NA> extravascular        0
     Trt 1  4    1    0        0    <NA> extravascular        0
     Trt 1  5    1    0        0    <NA> extravascular        0
     Trt 2  1    2    0        0    <NA> extravascular        0",
                fixed=TRUE,
                info="Generic print.PKNCAdose works")
})

test_that("setDuration", {
  tmp.conc <- generate.conc(nsub=5, ntreat=2, time.points=0:24)
  tmp.dose <- generate.dose(tmp.conc)
  tmp.dose$nom_time <- tmp.dose$time
  rownames(tmp.dose) <- NULL
  mydose <- PKNCAdose(tmp.dose, formula=dose~time|treatment+ID)
  expect_message(
    expect_equal(
      setDuration(mydose),
      mydose,
      info="No changes with no arguments"
    ),
    class = "pknca_foundcolumn_duration"
  )
  expect_error(setDuration(mydose, duration="foo", rate="bar"),
               regexp="Both duration and rate cannot be given at the same time",
               fixed=TRUE,
               info="Cannot give both duration and rate")
  expect_message(expect_error(
    setDuration(mydose, duration="foobar"),
    regexp="duration must be numeric without missing (NA) or infinite values, and all values must be >= 0",
    fixed=TRUE,
    info="Cannot give both duration as non-numeric"),
    class = "pknca_foundcolumn_duration"
  )

  duration_example <- suppressMessages(setDuration(mydose, rate=2))
  expect_true(all(mydose$data$duration == 0))
  expect_equal(
    duration_example$data$duration,
    duration_example$data$dose/2
  )
  expect_equal(duration_example$columns$duration, "duration")
})

test_that("Test uniqueness after excluding rows (#298)", {
  repeated_with_exclusion <-
    data.frame(
      dose = 1,
      time = c(0, 0),
      id = 1,
      exclude = c(NA, "duplicate")
    )
  expect_error(
    PKNCAdose(repeated_with_exclusion, formula=dose~time|id),
    regexp="Rows that are not unique per group and time.*dosing"
  )
  expect_s3_class(
    PKNCAdose(repeated_with_exclusion, formula=dose~time|id, exclude = "exclude"),
    class = "PKNCAdose"
  )
  repeated_with_exclusion_firstrow <-
    data.frame(
      dose = 1,
      time = c(0, 0),
      id = 1,
      exclude = c("duplicate", NA)
    )
  expect_s3_class(
    PKNCAdose(repeated_with_exclusion_firstrow, formula=dose~time|id, exclude = "exclude"),
    class = "PKNCAdose"
  )

  repeated_with_exclusion_nogroup <-
    data.frame(
      dose = 1,
      time = c(0, 0),
      id = 1,
      exclude = c("duplicate", NA)
    )
  expect_s3_class(
    PKNCAdose(repeated_with_exclusion_firstrow, formula=dose~., exclude = "exclude"),
    class = "PKNCAdose"
  )
  expect_s3_class(
    PKNCAdose(repeated_with_exclusion_firstrow, formula=dose~.|id, exclude = "exclude"),
    class = "PKNCAdose"
  )
  repeated_with_exclusion_row2 <-
    data.frame(
      dose = 1,
      time = c(0, 0),
      id = 1,
      exclude = c(NA, "duplicate")
    )
  expect_s3_class(
    PKNCAdose(repeated_with_exclusion_row2, formula=dose~.|id, exclude = "exclude"),
    class = "PKNCAdose"
  )
})

test_that("PKNCAdose units (#336)", {
  d <- data.frame(dose = 1, time = 0, doseu_x = "A")

  # No units
  o_dose <- PKNCAdose(data = d, dose~time)
  expect_equal(o_dose$units, list())
  expect_null(o_dose$columns$doseu)

  # Each unit column individually
  o_dose <- PKNCAdose(data = d, dose~time, doseu = "doseu_x")
  expect_equal(o_dose$units, list())
  expect_equal(o_dose$columns$doseu, structure("doseu_x", unit_type = "column"))

  # Each unit as a value, not a column
  o_dose <- PKNCAdose(data = d, dose~time, doseu = "doseu_y")
  expect_equal(o_dose$units, list(doseu = structure("doseu_y", unit_type = "value")))
  expect_null(o_dose$columns$doseu)

  # Preferred units
  expect_error(
    PKNCAdose(data = d, dose~time, doseu_pref = "doseu_z"),
    regexp = "Preferred units may not be set unless original units are set: doseu_pref"
  )
  o_dose <- PKNCAdose(data = d, dose~time, doseu = "doseu_y", doseu_pref = "doseu_z")
  expect_equal(
    o_dose$units,
    list(
      doseu = structure("doseu_y", unit_type = "value"),
      doseu_pref = structure("doseu_z", unit_type = "value")
    )
  )
  expect_null(o_dose$columns$doseu)

  o_dose <- PKNCAdose(data = d, dose~time, doseu = "doseu_x", doseu_pref = "doseu_z")
  expect_equal(o_dose$units, list(doseu_pref = structure("doseu_z", unit_type = "value")))
  expect_equal(
    o_dose$columns$doseu,
    structure("doseu_x", unit_type = "column")
  )
})

# Internal methods for AUC integration

aucintegrate_linear <- function(conc.1, conc.2, time.1, time.2) {
  (time.2-time.1) * (conc.2+conc.1)/2
}

aucintegrate_log <- function(conc.1, conc.2, time.1, time.2) {
  (time.2-time.1) * (conc.2-conc.1)/log(conc.2/conc.1)
}

aucintegrate_inf <- function(clast, tlast, lambda.z) {
  clast/lambda.z
}

aumcintegrate_linear <- function(conc.1, conc.2, time.1, time.2) {
  (time.2-time.1) * (conc.2*time.2+conc.1*time.1)/2
}

aumcintegrate_log <- function(conc.1, conc.2, time.1, time.2) {
  ((time.2-time.1) * (conc.2*time.2-conc.1*time.1) / log(conc.2/conc.1)-
     (time.2-time.1)^2 * (conc.2-conc.1) / (log(conc.2/conc.1)^2))
}

aumcintegrate_inf <- function(conc.last, time.last, lambda.z) {
  (conc.last*time.last/lambda.z) + conc.last/(lambda.z^2)
}

#' Interpolate or extrapolate concentrations using the provided method
#'
#' @param conc_1,conc_2 The concentration at time1 and time2
#' @param time_1,time_2 The time value associated with conc1 and conc2
#' @param time_out Time when interpolation is requested
#' @param tlast The time of the last concentration above the lower limit of
#'   quantification (LOQ)
#' @returns The interpolated or extrapolated value using the correct method
#' @keywords Internal
#' @name interp_extrap_conc_method
NULL

#' @rdname interp_extrap_conc_method
interpolate_conc_linear <- function(conc_1, conc_2, time_1, time_2, time_out) {
  conc_1+(time_out-time_1)/(time_2-time_1)*(conc_2-conc_1)
}

#' @rdname interp_extrap_conc_method
interpolate_conc_log <- function(conc_1, conc_2, time_1, time_2, time_out) {
  exp(
    log(conc_1)+
      (time_out-time_1)/(time_2-time_1)*(log(conc_2)-log(conc_1))
  )
}

#' @inheritParams assert_lambdaz
#' @param clast The concentration at the last time above the lower LOQ
#' @rdname interp_extrap_conc_method
extrapolate_conc_lambdaz <- function(clast, lambda.z, tlast, time_out) {
  clast*exp(-lambda.z*(time_out - tlast))
}

#' Choose how to interpolate, extrapolate, or integrate data in each
#' concentration interval
#'
#' @inheritParams assert_conc_time
#' @inheritParams PKNCA.choose.option
#' @inheritParams assert_aucmethod
#' @param auc.type The type of AUC to compute.  Choices are 'AUCinf', 'AUClast',
#'   and 'AUCall'.
#' @param tlast Time of last concentration above the limit of quantification
#'   (will be calculated, if not provided)
#' @keywords Internal
#' @returns A character vector of methods for interpolation/extrapolation
#'   methods that is the same length as `conc` which indicates how to
#'   interpolate/integrate between each of the concentrations (all but the last
#'   value in the vector) and how to extrapolate after `tlast` (the last item in
#'   the vector).  Possible values in the vector are: 'zero', 'linear', 'log',
#'   and 'extrap_log'
choose_interval_method <- function(conc, time, tlast, method, auc.type, options) {
  # Input checking
  stopifnot(is.numeric(conc))
  stopifnot(is.numeric(time))
  stopifnot(!any(is.na(time)))
  stopifnot(!any(is.na(conc)))
  stopifnot(length(conc) == length(time))
  assert_aucmethod(method)
  stopifnot(length(auc.type) == 1)
  stopifnot(auc.type %in% c("AUCinf", "AUClast", "AUCall"))

  if (missing(tlast)) {
    tlast <- pk.calc.tlast(conc, time, check=FALSE)
  } else {
    stopifnot(is.numeric(tlast))
    stopifnot(length(tlast) == 1)
  }

  # Where is tlast in the data?
  idx_tlast <- which(time == tlast)

  ret <- rep(NA_character_, length(conc))
  # Handle all interpolation
  idx_1 <- seq(1, length(conc) - 1)
  idx_2 <- idx_1 + 1
  mask_zero <- conc[idx_1] == 0 & conc[idx_2] == 0
  if (all(conc %in% 0)) {
    ret[] <- "zero"
    # short circuit other options
    return(ret)
  } else if (method == "linear") {
    ret[seq_len(idx_tlast - 1)] <- "linear"
  } else if (method == "lin up/log down") {
    mask_down <- conc[idx_2] < conc[idx_1] & conc[idx_2] != 0
    mask_up <- !(mask_down | mask_zero)
    ret[c(mask_down, FALSE)] <- "log"
    ret[c(mask_up, FALSE)] <- "linear"
  } else if (method == "lin-log") {
    # We only need tmax for lin-log, and it is always recalculated to prevent
    # potential issues where tmax may be relative to the start of an interval vs
    # relative to absolute time.
    tmax <- pk.calc.tmax(conc = conc, time = time, options = options)
    mask_pre_tmax <- time[idx_2] <= tmax
    mask_post_tmax <- !mask_pre_tmax
    mask_zero_start_end <- conc[idx_1] %in% 0 | conc[idx_2] %in% 0
    mask_linear <- mask_pre_tmax | mask_zero_start_end
    mask_log <- mask_post_tmax & !mask_zero_start_end
    ret[c(mask_linear, FALSE)] <- "linear"
    ret[c(mask_log, FALSE)] <- "log"
  } else {
    stop("Unknown integration method, please report a bug: ", method) # nocov
  }
  ret[c(mask_zero, FALSE)] <- "zero"
  # What happens after tlast?
  if (idx_tlast < (length(ret) - 1)) {
    ret[seq(idx_tlast + 1, length(ret) - 1)] <- "zero"
  }
  # what happens at tlast?
  if (idx_tlast < length(conc)) {
    if (auc.type == "AUCall") {
      ret[idx_tlast] <- "linear"
    } else {
      ret[idx_tlast] <- "zero"
    }
  }

  # Handle extrapolation
  if (auc.type == "AUCinf") {
    ret[length(ret)] <- "extrap_log"
  } else {
    # Extrapolation is zero, except for AUCinf
    ret[length(ret)] <- "zero"
  }
  ret
}

#' Support function for AUC integration
#'
#' @inheritParams choose_interval_method
#' @inheritParams pk.calc.auxc
#' @param clast The last concentration above the limit of quantification
#' @param interval_method The method for integrating each interval of `conc`
#' @keywords Internal
auc_integrate <- function(conc, time, clast, tlast, lambda.z, interval_method, fun_linear, fun_log, fun_inf) {
  assert_lambdaz(lambda.z = lambda.z)
  interval_method_within <- interval_method[-length(interval_method)]
  interval_method_extrap <- interval_method[length(interval_method)]
  idx_1 <- seq_len(length(conc) - 1)
  idx_1_linear <- idx_1[interval_method_within == "linear"]
  idx_1_log <- idx_1[interval_method_within == "log"]

  ret <-
    c(
      fun_linear(conc[idx_1_linear], conc[idx_1_linear + 1],
                 time[idx_1_linear], time[idx_1_linear + 1]),
      fun_log(conc[idx_1_log], conc[idx_1_log + 1],
              time[idx_1_log], time[idx_1_log + 1])
    )

  if (interval_method_extrap %in% "extrap_log") {
    # Whether AUCinf,obs or AUCinf,pred is calculated depends on if clast,obs
    # or clast,pred is passed in.
    ret[length(ret)+1] <- fun_inf(clast, tlast, lambda.z)
  } else if (interval_method_extrap != "zero") {
    stop("Invalid interval_method_extrap, please report a bug: ", interval_method_extrap) # nocov
  }
  ret <- sum(ret)
  ret
}

test_that("choose_interval_method", {
  # All zeros ####
  expect_equal(
    choose_interval_method(conc = rep(0, 2), time = 1:2, method = "linear", auc.type = "AUCinf"),
    c("zero", "zero")
  )
  # Increasing, no zeros, AUCinf ####
  expect_equal(
    choose_interval_method(conc = 1:2, time = 1:2, method = "linear", auc.type = "AUCinf"),
    c("linear", "extrap_log")
  )
  expect_equal(
    choose_interval_method(conc = 1:3, time = 1:3, method = "linear", auc.type = "AUCinf"),
    c("linear", "linear", "extrap_log")
  )
  expect_equal(
    choose_interval_method(conc = 1:2, time = 1:2, method = "lin up/log down", auc.type = "AUCinf"),
    c("linear", "extrap_log")
  )
  expect_equal(
    choose_interval_method(conc = 1:3, time = 1:3, method = "lin up/log down", auc.type = "AUCinf"),
    c("linear", "linear", "extrap_log")
  )
  # Increasing, no zeros, AUCall ####
  expect_equal(
    choose_interval_method(conc = 1:2, time = 1:2, method = "linear", auc.type = "AUCall"),
    c("linear", "zero")
  )
  expect_equal(
    choose_interval_method(conc = 1:3, time = 1:3, method = "linear", auc.type = "AUCall"),
    c("linear", "linear", "zero")
  )
  # Decreasing, no zeros ####
  expect_equal(
    choose_interval_method(conc = 2:1, time = 1:2, method = "linear", auc.type = "AUCinf"),
    c("linear", "extrap_log")
  )
  expect_equal(
    choose_interval_method(conc = 3:1, time = 1:3, method = "linear", auc.type = "AUCinf"),
    c("linear", "linear", "extrap_log")
  )
  expect_equal(
    choose_interval_method(conc = 2:1, time = 1:2, method = "lin up/log down", auc.type = "AUCinf"),
    c("log", "extrap_log")
  )
  expect_equal(
    choose_interval_method(conc = 3:1, time = 1:3, method = "lin up/log down", auc.type = "AUCinf"),
    c("log", "log", "extrap_log")
  )
  # Increasing, one initial zero ####
  expect_equal(
    choose_interval_method(conc = c(0, 1:2), time = 1:3, method = "linear", auc.type = "AUCinf"),
    c("linear", "linear", "extrap_log")
  )
  expect_equal(
    choose_interval_method(conc = c(0, 1:2), time = 1:3, method = "lin up/log down", auc.type = "AUCinf"),
    c("linear", "linear", "extrap_log")
  )
  # Increasing, two initial zeros ####
  expect_equal(
    choose_interval_method(conc = c(0, 0, 1:2), time = 1:4, method = "linear", auc.type = "AUCinf"),
    c("zero", "linear", "linear", "extrap_log")
  )
  expect_equal(
    choose_interval_method(conc = c(0, 0, 1:2), time = 1:4, method = "lin up/log down", auc.type = "AUCinf"),
    c("zero", "linear", "linear", "extrap_log")
  )
  # Increasing and decreasing, two initial zeros ####
  expect_equal(
    choose_interval_method(conc = c(0, 0, 1:2, 1), time = 1:5, method = "linear", auc.type = "AUCinf"),
    c("zero", "linear", "linear", "linear", "extrap_log")
  )
  expect_equal(
    choose_interval_method(conc = c(0, 0, 1:2, 1), time = 1:5, method = "lin up/log down", auc.type = "AUCinf"),
    c("zero", "linear", "linear", "log", "extrap_log")
  )
  # Increasing and decreasing, two initial zeros, one final zero, AUCinf ####
  expect_equal(
    choose_interval_method(conc = c(0, 0, 1:2, 1, 0), time = 1:6, method = "linear", auc.type = "AUCinf"),
    c("zero", "linear", "linear", "linear", "zero", "extrap_log")
  )
  expect_equal(
    choose_interval_method(conc = c(0, 0, 1:2, 1, 0), time = 1:6, method = "lin up/log down", auc.type = "AUCinf"),
    c("zero", "linear", "linear", "log", "zero", "extrap_log")
  )
  # Increasing and decreasing, two initial zeros, one final zero, AUClast ####
  expect_equal(
    choose_interval_method(conc = c(0, 0, 1:2, 1, 0), time = 1:6, method = "linear", auc.type = "AUClast"),
    c("zero", "linear", "linear", "linear", "zero", "zero")
  )
  expect_equal(
    choose_interval_method(conc = c(0, 0, 1:2, 1, 0), time = 1:6, method = "lin up/log down", auc.type = "AUClast"),
    c("zero", "linear", "linear", "log", "zero", "zero")
  )
  # Increasing and decreasing, two initial zeros, one final zero, AUCall ####
  expect_equal(
    choose_interval_method(conc = c(0, 0, 1:2, 1, 0), time = 1:6, method = "linear", auc.type = "AUCall"),
    c("zero", "linear", "linear", "linear", "linear", "zero")
  )
  expect_equal(
    choose_interval_method(conc = c(0, 0, 1:2, 1, 0), time = 1:6, method = "lin up/log down", auc.type = "AUCall"),
    c("zero", "linear", "linear", "log", "linear", "zero")
  )
  # Increasing and decreasing, two initial zeros, two final zeros, AUCinf ####
  expect_equal(
    choose_interval_method(conc = c(0, 0, 1:2, 1, 0, 0), time = 1:7, method = "linear", auc.type = "AUCinf"),
    c("zero", "linear", "linear", "linear", "zero", "zero", "extrap_log")
  )
  expect_equal(
    choose_interval_method(conc = c(0, 0, 1:2, 1, 0, 0), time = 1:7, method = "lin up/log down", auc.type = "AUCinf"),
    c("zero", "linear", "linear", "log", "zero", "zero", "extrap_log")
  )

  # Increasing and decreasing, two initial zeros, one middle zero, two final zeros, AUCinf ####
  expect_equal(
    choose_interval_method(conc = c(0, 0, 1:2, 0, 2:1, 0, 0), time = 1:9, method = "linear", auc.type = "AUCinf"),
    c("zero", "linear", "linear", "linear", "linear", "linear", "zero", "zero", "extrap_log")
  )
  expect_equal(
    choose_interval_method(conc = c(0, 0, 1:2, 0, 2:1, 0, 0), time = 1:9, method = "lin up/log down", auc.type = "AUCinf"),
    c("zero", "linear", "linear", "linear", "linear", "log", "zero", "zero", "extrap_log")
  )
  expect_equal(
    choose_interval_method(conc = c(0, 0, 1:2, 0, 2:1, 0, 0), time = 1:9, method = "lin-log", auc.type = "AUCinf", options = list()),
    c("zero", "linear", "linear", "linear", "linear", "log", "zero", "zero", "extrap_log")
  )
  # Increasing and decreasing, two initial zeros, one middle zero, two final zeros, AUCall ####
  expect_equal(
    choose_interval_method(conc = c(0, 0, 1:2, 0, 2:1, 0, 0), time = 1:9, method = "linear", auc.type = "AUCall"),
    c("zero", "linear", "linear", "linear", "linear", "linear", "linear", "zero", "zero")
  )
  expect_equal(
    choose_interval_method(conc = c(0, 0, 1:2, 0, 2:1, 0, 0), time = 1:9, method = "lin up/log down", auc.type = "AUCall"),
    c("zero", "linear", "linear", "linear", "linear", "log", "linear", "zero", "zero")
  )
  # Increasing and decreasing, two initial zeros, one middle zero, two final zeros, AUClast ####
  expect_equal(
    choose_interval_method(conc = c(0, 0, 1:2, 0, 2:1, 0, 0), time = 1:9, method = "linear", auc.type = "AUClast"),
    c("zero", "linear", "linear", "linear", "linear", "linear", "zero", "zero", "zero")
  )
  expect_equal(
    choose_interval_method(conc = c(0, 0, 1:2, 0, 2:1, 0, 0), time = 1:9, method = "lin up/log down", auc.type = "AUClast"),
    c("zero", "linear", "linear", "linear", "linear", "log", "zero", "zero", "zero")
  )
  # Increasing and decreasing before and after Tmax, two initial zeros, two middle zeros, two final zeros, AUClast ####
  expect_equal(
    choose_interval_method(
      conc = c(0, 0,   1,   1.5,
               1, 2.5, 0,   0,
               2, 1,   1.5, 0,
               0),
      time = 1:13, method = "linear", auc.type = "AUClast"
    ),
    c(
      "zero", "linear", "linear",
      "linear",
      "linear", "linear", "zero",
      "linear",
      "linear", "linear", "zero",
      "zero",
      "zero"
    )
  )
  expect_equal(
    choose_interval_method(
      conc = c(0, 0,   1,   1.5,
               1, 2.5, 0,   0,
               2, 1,   1.5, 0,
               0),
      time = 1:13,
      method = "lin up/log down", auc.type = "AUClast"
    ),
    c(
      "zero", "linear", "linear",
      "log",
      "linear", "linear", "zero",
      "linear",
      "log", "linear", "zero",
      "zero",
      "zero"
    )
  )
  expect_equal(
    choose_interval_method(
      conc = c(0, 0,   1,   1.5,
               1, 2.5, 0,   0,
               2, 1,   1.5, 0,
               0),
      time = 1:13,
      method = "lin-log", auc.type = "AUClast",
      options = list()
    ),
    c(
      "zero", "linear", "linear",
      "linear",
      "linear", "linear", "zero",
      "linear",
      "log", "log", "zero",
      "zero",
      "zero"
    )
  )
  # lin-log respects Tmax options; increasing and decreasing before and after Tmax, two initial zeros, two middle zeros, two final zeros, AUClast ####
  expect_equal(
    choose_interval_method(
      conc = c(0, 0,   1,   2.5,
               1, 2.5, 0,   0,
               2, 1,   1.5, 0,
               0),
      time = 1:13,
      method = "lin-log", auc.type = "AUClast",
      options = list(first.tmax = TRUE)
    ),
    c(
      "zero", "linear", "linear",
      "log",
      "log", "linear", "zero",
      "linear",
      "log", "log", "zero",
      "zero",
      "zero"
    )
  )
  expect_equal(
    choose_interval_method(
      conc = c(0, 0,   1,   2.5,
               1, 2.5, 0,   0,
               2, 1,   1.5, 0,
               0),
      time = 1:13,
      method = "lin-log", auc.type = "AUClast",
      options = list(first.tmax = FALSE)
    ),
    c(
      "zero", "linear", "linear",
      "linear",
      "linear", "linear", "zero",
      "linear",
      "log", "log", "zero",
      "zero",
      "zero"
    )
  )
})

test_that("choose_interval_method expected errors", {
  expect_error(choose_interval_method())
  expect_error(choose_interval_method(conc = "A"))
  expect_error(choose_interval_method(conc = 1, time = "A"))
  expect_error(choose_interval_method(conc = NA_real_, time = 1))
  expect_error(choose_interval_method(conc = 1, time = NA_real_))
  expect_error(choose_interval_method(conc = 1, time = 1:2))
  expect_error(choose_interval_method(conc = 1:2, time = 1))
  expect_error(choose_interval_method(conc = 1:2, time = 1:2))
  expect_error(choose_interval_method(conc = 1:2, time = 1:2, method = c("foo", "bar")))
  expect_error(choose_interval_method(conc = 1:2, time = 1:2, method = "foo"))
  expect_error(choose_interval_method(conc = 1:2, time = 1:2, method = "linear"))
  expect_error(choose_interval_method(conc = 1:2, time = 1:2, method = "linear", auc.type = c("foo", "bar")))
  expect_error(choose_interval_method(conc = 1:2, time = 1:2, method = "linear", auc.type = "foo"))
  expect_error(choose_interval_method(conc = 1:2, time = 1:2, method = "linear", auc.type = "AUCinf", tlast = "A"))
  expect_error(choose_interval_method(conc = 1:2, time = 1:2, method = "linear", auc.type = "AUCinf", tlast = 1:2))
})

#' Compute the half-life and associated parameters
#'
#' The terminal elimination half-life is estimated from the final points in the
#' concentration-time curve using semi-log regression (`log(conc)~time`)
#' with automated selection of the points for calculation (unless
#' `manually.selected.points` is `TRUE`).
#'
#' See the "Half-Life Calculation" vignette for more details on the calculation
#' methods used.
#'
#' @details If `manually.selected.points` is `FALSE` (default), the
#' half-life is calculated by computing the best fit line for all points at or
#' after tmax (based on the value of `allow.tmax.in.half.life`).  The best
#' half-life is chosen by the following rules in order:
#'
#' \itemize{
#'  \item{At least `min.hl.points` points included}
#'  \item{A `lambda.z` > 0 and at the same time the best adjusted r-squared
#'  (within `adj.r.squared.factor`)}
#'  \item{The one with the most points included}
#' }
#'
#' If `manually.selected.points` is `TRUE`, the `conc`
#' and `time` data are used as-is without any form of selection for
#' the best-fit half-life.
#'
#' @inheritParams assert_conc_time
#' @inheritParams choose_interval_method
#' @inheritParams PKNCA.choose.option
#' @inheritParams pk.nca.interval
#' @param tmax Time of maximum concentration (will be calculated and
#'   included in the return data frame if not given)
#' @param tlast Time of last concentration above the limit of
#'   quantification (will be calculated and included in the return data
#'   frame if not given)
#' @param manually.selected.points Have the input points (`conc` and
#'   `time`) been manually selected?  The impact of setting this to
#'   `TRUE` is that no selection for the best points will be done.  When
#'   `TRUE`, this option causes the options of `adj.r.squared.factor`,
#'   `min.hl.points`, and `allow.tmax.in.half.life` to be ignored.
#' @param min.hl.points The minimum number of points that must be
#'   included to calculate the half-life
#' @param adj.r.squared.factor The allowance in adjusted r-squared for
#'   adding another point.
#' @param conc.blq See [clean.conc.blq()]
#' @param conc.na See [clean.conc.na()]
#' @param check Run [assert_conc_time()],
#'   [clean.conc.blq()], and [clean.conc.na()]?
#' @param first.tmax See [pk.calc.tmax()].
#' @param allow.tmax.in.half.life Allow the concentration point for tmax
#'   to be included in the half-life slope calculation.
#' @return A data frame with one row and columns for
#'  \describe{
#'   \item{tmax}{Time of maximum observed concentration (only included
#'     if not given as an input)}
#'   \item{tlast}{Time of last observed concentration above the LOQ (only
#'     included if not given as an input)}
#'   \item{r.squared}{coefficient of determination}
#'   \item{adj.r.squared}{adjusted coefficient of determination}
#'   \item{lambda.z}{elimination rate}
#'   \item{lambda.z.time.first}{first time for half-life calculation}
#'   \item{lambda.z.time.last}{last time for half-life calculation}
#'   \item{lambda.z.n.points}{number of points in half-life calculation}
#'   \item{clast.pred}{Concentration at tlast as predicted by the half-life
#'     line}
#'   \item{half.life}{half-life}
#'   \item{span.ratio}{span ratio [ratio of half-life to time used for
#'     half-life calculation}
#'  }
#' @references
#'
#' Gabrielsson J, Weiner D.  "Section 2.8.4 Strategies for estimation of
#' lambda-z."  Pharmacokinetic & Pharmacodynamic Data Analysis: Concepts
#' and Applications, 4th Edition.  Stockholm, Sweden: Swedish
#' Pharmaceutical Press, 2000.  167-9.
#' @family NCA parameter calculations
#' @export
pk.calc.half.life <- function(conc, time, tmax, tlast,
                              time.dose=NULL,
                              duration.dose=0,
                              manually.selected.points=FALSE,
                              options=list(),
                              min.hl.points=NULL,
                              adj.r.squared.factor=NULL,
                              conc.blq=NULL,
                              conc.na=NULL,
                              first.tmax=NULL,
                              allow.tmax.in.half.life=NULL,
                              check=TRUE) {
  # Check inputs
  min.hl.points <-
    PKNCA.choose.option(
      name="min.hl.points", value=min.hl.points, options=options
    )
  adj.r.squared.factor <-
    PKNCA.choose.option(
      name="adj.r.squared.factor", value=adj.r.squared.factor, options=options
    )
  conc.blq <-
    PKNCA.choose.option(
      name="conc.blq", value=conc.blq, options=options
    )
  conc.na <-
    PKNCA.choose.option(
      name="conc.na", value=conc.na, options=options
    )
  first.tmax <-
    PKNCA.choose.option(
      name="first.tmax", value=first.tmax, options=options
    )
  allow.tmax.in.half.life <-
    PKNCA.choose.option(
      name="allow.tmax.in.half.life", value=allow.tmax.in.half.life, options=options
    )
  if (check) {
    assert_conc_time(conc = conc, time = time)
    data <- clean.conc.blq(conc, time, conc.blq=conc.blq, conc.na=conc.na)
  } else {
    data <- data.frame(conc, time)
  }
  # if (inherits(data$conc, "units")) {
  #   conc_units <- units(data$conc)
  # } else {
  conc_units <- NULL
  #}
  data$log_conc <- log(data$conc)
  # Filter out points with 0 concentration. as.numeric() to handle units objects
  data <- data[as.numeric(data$conc) > 0,]
  if (!is.null(time.dose)) {
    # If multiple doses are within the interval, consider only the one ending the latest
    end.dose <- as.numeric(time.dose) + as.numeric(duration.dose)
    if (any(!is.na(end.dose))) {
      data <- data[as.numeric(data$time) > max(end.dose, na.rm = TRUE), ]
    }
  }
  # Prepare the return values
  ret <- data.frame(
    # Terminal elimination slope
    lambda.z=NA,
    # R-squared of terminal elimination slope
    r.squared=NA,
    # Adjusted r-squared of terminal elimination slope
    adj.r.squared=NA,
    # First time point used in the slope estimation (for plotting later)
    lambda.z.time.first=NA,
    # Last time point used in the slope estimation (for plotting later)
    lambda.z.time.last=NA,
    # Number of points in the half-life estimate
    lambda.z.n.points=NA,
    # Concentration at Tlast predicted by the half-life
    clast.pred=NA,
    # Half-life
    half.life=NA,
    # T1/2 span ratio
    span.ratio=NA)
  ret_replacements <-
    c("lambda.z", "r.squared", "adj.r.squared", "lambda.z.time.first",
      "lambda.z.time.last", "lambda.z.n.points", "clast.pred", "half.life", "span.ratio")
  if (missing(tmax)) {
    ret$tmax <-
      pk.calc.tmax(
        data$conc, data$time, first.tmax=first.tmax, check=FALSE
      )
  } else {
    ret$tmax <- tmax
  }
  if (missing(tlast)) {
    ret$tlast <- pk.calc.tlast(data$conc, data$time, check=FALSE)
  } else {
    ret$tlast <- tlast
  }
  # Data frame to use for computation of half-life
  if (allow.tmax.in.half.life) {
    # as.numeric is for units handling
    dfK <- data[as.numeric(data$time) >= as.numeric(ret$tmax), ]
  } else {
    # as.numeric is for units handling
    dfK <- data[as.numeric(data$time) > as.numeric(ret$tmax), ]
  }
  if (manually.selected.points) {
    if (nrow(data) > 0) {
      fit <- fit_half_life(data=data, tlast=ret$tlast, conc_units=conc_units)
      ret[,ret_replacements] <- fit[,ret_replacements]
      if (ret$half.life <= 0) {
        attr(ret, "exclude") <- "Negative half-life estimated with manually-selected points"
      }
    } else {
      warning("No data to manually fit for half-life (all concentrations may be 0 or excluded)")
      ret <-
        structure(
          ret,
          exclude="No data to manually fit for half-life (all concentrations may be 0 or excluded)"
        )
    }
  } else if (nrow(dfK) >= min.hl.points) {
    # If we have enough data to estimate a slope, then
    half_lives_for_selection <-
      data.frame(
        r.squared=-Inf,
        adj.r.squared=-Inf,
        clast.pred=NA_real_,
        lambda.z=-Inf,
        lambda.z.n.points=NA_integer_,
        lambda.z.time.first=dfK$time,
        lambda.z.time.last=NA_real_,
        log_conc=dfK$log_conc,
        span.ratio=NA_real_,
        half.life=NA_real_
      )
    half_lives_for_selection <-
      half_lives_for_selection[order(-half_lives_for_selection$lambda.z.time.first), ]
    for(i in min.hl.points:nrow(half_lives_for_selection)) {
      # Fit the terminal slopes until the adjusted r-squared value
      # is not improving (or it only gets worse by a small factor).
      fit <-
        fit_half_life(
          data=
            data.frame(
              # pass in the conc so that we can use its units, if applicable
              log_conc=half_lives_for_selection$log_conc[1:i],
              time=half_lives_for_selection$lambda.z.time.first[1:i]
            ),
          tlast=ret$tlast,
          conc_units=conc_units
        )
      half_lives_for_selection[i,names(fit)] <- fit
    }
    # Find the best model
    mask_best <-
      half_lives_for_selection$lambda.z > 0 &
      if (min.hl.points == 2 & nrow(half_lives_for_selection) == 2) {
        rlang::warn(
          message = "2 points used for half-life calculation",
          class = "pknca_halflife_2points"
        )
        TRUE
      } else {
        half_lives_for_selection$adj.r.squared >
          (max(half_lives_for_selection$adj.r.squared, na.rm=TRUE) - adj.r.squared.factor)
      }
    # Missing values are not the best
    mask_best[is.na(mask_best)] <- FALSE
    if (sum(mask_best) > 1) {
      # If more than one models qualify, choose the one with the
      # most points used.
      mask_best <-
        (mask_best &
           half_lives_for_selection$lambda.z.n.points == max(half_lives_for_selection$lambda.z.n.points[mask_best]))
    }
    # If the half-life fit, set all associated parameters
    if (any(mask_best)) {
      # Put in all the computed values
      ret[,ret_replacements] <- half_lives_for_selection[mask_best, ret_replacements]
    }
  } else {
    attr(ret, "exclude") <-
      sprintf(
        "Too few points for half-life calculation (min.hl.points=%g with only %g points)",
        min.hl.points, nrow(dfK)
      )
    rlang::warn(
      message = attr(ret, "exclude"),
      class = "pknca_halflife_too_few_points"
    )
  }
  # Drop the inputs of tmax and tlast, if given.
  if (!missing(tmax))
    ret$tmax <- NULL
  if (!missing(tlast))
    ret$tlast <- NULL
  ret
}

#' Perform the half-life fit given the data.  The function simply fits
#' the data without any validation.  No selection of points or any other
#' components are done.
#'
#' @param data The data to fit.  Must have two columns named "log_conc"
#'   and "time"
#' @param tlast The time of last observed concentration above the limit
#'   of quantification.
#' @param conc_units NULL or the units to set for concentration measures
#' @return A data.frame with one row and columns named "r.squared",
#'   "adj.r.squared", "PROB", "lambda.z", "clast.pred",
#'   "lambda.z.n.points", "half.life", "span.ratio"
#' @seealso [pk.calc.half.life()]
fit_half_life <- function(data, tlast, conc_units) {
  fit <- stats::.lm.fit(x=cbind(1, data$time), y=data$log_conc)
  # unit handling
  # if (inherits(tlast, "units")) {
  #   time_units <- units(tlast)
  # } else if (inherits(tlast, "mixed_units")) {
  #   time_units <- units(units::as_units(tlast))
  # } else {
  time_units <- NULL
  # }
  # if (!is.null(time_units)) {
  #   inverse_time_units <- time_units
  #   inverse_time_units$numerator <- time_units$denominator
  #   inverse_time_units$denominator <- time_units$numerator
  # } else {
  inverse_time_units <- NULL
  # }

  # as.numeric is so that it works for units objects
  r_squared <- 1 - as.numeric(sum(fit$residuals^2))/as.numeric(sum((data$log_conc - mean(data$log_conc))^2))
  clast_pred <- exp(sum(fit$coefficients*c(1, as.numeric(tlast))))
  # if (!is.null(conc_units)) {
  #   clast_pred <- units::set_units(clast_pred, conc_units, mode="standard")
  # }
  lambda_z <- -fit$coefficients[2]
  # if (!is.null(inverse_time_units)) {
  #   lambda_z <- units::set_units(lambda_z, inverse_time_units, mode="standard")
  # }
  ret <-
    data.frame(
      r.squared=r_squared,
      adj.r.squared=adj.r.squared(r_squared, nrow(data)),
      lambda.z=lambda_z,
      clast.pred=clast_pred,
      lambda.z.time.first=min(data$time, na.rm=TRUE),
      lambda.z.time.last=max(data$time, na.rm=TRUE),
      lambda.z.n.points=nrow(data)
    )
  ret$half.life <- log(2)/ret$lambda.z
  ret$span.ratio <- (max(data$time) - min(data$time))/ret$half.life
  ret
}

# Add the column to the interval specification
add.interval.col("half.life",
                 FUN="pk.calc.half.life",
                 values=c(FALSE, TRUE),
                 unit_type="time",
                 pretty_name="Half-life",
                 desc="The (terminal) half-life",
                 depends=c("tmax", "tlast"))
PKNCA.set.summary(
  name="half.life",
  description="arithmetic mean and standard deviation",
  point=business.mean,
  spread=business.sd
)
add.interval.col("r.squared",
                 FUN=NA,
                 values=c(FALSE, TRUE),
                 unit_type="unitless",
                 pretty_name="$r^2$",
                 desc="The r^2 value of the half-life calculation",
                 depends="half.life")
PKNCA.set.summary(
  name="r.squared",
  description="arithmetic mean and standard deviation",
  point=business.mean,
  spread=business.sd
)
add.interval.col("adj.r.squared",
                 FUN=NA,
                 values=c(FALSE, TRUE),
                 unit_type="unitless",
                 pretty_name="$r^2_{adj}$",
                 desc="The adjusted r^2 value of the half-life calculation",
                 depends="half.life")
PKNCA.set.summary(
  name="adj.r.squared",
  description="arithmetic mean and standard deviation",
  point=business.mean,
  spread=business.sd
)
add.interval.col("lambda.z",
                 FUN=NA,
                 values=c(FALSE, TRUE),
                 unit_type="inverse_time",
                 pretty_name="$\\lambda_z$",
                 desc="The elimination rate of the terminal half-life",
                 depends="half.life")
PKNCA.set.summary(
  name="lambda.z",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)
add.interval.col("lambda.z.time.first",
                 FUN=NA,
                 values=c(FALSE, TRUE),
                 unit_type="time",
                 pretty_name="First time for $\\lambda_z$",
                 desc="The first time point used for the calculation of half-life",
                 depends="half.life")
PKNCA.set.summary(
  name="lambda.z.time.first",
  description="median and range",
  point=business.median,
  spread=business.range
)
add.interval.col("lambda.z.time.last",
                 FUN=NA,
                 values=c(FALSE, TRUE),
                 unit_type="time",
                 pretty_name="Last time for $\\lambda_z$",
                 desc="The last time point used for the calculation of half-life",
                 depends="half.life")
PKNCA.set.summary(
  name="lambda.z.time.last",
  description="median and range",
  point=business.median,
  spread=business.range
)
add.interval.col("lambda.z.n.points",
                 FUN=NA,
                 values=c(FALSE, TRUE),
                 unit_type="count",
                 pretty_name="Number of points used for lambda_z",
                 desc="The number of points used for the calculation of half-life",
                 depends="half.life")
PKNCA.set.summary(
  name="lambda.z.n.points",
  description="median and range",
  point=business.median,
  spread=business.range
)
add.interval.col("clast.pred",
                 FUN=NA,
                 values=c(FALSE, TRUE),
                 unit_type="conc",
                 pretty_name="Clast,pred",
                 desc="The concentration at Tlast as predicted by the half-life",
                 depends="half.life")
PKNCA.set.summary(
  name="clast.pred",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)
add.interval.col("span.ratio",
                 FUN=NA,
                 values=c(FALSE, TRUE),
                 unit_type="fraction",
                 pretty_name="Span ratio",
                 desc="The ratio of the half-life to the duration used for half-life calculation",
                 depends="half.life")
PKNCA.set.summary(
  name="span.ratio",
  description="geometric mean and geometric coefficient of variation",
  point=business.geomean,
  spread=business.geocv
)

#' Determine which concentrations were used for half-life calculation
#'
#' @param object A PKNCAresults object
#' @returns A logical vector with `TRUE` if the point was used for half-life
#'   (including concentrations below the limit of quantification within the
#'   range of times for calculation), `FALSE` if it was not used for half-life
#'   but the half-life was calculated for the interval, and `NA` if half-life
#'   was not calculated for the interval. If a row is excluded from all
#'   calculations, it is set to `NA` as well.
#' @examples
#' o_conc <- PKNCAconc(Theoph, conc~Time|Subject)
#' o_data <- PKNCAdata(o_conc, intervals = data.frame(start = 0, end = Inf, half.life = TRUE))
#' o_nca <- pk.nca(o_data)
#' get_halflife_points(o_nca)
#' @export
get_halflife_points <- function(object) {
  # Insert a ROWID column so that we can reconstruct the order at the end
  rowid_col <- paste0(max(names(as.data.frame(as_PKNCAconc(object)))), "ROWID")
  object$data$conc$data[[rowid_col]] <- seq_len(nrow(object$data$conc$data))

  # Find the concentrations and results that go together
  splitdata <- full_join_PKNCAdata(as_PKNCAdata(object), extra_conc_cols = rowid_col)
  splitresults_prep <- as.data.frame(object)
  splitresults <-
    tidyr::nest(
      splitresults_prep,
      data_results = !c(intersect(names(splitresults_prep), names(splitdata)), "start", "end")
    )
  base_results <-
    dplyr::inner_join(
      splitdata, splitresults,
      by = intersect(names(splitdata), names(splitresults))
    )

  ret <- rep(NA, nrow(as.data.frame(as_PKNCAconc(object))))
  for (idx in seq_len(nrow(base_results))) {
    ret_current <-
      get_halflife_points_single(
        conc = base_results$data_conc[[idx]],
        results = base_results$data_results[[idx]],
        rowid_col = rowid_col
      )
    if (any(!is.na(ret[ret_current$rowid]))) {
      stop(
        "More than one half-life calculation was attempted on the following rows: ",
        paste(ret_current$rowid, collapse = ", ")
      )
    }
    ret[ret_current$rowid] <- ret_current$hl_used
  }
  ret
}

# Get the half-life points for a single interval
get_halflife_points_single <- function(conc, results, rowid_col) {
  # "include_half.life" and "exclude_half.life" columns are present in conc, if
  # they apply. That comes from `full_join_PKNCAdata()`
  ret <- data.frame(hl_used = NA, rowid = conc[[rowid_col]])
  if ("half.life" %in% results$PPTESTCD) {
    if ("include_half.life" %in% names(conc) && !all(is.na(conc$include_half.life))) {
      ret$hl_used <- conc$include_half.life %in% TRUE
    } else {
      time_first <- results$PPORRES[results$PPTESTCD %in% "lambda.z.time.first"]
      time_last <- results$PPORRES[results$PPTESTCD %in% "lambda.z.time.last"]
      excluded <-
        if ("exclude_half.life" %in% names(conc)) {
          conc$exclude_half.life %in% TRUE
        } else {
          FALSE
        }
      ret$hl_used <- (time_first <= conc$time) & (conc$time <= time_last) & !excluded
    }
  }
  ret
}

test_that("pk.calc.half.life", {
  # Confirm that half-life is correctly calculated with a simple
  # exponential decay
  v1 <-
    pk.calc.half.life(conc=c(1, 0.5, 0.25),
                      time=c(0, 1, 2),
                      min.hl.points=3,
                      allow.tmax.in.half.life=TRUE,
                      adj.r.squared.factor=0.0001)$half.life
  expect_equal(v1, 1)

  # Ensure that when input data is not checked, the code works
  # correctly.
  v2 <-
    pk.calc.half.life(conc=c(1, 0.5, 0.25),
                      time=c(0, 1, 2),
                      min.hl.points=3,
                      allow.tmax.in.half.life=TRUE,
                      adj.r.squared.factor=0.0001,
                      check=FALSE)$half.life

  expect_equal(v2, 1)

  # Ensure that min.hl.points is respected
  expect_warning(pk.calc.half.life(conc=c(1, 0.5, 0.25),
                                   time=c(0, 1, 2),
                                   min.hl.points=4),
                 regexp="Too few points for half-life calculation")

  # Ensure that when there are more than one best models by adjusted
  # r-squared, the one with the most points is used.
  expect_warning(
    expect_equal(pk.calc.half.life(conc=c(1, 0.5, 0.25, 0.1251),
                                   time=c(0, 1, 2, 3),
                                   min.hl.points=3,
                                   allow.tmax.in.half.life=TRUE,
                                   adj.r.squared.factor=0.1,
                                   check=FALSE)$half.life,
                 1.000346,
                 tolerance=0.0001))

  # Ensure that the allow.tmax.in.half.life parameter is followed
  expect_equal(pk.calc.half.life(conc=c(1, 0.5, 0.25, 0.1251),
                                 time=c(0, 1, 2, 3),
                                 min.hl.points=3,
                                 allow.tmax.in.half.life=FALSE,
                                 check=FALSE)$half.life,
               1.000577,
               tolerance=0.00001)
  expect_warning(
    expect_equal(pk.calc.half.life(conc=c(1, 0.5, 0.25, 0.1251),
                                   time=c(0, 1, 2, 3),
                                   min.hl.points=3,
                                   allow.tmax.in.half.life=TRUE,
                                   adj.r.squared.factor=0.1,
                                   check=FALSE)$half.life,
                 1.000346,
                 tolerance=0.00001))


  # Make sure that when tmax and tlast are given that they are
  # automatically included and not recalculated
  expect_equal(pk.calc.half.life(conc=c(1, 0.5, 0.25, 0.1251),
                                 time=c(0, 1, 2, 3),
                                 min.hl.points=3,
                                 tmax=0,
                                 tlast=3,
                                 allow.tmax.in.half.life=TRUE,
                                 adj.r.squared.factor=0.01,
                                 check=FALSE),
               data.frame(lambda.z=0.6929073,
                          r.squared=0.9999999,
                          adj.r.squared=0.9999999,
                          lambda.z.time.first=0,
                          lambda.z.time.last=3,
                          lambda.z.n.points=4,
                          clast.pred=0.12507,
                          half.life=1.000346,
                          span.ratio=2.998962),
               tolerance=0.0001)
  # It only gives tlast or tmax if you don't give them as inputs.
  expect_equal(pk.calc.half.life(conc=c(1, 0.5, 0.25, 0.1251),
                                 time=c(0, 1, 2, 3),
                                 min.hl.points=3,
                                 tlast=3,
                                 allow.tmax.in.half.life=TRUE,
                                 adj.r.squared.factor=0.01,
                                 check=FALSE),
               data.frame(lambda.z=0.6929073,
                          r.squared=0.9999999,
                          adj.r.squared=0.9999999,
                          lambda.z.time.first=0,
                          lambda.z.time.last=3,
                          lambda.z.n.points=4,
                          clast.pred=0.12507,
                          half.life=1.000346,
                          span.ratio=2.998962,
                          tmax=0),
               tolerance=0.0001)
    expect_equal(pk.calc.half.life(conc=c(1, 0.5, 0.25, 0.1251),
                                 time=c(0, 1, 2, 3),
                                 min.hl.points=3,
                                 tmax=0,
                                 allow.tmax.in.half.life=TRUE,
                                 adj.r.squared.factor=0.01,
                                 check=FALSE),
               data.frame(lambda.z=0.6929073,
                          r.squared=0.9999999,
                          adj.r.squared=0.9999999,
                          lambda.z.time.first=0,
                          lambda.z.time.last=3,
                          lambda.z.n.points=4,
                          clast.pred=0.12507,
                          half.life=1.000346,
                          span.ratio=2.998962,
                          tlast=3),
               tolerance=0.0001)

})

test_that("half-life manual point selection", {
  expect_equal(
      pk.calc.half.life(conc=c(3, 1, 0.5, 0.13, 0.12, 0.113),
                        time=c(0, 1, 2, 3, 4, 5),
                        manually.selected.points=TRUE,
                        min.hl.points=3,
                        allow.tmax.in.half.life=FALSE,
                        check=FALSE)$half.life,
      1.00653,
      tolerance=0.0001,
      info="manual selection uses the given points as is")
  expect_true(
    pk.calc.half.life(conc=c(3, 1, 0.5, 0.13, 0.12, 0.113),
                      time=c(0, 1, 2, 3, 4, 5),
                      manually.selected.points=TRUE,
                      min.hl.points=3,
                      allow.tmax.in.half.life=FALSE,
                      check=FALSE)$half.life !=
      pk.calc.half.life(conc=c(3, 1, 0.5, 0.13, 0.12, 0.113),
                        time=c(0, 1, 2, 3, 4, 5),
                        min.hl.points=3,
                        allow.tmax.in.half.life=FALSE,
                        check=FALSE)$half.life,
    info="manual selection is different than automatic selection")
  expect_true(
    pk.calc.half.life(conc=c(3, 1, 0.5, 0.13, 0.12, 0.113),
                      time=c(0, 1, 2, 3, 4, 5),
                      manually.selected.points=TRUE,
                      min.hl.points=3,
                      tlast=20,
                      allow.tmax.in.half.life=FALSE,
                      check=FALSE)$clast.pred <
      pk.calc.half.life(conc=c(3, 1, 0.5, 0.13, 0.12, 0.113),
                        time=c(0, 1, 2, 3, 4, 5),
                        manually.selected.points=TRUE,
                        min.hl.points=3,
                        allow.tmax.in.half.life=FALSE,
                        check=FALSE)$clast.pred,
    info="manually-selected half-life respects tlast and generates a different clast.pred")
  expect_warning(
    manual_blq <-
      pk.calc.half.life(
        conc=rep(0, 6),
        time=c(0, 1, 2, 3, 4, 5),
        manually.selected.points=TRUE,
        min.hl.points=3,
        tlast=20,
        allow.tmax.in.half.life=FALSE,
        check=FALSE
      ),
    regexp="No data to manually fit for half-life (all concentrations may be 0 or excluded)",
    fixed=TRUE,
    info="All BLQ with manual point selection gives a warning"
  )
  expect_true(all(is.na(unlist(manual_blq))),
              info="All BLQ with manual point selection gives all NA results")

  excluded_result <-
    data.frame(
      lambda.z = -log(2),
      r.squared = 1,
      adj.r.squared = 1,
      lambda.z.time.first = 1L,
      lambda.z.time.last = 3L,
      lambda.z.n.points = 3L,
      clast.pred = 8,
      half.life = -1,
      span.ratio = -2,
      tmax = 3L,
      tlast = 3L
    )
  attr(excluded_result, "exclude") <- "Negative half-life estimated with manually-selected points"
  expect_equal(
    pk.calc.half.life(conc = 2^(1:3), time = 1:3, manually.selected.points = TRUE),
    excluded_result
  )
})

test_that("two-point half-life succeeds (fix #114)", {
  expect_warning(expect_warning(
    expect_equal(
      pk.calc.half.life(
        conc=c(1, 0.5),
        time=c(0, 1),
        min.hl.points=2,
        allow.tmax.in.half.life=TRUE,
        check=FALSE
      ),
      data.frame(
        lambda.z=log(2),
        r.squared=1,
        adj.r.squared=NA_real_,
        lambda.z.time.first=0,
        lambda.z.time.last=1,
        lambda.z.n.points=2,
        clast.pred=0.5,
        half.life=1,
        span.ratio=1,
        tmax=0,
        tlast=1
      )
    ),
    class = "pknca_halflife_2points"),
    class = "pknca_adjr2_2points"
  )
})

test_that("When tlast is excluded, lambda.z.time.last != tlast", {
  d_conc <- data.frame(
    conc = c(1, 0.5, 0.25, 0.1251, 0.05, 0),
    time = c(0, 1, 2, 3, 4, 5),
    exclude_hl = c(FALSE, FALSE, FALSE, FALSE, TRUE, FALSE),
    subject = 1
  )
  d_dose <- data.frame(dose = 1, time = 0, subject = 1)

  o_conc <- PKNCAconc(d_conc, formula = conc ~ time | subject, exclude_half.life = "exclude_hl")
  o_dose <- PKNCAdose(d_dose, formula = dose ~ time | subject)
  o_data <- PKNCAdata(o_conc, o_dose)
  myresult <- pk.nca(o_data)
  expect_equal(
    myresult$result[myresult$result$PPTESTCD == "lambda.z.time.last",]$PPORRES,
    3
  )
})


test_that("lambda.z.time.first is after the absorption phase in infusion studies (shifts with duration)", {
  # Simulate an infusion with two different durations
  d_conc <- data.frame(
    conc = c(0.5, 1.0, 0.5, 0.12, 0.25, 0.05),
    time = 1:6,
    subject = 1
  )
  d_dose <- data.frame(
    dose = 1,
    time = 0,
    duration1 = 2,  # Duration for case 1
    duration2 = 3,  # Duration for case 2
    route = "intravascular",
    subject = 1
  )

  # First case: duration = 2, so lambda.z.time.first should be after 2
  o_data1 <- PKNCAdata(
    PKNCAconc(d_conc, conc ~ time | subject),
    PKNCAdose(d_dose, dose ~ time | subject, route = "route", duration = "duration1"),
    intervals = data.frame(start = 0, end = Inf, half.life = TRUE)
  )
  res1 <- pk.nca(o_data1)
  lz_first1 <- res1$result$PPORRES[res1$result$PPTESTCD == "lambda.z.time.first"]
  expect_equal(lz_first1, 3)

  # Second case: duration = 4, so lambda.z.time.first should be after 3
  o_data2 <- PKNCAdata(
    PKNCAconc(d_conc, conc ~ time | subject),
    PKNCAdose(d_dose, dose ~ time | subject, route = "route", duration = "duration2"),
    intervals = data.frame(start = 0, end = Inf, half.life = TRUE)
  )
  res2 <- pk.nca(o_data2)
  lz_first2 <- res2$result$PPORRES[res2$result$PPTESTCD == "lambda.z.time.first"]
  expect_equal(lz_first2, 4)

  # And the second case should be later than the first
  expect_true(lz_first2 > lz_first1)
})

test_that("for an interval with multiple doses lambda.z is calculated for the latest one", {
  # Simulate two infusions: one at time 0 (duration 2), one at time 4 (duration 2)
  d_conc <- data.frame(
    conc = c(0.5, 1.0, 0.99, 0.98, 0.97, 0.96,
             0.5, 1.0, 0.8, 0.5, 0.2, 0.1),
    time = 1:12,
    subject = 1
  )
  d_dose <- data.frame(
    dose = c(1, 1),
    time = c(6, 0),
    duration = c(2, 2),
    route = "intravascular",
    subject = 1
  )
  o_data <- PKNCAdata(
    PKNCAconc(d_conc, conc ~ time | subject),
    PKNCAdose(d_dose, dose ~ time | subject, route = "route", duration = "duration"),
    intervals = data.frame(start = c(0, 0, 6),
                           end = c(Inf, 6, 12),
                           half.life = TRUE)
  )
  res <- pk.nca(o_data)
  lzs <- res$result$PPORRES[res$result$PPTESTCD == "lambda.z"]
  # Lambda z for the first interval (0 to Inf) should be the same as for the third interval (6 to 12)
  expect_equal(lzs[1], lzs[3])
})

test_that("get_halflife_points", {
  o_conc <- PKNCAconc(Theoph, conc~Time|Subject)
  o_data <- PKNCAdata(o_conc, intervals = data.frame(start = 0, end = Inf, half.life = TRUE))
  o_nca <- suppressMessages(pk.nca(o_data))
  hl_points <- suppressMessages(get_halflife_points(o_nca))

  # Visualize the results
  # d_all <- Theoph
  # d_all$hl_points <- hl_points
  # ggplot(d_all, aes(x = Time, y = conc, colour = hl_points, groups = Subject)) +
  #   geom_point() + geom_line()
  expect_equal(
    hl_points,
    c(FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, TRUE,
      TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE,
      TRUE, TRUE, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE,
      FALSE, FALSE, TRUE, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE,
      FALSE, FALSE, FALSE, TRUE, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE,
      FALSE, FALSE, FALSE, TRUE, TRUE, TRUE, TRUE, FALSE, FALSE, FALSE,
      FALSE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, FALSE, FALSE,
      FALSE, FALSE, FALSE, FALSE, FALSE, TRUE, TRUE, TRUE, TRUE, FALSE,
      FALSE, FALSE, FALSE, FALSE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE,
      FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, TRUE,
      TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE,
      FALSE, TRUE, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE,
      FALSE, FALSE, TRUE, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE,
      FALSE, FALSE, FALSE, TRUE, TRUE, TRUE)
  )

  # Setup for the remaining tests
  d_conc <- as.data.frame(Theoph[Theoph$Subject %in% Theoph$Subject[1], ])
  d_conc$incl <- c(FALSE, TRUE, rep(NA, 8), TRUE)
  d_conc$excl_txt <- c(NA, "x", rep(NA, 8), "x")
  d_conc$excl <- c(NA, TRUE, rep(NA, 6), TRUE, FALSE, TRUE)

  # No modification
  o_conc <- PKNCAconc(d_conc, conc~Time|Subject)
  o_data <- PKNCAdata(o_conc, intervals = data.frame(start = 0, end = Inf, half.life = TRUE))
  o_nca <- suppressMessages(suppressWarnings(pk.nca(o_data)))
  expect_equal(
    suppressMessages(get_halflife_points(o_nca)),
    c(rep(FALSE, 8), rep(TRUE, 3))
  )

  # Test include_half.life
  o_conc <- PKNCAconc(d_conc, conc~Time|Subject, include_half.life = "incl")
  o_data <- PKNCAdata(o_conc, intervals = data.frame(start = 0, end = Inf, half.life = TRUE))
  o_nca <- suppressMessages(suppressWarnings(pk.nca(o_data)))
  expect_equal(
    suppressMessages(get_halflife_points(o_nca)),
    c(FALSE, TRUE, rep(FALSE, 8), TRUE)
  )

  # Test exclude
  o_conc <- PKNCAconc(d_conc, conc~Time|Subject, exclude = "excl_txt")
  o_data <- PKNCAdata(o_conc, intervals = data.frame(start = 0, end = Inf, half.life = TRUE))
  o_nca <- suppressMessages(pk.nca(o_data))
  expect_equal(
    suppressMessages(get_halflife_points(o_nca)),
    # NA values indicate that the point was excluded from all calculations
    c(FALSE, NA, FALSE, FALSE, FALSE, FALSE, FALSE, TRUE, TRUE, TRUE, NA)
  )

  # Test exclude_half.life
  o_conc <- PKNCAconc(d_conc, conc~Time|Subject, exclude_half.life = "excl")
  o_data <- PKNCAdata(o_conc, intervals = data.frame(start = 0, end = Inf, half.life = TRUE))
  o_nca <- suppressMessages(pk.nca(o_data))
  expect_equal(
    suppressMessages(get_halflife_points(o_nca)),
    # NA values indicate that the point was excluded from all calculations
    c(FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, TRUE, TRUE, FALSE, TRUE, FALSE)
  )

  # Multiple, overlapping half-life calculations gives an error
  o_conc <- PKNCAconc(d_conc, conc~Time|Subject)
  o_data <-
    PKNCAdata(
      o_conc,
      intervals =
        data.frame(
          start = 0, end = c(24, Inf),
          half.life = TRUE
        )
    )
  o_nca <- suppressMessages(suppressWarnings(pk.nca(o_data)))
  expect_error(
    suppressMessages(get_halflife_points(o_nca)),
    regexp = "More than one half-life calculation was attempted on the following rows: 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11"
  )

  # Half-life points are selected right even if lambda.z.time.last != tlast (#448)
  d_conc <- data.frame(
    conc = c(1, 0.5, 0.25, 0.125, 0.07, 0, 0.01),
    time = c(0, 1, 2, 3, 4, 5, 6),
    exclude_hl = c(FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, TRUE),
    subject = 1
  )
  # without exclusion of Tlast
  o_conc <- PKNCAconc(d_conc, formula = conc ~ time | subject)
  o_data <- PKNCAdata(o_conc, intervals = data.frame(start = 0, end = Inf, half.life = TRUE))
  o_nca <- suppressMessages(pk.nca(o_data))
  hl_points <- suppressMessages(get_halflife_points(o_nca))
  # Note that BLQ times are included
  expect_equal(
    d_conc$time[which(hl_points)],
    1:6,
    info = "get_halflife_points uses lambda.z.time.last, not tlast"
  )
  # with exclusion of Tlast
  o_conc <- PKNCAconc(d_conc, formula = conc ~ time | subject, exclude_half.life = "exclude_hl")
  o_data <- PKNCAdata(o_conc, intervals = data.frame(start = 0, end = Inf, half.life = TRUE))
  o_nca <- suppressMessages(pk.nca(o_data))
  hl_points <- suppressMessages(get_halflife_points(o_nca))
  # lambda.z.time.last should be 4, tlast is 6 and excluded
  expect_equal(
    d_conc$time[which(hl_points)],
    1:4,
    info = "get_halflife_points uses lambda.z.time.last, not tlast"
  )
})

#' Clean up the time to steady-state parameters and return a data frame for use
#' by the tss calculators.
#'
#' @inheritParams assert_conc_time
#' @inheritParams PKNCA.choose.option
#' @param subject Subject identifiers (used as a random effect in the model)
#' @param treatment Treatment description (if missing, all subjects are assumed
#'   to be on the same treatment)
#' @param subject.dosing Subject number for dosing
#' @param time.dosing Time of dosing
#' @param conc.blq See [clean.conc.blq()]
#' @param conc.na See [clean.conc.na()]
#' @param check Run [assert_conc_time()]?
#' @param \dots Discarded inputs to allow generic calls between tss methods.
#' @returns a data frame with columns for `conc`entration, `time`, `subject`,
#'   and `treatment`.
pk.tss.data.prep <- function(conc, time, subject, treatment,
                             subject.dosing, time.dosing,
                             options=list(),
                             conc.blq=NULL,
                             conc.na=NULL,
                             check=TRUE, ...) {
  if (check) {
    # When subject and time are not given, then monotonicity tests for
    # time are not required.
    sorted_time <- missing(subject) & missing(treatment)
    assert_conc_time(conc = conc, time = time, sorted_time = sorted_time)
  }
  if (!missing(subject.dosing) & missing(subject)) {
    stop("Cannot give subject.dosing without subject")
  }
  if (any(is.na(time.dosing))) {
    stop("time.dosing may not contain any NA values")
  }
  if (!missing(subject)) {
    if (!missing(treatment)) {
      ret <-
        clean.conc.blq(
          conc = conc, time = time,
          conc.blq = conc.blq, conc.na = conc.na, options = options,
          check = FALSE,
          subject, treatment
        )
    } else {
      ret <-
        clean.conc.blq(
          conc = conc, time = time,
          conc.blq = conc.blq, conc.na = conc.na, options = options,
          check = FALSE,
          subject
        )
    }
  } else if (missing(treatment)) {
    ret <-
      clean.conc.blq(
        conc = conc, time = time,
        conc.blq = conc.blq, conc.na = conc.na, options = options,
        check = FALSE
      )
  } else {
    ret <-
      clean.conc.blq(
        conc = conc, time = time,
        conc.blq = conc.blq, conc.na = conc.na, options = options,
        check = FALSE,
        treatment
      )
  }
  if (missing(subject.dosing)) {
    # Shrink the data to just the predose data
    ret <- subset(ret, time %in% time.dosing)
  } else {
    dosing <-
      data.frame(
        subject=subject.dosing,
        time=time.dosing,
        stringsAsFactors=FALSE
      )
    # Shrink the data to just the predose data (by subject)
    ret <- merge(ret, dosing)
  }
  # Clean out unnecessary parts of the output data frame
  if ("subject" %in% names(ret)) {
    if (length(unique(ret$subject)) == 1) {
      # Drop the "subject" column from single-subject data
      ret$subject <- NULL
    } else if (!is.factor(ret$subject)) {
      # Make sure that it is a factor made from a character vector because the
      # output subject numbering will come from row.names of the random
      # effects.
      ret$subject <- factor(as.character(ret$subject))
    }
  }
  if ("treatment" %in% names(ret)) {
    if (length(unique(ret$treatment)) == 1) {
      # Drop the "treatment" column from single-treatment data
      ret$treatment <- NULL
    } else if (!is.factor(ret$treatment)) {
      # Make sure that it is a factor otherwise
      ret$treatment <- factor(ret$treatment)
    }
  }
  ret
}

#' Compute the time to steady-state (tss)
#'
#' @param \dots Passed to [pk.tss.monoexponential()] or
#'   [pk.tss.stepwise.linear()].
#' @param check See [pk.tss.data.prep()]
#' @param type The type of Tss to calculate, either `stepwise.linear` or
#'   `monoexponential`
#' @returns A data frame with columns as defined from `pk.tss.monoexponential`
#'   and/or `pk.tss.stepwise.linear`.
#' @family Time to steady-state calculations
#' @export
pk.tss <- function(...,
                   type=c("monoexponential", "stepwise.linear"),
                   check=TRUE) {
  type <- match.arg(type, several.ok=TRUE)
  ret <- NA
  if ("monoexponential" %in% type) {
    ret_monoexponential <- pk.tss.monoexponential(..., check=check)
    if (identical(NA, ret)) {
      ret <- ret_monoexponential
    } else {
      stop("Bug in pk.tss where ret is set to non-NA too early.  Please report the bug with a reproducible example.") # nocov
    }
    # Set check to FALSE if it has already been checked (so that it
    # doesn't happen again in stepwise.linear)
    check <- FALSE
  }
  if ("stepwise.linear" %in% type) {
    ret_stepwise <- pk.tss.stepwise.linear(..., check=check)
    if (identical(NA, ret)) {
      ret <- ret_stepwise
    } else {
      ret <- merge(ret, ret_stepwise, all=TRUE)
    }
  }
  ret
}

---
title: "AUC integration methods"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{AUC integration methods}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r knitr-setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  warning = FALSE,
  comment = "#>"
)
```

```{r setup, include = FALSE}
library(PKNCA)
library(ggplot2)
auc_methods <- c("lin up/log down", "linear", "lin-log")
auc_types <- c("AUCinf", "AUClast", "AUCall")

comma_and <- function(x) {
  paste(
    paste(x[-length(x)], collapse = ", "),
    ngettext(length(x) - 1, ", and", "and"),
    x[length(x)]
  )
}
```

# Integration methods for Area Under the Concentration-Time curve (AUC)

There are `r length(auc_methods)` methods for choosing how to integrate the AUC
between two concentration measurements in `PKNCA`.  They are
`r comma_and(auc_methods)`.  Additionally, there are `r length(auc_types)`
methods for extrapolating after the last concentration above the limit of
quantification.  They are `r comma_and(auc_types)`.  Other methods of
calculating AUC (such as AUC~tau~ and AUC~int~) are made with variants of these.

# Definitions and abbreviations

* AUC: Area under the concentration-time curve
* BLQ: Below the lower limit of quantification
* LLOQ: lower limit of quantification
* NCA: Noncompartmental analysis
* Profile: A set of concentration-time points for calculation
* T~last~: The last concentration above the limit of quantification within a profile
* T~max~: The time of the maximum concentration

# Description of methods of integrating between two concentrations before T~last~

Note that other NCA tools may not describe interpolation as zero.  The
zero-interpolation rules are used by PKNCA to assist with other methods used
across the suite of tools for interpolation and data cleaning within PKNCA.  The
zero-interpolation rules could be swapped for linear trapezoidal rules with the
same effects here.

## Linear up/logarithmic down (`"lin up/log down"`) interpolation

Linear up/logarithmic down interpolation is the most commonly used method for
PK, and it is the default for `PKNCA`.

Linear up/logarithmic down interpolation is often used when an exogenous
substance is dosed and measured, and when the elimination likely occurs by
first-order elimination from the body.

Linear up/logarithmic down interpolation uses the following rules in order for
each pair of concentrations through T~last~:

1. If concentrations are both zero, interpolate as zero;
2. If concentrations are decreasing and the second concentration is not zero,
   use logarithmic interpolation; and
3. If concentrations are decreasing before T~last~ or increasing ever, use
   linear interpolation.

## Linear trapezoidal (`"linear"`) interpolation

Linear trapezoidal interpolation is often used when an endogenous substance is
measured (and possibly dosed), and when the elimination may not occur by
first-order elimination processes.

Linear trapezoidal interpolation uses the following rules in order for each pair
of concentrations through T~last~:

1. If concentrations are both zero, interpolate as zero; and
2. Use linear interpolation for all other times (this could be the only rule).

## Linear to T~max~/logarithmic after T~max~ (`"lin-log"`) interpolation

Linear to T~max~/logarithmic after T~max~ interpolation is infrequently used.
It uses the following rules in order for each pair of concentrations through
T~last~:

1. If concentrations are both zero, interpolate as zero;
2. If concentrations are before T~max~, use linear interpolation;
3. If concentrations are after T~max~ (and before T~last~) and either concentration is zero, use linear interpolation; and
4. If concentrations are after T~max~ and neither is zero, use logarithmic interpolation.

# Description of methods of integrating between two concentrations after T~last~

## `"AUClast"` extrapolation

AUClast extrapolation after T~last~ is the simplest.  It is no extrapolation;
the extrapolated AUC integral is zero.

## `"AUCall"` extrapolation

AUCall extrapolation after T~last~ has two rules:

1. If the last concentration measured is above the limit of quantification (in
   other words, the last time is T~last~), then no extrapolation is done; otherwise
2. Integrate linearly the triangle between T~last~ and the time of zero
   concentration after T~last~.

## `"AUCinf"` extrapolation

AUCinf extrapolation requires estimation of a half-life.  It extrapolates using the equation

$Extrap = \frac{C_{last}}{\lambda_z}$

# Examples

```{r setup-example, include=FALSE}
examples <-
  list(
    "Clast is above the LLOQ" =
      data.frame(
        conc = c(0, 1.8, 3, 2, 1, 0.5, 0.25),
        time = 0:6
      ),
    "Clast is below the LLOQ" =
      data.frame(
        conc = c(0, 1.8, 3, 2, 1, 0.5, 0),
        time = 0:6
      ),
    "Concentrations increase and decrease after T~max~ with zeros in the middle" =
      data.frame(
        conc = c(0, 1.8, 0, 0, 3, 2, 2.5, 0, 0, 2.5, 1, 0.5, 0),
        time = 0:12
      )
  )
```

```{r example-figures, fig.cap = caption, echo = FALSE}
caption <- character()
for (current_example_nm in names(examples)) {
  for (current_method in auc_methods) {
    for (current_type in auc_types) {
      caption <-
        c(
          caption,
          sprintf(
            "Example PK where %s; %s interpolation method; %s extrapolation method (AUC type); and keeping all BLQ values (not default BLQ handling)",
            current_example_nm, current_method, current_type
          )
        )
      current_example <- examples[[current_example_nm]]

      # Handle BLQ
      current_example_noblq <- clean.conc.blq(conc = current_example$conc, time = current_example$time)
      current_example_noblq$BLQ <- "Above LLOQ"
      current_example_points <-
        merge(current_example, current_example_noblq, all = TRUE)
      current_example_points$BLQ[is.na(current_example_points$BLQ)] <- "Below LLOQ"
      
      # Calculate some NCA parameters required for interpolation and extrapolation
      tlast <- pk.calc.tlast(conc = current_example$conc, time = current_example$time)
      half_life <- pk.calc.half.life(conc = current_example$conc, time = current_example$time)
      
      # Define the methods used for interpolation and extrapolation
      current_example$method <-
        factor(
          PKNCA:::choose_interval_method(
            conc = current_example$conc,
            time = current_example$time,
            tlast = tlast,
            method = current_method,
            auc.type = current_type,
            options = list()
          ),
          levels = c("zero", "linear", "log", "extrap_log")
        )
      stopifnot(!any(is.na(current_example$method)))
      current_example_interpolate <-
        data.frame(
          time =
            sort(c(
              # make lines almost directly down at the transition after Tlast to
              # zero, where applicable
              current_example$time + sqrt(.Machine$double.eps),
              seq(min(current_example$time), max(current_example$time) + diff(current_example$time)[1], length.out = 200)
            ))
        )
      current_example_interpolate$conc <-
        interp.extrap.conc(
          conc = current_example$conc,
          time = current_example$time,
          time.out = current_example_interpolate$time,
          lambda.z = half_life$lambda.z,
          method = current_method,
          auc.type = current_type,
          conc.blq = "keep"
        )
      current_example$time_after <- c(current_example$time[-1], Inf)
      if (current_example$method[nrow(current_example)] == "zero") {
        current_example_plot <- current_example
      } else {
        current_example_before_tlast <- current_example[current_example$time < tlast, ]
        current_example_after_tlast <- current_example[current_example$time >= tlast, ][1, ]
        current_example_after_tlast$method <- current_example$method[nrow(current_example)]
        current_example_after_tlast$time_after <- Inf
        current_example_plot <-
          dplyr::bind_rows(
            current_example_before_tlast,
            current_example_after_tlast
          )
      }
      print(
        ggplot(current_example_points, aes(x = time, y = conc)) +
          geom_rect(data = current_example_plot, aes(xmin = time, xmax = time_after, ymin = -Inf, ymax = Inf, fill = method), alpha = 0.2) +
          geom_point(aes(shape = BLQ)) +
          geom_line(data = current_example_interpolate) +
          scale_colour_hue(drop = FALSE)
      )
    }
  }
}
```

---
title: "Superposition of Pharmacokinetic Data"
author: "Bill Denney"
output:
  rmarkdown::html_vignette:
    toc: yes
    toc_depth: 6
vignette: >
  %\VignetteIndexEntry{Superposition of Pharmacokinetic Data}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, echo=FALSE, include=FALSE}
library(PKNCA)
library(knitr)
library(ggplot2)
scale_colour_discrete <- scale_colour_hue
scale_fill_discrete <- scale_fill_hue
scale_colour_ordinal <- scale_colour_hue
scale_fill_ordinal <- scale_fill_hue
```

Examples simplify understanding.  Below is an example of how to use
the theophylline dataset to estimate the concentration for each
subject after multiple doses.

```{r check-ggplot, include=!requireNamespace("ggplot2"), results="asis"}
cat("ggplot2 is required for this vignette to work correctly.  Please install the ggplot2 library and retry building the vignette.")
```

# Load the data

```{r showtheoph}
## It is always a good idea to look at the data
knitr::kable(head(datasets::Theoph))
```

The columns that we will be interested in for our analysis are conc,
Time, and Subject in the concentration data.

```{r setupconc}
## By default it is groupedData; convert it to a data frame for use
conc_obj <- PKNCAconc(as.data.frame(datasets::Theoph), conc~Time|Subject)
```

# Compute the Superposition from Single-Dose Data to Steady-State

With a simple call, we can have the estimated steady-state
concentration for each subject.  At minimum, the time between dosing
(`tau`) must be provided.

```{r superposition, error=TRUE}
steady_state <- superposition(conc_obj, tau=24)
```

The error noting that the first concentration must be zero is due to
the fact that superposition usually occurs with single-dose data.  If
the first concentration is nonzero, the data are not likely to be
single-dose (or a data error should be fixed).  Let's find the offending data.

```{r findnonzero}
knitr::kable(subset(datasets::Theoph, Time == 0 & conc > 0),
             caption="Nonzero predose measurements",
             row.names=FALSE)
```

For this example, we will assume that these were errors, correct them to zero, and recalculate.

```{r allownonzero}
## Correct nonzero concentrations at time 0 to be BLQ.
theoph_corrected <- as.data.frame(datasets::Theoph)
theoph_corrected$conc[theoph_corrected$Time == 0] <- 0
conc_obj_corrected <- PKNCAconc(theoph_corrected, conc~Time|Subject)

## Calculate the new steady-state concentrations with 24 hour dosing
steady_state <- superposition(conc_obj_corrected, tau=24)
knitr::kable(head(steady_state, n=14),
             caption="Superposition at steady-state")
```

The output is a `r class(steady_state)` including all the grouping factors as columns, a column for `conc`entration, and a column for `time`.  Time point selection ensures that the beginning and end of the interval are included and that every measured time that contributes to the interval is included.  The points at the beginning and end of the interval are very similar; they are within a tolerance of `r formals(PKNCA:::superposition.numeric)$steady.state.tol` as defined by the `steady.state.tol` argument to superposition.

# Nonstandard Superposition Computations

## Compute the Superposition from Single-Dose Data to a Specific Dose

If simulation to a specific dose is needed, the number of dosing intervals (`n.tau`) can be specified.

```{r UnsteadyState}
## Calculate the unsteady-state concentrations with 24 hour dosing
unsteady_state <- superposition(conc_obj_corrected, tau=24, n.tau=2)
knitr::kable(head(unsteady_state, n=14),
             caption="Superposition before steady-state")
```

## Compute the Superposition from Single-Dose Data with >1 Dose Per Interval

Some dosing intervals are more complex than once per X hours (or days or weeks or...).  To predict more complex dosing with superposition, give the dose times within the interval.  The `dose.times` must all be less than `tau` (otherwise they are not in the interval).

```{r ComplexInterval}
## Calculate the new steady-state concentrations with 24 hour dosing
complex_interval_steady_state <- superposition(conc_obj_corrected, tau=24, dose.times=c(0, 2, 4))
knitr::kable(head(complex_interval_steady_state, n=10),
             caption="Superposition at steady-state with complex dosing")
```
```{r ComplexInterval-visualization, eval=requireNamespace("ggplot2")}
ggplot(complex_interval_steady_state,
       aes(y=conc, x=time, colour=Subject)) +
  geom_line()
```

With this more complex dosing interval, the number of time points estimated increases.  The next section describes the selection of time points.

### Show the Curve to Steady-State

To determine the concentration curve to get to steady-state, you can give all the dose times considered required to get to steady-state.  To do this, specify tau as the total time to steady-state, specify `n.tau` as `1` to indicate that only one round of dosing should be administered.

This command does not technically go to steady-state; if the `dose.times` are not sufficiently long to reach steady-state, it only goes for as many doses as requested.

```{r to_steady_state}
up_to_steady_state <- superposition(conc_obj_corrected,
                                    tau=4*24,
                                    n.tau=1,
                                    dose.times=seq(0, 3*24, by=12))
```
```{r to_steady_state-visualization, eval=requireNamespace("ggplot2")}
ggplot(up_to_steady_state, aes(x=time, y=conc, colour=Subject)) +
  geom_line()
```

# Time Point Selection and Addition

Superposition is often used to estimate NCA parameters with nonparametric methods.  To ensure that estimated parameters are as accurate as possible (especially $C_{max}$), each dose has every post-dose time point included.  Specifically, each dose will have the following times:

* 0 (zero) and `tau`,
* The time of each dose (the `dose.times` argument)
* Every value from the time column of the data modulo `tau` (shifting the time for each measurement to be within the dosing interval) repeated for each dose, and
* each time from the `additional.times` argument.

How the number of time points increases can be seen by comparing the time points for subject 1 in the steady-state single dosing and the complex dosing examples above.

```{r time-point-selection}
steady_state$time[steady_state$Subject == 1]
sum(steady_state$Subject == 1)
complex_interval_steady_state$time[complex_interval_steady_state$Subject == 1]
sum(complex_interval_steady_state$Subject == 1)
```

# Interpolation and Extrapolation Methods

The interpolation and extrapolation methods align with those used for calculating the AUC.  By default, interpolation uses the `PKNCA.options` selection for `"auc.method"` and extrapolation follows the curve of $AUC_{inf}$.  These can be modified with the `interp.method` and `extrap.method` arguments.

---
title: "Selection of Calculation Intervals"
author: "Bill Denney"
output:
  rmarkdown::html_vignette:
    toc: yes
    toc_depth: 6
vignette: >
  %\VignetteIndexEntry{Selection of Calculation Intervals}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo=TRUE,
                      fig.width=6,
                      fig.height=4)
library(cowplot)
library(PKNCA)
library(knitr)
library(ggplot2)
source("../tests/testthat/helper-generate_data.R")
scale_colour_discrete <- scale_colour_hue
scale_fill_discrete <- scale_fill_hue
```

```{r check-ggplot, include=!requireNamespace("ggplot2"), results="asis"}
cat("ggplot2 is required for this vignette to work correctly.  Please install the ggplot2 library and retry building the vignette.")
```

# Introduction

PKNCA considers two types of data grouping within data sets:  the group and the interval.  A group typically identifies a single subject given a single intervention type (a "treatment") with a single analyte.  An interval subsets a group by times within the group, and primary noncompartmental analysis (NCA) calculations are performed within an interval.

As a concrete example, consider the figure below shows the concentration-time profile of a study subject in a multiple-dose study.  The group is all points in the figure, and the interval for the last day (144 to 168 hr) is the area with blue shading.

```{r intro_interval_plot, echo=FALSE}
# Simulate concentration-time data and setup the PKNCAconc object
d_conc <-
  PKNCAconc(
    generate.conc(nsub=1,
                  ntreat=1,
                  time.points=c(0, 1, 2, 4, 6, 8, 12, 24, 36, 48),
                  nstudies=1,
                  nanalytes=1,
                  resid=0),
    conc~time|treatment+ID)
print(d_conc)
# Use superposition to simulate multiple doses
d_conc_multi <-
  superposition(d_conc,
                tau=168,
                dose.times=seq(0, 168-24, by=24),
                n.tau=1)
```
```{r intro_interval_plot-visualization, eval=requireNamespace("ggplot2")}
# Plot the concentration-time data and the interval
ggplot(d_conc_multi, aes(x=time, y=conc)) +
  geom_ribbon(data=d_conc_multi[d_conc_multi$time >= 144,],
              aes(ymax=conc, ymin=0),
              fill="skyblue") +
  geom_point() + geom_line() +
  scale_x_continuous(breaks=seq(0, 168, by=12)) +
  scale_y_continuous(limits=c(0, NA)) +
  labs(x="Time Since First Dose (hr)",
       y="Concentration\n(arbitrary units)")
```
```{r intro_interval_spec}
intervals_manual <- data.frame(start=144, end=168, auclast=TRUE)
knitr::kable(intervals_manual)

PKNCAdata(d_conc, intervals=intervals_manual)
```

# Group Matching

Group matching occurs by matching all overlapping column names between the groups and the interval data.frame.  (Note that grouping columns cannot be the word `start`, `end`, or share a name with an NCA parameter.)

## Selecting the Subjects for an Interval

The groups for an interval prepare for summarization.  Typically the groups will take a structure similar to the preferred summarization structure with groups nested in the logical method for summary.  As an example, the group structure may be: study, treatment, day, analyte, and subject.  The grouping names for an interval must be the same as or a subset of the grouping names used for the concentration data.

As the matching occurs with all available columns, the grouping columns names are only required to the level of specificity for the calculations desired.  As an example, if you want AUC~inf,obs~ in subjects who received single doses and AUC~last~ on days 1 (0 to 24 hours) and 10 (216 to 240 hours) in subjects who received multiple doses, with treatment defined as "Drug 1 Single" or "Drug 1 Multiple", the intervals could be defined as below.

```{r select_group}
intervals_manual <-
  data.frame(
    treatment=c("Drug 1 Single", "Drug 1 Multiple", "Drug 1 Multiple"),
    start=c(0, 0, 216),
    end=c(Inf, 24, 240),
    aucinf.obs=c(TRUE, FALSE, FALSE),
    auclast=c(FALSE, TRUE, TRUE)
  )
knitr::kable(intervals_manual)
```

# Intervals

Intervals are defined by `data.frame`s with one row per interval, zero or more columns to match the groups from the `PKNCAdata` object, and one or more NCA parameters to calculate.

Selection of points within an interval occurs by choosing any point at or after the `start` and at or before the `end`.

## To Infinity

The end of an interval may be infinity.  An interval to infinity works the same as any other interval in that points are selected by being at or after the `start` and at or before the `end` of the interval.  Selecting `Inf` or any value at or after the maximum time yields no difference in effect, but `Inf` is simpler when scripting to ensure that all points are selected.

```{r infinity_interval_plot, echo=FALSE}
# Simulate concentration-time data and setup the PKNCAconc object
d_conc <-
  PKNCAconc(
    generate.conc(nsub=1,
                  ntreat=1,
                  time.points=c(0, 1, 2, 4, 6, 8, 12, 24, 36, 48),
                  nstudies=1,
                  nanalytes=1,
                  resid=0),
    conc~time|treatment+ID)
print(d_conc)
```
```{r infinity_interval_plot-visualization, eval=requireNamespace("ggplot2")}
# Use superposition to simulate multiple doses
ggplot(as.data.frame(d_conc)[as.data.frame(d_conc)$time <= 48,], aes(x=time, y=conc)) +
  geom_ribbon(data=as.data.frame(d_conc),
              aes(ymax=conc, ymin=0),
              fill="skyblue") +
  geom_point() + geom_line() +
  scale_x_continuous(breaks=seq(0, 72, by=12)) +
  scale_y_continuous(limits=c(0, NA)) +
  labs(x="Time Since First Dose (hr)",
       y="Concentration\n(arbitrary units)")
```
```{r infinity_interval_spec}
intervals_manual <-
  data.frame(
    start=0,
    end=Inf,
    auclast=TRUE,
    aucinf.obs=TRUE
  )
print(intervals_manual)

my.data <- PKNCAdata(d_conc, intervals=intervals_manual)
```

## Multiple Intervals

More than one interval may be specified for the same subject or group of subjects by providing more than one row of interval specifications.  In the figure below, the blue and green shaded regions indicate the first and second rows of the intervals, respectively.

```{r multiple_intervals_plot, echo=FALSE}
# Simulate concentration-time data and setup the PKNCAconc object
d_conc <-
  PKNCAconc(
    generate.conc(nsub=1,
                  ntreat=1,
                  time.points=c(0, 1, 2, 4, 6, 8, 12, 24, 36, 48),
                  nstudies=1,
                  nanalytes=1,
                  resid=0),
    conc~time|treatment+ID)
print(d_conc)
# Use superposition to simulate multiple doses
d_conc_multi <-
  superposition(d_conc,
                tau=168,
                dose.times=seq(0, 168-24, by=24),
                n.tau=1)
```
```{r multiple_intervals_plot-visualization, eval=requireNamespace("ggplot2")}
# Plot the concentration-time data and the interval
ggplot(d_conc_multi, aes(x=time, y=conc)) +
  geom_ribbon(data=d_conc_multi[d_conc_multi$time <= 24,],
              aes(ymax=conc, ymin=0),
              fill="skyblue") +
  geom_ribbon(data=d_conc_multi[d_conc_multi$time >= 144,],
              aes(ymax=conc, ymin=0),
              fill="lightgreen") +
  geom_point() + geom_line() +
  scale_x_continuous(breaks=seq(0, 168, by=12)) +
  scale_y_continuous(limits=c(0, NA)) +
  labs(x="Time Since First Dose (hr)",
       y="Concentration\n(arbitrary units)")
```
```{r multiple_intervals_spec}
intervals_manual <-
  data.frame(
    start=c(0, 144),
    end=c(24, 168),
    auclast=TRUE
  )
knitr::kable(intervals_manual)

my.data <- PKNCAdata(d_conc, intervals=intervals_manual)
```

# Overlapping Intervals and Different Calculations by Interval

In some scenarios, multiple intervals may be needed where some intervals overlap.  There is no issue with an interval specification that has two rows with overlapping times; the rows are considered separately.  In the example below, the 0-24 interval is shared between both the first and second (shaded blue-green).

The example of overlapping intervals also illustrates that different calculations can be performed in different intervals.  In this case, `auclast` is calculated in both intervals while `aucinf.obs` is only calculated in the 0-Inf interval.

```{r overlapping_intervals_plot, echo=FALSE}
# Simulate concentration-time data and setup the PKNCAconc object
d_conc <-
  PKNCAconc(
    generate.conc(nsub=1,
                  ntreat=1,
                  time.points=c(0, 1, 2, 4, 6, 8, 12, 24, 36, 48),
                  nstudies=1,
                  nanalytes=1,
                  resid=0),
    conc~time|treatment+ID)
print(d_conc)
```
```{r overlapping_intervals_plot-visualization, eval=requireNamespace("ggplot2")}
# Use superposition to simulate multiple doses
ggplot(as.data.frame(d_conc), aes(x=time, y=conc)) +
  geom_ribbon(data=as.data.frame(d_conc),
              aes(ymax=conc, ymin=0),
              fill="lightgreen",
              alpha=0.5) +
  geom_ribbon(data=as.data.frame(d_conc)[as.data.frame(d_conc)$time <= 24,],
              aes(ymax=conc, ymin=0),
              fill="skyblue",
              alpha=0.5) +
  geom_point() + geom_line() +
  scale_x_continuous(breaks=seq(0, 168, by=12)) +
  scale_y_continuous(limits=c(0, NA)) +
  labs(x="Time Since First Dose (hr)",
       y="Concentration\n(arbitrary units)")
```
```{r overlapping_intervals_spec}
intervals_manual <-
  data.frame(
    start=0,
    end=c(24, Inf),
    auclast=TRUE,
    aucinf.obs=c(FALSE, TRUE)
  )
knitr::kable(intervals_manual)

my.data <- PKNCAdata(d_conc, intervals=intervals_manual)
```

# Intervals with Duration

Some events have durations of times rather than instants in time associated with them.  Two typical examples of duration data in NCA are intravenous infusions and urine or fecal sample collections.  Inform PKNCA of durations with the `duration` argument to the `PKNCAdose` and `PKNCAconc` functions.

Durations data are selected based on both the beginning and ending of the duration existing within the interval.

```{r interval_yes_no, echo=FALSE, eval=requireNamespace("ggplot2")}
ggplot_intervals <- function(definition, intervals) {
  intervals$within <-
    c("No", "Yes")[intervals$interval.start >= min(definition$start.x) &
                     intervals$interval.end >= max(definition$start.x)]
  intervals$within <-
    factor(intervals$within, levels=c("Yes", "No"))
  ggplot(intervals,
         aes(x=interval.start,
             xend=interval.end,
             y=y,
             yend=y,
             linetype=within)) +
    geom_segment() +
    geom_segment(data=definition,
                 mapping=aes(x=start.x, xend=start.x, y=start.y, yend=end.y),
                 arrow=arrow(),
                 inherit.aes=FALSE) +
    geom_point(aes(x=interval.start, y=1),
               shape="|",
               size=4) +
    geom_point(aes(x=max(interval.end), y=1),
               shape="|",
               size=4) +
    scale_y_continuous(breaks=NULL) +
    labs(x="Time",
         y=NULL)
}

interval_definition <-
  data.frame(start.x=c(0, 24),
             start.y=0.5,
             end.y=0.9)
show_intervals <-
  data.frame(interval.start=c(0, 4, 12, 24),
             interval.end=c(4, 12, 24, 48),
             y=1)
ggplot_intervals(interval_definition,
                 show_intervals)

interval_definition <-
  data.frame(start.x=c(0, 16),
             start.y=0.5,
             end.y=0.9)
show_intervals <-
  data.frame(interval.start=c(0, 4, 12, 24),
             interval.end=c(4, 12, 24, 48),
             y=1)
ggplot_intervals(interval_definition,
                 show_intervals)

```

# Parameters Available for Calculation in an Interval

The following parameters are available in an interval.  For more information about the parameter, see the documentation for the function.

```{r parameters-available, echo=FALSE}
interval_spec <- get.interval.cols()
interval_spec <- interval_spec[sort(names(interval_spec))]

get_function_for_calc <- function(x) {
  if (is.na(x$FUN)) {
    if (is.null(x$depends)) {
      "(none)"
    } else {
      paste("See the parameter name", x$depends)
    }
  } else {
    x$FUN
  }
}

kable(
  data.frame(
    `Parameter Name`=names(interval_spec),
    `Unit Type`=vapply(X = interval_spec, FUN = "[[", "unit_type", FUN.VALUE = ""),
    `Parameter Description`=vapply(X = interval_spec, FUN = "[[", "desc", FUN.VALUE = ""),
    `Function for Calculation`=vapply(X = interval_spec, FUN = get_function_for_calc, FUN.VALUE = ""),
    check.names=FALSE,
    stringsAsFactors=FALSE
  ),
  row.names=FALSE
)
```

---
title: "Half-life calculation with Tobit regression"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Half-life calculation with Tobit regression}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r pre-setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup, include=FALSE}
library(PKNCA)
```

# Half-life calculation with Tobit regression

Half-life calculation with Tobit regression allows inclusion of concentrations
that are below the lower limit of quantification in the half-life estimate.

## Comparison to semi-log regression

Typical half-life calculation uses curve-stripping semi-log regression of the
natural logarithm of the concentration by time.  Based on the use of the
logarithm of the concentration, concentrations below the lower limit of
quantification (LLOQ) which are set to zero are ignored.

Tobit regression allows inclusion of the concentrations below the LLOQ in the
half-life calculation.  Tobit regression for half-life is equivalent to using
Beal's M3 method in population pharmacokinetic (PK) models.

With Tobit regression, a line is fit using maximum likelihood.  For points above
the LLOQ, the likelihood is based on the probability density at the observed
concentration.  For points below the LLOQ, the likelihood is based on the
cumulative probability distribution function from negative infinity to the limit
of quantification.

## Automatic point selection

## Automatic point selection with semi-log regression

With semi-log regression, the typical method used to automatically select
concentrations for inclusion in the half-life estimate is to:

1. Omit all concentrations that are missing.
2. Omit all concentrations that are below the LLOQ.
3. Estimate the half-life for each set of points from the first concentration
   measure after T~max~ to the third measure before T~last~.
4. Select the best half-life with the following criteria, in order:
    1. The adjusted r-squared must be within a tolerance factor (typically
       0.0001) of the largest adjusted r-squared.
    2. The $\lambda_z$ value (slope for the half-life line) must be positive;
       in other words, the half-life slope must be decreasing.
    3. If multiple choices of points fit the above criteria, choose the one with
       the most concentration measurements included.

For comparison with PKNCA, note that Phoenix WinNonlin switches the order for
selection of 4.1 and 4.2 above.  So, if the best adjusted r-squared is for an
increasing slope but there is another adjusted r-squared with a decreasing
slope, Phoenix will report the half-life.

## Automatic point selection with Tobit regression

With Tobit regression, the method is generally similar to the semi-log
regression with two changes.  The first change is that concentrations below the
LLOQ are retained in the estimate.  The second change is that the adjusted
r-squared is not possible to calculate when including points below the LLOQ, so
the minimum standard deviation estimate is used.

The selection method below results in effectively the same estimates for
half-life when all points are above the LLOQ and improved estimates for
half-life when some points are below the LLOQ.  Future research may investigate
optimization of this method.

The steps for Tobit regression are:

1. Omit all concentrations that are missing.
2. Estimate the half-life for each set of points from the first concentration
   measure after T~max~ to the third measure before T~last~ while including all
   points below the LLOQ after T~last~.
3. Select the best half-life with the following criteria, in order:
    1. The estimated standard deviation of the slope is minimized.
    2. The $\lambda_z$ value (slope for the half-life line) must be positive;
       in other words, the half-life slope must be decreasing.

# Comparison of Tobit and semi-log regression

In almost all scenarios, Tobit regression using the algorithm above improves the
half-life estimate compared to semi-log regression.  In the figure below,
concentration-time profiles were simulated with 1-, 2-, and 3-compartment linear
PK models with intravenous or extravascular administration and a variety of
compartmental model parameters.  The true half-life was calculated based on the
compartmental model parameters.  Then, the ratio of the estimated to true
half-life was calculated.  Values closer to 1 indicate a better fit and values
farther from 1 indicate a poorer fit.

![The empirical cumulative distribution function for ratio of estimated to theoretical half-lives for Tobit regression (aqua line) and semi-log regression (red line) are shown.](v06-half-life-calculation-tobit_figure_1.svg)

Tobit regression performs universally better than least-squares up to the
estimated theoretical half-life, and better at >2-fold above the theoretical
half-life while least-squares performs slightly better between the theoretical
and 2-fold above.  The fact that the Tobit regression cumulative distribution
function is closer to 1 across the range of simulations indicates that Tobit
regression provides better half-life estimate across a broad range of data.

---
title: "Data Imputation"
author: "Bill Denney"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Data Imputation}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

Imputation may be required for noncompartmental analysis (NCA) calculations.
Typical imputations may require setting the concentration before the first dose
to zero or shifting actual time predose concentrations to the beginning of the
dosing interval.

PKNCA supports imputation either for the full analysis dataset or per
calculation interval.

The current list of imputation methods built into PKNCA can be found by looking
at `?PKNCA_impute_method`:

```{r results='markup'}
library(PKNCA)
cat(paste(
  "*", ls("package:PKNCA", pattern = "^PKNCA_impute_method")
), sep = "\n")
```

## How does imputation occur?

(You can skip this section if you don't desire the details of the methods of
imputation.)

Imputation occurs just before calculations are performed within PKNCA.
Imputation occurs only on a single interval definition at a time, so the same
group (usually meaning the same subject with the same analyte) at the same time
range can have different imputations for different parameter calculations.

The reason that this is done is to ensure that there are no unintentional
modifications to the data.  As an example, if an AUC~0-24~ were calculated on
Day 1 and Day 2 of a study with actual times, the nominal 24 hour sample may be
collected at 23.5 hours.  It may be preferable to keep the 23.5 hour sample at
23.5 hours for the Day 1 calculation, and at the same time, it may be preferred
to shift the same 23.5 hr sample to 24 hours (time 0 on Day 2) for the Day 2
calculation.

## How to select imputation methods to use

The selection of imputation methods uses a string of text with commas or spaces
(or both) separating the imputation methods to use.  No imputation will be
performed if the imputation method is requested as `NA` or `""`.

* To select no imputation (the default), indicate the imputation by `NA` or
  `""`.
* To set imputation on the full dataset, use the `impute` argument to
  `PKNCAdata()` to specify the methods to use.
* To set imputation by interval, use the `impute` argument to `PKNCAdata()` to
  specify the column in the intervals dataset to use for imputation.
* You cannot specify imputation for both the full dataset and by interval at the
  same time.  And, if a column name in the dataset matches the `impute` argument
  to `PKNCAdata()`, that will be used.

Imputation method functions are named `PKNCA_impute_method_[method name]`.  For
example, the method to impute a concentration of 0 at time 0 is named
`PKNCA_impute_method_start_conc0`.  When specifying the imputation method to
use, give the `[method name]` part of the function name.  So for the example
above, use `"start_conc0"`.

To specify more than one, give all the methods in order with a comma or space
separating them.  For example, to first move a predose concentration up to the
time of dosing and then set time 0 to concentration 0, use
`"start_predose,start_conc0"`, and the two methods will be applied in order.

## Imputation for the full dataset

If an imputation applies to the full dataset, it can be provided in the `impute`
argument to `PKNCAdata()`:

```{r impute-full-data}
library(PKNCA)
# Remove time 0 to illustrate that imputation works
d_conc <- as.data.frame(datasets::Theoph)[!datasets::Theoph$Time == 0, ]
conc_obj <- PKNCAconc(d_conc, conc~Time|Subject)
d_dose <- unique(datasets::Theoph[datasets::Theoph$Time == 0,
                                  c("Dose", "Time", "Subject")])
dose_obj <- PKNCAdose(d_dose, Dose~Time|Subject)
data_obj <- PKNCAdata(conc_obj, dose_obj, impute = "start_predose,start_conc0")
nca_obj <- pk.nca(data_obj)
summary(nca_obj)
```

## Imputation by calculation interval

If an imputation applies to specific intervals, the column in the interval
data.frame can be provided in the `impute` argument to `PKNCAdata()`:

```{r impute-by-interval}
library(PKNCA)
# Remove time 0 to illustrate that imputation works
d_conc <- as.data.frame(datasets::Theoph)[!datasets::Theoph$Time == 0, ]
conc_obj <- PKNCAconc(d_conc, conc~Time|Subject)
d_dose <- unique(datasets::Theoph[datasets::Theoph$Time == 0,
                                  c("Dose", "Time", "Subject")])
dose_obj <- PKNCAdose(d_dose, Dose~Time|Subject)

d_intervals <-
  data.frame(
    start=0, end=c(24, 24.1),
    auclast=TRUE,
    impute=c(NA, "start_conc0")
  )

data_obj <- PKNCAdata(conc_obj, dose_obj, intervals = d_intervals, impute = "impute")
nca_obj <- pk.nca(data_obj)
# PKNCA does not impute time 0 by default, so AUClast in the 0-24 interval is
# not calculated
summary(nca_obj)
```

## Advanced: Writing your own imputation functions

Writing your own imputation function is intended to be a simple process.  To
create an imputation function requires the following steps:

1. Write a function where the name starts with `PKNCA_impute_method_` and the
  remainder of the function name is a brief description of the method.  (Such as
  `PKNCA_impute_method_start_conc0`.)
2. The function should have 4 arguments:  `conc`, `time`, `...`, and `options`.
3. The function should return a single data.frame with two columns named `conc`
   and `time`.  The rows in the data.frame must be sorted by `time`.

In addition to the above, the function may take named arguments of:

* `start` and `end` to indicate the start and end time of the interval, and
* `conc.group` and `time.group` to indicate the concentrations and times that
  have not been filtered for the interval.

---
title: "PKNCA Training Sessions"
author: "William Denney"
date: "18 October 2023"
output:
  ioslides_presentation:
    widescreen: true
vignette: >
  %\VignetteIndexEntry{PKNCA Training Sessions}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = FALSE)
requireNamespace("pmxTools")
library(PKNCA)
library(dplyr)
library(ggplot2)
library(tidyr)
library(purrr)
breaks_hours <- function(n=5, Q=c(1, 6, 4, 12, 2, 24, 168), ...) {
  n_default <- n
  Q_default <- Q
  function(x, n = n_default, Q=Q_default) {
    x <- x[is.finite(x)]
    if (length(x) == 0) {
      return(numeric())
    }
    rng <- range(x)
    labeling::extended(rng[1], rng[2], m=n, Q=Q, ...)
  }
}

scale_x_hours <- function(..., breaks=breaks_hours()) {
  ggplot2::scale_x_continuous(..., breaks=breaks)
}
```

<style>
.forceBreak { -webkit-column-break-after: always; break-after: column; }

<!-- from https://stackoverflow.com/questions/1909648/stacking-divs-on-top-of-each-other -->
.container {
  width: 300px;
  height: 300px;
  margin: 0 auto;
  background-color: yellow;
  /* important part */
  display: grid;
  place-items: center;
  grid-template-areas: "inner-div";
}

.inner {
  /* important part */
  grid-area: inner-div;
}

.bigStrikethroughOuter {
  <!-- text-decoration: line-through; -->
  font-size: 8em;
  text-align: center;
  color: red;
  background: 
  linear-gradient(to top left,
    transparent 0%,
    transparent calc(50% - 0.05em),
    red calc(50% - 0.05em),
    red 50%,
    red calc(50% + 0.05em),
    transparent calc(50% + 0.05em),
    transparent 100%),
  linear-gradient(to top right,
    transparent 0%,
    transparent calc(50% - 0.05em),
    red calc(50% - 0.05em),
    red 50%,
    red calc(50% + 0.05em),
    transparent calc(50% + 0.05em),
    transparent 100%);
}

.bigStrikethroughInner {
  color: blue;
  font-size: 8em;
}

.autoImageWidth {
  width: auto !important; /*override the width below*/
  max-width: 100%;
  float: left;
  clear: both;
  text-align: center;
}

<!-- imessages -->

.imessage {
  font-family: helvetica;
  display: flex ;
  flex-direction: column;
  align-items: center;
}

.chat {
  width: 300px;
  border: solid 1px #EEE;
  display: flex;
  flex-direction: column;
  padding: 10px;
}

.messages {
  margin-top: 30px;
  display: flex;
  flex-direction: column;
}

.message {
  border-radius: 20px;
  padding: 8px 15px;
  margin-top: 5px;
  margin-bottom: 5px;
  display: inline-block;
}

.yours {
  align-items: flex-start;
}

.yours .message {
  margin-right: 25%;
  background-color: #eee;
  position: relative;
}

.yours .message.last:before {
  content: "";
  position: absolute;
  z-index: 0;
  bottom: 0;
  left: -7px;
  height: 20px;
  width: 20px;
  background: #eee;
  border-bottom-right-radius: 15px;
}
.yours .message.last:after {
  content: "";
  position: absolute;
  z-index: 1;
  bottom: 0;
  left: -10px;
  width: 10px;
  height: 20px;
  background: white;
  border-bottom-right-radius: 10px;
}

.mine {
  align-items: flex-end;
}

.mine .message {
  color: white;
  margin-left: 25%;
  background: linear-gradient(to bottom, #00D0EA 0%, #0085D1 100%);
  background-attachment: fixed;
  position: relative;
}

.mine .message.last:before {
  content: "";
  position: absolute;
  z-index: 0;
  bottom: 0;
  right: -8px;
  height: 20px;
  width: 20px;
  background: linear-gradient(to bottom, #00D0EA 0%, #0085D1 100%);
  background-attachment: fixed;
  border-bottom-left-radius: 15px;
}

.mine .message.last:after {
  content: "";
  position: absolute;
  z-index: 1;
  bottom: 0;
  right: -10px;
  width: 10px;
  height: 20px;
  background: white;
  border-bottom-left-radius: 10px;
}

</style>

# Introduction to PKNCA and Basics of Its Use

Creation of these materials were partially supported by funding from the Metrum Research Group.

## Introduction to PKNCA {.build .smaller}

PKNCA is a tool for calculating noncompartmental analysis (NCA) results for pharmacokinetic (PK) data.

... but, you already knew that or you wouldn't be here.

PKNCA has several foci:

* be regulatory-ready
    * it has approximately 100% test coverage.
* be reproducible
    * it has a focus on being scriptable.
* get the right answer or none at all
    * it will try to know what you want,
    * but all decisions can be overridden, and
    * if there is a question that may cause an error or an unanticipated result, either no result will output or an error will be raised.

## Enjoy! {.build}

I hope that you have a whale of a good time during this training.

![](https://apps-afsc.fisheries.noaa.gov/Quarterly/amj2005/images/killerwhales.jpg)

(Foreshadowing...)

## Some NCA Definitions

* **C~max~**: The maximum observed concentration
* **T~max~**: The time of the maximum observed concentration
* **t~last~**: The time of the last concentration above the limit of quantification
* **AUC**: Area under the concentration-time curve.  Some important AUC variants are:
    * **AUC~last~**: AUC from time zero to t~last~
    * **AUC~int~**: AUC from time zero to the end of an interval of time, often extrapolated or interpolated (e.g. AUC~0-24hr~)
    * **AUC~&infin;~**: AUC from time zero to t~last~ then extrapolated from t~last~ to time infinity using the half life

# Dataset Basics

## NCA Data are Not Tidy ***as a Single Dataset***

"Tidy datasets... have a specific structure: each variable is a column, each observation is a row, and each type of observational unit is a table." - Hadley Wickham (https://doi.org/10.18637/jss.v059.i10)

CDISC has NCA tidied, and PKNCA follows that model:

* concentration-time is a dataset (PC domain; `PKNCAconc()` object)
* dose-time is a dataset (EX/EC domains; `PKNCAdose()` object)
* NCA results are a dataset (PP domain; `pk.nca()` output)

## Dataset Basics: Minimum data

PKNCA requires at least the concentration, time, and what you want to calculate.

```{r fig.width=6, fig.height=4}
conc <-
  datasets::Theoph %>%
  filter(Subject %in% 1)
ggplot(conc, aes(x=Time, y=conc)) +
  geom_line() +
  scale_x_hours()
```

## Dataset Basics: What columns are needed?

Column names are provided by the input to `PKNCAconc()` and `PKNCAdose()`; they are not hard-coded.

Columns that can be used include:

* `PKNCAconc()`: concentration, time, groups; data exclusions; half-life inclusion and exclusion
* `PKNCAdose()`: dose, time, groups; route, rate/duration of infusion; data exclusions
* intervals given to `PKNCAdata()`: groups, start, end, and any NCA parameters to calculate

## Dataset Basics:  Example data

In the following slides, abbreviated data from an example study where two treatments ("A" and "B") are administered to two subjects (1 and 2).

* For PKNCA, the groups will be **Treatment** and **Subject**.
    * PKNCA considers groups in order with the subject identifier as the last group (or the last group before a forward slash, `/`, if `/` is present).
    * When indicated in order (`...|Treatment+Subject`), PKNCA automatically knows to keep **Treatment** and drop **Subject** for summaries (more on that later).

## Dataset Basics:  Example concentration data {.columns-2}

```{r}
conc_data <-
  withr::with_seed(5, {
    data.frame(
      Subject=rep(1:2, each=6),
      Treatment=rep(c("A", "B", "A", "B"), each=3),
      Time=rep(c(0, 2, 8), 4),
      Conc=rep(c(0, 2, 0.5), 4)*exp(rnorm(n=12, sd=0.05))
    )
  })
```

```{r}
pander::pander(conc_data %>% filter(Subject == 1))
```

<p class="forceBreak"></p>

```{r}
pander::pander(conc_data %>% filter(Subject == 2))
```

## Dataset Basics:  Example dosing data

```{r}
dose_data <-
  data.frame(
    Subject=rep(1:2, each=2),
    Treatment=c("A", "B", "A", "B"),
    Time=0,
    Dose=10
  )
```

```{r}
pander::pander(dose_data %>% filter(Subject == 1))
```

<p class="forceBreak"></p>

```{r}
pander::pander(dose_data %>% filter(Subject == 2))
```

## Dataset Basics:  Example interval data

```{r, echo=TRUE}
d_interval_1 <-
  data.frame(
    start=0, end=8,
    cmax=TRUE, tmax=TRUE, auclast=TRUE
  )
```

```{r}
pander::pander(d_interval_1)
```

Groups are not required, if you want the same intervals calculated for each group.

## Hands-on: First NCA calculation with PKNCA

```{r echo=TRUE}
library(dplyr)
library(ggplot2)
library(tidyr)
library(purrr)
library(PKNCA)
# Concentration data setup
d_conc <-
  datasets::Theoph %>%
  filter(Subject %in% 1)
o_conc <- PKNCAconc(conc~Time, data=d_conc)
# Setup intervals for calculation
d_intervals <- data.frame(start=0, end=24, cmax=TRUE, tmax=TRUE,
                          auclast=TRUE, aucint.inf.obs=TRUE)
# Combine concentration and dose
o_data <- PKNCAdata(o_conc, intervals=d_intervals)
# Calculate the results (suppressMessages() hides a message that isn't needed now)
o_result <- suppressMessages(pk.nca(o_data))
# summary(o_result)
```

# PKNCA Functions

## What functions are the most used?

* `PKNCAconc()`: define a concentration-time `PKNCAconc` object
    * All information about concentration data are given:  concentration, time
    * Optional information includes: grouping information (usually given), data to exclude, half-life inclusion and exclusion columns
* `PKNCAdose()`: define a dose-time `PKNCAdose` object (optional)
    * dose amount and time are both optional
    * Optional information includes: rate or duration of infusion, data to exclude
* `PKNCAdata()`: combine `PKNCAconc`, optionally `PKNCAdose`, and optionally `intervals` into a `PKNCAdata` object
    * the `PKNCAconc` object must be given; the `PKNCAdose` object is optional; interval definitions are usually given; calculation options may be given
* `pk.nca()`: calculate the NCA parameters from a data object into a `PKNCAresult` object

## How do I do a simple calculation? all steps

We will break this down in subsequent slides.

```{r echo=TRUE}
# Concentration data setup
d_conc <-
  datasets::Theoph %>%
  filter(Subject %in% 1)
o_conc <- PKNCAconc(conc~Time, data=d_conc)
# Dose data setup
d_dose <-
  datasets::Theoph %>%
  filter(Subject %in% 1) %>%
  filter(Time == 0)
o_dose <- PKNCAdose(Dose~Time, data=d_dose)
# Combine concentration and dose
o_data <- PKNCAdata(o_conc, o_dose)
# Calculate the results
o_result <- pk.nca(o_data)
```

## How do I do a simple calculation? Concentration data {.smaller}

```{r echo=TRUE}
# Load your dataset as a data.frame
d_conc <-
  datasets::Theoph %>%
  filter(Subject %in% 1)
# Take a look at the data
pander::pander(head(d_conc, 2))
# Define the PKNCAconc object indicating the concentration and time columns, the
# dataset, and any other options.
o_conc <- PKNCAconc(conc~Time, data=d_conc)
```

## How do I do a simple calculation? Dose data {.smaller}

```{r echo=TRUE}
# Load your dataset as a data.frame
d_dose <-
  datasets::Theoph %>%
  filter(Subject %in% 1) %>%
  filter(Time == 0)
# Take a look at the data
pander::pander(d_dose)
# Define the PKNCAdose object indicating the dose amount and time columns, the
# dataset, and any other options.
o_dose <- PKNCAdose(Dose~Time, data=d_dose)
```

## How do I do a simple calculation? Calculate results {.smaller}

```{r echo=TRUE}
# Combine the PKNCAconc and PKNCAdose objects.  You can add interval
# specifications and calculation options here.
o_data <- PKNCAdata(o_conc, o_dose)
# Calculate the results
o_result <- pk.nca(o_data)
```

## How do I do a simple calculation?  Get results

To calculate summary statistics, use `summary()`; to extract all individual-level results, use `as.data.frame()`.

The `"caption"` attribute of the summary describes how the summary statistics were calculated for each parameter.  (Hint: `pander::pander()` knows how to use that to put the caption on a table in a report.)

The individual results contain the columns for start time, end time, grouping variables (none in this example), parameter names, values, and if the value should be excluded.

## How do I do a simple calculation?  Get summary results {.smaller}

```{r echo=TRUE}
# Look at summarized results
pander::pander(summary(o_result))
```

## How do I do a simple calculation?  Get individual results {.smaller}

Use `as.data.frame()` to get the individual NCA parameter results.

```{r echo=TRUE}
# Look at individual results
pander::pander(head(
  as.data.frame(o_result),
  n=3
))
```

# PKNCA datasets

## How does PKNCA think about data?

Three types of data are inputs for calculation in PKNCA:

* concentration-time (`PKNCAconc`),
* dose-time (`PKNCAdose`), and
* intervals.

`PKNCAconc` and `PKNCAdose` objects can optionally have groups.  The groups in a `PKNCAdose` object must be the same or fewer than the groups in `PKNCAconc` object (for example, all subjects in a treatment arm may receive the same dose).

## What is an "interval" and how is it different than a "group"? {.columns-2 .smaller}

```{r interval-vs-groups-setup, echo=FALSE}
last_dose_time <- 24
dose_interval <- 8
dose_times <- seq(0, last_dose_time-dose_interval, by=dose_interval)
d_conc_superposition <-
  superposition(
    o_conc,
    dose.times=dose_times,
    tau=last_dose_time,
    check.blq=FALSE,
    n.tau=1
  )
```

A **group** separates one full concentration-time profile for a subject that you may ever want to consider at the same time.  Usually, it groups by study, treatment, analyte, and subject (other groups can be useful depending on the study design).

An **interval** selects a time range within a **group**.

One time can be in zero or more intervals, but only zero or one group.  Intervals can be adjacent (0-12 and 12-24) or overlap (0-12 and 0-24).  In other words, one sample may be used in more than one interval, but one sample will never be used in more than one group.

**Legend:**  The group contains all points on the figure.  Shaded regions indicate intervals.  Arrows indicate points shared between intervals within the group.

<p class="forceBreak"></p>

```{r fig.width=4, fig.height=4}
d_intervals <-
  tibble(
    start=dose_times,
    end=dose_times + dose_interval
  ) %>%
  mutate(
    name=sprintf("Interval %g", row_number()),
    height=max(d_conc_superposition$conc)*1.03,
    width=dose_interval,
    x=(start+end)/2,
    y=height/2
  )
d_interval_arrows <-
  d_conc_superposition %>%
  filter(time != 0 & time %in% dose_times) %>%
  mutate(
    name1=sprintf("Interval %g", row_number()),
    name2=sprintf("Interval %g", row_number() + 1),
  )
ggplot(d_conc_superposition, aes(x=time, y=conc)) +
  geom_tile(
    data=d_intervals,
    aes(x=x, y=y, width=width, height=height, colour=name, fill=name),
    alpha=0.2,
    inherit.aes=FALSE,
    show.legend=FALSE
  ) +
  geom_segment(
    data=d_interval_arrows,
    aes(x=time - 0.8, xend=time - 0.1, y=conc-2.1, yend=conc - 0.1, colour=name2),
    arrow=arrow(length=unit(0.1, "inches")),
    inherit.aes=FALSE,
    show.legend=FALSE
  ) +
  geom_segment(
    data=d_interval_arrows,
    aes(x=time + 0.8, xend=time + 0.1, y=conc-2.1, yend=conc - 0.1, colour=name1),
    arrow=arrow(length=unit(0.1, "inches")),
    inherit.aes=FALSE,
    show.legend=FALSE
  ) +
  geom_line() +
  geom_point() +
  scale_x_hours() +
  labs(
    title=sprintf("Dosing Q%gH", dose_interval)
  )
```

## Common data management requirements before sending data to PKNCA {.smaller}

1. Time must not be missing for `PKNCAconc` (if given to `PKNCAdose`, it must not be missing).
2. Below the limit of quantification (BLQ) concentrations must be set to zero (not `NA`).
3. Imputation of time zero is required for AUC calculation.
4. Especially for actual-time calculations, imputation of the beginning of the interval is usually needed.

Columns must be created for:

* Concentration or dose,
* Time
* Groups
    * usually columns for study, treatment arm, subject;
    * sometimes analyte, formulation, period (needed in case the same subject receives the same treatment arm multiple times)

## Setup your concentration data {.columns-2}

* Concentration data must be numeric

<p class="forceBreak"></p>

<div class="bigStrikethroughOuter">
<div class="bigStrikethroughInner">A</div>
</div>

## Setup your concentration data {.columns-2}

* Concentration data must be numeric
* Time must be numeric and not be missing

<p class="forceBreak"></p>

<div class="bigStrikethroughOuter">
<div class="bigStrikethroughInner">NA</div>
</div>

## Setup your concentration data {.columns-2}

* Concentration data must be numeric
* Time must be numeric and not be missing
* Groups can be anything, setup at the level of the individual

<p class="forceBreak"></p>

<div class="autoImageWidth">
![](https://apps-afsc.fisheries.noaa.gov/Quarterly/amj2005/images/killerwhales.jpg)<br />
Group: <span style="color: green;">&#128504;</span> a pod of killer whales
</div>

## Setup your dosing data (if you have it and even if you don't) {.smaller}

Normal dosing data setup: `PKNCAdose(dose~time|actarm+usubjid, data=d_dose)`

* Dose amount must be numeric &mdash; or it can be omitted
    * `PKNCAdose(~time|actarm+usubjid, data=d_dose)`
* Time must be numeric and not be missing &mdash; or it can be omitted
    * `PKNCAdose(dose~.|actarm+usubjid, data=d_dose)`
* Groups can be anything &mdash; may be grouped at a higher level than the individual
    * Useful when all dose amounts and times are the same within an arm: `PKNCAdose(dose~time|actarm, data=d_dose)`
    * Useful dose amount is the same at all times within an arm: `PKNCAdose(dose~.|actarm, data=d_dose)`
    * Useful when times are all the same within an arm but dose may differ: `PKNCAdose(~time|actarm, data=d_dose)`

## Define your intervals

Intervals have columns for:

* `start` and `end` times for the interval,
* groups matching any level of grouping; intervals apply by a merge/join with the groups
* parameters to calculate (`TRUE` means to calculate it; `FALSE` means don't).  The full list of available parameters is in the [selection of calculation intervals vignette](http://humanpred.github.io/pknca/articles/Selection-of-Calculation-Intervals.html#parameters-available-for-calculation-in-an-interval-1).
    * You only have to specify the parameter you want, not all parameters.

## Define your intervals: example

* For time 0 to 24, calculate AUClast
* For time 0 to infinity, calculate cmax, tmax, half.life, and aucinf.obs

```{r}
PKNCA.options("single.dose.aucs") %>%
  select(c(all_of(c("start", "end")), where(~is.logical(.x) && any(.x)))) %>%
  pander::pander()
```

# Calculations above the hood

## Prepare your data for calculation

```{r echo=TRUE}
d_conc <-
  datasets::Theoph %>%
  mutate(
    Treatment=
      case_when(
        Dose <= median(Dose)~"Low dose",
        TRUE~"High dose"
      )
  )
# The study was single-dose
d_dose <-
  d_conc %>%
  select(Treatment, Subject, Dose) %>%
  unique() %>%
  mutate(dose_time=0)
```

## Calculate without dosing data {.build}

```{r echo=TRUE}
o_conc <- PKNCAconc(conc~Time|Treatment+Subject, data=d_conc)
try({
  o_data <- PKNCAdata(o_conc)
  summary(pk.nca(o_data))
})
```

Whoops!  Without dosing, we need intervals.

## Calculate without dosing data, try 2

```{r echo=TRUE}
o_conc <- PKNCAconc(conc~Time|Treatment+Subject, data=d_conc)
d_intervals <- data.frame(start=0, end=Inf, cmax=TRUE, tmax=TRUE,
                          half.life=TRUE, aucinf.obs=TRUE)
o_data_manual_intervals <- PKNCAdata(o_conc, intervals=d_intervals)
summary(pk.nca(o_data_manual_intervals))
```

## Dosing data helps with interval setup

```{r echo=TRUE}
o_conc <- PKNCAconc(conc~Time|Treatment+Subject, data=d_conc)
o_dose <- PKNCAdose(Dose~dose_time|Treatment+Subject, data=d_dose)
o_data_auto_intervals <- PKNCAdata(o_conc, o_dose)
o_data_auto_intervals$intervals$aucint.inf.obs <- TRUE
summary(pk.nca(o_data_auto_intervals))
```

## AUC considerations with PKNCA (1/3) {.columns-2}

```{r auc-considerations-setup, warning=FALSE}
d_conc <-
  datasets::Theoph %>%
  filter(Subject == 1)
o_conc <- PKNCAconc(conc~Time, data=d_conc)
d_interval_int <- data.frame(start=0, end=Inf, half.life=TRUE)
o_data_int <- PKNCAdata(o_conc, intervals=d_interval_int)
o_nca_int <- suppressMessages(pk.nca(o_data_int))
lambda_z_int <-
  o_nca_int %>%
  as.data.frame() %>%
  filter(PPTESTCD %in% "lambda.z") %>%
  "[["("PPORRES")

d_interval_inf <- data.frame(start=0, end=24, half.life=TRUE)
o_data_inf <- PKNCAdata(o_conc, intervals=d_interval_inf)
o_nca_inf <- suppressMessages(pk.nca(o_data_inf))
lambda_z_inf <-
  o_nca_inf %>%
  as.data.frame() %>%
  filter(PPTESTCD %in% "lambda.z") %>%
  "[["("PPORRES")

tlast <- 
  o_nca_inf %>%
  as.data.frame() %>%
  filter(PPTESTCD %in% "tlast") %>%
  "[["("PPORRES")

d_auc_calcs <-
  d_conc %>%
  bind_rows(
    tibble(Time=seq(12, 60))
  ) %>%
  mutate(
    conc_all_int=
      interp.extrap.conc(
        conc=conc[!is.na(conc)],
        time=Time[!is.na(conc)],
        time.out=Time,
        lambda.z=lambda_z_int
      ),
    conc_all_inf=
      interp.extrap.conc(
        conc=conc[!is.na(conc) & Time <= 24],
        time=Time[!is.na(conc) & Time <= 24],
        time.out=Time,
        lambda.z=lambda_z_inf
      ),
    conc_last=
      case_when(
        Time <= 24~conc,
        TRUE~NA_real_
      ),
    conc_int=
      case_when(
        Time <= 24 & Time >= tlast~conc_all_int,
        TRUE~NA_real_
      ),
    conc_inf=
      case_when(
        Time >= tlast~conc_all_inf,
        TRUE~NA_real_
      )
  ) %>%
  arrange(Time)
auc_figure_time_max <- 36
p_auc_calcs <-
  ggplot(d_auc_calcs, aes(x=Time, y=conc)) +
  # AUCinf (with a work-around for https://github.com/tidyverse/ggplot2/issues/4661)
  geom_area(
    data=d_auc_calcs %>% filter(Time <= auc_figure_time_max),
    aes(y=conc_inf, colour="AUCinf", fill="AUCinf"),
    alpha=0.2,
    na.rm=TRUE
  ) +
  geom_line(
    data=d_auc_calcs,
    aes(y=conc_inf, colour="AUCinf"),
    na.rm=TRUE
  ) +
  # AUCint
  geom_area(
    aes(y=conc_int, colour="AUCint", fill="AUCint"),
    alpha=0.2,
    na.rm=TRUE
  ) +
  # AUClast
  geom_area(
    aes(y=conc_last, colour="AUClast", fill="AUClast"),
    na.rm=TRUE
  ) +
  geom_point(show.legend=FALSE,
    na.rm=TRUE) +
  geom_line(show.legend=FALSE,
    na.rm=TRUE) +
  geom_vline(xintercept=24, linetype="63") +
  scale_x_continuous(breaks=seq(0, auc_figure_time_max, by=6)) +
  coord_cartesian(xlim=c(0, auc_figure_time_max)) +
  labs(
    colour="AUC type",
    fill="AUC type"
  )
```

```{r warning=FALSE, out.width="100%"}
p_auc_calcs
```

<p class="forceBreak"></p>

The considerations below mainly apply to actual-time data; nominal-time data usually have measurements at the start and end time for the interval.

With an interval start and end of 0 and 24 (and the last measurement time just after 24 hours):

* **AUC~last~** is calculated only based on points within the interval (the AUClast color in the figure)

## AUC considerations with PKNCA (2/3) {.columns-2}

```{r warning=FALSE, out.width="100%"}
p_auc_calcs
```

<p class="forceBreak"></p>

The considerations below mainly apply to actual-time data; nominal-time data usually have measurements at the start and end time for the interval.

With an interval start and end of 0 and 24 (and the last measurement time just after 24 hours):

* **AUC~int~** looks at the points in the interval, and if there is no measurement at the interval end time, interpolates or extrapolates to the interval end time (the AUClast and AUCint color in the figure)

## AUC considerations with PKNCA (3/3) {.columns-2}

```{r warning=FALSE, out.width="100%"}
p_auc_calcs
```

<p class="forceBreak"></p>

The considerations below mainly apply to actual-time data; nominal-time data usually have measurements at the start and end time for the interval.

With an interval start and end of 0 and 24 (and the last measurement time just after 24 hours):

* **AUC~&infin;~** is calculated based on AUC~last~, t~last~, and the half-life from t~last~, only using data within the interval-- no data after the end of the interval.
* Ensure that the interval used for calculating AUC~&infin;~ includes all the points desired (usually, `end=Inf`).

# Hands-on workshop

## Steady-state intramuscular administration

The data for the exercise are from a PK study of amikacin in a killer whale and a beluga whale.  (DOI: 10.1638/03-078)

![](https://apps-afsc.fisheries.noaa.gov/Quarterly/amj2005/images/killerwhales.jpg)

(Callback...)

## Steady-state intramuscular administration

```{r eval=FALSE, echo=TRUE}
library(PKNCA)

d_conc <- read.csv("c:/tmp/whale_conc.csv")
d_dose <- read.csv("c:/tmp/whale_dose.csv")
head(d_conc)
head(d_dose)

o_conc <- PKNCAconc(concentration~time|Animal, data=d_conc)
o_dose <- PKNCAdose(dose~time|Animal, data=d_dose)
o_data <- PKNCAdata(o_conc, o_dose)
o_data$intervals
o_nca <- pk.nca(o_data)
summary(o_nca)
summary(o_nca, drop.group=c())
as.data.frame(o_nca)
```

# Day 2 Start

# Control your data

## Including and excluding data points

Data may be included/excluded in two ways:

* Overall: excluded a row of data from all analyses
* Half-life: excluded from half-life calculations, but included in all other analyses

For both ways of including/excluding data, it is defined by a column in the input data.  The column is either `NA` or an empty string (`""`) to indicate "no" or any other text to indicate "yes".

## Exclude data points overall {.columns-2}

Use the `exclude` argument for `PKNCAconc()` or `PKNCAdose()`.

When you use `exclude`, this is how you give your data to PKNCA:

```{r exclude-example-1, echo=TRUE}
d_before_exclude <-
 data.frame(
  time=0:4,
  conc=c(0, 2, 1, 0.5, 0.25),
  not_this=c(NA, "Not this", rep(NA, 3))
 )
o_conc <-
 PKNCAconc(
  data=d_before_exclude,
  conc~time,
  exclude="not_this"
 )
```

<p class="forceBreak"></p>

And, this is how PKNCA thinks about it:

```{r exclude-example-2, echo=TRUE}
pander::pander(
  d_before_exclude %>%
    filter(is.na(not_this))
)
```

## Exclude data points overall

```{r, echo=TRUE, eval=FALSE}
o_conc <- PKNCAconc(data=d_before_exclude, conc~time, exclude="not_this")
```

<div class="imessage">
<div class="yours messages">
  <div class="message last">Hey babe, did you get my 5 rows of data?</div>
</div>
<div class="mine messages">
  <div class="message last">I only saw 4 rows.  Are you sure you sent 5?</div>
</div>
<div class="yours messages">
  <div class="message last">Yep, definitely 5 check that last slide. &#128544;</div>
</div>
<div class="mine messages">
  <div class="message">...</div>
  <div class="message last">New phone.  Who dis?</div>
</div>
</div>

## Digression: How is &lambda;z automatically calculated? {.smaller}

* Filter the data from the first point after t~max~ (or from t~max~ if `allow.tmax.in.half.life=TRUE`) to t~last~ and excluding BLQ in the middle.
* Fit the semi-log line from 3 points before t~last~ (3 can be changed with the `min.hl.points` option) to t~last~.
    * Repeat for all sets of points from there to the first point included.
    * If that 3 points are not available, it is not calculated.
* Among the fits, select the best adjusted r^2^ (within a tolerance of `adj.r.squared.factor`).
* Require &lambda;z` > 0`.
* If more than one fit is available at this point, select the one with the most points included.

Note: WinNonlin first requires &lambda;z` > 0` then selects for adjusted r^2^.  Therefore, WinNonlin will occasionally provide a half-life when PKNCA will not, but the fit line is not as good (as measured by r^2^).  The selection of filtering order is an intentional feature with PKNCA, and it generally has minimal impact on summary statistics because the quality of the half-life fit is usually low in this scenario.

## &lambda;z control (manual exclusions and inclusions of data points)

Use the `exclude_half.life` or `include_half.life` argument for `PKNCAconc()`.  The two arguments behave very differently in how points are selected for half-life.

`exclude_half.life` uses the same automatic point selection method of curve stripping (described before), but it excludes individual points from that calculation.

`include_half.life` uses no automatic point selection method, and only points specifically noted by the analyst are included.

# Less-common calculations

## Urine calculations

```{r echo=TRUE}
d_urine <-
  data.frame(
    conc=c(1, 2, 3),
    urine_volume=c(200, 100, 300),
    time=c(1, 2, 3)
  )
o_conc <- PKNCAconc(data=d_urine, conc~time, volume="urine_volume")
d_intervals <- data.frame(start=0, end=24, ae=TRUE)
o_data <- PKNCAdata(o_conc, intervals=d_intervals)
o_nca <- suppressMessages(pk.nca(o_data))
summary(o_nca)
```

## Urine calculations: understanding what is happening and potential hiccups

Intervals for urine are treated the same as any other interval type.  Specifically, PKNCA does not look outside the start and end of the interval.

* Watch out for e.g. a 24-hour urine amount to be included in more than one interval because start = 0 and end = 24.
* Watch out for an actual start or end time to be outside of the interval and therefore to be omitted from calculations.

# Calculations below the hood

## PKNCA only calculates what is required, not every possible parameter (1 of 2)

If you don't need a parameter, PKNCA won't calculate it.

For example, if all you need is `cmax`, all you'll get is `cmax`.

```{r echo=TRUE}
o_conc <- PKNCAconc(data=data.frame(conc=2^-(1:4), time=0:3), conc~time)
o_data <- PKNCAdata(o_conc, intervals=data.frame(start=0, end=Inf, cmax=TRUE))
o_nca <- suppressMessages(pk.nca(o_data))
as.data.frame(o_nca)
```

## PKNCA only calculates what is required, not every possible parameter (2 of 2) {.columns-2 .smaller}

If you need AUC~0-\infty~, PKNCA will calculate other required parameters behind the scenes.

```{r echo=TRUE}
o_data <-
  PKNCAdata(
    o_conc,
    intervals=
      data.frame(
        start=0, end=Inf,
        aucinf.obs=TRUE
      )
  )
o_nca <- suppressMessages(pk.nca(o_data))
```

<p class="forceBreak"></p>

```{r echo=TRUE}
as.data.frame(o_nca)
```

## How to select the correct parameters for calculations (aka, why are there `r sum(grepl(x=names(PKNCA.options("single.dose.aucs")), pattern="^auc"))` types of AUC in PKNCA?)

CDISC has one set of names, but they are not precise (e.g. AUCINT doesn't tell the interpolation/extrapolation method).

PKNCA tries to be everything to everyone (in terms of parameters calculated), and it simultaneously tries to be precise.  That yields many parameters.

See the Selection of Calculation Intervals vignette in the [Parameters Available for Calculation in an Interval section](http://humanpred.github.io/pknca/articles/Selection-of-Calculation-Intervals.html#parameters-available-for-calculation-in-an-interval-1) for all available parameters.

## When are intervals (partly) ignored? {.smaller}

Very few parameters reach outside of the `start` and `end` of an interval for additional information about what is being calculated.  As of the writing of these training materials (PKNCA version 0.9.5), the only parameters that look outside are the `aucint` class of parameters.

AUC~int~ may look after the end of the interval to calculate the concentration at `end`.

Note: Watch out for a dose before the next concentration (e.g. a dose at 24 hours but the prior sample is around 12 and the next is around 25):

```{r fig.height=3}
d_prep <-
  datasets::Theoph %>%
  filter(Subject == 1) %>%
  mutate(
    conc=
      case_when(
        Time == 0~0,
        TRUE~conc
      )
  )
d_plot <-
  superposition(
    conc=d_prep$conc,
    time=d_prep$Time,
    tau=48,
    n.tau=1,
    dose.times=24*(0:1)
  ) %>%
  # Pretend that we missed the predose sample
  filter(
    !between(time, 23, 24.5)
  )
# Pretend that there was a dose at 24
ggplot(d_plot, aes(x=time, y=conc)) +
  geom_point() + geom_line() +
  geom_vline(xintercept=24) +
  scale_x_hours()
```

# Control your results

## Excluding results (Not the best way) {.smaller}

A simple way to exclude a value from results is to convert the results to a data.frame and then drop the rows you don't want:

```{r exclude-setup}
o_conc <- PKNCAconc(data=data.frame(conc=2^-(1:4), time=0:3), conc~time)
o_data <-
  PKNCAdata(
    o_conc,
    intervals=
      data.frame(
        start=0, end=Inf,
        aucinf.obs=TRUE
      )
  )
o_nca <- suppressMessages(pk.nca(o_data))
```

```{r exclude-not-best, echo=TRUE}
as.data.frame(o_nca) %>%
  filter(PPTESTCD != "half.life")
```

But, parameters derived from half-life remain.

## Excluding results (The best way, 1/2) {.columns-2 .smaller}

When you use the `exclude()` function, parameters that are dependent on an excluded parameter will be excluded.

```{r exclude-best, echo=TRUE}
o_nca_excluded <-
  o_nca %>%
  exclude(FUN=exclude_nca_span.ratio(3))
as.data.frame(o_nca_excluded)
```

## Excluding results (The best way, 2/2) {.smaller}

Now, everything dependent on the half-life is excluded in summaries.

```{r echo=TRUE}
summary(o_nca)
summary(o_nca_excluded)
```

# NCA-related calculations

## Superposition {.smaller .columns-2}

Superposition assumes linear kinetics and can convert a single-dose profile to multi-dose.

```{r echo=TRUE}
# Subject 2 is selected for a BLQ time=0 concentration
d_prep <-
  datasets::Theoph %>%
  filter(Subject == 2)
# Superposition to steady-state is the default
d_ss <-
  superposition(
    conc=d_prep$conc,
    time=d_prep$Time,
    tau=24
  )
# Going to steady-state is also an option
# (n.tau=2 means the second dose)
d_second_dose <-
  superposition(
    conc=d_prep$conc,
    time=d_prep$Time,
    tau=24,
    n.tau=2
  )
```

<p class="forceBreak"></p>

```{r echo=TRUE}
# Want the profile for the first two doses
# together?
d_first_two <-
  superposition(
    conc=d_prep$conc,
    time=d_prep$Time,
    tau=48, # 48 hours
    n.tau=1, # One tau interval (0 to 48 hours)
    dose.times=c(0, 24)
  )
```

```{r, fig.height=3, fig.width=3}
ggplot(d_ss, aes(x=time, y=conc)) +
  geom_point() + geom_line() +
  scale_y_continuous(limits=c(0, NA))
```

## Time-to-Steady-state calculations {.smaller}

Time-to-steady-state (tss) can be useful as a method to confirm that a subject is at steady-state.  PKNCA can calculate tss using trough concentrations either with a monoexponential increase toward steady-state (preferred) or a linear trend back from the final point.

```{r echo=TRUE}
dose_times <- seq(0, 96-1, by=6)
d_multidose <-
  superposition(
    conc=d_prep$conc,
    time=d_prep$Time,
    tau=96, # 48 hours
    n.tau=1, # One tau interval (0 to 48 hours)
    dose.times=dose_times
  )
pk.tss.monoexponential(
  conc=d_multidose$conc, time=d_multidose$time, subject=rep(1, nrow(d_multidose)),
  time.dosing=dose_times, subject.dosing=rep(1, length(dose_times)),
  output="single"
)
```

# Reporting

## Graphics are intentionally not part of PKNCA, but there are some tricks that can help...

Generate all individual profiles using the groups that you defined:

```{r echo=TRUE}
o_conc <- PKNCAconc(conc~Time|Subject, data=datasets::Theoph)
d_plot <-
  grouped_df(data=datasets::Theoph, vars=names(getGroups(o_conc))) %>%
  nest() %>%
  mutate(
    figure=
      lapply(
        pmap(.l=list(data=data), .f=ggplot,aes(x=Time, y=conc)),
        FUN="+",
        geom_line()
      )
  )
# d_plot$figure
```

## Best practices for Data -> PKNCA -> knitr

Make summary tables using the `summary()` function on the NCA results, and use `pander::pander()` to make a pretty table with captions.

```{r echo=TRUE}
pander::pander(summary(o_nca))
```

Make an NCA data listing using the `as.data.frame()` function on the NCA results.

```{r echo=TRUE, eval=FALSE}
pander::pander(as.data.frame(o_nca))
```

## Units (especially clearance)

PKNCA supports units with the `pknca_units_table()` function.  See the
[Unit Assignment and Conversion with PKNCA](http://humanpred.github.io/pknca/articles/v07-unit-conversion.html)
vignette for more information.

When units are not specified, The most common place where that becomes an issue
is with clearance which ends up having unusual units like "mg/(hr*ng/mL)" (with
units of mg for dosing, hr for time, and ng/mL for concentration).

## Data imputation

Some data points are required for inputs such as:

* the start of the interval for AUC,
* the end of the interval for some calculations (e.g. AUC~last~ vs AUC~int~), and
* interpolated concentrations at the end of a urine interval for urinary PK calculations.

## IV bolus AUC (need to add C0)

Due to the need for back-extrapolation to C~0~, AUCs for IV bolus dosing need to
use different AUC parameters such as `"aucivlast"` instead of `"auclast"`.

## Combined, multi-subject data (e.g. sparse animal sampling)

Sparse NCA calculations are supported in PKNCA.  See the
[Sparse NCA Calculations](http://humanpred.github.io/pknca/articles/v04-sparse.html)
vignette for more information.

# Limitations

## Secondary parameters (e.g. bioavailability and renal clearance)

PKNCA does not (yet) have the ability to calculate secondary PK parameters that require looking at more than one group/interval at a time.

## Validation of PKNCA

PKNCA has an extensive testing and validation suite built-in.  To run the
testing and validation suite of tests with a full report generated, see the
[PKNCA Validation](http://humanpred.github.io/pknca/articles/v60-PKNCA-validation.html)
vignette.

# Hands-on

## Single- and Multiple-dose, single analyte: Setup the underlying datasets {.smaller}

```{r, echo=TRUE}
d_conc <-
  datasets::Theoph %>%
  rename(time=Time) %>%
  mutate(
    Subject=as.character(Subject)
  )
d_multidose <-
  PKNCAconc(conc~time|Subject, data=d_conc) %>%
  superposition(tau=24, check.blq=FALSE)
d_singledose_single_analyte <-
  d_conc %>%
  mutate(
    Study_Part="Single"
  )
d_multidose_single_analyte <-
  d_conc %>%
  mutate(Day=1) %>%
  bind_rows(
    d_multidose %>% mutate(time=time + 120, Day=6)
  ) %>%
  mutate(
    Study_Part="Multiple"
  )
```

## Single- and Multiple-dose, single analyte: Setup the concentration and dose datasets {.smaller}

```{r, echo=TRUE}
d_single_multi_conc <- bind_rows(d_singledose_single_analyte, d_multidose_single_analyte)
d_single_multi_dose <-
  d_single_multi_conc %>%
  filter(
    (Study_Part %in% "Single" & time == 0) |
      (Study_Part %in% "Multiple" & (time %% 24) == 0)
  )
```

## Single- and Multiple-dose, single analyte: Perform basic analysis {.smaller}

```{r, echo=TRUE}
o_conc <- PKNCAconc(data=d_single_multi_conc, conc~time|Study_Part+Subject)
o_dose <- PKNCAdose(data=d_single_multi_dose, Dose~time|Study_Part+Subject)
o_data <- PKNCAdata(o_conc, o_dose)
o_data$intervals %>% select(-Subject) %>% unique() %>% as.data.frame()
o_nca <- pk.nca(o_data)
```

## Single- and Multiple-dose, single analyte: Use intervals for fewer subjects {.smaller}

```{r echo=TRUE}
d_intervals <-
  data.frame(
    start=0,
    end=24,
    Subject=c("1", "2"),
    Study_Part="Single",
    aucinf.obs=TRUE
  )
o_data <- PKNCAdata(o_conc, o_dose, intervals=d_intervals)
o_nca <- pk.nca(o_data)
summary(o_nca)
```

## Single- and Multiple-dose, single analyte: Use custom intervals per subjects {.smaller}

```{r echo=TRUE}
# Find the time closest to 12 hours
d_intervals_prep <-
  d_single_multi_conc %>%
  filter(Study_Part == "Single") %>%
  mutate(
    time_deviation=abs(time-12)
  ) %>%
  group_by(Subject, Study_Part) %>%
  filter(time %in% time[time_deviation == min(time_deviation)])
d_intervals <-
  d_intervals_prep %>%
  select(Study_Part, Subject, end=time) %>%
  mutate(
    start=0,
    aucinf.obs=TRUE
  )
o_data <- PKNCAdata(o_conc, o_dose, intervals=d_intervals)

o_nca <- pk.nca(o_data)
summary(o_nca, drop.group=c("Subject", "end"))
```

## Single- and Multiple-dose, parent and metabolite

```{r, echo=TRUE}
d_single_multi_conc_multi_analyte <-
  bind_rows(
    d_single_multi_conc %>% mutate(Analyte="Parent"),
    d_single_multi_conc %>%
      mutate(
        Analyte="Metabolite",
        conc=conc/2
      )
  )
o_conc <-
  PKNCAconc(
    data=d_single_multi_conc_multi_analyte,
    conc~time|Study_Part+Subject/Analyte
  )
o_dose <- PKNCAdose(data=d_single_multi_dose, Dose~time|Study_Part+Subject)
o_data <- PKNCAdata(o_conc, o_dose)
o_nca <- pk.nca(o_data)
summary(o_nca)
```

---
title: "Computing NCA Parameters for Theophylline"
author: "Bill Denney"
output:
  rmarkdown::html_vignette:
    toc: yes
    toc_depth: 6
vignette: >
  %\VignetteIndexEntry{Computing NCA Parameters for Theophylline}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, echo=FALSE, include=FALSE}
library(PKNCA)
library(knitr)
```

Examples simplify understanding.  Below is an example of how to use the theophylline dataset to generate NCA parameters.

# Load the data

```{r showtheoph}
## It is always a good idea to look at the data
knitr::kable(head(datasets::Theoph))
```

The columns that we will be interested in for our analysis are conc, Time, and Subject in the concentration data set and Dose, Time, and Subject for the dosing data set.

```{r setupconcdose}
## By default it is groupedData; convert it to a data frame for use
conc_obj <- PKNCAconc(as.data.frame(datasets::Theoph), conc~Time|Subject)

## Dosing data needs to only have one row per dose, so subset for
## that first.
d_dose <- unique(datasets::Theoph[datasets::Theoph$Time == 0,
                                  c("Dose", "Time", "Subject")])
knitr::kable(d_dose,
             caption="Example dosing data extracted from theophylline data set")
dose_obj <- PKNCAdose(d_dose, Dose~Time|Subject)
```

# Merge the Concentration and Dose

After loading the data, they must be combined to prepare for parameter calculation.  Intervals for calculation will automatically be selected based on the `single.dose.aucs setting` in `PKNCA.options`

```{r autointervals}
data_obj_automatic <- PKNCAdata(conc_obj, dose_obj)
knitr::kable(PKNCA.options("single.dose.aucs"))
knitr::kable(data_obj_automatic$intervals)
```

Intervals for calculation can also be specified manually.  Manual specification requires at least columns for `start` time, `end` time, and the parameters requested.  The manual specification can also include any grouping factors from the concentration data set.  Column order of the intervals is not important.  When intervals are manually specified, they are expanded to the full interval set when added to a PKNCAdata object (in other words, a column is created for each parameter.  Also, PKNCA automatically calculates parameters required for the NCA, so while lambda.z is required for calculating AUC~0-$\infty$~, you do not have to specify it in the parameters requested.

```{r manualintervals}
intervals_manual <- data.frame(start=0,
                               end=Inf,
                               cmax=TRUE,
                               tmax=TRUE,
                               aucinf.obs=TRUE,
                               auclast=TRUE)
data_obj_manual <- PKNCAdata(conc_obj, dose_obj,
                             intervals=intervals_manual)
knitr::kable(data_obj_manual$intervals)
```

# Compute the parameters

Parameter calculation will automatically split the data by the grouping factor(s), subset by the interval, calculate all required parameters, record all options used for the calculations, and include data provenance to show that the calculation was performed as described.  For all this, just call the `pk.nca` function with your PKNCAdata object.

```{r calculationauto}
results_obj_automatic <- pk.nca(data_obj_automatic)
knitr::kable(head(as.data.frame(results_obj_automatic)))
```
```{r calculationautoshow, eval=FALSE}
summary(results_obj_automatic)
```
```{r calculationautoshowpretty, echo=FALSE}
## Make a pretty table instead of the data.frame preformatted printout
knitr::kable(summary(results_obj_automatic))
```

```{r calculationmanual}
results_obj_manual <- pk.nca(data_obj_manual)
knitr::kable(head(as.data.frame(results_obj_manual)))
```
```{r calculationmanualshow, eval=FALSE}
summary(results_obj_manual)
```
```{r calculationmanualshowpretty, echo=FALSE}
## Make a pretty table instead of the data.frame preformatted printout
knitr::kable(summary(results_obj_manual))
```

# Multiple Dose Example

Assessing multiple dose pharmacokinetics is conceptually the same as single-dose in PKNCA.

To assess multiple dose PK, the theophylline data will be extended from single to multiple doses using superposition (see the [superposition](v20-superposition.html) vignette for more information).

```{r superposition}
d_conc <- PKNCAconc(as.data.frame(Theoph), conc~Time|Subject)
conc_obj_multi <-
  PKNCAconc(
    superposition(d_conc,
                  tau=168,
                  dose.times=seq(0, 144, by=24),
                  n.tau=1,
                  check.blq=FALSE),
    conc~time|Subject)
conc_obj_multi
dose_obj_multi <- PKNCAdose(expand.grid(Subject=unique(as.data.frame(conc_obj_multi)$Subject),
                                      time=seq(0, 144, by=24)),
                          ~time|Subject)
dose_obj_multi
```

The superposition-simulated scenario is not especially realistic as it includes dense sampling on every day.  With this scenario, the intervals automatically selected have an interval for every subject on every day.

```{r multi_auto_choose_intervals}
data_obj <- PKNCAdata(conc_obj_multi, dose_obj_multi)
data_obj$intervals[,c("Subject", "start", "end")]
```

In a more realistic scenario, dense PK sampling occurs for every subject on the first and last days.  To select those intervals manually, specify the intervals of interest in the `intervals` argument to the PKNCAdata function call.  The intervals are automatically expanded not to calculate anything that was not requested.

```{r multi_manual_choose_intervals}
intervals_manual <- data.frame(start=c(0, 144),
                               end=c(24, 168),
                               cmax=TRUE,
                               auclast=TRUE)
data_obj <- PKNCAdata(conc_obj_multi, dose_obj_multi,
                      intervals=intervals_manual)
data_obj$intervals
```

After the data is ready, the calculations and summary can progress.

```{r multi_calc}
results_obj <- pk.nca(data_obj)
print(results_obj)

summary(results_obj)
```

---
title: "Writing PKNCA Parameter Functions"
author: "Bill Denney"
output:
  rmarkdown::html_vignette:
    toc: yes
    toc_depth: 6
vignette: >
  %\VignetteIndexEntry{Writing PKNCA Parameter Functions}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
library(PKNCA)
```

# Writing PKNCA Parameter Functions

The PKNCA package is designed to be comprehensive in its coverage of the needs of an noncompartmental analysis (NCA) specialist.  While it has many NCA parameters specified, it may not have all parameters defined, and its design is modular to accept new parameter definitions.  From its inception, PKNCA is built in modules to allow addition of new components (or removal of unnecessary ones).  Defining new NCA parameters is straight-forward, and this guide will describe how it is done.  The three parts to writing a new NCA parameter in PKNCA are described below.

# Writing the Parameter Function

## Requirements

The starting point to writing a new NCA parameter is writing the function that calculates the parameter value.  The function can be passed any of the following arguments.  The arguments must be named as described below:

* `conc` is the numeric vector of plasma concentrations for an interval for a single group (usually a single analyte for a single subject in a single study).
* `time` is the numeric vector of the time for plasma concentration measurements.
* `duration.conc` is the duration of a concentration measurement (usually for urine or fecal measurements)
* `dose` is the numeric vector of dose amounts for an interval for a single group.  NOTE: This is a vector and not always a scalar.  If your function expects a scalar, you should usually take the sum of the dose argument.
* `time.dose` is the numeric vector of time for the doses.
* `duration.dose` is the duration of a dose (usually for intravenous infusions)
* `start` and `end` are the scalar numbers for the start and end time of the current interval.  NOTE: `end` may be `Inf` (infinity).
* `options` are the PKNCA options used for the current calculation usually as defined by the `PKNCA.option` function (though these options may be over-ridden by the `options` argument to the `PKNCAdata` function.
* Or, any NCA parameters by name (as given by `names(get.interval.cols())`).

The function should return either a scalar which is the value for the parameter (usually the case) or a data.frame with parameters named for each parameter calculated.  For an example of returning a data.frame, see the `half.life` function.

The return value may have an attribute of `exclude` (set by `attr(return_value, "exclude") <- "reason"`).  If the `exclude` attribute is set to a character string, then that string will be included in the `exclude` column for results.  If any of the input parameters have an exclude attribute set, then those are also added to the `exclude` column.  The exception to the setting of the `exclude` column is if the `exclude` attribute is `"DO NOT EXCLUDE"`, then the `exclude` column is set to `NA_character_`.

## Best Practices

* Use the function `assert_conc_time` if the function takes either `conc` or `time` as an input.
* Make sure that you check for missing values (`NA`) in your inputs.
* Don't recalculate other NCA parameters within your function unless you absolutely must.  Take the NCA parameter as an input.  That way, PKNCA will track the calculation dependencies.
* For consistency with the rest of PKNCA, start the function name with "pk.calc" (like "pk.calc.cmax").

# Tell PKNCA about the Parameter

Just writing a function doesn't connect it to the rest of PKNCA.  You have to tell PKNCA that the function exists and a few more details about it.  To do this, you need to use the `add.interval.col` function.  The function takes up to seven arguments:

* `name` is the name of the parameter (as a character string).
* `FUN` is the function name (as a character string).
* `values` are the possible values for the interval column (currently only `TRUE` and `FALSE` are supported).
* `depends` is a character vector of columns that must exist before this column can be created.  Use this to tell PKNCA about calculation dependencies (parameter X must be calculated to be able to calculate parameter Y).
* `formalsmap` remaps the (formal) function arguments.  `formalsmap` is usually used when the same function may be used for multiple different parameters, for example the function `pk.calc.thalf.eff` is used to calculate the parameters `thalf.eff.obs`, `thalf.eff.pred`, `thalf.eff.last`, `thalf.eff.iv.obs`, `thalf.eff.iv.pred`, and `thalf.eff.iv.last` with different mean residence time inputs.
* `desc` is a text description of the parameter.

# Tell PKNCA How to Summarize the Parameter

For any parameter, PKNCA needs to know how to summarize it for the `summary` function of the `PKNCAresults` class.  To tell PKNCA how to summarize a parameter, use the `PKNCA.set.summary` function.  It takes at least these four arguments:

* `name` must match an already existing parameter name (added by the `add.interval.col` function).
* `description` is a human-readable description of the `point` and `spread` for use in table captions.
* `point` is the function to calculate the point estimate (called as `point(x)`, and it must return a scalar).
* `spread` is the function to calculate the spread (or variability).  The function will be called as `spread(x)` and must return a scalar or a two-long vector.

# Putting It Together

One of the most common examples is the function to calculate C~max~:

    #' Determine maximum observed PK concentration
    #'
    #' @inheritParams assert_conc_time
    #' @param check Run \code{\link{assert_conc_time}}?
    #' @return a number for the maximum concentration or NA if all
    #' concentrations are missing
    #' @export
    pk.calc.cmax <- function(conc, check=TRUE) {
      if (check)
        assert_conc_time(conc=conc)
      if (length(conc) == 0 | all(is.na(conc))) {
        NA
      } else {
        max(conc, na.rm=TRUE)
      }
    }
    ## Add the column to the interval specification
    add.interval.col("cmax",
                     FUN="pk.calc.cmax",
                     values=c(FALSE, TRUE),
                     unit_type="conc",
                     pretty_name="Cmax",
                     desc="Maximum observed concentration",
                     depends=c())
    PKNCA.set.summary("cmax", "geometric mean and geometric coefficient of variation", business.geomean, business.geocv)

---
title: "Unit Assignment and Conversion with PKNCA"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Unit Assignment and Conversion with PKNCA}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
suppressPackageStartupMessages(
  library(PKNCA)
)
```

# Introduction

PKNCA can assign and convert units for reporting. There are two ways to provide
units to PKNCA: via the `units` argument to `PKNCAdata()` or by specifying units
with `PKNCAconc()` and/or `PKNCAdose()`. If you provide the units argument to
`PKNCAdata()`, units given to `PKNCAconc()` or `PKNCAdose()` are ignored.

# Examples of each way to add units

## Steps to add units to an NCA analysis from the data

For more details on parts of this NCA calculation example unrelated to units, see the [theophylline example vignette](v02-example-theophylline.html).

Provide the units for concentration (`concu`), time (`timeu`), and amount
(`amountu`) to the `PKNCAconc()` function and for dose (`doseu`) to the
`PKNCAdose()` function.

```{r create-conc-dose-obj-auto}
d_conc <- as.data.frame(datasets::Theoph)
d_conc$concu_col <- "mg/L"
d_conc$timeu_col <- "hr"
d_dose <- datasets::Theoph[datasets::Theoph$Time == 0, c("Dose", "Time", "Subject")]
d_dose$doseu_col <- "mg/kg"
o_conc <- PKNCAconc(d_conc, conc~Time|Subject, concu = "concu_col", timeu = "timeu_col")
o_dose <- PKNCAdose(d_dose, Dose~Time|Subject, doseu = "doseu_col")
```

Then, create the data object the same way as typical. Results will have units.

```{r create-data-obj-auto}
o_data <- PKNCAdata(o_conc, o_dose)
o_nca <- pk.nca(o_data)
summary(o_nca)
```

It is also possible to specify the units without them coming from columns in the
data.

```{r create-auto-direct}
d_conc <- as.data.frame(datasets::Theoph)
d_dose <- datasets::Theoph[datasets::Theoph$Time == 0, c("Dose", "Time", "Subject")]
o_conc <- PKNCAconc(d_conc, conc~Time|Subject, concu = "mg/L", timeu = "hr")
o_dose <- PKNCAdose(d_dose, Dose~Time|Subject, doseu = "mg/kg")
o_data <- PKNCAdata(o_conc, o_dose)
o_nca <- pk.nca(o_data)
summary(o_nca)
```

And, you can perform automatic unit conversions as long as the unit conversions
are defined without more information (e.g. convert between mass or time units).
For more complex conversions, see the information below.

```{r create-auto-direct-pref}
d_conc <- as.data.frame(datasets::Theoph)
d_dose <- datasets::Theoph[datasets::Theoph$Time == 0, c("Dose", "Time", "Subject")]
o_conc <- PKNCAconc(d_conc, conc~Time|Subject, concu = "mg/L", timeu = "hr", concu_pref = "ug/L", timeu_pref = "day")
o_dose <- PKNCAdose(d_dose, Dose~Time|Subject, doseu = "mg/kg")
o_data <- PKNCAdata(o_conc, o_dose)
o_nca <- pk.nca(o_data)
summary(o_nca)
```

## Steps to manually add units to an NCA analysis

For more details on parts of this NCA calculation example unrelated to units, see the [theophylline example vignette](v02-example-theophylline.html).

```{r create-conc-dose-obj}
o_conc <- PKNCAconc(as.data.frame(datasets::Theoph), conc~Time|Subject)
d_dose <- datasets::Theoph[datasets::Theoph$Time == 0, c("Dose", "Time", "Subject")]
o_dose <- PKNCAdose(d_dose, Dose~Time|Subject)
```

The difference from a calculation without units comes when setting up the `PKNCAdata` object.  You will add the units with the `units` argument.  

Since no urine or other similar collection is performed, the `amountu` argument is omitted for `pknca_units_table()`.

```{r create-data-obj}
d_units <-
  pknca_units_table(
    concu="mg/L", doseu="mg/kg", timeu="hr",
    # use molar units for concentrations and AUCs
    conversions=
      data.frame(
        PPORRESU=c("(mg/kg)/(hr*mg/L)", "(mg/kg)/(mg/L)", "mg/L", "hr*mg/L"),
        PPSTRESU=c("L/hr/kg", "L/kg", "mmol/L", "hr*mmol/L"),
        conversion_factor=c(NA, NA, 1/180.164, 1/180.164)
      )
  )

o_data <- PKNCAdata(o_conc, o_dose, units=d_units)
o_nca <- pk.nca(o_data)
summary(o_nca)
```

# Prepare a Unit Assignment and Conversion Table

A unit assignment and conversion table can be generated as a data.frame to use
with the `pknca_units_table()` function or manually.

The simplest method each of the types of units for inputs and automatically
generates the units for each NCA parameter.

```{r create-units-auto}
d_units_auto <- pknca_units_table(concu="ng/mL", doseu="mg", amountu="mg", timeu="hr")
# Show a selection of the units generated
d_units_auto[d_units_auto$PPTESTCD %in% c("cmax", "tmax", "auclast", "cl.obs", "vd.obs"), ]
```

As you see above, the default units table has a column for the `PPTESTCD`
indicating the parameter.  And, the column `PPORRESU` indicates what the default
units are.

Without unit conversion, the units for some parameters (notably clearances and
volumes) are not so useful.  You can add a conversion table to make any units
into the desired units.  For automatic conversion to work, the units must always
be convertible (by the `units` library).  Notably for automatic conversion, you
cannot go from mass to molar units since there is not a unique conversion from
mass to moles.

Loading the PKNCA package adds a unit of `"fraction"` so that it is usable for
fraction excreted (`fe`).

```{r create-units-semi-manual}
d_units_clean <-
  pknca_units_table(
    concu="ng/mL", doseu="mg", amountu="ng", timeu="hr",
    conversions=
      data.frame(
        PPORRESU=c("mg/(hr*ng/mL)", "mg/(ng/mL)", "hr", "ng/mg"),
        PPSTRESU=c("L/hr", "L", "day", "fraction")
      )
  )
# Show a selection of the units generated
d_units_clean[d_units_clean$PPTESTCD %in% c("cmax", "tmax", "auclast", "cl.obs", "vd.obs", "fe"), ]
```

Now, the units are much cleaner to look at.

To do a conversion that is not possible directly with the `units` library, you can add the conversion factor manually by adding the `conversion_factor` column.  You can mix-and-match manual and automatic modification by setting the `conversion_factor` column to `NA` when you want automatic conversion.  In the example below, we convert concentration units to molar.  Note that AUC units are not set to molar because we did not specify that conversion; all conversions must be specified.

```{r create-units-manual}
d_units_clean_manual <-
  pknca_units_table(
    concu="ng/mL", doseu="mg", amountu="mg", timeu="hr",
    conversions=
      data.frame(
        PPORRESU=c("mg/(hr*ng/mL)", "mg/(ng/mL)", "hr", "ng/mL"),
        PPSTRESU=c("L/hr", "L", "day", "nmol/L"),
        conversion_factor=c(NA, NA, NA, 1000/123)
      )
  )
# Show a selection of the units generated
d_units_clean_manual[d_units_clean_manual$PPTESTCD %in% c("cmax", "tmax", "auclast", "cl.obs", "vd.obs"), ]
```

# How do I add different unit conversions for different analytes?

Sometimes, when multiple analytes are used and, for example, molar outputs are desired while inputs are in mass units.  Different unit conversions may be required for different inputs.

Different unit conversions can be used by adding the grouping column to the units specification.

Start by setting up a concentration dataset with two analytes.  Since the dosing doesn't have an "Analyte" column, it will be matched to all concentration measures for the subject.

```{r different-units-conc-dose-setup}
d_conc_theoph <- as.data.frame(datasets::Theoph)
d_conc_theoph$Analyte <- "Theophylline"
# Approximately 6% of theophylline is metabolized to caffeine
# (https://www.pharmgkb.org/pathway/PA165958541).  Let's pretend that means it
# has 6% of the theophylline concentration at all times.
d_conc_caffeine <- as.data.frame(datasets::Theoph)
d_conc_caffeine$conc <- 0.06*d_conc_caffeine$conc
d_conc_caffeine$Analyte <- "Caffeine"
d_conc <- rbind(d_conc_theoph, d_conc_caffeine)

d_dose <- unique(datasets::Theoph[datasets::Theoph$Time == 0,
                                  c("Dose", "Time", "Subject")])
```

Setup the units with an "Analyte" column to separate the units used.

```{r different-units-units-setup}
d_units_theoph <-
  pknca_units_table(
    concu="mg/L", doseu="mg/kg", timeu="hr",
    # use molar units for concentrations and AUCs
    conversions=
      data.frame(
        PPORRESU=c("(mg/kg)/(hr*mg/L)", "(mg/kg)/(mg/L)", "mg/L", "hr*mg/L"),
        PPSTRESU=c("L/hr/kg", "L/kg", "mmol/L", "hr*mmol/L"),
        conversion_factor=c(NA, NA, 1/180.164, 1/180.164)
      )
  )
d_units_theoph$Analyte <- "Theophylline"
d_units_caffeine <-
  pknca_units_table(
    concu="mg/L", doseu="mg/kg", timeu="hr",
    # use molar units for concentrations and AUCs
    conversions=
      data.frame(
        PPORRESU=c("(mg/kg)/(hr*mg/L)", "(mg/kg)/(mg/L)", "mg/L", "hr*mg/L"),
        PPSTRESU=c("L/hr/kg", "L/kg", "mmol/L", "hr*mmol/L"),
        conversion_factor=c(NA, NA, 1/194.19, 1/194.19)
      )
  )
d_units_caffeine$Analyte <- "Caffeine"
d_units <- rbind(d_units_theoph, d_units_caffeine)
```

Now, calculate adding the different units per analyte to the data object.

```{r different-units-calc}
o_conc <- PKNCAconc(d_conc, conc~Time|Subject/Analyte)
o_dose <- PKNCAdose(d_dose, Dose~Time|Subject)
o_data <- PKNCAdata(o_conc, o_dose, units=d_units)
o_nca <- pk.nca(o_data)
summary(o_nca)
```

---
title: "Post-Processing"
author: "Bill Denney"
output:
  rmarkdown::html_vignette:
    toc: yes
    toc_depth: 6
vignette: >
  %\VignetteIndexEntry{Post-Processing}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, echo=FALSE, include=FALSE}
library(PKNCA)
library(dplyr)
```

# Introduction

Once a calculation is complete, the steps to extract the desired results from the results object are described below.

# Setup

To extract results, first results must be available.  The example below comes from the Introduction and Usage vignette and is reproduced here simply to have results available.  See that vignette for more details about running PKNCA.

```{r setup-example}
library(PKNCA)
suppressPackageStartupMessages(library(dplyr))

d_conc <-
  as.data.frame(datasets::Theoph) %>%
  mutate(Subject=as.numeric(as.character(Subject)))
## Generate the dosing data
d_dose <- d_conc[d_conc$Time == 0,]
d_dose$Time <- 0

conc_obj <-
  PKNCAconc(
    d_conc,
    conc~Time|Subject
  )
dose_obj <-
  PKNCAdose(
    d_dose,
    Dose~Time|Subject
  )
data_obj <- PKNCAdata(conc_obj, dose_obj)
results_obj <- pk.nca(data_obj)
```

# Modifying Results

## Exclusion of Select Results

In many scenarios, individual results may need to be excluded from summaries.  To exclude results, use the `exclude()` function.

### Exclusion Functions

Several exclusion functions are built into PKNCA.  The built-in functions will exclude all results that either apply to the current value or are dependencies of the current value.  For example, $AUC_\infty$ depends on $\lambda_z$, and excluding based on span ratio will exclude all parameters that depend on $\lambda_z$, including $AUC_\infty$.

To see the built-in functions, type `?exclude_nca` at the R command line and review that help page.  To use them, provide the function to the `FUN` argument of `exclude()` as illustrated below.

```{r exclude-function}
results_excl_span <- exclude(results_obj, FUN=exclude_nca_span.ratio())

# Without any exclusions applied, the 'exclude' column is all NA.
as.data.frame(results_obj) %>%
  filter(Subject == 1)
# With exclusions applied, the 'exclude' column has the reason for exclusion.
as.data.frame(results_excl_span) %>%
  filter(Subject == 1)
```

You may also write your own exclusion function.  The exclusion functions built-into PKNCA are a bit more complex than required because they handle options and manage general functionality that may not apply to a user-specific need.  To write your own exclusion function, it should follow the description of how to write your own exclusion function as described in the details section of `?exclude`.

### Excluding Specific Results

Excluding specific results has the benefit that full control is provided.  But, excluding specific points allows for errors to also enter the analysis because parameters that depend on the excluded parameter will not be excluded.

```{r exclude-specific}
mask_exclude_cmax <-
  results_obj %>%
  as.data.frame() %>%
  dplyr::mutate(
    mask_exclude=Subject == 1 & PPTESTCD == "cmax"
  ) %>%
  "[["("mask_exclude")
results_excl_specific <-
  exclude(
    results_obj,
    mask=mask_exclude_cmax,
    reason="Cmax was actually above the ULOQ"
  )

# Without any exclusions applied, the 'exclude' column is all NA.
results_obj %>%
  as.data.frame() %>%
  filter(Subject == 1)
# With exclusions applied, the 'exclude' column has the reason for exclusion.
results_excl_specific %>%
  as.data.frame() %>%
  filter(Subject == 1)
```

### Multiple Exclusions

More than one exclusion can be applied sequentially to results as in the example below.

```{r exclude-multi}
mask_exclude_lz <-
  results_obj %>%
  as.data.frame() %>%
  dplyr::mutate(
    mask_exclude=Subject == 1 & PPTESTCD == "lambda.z"
  ) %>%
  "[["("mask_exclude")

# Starting from the exclusion example above where short span ratios were
# excluded, exclude Cmax for Subject 1, too.
results_excl_multi <-
  exclude(
    results_excl_span,
    mask=mask_exclude_cmax,
    reason="Cmax was actually above the ULOQ"
  )
results_excl_multi <-
  exclude(
    results_excl_multi,
    mask=mask_exclude_lz,
    reason="Issue with lambda.z fit"
  )

# With exclusions applied, the 'exclude' column has the reason for exclusion.
# More than one reason may appear if more than one exclusion is applied.
results_excl_multi %>%
  as.data.frame() %>%
  filter(Subject == 1)
```

# Extracting Results

## Summary Results

Summary results are obtained using the aptly named `summary()` function.  It will output a `summary_PKNCAresults` object that is simply a data.frame with an attribute of `caption`.  The summary is generated by evaluating summary statistics on each requested parameter.  Which summary statistics are calculated for each parameter are set with `PKNCA.set.summary()`, and they are described in the caption.  When a parameter is not requested for a given interval, it is illustrated with a period (`.`), by default (customizable with the `not.requested.string` argument to `summary()`).  When a parameter is required to calculate another parameter, but it is not specifically requested, it will not be shown in the summary.

The summary will have one column for each grouping variable other than the subject grouping variable; one column each for the start and end time; and one column per parameter calculated.

```{r summary}
summary(results_obj)
```

When values are excluded as described above, the excluded values are not included in the summary (note that half.life and aucinf.obs differ).

```{r summary-exclusion}
summary(results_excl_span)
```

## Listing of Results

A listing of all calculated values is available using `as.data.frame()`.

```{r listing}
as.data.frame(results_obj) %>%
  head(20)
```

Excluded values remain in the listing.

```{r listing-exclusion}
as.data.frame(results_excl_span) %>%
  head(20)
```

---
title: "Sparse NCA Calculations"
author: "Bill Denney"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Sparse NCA Calculations}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

# Sparse NCA Calculations

Sparse noncompartmental analysis (NCA) is performed when multiple individuals contribute to a single concentration-time profile due to the fact that there are only one or a subset of the full profile samples taken per animal.  A typical example is when three mice have PK drawn per time point, but no animals have more than one sample drawn.  Another typical example is when animals may have two or three samples during an interval, but no animal has the full profile.

## Sparse NCA Setup

Sparse NCA is setup similarly to how normal, dense PK sampling is setup with PKNCA.  The only difference are that you give the `sparse` option to `PKNCAconc()`, and in your interval calculations, you will request the sparse variants of the parameters.  As of the writing of this vignette, the only sparse parameter for calculation is `sparse_auclast`.  Any of the non-sparse parameters will be calculated based on the mean profile of the animals in a group.

The example below uses data extracted from Holder D. J., Hsuan F., Dixit R. and Soper K. (1999). A method for estimating and testing area under the curve in serial sacrifice, batch, and complete data designs. Journal of Biopharmaceutical Statistics, 9(3):451-464.

```{r}
# Setup the data
d_sparse <-
    data.frame(
      id = c(1L, 2L, 3L, 1L, 2L, 3L, 1L, 2L, 3L, 4L, 5L, 6L, 4L, 5L, 6L, 7L, 8L, 9L, 7L, 8L, 9L),
      conc = c(0, 0, 0,  1.75, 2.2, 1.58, 4.63, 2.99, 1.52, 3.03, 1.98, 2.22, 3.34, 1.3, 1.22, 3.54, 2.84, 2.55, 0.3, 0.0421, 0.231),
      time = c(0, 0, 0, 1, 1, 1, 6, 6, 6, 2, 2, 2, 10, 10, 10, 4, 4, 4, 24, 24, 24),
      dose = c(100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100)
    )
```

Look at your data.  (This is not technically a required step, but it's good practice.)

```{r}
library(ggplot2)
ggplot(d_sparse, aes(x=time, y=conc, group=id)) +
  geom_point() +
  geom_line() +
  scale_x_continuous(breaks=seq(0, 24, by=6))
```

## Data Setup Note

Sparse NCA requires that subject numbers (or animal numbers) are given, even if each subject only contributes a single sample.  The reason for this requirement is that which subject contributes to which time point changes the standard error calculation.  If all individuals contribute a single sample, a simple way to handle this is by setting a column with sequential numbers and giving that as the subject identifier:

```{r eval=FALSE}
d_sparse$id <- 1:nrow(d_sparse)
```

# Calculate!

Setup PKNCA for calculations and then calculate!

```{r}
library(PKNCA)
o_conc_sparse <- PKNCAconc(d_sparse, conc~time|id, sparse=TRUE)
d_intervals <-
  data.frame(
    start=0,
    end=24,
    aucinf.obs=TRUE,
    cmax=TRUE,
    sparse_auclast=TRUE
  )
o_data_sparse <- PKNCAdata(o_conc_sparse, intervals=d_intervals)
o_nca <- pk.nca(o_data_sparse)
```

# Results

As with any other PKNCA result, the data are available through the `summary()` function:

```{r}
summary(o_nca)
```

or individual results are available through the `as.data.frame()` function:

```{r}
as.data.frame(o_nca)
```

---
title: "Options for Controlling PKNCA"
author: "Bill Denney"
output:
  rmarkdown::html_vignette:
    toc: yes
    toc_depth: 6
vignette: >
  %\VignetteIndexEntry{Options for Controlling PKNCA}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
library(PKNCA)
```

# Summary

PKNCA has many options that control its function.  These options have effects throughout the package.  The options are controlled using either the `PKNCA.options` function or by passing the `options` argument to any of the functions with that as an argument.  All options supported by the current version of PKNCA (`r sessionInfo(package="PKNCA")$otherPkgs$PKNCA$Version`) are listed below with their descriptions.

# Options

```{r options, echo=FALSE, results='asis'}
PKNCA.options(default=TRUE)
opts <- PKNCA.options()
for (n in names(opts)) {
  cat("## ", n, "\n", sep="")
  cat("\n")
  cat(PKNCA:::PKNCA.options.describe(n), "\n", sep="")
  cat("\n")
  if (is.character(opts[[n]]) | is.numeric(opts[[n]]) | is.factor(opts[[n]]) | is.logical(opts[[n]])) {
    cat("The default value is: ", opts[[n]], "\n", sep="")
  } else if (is.data.frame(opts[[n]])) {
    print(knitr::kable(opts[[n]]))
  } else {
    # Probably a list
    print(opts[[n]])
  }
  cat("\n")
}
```

---
title: "Noncompartmental evaluation of time to steady-state"
author: "Bill Denney"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Noncompartmental evaluation of time to steady-state}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
library(PKNCA)
library(knitr)
library(ggplot2)
scale_colour_discrete <- scale_colour_hue
scale_fill_discrete <- scale_fill_hue
scale_colour_ordinal <- scale_colour_hue
scale_fill_ordinal <- scale_fill_hue
```

Time to steady-state (TSS) can be estimated with PKNCA using either a monoexponential
increase toward an asymptote or by a linear regression of the last points.
According to Maganti (2008), the monoexponential method is preferred.

TSS can be estimated using either method using the `pk.tss()` function in PKNCA.

# Example

## Data setup

Illustrating time to steady-state, the example from the
[superposition vignette](v20-superposition.html) will be used.

```{r superposition, error=TRUE}
library(PKNCA)
theoph_corrected <- as.data.frame(datasets::Theoph)
theoph_corrected$conc[theoph_corrected$Time == 0] <- 0
conc_obj <- PKNCAconc(theoph_corrected, conc~Time|Subject)
steady_state <- superposition(conc_obj, dose.times = seq(0, 168 - 12, by=12), tau=168, n.tau=1)
# Add some noise to the data so that it seems more reasonable
steady_state_noise <- steady_state
steady_state_noise$conc <-
  withr::with_seed(
    seed = 5,
    steady_state_noise$conc*exp(rnorm(nrow(steady_state_noise), mean = 0, sd = 0.1))
  )
```

Examine the data graphically.

```{r superposition-fig}
library(ggplot2)
ggplot(steady_state_noise, aes(x=time, y=conc, groups=Subject)) + geom_line()
```

## Estimate time to Steady State

### Monoexponential

The below code estimates four different types of monoexponential time to
steady-state:

1. tss.monoexponential.population: The population estimate of TSS using a nonlinear mixed effects model (one value for all subjects)
2. tss.monoexponential.popind: The individual estimate from a nonlinear mixed effects model (one value per subject)
3. tss.monoexponential.individual: The individual estimate using a gnls model to estimate each subject separately (one value per subject)
4. tss.monoexponential.single: The mean estimate of TSS using a nonlinear model

```{r tss-mono}
tss_mono <-
  pk.tss.monoexponential(
    conc = steady_state_noise$conc,
    time = steady_state_noise$time,
    subject = steady_state_noise$Subject,
    time.dosing = seq(0, 168 - 12, by=12)
  )
tss_mono
```

The fraction of steady-state required for time to steady-state can be changed
with the `tss.fraction` argument (see `?pk.tss.monoexponential`).

### Stepwise Linear

The stepwise linear method estimates if the slope of the predose concentrations
is statistically significant starting from the last measurement and moving
backward in time.  It has bias in that more individuals will move the time to
steady-state to a late time point.

```{r tss-step}
tss_step <-
  pk.tss.stepwise.linear(
    conc = steady_state_noise$conc,
    time = steady_state_noise$time,
    subject = steady_state_noise$Subject,
    time.dosing = seq(0, 168 - 12, by=12)
  )
tss_step
```

# References

1. Maganti, L., Panebianco, D.L. & Maes, A.L. Evaluation of Methods for Estimating Time to Steady State with Examples from Phase 1 Studies. AAPS J 10, 141–147 (2008). https://doi.org/10.1208/s12248-008-9014-y

---
title: "Methods Used for Dose-Aware Concentration Interpolation/Extrapolation"
author: "Bill Denney"
output:
  rmarkdown::html_vignette:
    toc: yes
    toc_depth: 6
vignette: >
  %\VignetteIndexEntry{Methods Used for Dose-Aware Concentration Interpolation/Extrapolation}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
library(PKNCA)
library(dplyr)
```

# Introduction

Interpolation and extrapolation with awareness of doses occurring before, at the same time of, or after the requested interpolation time point must account for many interactions.  To ensure clarity in the interpolation/extrapolation methods and the decisions made by the algorithm, each potential choice is listed below with its accompanying calculation method.  The code used to generate the table is the same as the code within the function.

```{r generatecombs, echo=FALSE}
method.choices <- names(PKNCA:::interp.extrap.conc.dose.select)
method.choices <-
  factor(method.choices, levels=method.choices, ordered=TRUE)
         
all_combs <-
  expand.grid(event_before=setdiff(unlist(PKNCA:::event_choices_interp.extrap.conc.dose), "output_only"),
              event=setdiff(unlist(PKNCA:::event_choices_interp.extrap.conc.dose), "none"),
              event_after=setdiff(unlist(PKNCA:::event_choices_interp.extrap.conc.dose), "output_only"),
              Method="",
              stringsAsFactors=FALSE)

for (n in method.choices) {
  mask <-
    do.call(PKNCA:::interp.extrap.conc.dose.select[[n]]$select,
            list(x=all_combs),
            envir=environment(pk.nca))
  all_combs$Method[mask] <- n
}

all_combs <-
  all_combs[do.call(order,
                    args=append(as.list(all_combs), list(na.last=FALSE))),]
```

# Methods

The method list below is described and sorted in order of how many scenarios the method is applied to in the list.

For each of the summary tables below, the column headers are as follows:

* **Event Before**: The type of event before the time of the requested output may be one of the following
* **Event At**: The equivalent to "Event Before" but for the event occurring at the requested output time.
* **Event After**: The equivalent to "Event Before" but for the next event occurring after the requested output time.

```{r methodsummary, results="asis", echo=FALSE}
methodorder <- names(sort(summary(factor(all_combs$Method)), decreasing=TRUE))

for (n in methodorder) {
  cat("## ", n, "\n\n", sep="")
  cat(PKNCA:::interp.extrap.conc.dose.select[[n]]$description, "\n\n", sep="")
  print(knitr::kable(
    all_combs[all_combs$Method %in% n,
              c("event_before","event", "event_after")],
    row.names=FALSE,
    col.names=c("Event Before", "Event At", "Event After")))
  cat("\n")
}
```

# Appendix: Complete Methods Table

```{r methodlisting, echo=FALSE}
knitr::kable(
  all_combs[,c("event_before", "event", "event_after", "Method")],
    row.names=FALSE,
    col.names=c("Event Before", "Event At", "Event After", "Method Used"))
```

---
title: "PKNCA Validation"
author: "Bill Denney"
output:
  rmarkdown::html_vignette:
    toc: yes
    toc_depth: 6
vignette: >
  %\VignetteIndexEntry{PKNCA Validation}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo=TRUE,
                      fig.width=6,
                      fig.height=4)
library(PKNCA)
library(knitr)
library(testthat)
library(dplyr)

run_tests <- dir.exists(system.file(package="PKNCA", "tests/"))
testing_results <- "Tests were not run because tests are not installed."
```

# Introduction

To run the tests, the package must be installed with its tests:

* To install from CRAN:
    * `install.packages(pkgs="PKNCA", INSTALL_opts="--install-tests", type="source")`
* To install from GitHub:
    * `library(devtools)`
    * `install_github("humanpred/pknca", INSTALL_opts="--install-tests")`

Testing and validation that results match in a local environment compared to the original environment is an important part of confirmation that a package works as expected.

Re-running this vignette in your local environment will confirm that local results match those in the original package development.  Test success is confirmed by the existence of no failed tests; warnings are expected during testing (and not shown in this vignette for that reason); and some tests may be skipped, but those are expected as well.

# Summary of Testing

```{r validation, include=FALSE, eval=run_tests}
testresult <- test_package(package="PKNCA")
test_log <- as.data.frame(testresult)
failed_tests <- sum(test_log$failed)
testing_results <-
  if (failed_tests == 0) {
    "All tests passed."
  } else {
    "Some tests failed (see below)."
  }
```

The following sentence is dynamically generated to summarize the testing results:  `r testing_results`

```{r testing_log, results="asis", echo=FALSE, eval=run_tests}
cat("# Testing Log\n\n")

all_contexts <- unique(test_log$context)
test_log_printing <-
  test_log %>%
  select(context, file, test, nb, failed) %>%
  rename(`Testing Filename`=file,
         `Test Description`=test,
         `Number of Tests`=nb,
         `Number of Failed Tests`=failed)

for (current_context in all_contexts) {
  current_log <-
    test_log_printing[
      test_log_printing$context %in% current_context,
      setdiff(names(test_log_printing), "context"), # context is in the header
      drop=FALSE]
  cat("## ", current_context, "\n\n")
  print(kable(x=current_log,
              row.names=FALSE))
  cat("\n\n")
}
```

# Session Information

```{r session_info}
Sys.Date()
sessionInfo()
```

---
title: "Introduction to PKNCA and Usage Instructions"
author: "Bill Denney"
output:
  rmarkdown::html_vignette:
    toc: yes
    toc_depth: 6
vignette: >
  %\VignetteIndexEntry{Introduction to PKNCA and Usage Instructions}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, echo=FALSE, include=FALSE}
library(PKNCA)
```

PKNCA provides functions to complete noncompartmental analysis (NCA)
for pharmacokinetic (PK) data.  Its intent is to provide a complete 
R-based solution-enabling data provenance for NCA. This will include 
the tracking of data cleaning, enabling of calculations, exporting
of results, and general reporting.  The library is designed to give
a reasonable answer without user intervention (load, calculate, and summarize),
but it allows the user to override the automatic selections at any point.

The library design is modular to allow expansion based on needs
unforseen by the authors including new NCA parameters, novel data
cleaning methods, and modular summarization decisions.  Expanding the
library will be discussed in a separate vignette.

# Quick Start

The simplest analysis requires concentration and dosing data at a minimum. Given this, it then takes five function calls to provide summarized results.  (Please note that this and the other examples in this document are intended to show the typical workflow, but they are not intended to run directly.  For an example to run directly, please see [the theophylline example](v02-example-theophylline.html).)

```{r setup_data}
library(PKNCA)
library(dplyr, quietly=TRUE)

## Load the PK concentration data
d_conc <- datasets::Theoph
## Generate the dosing data
d_dose <- d_conc[d_conc$Time == 0,]

## Create a concentration object specifying the concentration, time, and
## subject columns.  (Note that any number of grouping levels is
## supported; you are not restricted to just grouping by subject.)
conc_obj <-
  PKNCAconc(
    d_conc,
    conc~Time|Subject
  )
## Create a dosing object specifying the dose, time, and subject
## columns.  (Note that the grouping factors should be the same as or a
## subset of the grouping factors for concentration, and the grouping
## columns must have the same names between concentration and dose
## objects.)
dose_obj <-
  PKNCAdose(
    d_dose,
    Dose~Time|Subject
  )
## Combine the concentration and dosing information both to
## automatically define the intervals for NCA calculation and provide
## doses for calculations requiring dose.
data_obj <- PKNCAdata(conc_obj, dose_obj)

## Calculate the NCA parameters
results_obj <- pk.nca(data_obj)

## Summarize the results
summary(results_obj)
```

# Data Handling

After loading data, it must be in the right form.  The minimum requirements are
that concentration, dose, and time must all be numeric (and not factors).
Grouping variables have no specific requirements; they can be any mode.

Values below the limit of quantification are coded as zeros (`0`), and missing
values are coded as `NA`.

# Options: Make PKNCA Work Your Way

## Calculation Options: the PKNCA.options Function

Different organizations have different requirements for computation and
summarization of NCA.  Options for how to perform calculations and summaries are
handled by the `PKNCA.options` command.

Default options have been set to commonly-used standard parameters. The current
value for options can be found by running the command with no arguments:

```{r}
PKNCA.options()
```

And, to reset the current values to the library defaults, run the function with
the default argument set to `TRUE`.

```{r eval=FALSE}
PKNCA.options(default=TRUE)
```

Each of the options is documented where it is used; for example, the first.tmax
option is documented in the `pk.calc.tmax` function.

## Summarization Options: the PKNCA.set.summary Function

On top of methods of calculation, summarization method preferences differ.
Typical summarization preferences include selection of the measurement of
central tendency and dispersion, handling of missing values, handling of values
below the limit of quantification, and more.  Beyond the method for
summarization, presentation is managed through user preferences.  Presentation
is typically controlled by rounding to either a defined number of decimal places
or significant figures.

An example is that C~max~ may be summarized by the geometric mean with the
geometric CV using three significant figures, and having a summary result
requires that at least half of the available values are present (not missing).
The code below will set this example.

```{r eval=FALSE}
PKNCA.set.summary(
  name = "cmax",
  description = "geometric mean and geometric coefficient of variation",
  point = business.geomean,
  spread = business.geocv,
  rounding = list(signif=3)
)
```

Another example is that T~max~ is usually summarized by the median and
range, and as measurements are often taken with minute resolution and
recorded in hours, reporting is usually to the second decimal place.

```{r eval=FALSE}
PKNCA.set.summary(
  name = "tmax",
  description = "median and range",
  point = business.median,
  spread = business.range,
  rounding = list(round=2)
)
```

If the functions or default rounding options provided in the library
do not meet the summarization needs, a user-supplied function can be
used for rounding.

```{r custom_summary_fun, eval=FALSE}
median_na <- function(x) {
  median(x, na.rm = TRUE)
}
quantprob_na <- function(x) {
  quantile(x, probs = c(0.05, 0.95), na.rm=TRUE)
}
PKNCA.set.summary(
  name="auclast",
  description = "median and 5th to 95th percentile",
  point=median_na,
  spread=quantprob_na,
  rounding=list(signif=3)
)
```

In some cases multiple parameters may need the same summary functions
(as often occurs with simulated data).  Many parameters can be set
simultaneously by specifying a vector of `name`s.

```{r multi_summary_settings, eval=FALSE}
median_na <- function(x) {
  median(x, na.rm=TRUE)
}
quantprob_na <- function(x) {
  quantile(x, probs=c(0.05, 0.95), na.rm=TRUE)
}
PKNCA.set.summary(
  name=c("auclast", "cmax", "tmax", "half.life", "aucinf.pred"),
  description = "median and 5th to 95th percentile",
  point=median_na,
  spread=quantprob_na,
  rounding=list(signif=3)
)
```

# Grouping NCA Data

As described in the [quick start](#quick-start), concentration and dose
data are generally grouped to identify how to separate the data.
Typical groups for concentration data include study, treatment,
subject, and analyte.  Typical groups for dose data include study,
treatment, and subject.  By default, summaries are produced based on
the concentration groups dropping the subject (so that averages are
taken across subjects within the other parameters).

The quick start example can be extended to include multiple analytes
as follows.  The only difference is the `/analyte`
formula element in the concentration data.  The reason for the slash
instead of the plus is that the last element before a slash is assumed
to be the subject, and as noted before, the subject is (by default)
excluded from the summary grouping (so that summaries are grouped by
study, treatment, etc., but not by subject).

```{r grouping, eval=FALSE}
## Generate a faux multi-study, multi-analyte dataset.
d_conc_parent <- d_conc
d_conc_parent$Subject <- as.numeric(as.character(d_conc_parent$Subject))
d_conc_parent$Study <- d_conc_parent$Subject <= 6
d_conc_parent$Analyte <- "Parent"
d_conc_metabolite <- d_conc_parent
d_conc_metabolite$conc <- d_conc_metabolite$conc/2
d_conc_metabolite$Analyte <- "Metabolite"
d_conc_both <- rbind(d_conc_parent, d_conc_metabolite)
d_conc_both <- d_conc_both[with(d_conc_both, order(Study, Subject, Time, Analyte)),]
d_dose_both <- d_conc_both[d_conc_both$Time == 0 & d_conc_both$Analyte %in% "Parent",
                           c("Study", "Subject", "Dose", "Time")]

## Create a concentration object specifying the concentration, time,
## study, and subject columns.  (Note that any number of grouping
## levels is supporting; you are not restricted to this list.)
conc_obj <- PKNCAconc(d_conc_both,
                      conc~Time|Study+Subject/Analyte)
## Create a dosing object specifying the dose, time, study, and
## subject columns.  (Note that the grouping factors should be a
## subset of the grouping factors for concentration, and the columns
## must have the same names between concentration and dose objects.)
dose_obj <- PKNCAdose(d_dose_both,
                     Dose~Time|Study+Subject)

# Perform and summarize the PK data as previously described
data_obj <- PKNCAdata(conc_obj, dose_obj)
results_obj <- pk.nca(data_obj)
summary(results_obj)
```

# Selecting Calculation Intervals

All NCA calculations require the interval over which they are
calculated.  When the concentration and dosing information are
combined to the PKNCAdata object, intervals are automatically
determined.  The exception to this automatic determination is if the
user provides intervals.

When selected either automatically or manually, intervals define at
minimum a start time, an end time, and the parameters to be
calculated.  The parameter list is available from the `get.interval.cols`
function.  The parameters requested are specified by setting the entry in
a data.frame as requested.

```{r}
intervals <-
  data.frame(
    start=0, end=c(24, Inf),
    cmax=c(FALSE, TRUE),
    tmax=c(FALSE, TRUE),
    auclast=TRUE,
    aucinf.obs=c(FALSE, TRUE)
  )
```

Intervals like the one above are sufficient for designs with a
single type of treatment– such as single doses.  For more
complex treatments in a single analysis, like the combination of
single and multiple doses, include a treatment column matching
the treatment column name from the concentration data set.  See
the [Manual Interval Specification](#manual-interval-specification) section
below for more details.

## Selection of Data Used for Calculation

When choosing which data is used for a calculation, PKNCA will never look beyond
the data specified in the group and interval.  Groups are defined by the call to
the `PKNCAconc` function, and they will typically define the measurement of a
single analyte from a single individual receiving a single treatment.  Intervals
are subsets within a group by start and end time.  PKNCA never examines data
outside of the group and interval for standard NCA calculations.  As an example,
with data from 0 to 48 hours and an interval set to `start` at 0 and `end` at 24
with the calculation of `aucinf.obs`, any data after 24 hours will not be used
for the half-life or AUC~inf~ calculations.

A few functions look at data outside of a single interval, but these functions
do not look at data outside of a single group, and these functions are typically
used during preparation for NCA calculations not for the calculations
themselves.  Functions that look at a group as a whole include
`choose.auc.intervals`, `find.tau`, and `pk.tss`.

## Automatic Interval Determination

If intervals are not specified when combining the concentration and dosing data,
they will automatically be found from the concentration and dosing data.

Single dose data has a simple interval selection:  the option `single.dose.aucs`
is used from the `PKNCA.options`.

```{r asis=TRUE, echo=FALSE}
knitr::kable(PKNCA.options()$single.dose.aucs)
```

For multiple-dose studies, PKNCA selects one group at a time and compares the
concentration and dosing times.  When there is a concentration measurement
between doses, an interval row is added.  The dosing interval ($\tau$) is
determined by looking for pattern repeats within the dosing data using the
`find.tau` function.

```{r}
## find.tau can work when all doses have the same interval...
dose_times <- seq(0, 168, by=24)
print(dose_times)
PKNCA::find.tau(dose_times)

## or when the doses have mixed intervals (10 and 24 hours).
dose_times <- sort(c(seq(0, 168, by=24),
                     seq(10, 178, by=24)))
print(dose_times)
PKNCA::find.tau(dose_times)
```

After finding $\tau$, PKNCA will also look after the last dose (or the beginning
of the last dosing interval), and two additional intervals may be added:

* one interval for the dosing interval after the beginning of the last dosing
  interval (if there are concentrations measured in the interval)
* one interval for the half-life after the last dosing interval (if there are
  concentration more than $\tau$ after the beginning of the last interval).

One consequence of automatic interval selection is that many rows are generated
for intervals; one row is generated per interval per subject.  The benefit of
the method producing a large number of rows is that it is fully flexible to the
actual study results.  If a subject has a different schedule than the others for
the same treatment (e.g. measurements that were nominally scheduled for day 14
occurred on day 13), those differences will be found.

## Manual Interval Specification

Intervals can also be specified manually.  Two use cases are common for manual
specification:  fully manual (never requesting the automatic intervals) and
updating the automatic intervals.

Fully manual intervals can be specified by providing it to the `PKNCAdata` call.

```{r eval=FALSE}
intervals_manual <-
  data.frame(
    start=0, end=c(24, Inf),
    cmax=c(FALSE, TRUE),
    tmax=c(FALSE, TRUE),
    auclast=TRUE,
    aucinf.obs=c(FALSE, TRUE)
  )
data_obj <-
  PKNCAdata(
    conc_obj, dose_obj, 
    intervals=intervals_manual
  )
```

To update the automatically-selected intervals, extract the intervals, modify them, and put them back.

```{r eval=FALSE}
data_obj <- PKNCAdata(conc_obj, dose_obj)
intervals_manual <- data_obj$intervals
intervals_manual$aucinf.obs[1] <- TRUE
data_obj$intervals <- intervals_manual
```

## Keeping a column from intervals

When computing NCA using actual times, grouping by start and end time in
summaries (see layer) is less helpful because everyone could have different
start and end times.  So, you may keep the interval columns using the option
`"keep_interval_cols"` as follows (where "dosetype" must be a column name in the
intervals):

```{r eval=FALSE}
data_obj <- PKNCAdata(conc_obj, dose_obj, options = list(keep_interval_cols = "dosetype"))
```

# Summarizing results

When NCA has been calculated, you can summarize the results with the `summary()`
function.

```{r eval=FALSE}
summary(o_nca)
```

By default, it will count the number of unique subjects (`N`) in the summary, and
when the number of subjects differs from the number of measurements included in
a summary (`n`), it will summarize `n` for the given parameters.  Note that
counting of "n" includes all non-missing values that were not excluded from
summarization; this will included all zeros that are e.g. excluded from
geometric statistics.

Edge cases like two unique subjects where one has an excluded value and one has
duplicated values (`N = 2` and `n = 2` even though both measurements come from a
single subject) are to be handled by the user.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pk.calc.all.R
\name{pk.nca}
\alias{pk.nca}
\title{Compute NCA parameters for each interval for each subject.}
\usage{
pk.nca(data, verbose = FALSE)
}
\arguments{
\item{data}{A PKNCAdata object}

\item{verbose}{Indicate, by \code{message()}, the current state of calculation.}
}
\value{
A \code{PKNCAresults} object.
}
\description{
The \code{pk.nca} function computes the NCA parameters from a \code{PKNCAdata} object.
All options for the calculation and input data are set in prior functions
(\code{PKNCAconc}, \code{PKNCAdose}, and \code{PKNCAdata}).  Options for calculations are
set either in \code{PKNCAdata} or with the current default options in
\code{PKNCA.options}.
}
\details{
When performing calculations, all time results are relative to the start of
the interval.  For example, if an interval starts at 168 hours, ends at 192
hours, and and the maximum concentration is at 169 hours, \code{tmax=169-168=1}.
}
\seealso{
\code{\link[=PKNCAdata]{PKNCAdata()}}, \code{\link[=PKNCA.options]{PKNCA.options()}}, \code{\link[=summary.PKNCAresults]{summary.PKNCAresults()}},
\code{\link[=as.data.frame.PKNCAresults]{as.data.frame.PKNCAresults()}}, \code{\link[=exclude]{exclude()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exclude.R
\name{normalize_exclude}
\alias{normalize_exclude}
\title{Normalize the exclude column by setting blanks to NA}
\usage{
normalize_exclude(object)
}
\arguments{
\item{object}{The object to extract the exclude column from}
}
\value{
The exclude vector where \code{NA} indicates not to exclude and anything
else indicates to exclude.
}
\description{
Normalize the exclude column by setting blanks to NA
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sparse.R
\name{sparse_mean}
\alias{sparse_mean}
\title{Calculate the mean concentration at all time points for use in sparse NCA
calculations}
\usage{
sparse_mean(
  sparse_pk,
  sparse_mean_method = c("arithmetic mean, <=50\% BLQ", "arithmetic mean")
)
}
\arguments{
\item{sparse_pk}{A sparse_pk object from \code{\link[=as_sparse_pk]{as_sparse_pk()}}}

\item{sparse_mean_method}{The method used to calculate the sparse mean (see
details)}
}
\value{
A vector the same length as \code{sparse_pk} with the mean concentration
at each of those times.
}
\description{
Choices for the method of calculation (the argument \code{sparse_mean_method})
are:
}
\details{
\describe{
\item{"arithmetic mean"}{Arithmetic mean (ignoring number of BLQ samples)}
\item{"arithmetic mean, <=50\% BLQ"}{If >= 50\% of the measurements are BLQ, zero.  Otherwise, the arithmetic mean of all samples (including the BLQ as zero).}
}
}
\seealso{
Other Sparse Methods: 
\code{\link{as_sparse_pk}()},
\code{\link{pk.calc.sparse_auc}()},
\code{\link{sparse_auc_weight_linear}()}
}
\concept{Sparse Methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-PKNCAconc.R, R/class-PKNCAdata.R,
%   R/class-PKNCAdose.R, R/class-PKNCAresults.R
\name{group_vars.PKNCAconc}
\alias{group_vars.PKNCAconc}
\alias{group_vars.PKNCAdata}
\alias{group_vars.PKNCAdose}
\alias{group_vars.PKNCAresults}
\title{Get grouping variables for a PKNCA object}
\usage{
\method{group_vars}{PKNCAconc}(x)

\method{group_vars}{PKNCAdata}(x)

\method{group_vars}{PKNCAdose}(x)

\method{group_vars}{PKNCAresults}(x)
}
\arguments{
\item{x}{The PKNCA object}
}
\value{
A character vector (possibly empty) of the grouping variables
}
\description{
Get grouping variables for a PKNCA object
}
\section{Functions}{
\itemize{
\item \code{group_vars(PKNCAdata)}: Get group_vars for a PKNCAdata object
from the PKNCAconc object within

\item \code{group_vars(PKNCAdose)}: Get group_vars for a PKNCAdose object

\item \code{group_vars(PKNCAresults)}: Get group_vars for a PKNCAresults object
from the PKNCAconc object within

}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PKNCA.options.R
\name{PKNCA.choose.option}
\alias{PKNCA.choose.option}
\title{Choose either the value from an option list or the current set value for an
option.}
\usage{
PKNCA.choose.option(name, value = NULL, options = list())
}
\arguments{
\item{name}{The option name requested.}

\item{value}{A value to check for the option (\code{NULL} to choose not to check
the value).}

\item{options}{List of changes to the default PKNCA options (see
\code{PKNCA.options()})}
}
\value{
The value of the option first from the \code{options} list and if it is
not there then from the current settings.
}
\description{
Choose either the value from an option list or the current set value for an
option.
}
\seealso{
Other PKNCA calculation and summary settings: 
\code{\link{PKNCA.options}()},
\code{\link{PKNCA.set.summary}()}
}
\concept{PKNCA calculation and summary settings}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/001-add.interval.col.R
\name{sort.interval.cols}
\alias{sort.interval.cols}
\title{Sort the interval columns by dependencies.}
\usage{
\method{sort}{interval.cols}()
}
\description{
Columns are always to the right of columns that they depend on.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cleaners.R
\name{clean.conc.blq}
\alias{clean.conc.blq}
\title{Handle BLQ values in the concentration measurements as requested by the user.}
\usage{
clean.conc.blq(
  conc,
  time,
  ...,
  options = list(),
  conc.blq = NULL,
  conc.na = NULL,
  check = TRUE
)
}
\arguments{
\item{conc}{Measured concentrations}

\item{time}{Time of the measurement of the concentrations}

\item{\dots}{Additional arguments passed to clean.conc.na}

\item{options}{List of changes to the default PKNCA options (see
\code{PKNCA.options()})}

\item{conc.blq}{How to handle a BLQ value that is between above LOQ values?
See details for description.}

\item{conc.na}{How to handle NA concentrations.  (See \code{\link[=clean.conc.na]{clean.conc.na()}})}

\item{check}{Run \code{\link[=assert_conc_time]{assert_conc_time()}}?}
}
\value{
The concentration and time measurements (data frame) filtered and
cleaned as requested relative to BLQ in the middle.
}
\description{
Handle BLQ values in the concentration measurements as requested by the user.
}
\details{
\code{NA} concentrations (and their associated times) will be handled as
described in \code{\link[=clean.conc.na]{clean.conc.na()}} before working with the BLQ values.  The
method for handling NA concentrations can affect the output of which points
are considered BLQ and which are considered "middle".  Values are
considered BLQ if they are 0.

\code{conc.blq} can be set either a scalar indicating what should be done for
all BLQ values or a list with elements either named "first", "middle" and "last" or "before.tmax" and "after.tmax"
each set to a scalar.

The meaning of each of the list elements is:
\describe{
\item{first}{Values up to the first non-BLQ value.  Note
that if all values are BLQ, this includes all values.}
\item{middle}{Values that are BLQ between the first and last
non-BLQ values.}
\item{last}{Values that are BLQ after the last non-BLQ value}
\item{before.tmax}{Values that are BLQ before the time at first maximum concentration}
\item{after.tmax}{Values that are BLQ after the time at first maximum concentration}
}

The valid settings for each are:
\describe{
\item{"drop"}{Drop the BLQ values}
\item{"keep"}{Keep the BLQ values}
\item{a number}{Set the BLQ values to that number}
}
}
\seealso{
Other Data cleaners: 
\code{\link{clean.conc.na}()}
}
\concept{Data cleaners}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pk.calc.simple.R
\name{pk.calc.aucpext}
\alias{pk.calc.aucpext}
\title{Calculate the AUC percent extrapolated}
\usage{
pk.calc.aucpext(auclast, aucinf)
}
\arguments{
\item{auclast}{the area under the curve from time 0 to the last measurement
above the limit of quantification}

\item{aucinf}{the area under the curve from time 0 to infinity}
}
\value{
The numeric value of the AUC percent extrapolated or \code{NA_real_} if
any of the following are true \code{is.na(aucinf)}, \code{is.na(auclast)},
\code{aucinf <= 0}, or \code{auclast <= 0}.
}
\description{
Calculate the AUC percent extrapolated
}
\details{
aucpext is \code{100*(1-auclast/aucinf)}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pk.calc.simple.R
\name{pk.calc.tlag}
\alias{pk.calc.tlag}
\title{Determine the observed lag time (time before the first
concentration above the limit of quantification or above the first
concentration in the interval)}
\usage{
pk.calc.tlag(conc, time)
}
\arguments{
\item{conc}{Measured concentrations}

\item{time}{Time of the measurement of the concentrations}
}
\value{
The time associated with the first increasing concentration
}
\description{
Determine the observed lag time (time before the first
concentration above the limit of quantification or above the first
concentration in the interval)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pk.calc.simple.R
\name{pk.calc.aucabove}
\alias{pk.calc.aucabove}
\title{Calculate the AUC above a given concentration}
\usage{
pk.calc.aucabove(conc, time, conc_above = NA_real_, ..., options = list())
}
\arguments{
\item{conc}{Measured concentrations}

\item{time}{Time of the measurement of the concentrations}

\item{conc_above}{The concentration to be above}

\item{...}{Extra arguments.  Currently, the only extra argument that is used
is \code{method} as described in the details section.}

\item{options}{List of changes to the default PKNCA options (see
\code{PKNCA.options()})}
}
\value{
The AUC of the concentration above the limit
}
\description{
Concentrations below the given concentration (\code{conc_above}) will be set
to zero.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defunct.R
\name{defunct}
\alias{defunct}
\alias{check.conc.time}
\title{The following functions are defunct}
\usage{
check.conc.time(...)
}
\arguments{
\item{...}{Ignored}
}
\description{
The following functions are defunct
}
\section{Functions}{
\itemize{
\item \code{check.conc.time()}: Defunct as of version 0.11

}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/time_calc.R
\name{time_calc}
\alias{time_calc}
\title{Times relative to an event (typically dosing)}
\usage{
time_calc(time_event, time_obs, units = NULL)
}
\arguments{
\item{time_event}{A vector of times for events}

\item{time_obs}{A vector of times for observations}

\item{units}{Passed to \code{base::as.numeric.difftime()}}
}
\value{
A data.frame with columns for:
\describe{
\item{event_number_before}{The index of \code{time_event} that is the last one before \code{time_obs} or \code{NA} if none are before.}
\item{event_number_after}{The index of \code{time_event} that is the first one after \code{time_obs} or \code{NA} if none are after.}
\item{time_before}{The minimum time that the current \code{time_obs} is before a \code{time_event}, 0 if at least one \code{time_obs == time_event}.}
\item{time_after}{The minimum time that the current \code{time_obs} is after a \code{time_event}, 0 if at least one \code{time_obs == time_event}.}
\item{time_after_first}{The time after the first event (may be negative or positive).}
}

\code{time_after} and \code{time_before} are calculated if they are at the same time as
a dose, they equal zero, and otherwise, they are calculated relative to the
dose number in the \verb{event_number_*} columns.
}
\description{
Times relative to an event (typically dosing)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tss.monoexponential.R
\name{tss.monoexponential.generate.formula}
\alias{tss.monoexponential.generate.formula}
\title{A helper function to generate the formula and starting values for the
parameters in monoexponential models.}
\usage{
tss.monoexponential.generate.formula(data)
}
\arguments{
\item{data}{The data used for the model}
}
\value{
a list with elements for each of the variables
}
\description{
A helper function to generate the formula and starting values for the
parameters in monoexponential models.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-PKNCAconc.R, R/class-PKNCAdata.R,
%   R/class-PKNCAdose.R, R/class-PKNCAresults.R
\name{getGroups.PKNCAconc}
\alias{getGroups.PKNCAconc}
\alias{getGroups.PKNCAdata}
\alias{getGroups.PKNCAdose}
\alias{getGroups.PKNCAresults}
\title{Get the groups (right hand side after the \code{|} from a PKNCA
object).}
\usage{
\method{getGroups}{PKNCAconc}(
  object,
  form = stats::formula(object),
  level,
  data = as.data.frame(object),
  sep
)

\method{getGroups}{PKNCAdata}(object, ...)

\method{getGroups}{PKNCAdose}(...)

\method{getGroups}{PKNCAresults}(
  object,
  form = formula(object$data$conc),
  level,
  data = object$result,
  sep
)
}
\arguments{
\item{object}{The object to extract the data from}

\item{form}{The formula to extract the data from (defaults to the formula
from \code{object})}

\item{level}{optional.  If included, this specifies the level(s) of the
groups to include.  If a numeric scalar, include the first \code{level} number
of groups.  If a numeric vector, include each of the groups specified by
the number.  If a character vector, include the named group levels.}

\item{data}{The data to extract the groups from (defaults to the data from
\code{object})}

\item{sep}{Unused (kept for compatibility with the nlme package)}

\item{...}{Arguments passed to other getGroups functions}
}
\value{
A data frame with the (selected) group columns.

A data frame with the (selected) group columns.
}
\description{
Get the groups (right hand side after the \code{|} from a PKNCA
object).

Get the groups (right hand side after the \code{|} from a PKNCA
object).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pk.calc.c0.R
\name{pk.calc.c0}
\alias{pk.calc.c0}
\alias{pk.calc.c0.method.logslope}
\alias{pk.calc.c0.method.c0}
\alias{pk.calc.c0.method.c1}
\alias{pk.calc.c0.method.set0}
\alias{pk.calc.c0.method.cmin}
\title{Estimate the concentration at dosing time for an IV bolus dose.}
\usage{
pk.calc.c0(
  conc,
  time,
  time.dose = 0,
  method = c("c0", "logslope", "c1", "cmin", "set0"),
  check = TRUE
)

pk.calc.c0.method.logslope(conc, time, time.dose = 0, check = TRUE)

pk.calc.c0.method.c0(conc, time, time.dose = 0, check = TRUE)

pk.calc.c0.method.c1(conc, time, time.dose = 0, check = TRUE)

pk.calc.c0.method.set0(conc, time, time.dose = 0, check = TRUE)

pk.calc.c0.method.cmin(conc, time, time.dose = 0, check = TRUE)
}
\arguments{
\item{conc}{Measured concentrations}

\item{time}{Time of the measurement of the concentrations}

\item{time.dose}{The time when dosing occurred}

\item{method}{The order of methods to test (see details)}

\item{check}{Check the \code{conc} and \code{time} inputs}
}
\value{
The estimated concentration at time 0.
}
\description{
Estimate the concentration at dosing time for an IV bolus dose.
}
\details{
Methods available for interpolation are below, and each
has its own specific function.

\describe{
\item{\code{c0}}{If the observed \code{conc} at \code{time.dose} is nonzero, return that.  This method should usually be used first for single-dose IV bolus data in case nominal time zero is measured.}
\item{\code{logslope}}{Compute the semilog line between the first two measured times, and use that line to extrapolate backward to \code{time.dose}}
\item{\code{c1}}{Use the first point after \code{time.dose}}
\item{\code{cmin}}{Set c0 to cmin during the interval.  This method should usually be used for multiple-dose oral data and IV infusion data.}
\item{\code{set0}}{Set c0 to zero (regardless of any other data).  This method should usually be used first for single-dose oral data.}
}
}
\section{Functions}{
\itemize{
\item \code{pk.calc.c0.method.logslope()}: Semilog regress the first and second points
after time.dose.  This method will return \code{NA} if the second
\code{conc} after \code{time.dose} is 0 or greater than the first.

\item \code{pk.calc.c0.method.c0()}: Use \code{C0} = \code{conc[time \%in\% time.dose]} if it is
nonzero.

\item \code{pk.calc.c0.method.c1()}: Use \code{C0} = \code{C1}.

\item \code{pk.calc.c0.method.set0()}: Use \code{C0} = 0 (typically used for single dose oral and
IV infusion)

\item \code{pk.calc.c0.method.cmin()}: Use \code{C0} = Cmin (typically used for multiple dose oral
and IV infusion but not IV bolus)

}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unit-support.R
\name{pknca_units_table}
\alias{pknca_units_table}
\title{Create a unit assignment and conversion table}
\usage{
pknca_units_table(
  concu,
  doseu,
  amountu,
  timeu,
  concu_pref = NULL,
  doseu_pref = NULL,
  amountu_pref = NULL,
  timeu_pref = NULL,
  conversions = data.frame()
)
}
\arguments{
\item{concu, doseu, amountu, timeu}{Units for concentration, dose, amount, and
time in the source data}

\item{concu_pref, doseu_pref, amountu_pref, timeu_pref}{Preferred units for
reporting; \code{conversions} will be automatically.}

\item{conversions}{An optional data.frame with columns of c("PPORRESU",
"PPSTRESU", "conversion_factor") for the original calculation units, the
standardized units, and a conversion factor to multiply the initial value
by to get a standardized value.  This argument overrides any preferred unit
conversions from \code{concu_pref}, \code{doseu_pref}, \code{amountu_pref}, or
\code{timeu_pref}.}
}
\value{
A unit conversion table with columns for "PPTESTCD" and "PPORRESU"
if \code{conversions} is not given, and adding "PPSTRESU" and
"conversion_factor" if \code{conversions} is given.
}
\description{
This data.frame is typically used for the \code{units} argument for \code{\link[=PKNCAdata]{PKNCAdata()}}.
If a unit is not given, then all of the units derived from that unit will be
\code{NA}.
}
\examples{
pknca_units_table() # only parameters that are unitless
pknca_units_table(
  concu="ng/mL", doseu="mg/kg", amountu="mg", timeu="hr"
)
pknca_units_table(
  concu="ng/mL", doseu="mg/kg", amountu="mg", timeu="hr",
  # Convert clearance and volume units to more understandable units with
  # automatic unit conversion
  conversions=data.frame(
    PPORRESU=c("(mg/kg)/(hr*ng/mL)", "(mg/kg)/(ng/mL)"),
    PPSTRESU=c("mL/hr/kg", "mL/kg")
  )
)
pknca_units_table(
  concu="mg/L", doseu="mg/kg", amountu="mg", timeu="hr",
  # Convert clearance and volume units to molar units (assuming
  conversions=data.frame(
    PPORRESU=c("mg/L", "(mg/kg)/(hr*ng/mL)", "(mg/kg)/(ng/mL)"),
    PPSTRESU=c("mmol/L", "mL/hr/kg", "mL/kg"),
    # Manual conversion of concentration units from ng/mL to mmol/L (assuming
    # a molecular weight of 138.121 g/mol)
    conversion_factor=c(1/138.121, NA, NA)
  )
)

# This will make all time-related parameters use "day" even though the
# original units are "hr"
pknca_units_table(
  concu = "ng/mL", doseu = "mg/kg", timeu = "hr", amountu = "mg",
  timeu_pref = "day"
)
}
\seealso{
The \code{units} argument for \code{\link[=PKNCAdata]{PKNCAdata()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tss.monoexponential.R
\name{pk.tss.monoexponential}
\alias{pk.tss.monoexponential}
\title{Compute the time to steady state using nonlinear, mixed-effects modeling of
trough concentrations.}
\usage{
pk.tss.monoexponential(
  ...,
  tss.fraction = 0.9,
  output = c("population", "popind", "individual", "single"),
  check = TRUE,
  verbose = FALSE
)
}
\arguments{
\item{\dots}{See \code{\link[=pk.tss.data.prep]{pk.tss.data.prep()}}}

\item{tss.fraction}{The fraction of steady-state required for calling
steady-state}

\item{output}{Which types of outputs should be produced? \code{population} is the
population estimate for time to steady-state (from an nlme model), \code{popind}
is the individual estimate (from an nlme model), \code{individual} fits each
individual separately with a gnls model (requires more than one individual;
use \code{single} for one individual), and \code{single} fits all the data to a
single gnls model.}

\item{check}{See \code{\link[=pk.tss.data.prep]{pk.tss.data.prep()}}.}

\item{verbose}{Describe models as they are run, show convergence of the model
(passed to the nlme function), and additional details while running.}
}
\value{
A scalar float for the first time when steady-state is achieved or
\code{NA} if it is not observed.
}
\description{
Trough concentrations are selected as concentrations at the time of dosing.
An exponential curve is then fit through the data with a different magnitude
by treatment (as a factor) and a random steady-state concentration and time
to stead-state by subject (see \code{random.effects} argument).
}
\references{
Maganti, L., Panebianco, D.L. & Maes, A.L. Evaluation of Methods for
Estimating Time to Steady State with Examples from Phase 1 Studies. AAPS J
10, 141–147 (2008). https://doi.org/10.1208/s12248-008-9014-y
}
\seealso{
Other Time to steady-state calculations: 
\code{\link{pk.tss}()},
\code{\link{pk.tss.stepwise.linear}()}
}
\concept{Time to steady-state calculations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/choose.intervals.R
\name{find.tau}
\alias{find.tau}
\title{Find the repeating interval within a vector of doses}
\usage{
find.tau(x, na.action = stats::na.omit, options = list(), tau.choices = NULL)
}
\arguments{
\item{x}{the vector to find the interval within}

\item{na.action}{What to do with NAs in \code{x}}

\item{options}{List of changes to the default PKNCA options (see
\code{PKNCA.options()})}

\item{tau.choices}{the intervals to look for if the doses are not all equally
spaced.}
}
\value{
A scalar indicating the repeating interval with the most repetition.
\enumerate{
\item If all values are \code{NA} then NA is returned.
\item If all values are the same, then 0 is returned.
\item If all values are equally spaced, then that spacing is
returned.
\item If one of the \code{choices} can minimize the number of
unique values, then that is returned.
\item If none of the \code{choices} can minimize the number of
unique values, then -1 is returned.
}
}
\description{
This is intended to find the interval over which x repeats by the rule
\code{unique(mod(x, interval))} is minimized.
}
\seealso{
Other Interval determination: 
\code{\link{choose.auc.intervals}()}
}
\concept{Interval determination}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dplyr.R
\name{filter.PKNCAresults}
\alias{filter.PKNCAresults}
\alias{filter.PKNCAconc}
\alias{filter.PKNCAdose}
\title{dplyr filtering for PKNCA}
\usage{
\method{filter}{PKNCAresults}(.data, ..., .preserve = FALSE)

\method{filter}{PKNCAconc}(.data, ..., .preserve = FALSE)

\method{filter}{PKNCAdose}(.data, ..., .preserve = FALSE)
}
\arguments{
\item{.data}{A data frame, data frame extension (e.g. a tibble), or a
lazy data frame (e.g. from dbplyr or dtplyr). See \emph{Methods}, below, for
more details.}

\item{...}{<\code{\link[rlang:args_data_masking]{data-masking}}> Expressions that
return a logical value, and are defined in terms of the variables in
\code{.data}. If multiple expressions are included, they are combined with the
\code{&} operator. Only rows for which all conditions evaluate to \code{TRUE} are
kept.}

\item{.preserve}{Relevant when the \code{.data} input is grouped.
If \code{.preserve = FALSE} (the default), the grouping structure
is recalculated based on the resulting data, otherwise the grouping is kept as is.}
}
\description{
dplyr filtering for PKNCA
}
\seealso{
Other dplyr verbs: 
\code{\link{group_by.PKNCAresults}()},
\code{\link{inner_join.PKNCAresults}()},
\code{\link{mutate.PKNCAresults}()}
}
\concept{dplyr verbs}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pk.calc.urine.R
\name{pk.calc.fe}
\alias{pk.calc.fe}
\title{Calculate fraction excreted (typically in urine or feces)}
\usage{
pk.calc.fe(ae, dose)
}
\arguments{
\item{ae}{The amount excreted (as a numeric scalar or vector)}

\item{dose}{The dose (as a numeric scalar or vector)}
}
\value{
The fraction of dose excreted
}
\description{
Calculate fraction excreted (typically in urine or feces)
}
\details{
fe is \code{sum(ae)/dose}

The units for \code{ae} and \code{dose} should be the same so that \code{ae/dose}
is a unitless fraction.
}
\seealso{
\code{\link[=pk.calc.ae]{pk.calc.ae()}}, \code{\link[=pk.calc.clr]{pk.calc.clr()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pk.calc.all.R
\name{any_sparse_dense_in_interval}
\alias{any_sparse_dense_in_interval}
\title{Determine if there are any sparse or dense calculations requested within an interval}
\usage{
any_sparse_dense_in_interval(interval, sparse)
}
\arguments{
\item{interval}{An interval specification}

\item{sparse}{Are the concentration-time data sparse PK (commonly used in
small nonclinical species or with terminal or difficult sampling) or dense
PK (commonly used in clinical studies or larger nonclinical species)?}
}
\value{
A logical value indicating if the interval requests any sparse (if
\code{sparse=TRUE}) or dense (if \code{sparse=FALSE}) calculations.
}
\description{
Determine if there are any sparse or dense calculations requested within an interval
}
\keyword{Internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tss.R
\name{pk.tss}
\alias{pk.tss}
\title{Compute the time to steady-state (tss)}
\usage{
pk.tss(..., type = c("monoexponential", "stepwise.linear"), check = TRUE)
}
\arguments{
\item{\dots}{Passed to \code{\link[=pk.tss.monoexponential]{pk.tss.monoexponential()}} or
\code{\link[=pk.tss.stepwise.linear]{pk.tss.stepwise.linear()}}.}

\item{type}{The type of Tss to calculate, either \code{stepwise.linear} or
\code{monoexponential}}

\item{check}{See \code{\link[=pk.tss.data.prep]{pk.tss.data.prep()}}}
}
\value{
A data frame with columns as defined from \code{pk.tss.monoexponential}
and/or \code{pk.tss.stepwise.linear}.
}
\description{
Compute the time to steady-state (tss)
}
\seealso{
Other Time to steady-state calculations: 
\code{\link{pk.tss.monoexponential}()},
\code{\link{pk.tss.stepwise.linear}()}
}
\concept{Time to steady-state calculations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/impute.R
\name{add_impute_to_intervals}
\alias{add_impute_to_intervals}
\title{Add the imputation column to the intervals, if it is not already there}
\usage{
add_impute_to_intervals(object)
}
\arguments{
\item{object}{The PKNCAdata object to impute data within}
}
\value{
The PKNCAdata object with an impute column added to the intervals (if
it is not already there) and the object$impute set to that column name
}
\description{
Add the imputation column to the intervals, if it is not already there
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unit-support.R
\name{pknca_unit_conversion}
\alias{pknca_unit_conversion}
\title{Perform unit conversion (if possible) on PKNCA results}
\usage{
pknca_unit_conversion(result, units, allow_partial_missing_units = FALSE)
}
\arguments{
\item{result}{The results data.frame}

\item{units}{The unit conversion table}

\item{allow_partial_missing_units}{Should missing units be allowed for some
but not all parameters?}
}
\value{
The result table with units converted
}
\description{
Perform unit conversion (if possible) on PKNCA results
}
\keyword{Internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assertions.R
\name{assert_conc}
\alias{assert_conc}
\alias{assert_time}
\alias{assert_conc_time}
\title{Verify that concentration measurements are valid}
\usage{
assert_conc(conc, any_missing_conc = TRUE)

assert_time(time, sorted_time = TRUE)

assert_conc_time(conc, time, any_missing_conc = TRUE, sorted_time = TRUE)
}
\arguments{
\item{conc}{Measured concentrations}

\item{any_missing_conc}{Are any concentration values allowed to be \code{NA}?}

\item{time}{Time of the measurement of the concentrations}

\item{sorted_time}{Must the time be unique and monotonically increasing?}
}
\value{
\code{conc} or give an informative error

\code{time} or give an informative error

A data.frame with columns named "conc" and "time" or an informative
error
}
\description{
If the concentrations or times are invalid, will provide an error.
Reasons for being invalid are
\itemize{
\item \code{time} is not a number
\item \code{conc} is not a number
\item Any \code{time} value is NA
\item \code{time} is not monotonically increasing
\item \code{conc} and \code{time} are not the same length
}
}
\details{
Some cases may generate warnings but allow the data to proceed.
\itemize{
\item A negative concentration is often but not always an
error; it will generate a warning.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pk.calc.simple.R
\name{pk.calc.kel}
\alias{pk.calc.kel}
\title{Calculate the elimination rate (Kel)}
\usage{
pk.calc.kel(mrt)
}
\arguments{
\item{mrt}{the mean residence time}

\item{kel}{is \code{1/mrt}, not to be confused with lambda.z.}
}
\value{
the numeric value of the elimination rate
}
\description{
Calculate the elimination rate (Kel)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pk.calc.simple.R
\name{pk.calc.ptr}
\alias{pk.calc.ptr}
\title{Determine the peak-to-trough ratio}
\usage{
pk.calc.ptr(cmax, ctrough)
}
\arguments{
\item{cmax}{The maximum observed concentration}

\item{ctrough}{The last concentration in an interval}
}
\value{
The ratio of cmax to ctrough (if ctrough == 0, NA)
}
\description{
Determine the peak-to-trough ratio
}
\details{
ptr is \code{cmax/ctrough}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assertions.R
\name{assert_lambdaz}
\alias{assert_lambdaz}
\title{Assert that a lambda.z value is valid}
\usage{
assert_lambdaz(
  lambda.z,
  any.missing = TRUE,
  .var.name = checkmate::vname(lambda.z)
)
}
\arguments{
\item{lambda.z}{The elimination rate (in units of inverse time) for
extrapolation}

\item{any.missing}{[\code{logical(1)}]\cr
Are vectors with missing values allowed? Default is \code{TRUE}.}

\item{.var.name}{[\code{character(1)}]\cr
Name of the checked object to print in assertions. Defaults to
the heuristic implemented in \code{\link[checkmate]{vname}}.}
}
\value{
\code{lambda.z} or an informative error
}
\description{
Assert that a lambda.z value is valid
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-PKNCAconc.R, R/class-PKNCAdose.R
\name{setDuration.PKNCAconc}
\alias{setDuration.PKNCAconc}
\alias{setDuration}
\alias{setDuration.PKNCAdose}
\title{Set the duration of dosing or measurement}
\usage{
\method{setDuration}{PKNCAconc}(object, duration, ...)

setDuration(object, ...)

\method{setDuration}{PKNCAdose}(object, duration, rate, dose, ...)
}
\arguments{
\item{object}{An object to set a duration on}

\item{duration}{The value to set for the duration or the name of the column
in the data to use for the duration.}

\item{...}{Arguments passed to another setDuration function}

\item{rate}{(for PKNCAdose objects only) The rate of infusion}

\item{dose}{(for PKNCAdose objects only) The dose amount}
}
\value{
The object with duration set
}
\description{
Set the duration of dosing or measurement
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/superposition.R
\name{superposition}
\alias{superposition}
\alias{superposition.PKNCAconc}
\alias{superposition.numeric}
\title{Compute noncompartmental superposition for repeated dosing}
\usage{
superposition(conc, ...)

\method{superposition}{PKNCAconc}(conc, ...)

\method{superposition}{numeric}(
  conc,
  time,
  dose.input = NULL,
  tau,
  dose.times = 0,
  dose.amount,
  n.tau = Inf,
  options = list(),
  lambda.z,
  clast.pred = FALSE,
  tlast,
  additional.times = numeric(),
  check.blq = TRUE,
  method = NULL,
  auc.type = "AUCinf",
  steady.state.tol = 0.001,
  ...
)
}
\arguments{
\item{conc}{Measured concentrations}

\item{...}{Additional arguments passed to the \code{half.life} function if
required to compute \code{lambda.z}.}

\item{time}{Time of the measurement of the concentrations}

\item{dose.input}{The dose given to generate the \code{conc} and \code{time} inputs. If
missing, output doses will be assumed to be equal to the input dose.}

\item{tau}{The dosing interval}

\item{dose.times}{The time of dosing within the dosing interval. The
\code{min(dose.times)} must be >= 0, and the \code{max(dose.times)} must be < \code{tau}.
There may be more than one dose times given as a vector.}

\item{dose.amount}{The doses given for the output.  Linear proportionality
will be used from the input to output if they are not equal.  The length of
dose.amount must be either 1 or matching the length of \code{dose.times}.}

\item{n.tau}{The number of tau dosing intervals to simulate or \code{Inf} for
steady-state.}

\item{options}{List of changes to the default PKNCA options (see
\code{PKNCA.options()})}

\item{lambda.z}{The elimination rate (in units of inverse time) for
extrapolation}

\item{clast.pred}{To use predicted as opposed to observed Clast, either give
the value for clast.pred here or set it to true (for automatic calculation
from the half-life).}

\item{tlast}{The time of last observed concentration above the limit of
quantification.  This is calculated if not provided.}

\item{additional.times}{Times to include in the final outputs in addition to
the standard times (see details).  All \code{min(additional.times)} must be >=
0, and the \code{max(additional.times)} must be <= \code{tau}.}

\item{check.blq}{Must the first concentration measurement be below the limit
of quantification?}

\item{method}{The method for integration (one of 'lin up/log down',
'lin-log', or 'linear')}

\item{auc.type}{The type of AUC to compute.  Choices are 'AUCinf', 'AUClast',
and 'AUCall'.}

\item{steady.state.tol}{The tolerance for assessing if steady-state has been
achieved (between 0 and 1, exclusive).}
}
\value{
A data frame with columns named "conc" and "time".
}
\description{
Compute noncompartmental superposition for repeated dosing
}
\details{
The returned superposition times will include all of the following
times: 0 (zero), \code{dose.times}, \verb{time modulo tau} (shifting \code{time} for each
dose time as well), \code{additional.times}, and \code{tau}.
}
\seealso{
\code{\link[=interp.extrap.conc]{interp.extrap.conc()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pk.calc.simple.R
\name{pk.calc.mrt.md}
\alias{pk.calc.mrt.md}
\title{Calculate the mean residence time (MRT) for multiple-dose data with nonlinear
kinetics.}
\usage{
pk.calc.mrt.md(auctau, aumctau, aucinf, tau)
}
\arguments{
\item{auctau}{the AUC from time 0 to the end of the dosing interval (tau).}

\item{aumctau}{the AUMC from time 0 to the end of the dosing interval (tau).}

\item{aucinf}{the AUC from time 0 to infinity (typically using single-dose
data)}

\item{tau}{The dosing interval}
}
\description{
Calculate the mean residence time (MRT) for multiple-dose data with nonlinear
kinetics.
}
\details{
mrt.md is \code{aumctau/auctau + tau*(aucinf-auctau)/auctau} and should
only be used for multiple dosing with equal intervals between doses.

Note that if \code{aucinf == auctau} (as would be the assumption with
linear kinetics), the equation becomes the same as the single-dose MRT.
}
\seealso{
\code{\link[=pk.calc.mrt]{pk.calc.mrt()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/002-pk.business.rules.R
\name{business.mean}
\alias{business.mean}
\alias{business.sd}
\alias{business.cv}
\alias{business.geomean}
\alias{business.geocv}
\alias{business.min}
\alias{business.max}
\alias{business.median}
\alias{business.range}
\title{Generate functions to do the named function (e.g. mean) applying
the business rules.}
\usage{
business.mean(x, ...)

business.sd(x, ...)

business.cv(x, ...)

business.geomean(x, ...)

business.geocv(x, ...)

business.min(x, ...)

business.max(x, ...)

business.median(x, ...)

business.range(x, ...)
}
\arguments{
\item{x}{vector to be passed to the various functions}

\item{...}{Additional arguments to be passed to the underlying function.}
}
\value{
The value of the various functions or NA if too many values are
missing
}
\description{
Generate functions to do the named function (e.g. mean) applying
the business rules.
}
\section{Functions}{
\itemize{
\item \code{business.sd()}: Compute the standard deviation with business rules.

\item \code{business.cv()}: Compute the coefficient of variation with business rules.

\item \code{business.geomean()}: Compute the geometric mean with business rules.

\item \code{business.geocv()}: Compute the geometric coefficient of variation with business rules.

\item \code{business.min()}: Compute the minimum with business rules.

\item \code{business.max()}: Compute the maximum with business rules.

\item \code{business.median()}: Compute the median with business rules.

\item \code{business.range()}: Compute the range with business rules.

}}
\seealso{
\code{\link[=pk.business]{pk.business()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/001-add.interval.col.R
\name{add.interval.col}
\alias{add.interval.col}
\title{Add columns for calculations within PKNCA intervals}
\usage{
add.interval.col(
  name,
  FUN,
  values = c(FALSE, TRUE),
  unit_type,
  pretty_name,
  depends = NULL,
  desc = "",
  sparse = FALSE,
  formalsmap = list(),
  datatype = c("interval", "individual", "population")
)
}
\arguments{
\item{name}{The column name as a character string}

\item{FUN}{The function to run (as a character string) or \code{NA} if the
parameter is automatically calculated when calculating another parameter.}

\item{values}{Valid values for the column}

\item{unit_type}{The type of units to use for assigning and converting units.}

\item{pretty_name}{The name of the parameter to use for printing in summary
tables with units.  (If an analysis does not include units, then the normal
name is used.)}

\item{depends}{Character vector of columns that must be run before this
column.}

\item{desc}{A human-readable description of the parameter (<=40 characters to
comply with SDTM)}

\item{sparse}{Is the calculation for sparse PK?}

\item{formalsmap}{A named list mapping parameter names in the function call
to NCA parameter names.  See the details for information on use of
\code{formalsmap}.}

\item{datatype}{The type of data used for the calculation}
}
\value{
NULL (Calling this function has a side effect of changing the
available intervals for calculations)
}
\description{
Add columns for calculations within PKNCA intervals
}
\details{
The \code{formalsmap} argument enables mapping some alternate formal
argument names to parameters.  It is used to generalize functions that may
use multiple similar arguments (such as the variants of mean residence time).
The names of the list should correspond to function formal parameter names
and the values should be one of the following:

\itemize{
\item{For the current interval:}
\describe{
\item{character strings of NCA parameter name}{The value of the parameter calculated for the current interval.}
\item{"conc"}{Concentration measurements for the current interval.}
\item{"time"}{Times associated with concentration measurements for the current interval (values start at 0 at the beginning of the current interval).}
\item{"volume"}{Volume associated with concentration measurements for the current interval (typically applies for excretion parameters like urine).}
\item{"duration.conc"}{Durations associated with concentration measurements for the current interval.}
\item{"dose"}{Dose amounts assocuated with the current interval.}
\item{"time.dose"}{Time of dose start associated with the current interval (values start at 0 at the beginning of the current interval).}
\item{"duration.dose"}{Duration of dose (typically infusion duration) for doses in the current interval.}
\item{"route"}{Route of dosing for the current interval.}
\item{"start"}{Time of interval start.}
\item{"end"}{Time of interval end.}
\item{"options"}{PKNCA.options governing calculations.}
}
\item{For the current group:}
\describe{
\item{"conc.group"}{Concentration measurements for the current group.}
\item{"time.group"}{Times associated with concentration measurements for the current group (values start at 0 at the beginning of the current interval).}
\item{"volume.group"}{Volume associated with concentration measurements for the current interval (typically applies for excretion parameters like urine).}
\item{"duration.conc.group"}{Durations assocuated with concentration measurements for the current group.}
\item{"dose.group"}{Dose amounts assocuated with the current group.}
\item{"time.dose.group"}{Time of dose start associated with the current group (values start at 0 at the beginning of the current interval).}
\item{"duration.dose.group"}{Duration of dose (typically infusion duration) for doses in the current group.}
\item{"route.group"}{Route of dosing for the current group.}
}
}
}
\examples{
\dontrun{
add.interval.col("cmax",
                 FUN="pk.calc.cmax",
                 values=c(FALSE, TRUE),
                 unit_type="conc",
                 pretty_name="Cmax",
                 desc="Maximum observed concentration")
add.interval.col("cmax.dn",
                 FUN="pk.calc.dn",
                 values=c(FALSE, TRUE),
                 unit_type="conc_dosenorm",
                 pretty_name="Cmax (dose-normalized)",
                 desc="Maximum observed concentration, dose normalized",
                 formalsmap=list(parameter="cmax"),
                 depends="cmax")
}
}
\seealso{
Other Interval specifications: 
\code{\link{check.interval.deps}()},
\code{\link{check.interval.specification}()},
\code{\link{choose.auc.intervals}()},
\code{\link{get.interval.cols}()},
\code{\link{get.parameter.deps}()}
}
\concept{Interval specifications}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pk.calc.urine.R
\name{pk.calc.ae}
\alias{pk.calc.ae}
\title{Calculate amount excreted (typically in urine or feces)}
\usage{
pk.calc.ae(conc, volume, check = TRUE)
}
\arguments{
\item{conc}{Measured concentrations}

\item{volume}{The volume (or mass) of the sample}

\item{check}{Should the concentration and volume data be checked?}
}
\value{
The amount excreted during the interval
}
\description{
Calculate amount excreted (typically in urine or feces)
}
\details{
ae is \code{sum(conc*volume)}.

The units for the concentration and volume should match such that
\code{sum(conc*volume)} has units of mass or moles.
}
\seealso{
\code{\link[=pk.calc.clr]{pk.calc.clr()}}, \code{\link[=pk.calc.fe]{pk.calc.fe()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-PKNCAdata.R
\name{PKNCAdata}
\alias{PKNCAdata}
\alias{PKNCAdata.PKNCAconc}
\alias{PKNCAdata.PKNCAdose}
\alias{PKNCAdata.default}
\title{Create a PKNCAdata object.}
\usage{
PKNCAdata(data.conc, data.dose, ...)

\method{PKNCAdata}{PKNCAconc}(data.conc, data.dose, ...)

\method{PKNCAdata}{PKNCAdose}(data.conc, data.dose, ...)

\method{PKNCAdata}{default}(
  data.conc,
  data.dose,
  ...,
  formula.conc,
  formula.dose,
  impute = NA_character_,
  intervals,
  units,
  options = list()
)
}
\arguments{
\item{data.conc}{Concentration data as a \code{PKNCAconc} object or a data frame}

\item{data.dose}{Dosing data as a \code{PKNCAdose} object (see details)}

\item{...}{arguments passed to \code{PKNCAdata.default}}

\item{formula.conc}{Formula for making a \code{PKNCAconc} object with \code{data.conc}.
This must be given if \code{data.conc} is a data.frame, and it must not be given
if \code{data.conc} is a \code{PKNCAconc} object.}

\item{formula.dose}{Formula for making a \code{PKNCAdose} object with \code{data.dose}.
This must be given if \code{data.dose} is a data.frame, and it must not be given
if \code{data.dose} is a \code{PKNCAdose} object.}

\item{impute}{Methods for imputation.  \code{NA} for to search for the column
named "impute" in the intervals or no imputation if that column does not
exist, a comma-or space-separated list of names, or the name of a column in
the \code{intervals} data.frame.  See
\code{vignette("v08-data-imputation", package="PKNCA")} for more details.}

\item{intervals}{A data frame with the AUC interval specifications as defined
in \code{\link[=check.interval.specification]{check.interval.specification()}}.  If missing, this will be
automatically chosen by \code{\link[=choose.auc.intervals]{choose.auc.intervals()}}. (see details)}

\item{units}{A data.frame of unit assignments and conversions as created by
\code{\link[=pknca_units_table]{pknca_units_table()}}}

\item{options}{List of changes to the default PKNCA options (see
\code{PKNCA.options()})}
}
\value{
A PKNCAdata object with concentration, dose, interval, and
calculation options stored (note that PKNCAdata objects can also have
results after a NCA calculations are done to the data).
}
\description{
\code{PKNCAdata()} combines \code{PKNCAconc} and \code{PKNCAdose} objects and adds in the
intervals for PK calculations.
}
\details{
If \code{data.dose} is not given or is \code{NA}, then the \code{intervals} must be
given.  At least one of \code{data.dose} and \code{intervals} must be given.
}
\seealso{
\code{\link[=choose.auc.intervals]{choose.auc.intervals()}}, \code{\link[=pk.nca]{pk.nca()}}, \code{\link[=pknca_units_table]{pknca_units_table()}}

Other PKNCA objects: 
\code{\link{PKNCAconc}()},
\code{\link{PKNCAdose}()},
\code{\link{PKNCAresults}()}
}
\concept{PKNCA objects}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/auciv.R
\name{pk.calc.auciv}
\alias{pk.calc.auciv}
\alias{pk.calc.auciv_pbext}
\title{Calculate AUC for intravenous dosing}
\usage{
pk.calc.auciv(conc, time, c0, auc, ..., options = list(), check = TRUE)

pk.calc.auciv_pbext(auc, auciv)
}
\arguments{
\item{conc}{Measured concentrations}

\item{time}{Time of the measurement of the concentrations}

\item{c0}{The concentration at time 0, typically calculated using
\code{pk.calc.c0()}}

\item{auc}{The AUC calculated using \code{conc} and \code{time} without \code{c0} (it may be
calculated using any method)}

\item{...}{For functions other than \code{pk.calc.auxc}, these values are passed
to \code{pk.calc.auxc}}

\item{options}{List of changes to the default PKNCA options (see
\code{PKNCA.options()})}

\item{check}{Run \code{\link[=assert_conc_time]{assert_conc_time()}}, \code{\link[=clean.conc.blq]{clean.conc.blq()}}, and
\code{\link[=clean.conc.na]{clean.conc.na()}}?}

\item{auciv}{The AUC calculated using \code{c0}}
}
\value{
\code{pk.calc.auciv}: The AUC calculated using \code{c0}

\code{pk.calc.auciv_pctbackextrap}: The AUC percent back-extrapolated
}
\description{
Calculate AUC for intravenous dosing
}
\details{
The AUC for intravenous (IV) dosing extrapolates the AUC back from
the first measurement to time 0 using c0 and the AUC calculated by another
method (for example the auclast).

The calculation method takes the following steps:

\itemize{
\item{\code{time = 0} must be present in the data with a measured concentration.}
\item{The AUC between \code{time = 0} and the next time point is calculated (\code{auc_first}).}
\item{The AUC between \code{time = 0} with \code{c0} and the next time point is calculated (\code{auc_second}).}
\item{The final AUC is the initial AUC plus the difference between the two AUCs (\code{auc_final <- auc + auc_second - auc_first}).}
}

The calculation for back-extrapolation is \code{100*(1 - auc/auciv)}.
}
\section{Functions}{
\itemize{
\item \code{pk.calc.auciv_pbext()}: Calculate the percent back-extrapolated AUC for IV
administration

}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pk.calc.simple.R
\name{pk.calc.count_conc}
\alias{pk.calc.count_conc}
\alias{pk.calc.count_conc_measured}
\title{Count the number of concentration measurements in an interval}
\usage{
pk.calc.count_conc(conc, check = TRUE)

pk.calc.count_conc_measured(conc, check = TRUE)
}
\arguments{
\item{conc}{Measured concentrations}

\item{check}{Run \code{\link[=assert_conc]{assert_conc()}}?}
}
\value{
a count of the non-missing concentrations (0 if all concentrations
are missing)

a count of the non-missing, measured (not below or above the limit
of quantification) concentrations (0 if all concentrations are missing)
}
\description{
\code{count_conc} and \code{count_conc_measured} are typically used for quality control
on the data to ensure that there are a sufficient number of non-missing
samples for a calculation and to ensure that data are consistent between
individuals.
}
\section{Functions}{
\itemize{
\item \code{pk.calc.count_conc_measured()}: Count the number of concentration measurements
that are not missing, above, or below the limit of quantification in an
interval

}}
\seealso{
Other NCA parameters for concentrations during the intervals: 
\code{\link{pk.calc.clast.obs}()},
\code{\link{pk.calc.cmax}()},
\code{\link{pk.calc.cstart}()},
\code{\link{pk.calc.ctrough}()}
}
\concept{NCA parameters for concentrations during the intervals}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pk.calc.simple.R
\name{pk.calc.cstart}
\alias{pk.calc.cstart}
\title{Determine the concentration at the beginning of the interval}
\usage{
pk.calc.cstart(conc, time, start)
}
\arguments{
\item{conc}{Measured concentrations}

\item{time}{Time of the measurement of the concentrations}

\item{start}{The start time of the interval}
}
\value{
The concentration when \code{time == end}.  If none match, then \code{NA}
}
\description{
Determine the concentration at the beginning of the interval
}
\seealso{
Other NCA parameters for concentrations during the intervals: 
\code{\link{pk.calc.clast.obs}()},
\code{\link{pk.calc.cmax}()},
\code{\link{pk.calc.count_conc}()},
\code{\link{pk.calc.ctrough}()}
}
\concept{NCA parameters for concentrations during the intervals}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-PKNCAconc.R, R/class-PKNCAdata.R,
%   R/class-PKNCAresults.R, R/sparse.R
\name{is_sparse_pk.PKNCAconc}
\alias{is_sparse_pk.PKNCAconc}
\alias{is_sparse_pk.PKNCAdata}
\alias{is_sparse_pk.PKNCAresults}
\alias{is_sparse_pk}
\title{Is a PKNCA object used for sparse PK?}
\usage{
\method{is_sparse_pk}{PKNCAconc}(object)

\method{is_sparse_pk}{PKNCAdata}(object)

\method{is_sparse_pk}{PKNCAresults}(object)

is_sparse_pk(object)
}
\arguments{
\item{object}{The object to see if it includes sparse PK}
}
\value{
\code{TRUE} if sparse and \code{FALSE} if dense (not sparse)
}
\description{
Is a PKNCA object used for sparse PK?
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assertions.R
\name{assert_dosetau}
\alias{assert_dosetau}
\title{Assert that a value is a dosing interval}
\usage{
assert_dosetau(tau)
}
\arguments{
\item{tau}{The dosing interval}
}
\value{
\code{tau} or an informative error
}
\description{
Assert that a value is a dosing interval
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/impute.R
\name{PKNCA_impute_method}
\alias{PKNCA_impute_method}
\alias{PKNCA_impute_method_start_conc0}
\alias{PKNCA_impute_method_start_cmin}
\alias{PKNCA_impute_method_start_predose}
\title{Methods for imputation of data with PKNCA}
\usage{
PKNCA_impute_method_start_conc0(conc, time, start = 0, ..., options = list())

PKNCA_impute_method_start_cmin(conc, time, start, end, ..., options = list())

PKNCA_impute_method_start_predose(
  conc,
  time,
  start,
  end,
  conc.group,
  time.group,
  ...,
  max_shift = NA_real_,
  options = list()
)
}
\arguments{
\item{conc}{Measured concentrations}

\item{time}{Time of the measurement of the concentrations}

\item{start}{The start time of the interval}

\item{...}{ignored}

\item{options}{List of changes to the default PKNCA options (see
\code{PKNCA.options()})}

\item{end}{The end time of the interval}

\item{conc.group}{All concentrations measured for the group}

\item{time.group}{Time of all concentrations measured for the group}

\item{max_shift}{The maximum amount of time to shift a concentration forward
(defaults to 5\% of the interval duration, i.e. \code{0.05*(end - start)}, if
\code{is.finite(end)}, and when \code{is.infinite(end)}, defaults to 5\% of the time
from start to \code{max(time)})}
}
\value{
A data.frame with one column named conc with imputed concentrations
and one column named time with the times.
}
\description{
Methods for imputation of data with PKNCA
}
\section{Functions}{
\itemize{
\item \code{PKNCA_impute_method_start_conc0()}: Add a new concentration of 0 at the start
time, even if a nonzero concentration exists at that time (usually used
with single-dose data)

\item \code{PKNCA_impute_method_start_cmin()}: Add a new concentration of the minimum during
the interval at the start time (usually used with multiple-dose data)

\item \code{PKNCA_impute_method_start_predose()}: Shift a predose concentration to become the
time zero concentration (only if a time zero concentration does not exist)

}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-PKNCAconc.R, R/class-PKNCAdose.R
\name{model.frame.PKNCAconc}
\alias{model.frame.PKNCAconc}
\alias{model.frame.PKNCAdose}
\title{Extract the columns used in the formula (in order) from a PKNCAconc
or PKNCAdose object.}
\usage{
\method{model.frame}{PKNCAconc}(formula, ...)

\method{model.frame}{PKNCAdose}(formula, ...)
}
\arguments{
\item{formula}{The object to use (parameter name is \code{formula} to use the
generic function)}

\item{\dots}{Unused}
}
\value{
A data frame with the columns from the object in formula order.
}
\description{
Extract the columns used in the formula (in order) from a PKNCAconc
or PKNCAdose object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-summary_PKNCAresults.R
\name{print.summary_PKNCAresults}
\alias{print.summary_PKNCAresults}
\title{Print the results summary}
\usage{
\method{print}{summary_PKNCAresults}(x, ...)
}
\arguments{
\item{x}{A summary_PKNCAresults object}

\item{...}{passed to print.data.frame (\code{row.names} is always set to \code{FALSE})}
}
\value{
\code{x} invisibly
}
\description{
Print the results summary
}
\seealso{
\code{\link[=summary.PKNCAresults]{summary.PKNCAresults()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/half.life.R
\name{get_halflife_points}
\alias{get_halflife_points}
\title{Determine which concentrations were used for half-life calculation}
\usage{
get_halflife_points(object)
}
\arguments{
\item{object}{A PKNCAresults object}
}
\value{
A logical vector with \code{TRUE} if the point was used for half-life
(including concentrations below the limit of quantification within the
range of times for calculation), \code{FALSE} if it was not used for half-life
but the half-life was calculated for the interval, and \code{NA} if half-life
was not calculated for the interval. If a row is excluded from all
calculations, it is set to \code{NA} as well.
}
\description{
Determine which concentrations were used for half-life calculation
}
\examples{
o_conc <- PKNCAconc(Theoph, conc~Time|Subject)
o_data <- PKNCAdata(o_conc, intervals = data.frame(start = 0, end = Inf, half.life = TRUE))
o_nca <- pk.nca(o_data)
get_halflife_points(o_nca)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assertions.R
\name{assert_numeric_between}
\alias{assert_numeric_between}
\title{Confirm that a value is greater than another value}
\usage{
assert_numeric_between(
  x,
  any.missing = FALSE,
  null.ok = FALSE,
  lower_eq = -Inf,
  lower = -Inf,
  upper = Inf,
  upper_eq = Inf,
  ...,
  .var.name = checkmate::vname(x)
)
}
\arguments{
\item{x}{[any]\cr
Object to check.}

\item{any.missing}{[\code{logical(1)}]\cr
Are vectors with missing values allowed? Default is \code{TRUE}.}

\item{null.ok}{[\code{logical(1)}]\cr
If set to \code{TRUE}, \code{x} may also be \code{NULL}.
In this case only a type check of \code{x} is performed, all additional checks are disabled.}

\item{lower_eq, upper_eq}{Values where equality is not allowed}

\item{lower}{[\code{numeric(1)}]\cr
Lower value all elements of \code{x} must be greater than or equal to.}

\item{upper}{[\code{numeric(1)}]\cr
Upper value all elements of \code{x} must be lower than or equal to.}

\item{...}{Passed to \code{checkmate::assert_numeric()}}

\item{.var.name}{[\code{character(1)}]\cr
Name of the checked object to print in assertions. Defaults to
the heuristic implemented in \code{\link[checkmate]{vname}}.}
}
\value{
\code{x}
}
\description{
Confirm that a value is greater than another value
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pk.calc.simple.R
\name{pk.calc.mrt}
\alias{pk.calc.mrt}
\alias{pk.calc.mrt.iv}
\title{Calculate the mean residence time (MRT) for single-dose data or linear
multiple-dose data.}
\usage{
pk.calc.mrt(auc, aumc)

pk.calc.mrt.iv(auc, aumc, duration.dose)
}
\arguments{
\item{auc}{the AUC from 0 to infinity or 0 to tau}

\item{aumc}{the AUMC from 0 to infinity or 0 to tau}

\item{duration.dose}{The duration of the dose (usually an infusion duration
for an IV infusion)}
}
\value{
the numeric value of the mean residence time
}
\description{
Calculate the mean residence time (MRT) for single-dose data or linear
multiple-dose data.
}
\details{
mrt is \code{aumc/auc - duration.dose/2} where \code{duration.dose = 0} for oral administration.
}
\section{Functions}{
\itemize{
\item \code{pk.calc.mrt.iv()}: MRT for an IV infusion

}}
\seealso{
\code{\link[=pk.calc.mrt.md]{pk.calc.mrt.md()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unit-support.R
\name{pknca_units_add_paren}
\alias{pknca_units_add_paren}
\title{Add parentheses to a unit value, if needed}
\usage{
pknca_units_add_paren(unit)
}
\arguments{
\item{unit}{The text of the unit}
}
\value{
The unit with parentheses around it, if needed
}
\description{
Add parentheses to a unit value, if needed
}
\keyword{Internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-PKNCAconc.R, R/class-PKNCAdose.R,
%   R/class-PKNCAresults.R, R/class-general.R
\name{getDataName.PKNCAconc}
\alias{getDataName.PKNCAconc}
\alias{getDataName.PKNCAdose}
\alias{getDataName.PKNCAresults}
\alias{getDataName}
\alias{getDataName.default}
\title{Get the name of the element containing the data for the current
object.}
\usage{
\method{getDataName}{PKNCAconc}(object)

\method{getDataName}{PKNCAdose}(object)

\method{getDataName}{PKNCAresults}(object)

getDataName(object)

\method{getDataName}{default}(object)
}
\arguments{
\item{object}{The object to get the data name from.}
}
\value{
A character scalar with the name of the data object (or \code{NULL} if the
method does not apply).
}
\description{
Get the name of the element containing the data for the current
object.
}
\section{Methods (by class)}{
\itemize{
\item \code{getDataName(default)}: If no data name exists, returns NULL.

}}
\seealso{
Other PKNCA object extractors: 
\code{\link{getDepVar}()},
\code{\link{getIndepVar}()}
}
\concept{PKNCA object extractors}
\keyword{Internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PKNCA.options.R
\name{PKNCA.options.describe}
\alias{PKNCA.options.describe}
\title{Describe a PKNCA.options option by name.}
\usage{
PKNCA.options.describe(name)
}
\arguments{
\item{name}{The option name requested.}
}
\value{
A character string of the description.
}
\description{
Describe a PKNCA.options option by name.
}
\seealso{
\code{\link[=PKNCA.options]{PKNCA.options()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/general.functions.R
\name{roundString}
\alias{roundString}
\title{Round a value to a defined number of digits printing out trailing zeros, if
applicable.}
\usage{
roundString(x, digits = 0, sci_range = Inf, sci_sep = "e", si_range)
}
\arguments{
\item{x}{The number to round}

\item{digits}{integer indicating the number of decimal places}

\item{sci_range}{See help for \code{\link[=signifString]{signifString()}} (and you likely want
to round with \code{signifString} if you want to use this argument)}

\item{sci_sep}{The separator to use for scientific notation strings
(typically this will be either "e" or "x10^" for computer- or
human-readable output).}

\item{si_range}{Deprecated, please use \code{sci_range}}
}
\value{
A string with the value
}
\description{
Round a value to a defined number of digits printing out trailing zeros, if
applicable.
}
\details{
Values that are not standard numbers like \code{Inf}, \code{NA}, and
\code{NaN} are returned as \code{"Inf"}, \code{"NA"}, and \code{NaN}.
}
\seealso{
\code{\link[=round]{round()}}, \code{\link[=signifString]{signifString()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AIC.list.R
\name{get.best.model}
\alias{get.best.model}
\title{Extract the best model from a list of models using the AIC.}
\usage{
get.best.model(object, ...)
}
\arguments{
\item{object}{the list of models}

\item{\dots}{Parameters passed to AIC.list}
}
\value{
The model which is assessed as best.  If more than one are equal,
the first is chosen.
}
\description{
Extract the best model from a list of models using the AIC.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pk.calc.all.R
\name{pk.nca.intervals}
\alias{pk.nca.intervals}
\title{Compute NCA for multiple intervals}
\usage{
pk.nca.intervals(
  data_conc,
  data_dose,
  data_intervals,
  sparse,
  options,
  impute,
  verbose = FALSE
)
}
\arguments{
\item{data_conc}{A data.frame or tibble with standardized column names as
output from \code{prepare_PKNCAconc()}}

\item{data_dose}{A data.frame or tibble with standardized column names as
output from \code{prepare_PKNCAdose()}}

\item{data_intervals}{A data.frame or tibble with standardized column names
as output from \code{prepare_PKNCAintervals()}}

\item{sparse}{Should only sparse calculations be performed (TRUE) or only
dense calculations (FALSE)?}

\item{options}{List of changes to the default PKNCA options (see
\code{PKNCA.options()})}

\item{impute}{The column name in \code{data_intervals} to use for imputation}

\item{verbose}{Indicate, by \code{message()}, the current state of calculation.}
}
\value{
A data.frame with all NCA results
}
\description{
Compute NCA for multiple intervals
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exclude.R
\name{exclude}
\alias{exclude}
\alias{exclude.default}
\title{Exclude data points or results from calculations or summarization.}
\usage{
exclude(object, reason, mask, FUN)

\method{exclude}{default}(object, reason, mask, FUN)
}
\arguments{
\item{object}{The object to exclude data from.}

\item{reason}{The reason to add as a reason for exclusion.}

\item{mask}{A logical vector or numeric index of values to exclude (see
details).}

\item{FUN}{A function to operate on the data (one group at a time) to select
reasons for exclusions (see details).}
}
\value{
The object with updated information in the exclude column. The
exclude column will contain the \code{reason} if \code{mask} or \code{FUN} indicate.  If a
previous reason for exclusion was given, then subsequent reasons for
exclusion will be added to the first with a semicolon space ("; ")
separator.
}
\description{
Exclude data points or results from calculations or summarization.
}
\details{
Only one of \code{mask} or \code{FUN} may be given.  If \code{FUN} is given, it
will be called with two arguments:  a data.frame (or similar object) that
consists of a single group of the data and the full object (e.g. the
PKNCAconc object), \code{FUN(current_group, object)}, and it must return a
logical vector equivalent to \code{mask} or a character vector with the reason
text given when data should be excluded or \code{NA_character_} when the data
should be included (for the current exclusion test).
}
\section{Methods (by class)}{
\itemize{
\item \code{exclude(default)}: The general case for data exclusion

}}
\examples{
myconc <- PKNCAconc(data.frame(subject=1,
                               time=0:6,
                               conc=c(1, 2, 3, 2, 1, 0.5, 0.25)),
                    conc~time|subject)
exclude(myconc,
        reason="Carryover",
        mask=c(TRUE, rep(FALSE, 6)))
}
\seealso{
Other Result exclusions: 
\code{\link{exclude_nca}}
}
\concept{Result exclusions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/provenance.R
\name{addProvenance}
\alias{addProvenance}
\title{Add a hash and associated information to enable checking object provenance.}
\usage{
addProvenance(object, replace = FALSE)
}
\arguments{
\item{object}{The object to add provenance}

\item{replace}{Replace provenance if the object already has a provenance
attribute.  (If the object already has provenance and \code{replace} is \code{FALSE},
then an error will be raised.)}
}
\value{
The object with provenance as an added item
}
\description{
Add a hash and associated information to enable checking object provenance.
}
\seealso{
\code{\link[=checkProvenance]{checkProvenance()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/general.functions.R
\name{signifString}
\alias{signifString}
\alias{signifString.data.frame}
\alias{signifString.default}
\title{Round a value to a defined number of significant digits printing out trailing
zeros, if applicable.}
\usage{
signifString(x, ...)

\method{signifString}{data.frame}(x, ...)

\method{signifString}{default}(x, digits = 6, sci_range = 6, sci_sep = "e", si_range, ...)
}
\arguments{
\item{x}{The number to round}

\item{...}{Arguments passed to methods.}

\item{digits}{integer indicating the number of significant digits}

\item{sci_range}{integer (or \code{Inf}) indicating when to switch to
scientific notation instead of floating point. Zero indicates always use
scientific; \code{Inf} indicates to never use scientific notation;
otherwise, scientific notation is used when \code{abs(log10(x)) > si_range}.}

\item{sci_sep}{The separator to use for scientific notation strings
(typically this will be either "e" or "x10^" for computer- or
human-readable output).}

\item{si_range}{Deprecated, please use \code{sci_range}}
}
\value{
A string with the value
}
\description{
Round a value to a defined number of significant digits printing out trailing
zeros, if applicable.
}
\details{
Values that are not standard numbers like \code{Inf}, \code{NA}, and
\code{NaN} are returned as \code{"Inf"}, \code{"NA"}, and \code{NaN}.
}
\seealso{
\code{\link[=signif]{signif()}}, \code{\link[=roundString]{roundString()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sparse.R
\name{sparse_auc_weight_linear}
\alias{sparse_auc_weight_linear}
\title{Calculate the weight for sparse AUC calculation with the linear-trapezoidal
rule}
\usage{
sparse_auc_weight_linear(sparse_pk)
}
\arguments{
\item{sparse_pk}{A sparse_pk object from \code{\link[=as_sparse_pk]{as_sparse_pk()}}}
}
\value{
A numeric vector of weights for sparse AUC calculations the same
length as \code{sparse_pk}
}
\description{
The weight is used as the \eqn{w_i}{w_i} parameter in \code{\link[=pk.calc.sparse_auc]{pk.calc.sparse_auc()}}
}
\details{
\deqn{w_i = \frac{\delta_{time,i-1,i} + \delta_{time,i,i+1}}{2}}{w_i = (d_time[i-1,i] + d_time[i,i+1])/2}
\deqn{\delta_{time,i,i+1} = t_{i+1} - t_i}{d_time = t_[i+1] - t_i, and zero if i < 1 or i > K}

Where:

\describe{
\item{\eqn{w_i}{w_i}}{is the weight at time i}
\item{\eqn{\delta_{time,i-1,i}}{d_time[i-1,i]} and \eqn{\delta_{time,i,i+1}}{d_time[i,i+1]}}{are the changes between time i-1 and i or i and i+1 (zero outside of the time range)}
\item{\eqn{t_i}{t_i}}{is the time at time i}
}
}
\seealso{
Other Sparse Methods: 
\code{\link{as_sparse_pk}()},
\code{\link{pk.calc.sparse_auc}()},
\code{\link{sparse_mean}()}
}
\concept{Sparse Methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/general.functions.R
\name{check.conversion}
\alias{check.conversion}
\title{Check that the conversion to a data type does not change the number
of NA values}
\usage{
check.conversion(x, FUN, ...)
}
\arguments{
\item{x}{the value to convert}

\item{FUN}{the function to use for conversion}

\item{\dots}{arguments passed to \code{FUN}}
}
\value{
\code{FUN(x, ...)} or an error if the set of NAs change.
}
\description{
Check that the conversion to a data type does not change the number
of NA values
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/impute.R
\name{PKNCA_impute_fun_list}
\alias{PKNCA_impute_fun_list}
\title{Separate out a vector of PKNCA imputation methods into a list of functions}
\usage{
PKNCA_impute_fun_list(x)
}
\arguments{
\item{x}{The character vector of PKNCA imputation method functions (without
the \code{PKNCA_impute_method_} part)}
}
\value{
A list of character vectors of functions to run.
}
\description{
An error will be raised if the functions are not found.
}
\details{
This function is not for use by users of PKNCA.
}
\keyword{Internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pk.calc.simple.R
\name{pk.calc.ceoi}
\alias{pk.calc.ceoi}
\title{Determine the concentration at the end of infusion}
\usage{
pk.calc.ceoi(conc, time, duration.dose = NA, check = TRUE)
}
\arguments{
\item{conc}{Measured concentrations}

\item{time}{Time of the measurement of the concentrations}

\item{duration.dose}{The duration for the dosing administration (typically
from IV infusion)}

\item{check}{Run \code{\link[=assert_conc_time]{assert_conc_time()}}?}
}
\value{
The concentration at the end of the infusion, \code{NA} if
\code{duration.dose} is \code{NA}, or \code{NA} if all \code{time != duration.dose}
}
\description{
Determine the concentration at the end of infusion
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-PKNCAconc.R
\name{PKNCAconc}
\alias{PKNCAconc}
\alias{PKNCAconc.default}
\alias{PKNCAconc.tbl_df}
\alias{PKNCAconc.data.frame}
\title{Create a PKNCAconc object}
\usage{
PKNCAconc(data, ...)

\method{PKNCAconc}{default}(data, ...)

\method{PKNCAconc}{tbl_df}(data, ...)

\method{PKNCAconc}{data.frame}(
  data,
  formula,
  subject,
  time.nominal,
  exclude = NULL,
  duration,
  volume,
  exclude_half.life,
  include_half.life,
  sparse = FALSE,
  ...,
  concu = NULL,
  amountu = NULL,
  timeu = NULL,
  concu_pref = NULL,
  amountu_pref = NULL,
  timeu_pref = NULL
)
}
\arguments{
\item{data}{A data frame with concentration (or amount for urine/feces),
time, and the groups defined in \code{formula}.}

\item{...}{Ignored.}

\item{formula}{The formula defining the \code{concentration~time|groups} or
\code{amount~time|groups} for urine/feces (In the remainder of the
documentation, "concentration" will be used to describe concentration or
amount.)  One special aspect of the \code{groups} part of the formula is that
the last group is typically assumed to be the \code{subject}; see the
documentation for the \code{subject} argument for exceptions to this assumption.}

\item{subject}{The column indicating the subject number.  If not provided,
this defaults to the beginning of the inner groups: For example with
\code{concentration~time|Study+Subject/Analyte}, the inner groups start with the
first grouping variable before a \code{/}, \code{Subject}.  If there is only one
grouping variable, it is assumed to be the subject (e.g.
\code{concentration~time|Subject}), and if there are multiple grouping variables
without a \code{/}, subject is assumed to be the last one.  For single-subject
data, it is assigned as \code{NULL}.}

\item{time.nominal}{(optional) The name of the nominal time column (if the
main time variable is actual time.  The \code{time.nominal} is not used during
calculations; it is available to assist with data summary and checking.}

\item{exclude}{(optional) The name of a column with concentrations to exclude
from calculations and summarization.  If given, the column should have
values of \code{NA} or \code{""} for concentrations to include and non-empty text for
concentrations to exclude.}

\item{duration}{(optional) The duration of collection as is typically used
for concentration measurements in urine or feces.}

\item{volume}{(optional) The volume (or mass) of collection as is typically
used for urine or feces measurements.}

\item{exclude_half.life, include_half.life}{A character scalar for the column
name in the dataset of the points to exclude from the half-life calculation
(still using normal curve-stripping selection rules for the other points)
or to include for the half-life (using specifically those points and
bypassing automatic curve-stripping point selection).  See the "Half-Life
Calculation" vignette for more details on the use of these arguments.}

\item{sparse}{Are the concentration-time data sparse PK (commonly used in
small nonclinical species or with terminal or difficult sampling) or dense
PK (commonly used in clinical studies or larger nonclinical species)?}

\item{concu, amountu, timeu}{Either unit values (e.g. "ng/mL") or column names
within the data where units are provided.}

\item{concu_pref, amountu_pref, timeu_pref}{Preferred units for reporting (not
column names)}
}
\value{
A PKNCAconc object that can be used for automated NCA.
}
\description{
Create a PKNCAconc object
}
\seealso{
Other PKNCA objects: 
\code{\link{PKNCAdata}()},
\code{\link{PKNCAdose}()},
\code{\link{PKNCAresults}()}
}
\concept{PKNCA objects}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-general.R
\name{getDepVar}
\alias{getDepVar}
\title{Get the dependent variable (left hand side of the formula) from a
PKNCA object.}
\usage{
getDepVar(x, ...)
}
\arguments{
\item{x}{The object to extract the formula from}

\item{\dots}{Unused}
}
\value{
The vector of the dependent variable from the object.
}
\description{
Get the dependent variable (left hand side of the formula) from a
PKNCA object.
}
\seealso{
Other PKNCA object extractors: 
\code{\link{getDataName.PKNCAconc}()},
\code{\link{getIndepVar}()}
}
\concept{PKNCA object extractors}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PKNCA.R
\docType{package}
\name{PKNCA}
\alias{PKNCA-package}
\alias{PKNCA}
\title{Compute noncompartmental pharmacokinetics}
\description{
Compute pharmacokinetic (PK) noncompartmental analysis (NCA)
parameters.
}
\details{
PKNCA has been cross-validated with both Phoenix WinNonlin(R) and Pumas
(click here for the
\href{http://www.humanpredictions.com/wp-content/uploads/2020/01/ACOP_2019_T102_NCA_performance_evaluation_Yingbo_revised.pdf}{cross-validation article})

A common workflow would load data from a file or database into a
data.frame then run the following code.
}
\examples{
\dontrun{
# Load concentration-time data into a data.frame called d.conc
# with columns named "conc", "time", and "subject".
my.conc <- PKNCAconc(d.conc, conc~time|subject)
# Load dose-time data into a data.frame called d.dose
# with columns named "dose", "time", and "subject".
my.dose <- PKNCAdose(d.dose, dose~time|subject)
# Combine the concentration-time and dose-time data into an object
# ready for calculations.
my.data <- PKNCAdata(my.conc, my.dose)
# Perform the calculations
my.results <- pk.nca(my.data)
# Look at summary results
summary(my.results)
# Look at a listing of results
as.data.frame(my.results)
}
}
\seealso{
Useful links:
\itemize{
  \item \url{https://humanpred.github.io/pknca/}
  \item \url{https://github.com/humanpred/pknca}
  \item \url{http://humanpred.github.io/pknca/}
  \item Report bugs at \url{https://github.com/humanpred/pknca/issues}
}

}
\author{
\strong{Maintainer}: Bill Denney \email{wdenney@humanpredictions.com} (\href{https://orcid.org/0000-0002-5759-428X}{ORCID})

Authors:
\itemize{
  \item Clare Buckeridge \email{clare.buckeridge@pfizer.com}
  \item Gerardo Jose Rodriguez \email{gerardo.jrac@gmail.com} (\href{https://orcid.org/0000-0003-1413-0060}{ORCID})
}

Other contributors:
\itemize{
  \item Sridhar Duvvuri [contributor]
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check.intervals.R
\name{get.parameter.deps}
\alias{get.parameter.deps}
\title{Get all columns that depend on a parameter}
\usage{
get.parameter.deps(x)
}
\arguments{
\item{x}{The parameter name (as a character string)}
}
\value{
A character vector of parameter names that depend on the parameter
\code{x}.  If none depend on \code{x}, then the result will be an empty vector.
}
\description{
Get all columns that depend on a parameter
}
\seealso{
Other Interval specifications: 
\code{\link{add.interval.col}()},
\code{\link{check.interval.deps}()},
\code{\link{check.interval.specification}()},
\code{\link{choose.auc.intervals}()},
\code{\link{get.interval.cols}()}
}
\concept{Interval specifications}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pk.calc.simple.R
\name{pk.calc.swing}
\alias{pk.calc.swing}
\title{Determine the PK swing}
\usage{
pk.calc.swing(cmax, cmin)
}
\arguments{
\item{cmax}{The maximum observed concentration}

\item{cmin}{The minimum observed concentration}
}
\value{
The swing above the minimum concentration.  If \code{cmin} is zero, then
the result is infinity.
}
\description{
Determine the PK swing
}
\details{
swing is \code{100*(cmax - cmin)/cmin}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tss.stepwise.linear.R
\name{pk.tss.stepwise.linear}
\alias{pk.tss.stepwise.linear}
\title{Compute the time to steady state using stepwise test of linear trend}
\usage{
pk.tss.stepwise.linear(
  ...,
  min.points = 3,
  level = 0.95,
  verbose = FALSE,
  check = TRUE
)
}
\arguments{
\item{\dots}{See \code{\link[=pk.tss.data.prep]{pk.tss.data.prep()}}}

\item{min.points}{The minimum number of points required for the fit}

\item{level}{The confidence level required for assessment of steady-state}

\item{verbose}{Describe models as they are run, show convergence of the model
(passed to the nlme function), and additional details while running.}

\item{check}{See \code{\link[=pk.tss.data.prep]{pk.tss.data.prep()}}}
}
\value{
A scalar float for the first time when steady-state is achieved or
\code{NA} if it is not observed.
}
\description{
A linear slope is fit through the data to find when it becomes
non-significant.  Note that this is less preferred than the
\code{pk.tss.monoexponential} due to the fact that with more time or more subjects
the performance of the test changes (see reference).
}
\details{
The model is fit with a different magnitude by treatment (as a factor, if
given) and a random slope by subject (if given).  A minimum of \code{min.points}
is required to fit the model.
}
\references{
Maganti L, Panebianco DL, Maes AL.  Evaluation of Methods for
Estimating Time to Steady State with Examples from Phase 1 Studies. AAPS
Journal 10(1):141-7. doi:10.1208/s12248-008-9014-y
}
\seealso{
Other Time to steady-state calculations: 
\code{\link{pk.tss}()},
\code{\link{pk.tss.monoexponential}()}
}
\concept{Time to steady-state calculations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/choose.intervals.R
\name{choose.auc.intervals}
\alias{choose.auc.intervals}
\title{Choose intervals to compute AUCs from time and dosing information}
\usage{
choose.auc.intervals(
  time.conc,
  time.dosing,
  options = list(),
  single.dose.aucs = NULL
)
}
\arguments{
\item{time.conc}{Time of concentration measurement}

\item{time.dosing}{Time of dosing}

\item{options}{List of changes to the default PKNCA options (see
\code{PKNCA.options()})}

\item{single.dose.aucs}{The AUC specification for single dosing.}
}
\value{
A data frame with columns for \code{start}, \code{end}, \code{auc.type}, and
\code{half.life}.  See \code{\link[=check.interval.specification]{check.interval.specification()}} for column definitions.
The data frame may have zero rows if no intervals could be found.
}
\description{
Intervals for AUC are selected by the following metrics:
\enumerate{
\item If only one dose is administered, use the
\code{PKNCA.options("single.dose.aucs")}
\item If more than one dose is administered, estimate the AUC
between any two doses that have PK taken at both of the
dosing times and at least one time  between the doses.
\item For the final dose of multiple doses, try to determine the
dosing interval (\eqn{\tau}) and estimate the AUC in that
interval if multiple samples are taken in the interval.
\item If there are samples \eqn{> \tau} after the last dose,
calculate the half life after the last dose.
}
}
\seealso{
\code{\link[=pk.calc.auc]{pk.calc.auc()}}, \code{\link[=pk.calc.aumc]{pk.calc.aumc()}}, \code{\link[=pk.calc.half.life]{pk.calc.half.life()}},
\code{\link[=PKNCA.options]{PKNCA.options()}}

Other Interval specifications: 
\code{\link{add.interval.col}()},
\code{\link{check.interval.deps}()},
\code{\link{check.interval.specification}()},
\code{\link{get.interval.cols}()},
\code{\link{get.parameter.deps}()}

Other Interval determination: 
\code{\link{find.tau}()}
}
\concept{Interval determination}
\concept{Interval specifications}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pk.calc.simple.R
\name{pk.calc.f}
\alias{pk.calc.f}
\title{Calculate the absolute (or relative) bioavailability}
\usage{
pk.calc.f(dose1, auc1, dose2, auc2)
}
\arguments{
\item{dose1}{The dose administered in route or method 1}

\item{auc1}{The AUC from 0 to infinity or 0 to tau administered in route or
method 1}

\item{dose2}{The dose administered in route or method 2}

\item{auc2}{The AUC from 0 to infinity or 0 to tau administered in route or
method 2}
}
\description{
Calculate the absolute (or relative) bioavailability
}
\details{
f is \code{(auc2/dose2)/(auc1/dose1)}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check.intervals.R
\name{check.interval.specification}
\alias{check.interval.specification}
\title{Check the formatting of a calculation interval specification data frame.}
\usage{
check.interval.specification(x)
}
\arguments{
\item{x}{The data frame specifying what to calculate during each time
interval}
}
\value{
x The potentially updated data frame with the interval calculation
specification.
}
\description{
Calculation interval specifications are data frames defining what
calculations will be required and summarized from all time intervals. Note:
parameters which are not requested may be calculated if it is required for
(or computed at the same time as) a requested parameter.
}
\details{
\code{start} and \code{end} time must always be given as columns, and the \code{start} must
be before the \code{end}.  Other columns define the parameters to be calculated
and the groupings to apply the intervals to.
}
\seealso{
The vignette "Selection of Calculation Intervals"

Other Interval specifications: 
\code{\link{add.interval.col}()},
\code{\link{check.interval.deps}()},
\code{\link{choose.auc.intervals}()},
\code{\link{get.interval.cols}()},
\code{\link{get.parameter.deps}()}
}
\concept{Interval specifications}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pk.calc.simple.R
\name{pk.calc.cl}
\alias{pk.calc.cl}
\title{Calculate the (observed oral) clearance}
\usage{
pk.calc.cl(dose, auc)
}
\arguments{
\item{dose}{the dose administered}

\item{auc}{The area under the concentration-time curve.}
}
\value{
the numeric value of the total (CL) or observed oral clearance
(CL/F)
}
\description{
Calculate the (observed oral) clearance
}
\details{
cl is \code{dose/auc}.

If \code{dose} is the same length as the other inputs, then the output
will be the same length as all of the inputs; the function assumes that you
are calculating for multiple intervals simultaneously.  If the inputs other
than \code{dose} are scalars and \code{dose} is a vector, then the function assumes
multiple doses were given in a single interval, and the sum of the \code{dose}s
will be used for the calculation.
}
\references{
Gabrielsson J, Weiner D. "Section 2.5.1 Derivation of clearance."
Pharmacokinetic & Pharmacodynamic Data Analysis: Concepts and Applications,
4th Edition.  Stockholm, Sweden: Swedish Pharmaceutical Press, 2000. 86-7.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tss.monoexponential.R
\name{pk.tss.monoexponential.population}
\alias{pk.tss.monoexponential.population}
\title{A helper function to estimate population and popind outputs for
monoexponential time to steady-state.}
\usage{
pk.tss.monoexponential.population(
  data,
  output = c("population", "popind"),
  verbose = FALSE
)
}
\arguments{
\item{data}{a data frame as prepared by \code{\link[=pk.tss.data.prep]{pk.tss.data.prep()}}.  It must
contain at least columns for \code{subject}, \code{time}, \code{conc}, and \code{tss.constant}.}

\item{output}{a character vector requesting the output types.}

\item{verbose}{Show verbose output.}
}
\value{
A data frame with either one row (if \code{population} output is
provided) or one row per subject (if \code{popind} is provided).  The columns
will be named \code{tss.monoexponential.population} and/or
\code{tss.monoexponential.popind}.
}
\description{
This function is not intended to be called directly.  Please use
\code{pk.tss.monoexponential}.
}
\details{
If no model converges, then the \code{tss.monoexponential.population} column will
be set to NA. If the best model does not include a random effect for subject
on Tss then the \code{tss.monoexponential.popind} column of the output will be set
to NA.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pk.calc.simple.R
\name{pk.calc.totdose}
\alias{pk.calc.totdose}
\title{Extract the dose used for calculations}
\usage{
pk.calc.totdose(dose)
}
\arguments{
\item{dose}{the dose administered}
}
\value{
The total dose for an interval
}
\description{
Extract the dose used for calculations
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/auc_integrate.R
\name{choose_interval_method}
\alias{choose_interval_method}
\title{Choose how to interpolate, extrapolate, or integrate data in each
concentration interval}
\usage{
choose_interval_method(conc, time, tlast, method, auc.type, options)
}
\arguments{
\item{conc}{Measured concentrations}

\item{time}{Time of the measurement of the concentrations}

\item{tlast}{Time of last concentration above the limit of quantification
(will be calculated, if not provided)}

\item{method}{The method for integration (one of 'lin up/log down',
'lin-log', or 'linear')}

\item{auc.type}{The type of AUC to compute.  Choices are 'AUCinf', 'AUClast',
and 'AUCall'.}

\item{options}{List of changes to the default PKNCA options (see
\code{PKNCA.options()})}
}
\value{
A character vector of methods for interpolation/extrapolation
methods that is the same length as \code{conc} which indicates how to
interpolate/integrate between each of the concentrations (all but the last
value in the vector) and how to extrapolate after \code{tlast} (the last item in
the vector).  Possible values in the vector are: 'zero', 'linear', 'log',
and 'extrap_log'
}
\description{
Choose how to interpolate, extrapolate, or integrate data in each
concentration interval
}
\keyword{Internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pk.calc.simple.R
\name{pk.calc.tlast}
\alias{pk.calc.tlast}
\alias{pk.calc.tfirst}
\title{Determine time of last observed concentration above the limit of
quantification.}
\usage{
pk.calc.tlast(conc, time, check = TRUE)

pk.calc.tfirst(conc, time, check = TRUE)
}
\arguments{
\item{conc}{Measured concentrations}

\item{time}{Time of the measurement of the concentrations}

\item{check}{Run \code{\link[=assert_conc_time]{assert_conc_time()}}?}
}
\value{
The time of the last observed concentration measurement
}
\description{
\code{NA} will be returned if all \code{conc} are \code{NA} or 0.
}
\section{Functions}{
\itemize{
\item \code{pk.calc.tfirst()}: Determine the first concentration above
the limit of quantification.

}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sparse.R
\name{as_sparse_pk}
\alias{as_sparse_pk}
\title{Generate a sparse_pk object}
\usage{
as_sparse_pk(conc, time, subject)
}
\arguments{
\item{conc}{Measured concentrations}

\item{time}{Time of the measurement of the concentrations}

\item{subject}{Subject identifiers (may be any class; may not be null)}
}
\value{
A sparse_pk object which is a list of lists.  The inner lists have
elements named: "time", The time of measurement; "conc", The concentration
measured; "subject", The subject identifiers.  The object will usually be
modified by future functions to add more named elements to the inner list.
}
\description{
Generate a sparse_pk object
}
\seealso{
Other Sparse Methods: 
\code{\link{pk.calc.sparse_auc}()},
\code{\link{sparse_auc_weight_linear}()},
\code{\link{sparse_mean}()}
}
\concept{Sparse Methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_various.R
\name{as_PKNCAconc}
\alias{as_PKNCAconc}
\alias{as_PKNCAdose}
\alias{as_PKNCAdata}
\alias{as_PKNCAresults}
\title{Convert an object into a PKNCAconc object}
\usage{
as_PKNCAconc(x, ...)

as_PKNCAdose(x, ...)

as_PKNCAdata(x, ...)

as_PKNCAresults(x, ...)
}
\arguments{
\item{x}{The object to convert}

\item{...}{Passed to subsequent methods}
}
\value{
A converted object
}
\description{
Convert an object into a PKNCAconc object
}
\section{Functions}{
\itemize{
\item \code{as_PKNCAdose()}: Convert an object into a PKNCAdose object

\item \code{as_PKNCAdata()}: Convert an object into a PKNCAdata object

\item \code{as_PKNCAresults()}: Convert an object into a PKNCAresults object

}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check.intervals.R
\name{check.interval.deps}
\alias{check.interval.deps}
\title{Take in a single row of an interval specification and return that row updated
with any additional calculations that must be done to fulfill all
dependencies.}
\usage{
check.interval.deps(x)
}
\arguments{
\item{x}{A data frame with one or more rows of the PKNCA interval}
}
\value{
The interval specification with additional calculations added where
requested outputs require them.
}
\description{
Take in a single row of an interval specification and return that row updated
with any additional calculations that must be done to fulfill all
dependencies.
}
\seealso{
Other Interval specifications: 
\code{\link{add.interval.col}()},
\code{\link{check.interval.specification}()},
\code{\link{choose.auc.intervals}()},
\code{\link{get.interval.cols}()},
\code{\link{get.parameter.deps}()}
}
\concept{Interval specifications}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pk.calc.simple.R
\name{pk.calc.deg.fluc}
\alias{pk.calc.deg.fluc}
\title{Determine the degree of fluctuation}
\usage{
pk.calc.deg.fluc(cmax, cmin, cav)
}
\arguments{
\item{cmax}{The maximum observed concentration}

\item{cmin}{The minimum observed concentration}

\item{cav}{The average concentration in the interval}
}
\value{
The degree of fluctuation around the average concentration.
}
\description{
Determine the degree of fluctuation
}
\details{
deg.fluc is \code{100*(cmax - cmin)/cav}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-PKNCAdose.R
\name{PKNCAdose}
\alias{PKNCAdose}
\alias{PKNCAdose.default}
\alias{PKNCAdose.tbl_df}
\alias{PKNCAdose.data.frame}
\title{Create a PKNCAdose object}
\usage{
PKNCAdose(data, ...)

\method{PKNCAdose}{default}(data, ...)

\method{PKNCAdose}{tbl_df}(data, ...)

\method{PKNCAdose}{data.frame}(
  data,
  formula,
  route,
  rate,
  duration,
  time.nominal,
  exclude = NULL,
  ...,
  doseu = NULL,
  doseu_pref = NULL
)
}
\arguments{
\item{data}{A data frame with time and the groups defined in \code{formula}.}

\item{...}{Ignored.}

\item{formula}{The formula defining the \code{dose.amount~time|groups} where
\code{time} is the time of the dosing and \code{dose.amount} is the amount
administered at that time (see Details).}

\item{route}{Define the route of administration.  The value may be either a
column name from the \code{data} (checked first) or a character string of either
\code{"extravascular"} or \code{"intravascular"} (checked second).  If given as a
column name, then every value of the column must be either
\code{"extravascular"} or \code{"intravascular"}.}

\item{rate, duration}{(optional) for \code{"intravascular"} dosing, the rate or
duration of dosing.  If given as a character string, it is the name of a
column from the \code{data}, and if given as a number, it is the value for all
doses.  Only one may be given, and if neither is given, then the dose is
assumed to be a bolus (\code{duration=0}).  If \code{rate} is given, then the dose
amount must be given (the left hand side of the \code{formula}).}

\item{time.nominal}{(optional) The name of the nominal time column (if the
main time variable is actual time.  The \code{time.nominal} is not used during
calculations; it is available to assist with data summary and checking.}

\item{exclude}{(optional) The name of a column with concentrations to exclude
from calculations and summarization.  If given, the column should have
values of \code{NA} or \code{""} for concentrations to include and non-empty text for
concentrations to exclude.}

\item{doseu}{Either unit values (e.g. "mg") or column names within the data
where units are provided.}

\item{doseu_pref}{Preferred units for reporting (not column names)}
}
\value{
A PKNCAconc object that can be used for automated NCA.
}
\description{
Create a PKNCAdose object
}
\details{
The \code{formula} for a \code{PKNCAdose} object can be
given three ways: one-sided (missing left side), one-sided (missing
right side), or two-sided.  Each of the three ways can be given
with or without groups.  When given one-sided missing the left
side, the left side can either be omitted or can be given as a
period (\code{.}): \code{~time|treatment+subject} and
\code{.~time|treatment+subject} are identical, and dose-related NCA
parameters will all be reported as not calculable (for example,
clearance).  When given one-sided missing the right side, the right
side must be specified as a period (\code{.}):
\code{dose~.|treatment+subject}, and only a single row may be given
per group.  When the right side is missing, PKNCA assumes that the
same dose is given in every interval.  When given as a two-sided
formula
}
\seealso{
Other PKNCA objects: 
\code{\link{PKNCAconc}()},
\code{\link{PKNCAdata}()},
\code{\link{PKNCAresults}()}
}
\concept{PKNCA objects}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sparse.R
\name{sparse_pk_attribute}
\alias{sparse_pk_attribute}
\title{Set or get a sparse_pk object attribute}
\usage{
sparse_pk_attribute(sparse_pk, ...)
}
\arguments{
\item{sparse_pk}{A sparse_pk object from \code{\link[=as_sparse_pk]{as_sparse_pk()}}}

\item{...}{Either a character string (to get that value) or a named vector
the same length as \code{sparse_pk} to set the value.}
}
\value{
Either the attribute value or an updated \code{sparse_pk} object
}
\description{
Set or get a sparse_pk object attribute
}
\keyword{Internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dplyr.R
\name{mutate.PKNCAresults}
\alias{mutate.PKNCAresults}
\alias{mutate.PKNCAconc}
\alias{mutate.PKNCAdose}
\title{dplyr mutate-based modification for PKNCA}
\usage{
\method{mutate}{PKNCAresults}(.data, ...)

\method{mutate}{PKNCAconc}(.data, ...)

\method{mutate}{PKNCAdose}(.data, ...)
}
\arguments{
\item{.data}{A data frame, data frame extension (e.g. a tibble), or a
lazy data frame (e.g. from dbplyr or dtplyr). See \emph{Methods}, below, for
more details.}

\item{...}{<\code{\link[rlang:args_data_masking]{data-masking}}> Name-value pairs.
The name gives the name of the column in the output.

The value can be:
\itemize{
\item A vector of length 1, which will be recycled to the correct length.
\item A vector the same length as the current group (or the whole data frame
if ungrouped).
\item \code{NULL}, to remove the column.
\item A data frame or tibble, to create multiple columns in the output.
}}
}
\description{
dplyr mutate-based modification for PKNCA
}
\seealso{
Other dplyr verbs: 
\code{\link{filter.PKNCAresults}()},
\code{\link{group_by.PKNCAresults}()},
\code{\link{inner_join.PKNCAresults}()}
}
\concept{dplyr verbs}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-summary_PKNCAresults.R
\name{summary.PKNCAresults}
\alias{summary.PKNCAresults}
\title{Summarize PKNCA results}
\usage{
\method{summary}{PKNCAresults}(
  object,
  ...,
  drop_group = object$data$conc$columns$subject,
  drop_param = character(),
  summarize_n = NA,
  not_requested = ".",
  not_calculated = "NC",
  drop.group = deprecated(),
  summarize.n.per.group = deprecated(),
  not.requested.string = deprecated(),
  not.calculated.string = deprecated(),
  pretty_names = NULL
)
}
\arguments{
\item{object}{The results to summarize}

\item{...}{Ignored.}

\item{drop_group}{Which group(s) should be dropped from the formula?}

\item{drop_param}{Which parameters should be excluded from the summary?}

\item{summarize_n}{Should a column for \code{N} be added (\code{TRUE} or \code{FALSE})?
\code{NA} means to automatically detect adding \code{N} if the data has a subject
column indicated.  Note that \code{N} is maximum number of parameter results for
any parameter; if no parameters are requested for a group, then \code{N} will be
\code{NA}.}

\item{not_requested}{A character string to use when a parameter summary was
not requested for a parameter within an interval.}

\item{not_calculated}{A character string to use when a parameter summary was
requested, but the point estimate AND spread calculations (if applicable)
returned \code{NA}.}

\item{drop.group, summarize.n.per.group, not.requested.string, not.calculated.string}{Deprecated use \code{drop_group}, \code{not_requested}, \code{not_calculated}, or
\code{summarize_n}, instead}

\item{pretty_names}{Should pretty names (easier to understand in a report) be
used?  \code{TRUE} is yes, \code{FALSE} is no, and \code{NULL} is yes if units are used
and no if units are not used.}
}
\value{
A data frame of NCA parameter results summarized according to the
summarization settings.
}
\description{
Summarize PKNCA results
}
\details{
Excluded results will not be included in the summary.
}
\examples{
conc_obj <- PKNCAconc(as.data.frame(datasets::Theoph), conc ~ Time | Subject)
d_dose <-
  unique(datasets::Theoph[
    datasets::Theoph$Time == 0,
    c("Dose", "Time", "Subject")
  ])
dose_obj <- PKNCAdose(d_dose, Dose ~ Time | Subject)
data_obj_automatic <- PKNCAdata(conc_obj, dose_obj)
results_obj_automatic <- pk.nca(data_obj_automatic)
# To get standard results run summary
summary(results_obj_automatic)
# To enable numeric conversion and extraction, do not give a spread function
# and subsequently run as.numeric on the result columns.
PKNCA.set.summary(
  name = c("auclast", "cmax", "half.life", "aucinf.obs"),
  point = business.geomean,
  description = "geometric mean"
)
PKNCA.set.summary(
  name = c("tmax"),
  point = business.median,
  description = "median"
)
summary(results_obj_automatic, not_requested = "NA")
}
\seealso{
\code{\link[=PKNCA.set.summary]{PKNCA.set.summary()}}, \code{\link[=print.summary_PKNCAresults]{print.summary_PKNCAresults()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pk.calc.simple.R
\name{pk.calc.cmax}
\alias{pk.calc.cmax}
\alias{pk.calc.cmin}
\title{Determine maximum observed PK concentration}
\usage{
pk.calc.cmax(conc, check = TRUE)

pk.calc.cmin(conc, check = TRUE)
}
\arguments{
\item{conc}{Measured concentrations}

\item{check}{Run \code{\link[=assert_conc]{assert_conc()}}?}
}
\value{
a number for the maximum concentration or NA if all concentrations
are missing
}
\description{
Determine maximum observed PK concentration
}
\section{Functions}{
\itemize{
\item \code{pk.calc.cmin()}: Determine the minimum observed PK
concentration

}}
\examples{
conc_data <- Theoph[Theoph$Subject == 1,]
pk.calc.cmin(conc_data$conc)
}
\seealso{
Other NCA parameters for concentrations during the intervals: 
\code{\link{pk.calc.clast.obs}()},
\code{\link{pk.calc.count_conc}()},
\code{\link{pk.calc.cstart}()},
\code{\link{pk.calc.ctrough}()}

Other NCA parameters for concentrations during the intervals: 
\code{\link{pk.calc.clast.obs}()},
\code{\link{pk.calc.count_conc}()},
\code{\link{pk.calc.cstart}()},
\code{\link{pk.calc.ctrough}()}
}
\concept{NCA parameters for concentrations during the intervals}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sparse.R
\name{pk.calc.sparse_auc}
\alias{pk.calc.sparse_auc}
\alias{pk.calc.sparse_auclast}
\title{Calculate AUC and related parameters using sparse NCA methods}
\usage{
pk.calc.sparse_auc(
  conc,
  time,
  subject,
  method = NULL,
  auc.type = "AUClast",
  ...,
  options = list()
)

pk.calc.sparse_auclast(conc, time, subject, ..., options = list())
}
\arguments{
\item{conc}{Measured concentrations}

\item{time}{Time of the measurement of the concentrations}

\item{subject}{Subject identifiers (may be any class; may not be null)}

\item{method}{The method for integration (one of 'lin up/log down',
'lin-log', or 'linear')}

\item{auc.type}{The type of AUC to compute.  Choices are 'AUCinf', 'AUClast',
and 'AUCall'.}

\item{...}{For functions other than \code{pk.calc.auxc}, these values are passed
to \code{pk.calc.auxc}}

\item{options}{List of changes to the default PKNCA options (see
\code{PKNCA.options()})}
}
\description{
The AUC is calculated as:
}
\details{
\deqn{AUC=\sum\limits_{i} w_i \bar{C}_i}{AUC = sum(w_i * Cbar_i)}

Where:

\describe{
\item{\eqn{AUC}{AUC}}{is the estimated area under the concentration-time curve}
\item{\eqn{w_i}{w_i}}{is the weight applied to the concentration at time i (related to the time which it affects, see \code{\link[=sparse_auc_weight_linear]{sparse_auc_weight_linear()}})}
\item{\eqn{\bar{C}_i}{Cbar_i}}{is the average concentration at time i}
}
}
\section{Functions}{
\itemize{
\item \code{pk.calc.sparse_auclast()}: Compute the AUClast for sparse PK

}}
\seealso{
Other Sparse Methods: 
\code{\link{as_sparse_pk}()},
\code{\link{sparse_auc_weight_linear}()},
\code{\link{sparse_mean}()}
}
\concept{Sparse Methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unit-support.R
\name{pknca_find_units_param}
\alias{pknca_find_units_param}
\title{Find NCA parameters with a given unit type}
\usage{
pknca_find_units_param(unit_type)
}
\arguments{
\item{unit_type}{The type of unit as assigned with \code{add.interval.col}}
}
\value{
A character vector of parameters with a given unit type
}
\description{
Find NCA parameters with a given unit type
}
\keyword{Internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tss.R
\name{pk.tss.data.prep}
\alias{pk.tss.data.prep}
\title{Clean up the time to steady-state parameters and return a data frame for use
by the tss calculators.}
\usage{
pk.tss.data.prep(
  conc,
  time,
  subject,
  treatment,
  subject.dosing,
  time.dosing,
  options = list(),
  conc.blq = NULL,
  conc.na = NULL,
  check = TRUE,
  ...
)
}
\arguments{
\item{conc}{Measured concentrations}

\item{time}{Time of the measurement of the concentrations}

\item{subject}{Subject identifiers (used as a random effect in the model)}

\item{treatment}{Treatment description (if missing, all subjects are assumed
to be on the same treatment)}

\item{subject.dosing}{Subject number for dosing}

\item{time.dosing}{Time of dosing}

\item{options}{List of changes to the default PKNCA options (see
\code{PKNCA.options()})}

\item{conc.blq}{See \code{\link[=clean.conc.blq]{clean.conc.blq()}}}

\item{conc.na}{See \code{\link[=clean.conc.na]{clean.conc.na()}}}

\item{check}{Run \code{\link[=assert_conc_time]{assert_conc_time()}}?}

\item{\dots}{Discarded inputs to allow generic calls between tss methods.}
}
\value{
a data frame with columns for \code{conc}entration, \code{time}, \code{subject},
and \code{treatment}.
}
\description{
Clean up the time to steady-state parameters and return a data frame for use
by the tss calculators.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/001-add.interval.col.R
\name{get.interval.cols}
\alias{get.interval.cols}
\title{Get the columns that can be used in an interval specification}
\usage{
get.interval.cols()
}
\value{
A list with named elements for each parameter.  Each list element
contains the parameter definition.
}
\description{
Get the columns that can be used in an interval specification
}
\examples{
get.interval.cols()
}
\seealso{
\code{\link[=check.interval.specification]{check.interval.specification()}} and the vignette "Selection of
Calculation Intervals"

Other Interval specifications: 
\code{\link{add.interval.col}()},
\code{\link{check.interval.deps}()},
\code{\link{check.interval.specification}()},
\code{\link{choose.auc.intervals}()},
\code{\link{get.parameter.deps}()}
}
\concept{Interval specifications}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sparse.R
\name{sparse_to_dense_pk}
\alias{sparse_to_dense_pk}
\title{Extract the mean concentration-time profile as a data.frame}
\usage{
sparse_to_dense_pk(sparse_pk)
}
\arguments{
\item{sparse_pk}{A sparse_pk object from \code{\link[=as_sparse_pk]{as_sparse_pk()}}}
}
\value{
A data.frame with names of "conc" and "time"
}
\description{
Extract the mean concentration-time profile as a data.frame
}
\keyword{Internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AIC.list.R
\name{get.first.model}
\alias{get.first.model}
\title{Get the first model from a list of models}
\usage{
get.first.model(object)
}
\arguments{
\item{object}{the list of (lists of, ...) models}
}
\value{
The first item in the \code{object} that is not a list or \code{NA}.  If \code{NA}
is passed in or the list (of lists) is all \code{NA}, then \code{NA} is returned.
}
\description{
Get the first model from a list of models
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/auc_integrate.R
\name{interp_extrap_conc_method}
\alias{interp_extrap_conc_method}
\alias{interpolate_conc_linear}
\alias{interpolate_conc_log}
\alias{extrapolate_conc_lambdaz}
\title{Interpolate or extrapolate concentrations using the provided method}
\usage{
interpolate_conc_linear(conc_1, conc_2, time_1, time_2, time_out)

interpolate_conc_log(conc_1, conc_2, time_1, time_2, time_out)

extrapolate_conc_lambdaz(clast, lambda.z, tlast, time_out)
}
\arguments{
\item{conc_1, conc_2}{The concentration at time1 and time2}

\item{time_1, time_2}{The time value associated with conc1 and conc2}

\item{time_out}{Time when interpolation is requested}

\item{clast}{The concentration at the last time above the lower LOQ}

\item{lambda.z}{The elimination rate (in units of inverse time) for
extrapolation}

\item{tlast}{The time of the last concentration above the lower limit of
quantification (LOQ)}
}
\value{
The interpolated or extrapolated value using the correct method
}
\description{
Interpolate or extrapolate concentrations using the provided method
}
\keyword{Internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PKNCA.options.R
\name{PKNCA.options}
\alias{PKNCA.options}
\title{Set default options for PKNCA functions}
\usage{
PKNCA.options(..., default = FALSE, check = FALSE, name, value)
}
\arguments{
\item{\dots}{options to set or get the value for}

\item{default}{(re)sets all default options}

\item{check}{check a single option given, but do not set it (for validation
of the values when used in another function)}

\item{name}{An option name to use with the \code{value}.}

\item{value}{An option value (paired with the \code{name}) to set or check (if
\code{NULL}, ).}
}
\value{
If...
\describe{
\item{no arguments are given}{returns the current options.}
\item{a value is set (including the defaults)}{returns \code{NULL}}
\item{a single value is requested}{the current value of that option is returned as a scalar}
\item{multiple values are requested}{the current values of those options are returned as a list}
}
}
\description{
This function will set the default PKNCA options.  If given no inputs, it
will provide the current option set.  If given name/value pairs, it will set
the option (as in the \code{\link[=options]{options()}} function).  If given a name, it will return
the value for the parameter.  If given the \code{default} option as true, it will
provide the default options.
}
\details{
Options are either for calculation or summary functions. Calculation options
are required for a calculation function to report a result (otherwise the
reported value will be \code{NA}). Summary options are used during summarization
and are used for assessing what values are included in the summary.

See the vignette 'Options for Controlling PKNCA' for a current list of
options (\code{vignette("Options-for-Controlling-PKNCA", package="PKNCA")}).
}
\examples{

PKNCA.options()
PKNCA.options(default=TRUE)
PKNCA.options("auc.method")
PKNCA.options(name="auc.method")
PKNCA.options(auc.method="lin up/log down", min.hl.points=3)
}
\seealso{
\code{\link[=PKNCA.options.describe]{PKNCA.options.describe()}}

Other PKNCA calculation and summary settings: 
\code{\link{PKNCA.choose.option}()},
\code{\link{PKNCA.set.summary}()}
}
\concept{PKNCA calculation and summary settings}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pk.calc.simple.R
\name{adj.r.squared}
\alias{adj.r.squared}
\title{Calculate the adjusted r-squared value}
\usage{
adj.r.squared(r.sq, n)
}
\arguments{
\item{r.sq}{The r-squared value}

\item{n}{The number of points}
}
\value{
The numeric adjusted r-squared value
}
\description{
Calculate the adjusted r-squared value
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assertions.R
\name{assert_unit_col}
\alias{assert_unit_col}
\alias{assert_unit_value}
\alias{assert_unit}
\title{Assert that a value may either be a column name in the data (first) or a
single unit value (second)}
\usage{
assert_unit_col(unit, data)

assert_unit_value(unit)

assert_unit(unit, data)
}
\arguments{
\item{unit}{The column name or unit value}

\item{data}{The data.frame that contains a column named \code{unit}}
}
\value{
\code{unit} with an attribute of "unit_type" that is either "column" or
"value", or \code{NULL} if \code{is.null(unit)}
}
\description{
Assert that a value may either be a column name in the data (first) or a
single unit value (second)
}
\section{Functions}{
\itemize{
\item \code{assert_unit_col()}: Assert that a column name contains a character string
(that could be a unit specification)

\item \code{assert_unit_value()}: Assert that a value may be a single unit

The function does not verify that it is a real unit like "ng/mL" only that it
is a single character string.

}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parse_formula_to_cols.R
\name{findOperator}
\alias{findOperator}
\title{Find the first occurrence of an operator in a formula and return
the left, right, or both sides of the operator.}
\usage{
findOperator(x, op, side)
}
\arguments{
\item{x}{The formula to parse}

\item{op}{The operator to search for (e.g. \code{+}, \code{-}, \code{*}, \code{/}, ...)}

\item{side}{Which side of the operator would you like to see: 'left',
'right', or 'both'.}
}
\value{
The side of the operator requested, NA if requesting the left side
of a unary operator, and NULL if the operator is not found.
}
\description{
Find the first occurrence of an operator in a formula and return
the left, right, or both sides of the operator.
}
\seealso{
Other Formula parsing: 
\code{\link{parse_formula_to_cols}()}
}
\concept{Formula parsing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cleaners.R
\name{clean.conc.na}
\alias{clean.conc.na}
\title{Handle NA values in the concentration measurements as requested by the user.}
\usage{
clean.conc.na(conc, time, ..., options = list(), conc.na = NULL, check = TRUE)
}
\arguments{
\item{conc}{Measured concentrations}

\item{time}{Time of the measurement of the concentrations}

\item{\dots}{Additional items to add to the data frame}

\item{options}{List of changes to the default PKNCA options (see
\code{PKNCA.options()})}

\item{conc.na}{How to handle NA concentrations?  Either 'drop' or a number to
impute.}

\item{check}{Run \code{\link[=assert_conc_time]{assert_conc_time()}}?}
}
\value{
The concentration and time measurements (data frame) filtered
and cleaned as requested relative to NA in the concentration.
}
\description{
NA concentrations (and their associated times) will be removed then the BLQ
values in the middle
}
\seealso{
Other Data cleaners: 
\code{\link{clean.conc.blq}()}
}
\concept{Data cleaners}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pk.calc.simple.R
\name{pk.calc.vz}
\alias{pk.calc.vz}
\title{Calculate the terminal volume of distribution (Vz)}
\usage{
pk.calc.vz(cl, lambda.z)
}
\arguments{
\item{cl}{the clearance (or apparent observed clearance)}

\item{lambda.z}{The elimination rate (in units of inverse time) for
extrapolation}
}
\description{
Calculate the terminal volume of distribution (Vz)
}
\details{
vz is \code{cl/lambda.z}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-PKNCAdose.R
\name{setRoute}
\alias{setRoute}
\alias{setRoute.PKNCAdose}
\title{Set the dosing route}
\usage{
setRoute(object, ...)

\method{setRoute}{PKNCAdose}(object, route, ...)
}
\arguments{
\item{object}{A PKNCAdose object}

\item{...}{Arguments passed to another setRoute function}

\item{route}{A character string indicating one of the following:  the column
from the data which indicates the route of administration, a scalar
indicating the route of administration for all subjects, or a vector
indicating the route of administration for each dose in the dataset.}
}
\value{
The object with an updated route
}
\description{
Set the dosing route
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pk.calc.all.R
\name{pk_nca_result_to_df}
\alias{pk_nca_result_to_df}
\title{Convert the grouping info and list of results for each group into a results
data.frame}
\usage{
pk_nca_result_to_df(group_info, result)
}
\arguments{
\item{group_info}{A data.frame of grouping columns}

\item{result}{A list of data.frames with the results from NCA parameter
calculations}
}
\value{
A data.frame with group_info and result combined, warnings filtered
out, and results unnested.
}
\description{
Convert the grouping info and list of results for each group into a results
data.frame
}
\keyword{Internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-PKNCAconc.R, R/class-PKNCAdose.R
\name{formula.PKNCAconc}
\alias{formula.PKNCAconc}
\alias{formula.PKNCAdose}
\title{Extract the formula from a PKNCAconc object.}
\usage{
\method{formula}{PKNCAconc}(x, ...)

\method{formula}{PKNCAdose}(x, ...)
}
\arguments{
\item{x}{The object to extract the formula from.}

\item{\dots}{Unused}
}
\value{
A formula object
}
\description{
Extract the formula from a PKNCAconc object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-summary_PKNCAresults.R
\name{roundingSummarize}
\alias{roundingSummarize}
\title{During the summarization of PKNCAresults, do the rounding of values
based on the instructions given.}
\usage{
roundingSummarize(x, name)
}
\arguments{
\item{x}{The values to summarize}

\item{name}{The NCA parameter name (matching a parameter name in
\code{\link[=PKNCA.set.summary]{PKNCA.set.summary()}})}
}
\value{
A string of the rounded value
}
\description{
During the summarization of PKNCAresults, do the rounding of values
based on the instructions given.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interpolate.conc.R
\name{interp.extrap.conc}
\alias{interp.extrap.conc}
\alias{interpolate.conc}
\alias{extrapolate.conc}
\alias{interp.extrap.conc.dose}
\title{Interpolate concentrations between measurements or extrapolate concentrations
after the last measurement.}
\usage{
interp.extrap.conc(
  conc,
  time,
  time.out,
  lambda.z = NA,
  clast = pk.calc.clast.obs(conc, time),
  options = list(),
  method = NULL,
  auc.type = "AUCinf",
  interp.method,
  extrap.method,
  ...,
  conc.blq = NULL,
  conc.na = NULL,
  check = TRUE
)

interpolate.conc(
  conc,
  time,
  time.out,
  options = list(),
  method = NULL,
  interp.method,
  conc.blq = NULL,
  conc.na = NULL,
  conc.origin = 0,
  ...,
  check = TRUE
)

extrapolate.conc(
  conc,
  time,
  time.out,
  lambda.z = NA,
  clast = pk.calc.clast.obs(conc, time),
  auc.type = "AUCinf",
  extrap.method,
  options = list(),
  conc.na = NULL,
  conc.blq = NULL,
  ...,
  check = TRUE
)

interp.extrap.conc.dose(
  conc,
  time,
  time.dose,
  route.dose = "extravascular",
  duration.dose = NA,
  time.out,
  out.after = FALSE,
  options = list(),
  conc.blq = NULL,
  conc.na = NULL,
  ...,
  check = TRUE
)
}
\arguments{
\item{conc}{Measured concentrations}

\item{time}{Time of the measurement of the concentrations}

\item{time.out}{Time when interpolation is requested (vector for
\code{interp.extrap.conc()}, scalar otherwise)}

\item{lambda.z}{The elimination rate (in units of inverse time) for
extrapolation}

\item{clast}{The last observed concentration above the limit of
quantification.  If not given, \code{clast} is calculated from
\code{\link[=pk.calc.clast.obs]{pk.calc.clast.obs()}}}

\item{options}{List of changes to the default PKNCA options (see
\code{PKNCA.options()})}

\item{method}{The method for integration (one of 'lin up/log down',
'lin-log', or 'linear')}

\item{auc.type}{The type of AUC to compute.  Choices are 'AUCinf', 'AUClast',
and 'AUCall'.}

\item{interp.method, extrap.method}{deprecated in favor of method and auc.type}

\item{...}{Additional arguments passed to \code{interpolate.conc()} or
\code{extrapolate.conc()}.}

\item{conc.blq}{How to handle BLQ values. (See \code{\link[=clean.conc.blq]{clean.conc.blq()}} for usage
instructions.)}

\item{conc.na}{How to handle NA concentrations.  (See \code{\link[=clean.conc.na]{clean.conc.na()}})}

\item{check}{Run \code{\link[=assert_conc_time]{assert_conc_time()}}, \code{\link[=clean.conc.blq]{clean.conc.blq()}}, and
\code{\link[=clean.conc.na]{clean.conc.na()}}?}

\item{conc.origin}{The concentration before the first measurement.
\code{conc.origin} is typically used to set predose values to zero (default),
set a predose concentration for endogenous compounds, or set predose
concentrations to \code{NA} if otherwise unknown.}

\item{time.dose}{Time of the dose}

\item{route.dose}{What is the route of administration ("intravascular" or
"extravascular").  See the details for how this parameter is used.}

\item{duration.dose}{What is the duration of administration? See the details
for how this parameter is used.}

\item{out.after}{Should interpolation occur from the data before (\code{FALSE}) or
after (\code{TRUE}) the interpolated point?  See the details for how this
parameter is used.  It only has a meaningful effect at the instant of an IV
bolus dose.}
}
\value{
The interpolated or extrapolated concentration value as a scalar
double (or vector for \code{interp.extrap.conc()}).
}
\description{
\code{interpolate.conc()} and \code{extrapolate.conc()} returns an interpolated (or
extrapolated) concentration. \code{interp.extrap.conc()} will choose whether
interpolation or extrapolation is required and will also operate on many
concentrations.  These will typically be used to estimate the concentration
between two measured concentrations or after the last measured concentration.
Of note, these functions will not extrapolate prior to the first point.
}
\details{
An \code{NA} value for the \code{lambda.z} parameter will prevent extrapolation.

\describe{
\item{extrap.method}{
\describe{
\item{'AUCinf'}{Use lambda.z to extrapolate beyond the last point with the half-life.}
\item{'AUCall'}{If the last point is above the limit of quantification or missing, this is identical to 'AUCinf'. If the last point is below the limit of quantification, then linear interpolation between the Clast and the next BLQ is used for that  interval and all additional points are extrapolated as 0.}
\item{'AUClast'}{Extrapolates all points after the last above the  limit of quantification as 0.}
}
}
}

\code{duration.dose} and \code{direction.out} are ignored if \code{route.dose == "extravascular"}.  \code{direction.out} is ignored if \code{duration.dose > 0}.

\code{route.dose} and \code{duration.dose} affect how interpolation/extrapolation of
the concentration occurs at the time of dosing.  If \code{route.dose == "intravascular"} and \code{duration.dose == 0} then extrapolation occurs for an IV
bolus using \code{\link[=pk.calc.c0]{pk.calc.c0()}} with the data after dosing.  Otherwise (either
\code{route.dose == "extravascular"} or \code{duration.dose > 0}), extrapolation occurs
using the concentrations before dosing and estimating the half-life (or more
precisely, estimating \code{lambda.z}).  Finally, \code{direction.out} can change the
direction of interpolation in cases with \code{route.dose == "intravascular"} and
\code{duration.dose == 0}.  When \code{direction.out == "before"} interpolation occurs
only with data before the dose (as is the case for \code{route.dose == "extravascular"}), but if \code{direction.out == "after"} interpolation occurs
from the data after dosing.
}
\section{Functions}{
\itemize{
\item \code{interpolate.conc()}: Interpolate concentrations through Tlast (inclusive)

\item \code{extrapolate.conc()}: Extrapolate concentrations after Tlast

\item \code{interp.extrap.conc.dose()}: Interpolate and extrapolate concentrations
without interpolating or extrapolating beyond doses.

}}
\seealso{
\code{\link[=pk.calc.clast.obs]{pk.calc.clast.obs()}}, \code{\link[=pk.calc.half.life]{pk.calc.half.life()}}, \code{\link[=pk.calc.c0]{pk.calc.c0()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-PKNCAresults.R
\name{PKNCAresults}
\alias{PKNCAresults}
\title{Generate a PKNCAresults object}
\usage{
PKNCAresults(result, data, exclude = NULL)
}
\arguments{
\item{result}{a data frame with NCA calculation results and groups. Each row
is one interval and each column is a group name or the name of an NCA
parameter.}

\item{data}{The PKNCAdata used to generate the result}

\item{exclude}{(optional) The name of a column with concentrations to exclude
from calculations and summarization.  If given, the column should have
values of \code{NA} or \code{""} for concentrations to include and non-empty text for
concentrations to exclude.}
}
\value{
A PKNCAresults object with each of the above within.
}
\description{
This function should not be run directly.  The object is created for
summarization.
}
\seealso{
Other PKNCA objects: 
\code{\link{PKNCAconc}()},
\code{\link{PKNCAdata}()},
\code{\link{PKNCAdose}()}
}
\concept{PKNCA objects}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exclude.R
\name{setExcludeColumn}
\alias{setExcludeColumn}
\title{Set the exclude parameter on an object}
\usage{
setExcludeColumn(object, exclude = NULL, dataname = "data")
}
\arguments{
\item{object}{The object to set the exclude column on.}

\item{exclude}{The column name to set as the exclude value.}

\item{dataname}{The name of the data.frame within the object to add the
exclude column to.}
}
\value{
The object with an exclude column and attribute
}
\description{
This function adds the exclude column to an object.  To change the
exclude value, use the \code{\link[=exclude]{exclude()}} function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exclude_nca.R
\name{exclude_nca}
\alias{exclude_nca}
\alias{exclude_nca_span.ratio}
\alias{exclude_nca_max.aucinf.pext}
\alias{exclude_nca_count_conc_measured}
\alias{exclude_nca_min.hl.r.squared}
\alias{exclude_nca_min.hl.adj.r.squared}
\alias{exclude_nca_tmax_early}
\alias{exclude_nca_tmax_0}
\title{Exclude NCA parameters based on examining the parameter set.}
\usage{
exclude_nca_span.ratio(min.span.ratio)

exclude_nca_max.aucinf.pext(max.aucinf.pext)

exclude_nca_count_conc_measured(
  min_count,
  exclude_param_pattern = c("^aucall", "^aucinf", "^aucint", "^auciv", "^auclast",
    "^aumc", "^sparse_auc")
)

exclude_nca_min.hl.r.squared(min.hl.r.squared)

exclude_nca_min.hl.adj.r.squared(min.hl.adj.r.squared = 0.9)

exclude_nca_tmax_early(tmax_early = 0)

exclude_nca_tmax_0()
}
\arguments{
\item{min.span.ratio}{The minimum acceptable span ratio (uses
\code{PKNCA.options("min.span.ratio")} if not provided).}

\item{max.aucinf.pext}{The maximum acceptable percent AUC extrapolation (uses
\code{PKNCA.options("max.aucinf.pext")} if not provided).}

\item{min_count}{Minimum number of measured concentrations}

\item{exclude_param_pattern}{Character vector of regular expression patterns to exclude}

\item{min.hl.r.squared}{The minimum acceptable r-squared value for half-life
(uses \code{PKNCA.options("min.hl.r.squared")} if not provided).}

\item{min.hl.adj.r.squared}{The minimum acceptable adjusted r-squared for half-life
(uses 0.9 if not provided).}

\item{tmax_early}{The time for Tmax which is considered too early to be a valid NCA result}
}
\description{
Exclude NCA parameters based on examining the parameter set.
}
\section{Functions}{
\itemize{
\item \code{exclude_nca_span.ratio()}: Exclude based on span.ratio

\item \code{exclude_nca_max.aucinf.pext()}: Exclude based on AUC percent extrapolated (both observed and predicted)

\item \code{exclude_nca_count_conc_measured()}: Exclude AUC measurements based on count of concentrations measured and not below the lower limit of quantification

\item \code{exclude_nca_min.hl.r.squared()}: Exclude based on half-life r-squared

\item \code{exclude_nca_min.hl.adj.r.squared()}: Exclude based on half-life adjusted r-squared

\item \code{exclude_nca_tmax_early()}: Exclude based on implausibly early Tmax (often used for extravascular dosing with a Tmax value of 0)

\item \code{exclude_nca_tmax_0()}: Exclude based on implausibly early Tmax (special case for \code{tmax_early = 0})

}}
\examples{
my_conc <- PKNCAconc(data.frame(conc=1.1^(3:0),
                                time=0:3,
                                subject=1),
                     conc~time|subject)
my_data <- PKNCAdata(my_conc,
                     intervals=data.frame(start=0, end=Inf,
                                          aucinf.obs=TRUE,
                                          aucpext.obs=TRUE))
my_result <- pk.nca(my_data)
my_result_excluded <- exclude(my_result,
                              FUN=exclude_nca_max.aucinf.pext())
as.data.frame(my_result_excluded)
}
\seealso{
Other Result exclusions: 
\code{\link{exclude}()}
}
\concept{Result exclusions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-general.R
\name{getColumnValueOrNot}
\alias{getColumnValueOrNot}
\title{Get the value from a column in a data frame if the value is a column
there, otherwise, the value should be a scalar or the length of the
data.}
\usage{
getColumnValueOrNot(data, value, prefix = "X")
}
\arguments{
\item{data}{A data.frame or similar object}

\item{value}{A character string giving the name of a column in the \code{data}, a
scalar, or a vector the same length as the \code{data}}

\item{prefix}{The prefix to use if a column must be added (it will be used as
the full column name if it is not already in the dataset or it will be
prepended to the maximum column name if not.)}
}
\value{
A list with elements named "data", "name" giving the \code{data} with a
column named "name" with the value in that column.
}
\description{
Get the value from a column in a data frame if the value is a column
there, otherwise, the value should be a scalar or the length of the
data.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_and_assert_intervals.R
\name{set_intervals}
\alias{set_intervals}
\title{Set Intervals}
\usage{
set_intervals(data, intervals)
}
\arguments{
\item{data}{PKNCAdata object}

\item{intervals}{Proposed intervals}
}
\value{
The data object with the intervals set.
}
\description{
Takes in two objects, the PKNCAdata object and the proposed intervals.
It will then check that the intervals are valid, given the data object.
If the intervals are valid, it will set them in the object.
It will return the data object with the intervals set.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-PKNCAresults.R
\name{as.data.frame.PKNCAresults}
\alias{as.data.frame.PKNCAresults}
\title{Extract the parameter results from a PKNCAresults and return them as a
data.frame.}
\usage{
\method{as.data.frame}{PKNCAresults}(
  x,
  ...,
  out_format = c("long", "wide"),
  filter_requested = FALSE,
  filter_excluded = FALSE,
  out.format = deprecated()
)
}
\arguments{
\item{x}{The object to extract results from}

\item{...}{Ignored (for compatibility with generic \code{\link[=as.data.frame]{as.data.frame()}})}

\item{out_format}{Should the output be 'long' (default) or 'wide'?}

\item{filter_requested}{Only return rows with parameters that were
specifically requested?}

\item{filter_excluded}{Should excluded values be removed?}

\item{out.format}{Deprecated in favor of \code{out_format}}
}
\value{
A data.frame (or usually a tibble) of results
}
\description{
Extract the parameter results from a PKNCAresults and return them as a
data.frame.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/provenance.R
\name{print.provenance}
\alias{print.provenance}
\title{Print the summary of a provenance object}
\usage{
\method{print}{provenance}(x, ...)
}
\arguments{
\item{x}{The object to be printed}

\item{...}{Ignored}
}
\value{
invisible text of the printed information
}
\description{
Print the summary of a provenance object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/half.life.R
\name{fit_half_life}
\alias{fit_half_life}
\title{Perform the half-life fit given the data.  The function simply fits
the data without any validation.  No selection of points or any other
components are done.}
\usage{
fit_half_life(data, tlast, conc_units)
}
\arguments{
\item{data}{The data to fit.  Must have two columns named "log_conc"
and "time"}

\item{tlast}{The time of last observed concentration above the limit
of quantification.}

\item{conc_units}{NULL or the units to set for concentration measures}
}
\value{
A data.frame with one row and columns named "r.squared",
"adj.r.squared", "PROB", "lambda.z", "clast.pred",
"lambda.z.n.points", "half.life", "span.ratio"
}
\description{
Perform the half-life fit given the data.  The function simply fits
the data without any validation.  No selection of points or any other
components are done.
}
\seealso{
\code{\link[=pk.calc.half.life]{pk.calc.half.life()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/002-pk.business.rules.R
\name{geomean}
\alias{geomean}
\alias{geosd}
\alias{geocv}
\title{Compute the geometric mean, sd, and CV}
\usage{
geomean(x, na.rm = FALSE)

geosd(x, na.rm = FALSE)

geocv(x, na.rm = FALSE)
}
\arguments{
\item{x}{A vector to compute the geometric mean of}

\item{na.rm}{Should missing values be removed?}
}
\value{
The scalar value of the geometric mean, geometric standard deviation,
or geometric coefficient of variation.
}
\description{
Compute the geometric mean, sd, and CV
}
\section{Functions}{
\itemize{
\item \code{geosd()}: Compute the geometric standard deviation, \code{exp(sd(log(x)))}.

\item \code{geocv()}: Compute the geometric coefficient of variation,
\code{sqrt(exp(sd(log(x))^2)-1)*100}.

}}
\examples{
geomean(1:3)
geosd(1:3)
geocv(1:3)
}
\references{
Kirkwood T. B.L. Geometric means and measures of dispersion.
Biometrics 1979; 35: 908-909
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pk.calc.simple.R
\name{pk.calc.clast.obs}
\alias{pk.calc.clast.obs}
\title{Determine the last observed concentration above the limit of quantification
(LOQ).}
\usage{
pk.calc.clast.obs(conc, time, check = TRUE)
}
\arguments{
\item{conc}{Measured concentrations}

\item{time}{Time of the measurement of the concentrations}

\item{check}{Run \code{\link[=assert_conc_time]{assert_conc_time()}}?}
}
\value{
The last observed concentration above the LOQ
}
\description{
If all concentrations are missing, \code{NA_real_} is returned.  If all
concentrations are zero (below the limit of quantification) or missing, zero
is returned.  If Tlast is NA (due to no non-missing above LOQ measurements),
this will return \code{NA_real_}.
}
\seealso{
Other NCA parameters for concentrations during the intervals: 
\code{\link{pk.calc.cmax}()},
\code{\link{pk.calc.count_conc}()},
\code{\link{pk.calc.cstart}()},
\code{\link{pk.calc.ctrough}()}
}
\concept{NCA parameters for concentrations during the intervals}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dplyr.R
\name{group_by.PKNCAresults}
\alias{group_by.PKNCAresults}
\alias{group_by.PKNCAconc}
\alias{group_by.PKNCAdose}
\alias{ungroup.PKNCAresults}
\alias{ungroup.PKNCAconc}
\alias{ungroup.PKNCAdose}
\title{dplyr grouping for PKNCA}
\usage{
\method{group_by}{PKNCAresults}(.data, ..., .add = FALSE, .drop = dplyr::group_by_drop_default(.data))

\method{group_by}{PKNCAconc}(.data, ..., .add = FALSE, .drop = dplyr::group_by_drop_default(.data))

\method{group_by}{PKNCAdose}(.data, ..., .add = FALSE, .drop = dplyr::group_by_drop_default(.data))

\method{ungroup}{PKNCAresults}(x, ...)

\method{ungroup}{PKNCAconc}(x, ...)

\method{ungroup}{PKNCAdose}(x, ...)
}
\arguments{
\item{.data}{A data frame, data frame extension (e.g. a tibble), or a
lazy data frame (e.g. from dbplyr or dtplyr). See \emph{Methods}, below, for
more details.}

\item{...}{In \code{group_by()}, variables or computations to group by.
Computations are always done on the ungrouped data frame.
To perform computations on the grouped data, you need to use
a separate \code{mutate()} step before the \code{group_by()}.
Computations are not allowed in \code{nest_by()}.
In \code{ungroup()}, variables to remove from the grouping.}

\item{.add}{When \code{FALSE}, the default, \code{group_by()} will
override existing groups. To add to the existing groups, use
\code{.add = TRUE}.

This argument was previously called \code{add}, but that prevented
creating a new grouping variable called \code{add}, and conflicts with
our naming conventions.}

\item{.drop}{Drop groups formed by factor levels that don't appear in the
data? The default is \code{TRUE} except when \code{.data} has been previously
grouped with \code{.drop = FALSE}. See \code{\link[dplyr:group_by_drop_default]{group_by_drop_default()}} for details.}

\item{x}{A \code{\link[dplyr:tbl]{tbl()}}}
}
\description{
dplyr grouping for PKNCA
}
\seealso{
Other dplyr verbs: 
\code{\link{filter.PKNCAresults}()},
\code{\link{inner_join.PKNCAresults}()},
\code{\link{mutate.PKNCAresults}()}
}
\concept{dplyr verbs}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-PKNCAdata.R
\name{print.PKNCAdata}
\alias{print.PKNCAdata}
\title{Print a PKNCAdata object}
\usage{
\method{print}{PKNCAdata}(x, ...)
}
\arguments{
\item{x}{The object to print}

\item{...}{Arguments passed on to \code{\link[=print.PKNCAconc]{print.PKNCAconc()}} and \code{\link[=print.PKNCAdose]{print.PKNCAdose()}}}
}
\description{
Print a PKNCAdata object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assertions.R
\name{assert_intervaltime_single}
\alias{assert_intervaltime_single}
\title{Assert that an interval is accurately defined as an interval, and return the
interval}
\usage{
assert_intervaltime_single(interval = NULL, start = NULL, end = NULL)
}
\arguments{
\item{interval}{Numeric vector of two numbers for the start and end time of
integration}

\item{start}{The start time of the interval}

\item{end}{The end time of the interval}
}
\value{
\code{interval} (or \code{c(start, end)})
}
\description{
Assert that an interval is accurately defined as an interval, and return the
interval
}
\keyword{Internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-general.R
\name{setAttributeColumn}
\alias{setAttributeColumn}
\title{Add an attribute to an object where the attribute is added as a name
to the names of the object.}
\usage{
setAttributeColumn(
  object,
  attr_name,
  col_or_value,
  col_name,
  default_value,
  stop_if_default,
  warn_if_default,
  message_if_default
)
}
\arguments{
\item{object}{The object to set the attribute column on.}

\item{attr_name}{The attribute name to set}

\item{col_or_value}{If this exists as a column in the data, it is used as the
\code{col_name}.  If not, this becomes the \code{default_value}.}

\item{col_name}{The name of the column within the dataset to use (if missing,
uses \code{attr_name})}

\item{default_value}{The value to fill in the column if the column does not
exist (the column is filled with \code{NA} if it does not exist and no value is
provided).}

\item{stop_if_default, warn_if_default, message_if_default}{A character string
to provide as an error, a warning, or a message to the user if the
\code{default_value} is used.  They are tested in order (if stop, the code
stops; if warning, the message is ignored; and message last).}
}
\value{
The object with the attribute column added to the data.
}
\description{
Add an attribute to an object where the attribute is added as a name
to the names of the object.
}
\seealso{
\code{\link[=getAttributeColumn]{getAttributeColumn()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assertions.R
\name{assert_number_between}
\alias{assert_number_between}
\title{Confirm that a value is greater than another value}
\usage{
assert_number_between(
  x,
  ...,
  na.ok = FALSE,
  len = 1,
  .var.name = checkmate::vname(x)
)
}
\arguments{
\item{x}{[any]\cr
Object to check.}

\item{...}{Passed to \code{assert_numeric_between()}}

\item{na.ok}{[\code{logical(1)}]\cr
Are missing values allowed? Default is \code{FALSE}.}

\item{len}{Ignored (must be 1)}

\item{.var.name}{[\code{character(1)}]\cr
Name of the checked object to print in assertions. Defaults to
the heuristic implemented in \code{\link[checkmate]{vname}}.}
}
\value{
\code{x} or an informative error
}
\description{
Confirm that a value is greater than another value
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assertions.R
\name{assert_aucmethod}
\alias{assert_aucmethod}
\title{Assert that a value is a valid AUC method}
\usage{
assert_aucmethod(method = c("lin up/log down", "linear", "lin-log"))
}
\arguments{
\item{method}{The method for integration (one of 'lin up/log down',
'lin-log', or 'linear')}
}
\value{
\code{method} or an informative error
}
\description{
Assert that a value is a valid AUC method
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PKNCA.options.R
\name{PKNCA.set.summary}
\alias{PKNCA.set.summary}
\title{Define how NCA parameters are summarized.}
\usage{
PKNCA.set.summary(
  name,
  description,
  point,
  spread,
  rounding = list(signif = 3),
  reset = FALSE
)
}
\arguments{
\item{name}{The parameter name or a vector of parameter names.  It must have
already been defined (see \code{\link[=add.interval.col]{add.interval.col()}}).}

\item{description}{A single-line description of the summary}

\item{point}{The function to calculate the point estimate for the summary.
The function will be called as \code{point(x)} and must return a scalar value
(typically a number, NA, or a string).}

\item{spread}{Optional.  The function to calculate the spread (or
variability).  The function will be called as \code{spread(x)} and must return a
scalar or two-long vector (typically a number, NA, or a string).}

\item{rounding}{Instructions for how to round the value of point and spread.
It may either be a list or a function.  If it is a list, then it must have
a single entry with a name of either "signif" or "round" and a value of the
digits to round.  If a function, it is expected to return a scalar number
or character string with the correct results for an input of either a
scalar or a two-long vector.}

\item{reset}{Reset all the summary instructions}
}
\value{
All current summary settings (invisibly)
}
\description{
Define how NCA parameters are summarized.
}
\examples{
\dontrun{
PKNCA.set.summary(
  name="half.life",
  description="arithmetic mean and standard deviation",
  point=business.mean,
  spread=business.sd,
  rounding=list(signif=3)
)
}
}
\seealso{
\code{\link[=summary.PKNCAresults]{summary.PKNCAresults()}}

Other PKNCA calculation and summary settings: 
\code{\link{PKNCA.choose.option}()},
\code{\link{PKNCA.options}()}
}
\concept{PKNCA calculation and summary settings}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exclude_nca.R
\name{exclude_nca_by_param}
\alias{exclude_nca_by_param}
\title{Exclude NCA Results Based on Parameter Thresholds}
\usage{
exclude_nca_by_param(
  parameter,
  min_thr = NULL,
  max_thr = NULL,
  affected_parameters = parameter
)
}
\arguments{
\item{parameter}{The name of the PKNCA parameter to evaluate (e.g., "span.ratio").}

\item{min_thr}{The minimum acceptable value for the parameter. If not provided, is not applied.}

\item{max_thr}{The maximum acceptable value for the parameter. If not provided, is not applied.}

\item{affected_parameters}{Character vector of PKNCA parameters that will be marked as excluded.
By default is the defined parameter.}
}
\value{
A function that can be used with \code{PKNCA::exclude} to mark through the 'exclude'  column
the rows in the PKNCA results based on the specified thresholds for a parameter.
}
\description{
Exclude rows from NCA results based on specified thresholds for a given parameter.
This function allows users to define minimum and/or maximum acceptable values
for a parameter and excludes rows that fall outside these thresholds.
}
\examples{
# Example dataset
my_data <- PKNCA::PKNCAdata(
  PKNCA::PKNCAconc(data.frame(conc = 5:1,
                              time = 0:4,
                              subject = 1),
                   conc ~ time | subject),
  PKNCA::PKNCAdose(data.frame(subject = 1, dose = 100, time = 0),
                   dose ~ time | subject)
)
my_result <- PKNCA::pk.nca(my_data)

# Exclude rows where span.ratio is less than 2
excluded_result <- PKNCA::exclude(
  my_result,
  FUN = exclude_nca_by_param("span.ratio", min_thr = 2)
)
as.data.frame(excluded_result)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sparse.R
\name{cov_holder}
\alias{cov_holder}
\title{Calculate the covariance for two time points with sparse sampling}
\usage{
cov_holder(sparse_pk)
}
\arguments{
\item{sparse_pk}{A sparse_pk object from \code{\link[=as_sparse_pk]{as_sparse_pk()}}}
}
\value{
A matrix with one row and one column for each element of
\code{sparse_pk_attribute}.  The covariances are on the off diagonals, and for
simplicity of use, it also calculates the variance on the diagonal
elements.
}
\description{
The calculation follows equation A3 in Holder 2001 (see references below):
}
\details{
\deqn{\hat{\sigma}_{ij} = \sum\limits_{k=1}^{r_{ij}}{\frac{\left(x_{ik} - \bar{x}_i\right)\left(x_{jk} - \bar{x}_j\right)}{\left(r_{ij} - 1\right) + \left(1 - \frac{r_{ij}}{r_i}\right)\left(1 - \frac{r_{ij}}{r_j}\right)}}}{sigma_ij = sum_(k=1)^(r_ij)((x_ik-xbar_i)(x_jk-xbar_j)/((r_ij-1)+(1-r_ij/r_i)*(1-r_ij/r_j)))}

If \eqn{r_{ij} = 0}{r_ij = 0}, then \eqn{\hat{\sigma}_{ij}}{sigma_ij} is
defined as zero (rather than dividing by zero).

Where:
\describe{
\item{\eqn{\hat{\sigma}_{ij}}{sigma_ij}}{The covariance of times i and j}
\item{\eqn{r_i}{r_i} and \eqn{r_j}{r_j}}{The number of subjects (usually animals) at times i and j, respectively}
\item{\eqn{r_{ij}{r_ij}}}{The number of subjects (usually animals) at both times i and j}
\item{\eqn{x_{ik}}{x_ik} and \eqn{x_{jk}}{x_jk}}{The concentration measured for animal k at times i and j, respectively}
\item{\eqn{\bar{x}_i}{xbar_i} and \eqn{\bar{x}_j}{xbar_j}}{The mean of the concentrations at times i and j, respectively}
}

The Cauchy-Schwartz inequality is enforced for covariances to keep
correlation coefficients between -1 and 1, inclusive, as described in
equations 8 and 9 of Nedelman and Jia 1998.
}
\references{
Holder DJ. Comments on Nedelman and Jia’s Extension of Satterthwaite’s
Approximation Applied to Pharmacokinetics. Journal of Biopharmaceutical
Statistics. 2001;11(1-2):75-79. doi:10.1081/BIP-100104199

Nedelman JR, Jia X. An extension of Satterthwaite’s approximation applied to
pharmacokinetics. Journal of Biopharmaceutical Statistics. 1998;8(2):317-328.
doi:10.1080/10543409808835241
}
\keyword{Internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/half.life.R
\name{pk.calc.half.life}
\alias{pk.calc.half.life}
\title{Compute the half-life and associated parameters}
\usage{
pk.calc.half.life(
  conc,
  time,
  tmax,
  tlast,
  time.dose = NULL,
  duration.dose = 0,
  manually.selected.points = FALSE,
  options = list(),
  min.hl.points = NULL,
  adj.r.squared.factor = NULL,
  conc.blq = NULL,
  conc.na = NULL,
  first.tmax = NULL,
  allow.tmax.in.half.life = NULL,
  check = TRUE
)
}
\arguments{
\item{conc}{Measured concentrations}

\item{time}{Time of the measurement of the concentrations}

\item{tmax}{Time of maximum concentration (will be calculated and
included in the return data frame if not given)}

\item{tlast}{Time of last concentration above the limit of
quantification (will be calculated and included in the return data
frame if not given)}

\item{time.dose}{Time of the dose for the current interval (must be the same
length as \code{dose})}

\item{duration.dose}{The duration of the dose administration for the current
interval (typically zero for extravascular and intravascular bolus and
nonzero for intravascular infusion)}

\item{manually.selected.points}{Have the input points (\code{conc} and
\code{time}) been manually selected?  The impact of setting this to
\code{TRUE} is that no selection for the best points will be done.  When
\code{TRUE}, this option causes the options of \code{adj.r.squared.factor},
\code{min.hl.points}, and \code{allow.tmax.in.half.life} to be ignored.}

\item{options}{List of changes to the default PKNCA options (see
\code{PKNCA.options()})}

\item{min.hl.points}{The minimum number of points that must be
included to calculate the half-life}

\item{adj.r.squared.factor}{The allowance in adjusted r-squared for
adding another point.}

\item{conc.blq}{See \code{\link[=clean.conc.blq]{clean.conc.blq()}}}

\item{conc.na}{See \code{\link[=clean.conc.na]{clean.conc.na()}}}

\item{first.tmax}{See \code{\link[=pk.calc.tmax]{pk.calc.tmax()}}.}

\item{allow.tmax.in.half.life}{Allow the concentration point for tmax
to be included in the half-life slope calculation.}

\item{check}{Run \code{\link[=assert_conc_time]{assert_conc_time()}},
\code{\link[=clean.conc.blq]{clean.conc.blq()}}, and \code{\link[=clean.conc.na]{clean.conc.na()}}?}
}
\value{
A data frame with one row and columns for
\describe{
\item{tmax}{Time of maximum observed concentration (only included
if not given as an input)}
\item{tlast}{Time of last observed concentration above the LOQ (only
included if not given as an input)}
\item{r.squared}{coefficient of determination}
\item{adj.r.squared}{adjusted coefficient of determination}
\item{lambda.z}{elimination rate}
\item{lambda.z.time.first}{first time for half-life calculation}
\item{lambda.z.time.last}{last time for half-life calculation}
\item{lambda.z.n.points}{number of points in half-life calculation}
\item{clast.pred}{Concentration at tlast as predicted by the half-life
line}
\item{half.life}{half-life}
\item{span.ratio}{span ratio [ratio of half-life to time used for
half-life calculation}
}
}
\description{
The terminal elimination half-life is estimated from the final points in the
concentration-time curve using semi-log regression (\code{log(conc)~time})
with automated selection of the points for calculation (unless
\code{manually.selected.points} is \code{TRUE}).
}
\details{
See the "Half-Life Calculation" vignette for more details on the calculation
methods used.

If \code{manually.selected.points} is \code{FALSE} (default), the
half-life is calculated by computing the best fit line for all points at or
after tmax (based on the value of \code{allow.tmax.in.half.life}).  The best
half-life is chosen by the following rules in order:

\itemize{
\item{At least \code{min.hl.points} points included}
\item{A \code{lambda.z} > 0 and at the same time the best adjusted r-squared
(within \code{adj.r.squared.factor})}
\item{The one with the most points included}
}

If \code{manually.selected.points} is \code{TRUE}, the \code{conc}
and \code{time} data are used as-is without any form of selection for
the best-fit half-life.
}
\references{
Gabrielsson J, Weiner D.  "Section 2.8.4 Strategies for estimation of
lambda-z."  Pharmacokinetic & Pharmacodynamic Data Analysis: Concepts
and Applications, 4th Edition.  Stockholm, Sweden: Swedish
Pharmaceutical Press, 2000.  167-9.
}
\concept{NCA parameter calculations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assertions.R
\name{assert_PKNCAdata}
\alias{assert_PKNCAdata}
\title{Assert that an object is a PKNCAdata object}
\usage{
assert_PKNCAdata(object)
}
\arguments{
\item{object}{The PKNCAdata object}
}
\value{
The PKNCAdata object (confirmed to be usable)
}
\description{
Assert that an object is a PKNCAdata object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/002-pk.business.rules.R
\name{pk.business}
\alias{pk.business}
\title{Run any function with a maximum missing fraction of X and 0s possibly
counting as missing.  The maximum fraction missing comes from
\code{PKNCA.options("max.missing")}.}
\usage{
pk.business(FUN, zero.missing = FALSE, max.missing)
}
\arguments{
\item{FUN}{function to run.  The function is called as \code{FUN(x, ...)} with
missing values removed.}

\item{zero.missing}{Are zeros counted as missing?  If \code{TRUE} then include
them in the missing count.}

\item{max.missing}{The maximum fraction of the data allowed to be missing (a
number between 0 and 1, inclusive).}
}
\value{
A version of FUN that can be called with parameters that are checked
for missingness (and zeros) with missing (and zeros) removed before the
call.  If \code{max.missing} is exceeded, then NA is returned.
}
\description{
Note that all missing values are removed prior to calling the function.
}
\examples{
my_mean <- pk.business(FUN=mean)
mean(c(1:3, NA))
# Less than half missing results in the summary statistic of the available
# values.
my_mean(c(1:3, NA))
# More than half missing results in a missing value
my_mean(c(1:3, rep(NA, 4)))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/time.above.R
\name{pk.calc.time_above}
\alias{pk.calc.time_above}
\title{Determine time at or above a set value}
\usage{
pk.calc.time_above(conc, time, conc_above, ..., options = list(), check = TRUE)
}
\arguments{
\item{conc}{Measured concentrations}

\item{time}{Time of the measurement of the concentrations}

\item{conc_above}{The concentration to be above}

\item{...}{Extra arguments.  Currently, the only extra argument that is used
is \code{method} as described in the details section.}

\item{options}{List of changes to the default PKNCA options (see
\code{PKNCA.options()})}

\item{check}{Run \code{\link[=assert_conc_time]{assert_conc_time()}}, \code{\link[=clean.conc.blq]{clean.conc.blq()}}, and
\code{\link[=clean.conc.na]{clean.conc.na()}}?}
}
\value{
the time above the given concentration
}
\description{
Interpolation is performed aligning with \code{PKNCA.options("auc.method")}.
Extrapolation outside of the measured times is not yet implemented.  The
\code{method} may be changed by giving a named \code{method} argument, as well.
}
\details{
For \code{'lin up/log down'}, if \code{clast} is above \code{conc_above} and there are
concentrations BLQ after that, linear down is used to extrapolate to the BLQ
concentration (equivalent to AUCall).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-PKNCAconc.R, R/class-PKNCAdose.R
\name{print.PKNCAconc}
\alias{print.PKNCAconc}
\alias{summary.PKNCAconc}
\alias{print.PKNCAdose}
\alias{summary.PKNCAdose}
\title{Print and/or summarize a PKNCAconc or PKNCAdose object.}
\usage{
\method{print}{PKNCAconc}(x, n = 6, summarize = FALSE, ...)

\method{summary}{PKNCAconc}(object, n = 0, summarize = TRUE, ...)

\method{print}{PKNCAdose}(x, n = 6, summarize = FALSE, ...)

\method{summary}{PKNCAdose}(object, n = 0, summarize = TRUE, ...)
}
\arguments{
\item{x}{The object to print}

\item{n}{The number of rows of data to show (see
\code{\link[=head]{head()}})}

\item{summarize}{Summarize the nested number of groups}

\item{\dots}{Arguments passed to \code{print.formula} and
\code{print.data.frame}}

\item{object}{The object to summarize}
}
\description{
Print and/or summarize a PKNCAconc or PKNCAdose object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz-pk.calc.dn.R
\name{pk.calc.dn}
\alias{pk.calc.dn}
\title{Determine dose normalized NCA parameter}
\usage{
pk.calc.dn(parameter, dose)
}
\arguments{
\item{parameter}{Parameter to dose normalize}

\item{dose}{Dose in units compatible with the area under the curve}
}
\value{
a number for dose normalized AUC
}
\description{
Determine dose normalized NCA parameter
}
\examples{
pk.calc.dn(90, 10)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dplyr.R
\name{inner_join.PKNCAresults}
\alias{inner_join.PKNCAresults}
\alias{left_join.PKNCAresults}
\alias{right_join.PKNCAresults}
\alias{full_join.PKNCAresults}
\alias{inner_join.PKNCAconc}
\alias{left_join.PKNCAconc}
\alias{right_join.PKNCAconc}
\alias{full_join.PKNCAconc}
\alias{inner_join.PKNCAdose}
\alias{left_join.PKNCAdose}
\alias{right_join.PKNCAdose}
\alias{full_join.PKNCAdose}
\title{dplyr joins for PKNCA}
\usage{
\method{inner_join}{PKNCAresults}(
  x,
  y,
  by = NULL,
  copy = FALSE,
  suffix = c(".x", ".y"),
  ...,
  keep = FALSE
)

\method{left_join}{PKNCAresults}(
  x,
  y,
  by = NULL,
  copy = FALSE,
  suffix = c(".x", ".y"),
  ...,
  keep = FALSE
)

\method{right_join}{PKNCAresults}(
  x,
  y,
  by = NULL,
  copy = FALSE,
  suffix = c(".x", ".y"),
  ...,
  keep = FALSE
)

\method{full_join}{PKNCAresults}(
  x,
  y,
  by = NULL,
  copy = FALSE,
  suffix = c(".x", ".y"),
  ...,
  keep = FALSE
)

\method{inner_join}{PKNCAconc}(
  x,
  y,
  by = NULL,
  copy = FALSE,
  suffix = c(".x", ".y"),
  ...,
  keep = FALSE
)

\method{left_join}{PKNCAconc}(
  x,
  y,
  by = NULL,
  copy = FALSE,
  suffix = c(".x", ".y"),
  ...,
  keep = FALSE
)

\method{right_join}{PKNCAconc}(
  x,
  y,
  by = NULL,
  copy = FALSE,
  suffix = c(".x", ".y"),
  ...,
  keep = FALSE
)

\method{full_join}{PKNCAconc}(
  x,
  y,
  by = NULL,
  copy = FALSE,
  suffix = c(".x", ".y"),
  ...,
  keep = FALSE
)

\method{inner_join}{PKNCAdose}(
  x,
  y,
  by = NULL,
  copy = FALSE,
  suffix = c(".x", ".y"),
  ...,
  keep = FALSE
)

\method{left_join}{PKNCAdose}(
  x,
  y,
  by = NULL,
  copy = FALSE,
  suffix = c(".x", ".y"),
  ...,
  keep = FALSE
)

\method{right_join}{PKNCAdose}(
  x,
  y,
  by = NULL,
  copy = FALSE,
  suffix = c(".x", ".y"),
  ...,
  keep = FALSE
)

\method{full_join}{PKNCAdose}(
  x,
  y,
  by = NULL,
  copy = FALSE,
  suffix = c(".x", ".y"),
  ...,
  keep = FALSE
)
}
\arguments{
\item{x, y}{A pair of data frames, data frame extensions (e.g. a tibble), or
lazy data frames (e.g. from dbplyr or dtplyr). See \emph{Methods}, below, for
more details.}

\item{by}{A join specification created with \code{\link[dplyr:join_by]{join_by()}}, or a character
vector of variables to join by.

If \code{NULL}, the default, \verb{*_join()} will perform a natural join, using all
variables in common across \code{x} and \code{y}. A message lists the variables so
that you can check they're correct; suppress the message by supplying \code{by}
explicitly.

To join on different variables between \code{x} and \code{y}, use a \code{\link[dplyr:join_by]{join_by()}}
specification. For example, \code{join_by(a == b)} will match \code{x$a} to \code{y$b}.

To join by multiple variables, use a \code{\link[dplyr:join_by]{join_by()}} specification with
multiple expressions. For example, \code{join_by(a == b, c == d)} will match
\code{x$a} to \code{y$b} and \code{x$c} to \code{y$d}. If the column names are the same between
\code{x} and \code{y}, you can shorten this by listing only the variable names, like
\code{join_by(a, c)}.

\code{\link[dplyr:join_by]{join_by()}} can also be used to perform inequality, rolling, and overlap
joins. See the documentation at \link[dplyr:join_by]{?join_by} for details on
these types of joins.

For simple equality joins, you can alternatively specify a character vector
of variable names to join by. For example, \code{by = c("a", "b")} joins \code{x$a}
to \code{y$a} and \code{x$b} to \code{y$b}. If variable names differ between \code{x} and \code{y},
use a named character vector like \code{by = c("x_a" = "y_a", "x_b" = "y_b")}.

To perform a cross-join, generating all combinations of \code{x} and \code{y}, see
\code{\link[dplyr:cross_join]{cross_join()}}.}

\item{copy}{If \code{x} and \code{y} are not from the same data source,
and \code{copy} is \code{TRUE}, then \code{y} will be copied into the
same src as \code{x}.  This allows you to join tables across srcs, but
it is a potentially expensive operation so you must opt into it.}

\item{suffix}{If there are non-joined duplicate variables in \code{x} and
\code{y}, these suffixes will be added to the output to disambiguate them.
Should be a character vector of length 2.}

\item{...}{Other parameters passed onto methods.}

\item{keep}{Should the join keys from both \code{x} and \code{y} be preserved in the
output?
\itemize{
\item If \code{NULL}, the default, joins on equality retain only the keys from \code{x},
while joins on inequality retain the keys from both inputs.
\item If \code{TRUE}, all keys from both inputs are retained.
\item If \code{FALSE}, only keys from \code{x} are retained. For right and full joins,
the data in key columns corresponding to rows that only exist in \code{y} are
merged into the key columns from \code{x}. Can't be used when joining on
inequality conditions.
}}
}
\description{
dplyr joins for PKNCA
}
\seealso{
Other dplyr verbs: 
\code{\link{filter.PKNCAresults}()},
\code{\link{group_by.PKNCAresults}()},
\code{\link{mutate.PKNCAresults}()}
}
\concept{dplyr verbs}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pk.calc.all.R
\name{pk.nca.interval}
\alias{pk.nca.interval}
\title{Compute all PK parameters for a single concentration-time data set}
\usage{
pk.nca.interval(
  conc,
  time,
  volume,
  duration.conc,
  dose,
  time.dose,
  duration.dose,
  route,
  conc.group = NULL,
  time.group = NULL,
  volume.group = NULL,
  duration.conc.group = NULL,
  dose.group = NULL,
  time.dose.group = NULL,
  duration.dose.group = NULL,
  route.group = NULL,
  impute_method = NA_character_,
  include_half.life = NULL,
  exclude_half.life = NULL,
  subject,
  sparse,
  interval,
  options = list()
)
}
\arguments{
\item{conc}{Measured concentrations}

\item{time}{Time of the measurement of the concentrations}

\item{volume, volume.group}{The volume (or mass) of the concentration
measurement for the current interval or all data for the group (typically
for urine and fecal measurements)}

\item{duration.conc, duration.conc.group}{The duration of the concentration
measurement for the current interval or all data for the group (typically
for urine and fecal measurements)}

\item{dose, dose.group}{Dose amount (may be a scalar or vector) for the
current interval or all data for the group}

\item{time.dose}{Time of the dose for the current interval (must be the same
length as \code{dose})}

\item{duration.dose}{The duration of the dose administration for the current
interval (typically zero for extravascular and intravascular bolus and
nonzero for intravascular infusion)}

\item{route, route.group}{The route of dosing for the current interval or all
data for the group}

\item{conc.group}{All concentrations measured for the group}

\item{time.group}{Time of all concentrations measured for the group}

\item{time.dose.group}{Time of the dose for all data for the group (must be
the same length as \code{dose.group})}

\item{duration.dose.group}{The duration of the dose administration for all
data for the group (typically zero for extravascular and intravascular
bolus and nonzero for intravascular infusion)}

\item{impute_method}{The method to use for imputation as a character string}

\item{include_half.life}{An optional boolean vector of the concentration
measurements to include in the half-life calculation. If given, no
half-life point selection will occur.}

\item{exclude_half.life}{An optional boolean vector of the concentration
measurements to exclude from the half-life calculation.}

\item{subject}{Subject identifiers (used for sparse calculations)}

\item{sparse}{Should only sparse calculations be performed (TRUE) or only
dense calculations (FALSE)?}

\item{interval}{One row of an interval definition (see
\code{\link[=check.interval.specification]{check.interval.specification()}} for how to define the interval.}

\item{options}{List of changes to the default PKNCA options (see
\code{PKNCA.options()})}
}
\value{
A data frame with the start and end time along with all PK
parameters for the \code{interval}
}
\description{
For one subject/time range, compute all available PK parameters. All the
internal options should be set by \code{\link[=PKNCA.options]{PKNCA.options()}} prior to running.  The
only part that changes with a call to this function is the \code{conc}entration
and \code{time}.
}
\seealso{
\code{\link[=check.interval.specification]{check.interval.specification()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-PKNCAdata.R
\name{summary.PKNCAdata}
\alias{summary.PKNCAdata}
\title{Summarize a PKNCAdata object showing important details about the
concentration, dosing, and interval information.}
\usage{
\method{summary}{PKNCAdata}(object, ...)
}
\arguments{
\item{object}{The PKNCAdata object to summarize.}

\item{...}{arguments passed on to \code{\link[=print.PKNCAdata]{print.PKNCAdata()}}}
}
\description{
Summarize a PKNCAdata object showing important details about the
concentration, dosing, and interval information.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pk.calc.urine.R
\name{pk.calc.clr}
\alias{pk.calc.clr}
\title{Calculate renal clearance}
\usage{
pk.calc.clr(ae, auc)
}
\arguments{
\item{ae}{The amount excreted in urine (as a numeric scalar or vector)}

\item{auc}{The area under the curve (as a numeric scalar or vector)}
}
\value{
The renal clearance as a number
}
\description{
Calculate renal clearance
}
\details{
clr is \code{sum(ae)/auc}.

The units for the \code{ae} and \code{auc} should match such that \code{ae/auc} has
units of volume/time.
}
\seealso{
\code{\link[=pk.calc.ae]{pk.calc.ae()}}, \code{\link[=pk.calc.fe]{pk.calc.fe()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-general.R, R/dplyr.R
\docType{import}
\name{reexports}
\alias{reexports}
\alias{getGroups}
\alias{inner_join}
\alias{left_join}
\alias{right_join}
\alias{full_join}
\alias{filter}
\alias{group_by}
\alias{ungroup}
\alias{mutate}
\title{Objects exported from other packages}
\keyword{internal}
\description{
These objects are imported from other packages. Follow the links
below to see their documentation.

\describe{
  \item{dplyr}{\code{\link[dplyr]{filter}}, \code{\link[dplyr:mutate-joins]{full_join}}, \code{\link[dplyr]{group_by}}, \code{\link[dplyr:mutate-joins]{inner_join}}, \code{\link[dplyr:mutate-joins]{left_join}}, \code{\link[dplyr]{mutate}}, \code{\link[dplyr:mutate-joins]{right_join}}, \code{\link[dplyr:group_by]{ungroup}}}

  \item{nlme}{\code{\link[nlme]{getGroups}}}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pk.calc.simple.R
\name{pk.calc.tmax}
\alias{pk.calc.tmax}
\title{Determine time of maximum observed PK concentration}
\usage{
pk.calc.tmax(conc, time, options = list(), first.tmax = NULL, check = TRUE)
}
\arguments{
\item{conc}{Measured concentrations}

\item{time}{Time of the measurement of the concentrations}

\item{options}{List of changes to the default PKNCA options (see
\code{PKNCA.options()})}

\item{first.tmax}{If there is more than time that matches the maximum
concentration, should the first be considered as Tmax?  If not, then the
last is considered Tmax.}

\item{check}{Run \code{\link[=assert_conc_time]{assert_conc_time()}}?}
}
\value{
The time of the maximum concentration
}
\description{
Input restrictions are:
\enumerate{
\item the \code{conc} and \code{time} must be the same length,
\item the \code{time} may have no NAs,
}
\code{NA} will be returned if:
\enumerate{
\item the length of \code{conc} and \code{time} is 0
\item all \code{conc} is 0 or \code{NA}
}
}
\examples{
conc_data <- Theoph[Theoph$Subject == 1,]
pk.calc.tmax(conc = conc_data$conc, time = conc_data$Time)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/impute.R
\name{get_impute_method}
\alias{get_impute_method}
\title{Get the impute function from either the intervals column or from the method}
\usage{
get_impute_method(intervals, impute)
}
\arguments{
\item{intervals}{the data.frame of intervals}

\item{impute}{the imputation definition}
}
\value{
The imputation function vector
}
\description{
Get the impute function from either the intervals column or from the method
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/auc.R
\name{pk.calc.auxc}
\alias{pk.calc.auxc}
\alias{pk.calc.auc}
\alias{pk.calc.aumc}
\alias{pk.calc.auc.last}
\alias{pk.calc.auc.inf}
\alias{pk.calc.auc.inf.obs}
\alias{pk.calc.auc.inf.pred}
\alias{pk.calc.auc.all}
\alias{pk.calc.aumc.last}
\alias{pk.calc.aumc.inf}
\alias{pk.calc.aumc.inf.obs}
\alias{pk.calc.aumc.inf.pred}
\alias{pk.calc.aumc.all}
\title{A compute the Area Under the (Moment) Curve}
\usage{
pk.calc.auxc(
  conc,
  time,
  interval = c(0, Inf),
  clast = pk.calc.clast.obs(conc, time, check = FALSE),
  lambda.z = NA,
  auc.type = c("AUClast", "AUCinf", "AUCall"),
  options = list(),
  method = NULL,
  conc.blq = NULL,
  conc.na = NULL,
  check = TRUE,
  fun_linear,
  fun_log,
  fun_inf
)

pk.calc.auc(conc, time, ..., options = list())

pk.calc.auc.last(conc, time, ..., options = list())

pk.calc.auc.inf(conc, time, ..., options = list(), lambda.z)

pk.calc.auc.inf.obs(conc, time, clast.obs, ..., options = list(), lambda.z)

pk.calc.auc.inf.pred(conc, time, clast.pred, ..., options = list(), lambda.z)

pk.calc.auc.all(conc, time, ..., options = list())

pk.calc.aumc(conc, time, ..., options = list())

pk.calc.aumc.last(conc, time, ..., options = list())

pk.calc.aumc.inf(conc, time, ..., options = list(), lambda.z)

pk.calc.aumc.inf.obs(conc, time, clast.obs, ..., options = list(), lambda.z)

pk.calc.aumc.inf.pred(conc, time, clast.pred, ..., options = list(), lambda.z)

pk.calc.aumc.all(conc, time, ..., options = list())
}
\arguments{
\item{conc}{Measured concentrations}

\item{time}{Time of the measurement of the concentrations}

\item{interval}{Numeric vector of two numbers for the start and end time of
integration}

\item{clast, clast.obs, clast.pred}{The last concentration above the limit of
quantification; this is used for AUCinf calculations.  If provided as
clast.obs (observed clast value, default), AUCinf is AUCinf,obs. If
provided as clast.pred, AUCinf is AUCinf,pred.}

\item{lambda.z}{The elimination rate (in units of inverse time) for
extrapolation}

\item{auc.type}{The type of AUC to compute.  Choices are 'AUCinf', 'AUClast',
and 'AUCall'.}

\item{options}{List of changes to the default PKNCA options (see
\code{PKNCA.options()})}

\item{method}{The method for integration (one of 'lin up/log down',
'lin-log', or 'linear')}

\item{conc.blq}{How to handle BLQ values in between the first and last above
LOQ concentrations. (See \code{\link[=clean.conc.blq]{clean.conc.blq()}} for usage instructions.)}

\item{conc.na}{How to handle missing concentration values.  (See
\code{\link[=clean.conc.na]{clean.conc.na()}} for usage instructions.)}

\item{check}{Run \code{\link[=assert_conc_time]{assert_conc_time()}}, \code{\link[=clean.conc.blq]{clean.conc.blq()}}, and
\code{\link[=clean.conc.na]{clean.conc.na()}}?}

\item{fun_linear}{The function to use for integration of the linear part of
the curve (not required for AUC or AUMC functions)}

\item{fun_log}{The function to use for integration of the logarithmic part of
the curve (if log integration is used; not required for AUC or AUMC
functions)}

\item{fun_inf}{The function to use for extrapolation from the final
measurement to infinite time (not required for AUC or AUMC functions.}

\item{...}{For functions other than \code{pk.calc.auxc}, these values are passed
to \code{pk.calc.auxc}}
}
\value{
A numeric value for the AU(M)C.
}
\description{
Compute the area under the curve (AUC) and the area under the moment curve
(AUMC) for pharmacokinetic (PK) data.  AUC and AUMC are used for many
purposes when analyzing PK in drug development.
}
\details{
\code{pk.calc.auc.last} is simply a shortcut setting the \code{interval}
parameter to \code{c(0, "last")}.

Extrapolation beyond Clast occurs using the half-life and Clast,obs;
Clast,pred is not yet supported.

If all conc input are zero, then the AU(M)C is zero.

You probably do not want to call \code{pk.calc.auxc()}.  Usually, you will call
one of the other functions for calculating AUC like \code{pk.calc.auc.last()},
\code{pk.calc.auc.inf.obs()}, etc.
}
\section{Functions}{
\itemize{
\item \code{pk.calc.auc()}: Compute the area under the curve

\item \code{pk.calc.auc.last()}: Compute the AUClast.

\item \code{pk.calc.auc.inf()}: Compute the AUCinf

\item \code{pk.calc.auc.inf.obs()}: Compute the AUCinf with the observed Clast.

\item \code{pk.calc.auc.inf.pred()}: Compute the AUCinf with the predicted Clast.

\item \code{pk.calc.auc.all()}: Compute the AUCall.

\item \code{pk.calc.aumc()}: Compute the area under the moment curve

\item \code{pk.calc.aumc.last()}: Compute the AUMClast.

\item \code{pk.calc.aumc.inf()}: Compute the AUMCinf

\item \code{pk.calc.aumc.inf.obs()}: Compute the AUMCinf with the observed Clast.

\item \code{pk.calc.aumc.inf.pred()}: Compute the AUMCinf with the predicted Clast.

\item \code{pk.calc.aumc.all()}: Compute the AUMCall.

}}
\examples{
myconc <- c(0, 1, 2, 1, 0.5, 0.25, 0)
mytime <- c(0, 1, 2, 3, 4,   5,    6)
pk.calc.auc(myconc, mytime, interval=c(0, 6))
pk.calc.auc(myconc, mytime, interval=c(0, Inf))
}
\references{
Gabrielsson J, Weiner D.  "Section 2.8.1 Computation methods - Linear
trapezoidal rule."  Pharmacokinetic & Pharmacodynamic Data Analysis: Concepts
and Applications, 4th Edition.  Stockholm, Sweden: Swedish Pharmaceutical
Press, 2000.  162-4.

Gabrielsson J, Weiner D.  "Section 2.8.3 Computation methods - Log-linear
trapezoidal rule."  Pharmacokinetic & Pharmacodynamic Data Analysis: Concepts
and Applications, 4th Edition.  Stockholm, Sweden: Swedish Pharmaceutical
Press, 2000.  164-7.
}
\seealso{
\code{\link[=clean.conc.blq]{clean.conc.blq()}}

Other AUC calculations: 
\code{\link{pk.calc.aucint}()}
}
\concept{AUC calculations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/auc_integrate.R
\name{auc_integrate}
\alias{auc_integrate}
\title{Support function for AUC integration}
\usage{
auc_integrate(
  conc,
  time,
  clast,
  tlast,
  lambda.z,
  interval_method,
  fun_linear,
  fun_log,
  fun_inf
)
}
\arguments{
\item{conc}{Measured concentrations}

\item{time}{Time of the measurement of the concentrations}

\item{clast}{The last concentration above the limit of quantification}

\item{tlast}{Time of last concentration above the limit of quantification
(will be calculated, if not provided)}

\item{lambda.z}{The elimination rate (in units of inverse time) for
extrapolation}

\item{interval_method}{The method for integrating each interval of \code{conc}}

\item{fun_linear}{The function to use for integration of the linear part of
the curve (not required for AUC or AUMC functions)}

\item{fun_log}{The function to use for integration of the logarithmic part of
the curve (if log integration is used; not required for AUC or AUMC
functions)}

\item{fun_inf}{The function to use for extrapolation from the final
measurement to infinite time (not required for AUC or AUMC functions.}
}
\description{
Support function for AUC integration
}
\keyword{Internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_and_assert_intervals.R
\name{assert_intervals}
\alias{assert_intervals}
\title{Assert Intervals}
\usage{
assert_intervals(intervals, data)
}
\arguments{
\item{intervals}{Proposed intervals}

\item{data}{PKNCAdata object}
}
\value{
The intervals argument unchanged, or it will raise an error.
}
\description{
Verifies that an interval definition is valid for a PKNCAdata object.
Valid means that intervals are a data.frame (or data.frame-like object),
that the column names are either the groupings of the PKNCAconc part of
the PKNCAdata object or that they are one of the NCA parameters allowed
(i.e. names(get.interval.cols())).
It will return the intervals argument unchanged, or it will raise an error.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sparse.R
\name{var_sparse_auc}
\alias{var_sparse_auc}
\title{Calculate the variance for the AUC of sparsely sampled PK}
\usage{
var_sparse_auc(sparse_pk)
}
\arguments{
\item{sparse_pk}{A sparse_pk object from \code{\link[=as_sparse_pk]{as_sparse_pk()}}}
}
\description{
Equation 7.vii in Nedelman and Jia, 1998 is used for this calculation:
}
\details{
\deqn{var\left(\hat{AUC}\right) = \sum\limits_{i=0}^m\left(\frac{w_i^2 s_i^2}{r_i}\right) + 2\sum\limits_{i<j}\left(\frac{w_i w_j r_{ij} s_{ij}}{r_i r_j}\right)}{var(AUC) = sum_(i=0)^(m) ((w_i^2 * s_i^2)/(r_i) + + 2*sum_(i<j)((w_i * w_j * r_ij * s_ij)/(r_i * r_j))}

The degrees of freedom are calculated as described in equation 6 of the same
paper.
}
\references{
Nedelman JR, Jia X. An extension of Satterthwaite’s approximation applied to
pharmacokinetics. Journal of Biopharmaceutical Statistics. 1998;8(2):317-328.
doi:10.1080/10543409808835241
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-general.R
\name{getAttributeColumn}
\alias{getAttributeColumn}
\title{Retrieve the value of an attribute column.}
\usage{
getAttributeColumn(object, attr_name, warn_missing = c("attr", "column"))
}
\arguments{
\item{object}{The object to extract the attribute value from.}

\item{attr_name}{The name of the attribute to extract}

\item{warn_missing}{Give a warning if the "attr"ibute or "column" is missing.
Character vector with zero, one, or both of "attr" and "column".}
}
\value{
The value of the attribute (or \code{NULL} if the attribute is not set or
the column does not exist)
}
\description{
Retrieve the value of an attribute column.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aucint.R
\name{pk.calc.aucint}
\alias{pk.calc.aucint}
\alias{pk.calc.aucint.last}
\alias{pk.calc.aucint.all}
\alias{pk.calc.aucint.inf.obs}
\alias{pk.calc.aucint.inf.pred}
\title{Calculate the AUC over an interval with interpolation and/or
extrapolation of concentrations for the beginning and end of the
interval.}
\usage{
pk.calc.aucint(
  conc,
  time,
  interval = NULL,
  start = NULL,
  end = NULL,
  clast = pk.calc.clast.obs(conc, time),
  lambda.z = NA,
  time.dose = NULL,
  route = "extravascular",
  duration.dose = 0,
  method = NULL,
  auc.type = "AUClast",
  conc.blq = NULL,
  conc.na = NULL,
  check = TRUE,
  ...,
  options = list()
)

pk.calc.aucint.last(
  conc,
  time,
  start = NULL,
  end = NULL,
  time.dose,
  ...,
  options = list()
)

pk.calc.aucint.all(
  conc,
  time,
  start = NULL,
  end = NULL,
  time.dose,
  ...,
  options = list()
)

pk.calc.aucint.inf.obs(
  conc,
  time,
  start = NULL,
  end = NULL,
  time.dose,
  lambda.z,
  clast.obs,
  ...,
  options = list()
)

pk.calc.aucint.inf.pred(
  conc,
  time,
  start = NULL,
  end = NULL,
  time.dose,
  lambda.z,
  clast.pred,
  ...,
  options = list()
)
}
\arguments{
\item{conc}{Measured concentrations}

\item{time}{Time of the measurement of the concentrations}

\item{interval}{Numeric vector of two numbers for the start and end time of
integration}

\item{start}{The start time of the interval}

\item{end}{The end time of the interval}

\item{clast, clast.obs, clast.pred}{The last concentration above the limit of
quantification; this is used for AUCinf calculations.  If provided as
clast.obs (observed clast value, default), AUCinf is AUCinf,obs. If
provided as clast.pred, AUCinf is AUCinf,pred.}

\item{lambda.z}{The elimination rate (in units of inverse time) for
extrapolation}

\item{time.dose, route, duration.dose}{The time of doses, route of
administration, and duration of dose used with interpolation and
extrapolation of concentration data (see \code{\link[=interp.extrap.conc.dose]{interp.extrap.conc.dose()}}).  If
\code{NULL}, \code{\link[=interp.extrap.conc]{interp.extrap.conc()}} will be used instead (assuming that no doses
affecting concentrations are in the interval).}

\item{method}{The method for integration (one of 'lin up/log down',
'lin-log', or 'linear')}

\item{auc.type}{The type of AUC to compute.  Choices are 'AUCinf', 'AUClast',
and 'AUCall'.}

\item{conc.blq}{How to handle BLQ values in between the first and last above
LOQ concentrations. (See \code{\link[=clean.conc.blq]{clean.conc.blq()}} for usage instructions.)}

\item{conc.na}{How to handle missing concentration values.  (See
\code{\link[=clean.conc.na]{clean.conc.na()}} for usage instructions.)}

\item{check}{Run \code{\link[=assert_conc_time]{assert_conc_time()}}, \code{\link[=clean.conc.blq]{clean.conc.blq()}}, and
\code{\link[=clean.conc.na]{clean.conc.na()}}?}

\item{...}{Additional arguments passed to \code{pk.calc.auxc} and
\code{interp.extrap.conc}}

\item{options}{List of changes to the default PKNCA options (see
\code{PKNCA.options()})}
}
\value{
The AUC for an interval of time as a number
}
\description{
Calculate the AUC over an interval with interpolation and/or
extrapolation of concentrations for the beginning and end of the
interval.
}
\details{
When \code{pk.calc.aucint()} needs to extrapolate using \code{lambda.z} (in other
words, using the half-life), it will always extrapolate using the logarithmic
trapezoidal rule to align with using a half-life calculation for the
extrapolation.
}
\section{Functions}{
\itemize{
\item \code{pk.calc.aucint.last()}: Interpolate or extrapolate concentrations for
AUClast

\item \code{pk.calc.aucint.all()}: Interpolate or extrapolate concentrations for
AUCall

\item \code{pk.calc.aucint.inf.obs()}: Interpolate or extrapolate concentrations for
AUCinf.obs

\item \code{pk.calc.aucint.inf.pred()}: Interpolate or extrapolate concentrations for
AUCinf.pred

}}
\seealso{
\code{\link[=PKNCA.options]{PKNCA.options()}}, \code{\link[=interp.extrap.conc.dose]{interp.extrap.conc.dose()}}

Other AUC calculations: 
\code{\link{pk.calc.auxc}()}
}
\concept{AUC calculations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parse_formula_to_cols.R
\name{parse_formula_to_cols}
\alias{parse_formula_to_cols}
\title{Convert a formula representation to the columns for input data}
\usage{
parse_formula_to_cols(form)
}
\arguments{
\item{form}{the formula (or something coercible into a formula) to extract
into its parts}
}
\value{
A list of column names for various formula parts
}
\description{
Convert a formula representation to the columns for input data
}
\seealso{
Other Formula parsing: 
\code{\link{findOperator}()}
}
\concept{Formula parsing}
\keyword{Internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/provenance.R
\name{checkProvenance}
\alias{checkProvenance}
\title{Check the hash of an object to confirm its provenance.}
\usage{
checkProvenance(object)
}
\arguments{
\item{object}{The object to check provenance for}
}
\value{
\code{TRUE} if the provenance is confirmed to be consistent, \code{FALSE} if
the provenance is not consistent, or \code{NA} if provenance is not present.
}
\description{
Check the hash of an object to confirm its provenance.
}
\seealso{
\code{\link[=addProvenance]{addProvenance()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pk.calc.simple.R
\name{pk.calc.vss}
\alias{pk.calc.vss}
\title{Calculate the steady-state volume of distribution (Vss)}
\usage{
pk.calc.vss(cl, mrt)
}
\arguments{
\item{cl}{the clearance}

\item{mrt}{the mean residence time}
}
\value{
the volume of distribution at steady-state
}
\description{
Calculate the steady-state volume of distribution (Vss)
}
\details{
vss is \code{cl*mrt}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pk.calc.simple.R
\name{pk.calc.cav}
\alias{pk.calc.cav}
\title{Calculate the average concentration during an interval.}
\usage{
pk.calc.cav(auc, start, end)
}
\arguments{
\item{auc}{The area under the curve during the interval}

\item{start}{The start time of the interval}

\item{end}{The end time of the interval}
}
\value{
The Cav (average concentration during the interval)
}
\description{
Calculate the average concentration during an interval.
}
\details{
cav is \code{auc/(end-start)}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pk.calc.simple.R
\name{pk.calc.thalf.eff}
\alias{pk.calc.thalf.eff}
\title{Calculate the effective half-life}
\usage{
pk.calc.thalf.eff(mrt)
}
\arguments{
\item{mrt}{the mean residence time to infinity}
}
\value{
the numeric value of the effective half-life
}
\description{
Calculate the effective half-life
}
\details{
thalf.eff is \code{log(2)*mrt}.
}
